///////////////////////////////////////////////////////////
// This file was auto-generated by cmd/dumpstdlibast.go. //
// https://github.com/google/go-jsonnet#generated-stdlib //
//                                                       //
// --------------- DO NOT EDIT BY HAND! ---------------  //
///////////////////////////////////////////////////////////

package astgen

import (
	"github.com/google/go-jsonnet/ast"
)

var p7Var = "function <anonymous>"
var p7 = &p7Var
var p18Var = "thunk from <function <anonymous>>"
var p18 = &p18Var
var p23Var = "object <anonymous>"
var p23 = &p23Var
var p29Var = "function <anonymous>"
var p29 = &p29Var
var p38Var = "thunk from <function <anonymous>>"
var p38 = &p38Var
var p48Var = "function <anonymous>"
var p48 = &p48Var
var p57Var = "thunk from <function <anonymous>>"
var p57 = &p57Var
var p67Var = "function <anonymous>"
var p67 = &p67Var
var p76Var = "thunk from <function <anonymous>>"
var p76 = &p76Var
var p86Var = "function <anonymous>"
var p86 = &p86Var
var p95Var = "thunk from <function <anonymous>>"
var p95 = &p95Var
var p105Var = "function <anonymous>"
var p105 = &p105Var
var p114Var = "thunk from <function <anonymous>>"
var p114 = &p114Var
var p125Var = "function <anonymous>"
var p125 = &p125Var
var p134Var = "thunk from <function <anonymous>>"
var p134 = &p134Var
var p157Var = "function <anonymous>"
var p157 = &p157Var
var p161Var = "thunk from <function <anonymous>>"
var p161 = &p161Var
var p173Var = "thunk from <function <anonymous>>"
var p173 = &p173Var
var p185Var = "thunk from <function <anonymous>>"
var p185 = &p185Var
var p203Var = "thunk from <function <anonymous>>"
var p203 = &p203Var
var p216Var = "thunk from <thunk from <function <anonymous>>>"
var p216 = &p216Var
var p220Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p220 = &p220Var
var p229Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p229 = &p229Var
var p242Var = "thunk from <thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>>"
var p242 = &p242Var
var p251Var = "function <anonymous>"
var p251 = &p251Var
var p282Var = "thunk from <function <anonymous>>"
var p282 = &p282Var
var p299Var = "thunk from <function <anonymous>>"
var p299 = &p299Var
var p316Var = "thunk from <function <anonymous>>"
var p316 = &p316Var
var p334Var = "function <anonymous>"
var p334 = &p334Var
var p338Var = "thunk from <function <anonymous>>"
var p338 = &p338Var
var p349Var = "thunk from <function <anonymous>>"
var p349 = &p349Var
var p367Var = "thunk from <function <anonymous>>"
var p367 = &p367Var
var p378Var = "thunk from <thunk from <function <anonymous>>>"
var p378 = &p378Var
var p397Var = "function <anonymous>"
var p397 = &p397Var
var p401Var = "thunk from <function <anonymous>>"
var p401 = &p401Var
var p412Var = "thunk from <function <anonymous>>"
var p412 = &p412Var
var p430Var = "thunk from <function <anonymous>>"
var p430 = &p430Var
var p441Var = "thunk from <thunk from <function <anonymous>>>"
var p441 = &p441Var
var p452Var = "thunk from <thunk from <function <anonymous>>>"
var p452 = &p452Var
var p464Var = "thunk from <thunk from <function <anonymous>>>"
var p464 = &p464Var
var p483Var = "function <anonymous>"
var p483 = &p483Var
var p487Var = "thunk from <function <anonymous>>"
var p487 = &p487Var
var p505Var = "thunk from <function <anonymous>>"
var p505 = &p505Var
var p526Var = "thunk from <function <anonymous>>"
var p526 = &p526Var
var p552Var = "thunk <len> from <function <anonymous>>"
var p552 = &p552Var
var p556Var = "thunk from <thunk <len> from <function <anonymous>>>"
var p556 = &p556Var
var p566Var = "function <anonymous>"
var p566 = &p566Var
var p570Var = "thunk from <function <anonymous>>"
var p570 = &p570Var
var p604Var = "thunk from <function <anonymous>>"
var p604 = &p604Var
var p635Var = "function <anonymous>"
var p635 = &p635Var
var p643Var = "thunk from <function <anonymous>>"
var p643 = &p643Var
var p647Var = "thunk from <thunk from <function <anonymous>>>"
var p647 = &p647Var
var p665Var = "function <anonymous>"
var p665 = &p665Var
var p673Var = "thunk from <function <anonymous>>"
var p673 = &p673Var
var p677Var = "thunk from <thunk from <function <anonymous>>>"
var p677 = &p677Var
var p683Var = "function <anonymous>"
var p683 = &p683Var
var p701Var = "function <anonymous>"
var p701 = &p701Var
var p705Var = "thunk from <function <anonymous>>"
var p705 = &p705Var
var p725Var = "thunk from <function <anonymous>>"
var p725 = &p725Var
var p752Var = "thunk from <function <anonymous>>"
var p752 = &p752Var
var p768Var = "thunk from <function <anonymous>>"
var p768 = &p768Var
var p787Var = "thunk from <function <anonymous>>"
var p787 = &p787Var
var p803Var = "thunk from <function <anonymous>>"
var p803 = &p803Var
var p820Var = "function <anonymous>"
var p820 = &p820Var
var p824Var = "thunk from <function <anonymous>>"
var p824 = &p824Var
var p838Var = "thunk from <function <anonymous>>"
var p838 = &p838Var
var p848Var = "thunk from <function <anonymous>>"
var p848 = &p848Var
var p865Var = "thunk from <function <anonymous>>"
var p865 = &p865Var
var p882Var = "function <anonymous>"
var p882 = &p882Var
var p886Var = "thunk from <function <anonymous>>"
var p886 = &p886Var
var p900Var = "thunk from <function <anonymous>>"
var p900 = &p900Var
var p910Var = "thunk from <function <anonymous>>"
var p910 = &p910Var
var p927Var = "thunk from <function <anonymous>>"
var p927 = &p927Var
var p944Var = "function <anonymous>"
var p944 = &p944Var
var p948Var = "thunk from <function <anonymous>>"
var p948 = &p948Var
var p960Var = "thunk from <function <anonymous>>"
var p960 = &p960Var
var p974Var = "thunk from <function <anonymous>>"
var p974 = &p974Var
var p987Var = "thunk from <function <anonymous>>"
var p987 = &p987Var
var p1004Var = "thunk from <function <anonymous>>"
var p1004 = &p1004Var
var p1021Var = "thunk from <function <anonymous>>"
var p1021 = &p1021Var
var p1038Var = "thunk from <function <anonymous>>"
var p1038 = &p1038Var
var p1055Var = "function <anonymous>"
var p1055 = &p1055Var
var p1059Var = "thunk from <function <anonymous>>"
var p1059 = &p1059Var
var p1071Var = "thunk from <function <anonymous>>"
var p1071 = &p1071Var
var p1085Var = "thunk from <function <anonymous>>"
var p1085 = &p1085Var
var p1098Var = "thunk from <function <anonymous>>"
var p1098 = &p1098Var
var p1108Var = "thunk <strLen> from <function <anonymous>>"
var p1108 = &p1108Var
var p1112Var = "thunk from <thunk <strLen> from <function <anonymous>>>"
var p1112 = &p1112Var
var p1122Var = "thunk <cLen> from <function <anonymous>>"
var p1122 = &p1122Var
var p1126Var = "thunk from <thunk <cLen> from <function <anonymous>>>"
var p1126 = &p1126Var
var p1135Var = "function <aux>"
var p1135 = &p1135Var
var p1144Var = "thunk from <function <aux>>"
var p1144 = &p1144Var
var p1165Var = "thunk from <function <aux>>"
var p1165 = &p1165Var
var p1207Var = "thunk from <function <aux>>"
var p1207 = &p1207Var
var p1216Var = "thunk from <thunk from <function <aux>>>"
var p1216 = &p1216Var
var p1231Var = "thunk from <function <aux>>"
var p1231 = &p1231Var
var p1254Var = "thunk <aux> from <function <anonymous>>"
var p1254 = &p1254Var
var p1262Var = "thunk from <function <anonymous>>"
var p1262 = &p1262Var
var p1282Var = "thunk from <function <anonymous>>"
var p1282 = &p1282Var
var p1299Var = "thunk from <function <anonymous>>"
var p1299 = &p1299Var
var p1316Var = "thunk from <function <anonymous>>"
var p1316 = &p1316Var
var p1333Var = "thunk from <function <anonymous>>"
var p1333 = &p1333Var
var p1350Var = "function <anonymous>"
var p1350 = &p1350Var
var p1354Var = "thunk from <function <anonymous>>"
var p1354 = &p1354Var
var p1366Var = "thunk from <function <anonymous>>"
var p1366 = &p1366Var
var p1380Var = "thunk from <function <anonymous>>"
var p1380 = &p1380Var
var p1393Var = "thunk from <function <anonymous>>"
var p1393 = &p1393Var
var p1412Var = "thunk from <function <anonymous>>"
var p1412 = &p1412Var
var p1427Var = "function <revStr>"
var p1427 = &p1427Var
var p1431Var = "thunk from <function <revStr>>"
var p1431 = &p1431Var
var p1444Var = "thunk from <thunk from <function <revStr>>>"
var p1444 = &p1444Var
var p1448Var = "thunk from <thunk from <thunk from <function <revStr>>>>"
var p1448 = &p1448Var
var p1454Var = "thunk <revStr> from <function <anonymous>>"
var p1454 = &p1454Var
var p1467Var = "function <anonymous>"
var p1467 = &p1467Var
var p1471Var = "thunk from <function <anonymous>>"
var p1471 = &p1471Var
var p1475Var = "thunk from <function <anonymous>>"
var p1475 = &p1475Var
var p1489Var = "thunk from <thunk from <function <anonymous>>>"
var p1489 = &p1489Var
var p1494Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p1494 = &p1494Var
var p1498Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p1498 = &p1498Var
var p1506Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p1506 = &p1506Var
var p1529Var = "thunk from <function <anonymous>>"
var p1529 = &p1529Var
var p1546Var = "thunk from <function <anonymous>>"
var p1546 = &p1546Var
var p1563Var = "thunk from <function <anonymous>>"
var p1563 = &p1563Var
var p1580Var = "thunk from <function <anonymous>>"
var p1580 = &p1580Var
var p1597Var = "function <anonymous>"
var p1597 = &p1597Var
var p1601Var = "thunk from <function <anonymous>>"
var p1601 = &p1601Var
var p1613Var = "thunk from <function <anonymous>>"
var p1613 = &p1613Var
var p1625Var = "thunk from <function <anonymous>>"
var p1625 = &p1625Var
var p1641Var = "thunk <str_len> from <function <anonymous>>"
var p1641 = &p1641Var
var p1645Var = "thunk from <thunk <str_len> from <function <anonymous>>>"
var p1645 = &p1645Var
var p1655Var = "thunk <from_len> from <function <anonymous>>"
var p1655 = &p1655Var
var p1659Var = "thunk from <thunk <from_len> from <function <anonymous>>>"
var p1659 = &p1659Var
var p1667Var = "function <found_at>"
var p1667 = &p1667Var
var p1690Var = "thunk <found_at> from <function <anonymous>>"
var p1690 = &p1690Var
var p1698Var = "function <replace_after>"
var p1698 = &p1698Var
var p1729Var = "thunk from <function <replace_after>>"
var p1729 = &p1729Var
var p1740Var = "thunk <new_index> from <function <replace_after>>"
var p1740 = &p1740Var
var p1744Var = "thunk from <thunk <new_index> from <function <replace_after>>>"
var p1744 = &p1744Var
var p1756Var = "thunk from <function <replace_after>>"
var p1756 = &p1756Var
var p1792Var = "thunk from <function <replace_after>>"
var p1792 = &p1792Var
var p1808Var = "thunk <replace_after> from <function <anonymous>>"
var p1808 = &p1808Var
var p1825Var = "thunk from <function <anonymous>>"
var p1825 = &p1825Var
var p1835Var = "thunk from <thunk from <function <anonymous>>>"
var p1835 = &p1835Var
var p1847Var = "thunk from <function <anonymous>>"
var p1847 = &p1847Var
var p1890Var = "thunk <cp> from <function <anonymous>>"
var p1890 = &p1890Var
var p1899Var = "function <up_letter>"
var p1899 = &p1899Var
var p1905Var = "thunk from <function <up_letter>>"
var p1905 = &p1905Var
var p1916Var = "thunk from <function <up_letter>>"
var p1916 = &p1916Var
var p1931Var = "thunk from <function <up_letter>>"
var p1931 = &p1931Var
var p1937Var = "thunk from <thunk from <function <up_letter>>>"
var p1937 = &p1937Var
var p1948Var = "thunk <up_letter> from <function <anonymous>>"
var p1948 = &p1948Var
var p1956Var = "function <anonymous>"
var p1956 = &p1956Var
var p1960Var = "thunk from <function <anonymous>>"
var p1960 = &p1960Var
var p1969Var = "thunk from <thunk from <function <anonymous>>>"
var p1969 = &p1969Var
var p1979Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p1979 = &p1979Var
var p1998Var = "thunk <cp> from <function <anonymous>>"
var p1998 = &p1998Var
var p2007Var = "function <down_letter>"
var p2007 = &p2007Var
var p2013Var = "thunk from <function <down_letter>>"
var p2013 = &p2013Var
var p2024Var = "thunk from <function <down_letter>>"
var p2024 = &p2024Var
var p2039Var = "thunk from <function <down_letter>>"
var p2039 = &p2039Var
var p2045Var = "thunk from <thunk from <function <down_letter>>>"
var p2045 = &p2045Var
var p2056Var = "thunk <down_letter> from <function <anonymous>>"
var p2056 = &p2056Var
var p2064Var = "function <anonymous>"
var p2064 = &p2064Var
var p2068Var = "thunk from <function <anonymous>>"
var p2068 = &p2068Var
var p2077Var = "thunk from <thunk from <function <anonymous>>>"
var p2077 = &p2077Var
var p2087Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p2087 = &p2087Var
var p2106Var = "function <anonymous>"
var p2106 = &p2106Var
var p2111Var = "thunk from <function <anonymous>>"
var p2111 = &p2111Var
var p2122Var = "function <anonymous>"
var p2122 = &p2122Var
var p2142Var = "thunk <joiner> from <function <anonymous>>"
var p2142 = &p2142Var
var p2146Var = "thunk from <thunk <joiner> from <function <anonymous>>>"
var p2146 = &p2146Var
var p2159Var = "thunk from <thunk <joiner> from <function <anonymous>>>"
var p2159 = &p2159Var
var p2176Var = "function <anonymous>"
var p2176 = &p2176Var
var p2180Var = "thunk from <function <anonymous>>"
var p2180 = &p2180Var
var p2190Var = "thunk from <thunk from <function <anonymous>>>"
var p2190 = &p2190Var
var p2194Var = "function <anonymous>"
var p2194 = &p2194Var
var p2212Var = "object <invar>"
var p2212 = &p2212Var
var p2242Var = "thunk from <object <invar>>"
var p2242 = &p2242Var
var p2272Var = "thunk from <object <invar>>"
var p2272 = &p2272Var
var p2284Var = "thunk from <object <invar>>"
var p2284 = &p2284Var
var p2289Var = "thunk <invar> from <function <anonymous>>"
var p2289 = &p2289Var
var p2295Var = "function <anonymous>"
var p2295 = &p2295Var
var p2337Var = "thunk from <function <anonymous>>"
var p2337 = &p2337Var
var p2348Var = "thunk from <function <anonymous>>"
var p2348 = &p2348Var
var p2362Var = "function <build>"
var p2362 = &p2362Var
var p2388Var = "thunk from <function <build>>"
var p2388 = &p2388Var
var p2417Var = "thunk from <thunk from <function <build>>>"
var p2417 = &p2417Var
var p2446Var = "thunk <build> from <function <anonymous>>"
var p2446 = &p2446Var
var p2456Var = "thunk from <function <anonymous>>"
var p2456 = &p2456Var
var p2491Var = "thunk from <function <anonymous>>"
var p2491 = &p2491Var
var p2526Var = "thunk from <function <anonymous>>"
var p2526 = &p2526Var
var p2554Var = "function <anonymous>"
var p2554 = &p2554Var
var p2558Var = "thunk from <function <anonymous>>"
var p2558 = &p2558Var
var p2572Var = "thunk from <function <anonymous>>"
var p2572 = &p2572Var
var p2587Var = "thunk from <function <anonymous>>"
var p2587 = &p2587Var
var p2605Var = "thunk from <function <anonymous>>"
var p2605 = &p2605Var
var p2609Var = "thunk from <thunk from <function <anonymous>>>"
var p2609 = &p2609Var
var p2632Var = "function <anonymous>"
var p2632 = &p2632Var
var p2640Var = "thunk from <function <anonymous>>"
var p2640 = &p2640Var
var p2646Var = "function <anonymous>"
var p2646 = &p2646Var
var p2652Var = "thunk from <thunk from <function <anonymous>>>"
var p2652 = &p2652Var
var p2669Var = "function <anonymous>"
var p2669 = &p2669Var
var p2673Var = "thunk from <function <anonymous>>"
var p2673 = &p2673Var
var p2684Var = "thunk from <function <anonymous>>"
var p2684 = &p2684Var
var p2697Var = "thunk from <function <anonymous>>"
var p2697 = &p2697Var
var p2711Var = "thunk from <function <anonymous>>"
var p2711 = &p2711Var
var p2723Var = "thunk from <function <anonymous>>"
var p2723 = &p2723Var
var p2740Var = "thunk from <function <anonymous>>"
var p2740 = &p2740Var
var p2754Var = "thunk from <function <anonymous>>"
var p2754 = &p2754Var
var p2780Var = "function <anonymous>"
var p2780 = &p2780Var
var p2784Var = "thunk from <function <anonymous>>"
var p2784 = &p2784Var
var p2798Var = "thunk from <function <anonymous>>"
var p2798 = &p2798Var
var p2816Var = "thunk from <function <anonymous>>"
var p2816 = &p2816Var
var p2829Var = "thunk from <function <anonymous>>"
var p2829 = &p2829Var
var p2844Var = "thunk from <function <anonymous>>"
var p2844 = &p2844Var
var p2864Var = "thunk from <function <anonymous>>"
var p2864 = &p2864Var
var p2868Var = "thunk from <thunk from <function <anonymous>>>"
var p2868 = &p2868Var
var p2874Var = "function <anonymous>"
var p2874 = &p2874Var
var p2879Var = "thunk from <function <anonymous>>"
var p2879 = &p2879Var
var p2904Var = "function <anonymous>"
var p2904 = &p2904Var
var p2908Var = "thunk from <function <anonymous>>"
var p2908 = &p2908Var
var p2922Var = "thunk from <function <anonymous>>"
var p2922 = &p2922Var
var p2940Var = "thunk from <function <anonymous>>"
var p2940 = &p2940Var
var p2953Var = "thunk from <function <anonymous>>"
var p2953 = &p2953Var
var p2968Var = "thunk from <function <anonymous>>"
var p2968 = &p2968Var
var p2988Var = "thunk from <function <anonymous>>"
var p2988 = &p2988Var
var p2992Var = "thunk from <thunk from <function <anonymous>>>"
var p2992 = &p2992Var
var p2998Var = "function <anonymous>"
var p2998 = &p2998Var
var p3002Var = "thunk from <function <anonymous>>"
var p3002 = &p3002Var
var p3030Var = "function <anonymous>"
var p3030 = &p3030Var
var p3034Var = "thunk from <function <anonymous>>"
var p3034 = &p3034Var
var p3048Var = "thunk from <function <anonymous>>"
var p3048 = &p3048Var
var p3065Var = "thunk from <function <anonymous>>"
var p3065 = &p3065Var
var p3079Var = "thunk from <function <anonymous>>"
var p3079 = &p3079Var
var p3109Var = "object <anonymous>"
var p3109 = &p3109Var
var p3113Var = "thunk from <object <anonymous>>"
var p3113 = &p3113Var
var p3135Var = "thunk from <function <anonymous>>"
var p3135 = &p3135Var
var p3156Var = "function <anonymous>"
var p3156 = &p3156Var
var p3160Var = "thunk from <function <anonymous>>"
var p3160 = &p3160Var
var p3174Var = "thunk from <function <anonymous>>"
var p3174 = &p3174Var
var p3190Var = "thunk from <function <anonymous>>"
var p3190 = &p3190Var
var p3206Var = "thunk from <function <anonymous>>"
var p3206 = &p3206Var
var p3214Var = "thunk from <thunk from <function <anonymous>>>"
var p3214 = &p3214Var
var p3218Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p3218 = &p3218Var
var p3224Var = "function <anonymous>"
var p3224 = &p3224Var
var p3229Var = "thunk from <function <anonymous>>"
var p3229 = &p3229Var
var p3248Var = "thunk from <function <anonymous>>"
var p3248 = &p3248Var
var p3260Var = "thunk from <function <anonymous>>"
var p3260 = &p3260Var
var p3273Var = "thunk from <thunk from <function <anonymous>>>"
var p3273 = &p3273Var
var p3277Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p3277 = &p3277Var
var p3283Var = "function <anonymous>"
var p3283 = &p3283Var
var p3288Var = "thunk from <function <anonymous>>"
var p3288 = &p3288Var
var p3308Var = "thunk from <function <anonymous>>"
var p3308 = &p3308Var
var p3335Var = "function <aux>"
var p3335 = &p3335Var
var p3339Var = "thunk from <function <aux>>"
var p3339 = &p3339Var
var p3364Var = "thunk from <function <aux>>"
var p3364 = &p3364Var
var p3386Var = "thunk from <function <aux>>"
var p3386 = &p3386Var
var p3398Var = "thunk from <function <aux>>"
var p3398 = &p3398Var
var p3421Var = "thunk from <function <aux>>"
var p3421 = &p3421Var
var p3425Var = "thunk from <thunk from <function <aux>>>"
var p3425 = &p3425Var
var p3439Var = "thunk from <thunk from <function <aux>>>"
var p3439 = &p3439Var
var p3458Var = "thunk from <function <aux>>"
var p3458 = &p3458Var
var p3483Var = "thunk from <function <aux>>"
var p3483 = &p3483Var
var p3516Var = "thunk <aux> from <function <anonymous>>"
var p3516 = &p3516Var
var p3525Var = "function <anonymous>"
var p3525 = &p3525Var
var p3529Var = "thunk from <function <anonymous>>"
var p3529 = &p3529Var
var p3543Var = "thunk from <function <anonymous>>"
var p3543 = &p3543Var
var p3559Var = "thunk from <function <anonymous>>"
var p3559 = &p3559Var
var p3568Var = "thunk from <function <anonymous>>"
var p3568 = &p3568Var
var p3583Var = "thunk from <function <anonymous>>"
var p3583 = &p3583Var
var p3592Var = "thunk from <function <anonymous>>"
var p3592 = &p3592Var
var p3608Var = "thunk from <function <anonymous>>"
var p3608 = &p3608Var
var p3634Var = "function <anonymous>"
var p3634 = &p3634Var
var p3638Var = "thunk from <function <anonymous>>"
var p3638 = &p3638Var
var p3643Var = "thunk from <thunk from <function <anonymous>>>"
var p3643 = &p3643Var
var p3658Var = "function <anonymous>"
var p3658 = &p3658Var
var p3662Var = "thunk from <function <anonymous>>"
var p3662 = &p3662Var
var p3677Var = "thunk from <function <anonymous>>"
var p3677 = &p3677Var
var p3689Var = "thunk from <function <anonymous>>"
var p3689 = &p3689Var
var p3705Var = "thunk from <thunk from <function <anonymous>>>"
var p3705 = &p3705Var
var p3709Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p3709 = &p3709Var
var p3736Var = "thunk from <function <anonymous>>"
var p3736 = &p3736Var
var p3761Var = "function <try_parse_mapping_key>"
var p3761 = &p3761Var
var p3765Var = "thunk from <function <try_parse_mapping_key>>"
var p3765 = &p3765Var
var p3775Var = "thunk <c> from <function <try_parse_mapping_key>>"
var p3775 = &p3775Var
var p3796Var = "function <consume>"
var p3796 = &p3796Var
var p3800Var = "thunk from <function <consume>>"
var p3800 = &p3800Var
var p3813Var = "thunk <c> from <function <consume>>"
var p3813 = &p3813Var
var p3830Var = "thunk from <function <consume>>"
var p3830 = &p3830Var
var p3849Var = "object <anonymous>"
var p3849 = &p3849Var
var p3867Var = "thunk <consume> from <function <try_parse_mapping_key>>"
var p3867 = &p3867Var
var p3875Var = "thunk from <function <try_parse_mapping_key>>"
var p3875 = &p3875Var
var p3890Var = "object <anonymous>"
var p3890 = &p3890Var
var p3905Var = "thunk <try_parse_mapping_key> from <function <anonymous>>"
var p3905 = &p3905Var
var p3920Var = "function <consume>"
var p3920 = &p3920Var
var p3924Var = "thunk from <function <consume>>"
var p3924 = &p3924Var
var p3934Var = "thunk <c> from <function <consume>>"
var p3934 = &p3934Var
var p3951Var = "thunk from <function <consume>>"
var p3951 = &p3951Var
var p3963Var = "object <anonymous>"
var p3963 = &p3963Var
var p3980Var = "thunk from <function <consume>>"
var p3980 = &p3980Var
var p3992Var = "object <anonymous>"
var p3992 = &p3992Var
var p4009Var = "thunk from <function <consume>>"
var p4009 = &p4009Var
var p4021Var = "object <anonymous>"
var p4021 = &p4021Var
var p4038Var = "thunk from <function <consume>>"
var p4038 = &p4038Var
var p4050Var = "object <anonymous>"
var p4050 = &p4050Var
var p4067Var = "thunk from <function <consume>>"
var p4067 = &p4067Var
var p4079Var = "object <anonymous>"
var p4079 = &p4079Var
var p4088Var = "object <anonymous>"
var p4088 = &p4088Var
var p4112Var = "thunk <consume> from <function <try_parse_cflags>>"
var p4112 = &p4112Var
var p4117Var = "function <try_parse_cflags>"
var p4117 = &p4117Var
var p4121Var = "thunk from <function <try_parse_cflags>>"
var p4121 = &p4121Var
var p4129Var = "object <anonymous>"
var p4129 = &p4129Var
var p4142Var = "thunk <try_parse_cflags> from <function <anonymous>>"
var p4142 = &p4142Var
var p4151Var = "function <try_parse_field_width>"
var p4151 = &p4151Var
var p4168Var = "thunk from <function <try_parse_field_width>>"
var p4168 = &p4168Var
var p4180Var = "object <anonymous>"
var p4180 = &p4180Var
var p4198Var = "function <consume>"
var p4198 = &p4198Var
var p4202Var = "thunk from <function <consume>>"
var p4202 = &p4202Var
var p4212Var = "thunk <c> from <function <consume>>"
var p4212 = &p4212Var
var p4229Var = "thunk from <function <consume>>"
var p4229 = &p4229Var
var p4257Var = "thunk from <function <consume>>"
var p4257 = &p4257Var
var p4285Var = "thunk from <function <consume>>"
var p4285 = &p4285Var
var p4313Var = "thunk from <function <consume>>"
var p4313 = &p4313Var
var p4341Var = "thunk from <function <consume>>"
var p4341 = &p4341Var
var p4369Var = "thunk from <function <consume>>"
var p4369 = &p4369Var
var p4397Var = "thunk from <function <consume>>"
var p4397 = &p4397Var
var p4425Var = "thunk from <function <consume>>"
var p4425 = &p4425Var
var p4453Var = "thunk from <function <consume>>"
var p4453 = &p4453Var
var p4481Var = "thunk from <function <consume>>"
var p4481 = &p4481Var
var p4501Var = "object <anonymous>"
var p4501 = &p4501Var
var p4535Var = "thunk <consume> from <function <try_parse_field_width>>"
var p4535 = &p4535Var
var p4543Var = "thunk from <function <try_parse_field_width>>"
var p4543 = &p4543Var
var p4555Var = "thunk <try_parse_field_width> from <function <anonymous>>"
var p4555 = &p4555Var
var p4567Var = "function <try_parse_precision>"
var p4567 = &p4567Var
var p4571Var = "thunk from <function <try_parse_precision>>"
var p4571 = &p4571Var
var p4581Var = "thunk <c> from <function <try_parse_precision>>"
var p4581 = &p4581Var
var p4598Var = "thunk from <function <try_parse_precision>>"
var p4598 = &p4598Var
var p4610Var = "object <anonymous>"
var p4610 = &p4610Var
var p4625Var = "thunk <try_parse_precision> from <function <anonymous>>"
var p4625 = &p4625Var
var p4637Var = "function <try_parse_length_modifier>"
var p4637 = &p4637Var
var p4641Var = "thunk from <function <try_parse_length_modifier>>"
var p4641 = &p4641Var
var p4651Var = "thunk <c> from <function <try_parse_length_modifier>>"
var p4651 = &p4651Var
var p4694Var = "thunk <try_parse_length_modifier> from <function <anonymous>>"
var p4694 = &p4694Var
var p4706Var = "function <parse_conv_type>"
var p4706 = &p4706Var
var p4710Var = "thunk from <function <parse_conv_type>>"
var p4710 = &p4710Var
var p4720Var = "thunk <c> from <function <parse_conv_type>>"
var p4720 = &p4720Var
var p4750Var = "object <anonymous>"
var p4750 = &p4750Var
var p4771Var = "object <anonymous>"
var p4771 = &p4771Var
var p4792Var = "object <anonymous>"
var p4792 = &p4792Var
var p4813Var = "object <anonymous>"
var p4813 = &p4813Var
var p4834Var = "object <anonymous>"
var p4834 = &p4834Var
var p4855Var = "object <anonymous>"
var p4855 = &p4855Var
var p4876Var = "object <anonymous>"
var p4876 = &p4876Var
var p4897Var = "object <anonymous>"
var p4897 = &p4897Var
var p4918Var = "object <anonymous>"
var p4918 = &p4918Var
var p4939Var = "object <anonymous>"
var p4939 = &p4939Var
var p4960Var = "object <anonymous>"
var p4960 = &p4960Var
var p4981Var = "object <anonymous>"
var p4981 = &p4981Var
var p5002Var = "object <anonymous>"
var p5002 = &p5002Var
var p5053Var = "thunk <parse_conv_type> from <function <anonymous>>"
var p5053 = &p5053Var
var p5065Var = "function <parse_code>"
var p5065 = &p5065Var
var p5069Var = "thunk from <function <parse_code>>"
var p5069 = &p5069Var
var p5079Var = "thunk <mkey> from <function <parse_code>>"
var p5079 = &p5079Var
var p5083Var = "thunk from <thunk <mkey> from <function <parse_code>>>"
var p5083 = &p5083Var
var p5092Var = "thunk <cflags> from <function <parse_code>>"
var p5092 = &p5092Var
var p5096Var = "thunk from <thunk <cflags> from <function <parse_code>>>"
var p5096 = &p5096Var
var p5108Var = "thunk <fw> from <function <parse_code>>"
var p5108 = &p5108Var
var p5112Var = "thunk from <thunk <fw> from <function <parse_code>>>"
var p5112 = &p5112Var
var p5124Var = "thunk <prec> from <function <parse_code>>"
var p5124 = &p5124Var
var p5128Var = "thunk from <thunk <prec> from <function <parse_code>>>"
var p5128 = &p5128Var
var p5140Var = "thunk <len_mod> from <function <parse_code>>"
var p5140 = &p5140Var
var p5144Var = "thunk from <thunk <len_mod> from <function <parse_code>>>"
var p5144 = &p5144Var
var p5156Var = "thunk <ctype> from <function <parse_code>>"
var p5156 = &p5156Var
var p5160Var = "thunk from <thunk <ctype> from <function <parse_code>>>"
var p5160 = &p5160Var
var p5172Var = "object <anonymous>"
var p5172 = &p5172Var
var p5182Var = "object <anonymous>"
var p5182 = &p5182Var
var p5233Var = "thunk <parse_code> from <function <anonymous>>"
var p5233 = &p5233Var
var p5245Var = "function <parse_codes>"
var p5245 = &p5245Var
var p5249Var = "thunk from <function <parse_codes>>"
var p5249 = &p5249Var
var p5259Var = "thunk from <function <parse_codes>>"
var p5259 = &p5259Var
var p5270Var = "thunk <c> from <function <parse_codes>>"
var p5270 = &p5270Var
var p5285Var = "thunk <r> from <function <parse_codes>>"
var p5285 = &p5285Var
var p5289Var = "thunk from <thunk <r> from <function <parse_codes>>>"
var p5289 = &p5289Var
var p5303Var = "thunk from <function <parse_codes>>"
var p5303 = &p5303Var
var p5314Var = "thunk from <thunk from <function <parse_codes>>>"
var p5314 = &p5314Var
var p5335Var = "thunk from <function <parse_codes>>"
var p5335 = &p5335Var
var p5360Var = "thunk <parse_codes> from <function <anonymous>>"
var p5360 = &p5360Var
var p5366Var = "thunk <codes> from <function <anonymous>>"
var p5366 = &p5366Var
var p5370Var = "thunk from <thunk <codes> from <function <anonymous>>>"
var p5370 = &p5370Var
var p5385Var = "function <aux>"
var p5385 = &p5385Var
var p5399Var = "thunk from <function <aux>>"
var p5399 = &p5399Var
var p5414Var = "thunk <aux> from <function <padding>>"
var p5414 = &p5414Var
var p5419Var = "function <padding>"
var p5419 = &p5419Var
var p5423Var = "thunk from <function <padding>>"
var p5423 = &p5423Var
var p5430Var = "thunk <padding> from <function <anonymous>>"
var p5430 = &p5430Var
var p5436Var = "function <pad_left>"
var p5436 = &p5436Var
var p5449Var = "thunk from <function <pad_left>>"
var p5449 = &p5449Var
var p5453Var = "thunk from <thunk from <function <pad_left>>>"
var p5453 = &p5453Var
var p5464Var = "thunk <pad_left> from <function <anonymous>>"
var p5464 = &p5464Var
var p5472Var = "function <pad_right>"
var p5472 = &p5472Var
var p5481Var = "thunk from <function <pad_right>>"
var p5481 = &p5481Var
var p5485Var = "thunk from <thunk from <function <pad_right>>>"
var p5485 = &p5485Var
var p5499Var = "thunk <pad_right> from <function <anonymous>>"
var p5499 = &p5499Var
var p5509Var = "thunk <dec> from <function <render_int>>"
var p5509 = &p5509Var
var p5521Var = "function <aux>"
var p5521 = &p5521Var
var p5551Var = "thunk from <function <aux>>"
var p5551 = &p5551Var
var p5556Var = "thunk from <thunk from <function <aux>>>"
var p5556 = &p5556Var
var p5568Var = "thunk <aux> from <thunk <dec> from <function <render_int>>>"
var p5568 = &p5568Var
var p5576Var = "thunk from <thunk <dec> from <function <render_int>>>"
var p5576 = &p5576Var
var p5590Var = "thunk <zp> from <function <render_int>>"
var p5590 = &p5590Var
var p5612Var = "thunk <zp2> from <function <render_int>>"
var p5612 = &p5612Var
var p5616Var = "thunk from <thunk <zp2> from <function <render_int>>>"
var p5616 = &p5616Var
var p5625Var = "thunk <dec2> from <function <render_int>>"
var p5625 = &p5625Var
var p5629Var = "thunk from <thunk <dec2> from <function <render_int>>>"
var p5629 = &p5629Var
var p5637Var = "function <render_int>"
var p5637 = &p5637Var
var p5666Var = "thunk <render_int> from <function <anonymous>>"
var p5666 = &p5666Var
var p5676Var = "thunk <numerals> from <function <render_hex>>"
var p5676 = &p5676Var
var p5681Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5681 = &p5681Var
var p5690Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5690 = &p5690Var
var p5701Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5701 = &p5701Var
var p5720Var = "thunk <n_> from <function <render_hex>>"
var p5720 = &p5720Var
var p5724Var = "thunk from <thunk <n_> from <function <render_hex>>>"
var p5724 = &p5724Var
var p5733Var = "function <aux>"
var p5733 = &p5733Var
var p5765Var = "thunk from <function <aux>>"
var p5765 = &p5765Var
var p5770Var = "thunk from <thunk from <function <aux>>>"
var p5770 = &p5770Var
var p5781Var = "thunk <aux> from <function <render_hex>>"
var p5781 = &p5781Var
var p5788Var = "thunk <hex> from <function <render_hex>>"
var p5788 = &p5788Var
var p5797Var = "thunk from <thunk <hex> from <function <render_hex>>>"
var p5797 = &p5797Var
var p5811Var = "thunk from <thunk <hex> from <function <render_hex>>>"
var p5811 = &p5811Var
var p5815Var = "thunk from <thunk from <thunk <hex> from <function <render_hex>>>>"
var p5815 = &p5815Var
var p5824Var = "thunk <neg> from <function <render_hex>>"
var p5824 = &p5824Var
var p5833Var = "thunk <zp> from <function <render_hex>>"
var p5833 = &p5833Var
var p5865Var = "thunk <zp2> from <function <render_hex>>"
var p5865 = &p5865Var
var p5869Var = "thunk from <thunk <zp2> from <function <render_hex>>>"
var p5869 = &p5869Var
var p5879Var = "thunk <hex2> from <function <render_hex>>"
var p5879 = &p5879Var
var p5883Var = "thunk from <thunk <hex2> from <function <render_hex>>>"
var p5883 = &p5883Var
var p5904Var = "function <render_hex>"
var p5904 = &p5904Var
var p5940Var = "thunk <render_hex> from <function <anonymous>>"
var p5940 = &p5940Var
var p5951Var = "function <aux>"
var p5951 = &p5951Var
var p5973Var = "thunk from <function <aux>>"
var p5973 = &p5973Var
var p5990Var = "thunk from <function <aux>>"
var p5990 = &p5990Var
var p6006Var = "thunk <aux> from <function <strip_trailing_zero>>"
var p6006 = &p6006Var
var p6011Var = "function <strip_trailing_zero>"
var p6011 = &p6011Var
var p6015Var = "thunk from <function <strip_trailing_zero>>"
var p6015 = &p6015Var
var p6027Var = "thunk from <thunk from <function <strip_trailing_zero>>>"
var p6027 = &p6027Var
var p6035Var = "thunk <strip_trailing_zero> from <function <anonymous>>"
var p6035 = &p6035Var
var p6047Var = "thunk <n_> from <function <render_float_dec>>"
var p6047 = &p6047Var
var p6051Var = "thunk from <thunk <n_> from <function <render_float_dec>>>"
var p6051 = &p6051Var
var p6061Var = "thunk <whole> from <function <render_float_dec>>"
var p6061 = &p6061Var
var p6065Var = "thunk from <thunk <whole> from <function <render_float_dec>>>"
var p6065 = &p6065Var
var p6075Var = "thunk <denominator> from <function <render_float_dec>>"
var p6075 = &p6075Var
var p6079Var = "thunk from <thunk <denominator> from <function <render_float_dec>>>"
var p6079 = &p6079Var
var p6087Var = "thunk <numerator> from <function <render_float_dec>>"
var p6087 = &p6087Var
var p6099Var = "thunk from <thunk <numerator> from <function <render_float_dec>>>"
var p6099 = &p6099Var
var p6112Var = "thunk <whole> from <function <render_float_dec>>"
var p6112 = &p6112Var
var p6117Var = "thunk from <thunk <whole> from <function <render_float_dec>>>"
var p6117 = &p6117Var
var p6131Var = "thunk from <thunk <whole> from <function <render_float_dec>>>"
var p6131 = &p6131Var
var p6149Var = "thunk <frac> from <function <render_float_dec>>"
var p6149 = &p6149Var
var p6153Var = "thunk from <thunk <frac> from <function <render_float_dec>>>"
var p6153 = &p6153Var
var p6165Var = "thunk <dot_size> from <function <render_float_dec>>"
var p6165 = &p6165Var
var p6181Var = "thunk <zp> from <function <render_float_dec>>"
var p6181 = &p6181Var
var p6194Var = "thunk <str> from <function <render_float_dec>>"
var p6194 = &p6194Var
var p6199Var = "thunk from <thunk <str> from <function <render_float_dec>>>"
var p6199 = &p6199Var
var p6218Var = "function <render_float_dec>"
var p6218 = &p6218Var
var p6247Var = "thunk <frac_str> from <function <render_float_dec>>"
var p6247 = &p6247Var
var p6251Var = "thunk from <thunk <frac_str> from <function <render_float_dec>>>"
var p6251 = &p6251Var
var p6273Var = "thunk from <function <render_float_dec>>"
var p6273 = &p6273Var
var p6320Var = "thunk <render_float_dec> from <function <anonymous>>"
var p6320 = &p6320Var
var p6330Var = "thunk <exponent> from <function <render_float_sci>>"
var p6330 = &p6330Var
var p6348Var = "thunk from <thunk <exponent> from <function <render_float_sci>>>"
var p6348 = &p6348Var
var p6352Var = "thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>"
var p6352 = &p6352Var
var p6366Var = "thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>"
var p6366 = &p6366Var
var p6370Var = "thunk from <thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>>"
var p6370 = &p6370Var
var p6382Var = "thunk <suff> from <function <render_float_sci>>"
var p6382 = &p6382Var
var p6387Var = "thunk from <thunk <suff> from <function <render_float_sci>>>"
var p6387 = &p6387Var
var p6399Var = "thunk from <thunk from <thunk <suff> from <function <render_float_sci>>>>"
var p6399 = &p6399Var
var p6423Var = "thunk <mantissa> from <function <render_float_sci>>"
var p6423 = &p6423Var
var p6436Var = "thunk from <thunk <mantissa> from <function <render_float_sci>>>"
var p6436 = &p6436Var
var p6461Var = "thunk from <thunk <mantissa> from <function <render_float_sci>>>"
var p6461 = &p6461Var
var p6479Var = "thunk <zp2> from <function <render_float_sci>>"
var p6479 = &p6479Var
var p6483Var = "thunk from <thunk <zp2> from <function <render_float_sci>>>"
var p6483 = &p6483Var
var p6491Var = "function <render_float_sci>"
var p6491 = &p6491Var
var p6499Var = "thunk from <function <render_float_sci>>"
var p6499 = &p6499Var
var p6524Var = "thunk <render_float_sci> from <function <anonymous>>"
var p6524 = &p6524Var
var p6535Var = "thunk <cflags> from <function <format_code>>"
var p6535 = &p6535Var
var p6542Var = "thunk <fpprec> from <function <format_code>>"
var p6542 = &p6542Var
var p6555Var = "thunk <iprec> from <function <format_code>>"
var p6555 = &p6555Var
var p6572Var = "thunk <zp> from <function <format_code>>"
var p6572 = &p6572Var
var p6588Var = "function <format_code>"
var p6588 = &p6588Var
var p6604Var = "thunk from <function <format_code>>"
var p6604 = &p6604Var
var p6627Var = "thunk from <function <format_code>>"
var p6627 = &p6627Var
var p6641Var = "thunk from <function <format_code>>"
var p6641 = &p6641Var
var p6664Var = "thunk from <function <format_code>>"
var p6664 = &p6664Var
var p6680Var = "thunk from <thunk from <function <format_code>>>"
var p6680 = &p6680Var
var p6684Var = "thunk from <thunk from <thunk from <function <format_code>>>>"
var p6684 = &p6684Var
var p6728Var = "thunk from <function <format_code>>"
var p6728 = &p6728Var
var p6742Var = "thunk from <function <format_code>>"
var p6742 = &p6742Var
var p6764Var = "thunk <zero_prefix> from <function <format_code>>"
var p6764 = &p6764Var
var p6777Var = "thunk from <function <format_code>>"
var p6777 = &p6777Var
var p6793Var = "thunk from <thunk from <function <format_code>>>"
var p6793 = &p6793Var
var p6797Var = "thunk from <thunk from <thunk from <function <format_code>>>>"
var p6797 = &p6797Var
var p6844Var = "thunk from <function <format_code>>"
var p6844 = &p6844Var
var p6858Var = "thunk from <function <format_code>>"
var p6858 = &p6858Var
var p6883Var = "thunk from <function <format_code>>"
var p6883 = &p6883Var
var p6887Var = "thunk from <thunk from <function <format_code>>>"
var p6887 = &p6887Var
var p6944Var = "thunk from <function <format_code>>"
var p6944 = &p6944Var
var p6958Var = "thunk from <function <format_code>>"
var p6958 = &p6958Var
var p6979Var = "thunk from <function <format_code>>"
var p6979 = &p6979Var
var p7031Var = "thunk from <function <format_code>>"
var p7031 = &p7031Var
var p7045Var = "thunk from <function <format_code>>"
var p7045 = &p7045Var
var p7066Var = "thunk from <function <format_code>>"
var p7066 = &p7066Var
var p7124Var = "thunk from <function <format_code>>"
var p7124 = &p7124Var
var p7138Var = "thunk from <function <format_code>>"
var p7138 = &p7138Var
var p7160Var = "thunk <exponent> from <function <format_code>>"
var p7160 = &p7160Var
var p7169Var = "thunk from <thunk <exponent> from <function <format_code>>>"
var p7169 = &p7169Var
var p7173Var = "thunk from <thunk from <thunk <exponent> from <function <format_code>>>>"
var p7173 = &p7173Var
var p7187Var = "thunk from <thunk from <thunk <exponent> from <function <format_code>>>>"
var p7187 = &p7187Var
var p7191Var = "thunk from <thunk from <thunk from <thunk <exponent> from <function <format_code>>>>>"
var p7191 = &p7191Var
var p7218Var = "thunk from <function <format_code>>"
var p7218 = &p7218Var
var p7267Var = "thunk <digits_before_pt> from <function <format_code>>"
var p7267 = &p7267Var
var p7271Var = "thunk from <thunk <digits_before_pt> from <function <format_code>>>"
var p7271 = &p7271Var
var p7284Var = "thunk from <function <format_code>>"
var p7284 = &p7284Var
var p7351Var = "thunk from <function <format_code>>"
var p7351 = &p7351Var
var p7364Var = "thunk from <function <format_code>>"
var p7364 = &p7364Var
var p7378Var = "thunk from <function <format_code>>"
var p7378 = &p7378Var
var p7393Var = "thunk from <function <format_code>>"
var p7393 = &p7393Var
var p7410Var = "thunk from <function <format_code>>"
var p7410 = &p7410Var
var p7430Var = "thunk from <function <format_code>>"
var p7430 = &p7430Var
var p7479Var = "thunk <format_code> from <function <anonymous>>"
var p7479 = &p7479Var
var p7491Var = "function <format_codes_arr>"
var p7491 = &p7491Var
var p7495Var = "thunk from <function <format_codes_arr>>"
var p7495 = &p7495Var
var p7511Var = "thunk from <function <format_codes_arr>>"
var p7511 = &p7511Var
var p7532Var = "thunk from <function <format_codes_arr>>"
var p7532 = &p7532Var
var p7551Var = "thunk <code> from <function <format_codes_arr>>"
var p7551 = &p7551Var
var p7567Var = "thunk from <function <format_codes_arr>>"
var p7567 = &p7567Var
var p7577Var = "thunk from <function <format_codes_arr>>"
var p7577 = &p7577Var
var p7600Var = "thunk <tmp> from <function <format_codes_arr>>"
var p7600 = &p7600Var
var p7612Var = "object <anonymous>"
var p7612 = &p7612Var
var p7627Var = "thunk from <object <anonymous>>"
var p7627 = &p7627Var
var p7648Var = "thunk from <object <anonymous>>"
var p7648 = &p7648Var
var p7671Var = "object <anonymous>"
var p7671 = &p7671Var
var p7686Var = "thunk <tmp2> from <function <format_codes_arr>>"
var p7686 = &p7686Var
var p7698Var = "object <anonymous>"
var p7698 = &p7698Var
var p7716Var = "thunk from <object <anonymous>>"
var p7716 = &p7716Var
var p7743Var = "thunk from <object <anonymous>>"
var p7743 = &p7743Var
var p7772Var = "object <anonymous>"
var p7772 = &p7772Var
var p7788Var = "thunk <j2> from <function <format_codes_arr>>"
var p7788 = &p7788Var
var p7799Var = "thunk <val> from <function <format_codes_arr>>"
var p7799 = &p7799Var
var p7803Var = "thunk from <thunk <val> from <function <format_codes_arr>>>"
var p7803 = &p7803Var
var p7831Var = "thunk from <thunk <val> from <function <format_codes_arr>>>"
var p7831 = &p7831Var
var p7848Var = "thunk <s> from <function <format_codes_arr>>"
var p7848 = &p7848Var
var p7863Var = "thunk from <thunk <s> from <function <format_codes_arr>>>"
var p7863 = &p7863Var
var p7893Var = "thunk <s_padded> from <function <format_codes_arr>>"
var p7893 = &p7893Var
var p7901Var = "thunk from <thunk <s_padded> from <function <format_codes_arr>>>"
var p7901 = &p7901Var
var p7916Var = "thunk from <thunk <s_padded> from <function <format_codes_arr>>>"
var p7916 = &p7916Var
var p7933Var = "thunk <j3> from <function <format_codes_arr>>"
var p7933 = &p7933Var
var p7958Var = "thunk from <function <format_codes_arr>>"
var p7958 = &p7958Var
var p7999Var = "thunk <format_codes_arr> from <function <anonymous>>"
var p7999 = &p7999Var
var p8011Var = "function <format_codes_obj>"
var p8011 = &p8011Var
var p8015Var = "thunk from <function <format_codes_obj>>"
var p8015 = &p8015Var
var p8028Var = "thunk <code> from <function <format_codes_obj>>"
var p8028 = &p8028Var
var p8044Var = "thunk from <function <format_codes_obj>>"
var p8044 = &p8044Var
var p8054Var = "thunk from <function <format_codes_obj>>"
var p8054 = &p8054Var
var p8075Var = "thunk <f> from <function <format_codes_obj>>"
var p8075 = &p8075Var
var p8099Var = "thunk <fw> from <function <format_codes_obj>>"
var p8099 = &p8099Var
var p8123Var = "thunk <prec> from <function <format_codes_obj>>"
var p8123 = &p8123Var
var p8150Var = "thunk <val> from <function <format_codes_obj>>"
var p8150 = &p8150Var
var p8154Var = "thunk from <thunk <val> from <function <format_codes_obj>>>"
var p8154 = &p8154Var
var p8182Var = "thunk <s> from <function <format_codes_obj>>"
var p8182 = &p8182Var
var p8197Var = "thunk from <thunk <s> from <function <format_codes_obj>>>"
var p8197 = &p8197Var
var p8221Var = "thunk <s_padded> from <function <format_codes_obj>>"
var p8221 = &p8221Var
var p8229Var = "thunk from <thunk <s_padded> from <function <format_codes_obj>>>"
var p8229 = &p8229Var
var p8241Var = "thunk from <thunk <s_padded> from <function <format_codes_obj>>>"
var p8241 = &p8241Var
var p8256Var = "thunk from <function <format_codes_obj>>"
var p8256 = &p8256Var
var p8293Var = "thunk <format_codes_obj> from <function <anonymous>>"
var p8293 = &p8293Var
var p8301Var = "function <anonymous>"
var p8301 = &p8301Var
var p8305Var = "thunk from <function <anonymous>>"
var p8305 = &p8305Var
var p8314Var = "thunk from <function <anonymous>>"
var p8314 = &p8314Var
var p8331Var = "thunk from <function <anonymous>>"
var p8331 = &p8331Var
var p8340Var = "thunk from <function <anonymous>>"
var p8340 = &p8340Var
var p8353Var = "thunk from <function <anonymous>>"
var p8353 = &p8353Var
var p8358Var = "thunk from <thunk from <function <anonymous>>>"
var p8358 = &p8358Var
var p8451Var = "function <aux>"
var p8451 = &p8451Var
var p8464Var = "thunk from <function <aux>>"
var p8464 = &p8464Var
var p8474Var = "thunk from <thunk from <function <aux>>>"
var p8474 = &p8474Var
var p8492Var = "thunk <aux> from <function <anonymous>>"
var p8492 = &p8492Var
var p8497Var = "function <anonymous>"
var p8497 = &p8497Var
var p8501Var = "thunk from <function <anonymous>>"
var p8501 = &p8501Var
var p8517Var = "thunk from <thunk from <function <anonymous>>>"
var p8517 = &p8517Var
var p8538Var = "function <aux>"
var p8538 = &p8538Var
var p8542Var = "thunk from <function <aux>>"
var p8542 = &p8542Var
var p8557Var = "thunk from <function <aux>>"
var p8557 = &p8557Var
var p8566Var = "thunk from <thunk from <function <aux>>>"
var p8566 = &p8566Var
var p8585Var = "thunk <aux> from <function <anonymous>>"
var p8585 = &p8585Var
var p8590Var = "function <anonymous>"
var p8590 = &p8590Var
var p8594Var = "thunk from <function <anonymous>>"
var p8594 = &p8594Var
var p8615Var = "function <anonymous>"
var p8615 = &p8615Var
var p8619Var = "thunk from <function <anonymous>>"
var p8619 = &p8619Var
var p8633Var = "thunk from <function <anonymous>>"
var p8633 = &p8633Var
var p8650Var = "thunk from <function <anonymous>>"
var p8650 = &p8650Var
var p8664Var = "thunk from <function <anonymous>>"
var p8664 = &p8664Var
var p8681Var = "thunk from <function <anonymous>>"
var p8681 = &p8681Var
var p8695Var = "thunk from <function <anonymous>>"
var p8695 = &p8695Var
var p8711Var = "thunk from <function <anonymous>>"
var p8711 = &p8711Var
var p8721Var = "thunk from <thunk from <function <anonymous>>>"
var p8721 = &p8721Var
var p8741Var = "function <anonymous>"
var p8741 = &p8741Var
var p8776Var = "function <anonymous>"
var p8776 = &p8776Var
var p8780Var = "thunk from <function <anonymous>>"
var p8780 = &p8780Var
var p8794Var = "thunk from <function <anonymous>>"
var p8794 = &p8794Var
var p8829Var = "function <anonymous>"
var p8829 = &p8829Var
var p8833Var = "thunk from <function <anonymous>>"
var p8833 = &p8833Var
var p8847Var = "thunk from <function <anonymous>>"
var p8847 = &p8847Var
var p8891Var = "function <anonymous>"
var p8891 = &p8891Var
var p8895Var = "thunk from <function <anonymous>>"
var p8895 = &p8895Var
var p8909Var = "thunk from <function <anonymous>>"
var p8909 = &p8909Var
var p8926Var = "thunk from <function <anonymous>>"
var p8926 = &p8926Var
var p8940Var = "thunk from <function <anonymous>>"
var p8940 = &p8940Var
var p8976Var = "function <anonymous>"
var p8976 = &p8976Var
var p8980Var = "thunk from <function <anonymous>>"
var p8980 = &p8980Var
var p8994Var = "thunk from <function <anonymous>>"
var p8994 = &p8994Var
var p9011Var = "thunk from <function <anonymous>>"
var p9011 = &p9011Var
var p9025Var = "thunk from <function <anonymous>>"
var p9025 = &p9025Var
var p9057Var = "function <anonymous>"
var p9057 = &p9057Var
var p9089Var = "function <anonymous>"
var p9089 = &p9089Var
var p9095Var = "function <anonymous>"
var p9095 = &p9095Var
var p9101Var = "thunk from <function <anonymous>>"
var p9101 = &p9101Var
var p9119Var = "function <body_lines>"
var p9119 = &p9119Var
var p9123Var = "thunk from <function <body_lines>>"
var p9123 = &p9123Var
var p9138Var = "thunk <value_or_values> from <thunk from <thunk from <function <body_lines>>>>"
var p9138 = &p9138Var
var p9149Var = "thunk from <thunk from <function <body_lines>>>"
var p9149 = &p9149Var
var p9153Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p9153 = &p9153Var
var p9174Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p9174 = &p9174Var
var p9178Var = "thunk from <thunk from <thunk from <thunk from <function <body_lines>>>>>"
var p9178 = &p9178Var
var p9200Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p9200 = &p9200Var
var p9204Var = "thunk from <thunk from <thunk from <thunk from <function <body_lines>>>>>"
var p9204 = &p9204Var
var p9228Var = "thunk from <thunk from <function <body_lines>>>"
var p9228 = &p9228Var
var p9234Var = "thunk <body_lines> from <function <anonymous>>"
var p9234 = &p9234Var
var p9242Var = "function <section_lines>"
var p9242 = &p9242Var
var p9246Var = "thunk from <function <section_lines>>"
var p9246 = &p9246Var
var p9259Var = "thunk from <function <section_lines>>"
var p9259 = &p9259Var
var p9263Var = "thunk from <thunk from <function <section_lines>>>"
var p9263 = &p9263Var
var p9270Var = "thunk <section_lines> from <function <anonymous>>"
var p9270 = &p9270Var
var p9279Var = "thunk <main_body> from <function <anonymous>>"
var p9279 = &p9279Var
var p9283Var = "thunk from <thunk <main_body> from <function <anonymous>>>"
var p9283 = &p9283Var
var p9295Var = "thunk from <thunk <main_body> from <function <anonymous>>>"
var p9295 = &p9295Var
var p9314Var = "thunk from <thunk <all_sections> from <function <anonymous>>>"
var p9314 = &p9314Var
var p9318Var = "thunk from <thunk from <thunk <all_sections> from <function <anonymous>>>>"
var p9318 = &p9318Var
var p9338Var = "thunk <all_sections> from <function <anonymous>>"
var p9338 = &p9338Var
var p9345Var = "thunk from <thunk <all_sections> from <function <anonymous>>>"
var p9345 = &p9345Var
var p9355Var = "function <anonymous>"
var p9355 = &p9355Var
var p9359Var = "thunk from <function <anonymous>>"
var p9359 = &p9359Var
var p9364Var = "thunk from <thunk from <function <anonymous>>>"
var p9364 = &p9364Var
var p9374Var = "thunk from <thunk from <function <anonymous>>>"
var p9374 = &p9374Var
var p9395Var = "function <anonymous>"
var p9395 = &p9395Var
var p9399Var = "thunk from <function <anonymous>>"
var p9399 = &p9399Var
var p9414Var = "thunk <escapeStringToml> from <function <anonymous>>"
var p9414 = &p9414Var
var p9424Var = "thunk <bare_allowed> from <function <escapeKeyToml>>"
var p9424 = &p9424Var
var p9432Var = "thunk from <thunk <bare_allowed> from <function <escapeKeyToml>>>"
var p9432 = &p9432Var
var p9436Var = "thunk from <thunk from <thunk <bare_allowed> from <function <escapeKeyToml>>>>"
var p9436 = &p9436Var
var p9442Var = "function <escapeKeyToml>"
var p9442 = &p9442Var
var p9456Var = "thunk from <function <escapeKeyToml>>"
var p9456 = &p9456Var
var p9464Var = "thunk from <thunk from <function <escapeKeyToml>>>"
var p9464 = &p9464Var
var p9468Var = "thunk from <thunk from <thunk from <function <escapeKeyToml>>>>"
var p9468 = &p9468Var
var p9483Var = "thunk from <function <escapeKeyToml>>"
var p9483 = &p9483Var
var p9491Var = "thunk <escapeKeyToml> from <function <anonymous>>"
var p9491 = &p9491Var
var p9500Var = "function <isTableArray>"
var p9500 = &p9500Var
var p9508Var = "thunk from <function <isTableArray>>"
var p9508 = &p9508Var
var p9515Var = "thunk from <thunk from <function <isTableArray>>>"
var p9515 = &p9515Var
var p9532Var = "thunk from <function <isTableArray>>"
var p9532 = &p9532Var
var p9544Var = "thunk from <function <isTableArray>>"
var p9544 = &p9544Var
var p9550Var = "thunk <isTableArray> from <function <anonymous>>"
var p9550 = &p9550Var
var p9556Var = "function <isSection>"
var p9556 = &p9556Var
var p9560Var = "thunk from <function <isSection>>"
var p9560 = &p9560Var
var p9571Var = "thunk from <function <isSection>>"
var p9571 = &p9571Var
var p9576Var = "thunk <isSection> from <function <anonymous>>"
var p9576 = &p9576Var
var p9582Var = "function <renderValue>"
var p9582 = &p9582Var
var p9619Var = "thunk from <function <renderValue>>"
var p9619 = &p9619Var
var p9637Var = "thunk from <function <renderValue>>"
var p9637 = &p9637Var
var p9646Var = "thunk from <function <renderValue>>"
var p9646 = &p9646Var
var p9658Var = "thunk from <function <renderValue>>"
var p9658 = &p9658Var
var p9678Var = "thunk from <function <renderValue>>"
var p9678 = &p9678Var
var p9692Var = "thunk from <function <renderValue>>"
var p9692 = &p9692Var
var p9705Var = "thunk <range> from <function <renderValue>>"
var p9705 = &p9705Var
var p9709Var = "thunk from <thunk <range> from <function <renderValue>>>"
var p9709 = &p9709Var
var p9720Var = "thunk from <thunk from <thunk <range> from <function <renderValue>>>>"
var p9720 = &p9720Var
var p9729Var = "thunk <new_indent> from <function <renderValue>>"
var p9729 = &p9729Var
var p9743Var = "thunk <separator> from <function <renderValue>>"
var p9743 = &p9743Var
var p9755Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9755 = &p9755Var
var p9768Var = "thunk <lines> from <function <renderValue>>"
var p9768 = &p9768Var
var p9782Var = "thunk from <thunk from <thunk <lines> from <function <renderValue>>>>"
var p9782 = &p9782Var
var p9786Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9786 = &p9786Var
var p9803Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>"
var p9803 = &p9803Var
var p9808Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>>"
var p9808 = &p9808Var
var p9817Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>>>"
var p9817 = &p9817Var
var p9830Var = "thunk from <thunk from <thunk <lines> from <function <renderValue>>>>"
var p9830 = &p9830Var
var p9843Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9843 = &p9843Var
var p9861Var = "thunk from <function <renderValue>>"
var p9861 = &p9861Var
var p9885Var = "thunk from <function <renderValue>>"
var p9885 = &p9885Var
var p9894Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9894 = &p9894Var
var p9895Var = "thunk <lines> from <function <renderValue>>"
var p9895 = &p9895Var
var p9907Var = "thunk from <thunk from <thunk <lines> from <function <renderValue>>>>"
var p9907 = &p9907Var
var p9908Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9908 = &p9908Var
var p9924Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>"
var p9924 = &p9924Var
var p9929Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>>"
var p9929 = &p9929Var
var p9938Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>>>"
var p9938 = &p9938Var
var p9954Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <renderValue>>>>>>"
var p9954 = &p9954Var
var p9960Var = "thunk from <thunk from <thunk <lines> from <function <renderValue>>>>"
var p9960 = &p9960Var
var p9973Var = "thunk from <thunk from <thunk <lines> from <function <renderValue>>>>"
var p9973 = &p9973Var
var p9981Var = "thunk from <thunk <lines> from <function <renderValue>>>"
var p9981 = &p9981Var
var p9995Var = "thunk from <function <renderValue>>"
var p9995 = &p9995Var
var p10019Var = "thunk <renderValue> from <function <anonymous>>"
var p10019 = &p10019Var
var p10029Var = "thunk <kvp> from <function <renderTableInternal>>"
var p10029 = &p10029Var
var p10044Var = "thunk from <thunk <kvp> from <function <renderTableInternal>>>"
var p10044 = &p10044Var
var p10049Var = "thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>"
var p10049 = &p10049Var
var p10063Var = "thunk from <thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>>"
var p10063 = &p10063Var
var p10068Var = "thunk from <thunk from <thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>>>"
var p10068 = &p10068Var
var p10077Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>>>>"
var p10077 = &p10077Var
var p10095Var = "thunk from <thunk from <thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>>>"
var p10095 = &p10095Var
var p10104Var = "thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>"
var p10104 = &p10104Var
var p10119Var = "thunk from <thunk from <thunk <kvp> from <function <renderTableInternal>>>>"
var p10119 = &p10119Var
var p10138Var = "thunk <sections> from <function <renderTableInternal>>"
var p10138 = &p10138Var
var p10143Var = "thunk from <thunk <sections> from <function <renderTableInternal>>>"
var p10143 = &p10143Var
var p10157Var = "thunk from <thunk <sections> from <function <renderTableInternal>>>"
var p10157 = &p10157Var
var p10162Var = "thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>"
var p10162 = &p10162Var
var p10175Var = "thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>"
var p10175 = &p10175Var
var p10184Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>>"
var p10184 = &p10184Var
var p10194Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>>"
var p10194 = &p10194Var
var p10210Var = "thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>"
var p10210 = &p10210Var
var p10219Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>>"
var p10219 = &p10219Var
var p10229Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>>"
var p10229 = &p10229Var
var p10254Var = "thunk from <thunk <sections> from <function <renderTableInternal>>>"
var p10254 = &p10254Var
var p10265Var = "thunk from <thunk <sections> from <function <renderTableInternal>>>"
var p10265 = &p10265Var
var p10269Var = "thunk from <thunk from <thunk <sections> from <function <renderTableInternal>>>>"
var p10269 = &p10269Var
var p10281Var = "function <renderTableInternal>"
var p10281 = &p10281Var
var p10285Var = "thunk from <function <renderTableInternal>>"
var p10285 = &p10285Var
var p10294Var = "thunk <renderTableInternal> from <function <anonymous>>"
var p10294 = &p10294Var
var p10300Var = "function <renderTable>"
var p10300 = &p10300Var
var p10304Var = "thunk from <function <renderTable>>"
var p10304 = &p10304Var
var p10338Var = "thunk from <function <renderTable>>"
var p10338 = &p10338Var
var p10347Var = "thunk from <thunk from <function <renderTable>>>"
var p10347 = &p10347Var
var p10366Var = "thunk <renderTable> from <function <anonymous>>"
var p10366 = &p10366Var
var p10376Var = "thunk <range> from <function <renderTableArray>>"
var p10376 = &p10376Var
var p10380Var = "thunk from <thunk <range> from <function <renderTableArray>>>"
var p10380 = &p10380Var
var p10391Var = "thunk from <thunk from <thunk <range> from <function <renderTableArray>>>>"
var p10391 = &p10391Var
var p10411Var = "thunk from <thunk <sections> from <function <renderTableArray>>>"
var p10411 = &p10411Var
var p10416Var = "thunk from <thunk from <thunk <sections> from <function <renderTableArray>>>>"
var p10416 = &p10416Var
var p10427Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableArray>>>>>"
var p10427 = &p10427Var
var p10465Var = "thunk from <thunk from <thunk <sections> from <function <renderTableArray>>>>"
var p10465 = &p10465Var
var p10474Var = "thunk from <thunk from <thunk from <thunk <sections> from <function <renderTableArray>>>>>"
var p10474 = &p10474Var
var p10495Var = "thunk <sections> from <function <renderTableArray>>"
var p10495 = &p10495Var
var p10504Var = "function <renderTableArray>"
var p10504 = &p10504Var
var p10508Var = "thunk from <function <renderTableArray>>"
var p10508 = &p10508Var
var p10517Var = "thunk <renderTableArray> from <function <anonymous>>"
var p10517 = &p10517Var
var p10525Var = "function <anonymous>"
var p10525 = &p10525Var
var p10529Var = "thunk from <function <anonymous>>"
var p10529 = &p10529Var
var p10538Var = "thunk from <function <anonymous>>"
var p10538 = &p10538Var
var p10554Var = "thunk from <function <anonymous>>"
var p10554 = &p10554Var
var p10577Var = "thunk <str> from <function <anonymous>>"
var p10577 = &p10577Var
var p10581Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p10581 = &p10581Var
var p10590Var = "function <trans>"
var p10590 = &p10590Var
var p10651Var = "thunk <cp> from <function <trans>>"
var p10651 = &p10651Var
var p10655Var = "thunk from <thunk <cp> from <function <trans>>>"
var p10655 = &p10655Var
var p10690Var = "thunk from <function <trans>>"
var p10690 = &p10690Var
var p10718Var = "thunk <trans> from <function <anonymous>>"
var p10718 = &p10718Var
var p10729Var = "function <anonymous>"
var p10729 = &p10729Var
var p10738Var = "thunk from <function <anonymous>>"
var p10738 = &p10738Var
var p10751Var = "thunk from <thunk from <function <anonymous>>>"
var p10751 = &p10751Var
var p10755Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p10755 = &p10755Var
var p10769Var = "thunk from <thunk from <function <anonymous>>>"
var p10769 = &p10769Var
var p10789Var = "function <anonymous>"
var p10789 = &p10789Var
var p10793Var = "thunk from <function <anonymous>>"
var p10793 = &p10793Var
var p10807Var = "thunk <str> from <function <anonymous>>"
var p10807 = &p10807Var
var p10811Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p10811 = &p10811Var
var p10820Var = "function <trans>"
var p10820 = &p10820Var
var p10833Var = "thunk <trans> from <function <anonymous>>"
var p10833 = &p10833Var
var p10843Var = "function <anonymous>"
var p10843 = &p10843Var
var p10852Var = "thunk from <function <anonymous>>"
var p10852 = &p10852Var
var p10865Var = "thunk from <thunk from <function <anonymous>>>"
var p10865 = &p10865Var
var p10869Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p10869 = &p10869Var
var p10883Var = "thunk from <thunk from <function <anonymous>>>"
var p10883 = &p10883Var
var p10904Var = "thunk <str> from <function <anonymous>>"
var p10904 = &p10904Var
var p10908Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p10908 = &p10908Var
var p10917Var = "function <trans>"
var p10917 = &p10917Var
var p10930Var = "thunk <trans> from <function <anonymous>>"
var p10930 = &p10930Var
var p10937Var = "function <anonymous>"
var p10937 = &p10937Var
var p10944Var = "function <anonymous>"
var p10944 = &p10944Var
var p10948Var = "thunk from <function <anonymous>>"
var p10948 = &p10948Var
var p10955Var = "thunk from <function <anonymous>>"
var p10955 = &p10955Var
var p10965Var = "thunk from <thunk from <function <anonymous>>>"
var p10965 = &p10965Var
var p10985Var = "thunk <str> from <function <anonymous>>"
var p10985 = &p10985Var
var p10989Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p10989 = &p10989Var
var p10998Var = "function <anonymous>"
var p10998 = &p10998Var
var p11002Var = "thunk from <function <anonymous>>"
var p11002 = &p11002Var
var p11018Var = "thunk from <thunk from <function <anonymous>>>"
var p11018 = &p11018Var
var p11022Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p11022 = &p11022Var
var p11040Var = "thunk from <thunk from <function <anonymous>>>"
var p11040 = &p11040Var
var p11056Var = "function <anonymous>"
var p11056 = &p11056Var
var p11060Var = "thunk from <function <anonymous>>"
var p11060 = &p11060Var
var p11073Var = "function <anonymous>"
var p11073 = &p11073Var
var p11077Var = "thunk from <function <anonymous>>"
var p11077 = &p11077Var
var p11093Var = "function <aux>"
var p11093 = &p11093Var
var p11124Var = "thunk from <function <aux>>"
var p11124 = &p11124Var
var p11142Var = "thunk from <function <aux>>"
var p11142 = &p11142Var
var p11154Var = "thunk from <function <aux>>"
var p11154 = &p11154Var
var p11166Var = "thunk from <function <aux>>"
var p11166 = &p11166Var
var p11186Var = "thunk from <function <aux>>"
var p11186 = &p11186Var
var p11196Var = "thunk <range> from <function <aux>>"
var p11196 = &p11196Var
var p11200Var = "thunk from <thunk <range> from <function <aux>>>"
var p11200 = &p11200Var
var p11211Var = "thunk from <thunk from <thunk <range> from <function <aux>>>>"
var p11211 = &p11211Var
var p11220Var = "thunk <new_indent> from <function <aux>>"
var p11220 = &p11220Var
var p11232Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11232 = &p11232Var
var p11240Var = "thunk <lines> from <function <aux>>"
var p11240 = &p11240Var
var p11254Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p11254 = &p11254Var
var p11258Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11258 = &p11258Var
var p11275Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p11275 = &p11275Var
var p11280Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p11280 = &p11280Var
var p11289Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>>"
var p11289 = &p11289Var
var p11302Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p11302 = &p11302Var
var p11315Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11315 = &p11315Var
var p11333Var = "thunk from <function <aux>>"
var p11333 = &p11333Var
var p11352Var = "thunk from <function <aux>>"
var p11352 = &p11352Var
var p11362Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11362 = &p11362Var
var p11370Var = "thunk <lines> from <function <aux>>"
var p11370 = &p11370Var
var p11384Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p11384 = &p11384Var
var p11388Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11388 = &p11388Var
var p11405Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p11405 = &p11405Var
var p11410Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p11410 = &p11410Var
var p11419Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>>"
var p11419 = &p11419Var
var p11444Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p11444 = &p11444Var
var p11458Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p11458 = &p11458Var
var p11471Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p11471 = &p11471Var
var p11480Var = "thunk from <thunk <lines> from <function <aux>>>"
var p11480 = &p11480Var
var p11498Var = "thunk from <function <aux>>"
var p11498 = &p11498Var
var p11522Var = "thunk <aux> from <function <anonymous>>"
var p11522 = &p11522Var
var p11527Var = "function <anonymous>"
var p11527 = &p11527Var
var p11531Var = "thunk from <function <anonymous>>"
var p11531 = &p11531Var
var p11554Var = "function <onlyChars>"
var p11554 = &p11554Var
var p11558Var = "thunk from <function <onlyChars>>"
var p11558 = &p11558Var
var p11573Var = "thunk from <function <onlyChars>>"
var p11573 = &p11573Var
var p11577Var = "thunk from <thunk from <function <onlyChars>>>"
var p11577 = &p11577Var
var p11591Var = "thunk <onlyChars> from <function <anonymous>>"
var p11591 = &p11591Var
var p11603Var = "thunk from <thunk <reserved> from <function <isReserved>>>"
var p11603 = &p11603Var
var p11637Var = "thunk <reserved> from <function <isReserved>>"
var p11637 = &p11637Var
var p11655Var = "thunk <bad> from <function <isReserved>>"
var p11655 = &p11655Var
var p11659Var = "thunk from <thunk <bad> from <function <isReserved>>>"
var p11659 = &p11659Var
var p11668Var = "thunk from <thunk <bad> from <function <isReserved>>>"
var p11668 = &p11668Var
var p11681Var = "function <isReserved>"
var p11681 = &p11681Var
var p11690Var = "thunk from <function <isReserved>>"
var p11690 = &p11690Var
var p11708Var = "thunk <isReserved> from <function <anonymous>>"
var p11708 = &p11708Var
var p11718Var = "function <typeMatch>"
var p11718 = &p11718Var
var p11730Var = "thunk from <function <typeMatch>>"
var p11730 = &p11730Var
var p11747Var = "thunk from <function <typeMatch>>"
var p11747 = &p11747Var
var p11762Var = "thunk <typeMatch> from <function <anonymous>>"
var p11762 = &p11762Var
var p11774Var = "thunk <letters> from <function <bareSafe>>"
var p11774 = &p11774Var
var p11782Var = "thunk from <thunk <letters> from <function <bareSafe>>>"
var p11782 = &p11782Var
var p11786Var = "thunk from <thunk from <thunk <letters> from <function <bareSafe>>>>"
var p11786 = &p11786Var
var p11796Var = "thunk <digits> from <function <bareSafe>>"
var p11796 = &p11796Var
var p11804Var = "thunk from <thunk <digits> from <function <bareSafe>>>"
var p11804 = &p11804Var
var p11808Var = "thunk from <thunk from <thunk <digits> from <function <bareSafe>>>>"
var p11808 = &p11808Var
var p11818Var = "thunk <intChars> from <function <bareSafe>>"
var p11818 = &p11818Var
var p11827Var = "thunk from <thunk <intChars> from <function <bareSafe>>>"
var p11827 = &p11827Var
var p11831Var = "thunk from <thunk from <thunk <intChars> from <function <bareSafe>>>>"
var p11831 = &p11831Var
var p11844Var = "thunk <binChars> from <function <bareSafe>>"
var p11844 = &p11844Var
var p11853Var = "thunk from <thunk <binChars> from <function <bareSafe>>>"
var p11853 = &p11853Var
var p11857Var = "thunk from <thunk from <thunk <binChars> from <function <bareSafe>>>>"
var p11857 = &p11857Var
var p11870Var = "thunk <hexChars> from <function <bareSafe>>"
var p11870 = &p11870Var
var p11879Var = "thunk from <thunk <hexChars> from <function <bareSafe>>>"
var p11879 = &p11879Var
var p11883Var = "thunk from <thunk from <thunk <hexChars> from <function <bareSafe>>>>"
var p11883 = &p11883Var
var p11896Var = "thunk <floatChars> from <function <bareSafe>>"
var p11896 = &p11896Var
var p11905Var = "thunk from <thunk <floatChars> from <function <bareSafe>>>"
var p11905 = &p11905Var
var p11909Var = "thunk from <thunk from <thunk <floatChars> from <function <bareSafe>>>>"
var p11909 = &p11909Var
var p11922Var = "thunk <dateChars> from <function <bareSafe>>"
var p11922 = &p11922Var
var p11931Var = "thunk from <thunk <dateChars> from <function <bareSafe>>>"
var p11931 = &p11931Var
var p11935Var = "thunk from <thunk from <thunk <dateChars> from <function <bareSafe>>>>"
var p11935 = &p11935Var
var p11948Var = "thunk <safeChars> from <function <bareSafe>>"
var p11948 = &p11948Var
var p11953Var = "thunk from <thunk <safeChars> from <function <bareSafe>>>"
var p11953 = &p11953Var
var p11966Var = "thunk <keyLc> from <function <bareSafe>>"
var p11966 = &p11966Var
var p11970Var = "thunk from <thunk <keyLc> from <function <bareSafe>>>"
var p11970 = &p11970Var
var p11980Var = "thunk <keyChars> from <function <bareSafe>>"
var p11980 = &p11980Var
var p11984Var = "thunk from <thunk <keyChars> from <function <bareSafe>>>"
var p11984 = &p11984Var
var p11994Var = "thunk <keySet> from <function <bareSafe>>"
var p11994 = &p11994Var
var p11998Var = "thunk from <thunk <keySet> from <function <bareSafe>>>"
var p11998 = &p11998Var
var p12008Var = "thunk <keySetLc> from <function <bareSafe>>"
var p12008 = &p12008Var
var p12016Var = "thunk from <thunk <keySetLc> from <function <bareSafe>>>"
var p12016 = &p12016Var
var p12020Var = "thunk from <thunk from <thunk <keySetLc> from <function <bareSafe>>>>"
var p12020 = &p12020Var
var p12028Var = "function <bareSafe>"
var p12028 = &p12028Var
var p12032Var = "thunk from <function <bareSafe>>"
var p12032 = &p12032Var
var p12046Var = "thunk from <function <bareSafe>>"
var p12046 = &p12046Var
var p12067Var = "thunk from <function <bareSafe>>"
var p12067 = &p12067Var
var p12071Var = "thunk from <thunk from <function <bareSafe>>>"
var p12071 = &p12071Var
var p12082Var = "thunk from <function <bareSafe>>"
var p12082 = &p12082Var
var p12107Var = "thunk from <function <bareSafe>>"
var p12107 = &p12107Var
var p12111Var = "thunk from <thunk from <function <bareSafe>>>"
var p12111 = &p12111Var
var p12122Var = "thunk from <function <bareSafe>>"
var p12122 = &p12122Var
var p12138Var = "thunk from <function <bareSafe>>"
var p12138 = &p12138Var
var p12153Var = "thunk from <function <bareSafe>>"
var p12153 = &p12153Var
var p12162Var = "thunk from <function <bareSafe>>"
var p12162 = &p12162Var
var p12189Var = "thunk from <function <bareSafe>>"
var p12189 = &p12189Var
var p12193Var = "thunk from <thunk from <function <bareSafe>>>"
var p12193 = &p12193Var
var p12214Var = "thunk from <function <bareSafe>>"
var p12214 = &p12214Var
var p12218Var = "thunk from <thunk from <function <bareSafe>>>"
var p12218 = &p12218Var
var p12239Var = "thunk from <function <bareSafe>>"
var p12239 = &p12239Var
var p12243Var = "thunk from <thunk from <function <bareSafe>>>"
var p12243 = &p12243Var
var p12254Var = "thunk from <function <bareSafe>>"
var p12254 = &p12254Var
var p12274Var = "thunk from <function <bareSafe>>"
var p12274 = &p12274Var
var p12289Var = "thunk from <function <bareSafe>>"
var p12289 = &p12289Var
var p12308Var = "thunk from <function <bareSafe>>"
var p12308 = &p12308Var
var p12312Var = "thunk from <thunk from <function <bareSafe>>>"
var p12312 = &p12312Var
var p12323Var = "thunk from <function <bareSafe>>"
var p12323 = &p12323Var
var p12386Var = "thunk <bareSafe> from <function <anonymous>>"
var p12386 = &p12386Var
var p12394Var = "function <escapeKeyYaml>"
var p12394 = &p12394Var
var p12398Var = "thunk from <function <escapeKeyYaml>>"
var p12398 = &p12398Var
var p12411Var = "thunk from <function <escapeKeyYaml>>"
var p12411 = &p12411Var
var p12417Var = "thunk <escapeKeyYaml> from <function <anonymous>>"
var p12417 = &p12417Var
var p12425Var = "function <aux>"
var p12425 = &p12425Var
var p12456Var = "thunk from <function <aux>>"
var p12456 = &p12456Var
var p12474Var = "thunk from <function <aux>>"
var p12474 = &p12474Var
var p12484Var = "thunk <len> from <function <aux>>"
var p12484 = &p12484Var
var p12488Var = "thunk from <thunk <len> from <function <aux>>>"
var p12488 = &p12488Var
var p12519Var = "thunk <split> from <function <aux>>"
var p12519 = &p12519Var
var p12523Var = "thunk from <thunk <split> from <function <aux>>>"
var p12523 = &p12523Var
var p12537Var = "thunk from <function <aux>>"
var p12537 = &p12537Var
var p12565Var = "thunk from <thunk from <function <aux>>>"
var p12565 = &p12565Var
var p12574Var = "thunk from <thunk from <function <aux>>>"
var p12574 = &p12574Var
var p12588Var = "thunk from <function <aux>>"
var p12588 = &p12588Var
var p12607Var = "thunk from <function <aux>>"
var p12607 = &p12607Var
var p12627Var = "thunk from <function <aux>>"
var p12627 = &p12627Var
var p12641Var = "thunk from <function <aux>>"
var p12641 = &p12641Var
var p12654Var = "function <params>"
var p12654 = &p12654Var
var p12663Var = "thunk from <function <params>>"
var p12663 = &p12663Var
var p12675Var = "thunk from <function <params>>"
var p12675 = &p12675Var
var p12684Var = "object <anonymous>"
var p12684 = &p12684Var
var p12707Var = "thunk from <function <params>>"
var p12707 = &p12707Var
var p12719Var = "thunk from <function <params>>"
var p12719 = &p12719Var
var p12728Var = "object <anonymous>"
var p12728 = &p12728Var
var p12739Var = "object <anonymous>"
var p12739 = &p12739Var
var p12748Var = "thunk <params> from <function <aux>>"
var p12748 = &p12748Var
var p12757Var = "thunk <range> from <function <aux>>"
var p12757 = &p12757Var
var p12761Var = "thunk from <thunk <range> from <function <aux>>>"
var p12761 = &p12761Var
var p12772Var = "thunk from <thunk from <thunk <range> from <function <aux>>>>"
var p12772 = &p12772Var
var p12800Var = "thunk from <thunk <parts> from <function <aux>>>"
var p12800 = &p12800Var
var p12805Var = "thunk from <thunk from <thunk <parts> from <function <aux>>>>"
var p12805 = &p12805Var
var p12814Var = "thunk from <thunk from <thunk from <thunk <parts> from <function <aux>>>>>"
var p12814 = &p12814Var
var p12843Var = "thunk from <thunk <parts> from <function <aux>>>"
var p12843 = &p12843Var
var p12848Var = "thunk from <thunk from <thunk <parts> from <function <aux>>>>"
var p12848 = &p12848Var
var p12854Var = "thunk <parts> from <function <aux>>"
var p12854 = &p12854Var
var p12872Var = "thunk from <function <aux>>"
var p12872 = &p12872Var
var p12897Var = "thunk from <function <aux>>"
var p12897 = &p12897Var
var p12911Var = "thunk from <function <aux>>"
var p12911 = &p12911Var
var p12924Var = "function <params>"
var p12924 = &p12924Var
var p12933Var = "thunk from <function <params>>"
var p12933 = &p12933Var
var p12945Var = "thunk from <function <params>>"
var p12945 = &p12945Var
var p12954Var = "object <anonymous>"
var p12954 = &p12954Var
var p12983Var = "thunk from <function <params>>"
var p12983 = &p12983Var
var p12995Var = "thunk from <function <params>>"
var p12995 = &p12995Var
var p13004Var = "object <anonymous>"
var p13004 = &p13004Var
var p13019Var = "object <anonymous>"
var p13019 = &p13019Var
var p13028Var = "thunk <params> from <function <aux>>"
var p13028 = &p13028Var
var p13053Var = "thunk from <thunk <lines> from <function <aux>>>"
var p13053 = &p13053Var
var p13058Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p13058 = &p13058Var
var p13067Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p13067 = &p13067Var
var p13098Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p13098 = &p13098Var
var p13106Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p13106 = &p13106Var
var p13120Var = "thunk from <thunk <lines> from <function <aux>>>"
var p13120 = &p13120Var
var p13125Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p13125 = &p13125Var
var p13131Var = "thunk <lines> from <function <aux>>"
var p13131 = &p13131Var
var p13144Var = "thunk from <thunk <lines> from <function <aux>>>"
var p13144 = &p13144Var
var p13158Var = "thunk from <function <aux>>"
var p13158 = &p13158Var
var p13190Var = "thunk <aux> from <function <anonymous>>"
var p13190 = &p13190Var
var p13195Var = "function <anonymous>"
var p13195 = &p13195Var
var p13199Var = "thunk from <function <anonymous>>"
var p13199 = &p13199Var
var p13229Var = "function <anonymous>"
var p13229 = &p13229Var
var p13233Var = "thunk from <function <anonymous>>"
var p13233 = &p13233Var
var p13247Var = "thunk from <function <anonymous>>"
var p13247 = &p13247Var
var p13271Var = "thunk from <function <anonymous>>"
var p13271 = &p13271Var
var p13287Var = "thunk from <thunk from <function <anonymous>>>"
var p13287 = &p13287Var
var p13291Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p13291 = &p13291Var
var p13326Var = "function <anonymous>"
var p13326 = &p13326Var
var p13330Var = "thunk from <function <anonymous>>"
var p13330 = &p13330Var
var p13354Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p13354 = &p13354Var
var p13362Var = "thunk from <thunk from <thunk <fields> from <function <anonymous>>>>"
var p13362 = &p13362Var
var p13366Var = "thunk from <thunk from <thunk from <thunk <fields> from <function <anonymous>>>>>"
var p13366 = &p13366Var
var p13378Var = "thunk from <thunk from <thunk from <thunk <fields> from <function <anonymous>>>>>"
var p13378 = &p13378Var
var p13394Var = "thunk <fields> from <function <anonymous>>"
var p13394 = &p13394Var
var p13398Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p13398 = &p13398Var
var p13418Var = "thunk from <function <anonymous>>"
var p13418 = &p13418Var
var p13422Var = "thunk from <thunk from <function <anonymous>>>"
var p13422 = &p13422Var
var p13439Var = "thunk from <function <anonymous>>"
var p13439 = &p13439Var
var p13458Var = "thunk from <function <anonymous>>"
var p13458 = &p13458Var
var p13462Var = "thunk from <thunk from <function <anonymous>>>"
var p13462 = &p13462Var
var p13478Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p13478 = &p13478Var
var p13482Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p13482 = &p13482Var
var p13503Var = "thunk from <function <anonymous>>"
var p13503 = &p13503Var
var p13522Var = "thunk from <function <anonymous>>"
var p13522 = &p13522Var
var p13526Var = "thunk from <thunk from <function <anonymous>>>"
var p13526 = &p13526Var
var p13540Var = "thunk from <function <anonymous>>"
var p13540 = &p13540Var
var p13555Var = "thunk from <function <anonymous>>"
var p13555 = &p13555Var
var p13567Var = "thunk from <function <anonymous>>"
var p13567 = &p13567Var
var p13635Var = "thunk from <thunk <vars> from <function <anonymous>>>"
var p13635 = &p13635Var
var p13639Var = "thunk from <thunk from <thunk <vars> from <function <anonymous>>>>"
var p13639 = &p13639Var
var p13650Var = "thunk from <thunk from <thunk from <thunk <vars> from <function <anonymous>>>>>"
var p13650 = &p13650Var
var p13666Var = "thunk <vars> from <function <anonymous>>"
var p13666 = &p13666Var
var p13670Var = "thunk from <thunk <vars> from <function <anonymous>>>"
var p13670 = &p13670Var
var p13680Var = "function <anonymous>"
var p13680 = &p13680Var
var p13684Var = "thunk from <function <anonymous>>"
var p13684 = &p13684Var
var p13689Var = "thunk from <thunk from <function <anonymous>>>"
var p13689 = &p13689Var
var p13707Var = "function <anonymous>"
var p13707 = &p13707Var
var p13711Var = "thunk from <function <anonymous>>"
var p13711 = &p13711Var
var p13732Var = "thunk from <function <anonymous>>"
var p13732 = &p13732Var
var p13747Var = "function <aux>"
var p13747 = &p13747Var
var p13751Var = "thunk from <function <aux>>"
var p13751 = &p13751Var
var p13761Var = "thunk <tag> from <function <aux>>"
var p13761 = &p13761Var
var p13773Var = "thunk <has_attrs> from <function <aux>>"
var p13773 = &p13773Var
var p13778Var = "thunk from <thunk <has_attrs> from <function <aux>>>"
var p13778 = &p13778Var
var p13793Var = "thunk from <thunk <has_attrs> from <function <aux>>>"
var p13793 = &p13793Var
var p13802Var = "thunk <attrs> from <function <aux>>"
var p13802 = &p13802Var
var p13815Var = "thunk <children> from <function <aux>>"
var p13815 = &p13815Var
var p13852Var = "thunk <attrs_str> from <function <aux>>"
var p13852 = &p13852Var
var p13856Var = "thunk from <thunk <attrs_str> from <function <aux>>>"
var p13856 = &p13856Var
var p13875Var = "thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>"
var p13875 = &p13875Var
var p13879Var = "thunk from <thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>>"
var p13879 = &p13879Var
var p13900Var = "thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>"
var p13900 = &p13900Var
var p13916Var = "thunk from <thunk from <function <aux>>>"
var p13916 = &p13916Var
var p13934Var = "thunk from <thunk from <thunk from <function <aux>>>>"
var p13934 = &p13934Var
var p13938Var = "thunk from <thunk from <thunk from <thunk from <function <aux>>>>>"
var p13938 = &p13938Var
var p13951Var = "thunk from <function <aux>>"
var p13951 = &p13951Var
var p13968Var = "thunk <aux> from <function <anonymous>>"
var p13968 = &p13968Var
var p13976Var = "thunk from <function <anonymous>>"
var p13976 = &p13976Var
var p13996Var = "thunk <bytes> from <function <anonymous>>"
var p13996 = &p13996Var
var p14000Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p14000 = &p14000Var
var p14015Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p14015 = &p14015Var
var p14036Var = "function <aux>"
var p14036 = &p14036Var
var p14040Var = "thunk from <function <aux>>"
var p14040 = &p14040Var
var p14059Var = "thunk from <function <aux>>"
var p14059 = &p14059Var
var p14073Var = "thunk <str> from <function <aux>>"
var p14073 = &p14073Var
var p14119Var = "thunk from <function <aux>>"
var p14119 = &p14119Var
var p14145Var = "thunk from <function <aux>>"
var p14145 = &p14145Var
var p14159Var = "thunk <str> from <function <aux>>"
var p14159 = &p14159Var
var p14245Var = "thunk from <function <aux>>"
var p14245 = &p14245Var
var p14268Var = "thunk <str> from <function <aux>>"
var p14268 = &p14268Var
var p14385Var = "thunk from <function <aux>>"
var p14385 = &p14385Var
var p14409Var = "thunk <aux> from <function <anonymous>>"
var p14409 = &p14409Var
var p14418Var = "thunk <sanity> from <function <anonymous>>"
var p14418 = &p14418Var
var p14433Var = "thunk from <thunk from <thunk <sanity> from <function <anonymous>>>>"
var p14433 = &p14433Var
var p14440Var = "thunk from <thunk <sanity> from <function <anonymous>>>"
var p14440 = &p14440Var
var p14447Var = "function <anonymous>"
var p14447 = &p14447Var
var p14459Var = "thunk from <function <anonymous>>"
var p14459 = &p14459Var
var p14480Var = "function <anonymous>"
var p14480 = &p14480Var
var p14496Var = "thunk from <function <anonymous>>"
var p14496 = &p14496Var
var p14526Var = "function <aux>"
var p14526 = &p14526Var
var p14530Var = "thunk from <function <aux>>"
var p14530 = &p14530Var
var p14546Var = "thunk from <thunk <n1> from <function <aux>>>"
var p14546 = &p14546Var
var p14575Var = "thunk <n1> from <function <aux>>"
var p14575 = &p14575Var
var p14582Var = "thunk <n2> from <function <aux>>"
var p14582 = &p14582Var
var p14599Var = "thunk from <thunk <n2> from <function <aux>>>"
var p14599 = &p14599Var
var p14643Var = "thunk <n3> from <function <aux>>"
var p14643 = &p14643Var
var p14660Var = "thunk from <thunk <n3> from <function <aux>>>"
var p14660 = &p14660Var
var p14702Var = "thunk from <function <aux>>"
var p14702 = &p14702Var
var p14737Var = "thunk <aux> from <function <anonymous>>"
var p14737 = &p14737Var
var p14745Var = "thunk from <function <anonymous>>"
var p14745 = &p14745Var
var p14766Var = "thunk <bytes> from <function <anonymous>>"
var p14766 = &p14766Var
var p14770Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p14770 = &p14770Var
var p14779Var = "function <anonymous>"
var p14779 = &p14779Var
var p14783Var = "thunk from <function <anonymous>>"
var p14783 = &p14783Var
var p14795Var = "thunk from <thunk from <function <anonymous>>>"
var p14795 = &p14795Var
var p14814Var = "thunk <l> from <function <anonymous>>"
var p14814 = &p14814Var
var p14818Var = "thunk from <thunk <l> from <function <anonymous>>>"
var p14818 = &p14818Var
var p14827Var = "function <anonymous>"
var p14827 = &p14827Var
var p14831Var = "thunk from <function <anonymous>>"
var p14831 = &p14831Var
var p14836Var = "function <anonymous>"
var p14836 = &p14836Var
var p14867Var = "thunk <l> from <function <quickSort>>"
var p14867 = &p14867Var
var p14871Var = "thunk from <thunk <l> from <function <quickSort>>>"
var p14871 = &p14871Var
var p14877Var = "function <quickSort>"
var p14877 = &p14877Var
var p14886Var = "thunk from <function <quickSort>>"
var p14886 = &p14886Var
var p14896Var = "thunk <pos> from <function <quickSort>>"
var p14896 = &p14896Var
var p14901Var = "thunk <pivot> from <function <quickSort>>"
var p14901 = &p14901Var
var p14906Var = "thunk from <thunk <pivot> from <function <quickSort>>>"
var p14906 = &p14906Var
var p14919Var = "thunk <rest> from <function <quickSort>>"
var p14919 = &p14919Var
var p14924Var = "thunk from <thunk <rest> from <function <quickSort>>>"
var p14924 = &p14924Var
var p14932Var = "function <anonymous>"
var p14932 = &p14932Var
var p14963Var = "thunk <left> from <function <quickSort>>"
var p14963 = &p14963Var
var p14969Var = "function <anonymous>"
var p14969 = &p14969Var
var p14976Var = "thunk from <function <anonymous>>"
var p14976 = &p14976Var
var p14981Var = "thunk from <thunk <left> from <function <quickSort>>>"
var p14981 = &p14981Var
var p14993Var = "thunk <right> from <function <quickSort>>"
var p14993 = &p14993Var
var p14999Var = "function <anonymous>"
var p14999 = &p14999Var
var p15006Var = "thunk from <function <anonymous>>"
var p15006 = &p15006Var
var p15011Var = "thunk from <thunk <right> from <function <quickSort>>>"
var p15011 = &p15011Var
var p15022Var = "thunk from <function <quickSort>>"
var p15022 = &p15022Var
var p15032Var = "thunk from <function <quickSort>>"
var p15032 = &p15032Var
var p15044Var = "thunk from <function <quickSort>>"
var p15044 = &p15044Var
var p15069Var = "thunk <quickSort> from <function <anonymous>>"
var p15069 = &p15069Var
var p15081Var = "thunk <la> from <function <merge>>"
var p15081 = &p15081Var
var p15085Var = "thunk from <thunk <la> from <function <merge>>>"
var p15085 = &p15085Var
var p15093Var = "thunk <lb> from <function <merge>>"
var p15093 = &p15093Var
var p15097Var = "thunk from <thunk <lb> from <function <merge>>>"
var p15097 = &p15097Var
var p15106Var = "function <aux>"
var p15106 = &p15106Var
var p15164Var = "thunk from <function <aux>>"
var p15164 = &p15164Var
var p15176Var = "thunk from <function <aux>>"
var p15176 = &p15176Var
var p15190Var = "thunk from <function <aux>>"
var p15190 = &p15190Var
var p15201Var = "thunk from <thunk from <function <aux>>>"
var p15201 = &p15201Var
var p15217Var = "thunk from <function <aux>>"
var p15217 = &p15217Var
var p15229Var = "thunk from <thunk from <function <aux>>>"
var p15229 = &p15229Var
var p15248Var = "thunk <aux> from <function <merge>>"
var p15248 = &p15248Var
var p15253Var = "function <merge>"
var p15253 = &p15253Var
var p15257Var = "thunk from <function <merge>>"
var p15257 = &p15257Var
var p15266Var = "thunk <merge> from <function <anonymous>>"
var p15266 = &p15266Var
var p15275Var = "thunk <l> from <function <anonymous>>"
var p15275 = &p15275Var
var p15279Var = "thunk from <thunk <l> from <function <anonymous>>>"
var p15279 = &p15279Var
var p15285Var = "function <anonymous>"
var p15285 = &p15285Var
var p15294Var = "thunk from <function <anonymous>>"
var p15294 = &p15294Var
var p15304Var = "thunk from <function <anonymous>>"
var p15304 = &p15304Var
var p15317Var = "thunk <mid> from <function <anonymous>>"
var p15317 = &p15317Var
var p15322Var = "thunk from <thunk <mid> from <function <anonymous>>>"
var p15322 = &p15322Var
var p15337Var = "thunk <left> from <function <anonymous>>"
var p15337 = &p15337Var
var p15352Var = "thunk <right> from <function <anonymous>>"
var p15352 = &p15352Var
var p15369Var = "thunk from <function <anonymous>>"
var p15369 = &p15369Var
var p15373Var = "thunk from <thunk from <function <anonymous>>>"
var p15373 = &p15373Var
var p15387Var = "thunk from <thunk from <function <anonymous>>>"
var p15387 = &p15387Var
var p15419Var = "function <f>"
var p15419 = &p15419Var
var p15428Var = "thunk from <function <f>>"
var p15428 = &p15428Var
var p15435Var = "thunk from <function <f>>"
var p15435 = &p15435Var
var p15446Var = "thunk from <function <f>>"
var p15446 = &p15446Var
var p15455Var = "thunk from <function <f>>"
var p15455 = &p15455Var
var p15467Var = "thunk from <thunk from <function <f>>>"
var p15467 = &p15467Var
var p15481Var = "thunk from <function <f>>"
var p15481 = &p15481Var
var p15494Var = "thunk <f> from <function <anonymous>>"
var p15494 = &p15494Var
var p15502Var = "function <anonymous>"
var p15502 = &p15502Var
var p15506Var = "thunk from <function <anonymous>>"
var p15506 = &p15506Var
var p15526Var = "function <anonymous>"
var p15526 = &p15526Var
var p15534Var = "thunk from <function <anonymous>>"
var p15534 = &p15534Var
var p15538Var = "thunk from <thunk from <function <anonymous>>>"
var p15538 = &p15538Var
var p15554Var = "function <anonymous>"
var p15554 = &p15554Var
var p15569Var = "thunk from <function <anonymous>>"
var p15569 = &p15569Var
var p15575Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p15575 = &p15575Var
var p15577Var = "thunk from <thunk from <function <anonymous>>>"
var p15577 = &p15577Var
var p15602Var = "function <aux>"
var p15602 = &p15602Var
var p15606Var = "thunk from <function <aux>>"
var p15606 = &p15606Var
var p15641Var = "thunk from <function <aux>>"
var p15641 = &p15641Var
var p15670Var = "thunk <ak> from <function <aux>>"
var p15670 = &p15670Var
var p15675Var = "thunk from <thunk <ak> from <function <aux>>>"
var p15675 = &p15675Var
var p15685Var = "thunk <bk> from <function <aux>>"
var p15685 = &p15685Var
var p15690Var = "thunk from <thunk <bk> from <function <aux>>>"
var p15690 = &p15690Var
var p15709Var = "thunk from <function <aux>>"
var p15709 = &p15709Var
var p15728Var = "thunk from <thunk from <function <aux>>>"
var p15728 = &p15728Var
var p15751Var = "thunk from <function <aux>>"
var p15751 = &p15751Var
var p15767Var = "thunk from <thunk from <function <aux>>>"
var p15767 = &p15767Var
var p15783Var = "thunk from <function <aux>>"
var p15783 = &p15783Var
var p15799Var = "thunk from <thunk from <function <aux>>>"
var p15799 = &p15799Var
var p15824Var = "thunk <aux> from <function <anonymous>>"
var p15824 = &p15824Var
var p15829Var = "function <anonymous>"
var p15829 = &p15829Var
var p15833Var = "thunk from <function <anonymous>>"
var p15833 = &p15833Var
var p15861Var = "function <aux>"
var p15861 = &p15861Var
var p15865Var = "thunk from <function <aux>>"
var p15865 = &p15865Var
var p15880Var = "thunk from <function <aux>>"
var p15880 = &p15880Var
var p15898Var = "thunk from <function <aux>>"
var p15898 = &p15898Var
var p15910Var = "thunk from <function <aux>>"
var p15910 = &p15910Var
var p15923Var = "thunk from <function <aux>>"
var p15923 = &p15923Var
var p15942Var = "thunk from <thunk from <function <aux>>>"
var p15942 = &p15942Var
var p15960Var = "thunk from <function <aux>>"
var p15960 = &p15960Var
var p15972Var = "thunk from <function <aux>>"
var p15972 = &p15972Var
var p15985Var = "thunk from <function <aux>>"
var p15985 = &p15985Var
var p16005Var = "thunk from <function <aux>>"
var p16005 = &p16005Var
var p16028Var = "thunk <aux> from <function <anonymous>>"
var p16028 = &p16028Var
var p16033Var = "function <anonymous>"
var p16033 = &p16033Var
var p16037Var = "thunk from <function <anonymous>>"
var p16037 = &p16037Var
var p16063Var = "function <aux>"
var p16063 = &p16063Var
var p16067Var = "thunk from <function <aux>>"
var p16067 = &p16067Var
var p16086Var = "thunk from <function <aux>>"
var p16086 = &p16086Var
var p16119Var = "thunk from <function <aux>>"
var p16119 = &p16119Var
var p16131Var = "thunk from <function <aux>>"
var p16131 = &p16131Var
var p16144Var = "thunk from <function <aux>>"
var p16144 = &p16144Var
var p16169Var = "thunk from <function <aux>>"
var p16169 = &p16169Var
var p16181Var = "thunk from <function <aux>>"
var p16181 = &p16181Var
var p16194Var = "thunk from <function <aux>>"
var p16194 = &p16194Var
var p16210Var = "thunk from <thunk from <function <aux>>>"
var p16210 = &p16210Var
var p16226Var = "thunk from <function <aux>>"
var p16226 = &p16226Var
var p16251Var = "thunk <aux> from <function <anonymous>>"
var p16251 = &p16251Var
var p16256Var = "function <anonymous>"
var p16256 = &p16256Var
var p16260Var = "thunk from <function <anonymous>>"
var p16260 = &p16260Var
var p16282Var = "function <anonymous>"
var p16282 = &p16282Var
var p16286Var = "thunk from <function <anonymous>>"
var p16286 = &p16286Var
var p16297Var = "thunk <target_object> from <function <anonymous>>"
var p16297 = &p16297Var
var p16301Var = "thunk from <thunk <target_object> from <function <anonymous>>>"
var p16301 = &p16301Var
var p16317Var = "thunk <target_fields> from <function <anonymous>>"
var p16317 = &p16317Var
var p16321Var = "thunk from <thunk <target_fields> from <function <anonymous>>>"
var p16321 = &p16321Var
var p16332Var = "thunk from <thunk <target_fields> from <function <anonymous>>>"
var p16332 = &p16332Var
var p16351Var = "thunk <null_fields> from <function <anonymous>>"
var p16351 = &p16351Var
var p16362Var = "thunk from <thunk <null_fields> from <function <anonymous>>>"
var p16362 = &p16362Var
var p16377Var = "thunk from <thunk <null_fields> from <function <anonymous>>>"
var p16377 = &p16377Var
var p16388Var = "thunk <both_fields> from <function <anonymous>>"
var p16388 = &p16388Var
var p16392Var = "thunk from <thunk <both_fields> from <function <anonymous>>>"
var p16392 = &p16392Var
var p16402Var = "thunk from <thunk from <thunk <both_fields> from <function <anonymous>>>>"
var p16402 = &p16402Var
var p16434Var = "object <anonymous>"
var p16434 = &p16434Var
var p16438Var = "thunk from <object <anonymous>>"
var p16438 = &p16438Var
var p16461Var = "thunk from <object <anonymous>>"
var p16461 = &p16461Var
var p16476Var = "thunk from <object <anonymous>>"
var p16476 = &p16476Var
var p16494Var = "thunk from <object <anonymous>>"
var p16494 = &p16494Var
var p16524Var = "thunk from <function <anonymous>>"
var p16524 = &p16524Var
var p16555Var = "function <anonymous>"
var p16555 = &p16555Var
var p16559Var = "thunk from <function <anonymous>>"
var p16559 = &p16559Var
var p16588Var = "function <anonymous>"
var p16588 = &p16588Var
var p16592Var = "thunk from <function <anonymous>>"
var p16592 = &p16592Var
var p16606Var = "function <anonymous>"
var p16606 = &p16606Var
var p16610Var = "thunk from <function <anonymous>>"
var p16610 = &p16610Var
var p16624Var = "function <anonymous>"
var p16624 = &p16624Var
var p16628Var = "thunk from <function <anonymous>>"
var p16628 = &p16628Var
var p16644Var = "function <anonymous>"
var p16644 = &p16644Var
var p16648Var = "thunk from <function <anonymous>>"
var p16648 = &p16648Var
var p16670Var = "thunk from <function <anonymous>>"
var p16670 = &p16670Var
var p16683Var = "function <anonymous>"
var p16683 = &p16683Var
var p16687Var = "thunk from <function <anonymous>>"
var p16687 = &p16687Var
var p16707Var = "thunk from <function <anonymous>>"
var p16707 = &p16707Var
var p16720Var = "function <anonymous>"
var p16720 = &p16720Var
var p16724Var = "thunk from <function <anonymous>>"
var p16724 = &p16724Var
var p16746Var = "object <anonymous>"
var p16746 = &p16746Var
var p16755Var = "thunk from <function <anonymous>>"
var p16755 = &p16755Var
var p16765Var = "function <anonymous>"
var p16765 = &p16765Var
var p16769Var = "thunk from <function <anonymous>>"
var p16769 = &p16769Var
var p16791Var = "object <anonymous>"
var p16791 = &p16791Var
var p16800Var = "thunk from <function <anonymous>>"
var p16800 = &p16800Var
var p16810Var = "function <anonymous>"
var p16810 = &p16810Var
var p16814Var = "thunk from <function <anonymous>>"
var p16814 = &p16814Var
var p16829Var = "thunk <ta> from <function <anonymous>>"
var p16829 = &p16829Var
var p16833Var = "thunk from <thunk <ta> from <function <anonymous>>>"
var p16833 = &p16833Var
var p16843Var = "thunk <tb> from <function <anonymous>>"
var p16843 = &p16843Var
var p16847Var = "thunk from <thunk <tb> from <function <anonymous>>>"
var p16847 = &p16847Var
var p16857Var = "function <anonymous>"
var p16857 = &p16857Var
var p16861Var = "thunk from <function <anonymous>>"
var p16861 = &p16861Var
var p16878Var = "thunk from <function <anonymous>>"
var p16878 = &p16878Var
var p16889Var = "thunk <la> from <function <anonymous>>"
var p16889 = &p16889Var
var p16893Var = "thunk from <thunk <la> from <function <anonymous>>>"
var p16893 = &p16893Var
var p16906Var = "thunk from <function <anonymous>>"
var p16906 = &p16906Var
var p16916Var = "thunk from <thunk from <function <anonymous>>>"
var p16916 = &p16916Var
var p16929Var = "function <aux>"
var p16929 = &p16929Var
var p16959Var = "thunk from <function <aux>>"
var p16959 = &p16959Var
var p16975Var = "thunk <aux> from <function <anonymous>>"
var p16975 = &p16975Var
var p16983Var = "thunk from <function <anonymous>>"
var p16983 = &p16983Var
var p17005Var = "thunk from <function <anonymous>>"
var p17005 = &p17005Var
var p17016Var = "thunk <fields> from <function <anonymous>>"
var p17016 = &p17016Var
var p17020Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p17020 = &p17020Var
var p17030Var = "thunk <lfields> from <function <anonymous>>"
var p17030 = &p17030Var
var p17034Var = "thunk from <thunk <lfields> from <function <anonymous>>>"
var p17034 = &p17034Var
var p17047Var = "thunk from <function <anonymous>>"
var p17047 = &p17047Var
var p17061Var = "function <aux>"
var p17061 = &p17061Var
var p17073Var = "thunk <f> from <function <aux>>"
var p17073 = &p17073Var
var p17101Var = "thunk from <function <aux>>"
var p17101 = &p17101Var
var p17117Var = "thunk <aux> from <function <anonymous>>"
var p17117 = &p17117Var
var p17125Var = "thunk from <function <anonymous>>"
var p17125 = &p17125Var
var p17149Var = "thunk from <function <anonymous>>"
var p17149 = &p17149Var
var p17177Var = "thunk <arr> from <function <anonymous>>"
var p17177 = &p17177Var
var p17181Var = "thunk from <thunk <arr> from <function <anonymous>>>"
var p17181 = &p17181Var
var p17191Var = "function <anonymous>"
var p17191 = &p17191Var
var p17195Var = "thunk from <function <anonymous>>"
var p17195 = &p17195Var
var p17200Var = "thunk from <thunk from <function <anonymous>>>"
var p17200 = &p17200Var
var p17212Var = "thunk from <thunk from <function <anonymous>>>"
var p17212 = &p17212Var
var p17221Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p17221 = &p17221Var
var p17228Var = "function <anonymous>"
var p17228 = &p17228Var
var p17250Var = "function <isContent>"
var p17250 = &p17250Var
var p17265Var = "thunk from <function <isContent>>"
var p17265 = &p17265Var
var p17279Var = "thunk from <function <isContent>>"
var p17279 = &p17279Var
var p17292Var = "thunk from <function <isContent>>"
var p17292 = &p17292Var
var p17306Var = "thunk from <function <isContent>>"
var p17306 = &p17306Var
var p17320Var = "thunk <isContent> from <function <anonymous>>"
var p17320 = &p17320Var
var p17328Var = "function <anonymous>"
var p17328 = &p17328Var
var p17332Var = "thunk from <function <anonymous>>"
var p17332 = &p17332Var
var p17353Var = "thunk from <function <anonymous>>"
var p17353 = &p17353Var
var p17357Var = "thunk from <thunk from <function <anonymous>>>"
var p17357 = &p17357Var
var p17368Var = "thunk from <function <anonymous>>"
var p17368 = &p17368Var
var p17372Var = "thunk from <thunk from <function <anonymous>>>"
var p17372 = &p17372Var
var p17392Var = "thunk from <function <anonymous>>"
var p17392 = &p17392Var
var p17420Var = "thunk from <function <anonymous>>"
var p17420 = &p17420Var
var p17425Var = "thunk from <thunk from <function <anonymous>>>"
var p17425 = &p17425Var
var p17443Var = "object <anonymous>"
var p17443 = &p17443Var
var p17448Var = "thunk from <object <anonymous>>"
var p17448 = &p17448Var
var p17468Var = "thunk from <function <anonymous>>"
var p17468 = &p17468Var
var p17493Var = "function <anonymous>"
var p17493 = &p17493Var
var p17497Var = "thunk from <function <anonymous>>"
var p17497 = &p17497Var
var p17511Var = "thunk from <function <anonymous>>"
var p17511 = &p17511Var
var p17528Var = "thunk from <function <anonymous>>"
var p17528 = &p17528Var
var p17542Var = "thunk from <function <anonymous>>"
var p17542 = &p17542Var
var p17556Var = "thunk <pat_len> from <function <anonymous>>"
var p17556 = &p17556Var
var p17560Var = "thunk from <thunk <pat_len> from <function <anonymous>>>"
var p17560 = &p17560Var
var p17570Var = "thunk <str_len> from <function <anonymous>>"
var p17570 = &p17570Var
var p17574Var = "thunk from <thunk <str_len> from <function <anonymous>>>"
var p17574 = &p17574Var
var p17611Var = "function <anonymous>"
var p17611 = &p17611Var
var p17634Var = "thunk from <function <anonymous>>"
var p17634 = &p17634Var
var p17644Var = "thunk from <thunk from <function <anonymous>>>"
var p17644 = &p17644Var
var p17676Var = "function <anonymous>"
var p17676 = &p17676Var
var p17680Var = "thunk from <function <anonymous>>"
var p17680 = &p17680Var
var p17694Var = "thunk from <function <anonymous>>"
var p17694 = &p17694Var
var p17712Var = "function <anonymous>"
var p17712 = &p17712Var
var p17722Var = "thunk from <function <anonymous>>"
var p17722 = &p17722Var
var p17732Var = "thunk from <thunk from <function <anonymous>>>"
var p17732 = &p17732Var
var p17743Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p17743 = &p17743Var
var p17762Var = "function <anonymous>"
var p17762 = &p17762Var
var p17766Var = "thunk from <function <anonymous>>"
var p17766 = &p17766Var
var p17776Var = "thunk <arrLen> from <function <anonymous>>"
var p17776 = &p17776Var
var p17780Var = "thunk from <thunk <arrLen> from <function <anonymous>>>"
var p17780 = &p17780Var
var p17789Var = "function <aux>"
var p17789 = &p17789Var
var p17800Var = "thunk <e> from <function <aux>>"
var p17800 = &p17800Var
var p17814Var = "thunk from <function <aux>>"
var p17814 = &p17814Var
var p17831Var = "thunk from <function <aux>>"
var p17831 = &p17831Var
var p17848Var = "thunk from <function <aux>>"
var p17848 = &p17848Var
var p17859Var = "thunk from <thunk from <function <aux>>>"
var p17859 = &p17859Var
var p17871Var = "thunk <aux> from <function <anonymous>>"
var p17871 = &p17871Var
var p17879Var = "thunk from <function <anonymous>>"
var p17879 = &p17879Var
var p17895Var = "thunk from <function <anonymous>>"
var p17895 = &p17895Var
var p17912Var = "function <anonymous>"
var p17912 = &p17912Var
var p17916Var = "thunk from <function <anonymous>>"
var p17916 = &p17916Var
var p17926Var = "thunk <arrLen> from <function <anonymous>>"
var p17926 = &p17926Var
var p17930Var = "thunk from <thunk <arrLen> from <function <anonymous>>>"
var p17930 = &p17930Var
var p17939Var = "function <aux>"
var p17939 = &p17939Var
var p17950Var = "thunk <e> from <function <aux>>"
var p17950 = &p17950Var
var p17964Var = "thunk from <function <aux>>"
var p17964 = &p17964Var
var p17979Var = "thunk from <function <aux>>"
var p17979 = &p17979Var
var p17996Var = "thunk from <function <aux>>"
var p17996 = &p17996Var
var p18007Var = "thunk from <thunk from <function <aux>>>"
var p18007 = &p18007Var
var p18019Var = "thunk <aux> from <function <anonymous>>"
var p18019 = &p18019Var
var p18027Var = "thunk from <function <anonymous>>"
var p18027 = &p18027Var
var p18043Var = "thunk from <function <anonymous>>"
var p18043 = &p18043Var
var p18061Var = "thunk <t1> from <function <anonymous>>"
var p18061 = &p18061Var
var p18065Var = "thunk from <thunk <t1> from <function <anonymous>>>"
var p18065 = &p18065Var
var p18073Var = "thunk <t2> from <function <anonymous>>"
var p18073 = &p18073Var
var p18077Var = "thunk from <thunk <t2> from <function <anonymous>>>"
var p18077 = &p18077Var
var p18083Var = "function <anonymous>"
var p18083 = &p18083Var
var p18118Var = "thunk from <function <anonymous>>"
var p18118 = &p18118Var
var p18196Var = "thunk <len1> from <function <anonymous>>"
var p18196 = &p18196Var
var p18200Var = "thunk from <thunk <len1> from <function <anonymous>>>"
var p18200 = &p18200Var
var p18208Var = "thunk <len2> from <function <anonymous>>"
var p18208 = &p18208Var
var p18212Var = "thunk from <thunk <len2> from <function <anonymous>>>"
var p18212 = &p18212Var
var p18222Var = "thunk <minLen> from <function <anonymous>>"
var p18222 = &p18222Var
var p18226Var = "thunk from <thunk <minLen> from <function <anonymous>>>"
var p18226 = &p18226Var
var p18237Var = "function <aux>"
var p18237 = &p18237Var
var p18249Var = "thunk <cmpRes> from <function <aux>>"
var p18249 = &p18249Var
var p18254Var = "thunk from <thunk <cmpRes> from <function <aux>>>"
var p18254 = &p18254Var
var p18282Var = "thunk from <function <aux>>"
var p18282 = &p18282Var
var p18301Var = "thunk from <function <aux>>"
var p18301 = &p18301Var
var p18310Var = "thunk <aux> from <function <anonymous>>"
var p18310 = &p18310Var
var p18315Var = "function <anonymous>"
var p18315 = &p18315Var
var p18319Var = "thunk from <function <anonymous>>"
var p18319 = &p18319Var
var p18334Var = "function <anonymous>"
var p18334 = &p18334Var
var p18343Var = "thunk from <function <anonymous>>"
var p18343 = &p18343Var
var p18355Var = "function <anonymous>"
var p18355 = &p18355Var
var p18364Var = "thunk from <function <anonymous>>"
var p18364 = &p18364Var
var p18376Var = "function <anonymous>"
var p18376 = &p18376Var
var p18385Var = "thunk from <function <anonymous>>"
var p18385 = &p18385Var
var p18397Var = "function <anonymous>"
var p18397 = &p18397Var
var p18406Var = "thunk from <function <anonymous>>"
var p18406 = &p18406Var
var p18421Var = "function <anonymous>"
var p18421 = &p18421Var
var p18427Var = "function <anonymous>"
var p18427 = &p18427Var
var p18433Var = "thunk from <function <anonymous>>"
var p18433 = &p18433Var
var p18444Var = "function <anonymous>"
var p18444 = &p18444Var
var p18454Var = "function <anonymous>"
var p18454 = &p18454Var
var p18467Var = "function <anonymous>"
var p18467 = &p18467Var
var p18472Var = "thunk from <function <anonymous>>"
var p18472 = &p18472Var
var p18483Var = "function <anonymous>"
var p18483 = &p18483Var
var p18492Var = "thunk from <function <anonymous>>"
var p18492 = &p18492Var
var p18500Var = "object <anonymous>"
var p18500 = &p18500Var
var p18503Var = "function <anonymous>"
var p18503 = &p18503Var
var p18506Var = "object <anonymous>"
var p18506 = &p18506Var
var p18512Var = "function <anonymous>"
var p18512 = &p18512Var
var p18514Var = "function <anonymous>"
var p18514 = &p18514Var
var p18516Var = "function <anonymous>"
var p18516 = &p18516Var
var p18520Var = "function <anonymous>"
var p18520 = &p18520Var
var p18522Var = "function <anonymous>"
var p18522 = &p18522Var
var p18524Var = "function <anonymous>"
var p18524 = &p18524Var
var p18526Var = "function <anonymous>"
var p18526 = &p18526Var
var p18535Var = "thunk <zero_code> from <function <anonymous>>"
var p18535 = &p18535Var
var p18539Var = "thunk from <thunk <zero_code> from <function <anonymous>>>"
var p18539 = &p18539Var
var p18540Var = "thunk <zero_code> from <function <anonymous>>"
var p18540 = &p18540Var
var p18549Var = "thunk <upper_a_code> from <function <anonymous>>"
var p18549 = &p18549Var
var p18553Var = "thunk from <thunk <upper_a_code> from <function <anonymous>>>"
var p18553 = &p18553Var
var p18554Var = "thunk <upper_a_code> from <function <anonymous>>"
var p18554 = &p18554Var
var p18563Var = "thunk <lower_a_code> from <function <anonymous>>"
var p18563 = &p18563Var
var p18567Var = "thunk from <thunk <lower_a_code> from <function <anonymous>>>"
var p18567 = &p18567Var
var p18568Var = "thunk <lower_a_code> from <function <anonymous>>"
var p18568 = &p18568Var
var p18580Var = "thunk <code> from <function <addDigit>>"
var p18580 = &p18580Var
var p18584Var = "thunk from <thunk <code> from <function <addDigit>>>"
var p18584 = &p18584Var
var p18586Var = "thunk <code> from <function <addDigit>>"
var p18586 = &p18586Var
var p18593Var = "thunk <digit> from <function <addDigit>>"
var p18593 = &p18593Var
var p18596Var = "thunk <digit> from <function <addDigit>>"
var p18596 = &p18596Var
var p18598Var = "thunk <digit> from <function <addDigit>>"
var p18598 = &p18598Var
var p18602Var = "thunk <digit> from <function <addDigit>>"
var p18602 = &p18602Var
var p18605Var = "thunk <digit> from <function <addDigit>>"
var p18605 = &p18605Var
var p18609Var = "thunk <digit> from <function <addDigit>>"
var p18609 = &p18609Var
var p18611Var = "thunk <digit> from <function <addDigit>>"
var p18611 = &p18611Var
var p18613Var = "thunk <digit> from <function <addDigit>>"
var p18613 = &p18613Var
var p18618Var = "thunk <digit> from <function <addDigit>>"
var p18618 = &p18618Var
var p18621Var = "thunk <digit> from <function <addDigit>>"
var p18621 = &p18621Var
var p18623Var = "thunk <digit> from <function <addDigit>>"
var p18623 = &p18623Var
var p18627Var = "thunk <digit> from <function <addDigit>>"
var p18627 = &p18627Var
var p18630Var = "thunk <digit> from <function <addDigit>>"
var p18630 = &p18630Var
var p18634Var = "thunk <digit> from <function <addDigit>>"
var p18634 = &p18634Var
var p18636Var = "thunk <digit> from <function <addDigit>>"
var p18636 = &p18636Var
var p18638Var = "thunk <digit> from <function <addDigit>>"
var p18638 = &p18638Var
var p18642Var = "thunk <digit> from <function <addDigit>>"
var p18642 = &p18642Var
var p18646Var = "thunk <digit> from <function <addDigit>>"
var p18646 = &p18646Var
var p18648Var = "thunk <digit> from <function <addDigit>>"
var p18648 = &p18648Var
var p18651Var = "thunk <digit> from <function <addDigit>>"
var p18651 = &p18651Var
var p18654Var = "thunk <digit> from <function <addDigit>>"
var p18654 = &p18654Var
var p18660Var = "function <addDigit>"
var p18660 = &p18660Var
var p18663Var = "function <addDigit>"
var p18663 = &p18663Var
var p18665Var = "function <addDigit>"
var p18665 = &p18665Var
var p18669Var = "function <addDigit>"
var p18669 = &p18669Var
var p18671Var = "function <addDigit>"
var p18671 = &p18671Var
var p18673Var = "function <addDigit>"
var p18673 = &p18673Var
var p18675Var = "function <addDigit>"
var p18675 = &p18675Var
var p18679Var = "function <addDigit>"
var p18679 = &p18679Var
var p18683Var = "function <addDigit>"
var p18683 = &p18683Var
var p18687Var = "function <addDigit>"
var p18687 = &p18687Var
var p18689Var = "function <addDigit>"
var p18689 = &p18689Var
var p18691Var = "function <addDigit>"
var p18691 = &p18691Var
var p18702Var = "function <addDigit>"
var p18702 = &p18702Var
var p18706Var = "thunk from <function <addDigit>>"
var p18706 = &p18706Var
var p18709Var = "thunk from <function <addDigit>>"
var p18709 = &p18709Var
var p18711Var = "function <addDigit>"
var p18711 = &p18711Var
var p18717Var = "function <addDigit>"
var p18717 = &p18717Var
var p18720Var = "function <addDigit>"
var p18720 = &p18720Var
var p18723Var = "thunk <addDigit> from <function <anonymous>>"
var p18723 = &p18723Var
var p18731Var = "function <anonymous>"
var p18731 = &p18731Var
var p18735Var = "thunk from <function <anonymous>>"
var p18735 = &p18735Var
var p18742Var = "thunk from <function <anonymous>>"
var p18742 = &p18742Var
var p18746Var = "thunk from <thunk from <function <anonymous>>>"
var p18746 = &p18746Var
var p18748Var = "thunk from <function <anonymous>>"
var p18748 = &p18748Var
var p18751Var = "thunk from <function <anonymous>>"
var p18751 = &p18751Var
var p18752Var = "function <anonymous>"
var p18752 = &p18752Var
var p18755Var = "function <anonymous>"
var p18755 = &p18755Var
var p18758Var = "function <anonymous>"
var p18758 = &p18758Var
var p18761Var = "function <anonymous>"
var p18761 = &p18761Var
var p18765Var = "function <anonymous>"
var p18765 = &p18765Var
var p18776Var = "function <anonymous>"
var p18776 = &p18776Var
var p18778Var = "function <anonymous>"
var p18778 = &p18778Var
var p18784Var = "object <anonymous>"
var p18784 = &p18784Var
var p18790Var = "object <anonymous>"
var p18790 = &p18790Var
var p18792Var = "object <anonymous>"
var p18792 = &p18792Var
var p18795Var = "object <anonymous>"
var p18795 = &p18795Var
var p18797Var = "object <anonymous>"
var p18797 = &p18797Var
var p18800Var = "object <anonymous>"
var p18800 = &p18800Var
var p18802Var = "object <anonymous>"
var p18802 = &p18802Var
var p18805Var = "object <anonymous>"
var p18805 = &p18805Var
var p18807Var = "object <anonymous>"
var p18807 = &p18807Var
var p18810Var = "object <anonymous>"
var p18810 = &p18810Var
var p18812Var = "object <anonymous>"
var p18812 = &p18812Var
var p18813Var = "object <anonymous>"
var p18813 = &p18813Var
var p18815Var = "object <anonymous>"
var p18815 = &p18815Var
var p18837Var = "object <anonymous>"
var p18837 = &p18837Var
var p18840Var = "object <anonymous>"
var p18840 = &p18840Var
var p18842Var = "object <anonymous>"
var p18842 = &p18842Var
var p18845Var = "object <anonymous>"
var p18845 = &p18845Var
var p18847Var = "object <anonymous>"
var p18847 = &p18847Var
var p18857Var = "object <anonymous>"
var p18857 = &p18857Var
var p18861Var = "thunk from <object <anonymous>>"
var p18861 = &p18861Var
var p18863Var = "thunk from <object <anonymous>>"
var p18863 = &p18863Var
var p18864Var = "object <anonymous>"
var p18864 = &p18864Var
var p18872Var = "$"
var p18872 = &p18872Var
var p8 = &ast.Source{
	DiagnosticFileName: "<std>",
	Lines: []string{
		"/*\n",
		"Copyright 2015 Google Inc. All rights reserved.\n",
		"\n",
		"Licensed under the Apache License, Version 2.0 (the \"License\");\n",
		"you may not use this file except in compliance with the License.\n",
		"You may obtain a copy of the License at\n",
		"\n",
		"    http://www.apache.org/licenses/LICENSE-2.0\n",
		"\n",
		"Unless required by applicable law or agreed to in writing, software\n",
		"distributed under the License is distributed on an \"AS IS\" BASIS,\n",
		"WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
		"See the License for the specific language governing permissions and\n",
		"limitations under the License.\n",
		"*/\n",
		"\n",
		"/* This is the Jsonnet standard library, at least the parts of it that are written in Jsonnet.\n",
		" *\n",
		" * There are some native methods as well, which are defined in the interpreter and added to this\n",
		" * file.  It is never necessary to import std.jsonnet, it is embedded into the interpreter at\n",
		" * compile-time and automatically imported into all other Jsonnet programs.\n",
		" */\n",
		"{\n",
		"\n",
		"  local std = self,\n",
		"  local id = function(x) x,\n",
		"\n",
		"  isString(v):: std.type(v) == 'string',\n",
		"  isNumber(v):: std.type(v) == 'number',\n",
		"  isBoolean(v):: std.type(v) == 'boolean',\n",
		"  isObject(v):: std.type(v) == 'object',\n",
		"  isArray(v):: std.type(v) == 'array',\n",
		"  isFunction(v):: std.type(v) == 'function',\n",
		"\n",
		"  toString(a)::\n",
		"    if std.type(a) == 'string' then a else '' + a,\n",
		"\n",
		"  substr(str, from, len)::\n",
		"    assert std.isString(str) : 'substr first parameter should be a string, got ' + std.type(str);\n",
		"    assert std.isNumber(from) : 'substr second parameter should be a string, got ' + std.type(from);\n",
		"    assert std.isNumber(len) : 'substr third parameter should be a string, got ' + std.type(len);\n",
		"    assert len >= 0 : 'substr third parameter should be greater than zero, got ' + len;\n",
		"    std.join('', std.makeArray(std.max(0, std.min(len, std.length(str) - from)), function(i) str[i + from])),\n",
		"\n",
		"  startsWith(a, b)::\n",
		"    if std.length(a) < std.length(b) then\n",
		"      false\n",
		"    else\n",
		"      std.substr(a, 0, std.length(b)) == b,\n",
		"\n",
		"  endsWith(a, b)::\n",
		"    if std.length(a) < std.length(b) then\n",
		"      false\n",
		"    else\n",
		"      std.substr(a, std.length(a) - std.length(b), std.length(b)) == b,\n",
		"\n",
		"  lstripChars(str, chars)::\n",
		"    if std.length(str) > 0 && std.member(chars, str[0]) then\n",
		"      std.lstripChars(str[1:], chars) tailstrict\n",
		"    else\n",
		"      str,\n",
		"\n",
		"  rstripChars(str, chars)::\n",
		"    local len = std.length(str);\n",
		"    if len > 0 && std.member(chars, str[len - 1]) then\n",
		"      std.rstripChars(str[:len - 1], chars) tailstrict\n",
		"    else\n",
		"      str,\n",
		"\n",
		"  stripChars(str, chars)::\n",
		"    std.lstripChars(std.rstripChars(str, chars), chars),\n",
		"\n",
		"  stringChars(str)::\n",
		"    std.makeArray(std.length(str), function(i) str[i]),\n",
		"\n",
		"  local parse_nat(str, base) =\n",
		"    assert base > 0 && base <= 16 : 'integer base %d invalid' % base;\n",
		"    // These codepoints are in ascending order:\n",
		"    local zero_code = std.codepoint('0');\n",
		"    local upper_a_code = std.codepoint('A');\n",
		"    local lower_a_code = std.codepoint('a');\n",
		"    local addDigit(aggregate, char) =\n",
		"      local code = std.codepoint(char);\n",
		"      local digit = if code >= lower_a_code then\n",
		"        code - lower_a_code + 10\n",
		"      else if code >= upper_a_code then\n",
		"        code - upper_a_code + 10\n",
		"      else\n",
		"        code - zero_code;\n",
		"      assert digit >= 0 && digit < base : '%s is not a base %d integer' % [str, base];\n",
		"      base * aggregate + digit;\n",
		"    std.foldl(addDigit, std.stringChars(str), 0),\n",
		"\n",
		"  parseInt(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 && str != '-' : 'Not an integer: \"%s\"' % [str];\n",
		"    if str[0] == '-' then\n",
		"      -parse_nat(str[1:], 10)\n",
		"    else\n",
		"      parse_nat(str, 10),\n",
		"\n",
		"  parseOctal(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 : 'Not an octal number: \"\"';\n",
		"    parse_nat(str, 8),\n",
		"\n",
		"  parseHex(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 : 'Not hexadecimal: \"\"';\n",
		"    parse_nat(str, 16),\n",
		"\n",
		"  split(str, c)::\n",
		"    assert std.isString(str) : 'std.split first parameter must be a String, got ' + std.type(str);\n",
		"    assert std.isString(c) : 'std.split second parameter must be a String, got ' + std.type(c);\n",
		"    assert std.length(c) >= 1 : 'std.split second parameter must have length 1 or greater, got ' + std.length(c);\n",
		"    std.splitLimit(str, c, -1),\n",
		"\n",
		"  splitLimit(str, c, maxsplits)::\n",
		"    assert std.isString(str) : 'str.splitLimit first parameter must be a String, got ' + std.type(str);\n",
		"    assert std.isString(c) : 'str.splitLimit second parameter must be a String, got ' + std.type(c);\n",
		"    assert std.length(c) >= 1 : 'std.splitLimit second parameter must have length 1 or greater, got ' + std.length(c);\n",
		"    assert std.isNumber(maxsplits) : 'str.splitLimit third parameter must be a Number, got ' + std.type(maxsplits);\n",
		"    local strLen = std.length(str);\n",
		"    local cLen = std.length(c);\n",
		"    local aux(idx, ret, val) =\n",
		"      if idx >= strLen then\n",
		"        ret + [val]\n",
		"      else if str[idx:idx + cLen:1] == c &&\n",
		"              (maxsplits == -1 || std.length(ret) < maxsplits) then\n",
		"        aux(idx + cLen, ret + [val], '')\n",
		"      else\n",
		"        aux(idx + 1, ret, val + str[idx]);\n",
		"    aux(0, [], ''),\n",
		"\n",
		"  splitLimitR(str, c, maxsplits)::\n",
		"    assert std.isString(str) : 'str.splitLimitR first parameter must be a String, got ' + std.type(str);\n",
		"    assert std.isString(c) : 'str.splitLimitR second parameter must be a String, got ' + std.type(c);\n",
		"    assert std.length(c) >= 1 : 'std.splitLimitR second parameter must have length 1 or greater, got ' + std.length(c);\n",
		"    assert std.isNumber(maxsplits) : 'str.splitLimitR third parameter must be a Number, got ' + std.type(maxsplits);\n",
		"    if maxsplits == -1 then\n",
		"      std.splitLimit(str, c, -1)\n",
		"    else\n",
		"      local revStr(str) = std.join('', std.reverse(std.stringChars(str)));\n",
		"      std.map(function(e) revStr(e), std.reverse(std.splitLimit(revStr(str), revStr(c), maxsplits))),\n",
		"\n",
		"  strReplace(str, from, to)::\n",
		"    assert std.isString(str);\n",
		"    assert std.isString(from);\n",
		"    assert std.isString(to);\n",
		"    assert from != '' : \"'from' string must not be zero length.\";\n",
		"\n",
		"    // Cache for performance.\n",
		"    local str_len = std.length(str);\n",
		"    local from_len = std.length(from);\n",
		"\n",
		"    // True if from is at str[i].\n",
		"    local found_at(i) = str[i:i + from_len] == from;\n",
		"\n",
		"    // Return the remainder of 'str' starting with 'start_index' where\n",
		"    // all occurrences of 'from' after 'curr_index' are replaced with 'to'.\n",
		"    local replace_after(start_index, curr_index, acc) =\n",
		"      if curr_index > str_len then\n",
		"        acc + str[start_index:curr_index]\n",
		"      else if found_at(curr_index) then\n",
		"        local new_index = curr_index + std.length(from);\n",
		"        replace_after(new_index, new_index, acc + str[start_index:curr_index] + to) tailstrict\n",
		"      else\n",
		"        replace_after(start_index, curr_index + 1, acc) tailstrict;\n",
		"\n",
		"    // if from_len==1, then we replace by splitting and rejoining the\n",
		"    // string which is much faster than recursing on replace_after\n",
		"    if from_len == 1 then\n",
		"      std.join(to, std.split(str, from))\n",
		"    else\n",
		"      replace_after(0, 0, ''),\n",
		"\n",
		"  asciiUpper(str)::\n",
		"    local cp = std.codepoint;\n",
		"    local up_letter(c) = if cp(c) >= 97 && cp(c) < 123 then\n",
		"      std.char(cp(c) - 32)\n",
		"    else\n",
		"      c;\n",
		"    std.join('', std.map(up_letter, std.stringChars(str))),\n",
		"\n",
		"  asciiLower(str)::\n",
		"    local cp = std.codepoint;\n",
		"    local down_letter(c) = if cp(c) >= 65 && cp(c) < 91 then\n",
		"      std.char(cp(c) + 32)\n",
		"    else\n",
		"      c;\n",
		"    std.join('', std.map(down_letter, std.stringChars(str))),\n",
		"\n",
		"  range(from, to)::\n",
		"    std.makeArray(to - from + 1, function(i) i + from),\n",
		"\n",
		"  repeat(what, count)::\n",
		"    local joiner =\n",
		"      if std.isString(what) then ''\n",
		"      else if std.isArray(what) then []\n",
		"      else error 'std.repeat first argument must be an array or a string';\n",
		"    std.join(joiner, std.makeArray(count, function(i) what)),\n",
		"\n",
		"  slice(indexable, index, end, step)::\n",
		"    local invar =\n",
		"      // loop invariant with defaults applied\n",
		"      {\n",
		"        indexable: indexable,\n",
		"        index:\n",
		"          if index == null then 0\n",
		"          else index,\n",
		"        end:\n",
		"          if end == null then std.length(indexable)\n",
		"          else end,\n",
		"        step:\n",
		"          if step == null then 1\n",
		"          else step,\n",
		"        length: std.length(indexable),\n",
		"        type: std.type(indexable),\n",
		"      };\n",
		"    assert invar.index >= 0 && invar.end >= 0 && invar.step >= 0 : 'got [%s:%s:%s] but negative index, end, and steps are not supported' % [invar.index, invar.end, invar.step];\n",
		"    assert step != 0 : 'got %s but step must be greater than 0' % step;\n",
		"    assert std.isString(indexable) || std.isArray(indexable) : 'std.slice accepts a string or an array, but got: %s' % std.type(indexable);\n",
		"    local build(slice, cur) =\n",
		"      if cur >= invar.end || cur >= invar.length then\n",
		"        slice\n",
		"      else\n",
		"        build(\n",
		"          if invar.type == 'string' then\n",
		"            slice + invar.indexable[cur]\n",
		"          else\n",
		"            slice + [invar.indexable[cur]],\n",
		"          cur + invar.step\n",
		"        ) tailstrict;\n",
		"    build(if invar.type == 'string' then '' else [], invar.index),\n",
		"\n",
		"  member(arr, x)::\n",
		"    if std.isArray(arr) then\n",
		"      std.count(arr, x) > 0\n",
		"    else if std.isString(arr) then\n",
		"      std.length(std.findSubstr(x, arr)) > 0\n",
		"    else error 'std.member first argument must be an array or a string',\n",
		"\n",
		"  count(arr, x):: std.length(std.filter(function(v) v == x, arr)),\n",
		"\n",
		"  mod(a, b)::\n",
		"    if std.isNumber(a) && std.isNumber(b) then\n",
		"      std.modulo(a, b)\n",
		"    else if std.isString(a) then\n",
		"      std.format(a, b)\n",
		"    else\n",
		"      error 'Operator % cannot be used on types ' + std.type(a) + ' and ' + std.type(b) + '.',\n",
		"\n",
		"  map(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.map first param must be function, got ' + std.type(func))\n",
		"    else if !std.isArray(arr) && !std.isString(arr) then\n",
		"      error ('std.map second param must be array / string, got ' + std.type(arr))\n",
		"    else\n",
		"      std.makeArray(std.length(arr), function(i) func(arr[i])),\n",
		"\n",
		"  mapWithIndex(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.mapWithIndex first param must be function, got ' + std.type(func))\n",
		"    else if !std.isArray(arr) && !std.isString(arr) then\n",
		"      error ('std.mapWithIndex second param must be array, got ' + std.type(arr))\n",
		"    else\n",
		"      std.makeArray(std.length(arr), function(i) func(i, arr[i])),\n",
		"\n",
		"  mapWithKey(func, obj)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.mapWithKey first param must be function, got ' + std.type(func))\n",
		"    else if !std.isObject(obj) then\n",
		"      error ('std.mapWithKey second param must be object, got ' + std.type(obj))\n",
		"    else\n",
		"      { [k]: func(k, obj[k]) for k in std.objectFields(obj) },\n",
		"\n",
		"  flatMap(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.flatMap first param must be function, got ' + std.type(func))\n",
		"    else if std.isArray(arr) then\n",
		"      std.flattenArrays(std.makeArray(std.length(arr), function(i) func(arr[i])))\n",
		"    else if std.isString(arr) then\n",
		"      std.join('', std.makeArray(std.length(arr), function(i) func(arr[i])))\n",
		"    else error ('std.flatMap second param must be array / string, got ' + std.type(arr)),\n",
		"\n",
		"  join(sep, arr)::\n",
		"    local aux(arr, i, first, running) =\n",
		"      if i >= std.length(arr) then\n",
		"        running\n",
		"      else if arr[i] == null then\n",
		"        aux(arr, i + 1, first, running) tailstrict\n",
		"      else if std.type(arr[i]) != std.type(sep) then\n",
		"        error 'expected %s but arr[%d] was %s ' % [std.type(sep), i, std.type(arr[i])]\n",
		"      else if first then\n",
		"        aux(arr, i + 1, false, running + arr[i]) tailstrict\n",
		"      else\n",
		"        aux(arr, i + 1, false, running + sep + arr[i]) tailstrict;\n",
		"    if !std.isArray(arr) then\n",
		"      error 'join second parameter should be array, got ' + std.type(arr)\n",
		"    else if std.isString(sep) then\n",
		"      aux(arr, 0, true, '')\n",
		"    else if std.isArray(sep) then\n",
		"      aux(arr, 0, true, [])\n",
		"    else\n",
		"      error 'join first parameter should be string or array, got ' + std.type(sep),\n",
		"\n",
		"  lines(arr)::\n",
		"    std.join('\\n', arr + ['']),\n",
		"\n",
		"  deepJoin(arr)::\n",
		"    if std.isString(arr) then\n",
		"      arr\n",
		"    else if std.isArray(arr) then\n",
		"      std.join('', [std.deepJoin(x) for x in arr])\n",
		"    else\n",
		"      error 'Expected string or array, got %s' % std.type(arr),\n",
		"\n",
		"\n",
		"  format(str, vals)::\n",
		"\n",
		"    /////////////////////////////\n",
		"    // Parse the mini-language //\n",
		"    /////////////////////////////\n",
		"\n",
		"    local try_parse_mapping_key(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == '(' then\n",
		"        local consume(str, j, v) =\n",
		"          if j >= std.length(str) then\n",
		"            error 'Truncated format code.'\n",
		"          else\n",
		"            local c = str[j];\n",
		"            if c != ')' then\n",
		"              consume(str, j + 1, v + c)\n",
		"            else\n",
		"              { i: j + 1, v: v };\n",
		"        consume(str, i + 1, '')\n",
		"      else\n",
		"        { i: i, v: null };\n",
		"\n",
		"    local try_parse_cflags(str, i) =\n",
		"      local consume(str, j, v) =\n",
		"        assert j < std.length(str) : 'Truncated format code.';\n",
		"        local c = str[j];\n",
		"        if c == '#' then\n",
		"          consume(str, j + 1, v { alt: true })\n",
		"        else if c == '0' then\n",
		"          consume(str, j + 1, v { zero: true })\n",
		"        else if c == '-' then\n",
		"          consume(str, j + 1, v { left: true })\n",
		"        else if c == ' ' then\n",
		"          consume(str, j + 1, v { blank: true })\n",
		"        else if c == '+' then\n",
		"          consume(str, j + 1, v { plus: true })\n",
		"        else\n",
		"          { i: j, v: v };\n",
		"      consume(str, i, { alt: false, zero: false, left: false, blank: false, plus: false });\n",
		"\n",
		"    local try_parse_field_width(str, i) =\n",
		"      if i < std.length(str) && str[i] == '*' then\n",
		"        { i: i + 1, v: '*' }\n",
		"      else\n",
		"        local consume(str, j, v) =\n",
		"          assert j < std.length(str) : 'Truncated format code.';\n",
		"          local c = str[j];\n",
		"          if c == '0' then\n",
		"            consume(str, j + 1, v * 10 + 0)\n",
		"          else if c == '1' then\n",
		"            consume(str, j + 1, v * 10 + 1)\n",
		"          else if c == '2' then\n",
		"            consume(str, j + 1, v * 10 + 2)\n",
		"          else if c == '3' then\n",
		"            consume(str, j + 1, v * 10 + 3)\n",
		"          else if c == '4' then\n",
		"            consume(str, j + 1, v * 10 + 4)\n",
		"          else if c == '5' then\n",
		"            consume(str, j + 1, v * 10 + 5)\n",
		"          else if c == '6' then\n",
		"            consume(str, j + 1, v * 10 + 6)\n",
		"          else if c == '7' then\n",
		"            consume(str, j + 1, v * 10 + 7)\n",
		"          else if c == '8' then\n",
		"            consume(str, j + 1, v * 10 + 8)\n",
		"          else if c == '9' then\n",
		"            consume(str, j + 1, v * 10 + 9)\n",
		"          else\n",
		"            { i: j, v: v };\n",
		"        consume(str, i, 0);\n",
		"\n",
		"    local try_parse_precision(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == '.' then\n",
		"        try_parse_field_width(str, i + 1)\n",
		"      else\n",
		"        { i: i, v: null };\n",
		"\n",
		"    // Ignored, if it exists.\n",
		"    local try_parse_length_modifier(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == 'h' || c == 'l' || c == 'L' then\n",
		"        i + 1\n",
		"      else\n",
		"        i;\n",
		"\n",
		"    local parse_conv_type(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == 'd' || c == 'i' || c == 'u' then\n",
		"        { i: i + 1, v: 'd', caps: false }\n",
		"      else if c == 'o' then\n",
		"        { i: i + 1, v: 'o', caps: false }\n",
		"      else if c == 'x' then\n",
		"        { i: i + 1, v: 'x', caps: false }\n",
		"      else if c == 'X' then\n",
		"        { i: i + 1, v: 'x', caps: true }\n",
		"      else if c == 'e' then\n",
		"        { i: i + 1, v: 'e', caps: false }\n",
		"      else if c == 'E' then\n",
		"        { i: i + 1, v: 'e', caps: true }\n",
		"      else if c == 'f' then\n",
		"        { i: i + 1, v: 'f', caps: false }\n",
		"      else if c == 'F' then\n",
		"        { i: i + 1, v: 'f', caps: true }\n",
		"      else if c == 'g' then\n",
		"        { i: i + 1, v: 'g', caps: false }\n",
		"      else if c == 'G' then\n",
		"        { i: i + 1, v: 'g', caps: true }\n",
		"      else if c == 'c' then\n",
		"        { i: i + 1, v: 'c', caps: false }\n",
		"      else if c == 's' then\n",
		"        { i: i + 1, v: 's', caps: false }\n",
		"      else if c == '%' then\n",
		"        { i: i + 1, v: '%', caps: false }\n",
		"      else\n",
		"        error 'Unrecognised conversion type: ' + c;\n",
		"\n",
		"\n",
		"    // Parsed initial %, now the rest.\n",
		"    local parse_code(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local mkey = try_parse_mapping_key(str, i);\n",
		"      local cflags = try_parse_cflags(str, mkey.i);\n",
		"      local fw = try_parse_field_width(str, cflags.i);\n",
		"      local prec = try_parse_precision(str, fw.i);\n",
		"      local len_mod = try_parse_length_modifier(str, prec.i);\n",
		"      local ctype = parse_conv_type(str, len_mod);\n",
		"      {\n",
		"        i: ctype.i,\n",
		"        code: {\n",
		"          mkey: mkey.v,\n",
		"          cflags: cflags.v,\n",
		"          fw: fw.v,\n",
		"          prec: prec.v,\n",
		"          ctype: ctype.v,\n",
		"          caps: ctype.caps,\n",
		"        },\n",
		"      };\n",
		"\n",
		"    // Parse a format string (containing none or more % format tags).\n",
		"    local parse_codes(str, i, out, cur) =\n",
		"      if i >= std.length(str) then\n",
		"        out + [cur]\n",
		"      else\n",
		"        local c = str[i];\n",
		"        if c == '%' then\n",
		"          local r = parse_code(str, i + 1);\n",
		"          parse_codes(str, r.i, out + [cur, r.code], '') tailstrict\n",
		"        else\n",
		"          parse_codes(str, i + 1, out, cur + c) tailstrict;\n",
		"\n",
		"    local codes = parse_codes(str, 0, [], '');\n",
		"\n",
		"\n",
		"    ///////////////////////\n",
		"    // Format the values //\n",
		"    ///////////////////////\n",
		"\n",
		"    // Useful utilities\n",
		"    local padding(w, s) =\n",
		"      local aux(w, v) =\n",
		"        if w <= 0 then\n",
		"          v\n",
		"        else\n",
		"          aux(w - 1, v + s);\n",
		"      aux(w, '');\n",
		"\n",
		"    // Add s to the left of str so that its length is at least w.\n",
		"    local pad_left(str, w, s) =\n",
		"      padding(w - std.length(str), s) + str;\n",
		"\n",
		"    // Add s to the right of str so that its length is at least w.\n",
		"    local pad_right(str, w, s) =\n",
		"      str + padding(w - std.length(str), s);\n",
		"\n",
		"    // Render a sign & magnitude integer (radix ranges from decimal to binary).\n",
		"    // neg should be a boolean, and when true indicates that we should render a negative number.\n",
		"    // mag must always be a whole number >= 0, it's the magnitude of the integer to render\n",
		"    // min_chars must be a whole number >= 0\n",
		"    //   It is the field width, i.e. std.length() of the result should be >= min_chars\n",
		"    // min_digits must be a whole number >= 0. It's the number of zeroes to pad with.\n",
		"    // blank must be a boolean, if true adds an additional ' ' in front of a positive number, so\n",
		"    // that it is aligned with negative numbers with the same number of digits.\n",
		"    // plus must be a boolean, if true adds a '+' in front of a positive number, so that it is\n",
		"    // aligned with negative numbers with the same number of digits.  This takes precedence over\n",
		"    // blank, if both are true.\n",
		"    // radix must be a whole number >1 and <= 10.  It is the base of the system of numerals.\n",
		"    // zero_prefix is a string prefixed before the sign to all numbers that are not 0.\n",
		"    local render_int(neg, mag, min_chars, min_digits, blank, plus, radix, zero_prefix) =\n",
		"      // dec is the minimal string needed to represent the number as text.\n",
		"      local dec =\n",
		"        if mag == 0 then\n",
		"          '0'\n",
		"        else\n",
		"          local aux(n) =\n",
		"            if n == 0 then\n",
		"              zero_prefix\n",
		"            else\n",
		"              aux(std.floor(n / radix)) + (n % radix);\n",
		"          aux(mag);\n",
		"      local zp = min_chars - (if neg || blank || plus then 1 else 0);\n",
		"      local zp2 = std.max(zp, min_digits);\n",
		"      local dec2 = pad_left(dec, zp2, '0');\n",
		"      (if neg then '-' else if plus then '+' else if blank then ' ' else '') + dec2;\n",
		"\n",
		"    // Render an integer in hexadecimal.\n",
		"    local render_hex(n__, min_chars, min_digits, blank, plus, add_zerox, capitals) =\n",
		"      local numerals = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]\n",
		"                       + if capitals then ['A', 'B', 'C', 'D', 'E', 'F']\n",
		"                       else ['a', 'b', 'c', 'd', 'e', 'f'];\n",
		"      local n_ = std.abs(n__);\n",
		"      local aux(n) =\n",
		"        if n == 0 then\n",
		"          ''\n",
		"        else\n",
		"          aux(std.floor(n / 16)) + numerals[n % 16];\n",
		"      local hex = if std.floor(n_) == 0 then '0' else aux(std.floor(n_));\n",
		"      local neg = n__ < 0;\n",
		"      local zp = min_chars - (if neg || blank || plus then 1 else 0)\n",
		"                 - (if add_zerox then 2 else 0);\n",
		"      local zp2 = std.max(zp, min_digits);\n",
		"      local hex2 = (if add_zerox then (if capitals then '0X' else '0x') else '')\n",
		"                   + pad_left(hex, zp2, '0');\n",
		"      (if neg then '-' else if plus then '+' else if blank then ' ' else '') + hex2;\n",
		"\n",
		"    local strip_trailing_zero(str) =\n",
		"      local aux(str, i) =\n",
		"        if i < 0 then\n",
		"          ''\n",
		"        else\n",
		"          if str[i] == '0' then\n",
		"            aux(str, i - 1)\n",
		"          else\n",
		"            std.substr(str, 0, i + 1);\n",
		"      aux(str, std.length(str) - 1);\n",
		"\n",
		"    // Render floating point in decimal form\n",
		"    local render_float_dec(n__, zero_pad, blank, plus, ensure_pt, trailing, prec) =\n",
		"      local n_ = std.abs(n__);\n",
		"      local whole = std.floor(n_);\n",
		"      // Represent the rounded number as an integer * 1/10**prec.\n",
		"      // Note that it can also be equal to 10**prec and we'll need to carry\n",
		"      // over to the wholes.  We operate on the absolute numbers, so that we\n",
		"      // don't have trouble with the rounding direction.\n",
		"      local denominator = std.pow(10, prec);\n",
		"      local numerator = std.abs(n_) * denominator + 0.5;\n",
		"      local whole = std.sign(n_) * std.floor(numerator / denominator);\n",
		"      local frac = std.floor(numerator) % denominator;\n",
		"      local dot_size = if prec == 0 && !ensure_pt then 0 else 1;\n",
		"      local zp = zero_pad - prec - dot_size;\n",
		"      local str = render_int(n__ < 0, whole, zp, 0, blank, plus, 10, '');\n",
		"      if prec == 0 then\n",
		"        str + if ensure_pt then '.' else ''\n",
		"      else\n",
		"        if trailing || frac > 0 then\n",
		"          local frac_str = render_int(false, frac, prec, 0, false, false, 10, '');\n",
		"          str + '.' + if !trailing then strip_trailing_zero(frac_str) else frac_str\n",
		"        else\n",
		"          str;\n",
		"\n",
		"    // Render floating point in scientific form\n",
		"    local render_float_sci(n__, zero_pad, blank, plus, ensure_pt, trailing, caps, prec) =\n",
		"      local exponent = if n__ == 0 then 0 else std.floor(std.log(std.abs(n__)) / std.log(10));\n",
		"      local suff = (if caps then 'E' else 'e')\n",
		"                   + render_int(exponent < 0, std.abs(exponent), 3, 0, false, true, 10, '');\n",
		"      local mantissa = if exponent == -324 then\n",
		"        // Avoid a rounding error where std.pow(10, -324) is 0\n",
		"        // -324 is the smallest exponent possible.\n",
		"        n__ * 10 / std.pow(10, exponent + 1)\n",
		"      else\n",
		"        n__ / std.pow(10, exponent);\n",
		"      local zp2 = zero_pad - std.length(suff);\n",
		"      render_float_dec(mantissa, zp2, blank, plus, ensure_pt, trailing, prec) + suff;\n",
		"\n",
		"    // Render a value with an arbitrary format code.\n",
		"    local format_code(val, code, fw, prec_or_null, i) =\n",
		"      local cflags = code.cflags;\n",
		"      local fpprec = if prec_or_null != null then prec_or_null else 6;\n",
		"      local iprec = if prec_or_null != null then prec_or_null else 0;\n",
		"      local zp = if cflags.zero && !cflags.left then fw else 0;\n",
		"      if code.ctype == 's' then\n",
		"        std.toString(val)\n",
		"      else if code.ctype == 'd' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_int(val <= -1, std.floor(std.abs(val)), zp, iprec, cflags.blank, cflags.plus, 10, '')\n",
		"      else if code.ctype == 'o' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          local zero_prefix = if cflags.alt then '0' else '';\n",
		"          render_int(val <= -1, std.floor(std.abs(val)), zp, iprec, cflags.blank, cflags.plus, 8, zero_prefix)\n",
		"      else if code.ctype == 'x' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_hex(std.floor(val),\n",
		"                     zp,\n",
		"                     iprec,\n",
		"                     cflags.blank,\n",
		"                     cflags.plus,\n",
		"                     cflags.alt,\n",
		"                     code.caps)\n",
		"      else if code.ctype == 'f' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_float_dec(val,\n",
		"                           zp,\n",
		"                           cflags.blank,\n",
		"                           cflags.plus,\n",
		"                           cflags.alt,\n",
		"                           true,\n",
		"                           fpprec)\n",
		"      else if code.ctype == 'e' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_float_sci(val,\n",
		"                           zp,\n",
		"                           cflags.blank,\n",
		"                           cflags.plus,\n",
		"                           cflags.alt,\n",
		"                           true,\n",
		"                           code.caps,\n",
		"                           fpprec)\n",
		"      else if code.ctype == 'g' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          local exponent = std.floor(std.log(std.abs(val)) / std.log(10));\n",
		"          if exponent < -4 || exponent >= fpprec then\n",
		"            render_float_sci(val,\n",
		"                             zp,\n",
		"                             cflags.blank,\n",
		"                             cflags.plus,\n",
		"                             cflags.alt,\n",
		"                             cflags.alt,\n",
		"                             code.caps,\n",
		"                             fpprec - 1)\n",
		"          else\n",
		"            local digits_before_pt = std.max(1, exponent + 1);\n",
		"            render_float_dec(val,\n",
		"                             zp,\n",
		"                             cflags.blank,\n",
		"                             cflags.plus,\n",
		"                             cflags.alt,\n",
		"                             cflags.alt,\n",
		"                             fpprec - digits_before_pt)\n",
		"      else if code.ctype == 'c' then\n",
		"        if std.type(val) == 'number' then\n",
		"          std.char(val)\n",
		"        else if std.type(val) == 'string' then\n",
		"          if std.length(val) == 1 then\n",
		"            val\n",
		"          else\n",
		"            error '%c expected 1-sized string got: ' + std.length(val)\n",
		"        else\n",
		"          error '%c expected number / string, got: ' + std.type(val)\n",
		"      else\n",
		"        error 'Unknown code: ' + code.ctype;\n",
		"\n",
		"    // Render a parsed format string with an array of values.\n",
		"    local format_codes_arr(codes, arr, i, j, v) =\n",
		"      if i >= std.length(codes) then\n",
		"        if j < std.length(arr) then\n",
		"          error ('Too many values to format: ' + std.length(arr) + ', expected ' + j)\n",
		"        else\n",
		"          v\n",
		"      else\n",
		"        local code = codes[i];\n",
		"        if std.type(code) == 'string' then\n",
		"          format_codes_arr(codes, arr, i + 1, j, v + code) tailstrict\n",
		"        else\n",
		"          local tmp = if code.fw == '*' then {\n",
		"            j: j + 1,\n",
		"            fw: if j >= std.length(arr) then\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected at least ' + j)\n",
		"            else\n",
		"              arr[j],\n",
		"          } else {\n",
		"            j: j,\n",
		"            fw: code.fw,\n",
		"          };\n",
		"          local tmp2 = if code.prec == '*' then {\n",
		"            j: tmp.j + 1,\n",
		"            prec: if tmp.j >= std.length(arr) then\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected at least ' + tmp.j)\n",
		"            else\n",
		"              arr[tmp.j],\n",
		"          } else {\n",
		"            j: tmp.j,\n",
		"            prec: code.prec,\n",
		"          };\n",
		"          local j2 = tmp2.j;\n",
		"          local val =\n",
		"            if j2 < std.length(arr) then\n",
		"              arr[j2]\n",
		"            else\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected more than ' + j2);\n",
		"          local s =\n",
		"            if code.ctype == '%' then\n",
		"              '%'\n",
		"            else\n",
		"              format_code(val, code, tmp.fw, tmp2.prec, j2);\n",
		"          local s_padded =\n",
		"            if code.cflags.left then\n",
		"              pad_right(s, tmp.fw, ' ')\n",
		"            else\n",
		"              pad_left(s, tmp.fw, ' ');\n",
		"          local j3 =\n",
		"            if code.ctype == '%' then\n",
		"              j2\n",
		"            else\n",
		"              j2 + 1;\n",
		"          format_codes_arr(codes, arr, i + 1, j3, v + s_padded) tailstrict;\n",
		"\n",
		"    // Render a parsed format string with an object of values.\n",
		"    local format_codes_obj(codes, obj, i, v) =\n",
		"      if i >= std.length(codes) then\n",
		"        v\n",
		"      else\n",
		"        local code = codes[i];\n",
		"        if std.type(code) == 'string' then\n",
		"          format_codes_obj(codes, obj, i + 1, v + code) tailstrict\n",
		"        else\n",
		"          local f =\n",
		"            if code.mkey == null then\n",
		"              error 'Mapping keys required.'\n",
		"            else\n",
		"              code.mkey;\n",
		"          local fw =\n",
		"            if code.fw == '*' then\n",
		"              error 'Cannot use * field width with object.'\n",
		"            else\n",
		"              code.fw;\n",
		"          local prec =\n",
		"            if code.prec == '*' then\n",
		"              error 'Cannot use * precision with object.'\n",
		"            else\n",
		"              code.prec;\n",
		"          local val =\n",
		"            if std.objectHasAll(obj, f) then\n",
		"              obj[f]\n",
		"            else\n",
		"              error 'No such field: ' + f;\n",
		"          local s =\n",
		"            if code.ctype == '%' then\n",
		"              '%'\n",
		"            else\n",
		"              format_code(val, code, fw, prec, f);\n",
		"          local s_padded =\n",
		"            if code.cflags.left then\n",
		"              pad_right(s, fw, ' ')\n",
		"            else\n",
		"              pad_left(s, fw, ' ');\n",
		"          format_codes_obj(codes, obj, i + 1, v + s_padded) tailstrict;\n",
		"\n",
		"    if std.isArray(vals) then\n",
		"      format_codes_arr(codes, vals, 0, 0, '')\n",
		"    else if std.isObject(vals) then\n",
		"      format_codes_obj(codes, vals, 0, '')\n",
		"    else\n",
		"      format_codes_arr(codes, [vals], 0, 0, ''),\n",
		"\n",
		"  foldr(func, arr, init)::\n",
		"    local aux(func, arr, running, idx) =\n",
		"      if idx < 0 then\n",
		"        running\n",
		"      else\n",
		"        aux(func, arr, func(arr[idx], running), idx - 1) tailstrict;\n",
		"    aux(func, arr, init, std.length(arr) - 1),\n",
		"\n",
		"  foldl(func, arr, init)::\n",
		"    local aux(func, arr, running, idx) =\n",
		"      if idx >= std.length(arr) then\n",
		"        running\n",
		"      else\n",
		"        aux(func, arr, func(running, arr[idx]), idx + 1) tailstrict;\n",
		"    aux(func, arr, init, 0),\n",
		"\n",
		"\n",
		"  filterMap(filter_func, map_func, arr)::\n",
		"    if !std.isFunction(filter_func) then\n",
		"      error ('std.filterMap first param must be function, got ' + std.type(filter_func))\n",
		"    else if !std.isFunction(map_func) then\n",
		"      error ('std.filterMap second param must be function, got ' + std.type(map_func))\n",
		"    else if !std.isArray(arr) then\n",
		"      error ('std.filterMap third param must be array, got ' + std.type(arr))\n",
		"    else\n",
		"      std.map(map_func, std.filter(filter_func, arr)),\n",
		"\n",
		"  assertEqual(a, b)::\n",
		"    if a == b then\n",
		"      true\n",
		"    else\n",
		"      error 'Assertion failed. ' + a + ' != ' + b,\n",
		"\n",
		"  abs(n)::\n",
		"    if !std.isNumber(n) then\n",
		"      error 'std.abs expected number, got ' + std.type(n)\n",
		"    else\n",
		"      if n > 0 then n else -n,\n",
		"\n",
		"  sign(n)::\n",
		"    if !std.isNumber(n) then\n",
		"      error 'std.sign expected number, got ' + std.type(n)\n",
		"    else\n",
		"      if n > 0 then\n",
		"        1\n",
		"      else if n < 0 then\n",
		"        -1\n",
		"      else 0,\n",
		"\n",
		"  max(a, b)::\n",
		"    if !std.isNumber(a) then\n",
		"      error 'std.max first param expected number, got ' + std.type(a)\n",
		"    else if !std.isNumber(b) then\n",
		"      error 'std.max second param expected number, got ' + std.type(b)\n",
		"    else\n",
		"      if a > b then a else b,\n",
		"\n",
		"  min(a, b)::\n",
		"    if !std.isNumber(a) then\n",
		"      error 'std.min first param expected number, got ' + std.type(a)\n",
		"    else if !std.isNumber(b) then\n",
		"      error 'std.min second param expected number, got ' + std.type(b)\n",
		"    else\n",
		"      if a < b then a else b,\n",
		"\n",
		"  clamp(x, minVal, maxVal)::\n",
		"    if x < minVal then minVal\n",
		"    else if x > maxVal then maxVal\n",
		"    else x,\n",
		"\n",
		"  flattenArrays(arrs)::\n",
		"    std.foldl(function(a, b) a + b, arrs, []),\n",
		"\n",
		"  manifestIni(ini)::\n",
		"    local body_lines(body) =\n",
		"      std.join([], [\n",
		"        local value_or_values = body[k];\n",
		"        if std.isArray(value_or_values) then\n",
		"          ['%s = %s' % [k, value] for value in value_or_values]\n",
		"        else\n",
		"          ['%s = %s' % [k, value_or_values]]\n",
		"\n",
		"        for k in std.objectFields(body)\n",
		"      ]);\n",
		"\n",
		"    local section_lines(sname, sbody) = ['[%s]' % [sname]] + body_lines(sbody),\n",
		"          main_body = if std.objectHas(ini, 'main') then body_lines(ini.main) else [],\n",
		"          all_sections = [\n",
		"      section_lines(k, ini.sections[k])\n",
		"      for k in std.objectFields(ini.sections)\n",
		"    ];\n",
		"    std.join('\\n', main_body + std.flattenArrays(all_sections) + ['']),\n",
		"\n",
		"  manifestToml(value):: std.manifestTomlEx(value, '  '),\n",
		"\n",
		"  manifestTomlEx(value, indent)::\n",
		"    local\n",
		"      escapeStringToml = std.escapeStringJson,\n",
		"      escapeKeyToml(key) =\n",
		"        local bare_allowed = std.set(std.stringChars('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-'));\n",
		"        if std.setUnion(std.set(std.stringChars(key)), bare_allowed) == bare_allowed then key else escapeStringToml(key),\n",
		"      isTableArray(v) = std.isArray(v) && std.length(v) > 0 && std.all(std.map(std.isObject, v)),\n",
		"      isSection(v) = std.isObject(v) || isTableArray(v),\n",
		"      renderValue(v, indexedPath, inline, cindent) =\n",
		"        if v == true then\n",
		"          'true'\n",
		"        else if v == false then\n",
		"          'false'\n",
		"        else if v == null then\n",
		"          error 'Tried to manifest \"null\" at ' + indexedPath\n",
		"        else if std.isNumber(v) then\n",
		"          '' + v\n",
		"        else if std.isString(v) then\n",
		"          escapeStringToml(v)\n",
		"        else if std.isFunction(v) then\n",
		"          error 'Tried to manifest function at ' + indexedPath\n",
		"        else if std.isArray(v) then\n",
		"          if std.length(v) == 0 then\n",
		"            '[]'\n",
		"          else\n",
		"            local range = std.range(0, std.length(v) - 1);\n",
		"            local new_indent = if inline then '' else cindent + indent;\n",
		"            local separator = if inline then ' ' else '\\n';\n",
		"            local lines = ['[' + separator]\n",
		"                          + std.join([',' + separator],\n",
		"                                     [\n",
		"                                       [new_indent + renderValue(v[i], indexedPath + [i], true, '')]\n",
		"                                       for i in range\n",
		"                                     ])\n",
		"                          + [separator + (if inline then '' else cindent) + ']'];\n",
		"            std.join('', lines)\n",
		"        else if std.isObject(v) then\n",
		"          local lines = ['{ ']\n",
		"                        + std.join([', '],\n",
		"                                   [\n",
		"                                     [escapeKeyToml(k) + ' = ' + renderValue(v[k], indexedPath + [k], true, '')]\n",
		"                                     for k in std.objectFields(v)\n",
		"                                   ])\n",
		"                        + [' }'];\n",
		"          std.join('', lines),\n",
		"      renderTableInternal(v, path, indexedPath, cindent) =\n",
		"        local kvp = std.flattenArrays([\n",
		"          [cindent + escapeKeyToml(k) + ' = ' + renderValue(v[k], indexedPath + [k], false, cindent)]\n",
		"          for k in std.objectFields(v)\n",
		"          if !isSection(v[k])\n",
		"        ]);\n",
		"        local sections = [std.join('\\n', kvp)] + [\n",
		"          (\n",
		"            if std.isObject(v[k]) then\n",
		"              renderTable(v[k], path + [k], indexedPath + [k], cindent)\n",
		"            else\n",
		"              renderTableArray(v[k], path + [k], indexedPath + [k], cindent)\n",
		"          )\n",
		"          for k in std.objectFields(v)\n",
		"          if isSection(v[k])\n",
		"        ];\n",
		"        std.join('\\n\\n', sections),\n",
		"      renderTable(v, path, indexedPath, cindent) =\n",
		"        cindent + '[' + std.join('.', std.map(escapeKeyToml, path)) + ']'\n",
		"        + (if v == {} then '' else '\\n')\n",
		"        + renderTableInternal(v, path, indexedPath, cindent + indent),\n",
		"      renderTableArray(v, path, indexedPath, cindent) =\n",
		"        local range = std.range(0, std.length(v) - 1);\n",
		"        local sections = [\n",
		"          (cindent + '[[' + std.join('.', std.map(escapeKeyToml, path)) + ']]'\n",
		"           + (if v[i] == {} then '' else '\\n')\n",
		"           + renderTableInternal(v[i], path, indexedPath + [i], cindent + indent))\n",
		"          for i in range\n",
		"        ];\n",
		"        std.join('\\n\\n', sections);\n",
		"    if std.isObject(value) then\n",
		"      renderTableInternal(value, [], [], '')\n",
		"    else\n",
		"      error 'TOML body must be an object. Got ' + std.type(value),\n",
		"\n",
		"  escapeStringJson(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == '\"' then\n",
		"        '\\\\\"'\n",
		"      else if ch == '\\\\' then\n",
		"        '\\\\\\\\'\n",
		"      else if ch == '\\b' then\n",
		"        '\\\\b'\n",
		"      else if ch == '\\f' then\n",
		"        '\\\\f'\n",
		"      else if ch == '\\n' then\n",
		"        '\\\\n'\n",
		"      else if ch == '\\r' then\n",
		"        '\\\\r'\n",
		"      else if ch == '\\t' then\n",
		"        '\\\\t'\n",
		"      else\n",
		"        local cp = std.codepoint(ch);\n",
		"        if cp < 32 || (cp >= 127 && cp <= 159) then\n",
		"          '\\\\u%04x' % [cp]\n",
		"        else\n",
		"          ch;\n",
		"    '\"%s\"' % std.join('', [trans(ch) for ch in std.stringChars(str)]),\n",
		"\n",
		"  escapeStringPython(str)::\n",
		"    std.escapeStringJson(str),\n",
		"\n",
		"  escapeStringBash(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == \"'\" then\n",
		"        \"'\\\"'\\\"'\"\n",
		"      else\n",
		"        ch;\n",
		"    \"'%s'\" % std.join('', [trans(ch) for ch in std.stringChars(str)]),\n",
		"\n",
		"  escapeStringDollars(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == '$' then\n",
		"        '$$'\n",
		"      else\n",
		"        ch;\n",
		"    std.foldl(function(a, b) a + trans(b), std.stringChars(str), ''),\n",
		"\n",
		"  local xml_escapes = {\n",
		"    '<': '&lt;',\n",
		"    '>': '&gt;',\n",
		"    '&': '&amp;',\n",
		"    '\"': '&quot;',\n",
		"    \"'\": '&apos;',\n",
		"  },\n",
		"\n",
		"  escapeStringXML(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    std.join('', [std.get(xml_escapes, ch, ch) for ch in std.stringChars(str)]),\n",
		"\n",
		"  manifestJson(value):: std.manifestJsonEx(value, '    '),\n",
		"\n",
		"  manifestJsonMinified(value):: std.manifestJsonEx(value, '', '', ':'),\n",
		"\n",
		"  manifestJsonEx(value, indent, newline='\\n', key_val_sep=': ')::\n",
		"    local aux(v, path, cindent) =\n",
		"      if v == true then\n",
		"        'true'\n",
		"      else if v == false then\n",
		"        'false'\n",
		"      else if v == null then\n",
		"        'null'\n",
		"      else if std.isNumber(v) then\n",
		"        '' + v\n",
		"      else if std.isString(v) then\n",
		"        std.escapeStringJson(v)\n",
		"      else if std.isFunction(v) then\n",
		"        error 'Tried to manifest function at ' + path\n",
		"      else if std.isArray(v) then\n",
		"        local range = std.range(0, std.length(v) - 1);\n",
		"        local new_indent = cindent + indent;\n",
		"        local lines = ['[' + newline]\n",
		"                      + std.join([',' + newline],\n",
		"                                 [\n",
		"                                   [new_indent + aux(v[i], path + [i], new_indent)]\n",
		"                                   for i in range\n",
		"                                 ])\n",
		"                      + [newline + cindent + ']'];\n",
		"        std.join('', lines)\n",
		"      else if std.isObject(v) then\n",
		"        local lines = ['{' + newline]\n",
		"                      + std.join([',' + newline],\n",
		"                                 [\n",
		"                                   [cindent + indent + std.escapeStringJson(k) + key_val_sep\n",
		"                                    + aux(v[k], path + [k], cindent + indent)]\n",
		"                                   for k in std.objectFields(v)\n",
		"                                 ])\n",
		"                      + [newline + cindent + '}'];\n",
		"        std.join('', lines);\n",
		"    aux(value, [], ''),\n",
		"\n",
		"  manifestYamlDoc(value, indent_array_in_object=false, quote_keys=true)::\n",
		"    local onlyChars(charSet, strSet) =\n",
		"      if std.length(std.setInter(charSet, strSet)) == std.length(strSet) then\n",
		"        true\n",
		"      else false;\n",
		"    local isReserved(key) =\n",
		"      // NOTE: These values are checked for case insensitively.\n",
		"      // While this approach results in some false positives, it eliminates\n",
		"      // the risk of missing a permutation.\n",
		"      local reserved = [\n",
		"        // Boolean types taken from https://yaml.org/type/bool.html\n",
		"        'true',\n",
		"        'false',\n",
		"        'yes',\n",
		"        'no',\n",
		"        'on',\n",
		"        'off',\n",
		"        'y',\n",
		"        'n',\n",
		"        // Numerical words taken from https://yaml.org/type/float.html\n",
		"        '.nan',\n",
		"        '-.inf',\n",
		"        '+.inf',\n",
		"        '.inf',\n",
		"        'null',\n",
		"        // Invalid keys that contain no invalid characters\n",
		"        '-',\n",
		"        '---',\n",
		"        '',\n",
		"      ];\n",
		"      local bad = [word for word in reserved if word == std.asciiLower(key)];\n",
		"      if std.length(bad) > 0 then\n",
		"        true\n",
		"      else false;\n",
		"    local typeMatch(m_key, type) =\n",
		"      // Look for positive or negative numerical types (ex: 0x)\n",
		"      if std.substr(m_key, 0, 2) == type || std.substr(m_key, 0, 3) == '-' + type then\n",
		"        true\n",
		"      else false;\n",
		"    local bareSafe(key) =\n",
		"      /*\n",
		"      For a key to be considered safe to emit without quotes, the following must be true\n",
		"        - All characters must match [a-zA-Z0-9_/\\-]\n",
		"        - Not match the integer format defined in https://yaml.org/type/int.html\n",
		"        - Not match the float format defined in https://yaml.org/type/float.html\n",
		"        - Not match the timestamp format defined in https://yaml.org/type/timestamp.html\n",
		"        - Not match the boolean format defined in https://yaml.org/type/bool.html\n",
		"        - Not match the null format defined in https://yaml.org/type/null.html\n",
		"        - Not match (ignoring case) any reserved words which pass the above tests.\n",
		"          Reserved words are defined in isReserved() above.\n",
		"\n",
		"      Since the remaining YAML types require characters outside the set chosen as valid\n",
		"      for the elimination of quotes from the YAML output, the remaining types listed at\n",
		"      https://yaml.org/type/ are by default always quoted.\n",
		"      */\n",
		"      local letters = std.set(std.stringChars('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-/'));\n",
		"      local digits = std.set(std.stringChars('0123456789'));\n",
		"      local intChars = std.set(digits + std.stringChars('_-'));\n",
		"      local binChars = std.set(intChars + std.stringChars('b'));\n",
		"      local hexChars = std.set(digits + std.stringChars('abcdefx_-'));\n",
		"      local floatChars = std.set(digits + std.stringChars('e._-'));\n",
		"      local dateChars = std.set(digits + std.stringChars('-'));\n",
		"      local safeChars = std.set(letters + floatChars);\n",
		"      local keyLc = std.asciiLower(key);\n",
		"      local keyChars = std.stringChars(key);\n",
		"      local keySet = std.set(keyChars);\n",
		"      local keySetLc = std.set(std.stringChars(keyLc));\n",
		"      // Check for unsafe characters\n",
		"      if !onlyChars(safeChars, keySet) then\n",
		"        false\n",
		"      // Check for reserved words\n",
		"      else if isReserved(key) then\n",
		"        false\n",
		"      /* Check for timestamp values.  Since spaces and colons are already forbidden,\n",
		"         all that could potentially pass is the standard date format (ex MM-DD-YYYY, YYYY-DD-MM, etc).\n",
		"         This check is even more conservative: Keys that meet all of the following:\n",
		"           - all characters match [0-9\\-]\n",
		"           - has exactly 2 dashes\n",
		"         are considered dates.\n",
		"      */\n",
		"      else if onlyChars(dateChars, keySet)\n",
		"              && std.length(std.findSubstr('-', key)) == 2 then\n",
		"        false\n",
		"      /* Check for integers.  Keys that meet all of the following:\n",
		"           - all characters match [0-9_\\-]\n",
		"           - has at most 1 dash\n",
		"         are considered integers.\n",
		"      */\n",
		"      else if onlyChars(intChars, keySetLc)\n",
		"              && std.length(std.findSubstr('-', key)) < 2 then\n",
		"        false\n",
		"      /* Check for binary integers.  Keys that meet all of the following:\n",
		"           - all characters match [0-9b_\\-]\n",
		"           - has at least 3 characters\n",
		"           - starts with (-)0b\n",
		"         are considered binary integers.\n",
		"      */\n",
		"      else if onlyChars(binChars, keySetLc)\n",
		"              && std.length(key) > 2\n",
		"              && typeMatch(key, '0b') then\n",
		"        false\n",
		"      /* Check for floats. Keys that meet all of the following:\n",
		"           - all characters match [0-9e._\\-]\n",
		"           - has at most a single period\n",
		"           - has at most two dashes\n",
		"           - has at most 1 'e'\n",
		"         are considered floats.\n",
		"      */\n",
		"      else if onlyChars(floatChars, keySetLc)\n",
		"              && std.length(std.findSubstr('.', key)) == 1\n",
		"              && std.length(std.findSubstr('-', key)) < 3\n",
		"              && std.length(std.findSubstr('e', keyLc)) < 2 then\n",
		"        false\n",
		"      /* Check for hexadecimals.  Keys that meet all of the following:\n",
		"           - all characters match [0-9a-fx_\\-]\n",
		"           - has at most 1 dash\n",
		"           - has at least 3 characters\n",
		"           - starts with (-)0x\n",
		"         are considered hexadecimals.\n",
		"      */\n",
		"      else if onlyChars(hexChars, keySetLc)\n",
		"              && std.length(std.findSubstr('-', key)) < 2\n",
		"              && std.length(keyChars) > 2\n",
		"              && typeMatch(key, '0x') then\n",
		"        false\n",
		"      // All checks pass. Key is safe for emission without quotes.\n",
		"      else true;\n",
		"    local escapeKeyYaml(key) =\n",
		"      if bareSafe(key) then key else std.escapeStringJson(key);\n",
		"    local aux(v, path, cindent) =\n",
		"      if v == true then\n",
		"        'true'\n",
		"      else if v == false then\n",
		"        'false'\n",
		"      else if v == null then\n",
		"        'null'\n",
		"      else if std.isNumber(v) then\n",
		"        '' + v\n",
		"      else if std.isString(v) then\n",
		"        local len = std.length(v);\n",
		"        if len == 0 then\n",
		"          '\"\"'\n",
		"        else if v[len - 1] == '\\n' then\n",
		"          local split = std.split(v, '\\n');\n",
		"          std.join('\\n' + cindent + '  ', ['|'] + split[0:std.length(split) - 1])\n",
		"        else\n",
		"          std.escapeStringJson(v)\n",
		"      else if std.isFunction(v) then\n",
		"        error 'Tried to manifest function at ' + path\n",
		"      else if std.isArray(v) then\n",
		"        if std.length(v) == 0 then\n",
		"          '[]'\n",
		"        else\n",
		"          local params(value) =\n",
		"            if std.isArray(value) && std.length(value) > 0 then {\n",
		"              // While we could avoid the new line, it yields YAML that is\n",
		"              // hard to read, e.g.:\n",
		"              // - - - 1\n",
		"              //     - 2\n",
		"              //   - - 3\n",
		"              //     - 4\n",
		"              new_indent: cindent + '  ',\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else if std.isObject(value) && std.length(value) > 0 then {\n",
		"              new_indent: cindent + '  ',\n",
		"              // In this case we can start on the same line as the - because the indentation\n",
		"              // matches up then.  The converse is not true, because fields are not always\n",
		"              // 1 character long.\n",
		"              space: ' ',\n",
		"            } else {\n",
		"              // In this case, new_indent is only used in the case of multi-line strings.\n",
		"              new_indent: cindent,\n",
		"              space: ' ',\n",
		"            };\n",
		"          local range = std.range(0, std.length(v) - 1);\n",
		"          local parts = [\n",
		"            '-' + param.space + aux(v[i], path + [i], param.new_indent)\n",
		"            for i in range\n",
		"            for param in [params(v[i])]\n",
		"          ];\n",
		"          std.join('\\n' + cindent, parts)\n",
		"      else if std.isObject(v) then\n",
		"        if std.length(v) == 0 then\n",
		"          '{}'\n",
		"        else\n",
		"          local params(value) =\n",
		"            if std.isArray(value) && std.length(value) > 0 then {\n",
		"              // Not indenting allows e.g.\n",
		"              // ports:\n",
		"              // - 80\n",
		"              // instead of\n",
		"              // ports:\n",
		"              //   - 80\n",
		"              new_indent: if indent_array_in_object then cindent + '  ' else cindent,\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else if std.isObject(value) && std.length(value) > 0 then {\n",
		"              new_indent: cindent + '  ',\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else {\n",
		"              // In this case, new_indent is only used in the case of multi-line strings.\n",
		"              new_indent: cindent,\n",
		"              space: ' ',\n",
		"            };\n",
		"          local lines = [\n",
		"            (if quote_keys then std.escapeStringJson(k) else escapeKeyYaml(k)) + ':' + param.space + aux(v[k], path + [k], param.new_indent)\n",
		"            for k in std.objectFields(v)\n",
		"            for param in [params(v[k])]\n",
		"          ];\n",
		"          std.join('\\n' + cindent, lines);\n",
		"    aux(value, [], ''),\n",
		"\n",
		"  manifestYamlStream(value, indent_array_in_object=false, c_document_end=true, quote_keys=true)::\n",
		"    if !std.isArray(value) then\n",
		"      error 'manifestYamlStream only takes arrays, got ' + std.type(value)\n",
		"    else\n",
		"      '---\\n' + std.join(\n",
		"        '\\n---\\n', [std.manifestYamlDoc(e, indent_array_in_object, quote_keys) for e in value]\n",
		"      ) + if c_document_end then '\\n...\\n' else '\\n',\n",
		"\n",
		"\n",
		"  manifestPython(v)::\n",
		"    if std.isObject(v) then\n",
		"      local fields = [\n",
		"        '%s: %s' % [std.escapeStringPython(k), std.manifestPython(v[k])]\n",
		"        for k in std.objectFields(v)\n",
		"      ];\n",
		"      '{%s}' % [std.join(', ', fields)]\n",
		"    else if std.isArray(v) then\n",
		"      '[%s]' % [std.join(', ', [std.manifestPython(v2) for v2 in v])]\n",
		"    else if std.isString(v) then\n",
		"      '%s' % [std.escapeStringPython(v)]\n",
		"    else if std.isFunction(v) then\n",
		"      error 'cannot manifest function'\n",
		"    else if std.isNumber(v) then\n",
		"      std.toString(v)\n",
		"    else if v == true then\n",
		"      'True'\n",
		"    else if v == false then\n",
		"      'False'\n",
		"    else if v == null then\n",
		"      'None',\n",
		"\n",
		"  manifestPythonVars(conf)::\n",
		"    local vars = ['%s = %s' % [k, std.manifestPython(conf[k])] for k in std.objectFields(conf)];\n",
		"    std.join('\\n', vars + ['']),\n",
		"\n",
		"  manifestXmlJsonml(value)::\n",
		"    if !std.isArray(value) then\n",
		"      error 'Expected a JSONML value (an array), got %s' % std.type(value)\n",
		"    else\n",
		"      local aux(v) =\n",
		"        if std.isString(v) then\n",
		"          v\n",
		"        else\n",
		"          local tag = v[0];\n",
		"          local has_attrs = std.length(v) > 1 && std.isObject(v[1]);\n",
		"          local attrs = if has_attrs then v[1] else {};\n",
		"          local children = if has_attrs then v[2:] else v[1:];\n",
		"          local attrs_str =\n",
		"            std.join('', [' %s=\"%s\"' % [k, attrs[k]] for k in std.objectFields(attrs)]);\n",
		"          std.deepJoin(['<', tag, attrs_str, '>', [aux(x) for x in children], '</', tag, '>']);\n",
		"\n",
		"      aux(value),\n",
		"\n",
		"  local base64_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/',\n",
		"  local base64_inv = { [base64_table[i]]: i for i in std.range(0, 63) },\n",
		"\n",
		"  base64(input)::\n",
		"    local bytes =\n",
		"      if std.isString(input) then\n",
		"        std.map(std.codepoint, input)\n",
		"      else\n",
		"        input;\n",
		"\n",
		"    local aux(arr, i, r) =\n",
		"      if i >= std.length(arr) then\n",
		"        r\n",
		"      else if i + 1 >= std.length(arr) then\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i\n",
		"          base64_table[(arr[i] & 3) << 4] +\n",
		"          '==';\n",
		"        aux(arr, i + 3, r + str) tailstrict\n",
		"      else if i + 2 >= std.length(arr) then\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i, 4 MSB of i+1\n",
		"          base64_table[(arr[i] & 3) << 4 | (arr[i + 1] & 240) >> 4] +\n",
		"          // 4 LSB of i+1\n",
		"          base64_table[(arr[i + 1] & 15) << 2] +\n",
		"          '=';\n",
		"        aux(arr, i + 3, r + str) tailstrict\n",
		"      else\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i, 4 MSB of i+1\n",
		"          base64_table[(arr[i] & 3) << 4 | (arr[i + 1] & 240) >> 4] +\n",
		"          // 4 LSB of i+1, 2 MSB of i+2\n",
		"          base64_table[(arr[i + 1] & 15) << 2 | (arr[i + 2] & 192) >> 6] +\n",
		"          // 6 LSB of i+2\n",
		"          base64_table[(arr[i + 2] & 63)];\n",
		"        aux(arr, i + 3, r + str) tailstrict;\n",
		"\n",
		"    local sanity = std.all([a < 256 for a in bytes]);\n",
		"    if !sanity then\n",
		"      error 'Can only base64 encode strings / arrays of single bytes.'\n",
		"    else\n",
		"      aux(bytes, 0, ''),\n",
		"\n",
		"\n",
		"  base64DecodeBytes(str)::\n",
		"    if std.length(str) % 4 != 0 then\n",
		"      error 'Not a base64 encoded string \"%s\"' % str\n",
		"    else\n",
		"      local aux(str, i, r) =\n",
		"        if i >= std.length(str) then\n",
		"          r\n",
		"        else\n",
		"          // all 6 bits of i, 2 MSB of i+1\n",
		"          local n1 = [base64_inv[str[i]] << 2 | (base64_inv[str[i + 1]] >> 4)];\n",
		"          // 4 LSB of i+1, 4MSB of i+2\n",
		"          local n2 =\n",
		"            if str[i + 2] == '=' then []\n",
		"            else [(base64_inv[str[i + 1]] & 15) << 4 | (base64_inv[str[i + 2]] >> 2)];\n",
		"          // 2 LSB of i+2, all 6 bits of i+3\n",
		"          local n3 =\n",
		"            if str[i + 3] == '=' then []\n",
		"            else [(base64_inv[str[i + 2]] & 3) << 6 | base64_inv[str[i + 3]]];\n",
		"          aux(str, i + 4, r + n1 + n2 + n3) tailstrict;\n",
		"      aux(str, 0, []),\n",
		"\n",
		"  base64Decode(str)::\n",
		"    local bytes = std.base64DecodeBytes(str);\n",
		"    std.join('', std.map(std.char, bytes)),\n",
		"\n",
		"  reverse(arr)::\n",
		"    local l = std.length(arr);\n",
		"    std.makeArray(l, function(i) arr[l - i - 1]),\n",
		"\n",
		"  // Merge-sort for long arrays and naive quicksort for shorter ones\n",
		"  sort(arr, keyF=id)::\n",
		"    local quickSort(arr, keyF=id) =\n",
		"      local l = std.length(arr);\n",
		"      if std.length(arr) <= 1 then\n",
		"        arr\n",
		"      else\n",
		"        local pos = 0;\n",
		"        local pivot = keyF(arr[pos]);\n",
		"        local rest = std.makeArray(l - 1, function(i) if i < pos then arr[i] else arr[i + 1]);\n",
		"        local left = std.filter(function(x) keyF(x) < pivot, rest);\n",
		"        local right = std.filter(function(x) keyF(x) >= pivot, rest);\n",
		"        quickSort(left, keyF) + [arr[pos]] + quickSort(right, keyF);\n",
		"\n",
		"    local merge(a, b) =\n",
		"      local la = std.length(a), lb = std.length(b);\n",
		"      local aux(i, j, prefix) =\n",
		"        if i == la then\n",
		"          prefix + b[j:]\n",
		"        else if j == lb then\n",
		"          prefix + a[i:]\n",
		"        else\n",
		"          if keyF(a[i]) <= keyF(b[j]) then\n",
		"            aux(i + 1, j, prefix + [a[i]]) tailstrict\n",
		"          else\n",
		"            aux(i, j + 1, prefix + [b[j]]) tailstrict;\n",
		"      aux(0, 0, []);\n",
		"\n",
		"    local l = std.length(arr);\n",
		"    if std.length(arr) <= 30 then\n",
		"      quickSort(arr, keyF=keyF)\n",
		"    else\n",
		"      local mid = std.floor(l / 2);\n",
		"      local left = arr[:mid], right = arr[mid:];\n",
		"      merge(std.sort(left, keyF=keyF), std.sort(right, keyF=keyF)),\n",
		"\n",
		"  uniq(arr, keyF=id)::\n",
		"    local f(a, b) =\n",
		"      if std.length(a) == 0 then\n",
		"        [b]\n",
		"      else if keyF(a[std.length(a) - 1]) == keyF(b) then\n",
		"        a\n",
		"      else\n",
		"        a + [b];\n",
		"    std.foldl(f, arr, []),\n",
		"\n",
		"  set(arr, keyF=id)::\n",
		"    std.uniq(std.sort(arr, keyF), keyF),\n",
		"\n",
		"  setMember(x, arr, keyF=id)::\n",
		"    // TODO(dcunnin): Binary chop for O(log n) complexity\n",
		"    std.length(std.setInter([x], arr, keyF)) > 0,\n",
		"\n",
		"  setUnion(a, b, keyF=id)::\n",
		"    // NOTE: order matters, values in `a` win\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) then\n",
		"        acc + b[j:]\n",
		"      else if j >= std.length(b) then\n",
		"        acc + a[i:]\n",
		"      else\n",
		"        local ak = keyF(a[i]);\n",
		"        local bk = keyF(b[j]);\n",
		"        if ak == bk then\n",
		"          aux(a, b, i + 1, j + 1, acc + [a[i]]) tailstrict\n",
		"        else if ak < bk then\n",
		"          aux(a, b, i + 1, j, acc + [a[i]]) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc + [b[j]]) tailstrict;\n",
		"    aux(a, b, 0, 0, []),\n",
		"\n",
		"  setInter(a, b, keyF=id)::\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) || j >= std.length(b) then\n",
		"        acc\n",
		"      else\n",
		"        if keyF(a[i]) == keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j + 1, acc + [a[i]]) tailstrict\n",
		"        else if keyF(a[i]) < keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j, acc) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc) tailstrict;\n",
		"    aux(a, b, 0, 0, []) tailstrict,\n",
		"\n",
		"  setDiff(a, b, keyF=id)::\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) then\n",
		"        acc\n",
		"      else if j >= std.length(b) then\n",
		"        acc + a[i:]\n",
		"      else\n",
		"        if keyF(a[i]) == keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j + 1, acc) tailstrict\n",
		"        else if keyF(a[i]) < keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j, acc + [a[i]]) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc) tailstrict;\n",
		"    aux(a, b, 0, 0, []) tailstrict,\n",
		"\n",
		"  mergePatch(target, patch)::\n",
		"    if std.isObject(patch) then\n",
		"      local target_object =\n",
		"        if std.isObject(target) then target else {};\n",
		"\n",
		"      local target_fields =\n",
		"        if std.isObject(target_object) then std.objectFields(target_object) else [];\n",
		"\n",
		"      local null_fields = [k for k in std.objectFields(patch) if patch[k] == null];\n",
		"      local both_fields = std.setUnion(target_fields, std.objectFields(patch));\n",
		"\n",
		"      {\n",
		"        [k]:\n",
		"          if !std.objectHas(patch, k) then\n",
		"            target_object[k]\n",
		"          else if !std.objectHas(target_object, k) then\n",
		"            std.mergePatch(null, patch[k]) tailstrict\n",
		"          else\n",
		"            std.mergePatch(target_object[k], patch[k]) tailstrict\n",
		"        for k in std.setDiff(both_fields, null_fields)\n",
		"      }\n",
		"    else\n",
		"      patch,\n",
		"\n",
		"  get(o, f, default=null, inc_hidden=true)::\n",
		"    if std.objectHasEx(o, f, inc_hidden) then o[f] else default,\n",
		"\n",
		"  objectFields(o)::\n",
		"    std.objectFieldsEx(o, false),\n",
		"\n",
		"  objectFieldsAll(o)::\n",
		"    std.objectFieldsEx(o, true),\n",
		"\n",
		"  objectHas(o, f)::\n",
		"    std.objectHasEx(o, f, false),\n",
		"\n",
		"  objectHasAll(o, f)::\n",
		"    std.objectHasEx(o, f, true),\n",
		"\n",
		"  objectValues(o)::\n",
		"    [o[k] for k in std.objectFields(o)],\n",
		"\n",
		"  objectValuesAll(o)::\n",
		"    [o[k] for k in std.objectFieldsAll(o)],\n",
		"\n",
		"  objectKeysValues(o)::\n",
		"    [{ key: k, value: o[k] } for k in std.objectFields(o)],\n",
		"\n",
		"  objectKeysValuesAll(o)::\n",
		"    [{ key: k, value: o[k] } for k in std.objectFieldsAll(o)],\n",
		"\n",
		"  equals(a, b)::\n",
		"    local ta = std.type(a);\n",
		"    local tb = std.type(b);\n",
		"    if !std.primitiveEquals(ta, tb) then\n",
		"      false\n",
		"    else\n",
		"      if std.primitiveEquals(ta, 'array') then\n",
		"        local la = std.length(a);\n",
		"        if !std.primitiveEquals(la, std.length(b)) then\n",
		"          false\n",
		"        else\n",
		"          local aux(a, b, i) =\n",
		"            if i >= la then\n",
		"              true\n",
		"            else if a[i] != b[i] then\n",
		"              false\n",
		"            else\n",
		"              aux(a, b, i + 1) tailstrict;\n",
		"          aux(a, b, 0)\n",
		"      else if std.primitiveEquals(ta, 'object') then\n",
		"        local fields = std.objectFields(a);\n",
		"        local lfields = std.length(fields);\n",
		"        if fields != std.objectFields(b) then\n",
		"          false\n",
		"        else\n",
		"          local aux(a, b, i) =\n",
		"            if i >= lfields then\n",
		"              true\n",
		"            else if local f = fields[i]; a[f] != b[f] then\n",
		"              false\n",
		"            else\n",
		"              aux(a, b, i + 1) tailstrict;\n",
		"          aux(a, b, 0)\n",
		"      else\n",
		"        std.primitiveEquals(a, b),\n",
		"\n",
		"\n",
		"  resolvePath(f, r)::\n",
		"    local arr = std.split(f, '/');\n",
		"    std.join('/', std.makeArray(std.length(arr) - 1, function(i) arr[i]) + [r]),\n",
		"\n",
		"  prune(a)::\n",
		"    local isContent(b) =\n",
		"      if b == null then\n",
		"        false\n",
		"      else if std.isArray(b) then\n",
		"        std.length(b) > 0\n",
		"      else if std.isObject(b) then\n",
		"        std.length(b) > 0\n",
		"      else\n",
		"        true;\n",
		"    if std.isArray(a) then\n",
		"      [std.prune(x) for x in a if isContent($.prune(x))]\n",
		"    else if std.isObject(a) then {\n",
		"      [x]: $.prune(a[x])\n",
		"      for x in std.objectFields(a)\n",
		"      if isContent(std.prune(a[x]))\n",
		"    } else\n",
		"      a,\n",
		"\n",
		"  findSubstr(pat, str)::\n",
		"    if !std.isString(pat) then\n",
		"      error 'findSubstr first parameter should be a string, got ' + std.type(pat)\n",
		"    else if !std.isString(str) then\n",
		"      error 'findSubstr second parameter should be a string, got ' + std.type(str)\n",
		"    else\n",
		"      local pat_len = std.length(pat);\n",
		"      local str_len = std.length(str);\n",
		"      if pat_len == 0 || str_len == 0 || pat_len > str_len then\n",
		"        []\n",
		"      else\n",
		"        std.filter(function(i) str[i:i + pat_len] == pat, std.range(0, str_len - pat_len)),\n",
		"\n",
		"  find(value, arr)::\n",
		"    if !std.isArray(arr) then\n",
		"      error 'find second parameter should be an array, got ' + std.type(arr)\n",
		"    else\n",
		"      std.filter(function(i) arr[i] == value, std.range(0, std.length(arr) - 1)),\n",
		"\n",
		"  all(arr)::\n",
		"    assert std.isArray(arr) : 'all() parameter should be an array, got ' + std.type(arr);\n",
		"    local arrLen = std.length(arr);\n",
		"    local aux(idx) =\n",
		"      if idx >= arrLen then\n",
		"        true\n",
		"      else\n",
		"        local e = arr[idx];\n",
		"        assert std.isBoolean(e) : std.format('element \"%s\" of type %s is not a boolean', e, std.type(e));\n",
		"        if !e then\n",
		"          false\n",
		"        else\n",
		"          aux(idx + 1) tailstrict;\n",
		"    aux(0),\n",
		"\n",
		"  any(arr)::\n",
		"    assert std.isArray(arr) : 'any() parameter should be an array, got ' + std.type(arr);\n",
		"    local arrLen = std.length(arr);\n",
		"    local aux(idx) =\n",
		"      if idx >= arrLen then\n",
		"        false\n",
		"      else\n",
		"        local e = arr[idx];\n",
		"        assert std.isBoolean(e) : std.format('element \"%s\" of type %s is not a boolean', e, std.type(e));\n",
		"        if e then\n",
		"          true\n",
		"        else\n",
		"          aux(idx + 1) tailstrict;\n",
		"    aux(0),\n",
		"\n",
		"  // Three way comparison.\n",
		"  // TODO(sbarzowski): consider exposing and documenting it properly\n",
		"  __compare(v1, v2)::\n",
		"    local t1 = std.type(v1), t2 = std.type(v2);\n",
		"    if t1 != t2 then\n",
		"      error 'Comparison requires matching types. Got ' + t1 + ' and ' + t2\n",
		"    else if t1 == 'array' then\n",
		"      std.__compare_array(v1, v2)\n",
		"    else if t1 == 'function' || t1 == 'object' || t1 == 'boolean' then\n",
		"      error 'Values of type ' + t1 + ' are not comparable.'\n",
		"    else if v1 < v2 then -1\n",
		"    else if v1 > v2 then 1\n",
		"    else 0,\n",
		"\n",
		"  __compare_array(arr1, arr2)::\n",
		"    local len1 = std.length(arr1), len2 = std.length(arr2);\n",
		"    local minLen = std.min(len1, len2);\n",
		"    local aux(i) =\n",
		"      if i < minLen then\n",
		"        local cmpRes = std.__compare(arr1[i], arr2[i]);\n",
		"        if cmpRes != 0 then\n",
		"          cmpRes\n",
		"        else\n",
		"          aux(i + 1) tailstrict\n",
		"      else\n",
		"        std.__compare(len1, len2);\n",
		"    aux(0),\n",
		"\n",
		"  __array_less(arr1, arr2):: std.__compare_array(arr1, arr2) == -1,\n",
		"  __array_greater(arr1, arr2):: std.__compare_array(arr1, arr2) == 1,\n",
		"  __array_less_or_equal(arr1, arr2):: std.__compare_array(arr1, arr2) <= 0,\n",
		"  __array_greater_or_equal(arr1, arr2):: std.__compare_array(arr1, arr2) >= 0,\n",
		"\n",
		"  sum(arr):: std.foldl(function(a, b) a + b, arr, 0),\n",
		"\n",
		"  xor(x, y):: x != y,\n",
		"\n",
		"  xnor(x, y):: x == y,\n",
		"\n",
		"  round(x):: std.floor(x + 0.5),\n",
		"\n",
		"  isEmpty(str):: std.length(str) == 0,\n",
		"}\n",
		"\n",
	},
}

// StdAst is the AST for the standard library.
var StdAst = _StdAst
var _StdAst = &ast.DesugaredObject{
	Asserts: ast.Nodes{},
	Fields: ast.DesugaredObjectFields{
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isString",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "string",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p7,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(28),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(28),
									Column: int(40),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(28),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(28),
											Column: int(20),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p7,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(28),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(28),
										Column: int(25),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(28),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(28),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p7,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(28),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(28),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p7,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(28),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(28),
								Column: int(40),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(28),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(28),
								Column: int(13),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(28),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(28),
					Column: int(40),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isNumber",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "number",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p29,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(29),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(29),
									Column: int(40),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(29),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(29),
											Column: int(20),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p29,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(29),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(29),
										Column: int(25),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p38,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(29),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(29),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p29,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(29),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(29),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p29,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(29),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(29),
								Column: int(40),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(29),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(29),
								Column: int(13),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(29),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(29),
					Column: int(40),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isBoolean",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "boolean",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p48,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(30),
									Column: int(33),
								},
								End: ast.Location{
									Line: int(30),
									Column: int(42),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(30),
											Column: int(18),
										},
										End: ast.Location{
											Line: int(30),
											Column: int(21),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p48,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(30),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(30),
										Column: int(26),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p57,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(30),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(30),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p48,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(30),
									Column: int(18),
								},
								End: ast.Location{
									Line: int(30),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p48,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(30),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(30),
								Column: int(42),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(30),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(30),
								Column: int(14),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(30),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(30),
					Column: int(42),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isObject",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "object",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p67,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(31),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(31),
									Column: int(40),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(31),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(31),
											Column: int(20),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p67,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(31),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(31),
										Column: int(25),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p76,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(31),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(31),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p67,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(31),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(31),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p67,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(31),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(31),
								Column: int(40),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(31),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(31),
								Column: int(13),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(31),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(31),
					Column: int(40),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isArray",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "array",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p86,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(32),
									Column: int(31),
								},
								End: ast.Location{
									Line: int(32),
									Column: int(38),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(32),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(32),
											Column: int(19),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p86,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(32),
										Column: int(16),
									},
									End: ast.Location{
										Line: int(32),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p95,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(32),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(32),
													Column: int(26),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p86,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(32),
									Column: int(16),
								},
								End: ast.Location{
									Line: int(32),
									Column: int(27),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p86,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(32),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(32),
								Column: int(38),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(32),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(32),
								Column: int(12),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(32),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(32),
					Column: int(38),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isFunction",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralString{
						Value: "function",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p105,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(33),
									Column: int(34),
								},
								End: ast.Location{
									Line: int(33),
									Column: int(44),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(33),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(33),
											Column: int(22),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "type",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p105,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(33),
										Column: int(19),
									},
									End: ast.Location{
										Line: int(33),
										Column: int(27),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p114,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(33),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(33),
													Column: int(29),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p105,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(33),
									Column: int(19),
								},
								End: ast.Location{
									Line: int(33),
									Column: int(30),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p105,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(33),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(33),
								Column: int(44),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(33),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(33),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(33),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(33),
					Column: int(44),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "toString",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.LiteralString{
							Value: "string",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p125,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(36),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(31),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(36),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(36),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "type",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p125,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(36),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(36),
											Column: int(16),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p134,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(36),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(36),
														Column: int(18),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p125,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(36),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(19),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p125,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(36),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(31),
								},
							},
						},
						Op: ast.BinaryOp(12),
					},
					BranchTrue: &ast.Var{
						Id: "a",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p125,
							FreeVars: ast.Identifiers{
								"a",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(36),
									Column: int(37),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(38),
								},
							},
						},
					},
					BranchFalse: &ast.Binary{
						Right: &ast.Var{
							Id: "a",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p125,
								FreeVars: ast.Identifiers{
									"a",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(36),
										Column: int(49),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(50),
									},
								},
							},
						},
						Left: &ast.LiteralString{
							Value: "",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p125,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(36),
										Column: int(44),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(46),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p125,
							FreeVars: ast.Identifiers{
								"a",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(36),
									Column: int(44),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(50),
								},
							},
						},
						Op: ast.BinaryOp(3),
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p125,
						FreeVars: ast.Identifiers{
							"a",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(36),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(36),
								Column: int(50),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(35),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(35),
								Column: int(13),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(35),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(36),
					Column: int(50),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "substr",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(39),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p157,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(39),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(39),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p161,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(39),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(39),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p157,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(39),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(39),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(40),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(15),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p157,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(40),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(40),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "from",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p173,
												FreeVars: ast.Identifiers{
													"from",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(40),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(40),
														Column: int(29),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p157,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(40),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(40),
										Column: int(30),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(41),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(15),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isNumber",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(41),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(41),
												Column: int(24),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "len",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p185,
													FreeVars: ast.Identifiers{
														"len",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(41),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(41),
															Column: int(28),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p157,
									FreeVars: ast.Identifiers{
										"len",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(41),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(41),
											Column: int(29),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Conditional{
								Cond: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(42),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(20),
												},
											},
										},
									},
									Left: &ast.Var{
										Id: "len",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{
												"len",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(42),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(15),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"len",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(42),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(42),
												Column: int(20),
											},
										},
									},
									Op: ast.BinaryOp(8),
								},
								BranchTrue: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(43),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(43),
														Column: int(8),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "join",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(43),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(43),
													Column: int(13),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p203,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(43),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(43),
																Column: int(16),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(43),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(43),
																		Column: int(21),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "makeArray",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p203,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(43),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(43),
																	Column: int(31),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(35),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "max",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p216,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(39),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNumber{
																					OriginalString: "0",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p220,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(43),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(43),
																								Column: int(41),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Apply{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "std",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(43),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(43),
																										Column: int(46),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "min",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p220,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(43),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(43),
																									Column: int(50),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "len",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p229,
																										FreeVars: ast.Identifiers{
																											"len",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(43),
																												Column: int(51),
																											},
																											End: ast.Location{
																												Line: int(43),
																												Column: int(54),
																											},
																										},
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.Var{
																										Id: "from",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p229,
																											FreeVars: ast.Identifiers{
																												"from",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(43),
																													Column: int(74),
																												},
																												End: ast.Location{
																													Line: int(43),
																													Column: int(78),
																												},
																											},
																										},
																									},
																									Left: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(43),
																															Column: int(56),
																														},
																														End: ast.Location{
																															Line: int(43),
																															Column: int(59),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "length",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p229,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(43),
																														Column: int(56),
																													},
																													End: ast.Location{
																														Line: int(43),
																														Column: int(66),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "str",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p242,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(43),
																																	Column: int(67),
																																},
																																End: ast.Location{
																																	Line: int(43),
																																	Column: int(70),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p229,
																											FreeVars: ast.Identifiers{
																												"std",
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(43),
																													Column: int(56),
																												},
																												End: ast.Location{
																													Line: int(43),
																													Column: int(71),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p229,
																										FreeVars: ast.Identifiers{
																											"from",
																											"std",
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(43),
																												Column: int(56),
																											},
																											End: ast.Location{
																												Line: int(43),
																												Column: int(78),
																											},
																										},
																									},
																									Op: ast.BinaryOp(4),
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p220,
																						FreeVars: ast.Identifiers{
																							"from",
																							"len",
																							"std",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(43),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(43),
																								Column: int(79),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p216,
																		FreeVars: ast.Identifiers{
																			"from",
																			"len",
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(43),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(43),
																				Column: int(80),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Function{
																	ParenLeftFodder: ast.Fodder{},
																	ParenRightFodder: ast.Fodder{},
																	Body: &ast.Index{
																		Target: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p251,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(94),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(97),
																					},
																				},
																			},
																		},
																		Index: &ast.Binary{
																			Right: &ast.Var{
																				Id: "from",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p251,
																					FreeVars: ast.Identifiers{
																						"from",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(43),
																							Column: int(102),
																						},
																						End: ast.Location{
																							Line: int(43),
																							Column: int(106),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p251,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(43),
																							Column: int(98),
																						},
																						End: ast.Location{
																							Line: int(43),
																							Column: int(99),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p251,
																				FreeVars: ast.Identifiers{
																					"from",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(98),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(106),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p251,
																			FreeVars: ast.Identifiers{
																				"from",
																				"i",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(94),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(107),
																				},
																			},
																		},
																	},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "i",
																			CommaFodder: nil,
																			EqFodder: nil,
																			DefaultArg: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(91),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(92),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p216,
																		FreeVars: ast.Identifiers{
																			"from",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(43),
																				Column: int(82),
																			},
																			End: ast.Location{
																				Line: int(43),
																				Column: int(107),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p203,
														FreeVars: ast.Identifiers{
															"from",
															"len",
															"std",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(43),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(43),
																Column: int(108),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"from",
											"len",
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(43),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(43),
												Column: int(109),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchFalse: &ast.Error{
									Expr: &ast.Binary{
										Right: &ast.Var{
											Id: "len",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p157,
												FreeVars: ast.Identifiers{
													"len",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(42),
														Column: int(84),
													},
													End: ast.Location{
														Line: int(42),
														Column: int(87),
													},
												},
											},
										},
										Left: &ast.LiteralString{
											Value: "substr third parameter should be greater than zero, got ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p157,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(42),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(42),
														Column: int(81),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{
												"len",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(42),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(87),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"len",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(42),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(43),
												Column: int(109),
											},
										},
									},
								},
								ThenFodder: nil,
								ElseFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"from",
										"len",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(41),
															Column: int(84),
														},
														End: ast.Location{
															Line: int(41),
															Column: int(87),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "type",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p157,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(41),
														Column: int(84),
													},
													End: ast.Location{
														Line: int(41),
														Column: int(92),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "len",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p282,
															FreeVars: ast.Identifiers{
																"len",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(41),
																	Column: int(93),
																},
																End: ast.Location{
																	Line: int(41),
																	Column: int(96),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{
												"len",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(41),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(97),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "substr third parameter should be a string, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(41),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(81),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"len",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(41),
												Column: int(32),
											},
											End: ast.Location{
												Line: int(41),
												Column: int(97),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"len",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(41),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(43),
											Column: int(109),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"from",
									"len",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(40),
														Column: int(86),
													},
													End: ast.Location{
														Line: int(40),
														Column: int(89),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p157,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(40),
													Column: int(86),
												},
												End: ast.Location{
													Line: int(40),
													Column: int(94),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "from",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p299,
														FreeVars: ast.Identifiers{
															"from",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(40),
																Column: int(95),
															},
															End: ast.Location{
																Line: int(40),
																Column: int(99),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"from",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(40),
												Column: int(86),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(100),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "substr second parameter should be a string, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(40),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(83),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p157,
									FreeVars: ast.Identifiers{
										"from",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(40),
											Column: int(33),
										},
										End: ast.Location{
											Line: int(40),
											Column: int(100),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(40),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(43),
										Column: int(109),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"from",
								"len",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(39),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(39),
													Column: int(87),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p157,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(39),
												Column: int(84),
											},
											End: ast.Location{
												Line: int(39),
												Column: int(92),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p316,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(39),
															Column: int(93),
														},
														End: ast.Location{
															Line: int(39),
															Column: int(96),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p157,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(39),
											Column: int(84),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(97),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "substr first parameter should be a string, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p157,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(39),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(81),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p157,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(39),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(39),
										Column: int(97),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(39),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(43),
									Column: int(109),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"from",
							"len",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(38),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(38),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(19),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "len",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(38),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(38),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(43),
					Column: int(109),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "startsWith",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(46),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(46),
												Column: int(27),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "length",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p334,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(46),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(46),
											Column: int(34),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p338,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(46),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(46),
														Column: int(36),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p334,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(46),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(46),
										Column: int(37),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(46),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(46),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "length",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p334,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(46),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(46),
											Column: int(18),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p349,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(46),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(46),
														Column: int(20),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p334,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(46),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(46),
										Column: int(21),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p334,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(46),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(46),
									Column: int(37),
								},
							},
						},
						Op: ast.BinaryOp(9),
					},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p334,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(47),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(47),
									Column: int(12),
								},
							},
						},
						Value: false,
					},
					BranchFalse: &ast.Binary{
						Right: &ast.Var{
							Id: "b",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p334,
								FreeVars: ast.Identifiers{
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(49),
										Column: int(42),
									},
									End: ast.Location{
										Line: int(49),
										Column: int(43),
									},
								},
							},
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(49),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(49),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "substr",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p334,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(49),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(49),
											Column: int(17),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p367,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(49),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(19),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											OriginalString: "0",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p367,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(49),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(22),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(49),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(49),
																Column: int(27),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p367,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(49),
															Column: int(24),
														},
														End: ast.Location{
															Line: int(49),
															Column: int(34),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p378,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(49),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(49),
																		Column: int(36),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p367,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(49),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(37),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p334,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(49),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(49),
										Column: int(38),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p334,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(49),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(49),
									Column: int(43),
								},
							},
						},
						Op: ast.BinaryOp(12),
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p334,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(46),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(49),
								Column: int(43),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(45),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(45),
								Column: int(15),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(45),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(45),
								Column: int(18),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(45),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(49),
					Column: int(43),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "endsWith",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(52),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(52),
												Column: int(27),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "length",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p397,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(52),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(52),
											Column: int(34),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p401,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(52),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(52),
														Column: int(36),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p397,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(52),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(52),
										Column: int(37),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(52),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(52),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "length",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p397,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(52),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(52),
											Column: int(18),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p412,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(52),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(52),
														Column: int(20),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p397,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(52),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(52),
										Column: int(21),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p397,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(52),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(52),
									Column: int(37),
								},
							},
						},
						Op: ast.BinaryOp(9),
					},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p397,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(53),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(53),
									Column: int(12),
								},
							},
						},
						Value: false,
					},
					BranchFalse: &ast.Binary{
						Right: &ast.Var{
							Id: "b",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p397,
								FreeVars: ast.Identifiers{
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(55),
										Column: int(70),
									},
									End: ast.Location{
										Line: int(55),
										Column: int(71),
									},
								},
							},
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(55),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(55),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "substr",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p397,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(55),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(55),
											Column: int(17),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p430,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(55),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(19),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(55),
																	Column: int(37),
																},
																End: ast.Location{
																	Line: int(55),
																	Column: int(40),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p430,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(55),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(47),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p441,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(55),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(55),
																			Column: int(49),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p430,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(55),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(50),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(55),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(55),
																	Column: int(24),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p430,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(55),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(31),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p452,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(55),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(55),
																			Column: int(33),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p430,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(55),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(34),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p430,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(55),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(50),
													},
												},
											},
											Op: ast.BinaryOp(4),
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(55),
																Column: int(52),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(55),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p430,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(55),
															Column: int(52),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(62),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p464,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(55),
																		Column: int(63),
																	},
																	End: ast.Location{
																		Line: int(55),
																		Column: int(64),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p430,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(55),
														Column: int(52),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(65),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p397,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(55),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(55),
										Column: int(66),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p397,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(55),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(55),
									Column: int(71),
								},
							},
						},
						Op: ast.BinaryOp(12),
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p397,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(52),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(55),
								Column: int(71),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(51),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(51),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(51),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(51),
								Column: int(16),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(51),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(55),
					Column: int(71),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "lstripChars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(58),
												Column: int(31),
											},
											End: ast.Location{
												Line: int(58),
												Column: int(34),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "member",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p483,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(58),
											Column: int(31),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(41),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "chars",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p487,
												FreeVars: ast.Identifiers{
													"chars",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(58),
														Column: int(42),
													},
													End: ast.Location{
														Line: int(58),
														Column: int(47),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Index{
											Target: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p487,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(58),
															Column: int(49),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(52),
														},
													},
												},
											},
											Index: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p487,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(58),
															Column: int(53),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(54),
														},
													},
												},
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p487,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(58),
														Column: int(49),
													},
													End: ast.Location{
														Line: int(58),
														Column: int(55),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p483,
								FreeVars: ast.Identifiers{
									"chars",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(58),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(58),
										Column: int(56),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						Left: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p483,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(58),
											Column: int(26),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(27),
										},
									},
								},
							},
							Left: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(58),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(58),
													Column: int(11),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p483,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(58),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(58),
												Column: int(18),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p505,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(58),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(22),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p483,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(58),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(23),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p483,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(58),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(58),
										Column: int(27),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p483,
							FreeVars: ast.Identifiers{
								"chars",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(58),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(58),
									Column: int(56),
								},
							},
						},
						Op: ast.BinaryOp(17),
					},
					BranchTrue: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(59),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(59),
											Column: int(10),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "lstripChars",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p483,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(59),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(59),
										Column: int(22),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "$std",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "slice",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: nil,
											LeftBracketFodder: nil,
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "str",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p526,
															FreeVars: ast.Identifiers{
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(59),
																	Column: int(23),
																},
																End: ast.Location{
																	Line: int(59),
																	Column: int(26),
																},
															},
														},
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														OriginalString: "1",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p526,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(59),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(59),
																	Column: int(28),
																},
															},
														},
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: nil,
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(59),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(59),
													Column: int(30),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "chars",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p526,
											FreeVars: ast.Identifiers{
												"chars",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(59),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(59),
													Column: int(37),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p483,
							FreeVars: ast.Identifiers{
								"$std",
								"chars",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(59),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(59),
									Column: int(38),
								},
							},
						},
						TrailingComma: false,
						TailStrict: true,
					},
					BranchFalse: &ast.Var{
						Id: "str",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p483,
							FreeVars: ast.Identifiers{
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(61),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(61),
									Column: int(10),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p483,
						FreeVars: ast.Identifiers{
							"$std",
							"chars",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(58),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(61),
								Column: int(10),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(57),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(57),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(57),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(57),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(57),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(61),
					Column: int(10),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "rstripChars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(64),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(64),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p552,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(64),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(64),
												Column: int(27),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p556,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(64),
															Column: int(28),
														},
														End: ast.Location{
															Line: int(64),
															Column: int(31),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p552,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(64),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(64),
											Column: int(32),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "len",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(64),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(64),
									Column: int(32),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(65),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(65),
													Column: int(22),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "member",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p566,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(65),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(29),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "chars",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p570,
													FreeVars: ast.Identifiers{
														"chars",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(65),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(65),
															Column: int(35),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Index{
												Target: &ast.Var{
													Id: "str",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p570,
														FreeVars: ast.Identifiers{
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(65),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(65),
																Column: int(40),
															},
														},
													},
												},
												Index: &ast.Binary{
													Right: &ast.LiteralNumber{
														OriginalString: "1",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p570,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(65),
																	Column: int(47),
																},
																End: ast.Location{
																	Line: int(65),
																	Column: int(48),
																},
															},
														},
													},
													Left: &ast.Var{
														Id: "len",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p570,
															FreeVars: ast.Identifiers{
																"len",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(65),
																	Column: int(41),
																},
																End: ast.Location{
																	Line: int(65),
																	Column: int(44),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p570,
														FreeVars: ast.Identifiers{
															"len",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(65),
																Column: int(41),
															},
															End: ast.Location{
																Line: int(65),
																Column: int(48),
															},
														},
													},
													Op: ast.BinaryOp(4),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p570,
													FreeVars: ast.Identifiers{
														"len",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(65),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(65),
															Column: int(49),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p566,
									FreeVars: ast.Identifiers{
										"chars",
										"len",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(65),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(65),
											Column: int(50),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p566,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(65),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(15),
											},
										},
									},
								},
								Left: &ast.Var{
									Id: "len",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p566,
										FreeVars: ast.Identifiers{
											"len",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(65),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(11),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p566,
									FreeVars: ast.Identifiers{
										"len",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(65),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(65),
											Column: int(15),
										},
									},
								},
								Op: ast.BinaryOp(7),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p566,
								FreeVars: ast.Identifiers{
									"chars",
									"len",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(65),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(65),
										Column: int(50),
									},
								},
							},
							Op: ast.BinaryOp(17),
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(66),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(66),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "rstripChars",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p566,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(66),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(66),
											Column: int(22),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "$std",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "slice",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: nil,
												LeftBracketFodder: nil,
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p604,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(66),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(66),
																		Column: int(26),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p604,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(66),
																			Column: int(34),
																		},
																		End: ast.Location{
																			Line: int(66),
																			Column: int(35),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "len",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p604,
																	FreeVars: ast.Identifiers{
																		"len",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(66),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(66),
																			Column: int(31),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p604,
																FreeVars: ast.Identifiers{
																	"len",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(66),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(66),
																		Column: int(35),
																	},
																},
															},
															Op: ast.BinaryOp(4),
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: nil,
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
													"len",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(66),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(66),
														Column: int(36),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "chars",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p604,
												FreeVars: ast.Identifiers{
													"chars",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(66),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(66),
														Column: int(43),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p566,
								FreeVars: ast.Identifiers{
									"$std",
									"chars",
									"len",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(66),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(66),
										Column: int(44),
									},
								},
							},
							TrailingComma: false,
							TailStrict: true,
						},
						BranchFalse: &ast.Var{
							Id: "str",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p566,
								FreeVars: ast.Identifiers{
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(68),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(68),
										Column: int(10),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p566,
							FreeVars: ast.Identifiers{
								"$std",
								"chars",
								"len",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(65),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(68),
									Column: int(10),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p566,
						FreeVars: ast.Identifiers{
							"$std",
							"chars",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(64),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(68),
								Column: int(10),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(63),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(63),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(63),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(63),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(63),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(68),
					Column: int(10),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "stripChars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(71),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(71),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "lstripChars",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p635,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(71),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(71),
									Column: int(20),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(71),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(71),
														Column: int(24),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "rstripChars",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p643,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(71),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(71),
													Column: int(36),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "str",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p647,
														FreeVars: ast.Identifiers{
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(71),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(71),
																Column: int(40),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "chars",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p647,
														FreeVars: ast.Identifiers{
															"chars",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(71),
																Column: int(42),
															},
															End: ast.Location{
																Line: int(71),
																Column: int(47),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p643,
										FreeVars: ast.Identifiers{
											"chars",
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(71),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(71),
												Column: int(48),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "chars",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p643,
										FreeVars: ast.Identifiers{
											"chars",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(71),
												Column: int(50),
											},
											End: ast.Location{
												Line: int(71),
												Column: int(55),
											},
										},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p635,
						FreeVars: ast.Identifiers{
							"chars",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(71),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(71),
								Column: int(56),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(70),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(70),
								Column: int(17),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(70),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(70),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(70),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(71),
					Column: int(56),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "stringChars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(74),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(74),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "makeArray",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p665,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(74),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(74),
									Column: int(18),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(74),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(22),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p673,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(74),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(29),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "str",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p677,
														FreeVars: ast.Identifiers{
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(74),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(74),
																Column: int(33),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p673,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(74),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(74),
												Column: int(34),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Index{
										Target: &ast.Var{
											Id: "str",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p683,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(74),
														Column: int(48),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(51),
													},
												},
											},
										},
										Index: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p683,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(74),
														Column: int(52),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(53),
													},
												},
											},
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p683,
											FreeVars: ast.Identifiers{
												"i",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(74),
													Column: int(48),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(54),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(74),
													Column: int(45),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(46),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p673,
										FreeVars: ast.Identifiers{
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(74),
												Column: int(36),
											},
											End: ast.Location{
												Line: int(74),
												Column: int(54),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p665,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(74),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(74),
								Column: int(55),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(73),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(73),
								Column: int(18),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(73),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(74),
					Column: int(55),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "parseInt",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(95),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p701,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(95),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(95),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p705,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(95),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(95),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p701,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(95),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(95),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Binary{
								Right: &ast.LiteralString{
									Value: "-",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(96),
												Column: int(42),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(45),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								Left: &ast.Var{
									Id: "str",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(96),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(38),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(96),
											Column: int(35),
										},
										End: ast.Location{
											Line: int(96),
											Column: int(45),
										},
									},
								},
								Op: ast.BinaryOp(13),
							},
							Left: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(96),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(31),
											},
										},
									},
								},
								Left: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(96),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(96),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p701,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(96),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(96),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "str",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p725,
														FreeVars: ast.Identifiers{
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(96),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(96),
																Column: int(26),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(96),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(27),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(96),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(96),
											Column: int(31),
										},
									},
								},
								Op: ast.BinaryOp(7),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p701,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(96),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(96),
										Column: int(45),
									},
								},
							},
							Op: ast.BinaryOp(17),
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralString{
									Value: "-",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(97),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(97),
												Column: int(21),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								Left: &ast.Index{
									Target: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p701,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(97),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(97),
													Column: int(11),
												},
											},
										},
									},
									Index: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p701,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(97),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(97),
													Column: int(13),
												},
											},
										},
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(97),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(97),
												Column: int(14),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(97),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(97),
											Column: int(21),
										},
									},
								},
								Op: ast.BinaryOp(12),
							},
							BranchTrue: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Var{
										Id: "parse_nat",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p701,
											FreeVars: ast.Identifiers{
												"parse_nat",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(98),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(98),
													Column: int(17),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "slice",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "str",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p752,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(98),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(98),
																				Column: int(21),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p752,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(98),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(98),
																				Column: int(23),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(98),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(98),
																Column: int(25),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													OriginalString: "10",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p752,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(98),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(98),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"$std",
											"parse_nat",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(98),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(98),
												Column: int(30),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"$std",
										"parse_nat",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(98),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(98),
											Column: int(30),
										},
									},
								},
								Op: ast.UnaryOp(3),
							},
							BranchFalse: &ast.Apply{
								Target: &ast.Var{
									Id: "parse_nat",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"parse_nat",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(100),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(100),
												Column: int(16),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p768,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(100),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(100),
															Column: int(20),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												OriginalString: "10",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p768,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(100),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(100),
															Column: int(24),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"parse_nat",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(100),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(100),
											Column: int(25),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p701,
								FreeVars: ast.Identifiers{
									"$std",
									"parse_nat",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(97),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(100),
										Column: int(25),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "mod",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "Not an integer: \"%s\"",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p701,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(96),
															Column: int(48),
														},
														End: ast.Location{
															Line: int(96),
															Column: int(70),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p787,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(96),
																		Column: int(74),
																	},
																	End: ast.Location{
																		Line: int(96),
																		Column: int(77),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												CloseFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p701,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(96),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(96),
															Column: int(78),
														},
													},
												},
												TrailingComma: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(96),
											Column: int(48),
										},
										End: ast.Location{
											Line: int(96),
											Column: int(78),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(96),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(100),
										Column: int(25),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
								"parse_nat",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(95),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(95),
													Column: int(61),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p701,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(95),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(95),
												Column: int(66),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p803,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(95),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(95),
															Column: int(70),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(95),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(71),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "Expected string, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p701,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(95),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(55),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p701,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(95),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(95),
										Column: int(71),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(95),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(100),
									Column: int(25),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"parse_nat",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(94),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(94),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"parse_nat",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(94),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(100),
					Column: int(25),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "parseOctal",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(103),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p820,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(103),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(103),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p824,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(103),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(103),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p820,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(103),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(103),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p820,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(104),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(31),
										},
									},
								},
							},
							Left: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(104),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(104),
													Column: int(15),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p820,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(104),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(104),
												Column: int(22),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p838,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(104),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(104),
															Column: int(26),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p820,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(104),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(27),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p820,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(104),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(104),
										Column: int(31),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Var{
								Id: "parse_nat",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p820,
									FreeVars: ast.Identifiers{
										"parse_nat",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(105),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(105),
											Column: int(14),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "str",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p848,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(105),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(105),
														Column: int(18),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											OriginalString: "8",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p848,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(105),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(105),
														Column: int(21),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p820,
								FreeVars: ast.Identifiers{
									"parse_nat",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(105),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(105),
										Column: int(22),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Error{
							Expr: &ast.LiteralString{
								Value: "Not an octal number: \"\"",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p820,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(104),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(59),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(104),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(105),
										Column: int(22),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"parse_nat",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(103),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(103),
													Column: int(61),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p820,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(103),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(103),
												Column: int(66),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p865,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(103),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(103),
															Column: int(70),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p820,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(103),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(71),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "Expected string, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p820,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(103),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(55),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p820,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(103),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(103),
										Column: int(71),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(103),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(105),
									Column: int(22),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"parse_nat",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(102),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(102),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"parse_nat",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(102),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(105),
					Column: int(22),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "parseHex",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(108),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p882,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(108),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(108),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p886,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(108),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(108),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p882,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(108),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(108),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p882,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(109),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(31),
										},
									},
								},
							},
							Left: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(109),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(109),
													Column: int(15),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p882,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(109),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(109),
												Column: int(22),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p900,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(109),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(109),
															Column: int(26),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p882,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(109),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(27),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p882,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(109),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(109),
										Column: int(31),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Var{
								Id: "parse_nat",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p882,
									FreeVars: ast.Identifiers{
										"parse_nat",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(110),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(110),
											Column: int(14),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "str",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p910,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(110),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(110),
														Column: int(18),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											OriginalString: "16",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p910,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(110),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(110),
														Column: int(22),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p882,
								FreeVars: ast.Identifiers{
									"parse_nat",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(110),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(110),
										Column: int(23),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Error{
							Expr: &ast.LiteralString{
								Value: "Not hexadecimal: \"\"",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p882,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(109),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(55),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(109),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(110),
										Column: int(23),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"parse_nat",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(108),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(108),
													Column: int(61),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p882,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(108),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(108),
												Column: int(66),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p927,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(108),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(108),
															Column: int(70),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p882,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(108),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(71),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "Expected string, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p882,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(108),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(55),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p882,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(108),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(108),
										Column: int(71),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(108),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(110),
									Column: int(23),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"parse_nat",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(107),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(107),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"parse_nat",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(107),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(110),
					Column: int(23),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "split",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(113),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p944,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(113),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(113),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p948,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(113),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(113),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p944,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(113),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(113),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(114),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(15),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(114),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(114),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "c",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p960,
												FreeVars: ast.Identifiers{
													"c",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(114),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(114),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p944,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(114),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(114),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "1",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(115),
												Column: int(29),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(30),
											},
										},
									},
								},
								Left: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(115),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(115),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p944,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(115),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p974,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(115),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(115),
																Column: int(24),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(115),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(25),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(115),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(115),
											Column: int(30),
										},
									},
								},
								Op: ast.BinaryOp(8),
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(116),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(116),
													Column: int(8),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "splitLimit",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(116),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(116),
												Column: int(19),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p987,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(116),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(23),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "c",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p987,
													FreeVars: ast.Identifiers{
														"c",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(116),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(26),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Unary{
												Expr: &ast.LiteralNumber{
													OriginalString: "1",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p987,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(116),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(116),
																Column: int(30),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p987,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(116),
															Column: int(28),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(30),
														},
													},
												},
												Op: ast.UnaryOp(3),
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{
										"c",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(116),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(116),
											Column: int(31),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(115),
															Column: int(100),
														},
														End: ast.Location{
															Line: int(115),
															Column: int(103),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p944,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(115),
														Column: int(100),
													},
													End: ast.Location{
														Line: int(115),
														Column: int(110),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "c",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1004,
															FreeVars: ast.Identifiers{
																"c",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(115),
																	Column: int(111),
																},
																End: ast.Location{
																	Line: int(115),
																	Column: int(112),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p944,
											FreeVars: ast.Identifiers{
												"c",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(115),
													Column: int(100),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(113),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "std.split second parameter must have length 1 or greater, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p944,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(115),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(97),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(115),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(113),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(115),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(116),
											Column: int(31),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(114),
														Column: int(84),
													},
													End: ast.Location{
														Line: int(114),
														Column: int(87),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p944,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(114),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(114),
													Column: int(92),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1021,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(114),
																Column: int(93),
															},
															End: ast.Location{
																Line: int(114),
																Column: int(94),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(114),
												Column: int(84),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(95),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.split second parameter must be a String, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(114),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(81),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(114),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(114),
											Column: int(95),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(114),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(116),
										Column: int(31),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"c",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(113),
													Column: int(85),
												},
												End: ast.Location{
													Line: int(113),
													Column: int(88),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p944,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(113),
												Column: int(85),
											},
											End: ast.Location{
												Line: int(113),
												Column: int(93),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1038,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(113),
															Column: int(94),
														},
														End: ast.Location{
															Line: int(113),
															Column: int(97),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(113),
											Column: int(85),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(98),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.split first parameter must be a String, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p944,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(113),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(82),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p944,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(113),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(113),
										Column: int(98),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(113),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(116),
									Column: int(31),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"c",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(112),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(112),
								Column: int(12),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(112),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(112),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(112),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(116),
					Column: int(31),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "splitLimit",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(119),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1055,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(119),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(119),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1059,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(119),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(119),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p1055,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(119),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(119),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(120),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(15),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1055,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(120),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(120),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "c",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1071,
												FreeVars: ast.Identifiers{
													"c",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(120),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(120),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1055,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(120),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(120),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "1",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(121),
												Column: int(29),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(30),
											},
										},
									},
								},
								Left: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(121),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(121),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(121),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1085,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(121),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(121),
																Column: int(24),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(121),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(25),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1055,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(121),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(121),
											Column: int(30),
										},
									},
								},
								Op: ast.BinaryOp(8),
							},
							BranchTrue: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(122),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isNumber",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(122),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(122),
													Column: int(24),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "maxsplits",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1098,
														FreeVars: ast.Identifiers{
															"maxsplits",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(122),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(122),
																Column: int(34),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(122),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(122),
												Column: int(35),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(123),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(123),
																	Column: int(23),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1108,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(123),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(30),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "str",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1112,
																	FreeVars: ast.Identifiers{
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(123),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(123),
																			Column: int(34),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1108,
													FreeVars: ast.Identifiers{
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(123),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(123),
															Column: int(35),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "strLen",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(123),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(123),
													Column: int(35),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(124),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(124),
																		Column: int(21),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1122,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(124),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(124),
																	Column: int(28),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "c",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1126,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(124),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(124),
																				Column: int(30),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1122,
														FreeVars: ast.Identifiers{
															"c",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(124),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(124),
																Column: int(31),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "cLen",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(124),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(124),
														Column: int(31),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Body: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Var{
																	Id: "strLen",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"strLen",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(126),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(126),
																				Column: int(23),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "idx",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"idx",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(126),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(126),
																				Column: int(13),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1135,
																	FreeVars: ast.Identifiers{
																		"idx",
																		"strLen",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(126),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(126),
																			Column: int(23),
																		},
																	},
																},
																Op: ast.BinaryOp(8),
															},
															BranchTrue: &ast.Binary{
																Right: &ast.Array{
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "val",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1144,
																					FreeVars: ast.Identifiers{
																						"val",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(127),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(127),
																							Column: int(19),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	CloseFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"val",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(127),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(127),
																				Column: int(20),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																Left: &ast.Var{
																	Id: "ret",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"ret",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(127),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(127),
																				Column: int(12),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1135,
																	FreeVars: ast.Identifiers{
																		"ret",
																		"val",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(127),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(127),
																			Column: int(20),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Binary{
																		Right: &ast.Binary{
																			Right: &ast.Var{
																				Id: "maxsplits",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1135,
																					FreeVars: ast.Identifiers{
																						"maxsplits",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(129),
																							Column: int(53),
																						},
																						End: ast.Location{
																							Line: int(129),
																							Column: int(62),
																						},
																					},
																				},
																			},
																			Left: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(129),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(129),
																									Column: int(38),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "length",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1135,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(45),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "ret",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1165,
																									FreeVars: ast.Identifiers{
																										"ret",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(129),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(129),
																											Column: int(49),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1135,
																					FreeVars: ast.Identifiers{
																						"ret",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(129),
																							Column: int(35),
																						},
																						End: ast.Location{
																							Line: int(129),
																							Column: int(50),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1135,
																				FreeVars: ast.Identifiers{
																					"maxsplits",
																					"ret",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(129),
																						Column: int(35),
																					},
																					End: ast.Location{
																						Line: int(129),
																						Column: int(62),
																					},
																				},
																			},
																			Op: ast.BinaryOp(9),
																		},
																		Left: &ast.Binary{
																			Right: &ast.Unary{
																				Expr: &ast.LiteralNumber{
																					OriginalString: "1",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1135,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(30),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(31),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1135,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(129),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(129),
																							Column: int(31),
																						},
																					},
																				},
																				Op: ast.UnaryOp(3),
																			},
																			Left: &ast.Var{
																				Id: "maxsplits",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1135,
																					FreeVars: ast.Identifiers{
																						"maxsplits",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(129),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(129),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1135,
																				FreeVars: ast.Identifiers{
																					"maxsplits",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(129),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(129),
																						Column: int(31),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1135,
																			FreeVars: ast.Identifiers{
																				"maxsplits",
																				"ret",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(129),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(129),
																					Column: int(62),
																				},
																			},
																		},
																		Op: ast.BinaryOp(18),
																	},
																	Left: &ast.Binary{
																		Right: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1135,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(40),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(41),
																					},
																				},
																			},
																		},
																		Left: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "$std",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"$std",
																						},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "slice",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: nil,
																				LeftBracketFodder: nil,
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			FodderLeft: nil,
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1135,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(128),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(128),
																										Column: int(18),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "idx",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1135,
																								FreeVars: ast.Identifiers{
																									"idx",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(128),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(128),
																										Column: int(22),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.Var{
																								Id: "cLen",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1135,
																									FreeVars: ast.Identifiers{
																										"cLen",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(128),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(128),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "idx",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1135,
																									FreeVars: ast.Identifiers{
																										"idx",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(128),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(128),
																											Column: int(26),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1135,
																								FreeVars: ast.Identifiers{
																									"cLen",
																									"idx",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(128),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(128),
																										Column: int(33),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(128),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(128),
																										Column: int(35),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: nil,
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"cLen",
																					"idx",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(36),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1135,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"c",
																				"cLen",
																				"idx",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(128),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(128),
																					Column: int(41),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"c",
																			"cLen",
																			"idx",
																			"maxsplits",
																			"ret",
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(128),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(129),
																				Column: int(63),
																			},
																		},
																	},
																	Op: ast.BinaryOp(17),
																},
																BranchTrue: &ast.Apply{
																	Target: &ast.Var{
																		Id: "aux",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p1135,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(130),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(130),
																					Column: int(12),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.Var{
																						Id: "cLen",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1207,
																							FreeVars: ast.Identifiers{
																								"cLen",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(130),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(130),
																									Column: int(23),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "idx",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1207,
																							FreeVars: ast.Identifiers{
																								"idx",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(130),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(130),
																									Column: int(16),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1207,
																						FreeVars: ast.Identifiers{
																							"cLen",
																							"idx",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(130),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(130),
																								Column: int(23),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.Array{
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "val",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p1216,
																										FreeVars: ast.Identifiers{
																											"val",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(130),
																												Column: int(32),
																											},
																											End: ast.Location{
																												Line: int(130),
																												Column: int(35),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						CloseFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1207,
																							FreeVars: ast.Identifiers{
																								"val",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(130),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(130),
																									Column: int(36),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					Left: &ast.Var{
																						Id: "ret",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1207,
																							FreeVars: ast.Identifiers{
																								"ret",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(130),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(130),
																									Column: int(28),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1207,
																						FreeVars: ast.Identifiers{
																							"ret",
																							"val",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(130),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(130),
																								Column: int(36),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					Value: "",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1207,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(130),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(130),
																								Column: int(40),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"cLen",
																			"idx",
																			"ret",
																			"val",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(130),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(130),
																				Column: int(41),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																BranchFalse: &ast.Apply{
																	Target: &ast.Var{
																		Id: "aux",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p1135,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(132),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(132),
																					Column: int(12),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1231,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(132),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(132),
																									Column: int(20),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "idx",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1231,
																							FreeVars: ast.Identifiers{
																								"idx",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(132),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(132),
																									Column: int(16),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1231,
																						FreeVars: ast.Identifiers{
																							"idx",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(132),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(132),
																								Column: int(20),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "ret",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1231,
																						FreeVars: ast.Identifiers{
																							"ret",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(132),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(132),
																								Column: int(25),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.Index{
																						Target: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1231,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(132),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(132),
																										Column: int(36),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "idx",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1231,
																								FreeVars: ast.Identifiers{
																									"idx",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(132),
																										Column: int(37),
																									},
																									End: ast.Location{
																										Line: int(132),
																										Column: int(40),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1231,
																							FreeVars: ast.Identifiers{
																								"idx",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(132),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(132),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "val",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1231,
																							FreeVars: ast.Identifiers{
																								"val",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(132),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(132),
																									Column: int(30),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1231,
																						FreeVars: ast.Identifiers{
																							"idx",
																							"str",
																							"val",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(132),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(132),
																								Column: int(41),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1135,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"idx",
																			"ret",
																			"str",
																			"val",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(132),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(132),
																				Column: int(42),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1135,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"aux",
																		"c",
																		"cLen",
																		"idx",
																		"maxsplits",
																		"ret",
																		"std",
																		"str",
																		"val",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(128),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(132),
																			Column: int(42),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																Ctx: p1135,
																FreeVars: ast.Identifiers{
																	"$std",
																	"aux",
																	"c",
																	"cLen",
																	"idx",
																	"maxsplits",
																	"ret",
																	"std",
																	"str",
																	"strLen",
																	"val",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(126),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(132),
																		Column: int(42),
																	},
																},
															},
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "idx",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(125),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(125),
																		Column: int(18),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "ret",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(125),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(125),
																		Column: int(23),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "val",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(125),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(125),
																		Column: int(28),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: p1254,
															FreeVars: ast.Identifiers{
																"$std",
																"aux",
																"c",
																"cLen",
																"maxsplits",
																"std",
																"str",
																"strLen",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(125),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(42),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: nil,
													Variable: "aux",
													CloseFodder: nil,
													Fun: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Body: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(4),
															},
														},
														Ctx: p1055,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(133),
																Column: int(5),
															},
															End: ast.Location{
																Line: int(133),
																Column: int(8),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1262,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(133),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(133),
																			Column: int(10),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Array{
																Elements: nil,
																CloseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1262,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(133),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(133),
																			Column: int(14),
																		},
																	},
																},
																TrailingComma: false,
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1262,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(133),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(133),
																			Column: int(18),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1055,
													FreeVars: ast.Identifiers{
														"aux",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(133),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(133),
															Column: int(19),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p1055,
												FreeVars: ast.Identifiers{
													"$std",
													"c",
													"cLen",
													"maxsplits",
													"std",
													"str",
													"strLen",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(125),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(133),
														Column: int(19),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
												},
											},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"$std",
												"c",
												"maxsplits",
												"std",
												"str",
												"strLen",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(124),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(133),
													Column: int(19),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"$std",
											"c",
											"maxsplits",
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(123),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(133),
												Column: int(19),
											},
										},
									},
								},
								BranchFalse: &ast.Error{
									Expr: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(122),
																Column: int(96),
															},
															End: ast.Location{
																Line: int(122),
																Column: int(99),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "type",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1055,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(122),
															Column: int(96),
														},
														End: ast.Location{
															Line: int(122),
															Column: int(104),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "maxsplits",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1282,
																FreeVars: ast.Identifiers{
																	"maxsplits",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(122),
																		Column: int(105),
																	},
																	End: ast.Location{
																		Line: int(122),
																		Column: int(114),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1055,
												FreeVars: ast.Identifiers{
													"maxsplits",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(122),
														Column: int(96),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(115),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.LiteralString{
											Value: "str.splitLimit third parameter must be a Number, got ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1055,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(122),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(93),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"maxsplits",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(122),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(122),
													Column: int(115),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(122),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(133),
												Column: int(19),
											},
										},
									},
								},
								ThenFodder: nil,
								ElseFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"c",
										"maxsplits",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(121),
															Column: int(105),
														},
														End: ast.Location{
															Line: int(121),
															Column: int(108),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1055,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(121),
														Column: int(105),
													},
													End: ast.Location{
														Line: int(121),
														Column: int(115),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "c",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1299,
															FreeVars: ast.Identifiers{
																"c",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(121),
																	Column: int(116),
																},
																End: ast.Location{
																	Line: int(121),
																	Column: int(117),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"c",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(121),
													Column: int(105),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(118),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "std.splitLimit second parameter must have length 1 or greater, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(121),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(102),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(121),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(118),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(121),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(133),
											Column: int(19),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"c",
									"maxsplits",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(120),
														Column: int(89),
													},
													End: ast.Location{
														Line: int(120),
														Column: int(92),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1055,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(120),
													Column: int(89),
												},
												End: ast.Location{
													Line: int(120),
													Column: int(97),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1316,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(120),
																Column: int(98),
															},
															End: ast.Location{
																Line: int(120),
																Column: int(99),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(120),
												Column: int(89),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(100),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "str.splitLimit second parameter must be a String, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(120),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(86),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1055,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(120),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(120),
											Column: int(100),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(120),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(133),
										Column: int(19),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
								"c",
								"maxsplits",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(119),
													Column: int(90),
												},
												End: ast.Location{
													Line: int(119),
													Column: int(93),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1055,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(119),
												Column: int(90),
											},
											End: ast.Location{
												Line: int(119),
												Column: int(98),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1333,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(119),
															Column: int(99),
														},
														End: ast.Location{
															Line: int(119),
															Column: int(102),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1055,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(119),
											Column: int(90),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(103),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "str.splitLimit first parameter must be a String, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1055,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(119),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(87),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1055,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(119),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(119),
										Column: int(103),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(119),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(133),
									Column: int(19),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"c",
							"maxsplits",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(118),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(17),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(118),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(20),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "maxsplits",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(118),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(31),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(118),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(133),
					Column: int(19),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "splitLimitR",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(136),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(136),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1350,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(136),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(136),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1354,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(136),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(136),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p1350,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(136),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(136),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(137),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(137),
												Column: int(15),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1350,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(137),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(137),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "c",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1366,
												FreeVars: ast.Identifiers{
													"c",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(137),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(137),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1350,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(137),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(137),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "1",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(138),
												Column: int(29),
											},
											End: ast.Location{
												Line: int(138),
												Column: int(30),
											},
										},
									},
								},
								Left: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(138),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(138),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(138),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1380,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(138),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(138),
																Column: int(24),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(138),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(138),
												Column: int(25),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1350,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(138),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(138),
											Column: int(30),
										},
									},
								},
								Op: ast.BinaryOp(8),
							},
							BranchTrue: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(139),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(139),
														Column: int(15),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isNumber",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(139),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(139),
													Column: int(24),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "maxsplits",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1393,
														FreeVars: ast.Identifiers{
															"maxsplits",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(139),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(139),
																Column: int(34),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(139),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(139),
												Column: int(35),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Unary{
											Expr: &ast.LiteralNumber{
												OriginalString: "1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1350,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(140),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(140),
															Column: int(23),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(140),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(140),
														Column: int(23),
													},
												},
											},
											Op: ast.UnaryOp(3),
										},
										Left: &ast.Var{
											Id: "maxsplits",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{
													"maxsplits",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(140),
														Column: int(8),
													},
													End: ast.Location{
														Line: int(140),
														Column: int(17),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"maxsplits",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(140),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(140),
													Column: int(23),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(141),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(141),
															Column: int(10),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "splitLimit",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(141),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(141),
														Column: int(21),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "str",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1412,
															FreeVars: ast.Identifiers{
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(141),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(141),
																	Column: int(25),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "c",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1412,
															FreeVars: ast.Identifiers{
																"c",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(141),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(141),
																	Column: int(28),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Unary{
														Expr: &ast.LiteralNumber{
															OriginalString: "1",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1412,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(141),
																		Column: int(31),
																	},
																	End: ast.Location{
																		Line: int(141),
																		Column: int(32),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1412,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(141),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(141),
																	Column: int(32),
																},
															},
														},
														Op: ast.UnaryOp(3),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"c",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(141),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(141),
													Column: int(33),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchFalse: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Body: &ast.Function{
													ParenLeftFodder: ast.Fodder{},
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(143),
																			Column: int(27),
																		},
																		End: ast.Location{
																			Line: int(143),
																			Column: int(30),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "join",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1427,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(143),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(143),
																		Column: int(35),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		Value: "",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1431,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(143),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(143),
																					Column: int(38),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(143),
																							Column: int(40),
																						},
																						End: ast.Location{
																							Line: int(143),
																							Column: int(43),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "reverse",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1431,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(143),
																						Column: int(40),
																					},
																					End: ast.Location{
																						Line: int(143),
																						Column: int(51),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(143),
																											Column: int(52),
																										},
																										End: ast.Location{
																											Line: int(143),
																											Column: int(55),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "stringChars",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1444,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(143),
																										Column: int(52),
																									},
																									End: ast.Location{
																										Line: int(143),
																										Column: int(67),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "str",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p1448,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(143),
																													Column: int(68),
																												},
																												End: ast.Location{
																													Line: int(143),
																													Column: int(71),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1444,
																							FreeVars: ast.Identifiers{
																								"std",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(143),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(143),
																									Column: int(72),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1431,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(143),
																					Column: int(40),
																				},
																				End: ast.Location{
																					Line: int(143),
																					Column: int(73),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1427,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(143),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(143),
																	Column: int(74),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "str",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(143),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(143),
																	Column: int(23),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: p1454,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(143),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(143),
																Column: int(74),
															},
														},
													},
													TrailingComma: false,
												},
												EqFodder: nil,
												Variable: "revStr",
												CloseFodder: nil,
												Fun: nil,
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(144),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(144),
																Column: int(10),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "map",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1350,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(144),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(144),
															Column: int(14),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Var{
																	Id: "revStr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1467,
																		FreeVars: ast.Identifiers{
																			"revStr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(144),
																				Column: int(27),
																			},
																			End: ast.Location{
																				Line: int(144),
																				Column: int(33),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "e",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1471,
																					FreeVars: ast.Identifiers{
																						"e",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(144),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(144),
																							Column: int(35),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1467,
																	FreeVars: ast.Identifiers{
																		"e",
																		"revStr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(144),
																			Column: int(27),
																		},
																		End: ast.Location{
																			Line: int(144),
																			Column: int(36),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "e",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(144),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(144),
																			Column: int(25),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1475,
																FreeVars: ast.Identifiers{
																	"revStr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(144),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(144),
																		Column: int(36),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(144),
																				Column: int(38),
																			},
																			End: ast.Location{
																				Line: int(144),
																				Column: int(41),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "reverse",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1475,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(144),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(144),
																			Column: int(49),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(144),
																								Column: int(50),
																							},
																							End: ast.Location{
																								Line: int(144),
																								Column: int(53),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "splitLimit",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1489,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(144),
																							Column: int(50),
																						},
																						End: ast.Location{
																							Line: int(144),
																							Column: int(64),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Var{
																								Id: "revStr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1494,
																									FreeVars: ast.Identifiers{
																										"revStr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(144),
																											Column: int(65),
																										},
																										End: ast.Location{
																											Line: int(144),
																											Column: int(71),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "str",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p1498,
																												FreeVars: ast.Identifiers{
																													"str",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(144),
																														Column: int(72),
																													},
																													End: ast.Location{
																														Line: int(144),
																														Column: int(75),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1494,
																								FreeVars: ast.Identifiers{
																									"revStr",
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(144),
																										Column: int(65),
																									},
																									End: ast.Location{
																										Line: int(144),
																										Column: int(76),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Var{
																								Id: "revStr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1494,
																									FreeVars: ast.Identifiers{
																										"revStr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(144),
																											Column: int(78),
																										},
																										End: ast.Location{
																											Line: int(144),
																											Column: int(84),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "c",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p1506,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(144),
																														Column: int(85),
																													},
																													End: ast.Location{
																														Line: int(144),
																														Column: int(86),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1494,
																								FreeVars: ast.Identifiers{
																									"c",
																									"revStr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(144),
																										Column: int(78),
																									},
																									End: ast.Location{
																										Line: int(144),
																										Column: int(87),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "maxsplits",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1494,
																								FreeVars: ast.Identifiers{
																									"maxsplits",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(144),
																										Column: int(89),
																									},
																									End: ast.Location{
																										Line: int(144),
																										Column: int(98),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1489,
																				FreeVars: ast.Identifiers{
																					"c",
																					"maxsplits",
																					"revStr",
																					"std",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(144),
																						Column: int(50),
																					},
																					End: ast.Location{
																						Line: int(144),
																						Column: int(99),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1475,
																FreeVars: ast.Identifiers{
																	"c",
																	"maxsplits",
																	"revStr",
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(144),
																		Column: int(38),
																	},
																	End: ast.Location{
																		Line: int(144),
																		Column: int(100),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{
													"c",
													"maxsplits",
													"revStr",
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(144),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(144),
														Column: int(101),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"c",
												"maxsplits",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(143),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(144),
													Column: int(101),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"c",
											"maxsplits",
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(140),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(144),
												Column: int(101),
											},
										},
									},
								},
								BranchFalse: &ast.Error{
									Expr: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(139),
																Column: int(97),
															},
															End: ast.Location{
																Line: int(139),
																Column: int(100),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "type",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1350,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(139),
															Column: int(97),
														},
														End: ast.Location{
															Line: int(139),
															Column: int(105),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "maxsplits",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1529,
																FreeVars: ast.Identifiers{
																	"maxsplits",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(139),
																		Column: int(106),
																	},
																	End: ast.Location{
																		Line: int(139),
																		Column: int(115),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{
													"maxsplits",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(139),
														Column: int(97),
													},
													End: ast.Location{
														Line: int(139),
														Column: int(116),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.LiteralString{
											Value: "str.splitLimitR third parameter must be a Number, got ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(139),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(139),
														Column: int(94),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"maxsplits",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(139),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(139),
													Column: int(116),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(139),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(144),
												Column: int(101),
											},
										},
									},
								},
								ThenFodder: nil,
								ElseFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"maxsplits",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(138),
															Column: int(106),
														},
														End: ast.Location{
															Line: int(138),
															Column: int(109),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1350,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(138),
														Column: int(106),
													},
													End: ast.Location{
														Line: int(138),
														Column: int(116),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "c",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1546,
															FreeVars: ast.Identifiers{
																"c",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(138),
																	Column: int(117),
																},
																End: ast.Location{
																	Line: int(138),
																	Column: int(118),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"c",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(138),
													Column: int(106),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(119),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "std.splitLimitR second parameter must have length 1 or greater, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(138),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(103),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(138),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(138),
												Column: int(119),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(138),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(144),
											Column: int(101),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"maxsplits",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(137),
														Column: int(90),
													},
													End: ast.Location{
														Line: int(137),
														Column: int(93),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1350,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(137),
													Column: int(90),
												},
												End: ast.Location{
													Line: int(137),
													Column: int(98),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1563,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(137),
																Column: int(99),
															},
															End: ast.Location{
																Line: int(137),
																Column: int(100),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(137),
												Column: int(90),
											},
											End: ast.Location{
												Line: int(137),
												Column: int(101),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "str.splitLimitR second parameter must be a String, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(137),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(137),
												Column: int(87),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1350,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(137),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(137),
											Column: int(101),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(137),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(144),
										Column: int(101),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"c",
								"maxsplits",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(136),
													Column: int(91),
												},
												End: ast.Location{
													Line: int(136),
													Column: int(94),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1350,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(136),
												Column: int(91),
											},
											End: ast.Location{
												Line: int(136),
												Column: int(99),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1580,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(136),
															Column: int(100),
														},
														End: ast.Location{
															Line: int(136),
															Column: int(103),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1350,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(136),
											Column: int(91),
										},
										End: ast.Location{
											Line: int(136),
											Column: int(104),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "str.splitLimitR first parameter must be a String, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1350,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(136),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(136),
											Column: int(88),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1350,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(136),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(136),
										Column: int(104),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(136),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(144),
									Column: int(101),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"c",
							"maxsplits",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(135),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(135),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(135),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(135),
								Column: int(21),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "maxsplits",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(135),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(135),
								Column: int(32),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(135),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(144),
					Column: int(101),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "strReplace",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(147),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(147),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1597,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(147),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(147),
										Column: int(24),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1601,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(147),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(147),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p1597,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(147),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(147),
									Column: int(29),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(148),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(148),
												Column: int(15),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1597,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(148),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(148),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "from",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1613,
												FreeVars: ast.Identifiers{
													"from",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(148),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(148),
														Column: int(29),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1597,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(148),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(148),
										Column: int(30),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(149),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(149),
													Column: int(15),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1597,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(149),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(149),
												Column: int(24),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "to",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1625,
													FreeVars: ast.Identifiers{
														"to",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(149),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(149),
															Column: int(27),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1597,
									FreeVars: ast.Identifiers{
										"std",
										"to",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(149),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(149),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Conditional{
								Cond: &ast.Binary{
									Right: &ast.LiteralString{
										Value: "",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1597,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(150),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(150),
													Column: int(22),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									Left: &ast.Var{
										Id: "from",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1597,
											FreeVars: ast.Identifiers{
												"from",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(150),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(150),
													Column: int(16),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p1597,
										FreeVars: ast.Identifiers{
											"from",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(150),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(150),
												Column: int(22),
											},
										},
									},
									Op: ast.BinaryOp(13),
								},
								BranchTrue: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(153),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(153),
																	Column: int(24),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1641,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(153),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(153),
																Column: int(31),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "str",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1645,
																	FreeVars: ast.Identifiers{
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(153),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(153),
																			Column: int(35),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1641,
													FreeVars: ast.Identifiers{
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(153),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(153),
															Column: int(36),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "str_len",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(153),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(153),
													Column: int(36),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(154),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(154),
																		Column: int(25),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1655,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(154),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(154),
																	Column: int(32),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "from",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1659,
																		FreeVars: ast.Identifiers{
																			"from",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(154),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(154),
																				Column: int(37),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1655,
														FreeVars: ast.Identifiers{
															"from",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(154),
																Column: int(22),
															},
															End: ast.Location{
																Line: int(154),
																Column: int(38),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "from_len",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(154),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(154),
														Column: int(38),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Body: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Binary{
															Right: &ast.Var{
																Id: "from",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1667,
																	FreeVars: ast.Identifiers{
																		"from",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(157),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(157),
																			Column: int(52),
																		},
																	},
																},
															},
															Left: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "$std",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "slice",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: nil,
																	LeftBracketFodder: nil,
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1667,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(157),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(157),
																							Column: int(28),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1667,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(157),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(157),
																							Column: int(30),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				Right: &ast.Var{
																					Id: "from_len",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1667,
																						FreeVars: ast.Identifiers{
																							"from_len",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(157),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(157),
																								Column: int(43),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1667,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(157),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(157),
																								Column: int(32),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1667,
																					FreeVars: ast.Identifiers{
																						"from_len",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(157),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(157),
																							Column: int(43),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNull{
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: nil,
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"from_len",
																		"i",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(157),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(157),
																			Column: int(44),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1667,
																FreeVars: ast.Identifiers{
																	"$std",
																	"from",
																	"from_len",
																	"i",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(157),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(157),
																		Column: int(52),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(157),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(157),
																		Column: int(21),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: p1690,
															FreeVars: ast.Identifiers{
																"$std",
																"from",
																"from_len",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(157),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(157),
																	Column: int(52),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: nil,
													Variable: "found_at",
													CloseFodder: nil,
													Fun: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Body: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Var{
																		Id: "str_len",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"str_len",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(162),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(162),
																					Column: int(30),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "curr_index",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"curr_index",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(162),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(162),
																					Column: int(20),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1698,
																		FreeVars: ast.Identifiers{
																			"curr_index",
																			"str_len",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(162),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(162),
																				Column: int(30),
																			},
																		},
																	},
																	Op: ast.BinaryOp(7),
																},
																BranchTrue: &ast.Binary{
																	Right: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "$std",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "slice",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: nil,
																			LeftBracketFodder: nil,
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1698,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(163),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(163),
																									Column: int(18),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "start_index",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1698,
																							FreeVars: ast.Identifiers{
																								"start_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(163),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(163),
																									Column: int(30),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "curr_index",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1698,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(163),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(163),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: nil,
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"curr_index",
																				"start_index",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(163),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(163),
																					Column: int(42),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	Left: &ast.Var{
																		Id: "acc",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"acc",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(163),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(163),
																					Column: int(12),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1698,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"acc",
																			"curr_index",
																			"start_index",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(163),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(163),
																				Column: int(42),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																BranchFalse: &ast.Conditional{
																	Cond: &ast.Apply{
																		Target: &ast.Var{
																			Id: "found_at",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1698,
																				FreeVars: ast.Identifiers{
																					"found_at",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(164),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(164),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "curr_index",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1729,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(164),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(164),
																									Column: int(34),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"curr_index",
																				"found_at",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(164),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(164),
																					Column: int(35),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	BranchTrue: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Body: &ast.Binary{
																					Right: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(165),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(165),
																											Column: int(43),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "length",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1740,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(165),
																										Column: int(40),
																									},
																									End: ast.Location{
																										Line: int(165),
																										Column: int(50),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "from",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p1744,
																											FreeVars: ast.Identifiers{
																												"from",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(165),
																													Column: int(51),
																												},
																												End: ast.Location{
																													Line: int(165),
																													Column: int(55),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1740,
																							FreeVars: ast.Identifiers{
																								"from",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(165),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(165),
																									Column: int(56),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					Left: &ast.Var{
																						Id: "curr_index",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1740,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(165),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(165),
																									Column: int(37),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p1740,
																						FreeVars: ast.Identifiers{
																							"curr_index",
																							"from",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(165),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(165),
																								Column: int(56),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				EqFodder: ast.Fodder{},
																				Variable: "new_index",
																				CloseFodder: ast.Fodder{},
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(165),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(165),
																						Column: int(56),
																					},
																				},
																			},
																		},
																		Body: &ast.Apply{
																			Target: &ast.Var{
																				Id: "replace_after",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p1698,
																					FreeVars: ast.Identifiers{
																						"replace_after",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(166),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(166),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "new_index",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1756,
																								FreeVars: ast.Identifiers{
																									"new_index",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(166),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(166),
																										Column: int(32),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "new_index",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1756,
																								FreeVars: ast.Identifiers{
																									"new_index",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(166),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(166),
																										Column: int(43),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.Var{
																								Id: "to",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1756,
																									FreeVars: ast.Identifiers{
																										"to",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(166),
																											Column: int(81),
																										},
																										End: ast.Location{
																											Line: int(166),
																											Column: int(83),
																										},
																									},
																								},
																							},
																							Left: &ast.Binary{
																								Right: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "$std",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "slice",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: nil,
																										LeftBracketFodder: nil,
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$std",
																											},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																									},
																									FodderLeft: nil,
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "str",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p1756,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(166),
																																Column: int(51),
																															},
																															End: ast.Location{
																																Line: int(166),
																																Column: int(54),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "start_index",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p1756,
																														FreeVars: ast.Identifiers{
																															"start_index",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(166),
																																Column: int(55),
																															},
																															End: ast.Location{
																																Line: int(166),
																																Column: int(66),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "curr_index",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p1756,
																														FreeVars: ast.Identifiers{
																															"curr_index",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(166),
																																Column: int(67),
																															},
																															End: ast.Location{
																																Line: int(166),
																																Column: int(77),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNull{
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: nil,
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"$std",
																											"curr_index",
																											"start_index",
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(166),
																												Column: int(51),
																											},
																											End: ast.Location{
																												Line: int(166),
																												Column: int(78),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								Left: &ast.Var{
																									Id: "acc",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p1756,
																										FreeVars: ast.Identifiers{
																											"acc",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(166),
																												Column: int(45),
																											},
																											End: ast.Location{
																												Line: int(166),
																												Column: int(48),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p1756,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"acc",
																										"curr_index",
																										"start_index",
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(166),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(166),
																											Column: int(78),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1756,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"acc",
																									"curr_index",
																									"start_index",
																									"str",
																									"to",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(166),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(166),
																										Column: int(83),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1698,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"acc",
																					"curr_index",
																					"new_index",
																					"replace_after",
																					"start_index",
																					"str",
																					"to",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(166),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(166),
																						Column: int(84),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: true,
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"acc",
																				"curr_index",
																				"from",
																				"replace_after",
																				"start_index",
																				"std",
																				"str",
																				"to",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(165),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(166),
																					Column: int(84),
																				},
																			},
																		},
																	},
																	BranchFalse: &ast.Apply{
																		Target: &ast.Var{
																			Id: "replace_after",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				Ctx: p1698,
																				FreeVars: ast.Identifiers{
																					"replace_after",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(168),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(168),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "start_index",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1792,
																							FreeVars: ast.Identifiers{
																								"start_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(168),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(168),
																									Column: int(34),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1792,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(168),
																										Column: int(49),
																									},
																									End: ast.Location{
																										Line: int(168),
																										Column: int(50),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "curr_index",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p1792,
																								FreeVars: ast.Identifiers{
																									"curr_index",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(168),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(168),
																										Column: int(46),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1792,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(168),
																									Column: int(36),
																								},
																								End: ast.Location{
																									Line: int(168),
																									Column: int(50),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "acc",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1792,
																							FreeVars: ast.Identifiers{
																								"acc",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(168),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(168),
																									Column: int(55),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1698,
																			FreeVars: ast.Identifiers{
																				"acc",
																				"curr_index",
																				"replace_after",
																				"start_index",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(168),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(168),
																					Column: int(56),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: true,
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1698,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"acc",
																			"curr_index",
																			"found_at",
																			"from",
																			"replace_after",
																			"start_index",
																			"std",
																			"str",
																			"to",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(164),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(168),
																				Column: int(56),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: p1698,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"acc",
																		"curr_index",
																		"found_at",
																		"from",
																		"replace_after",
																		"start_index",
																		"std",
																		"str",
																		"str_len",
																		"to",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(162),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(168),
																			Column: int(56),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "start_index",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(161),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(161),
																			Column: int(36),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "curr_index",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(161),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(161),
																			Column: int(48),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "acc",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(161),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(161),
																			Column: int(53),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: p1808,
																FreeVars: ast.Identifiers{
																	"$std",
																	"found_at",
																	"from",
																	"replace_after",
																	"std",
																	"str",
																	"str_len",
																	"to",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(161),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(168),
																		Column: int(56),
																	},
																},
															},
															TrailingComma: false,
														},
														EqFodder: nil,
														Variable: "replace_after",
														CloseFodder: nil,
														Fun: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Body: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "1",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1597,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(172),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(172),
																		Column: int(21),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "from_len",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1597,
																FreeVars: ast.Identifiers{
																	"from_len",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(172),
																		Column: int(8),
																	},
																	End: ast.Location{
																		Line: int(172),
																		Column: int(16),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1597,
															FreeVars: ast.Identifiers{
																"from_len",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(172),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(172),
																	Column: int(21),
																},
															},
														},
														Op: ast.BinaryOp(12),
													},
													BranchTrue: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(173),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(173),
																			Column: int(10),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "join",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1597,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(173),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(173),
																		Column: int(15),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "to",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1825,
																			FreeVars: ast.Identifiers{
																				"to",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(173),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(173),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(173),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(173),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "split",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1825,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(173),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(173),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1835,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(173),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(173),
																									Column: int(33),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "from",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p1835,
																							FreeVars: ast.Identifiers{
																								"from",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(173),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(173),
																									Column: int(39),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1825,
																			FreeVars: ast.Identifiers{
																				"from",
																				"std",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(173),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(173),
																					Column: int(40),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1597,
															FreeVars: ast.Identifiers{
																"from",
																"std",
																"str",
																"to",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(173),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(173),
																	Column: int(41),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchFalse: &ast.Apply{
														Target: &ast.Var{
															Id: "replace_after",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																Ctx: p1597,
																FreeVars: ast.Identifiers{
																	"replace_after",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(175),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(175),
																		Column: int(20),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1847,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(175),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(175),
																					Column: int(22),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1847,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(175),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(175),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		Value: "",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p1847,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(175),
																					Column: int(27),
																				},
																				End: ast.Location{
																					Line: int(175),
																					Column: int(29),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1597,
															FreeVars: ast.Identifiers{
																"replace_after",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(175),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(175),
																	Column: int(30),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(4),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(1),
																Indent: int(4),
															},
															ast.FodderElement{
																Comment: []string{
																	"// if from_len==1, then we replace by splitting and rejoining the",
																},
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(4),
															},
															ast.FodderElement{
																Comment: []string{
																	"// string which is much faster than recursing on replace_after",
																},
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(4),
															},
														},
														Ctx: p1597,
														FreeVars: ast.Identifiers{
															"from",
															"from_len",
															"replace_after",
															"std",
															"str",
															"to",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(172),
																Column: int(5),
															},
															End: ast.Location{
																Line: int(175),
																Column: int(30),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(1),
															Indent: int(4),
														},
														ast.FodderElement{
															Comment: []string{
																"// Return the remainder of 'str' starting with 'start_index' where",
															},
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
														},
														ast.FodderElement{
															Comment: []string{
																"// all occurrences of 'from' after 'curr_index' are replaced with 'to'.",
															},
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
														},
													},
													Ctx: p1597,
													FreeVars: ast.Identifiers{
														"$std",
														"found_at",
														"from",
														"from_len",
														"std",
														"str",
														"str_len",
														"to",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(161),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(175),
															Column: int(30),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(1),
														Indent: int(4),
													},
													ast.FodderElement{
														Comment: []string{
															"// True if from is at str[i].",
														},
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p1597,
												FreeVars: ast.Identifiers{
													"$std",
													"from",
													"from_len",
													"std",
													"str",
													"str_len",
													"to",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(157),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(175),
														Column: int(30),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
												},
											},
											Ctx: p1597,
											FreeVars: ast.Identifiers{
												"$std",
												"from",
												"std",
												"str",
												"str_len",
												"to",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(154),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(175),
													Column: int(30),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(1),
												Indent: int(4),
											},
											ast.FodderElement{
												Comment: []string{
													"// Cache for performance.",
												},
												Kind: ast.FodderKind(2),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p1597,
										FreeVars: ast.Identifiers{
											"$std",
											"from",
											"std",
											"str",
											"to",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(153),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(175),
												Column: int(30),
											},
										},
									},
								},
								BranchFalse: &ast.Error{
									Expr: &ast.LiteralString{
										Value: "'from' string must not be zero length.",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1597,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(150),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(150),
													Column: int(65),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(150),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(175),
												Column: int(30),
											},
										},
									},
								},
								ThenFodder: nil,
								ElseFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"from",
										"std",
										"str",
										"to",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							BranchFalse: &ast.Error{
								Expr: &ast.LiteralString{
									Value: "Assertion failed",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(149),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(175),
											Column: int(30),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"from",
									"std",
									"str",
									"to",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.LiteralString{
								Value: "Assertion failed",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(148),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(175),
										Column: int(30),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
								"from",
								"std",
								"str",
								"to",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.LiteralString{
							Value: "Assertion failed",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(147),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(175),
									Column: int(30),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"from",
							"std",
							"str",
							"to",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(146),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(146),
								Column: int(17),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(146),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(146),
								Column: int(23),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "to",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(146),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(146),
								Column: int(27),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(146),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(175),
					Column: int(30),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "asciiUpper",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(178),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(178),
												Column: int(19),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "codepoint",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1890,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(178),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(178),
											Column: int(29),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "cp",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(178),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(178),
									Column: int(29),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "123",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1899,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(179),
																Column: int(52),
															},
															End: ast.Location{
																Line: int(179),
																Column: int(55),
															},
														},
													},
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "cp",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1899,
															FreeVars: ast.Identifiers{
																"cp",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(179),
																	Column: int(44),
																},
																End: ast.Location{
																	Line: int(179),
																	Column: int(46),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "c",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1905,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(179),
																				Column: int(47),
																			},
																			End: ast.Location{
																				Line: int(179),
																				Column: int(48),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1899,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(179),
																Column: int(44),
															},
															End: ast.Location{
																Line: int(179),
																Column: int(49),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1899,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(179),
															Column: int(44),
														},
														End: ast.Location{
															Line: int(179),
															Column: int(55),
														},
													},
												},
												Op: ast.BinaryOp(9),
											},
											Left: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "97",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1899,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(179),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(179),
																Column: int(40),
															},
														},
													},
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "cp",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p1899,
															FreeVars: ast.Identifiers{
																"cp",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(179),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(179),
																	Column: int(31),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "c",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1916,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(179),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(179),
																				Column: int(33),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p1899,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(179),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(179),
																Column: int(34),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1899,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(179),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(179),
															Column: int(40),
														},
													},
												},
												Op: ast.BinaryOp(8),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1899,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(179),
														Column: int(29),
													},
													End: ast.Location{
														Line: int(179),
														Column: int(55),
													},
												},
											},
											Op: ast.BinaryOp(17),
										},
										BranchTrue: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(180),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(180),
																Column: int(10),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "char",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1899,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(180),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(180),
															Column: int(15),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "32",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1931,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(180),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(180),
																			Column: int(26),
																		},
																	},
																},
															},
															Left: &ast.Apply{
																Target: &ast.Var{
																	Id: "cp",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p1931,
																		FreeVars: ast.Identifiers{
																			"cp",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(180),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(180),
																				Column: int(18),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p1937,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(180),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(180),
																							Column: int(20),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1931,
																	FreeVars: ast.Identifiers{
																		"c",
																		"cp",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(180),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(180),
																			Column: int(21),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1931,
																FreeVars: ast.Identifiers{
																	"c",
																	"cp",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(180),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(180),
																		Column: int(26),
																	},
																},
															},
															Op: ast.BinaryOp(4),
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1899,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(180),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(180),
														Column: int(27),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchFalse: &ast.Var{
											Id: "c",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p1899,
												FreeVars: ast.Identifiers{
													"c",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(182),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(182),
														Column: int(8),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p1899,
											FreeVars: ast.Identifiers{
												"c",
												"cp",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(179),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(182),
													Column: int(8),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "c",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(179),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(179),
													Column: int(22),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p1948,
										FreeVars: ast.Identifiers{
											"cp",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(179),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(182),
												Column: int(8),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "up_letter",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(183),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(183),
												Column: int(8),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "join",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1956,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(183),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(183),
											Column: int(13),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1960,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(183),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(183),
														Column: int(16),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(183),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(183),
																Column: int(21),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "map",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(183),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(183),
															Column: int(25),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "up_letter",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1969,
																FreeVars: ast.Identifiers{
																	"up_letter",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(183),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(183),
																		Column: int(35),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(183),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(183),
																				Column: int(40),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "stringChars",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p1969,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(183),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(183),
																			Column: int(52),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p1979,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(183),
																						Column: int(53),
																					},
																					End: ast.Location{
																						Line: int(183),
																						Column: int(56),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p1969,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(183),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(183),
																		Column: int(57),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"std",
													"str",
													"up_letter",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(183),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(183),
														Column: int(58),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p1956,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"up_letter",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(183),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(183),
										Column: int(59),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p1956,
							FreeVars: ast.Identifiers{
								"cp",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(179),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(183),
									Column: int(59),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p1956,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(178),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(183),
								Column: int(59),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(177),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(177),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(177),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(183),
					Column: int(59),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "asciiLower",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(186),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(186),
												Column: int(19),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "codepoint",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p1998,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(186),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(186),
											Column: int(29),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "cp",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(186),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(186),
									Column: int(29),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "91",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2007,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(187),
																Column: int(54),
															},
															End: ast.Location{
																Line: int(187),
																Column: int(56),
															},
														},
													},
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "cp",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2007,
															FreeVars: ast.Identifiers{
																"cp",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(187),
																	Column: int(46),
																},
																End: ast.Location{
																	Line: int(187),
																	Column: int(48),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "c",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p2013,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(187),
																				Column: int(49),
																			},
																			End: ast.Location{
																				Line: int(187),
																				Column: int(50),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2007,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(187),
																Column: int(46),
															},
															End: ast.Location{
																Line: int(187),
																Column: int(51),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2007,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(187),
															Column: int(46),
														},
														End: ast.Location{
															Line: int(187),
															Column: int(56),
														},
													},
												},
												Op: ast.BinaryOp(9),
											},
											Left: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "65",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2007,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(187),
																Column: int(40),
															},
															End: ast.Location{
																Line: int(187),
																Column: int(42),
															},
														},
													},
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "cp",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2007,
															FreeVars: ast.Identifiers{
																"cp",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(187),
																	Column: int(31),
																},
																End: ast.Location{
																	Line: int(187),
																	Column: int(33),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "c",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p2024,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(187),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(187),
																				Column: int(35),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2007,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(187),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(187),
																Column: int(36),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2007,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(187),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(187),
															Column: int(42),
														},
													},
												},
												Op: ast.BinaryOp(8),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2007,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(187),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(187),
														Column: int(56),
													},
												},
											},
											Op: ast.BinaryOp(17),
										},
										BranchTrue: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(188),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(188),
																Column: int(10),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "char",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2007,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(188),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(188),
															Column: int(15),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "32",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2039,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(188),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(188),
																			Column: int(26),
																		},
																	},
																},
															},
															Left: &ast.Apply{
																Target: &ast.Var{
																	Id: "cp",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p2039,
																		FreeVars: ast.Identifiers{
																			"cp",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(188),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(188),
																				Column: int(18),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p2045,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(188),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(188),
																							Column: int(20),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2039,
																	FreeVars: ast.Identifiers{
																		"c",
																		"cp",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(188),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(188),
																			Column: int(21),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2039,
																FreeVars: ast.Identifiers{
																	"c",
																	"cp",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(188),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(188),
																		Column: int(26),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2007,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(188),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(188),
														Column: int(27),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchFalse: &ast.Var{
											Id: "c",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p2007,
												FreeVars: ast.Identifiers{
													"c",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(190),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(190),
														Column: int(8),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2007,
											FreeVars: ast.Identifiers{
												"c",
												"cp",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(187),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(190),
													Column: int(8),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "c",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(187),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(187),
													Column: int(24),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p2056,
										FreeVars: ast.Identifiers{
											"cp",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(187),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(190),
												Column: int(8),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "down_letter",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(191),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(191),
												Column: int(8),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "join",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2064,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(191),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(191),
											Column: int(13),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2068,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(191),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(191),
														Column: int(16),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(191),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(191),
																Column: int(21),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "map",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2068,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(191),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(191),
															Column: int(25),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "down_letter",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2077,
																FreeVars: ast.Identifiers{
																	"down_letter",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(191),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(191),
																		Column: int(37),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(191),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(191),
																				Column: int(42),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "stringChars",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2077,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(191),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(191),
																			Column: int(54),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p2087,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(191),
																						Column: int(55),
																					},
																					End: ast.Location{
																						Line: int(191),
																						Column: int(58),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2077,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(191),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(191),
																		Column: int(59),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2068,
												FreeVars: ast.Identifiers{
													"down_letter",
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(191),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(191),
														Column: int(60),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2064,
								FreeVars: ast.Identifiers{
									"down_letter",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(191),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(191),
										Column: int(61),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p2064,
							FreeVars: ast.Identifiers{
								"cp",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(187),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(191),
									Column: int(61),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2064,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(186),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(191),
								Column: int(61),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(185),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(185),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(185),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(191),
					Column: int(61),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "range",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(194),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(194),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "makeArray",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2106,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(194),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(194),
									Column: int(18),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "1",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2111,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(194),
													Column: int(31),
												},
												End: ast.Location{
													Line: int(194),
													Column: int(32),
												},
											},
										},
									},
									Left: &ast.Binary{
										Right: &ast.Var{
											Id: "from",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2111,
												FreeVars: ast.Identifiers{
													"from",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(194),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(194),
														Column: int(28),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "to",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2111,
												FreeVars: ast.Identifiers{
													"to",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(194),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(194),
														Column: int(21),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2111,
											FreeVars: ast.Identifiers{
												"from",
												"to",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(194),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(194),
													Column: int(28),
												},
											},
										},
										Op: ast.BinaryOp(4),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2111,
										FreeVars: ast.Identifiers{
											"from",
											"to",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(194),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(194),
												Column: int(32),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										Right: &ast.Var{
											Id: "from",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2122,
												FreeVars: ast.Identifiers{
													"from",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(194),
														Column: int(50),
													},
													End: ast.Location{
														Line: int(194),
														Column: int(54),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2122,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(194),
														Column: int(46),
													},
													End: ast.Location{
														Line: int(194),
														Column: int(47),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2122,
											FreeVars: ast.Identifiers{
												"from",
												"i",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(194),
													Column: int(46),
												},
												End: ast.Location{
													Line: int(194),
													Column: int(54),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(194),
													Column: int(43),
												},
												End: ast.Location{
													Line: int(194),
													Column: int(44),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2111,
										FreeVars: ast.Identifiers{
											"from",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(194),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(194),
												Column: int(54),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p2106,
						FreeVars: ast.Identifiers{
							"from",
							"std",
							"to",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(194),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(194),
								Column: int(55),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(193),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(193),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "to",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(193),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(193),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(193),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(194),
					Column: int(55),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "repeat",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(198),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(198),
														Column: int(13),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isString",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2142,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(198),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(198),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "what",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2146,
														FreeVars: ast.Identifiers{
															"what",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(198),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(198),
																Column: int(27),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2142,
										FreeVars: ast.Identifiers{
											"std",
											"what",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(198),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(198),
												Column: int(28),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.LiteralString{
									Value: "",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2142,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(198),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(198),
												Column: int(36),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								BranchFalse: &ast.Conditional{
									Cond: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(199),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(199),
															Column: int(18),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isArray",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2142,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(199),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(199),
														Column: int(26),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "what",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2159,
															FreeVars: ast.Identifiers{
																"what",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(199),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(199),
																	Column: int(31),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2142,
											FreeVars: ast.Identifiers{
												"std",
												"what",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(199),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(199),
													Column: int(32),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchTrue: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2142,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(199),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(199),
													Column: int(40),
												},
											},
										},
										TrailingComma: false,
									},
									BranchFalse: &ast.Error{
										Expr: &ast.LiteralString{
											Value: "std.repeat first argument must be an array or a string",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2142,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(200),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(200),
														Column: int(74),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2142,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(200),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(200),
													Column: int(74),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2142,
										FreeVars: ast.Identifiers{
											"std",
											"what",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(199),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(200),
												Column: int(74),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p2142,
									FreeVars: ast.Identifiers{
										"std",
										"what",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(198),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(200),
											Column: int(74),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "joiner",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(197),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(200),
									Column: int(74),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(201),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(201),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "join",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2176,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(201),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(201),
										Column: int(13),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "joiner",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2180,
											FreeVars: ast.Identifiers{
												"joiner",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(201),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(201),
													Column: int(20),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(201),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(201),
															Column: int(25),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "makeArray",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2180,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(201),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(201),
														Column: int(35),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "count",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2190,
															FreeVars: ast.Identifiers{
																"count",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(201),
																	Column: int(36),
																},
																End: ast.Location{
																	Line: int(201),
																	Column: int(41),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Var{
															Id: "what",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2194,
																FreeVars: ast.Identifiers{
																	"what",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(201),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(201),
																		Column: int(59),
																	},
																},
															},
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(201),
																		Column: int(52),
																	},
																	End: ast.Location{
																		Line: int(201),
																		Column: int(53),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2190,
															FreeVars: ast.Identifiers{
																"what",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(201),
																	Column: int(43),
																},
																End: ast.Location{
																	Line: int(201),
																	Column: int(59),
																},
															},
														},
														TrailingComma: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2180,
											FreeVars: ast.Identifiers{
												"count",
												"std",
												"what",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(201),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(201),
													Column: int(60),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2176,
							FreeVars: ast.Identifiers{
								"count",
								"joiner",
								"std",
								"what",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(201),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(201),
									Column: int(61),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2176,
						FreeVars: ast.Identifiers{
							"count",
							"std",
							"what",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(197),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(201),
								Column: int(61),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "what",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(196),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(196),
								Column: int(14),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "count",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(196),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(196),
								Column: int(21),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(196),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(201),
					Column: int(61),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "slice",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.DesugaredObject{
								Asserts: ast.Nodes{},
								Fields: ast.DesugaredObjectFields{
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "indexable",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Var{
											Id: "indexable",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"indexable",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(207),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(207),
														Column: int(29),
													},
												},
											},
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(207),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(207),
												Column: int(29),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "index",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(209),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(209),
																Column: int(27),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "index",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{
															"index",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(209),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(209),
																Column: int(19),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"index",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(209),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(209),
															Column: int(27),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(209),
															Column: int(33),
														},
														End: ast.Location{
															Line: int(209),
															Column: int(34),
														},
													},
												},
											},
											BranchFalse: &ast.Var{
												Id: "index",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"index",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(210),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(210),
															Column: int(21),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"index",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(209),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(210),
														Column: int(21),
													},
												},
											},
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(208),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(210),
												Column: int(21),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "end",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(212),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(212),
																Column: int(25),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "end",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{
															"end",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(212),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(212),
																Column: int(17),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"end",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(212),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(212),
															Column: int(25),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(212),
																	Column: int(31),
																},
																End: ast.Location{
																	Line: int(212),
																	Column: int(34),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(212),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(212),
																Column: int(41),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "indexable",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2242,
																	FreeVars: ast.Identifiers{
																		"indexable",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(212),
																			Column: int(42),
																		},
																		End: ast.Location{
																			Line: int(212),
																			Column: int(51),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"indexable",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(212),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(212),
															Column: int(52),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											BranchFalse: &ast.Var{
												Id: "end",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"end",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(213),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(213),
															Column: int(19),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"end",
													"indexable",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(212),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(213),
														Column: int(19),
													},
												},
											},
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(211),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(213),
												Column: int(19),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "step",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(215),
																Column: int(22),
															},
															End: ast.Location{
																Line: int(215),
																Column: int(26),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "step",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2212,
														FreeVars: ast.Identifiers{
															"step",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(215),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(215),
																Column: int(18),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"step",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(215),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(215),
															Column: int(26),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralNumber{
												OriginalString: "1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(215),
															Column: int(32),
														},
														End: ast.Location{
															Line: int(215),
															Column: int(33),
														},
													},
												},
											},
											BranchFalse: &ast.Var{
												Id: "step",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"step",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(216),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(216),
															Column: int(20),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"step",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(215),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(216),
														Column: int(20),
													},
												},
											},
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(214),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(216),
												Column: int(20),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(217),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(217),
																Column: int(20),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(217),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(217),
															Column: int(27),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "indexable",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2272,
																FreeVars: ast.Identifiers{
																	"indexable",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(217),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(217),
																		Column: int(37),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"indexable",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(217),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(217),
														Column: int(38),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(217),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(217),
												Column: int(38),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
									ast.DesugaredObjectField{
										Name: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(218),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(218),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "type",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2212,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(218),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(218),
															Column: int(23),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "indexable",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2284,
																FreeVars: ast.Identifiers{
																	"indexable",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(218),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(218),
																		Column: int(33),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2212,
												FreeVars: ast.Identifiers{
													"indexable",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(218),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(218),
														Column: int(34),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(218),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(218),
												Column: int(34),
											},
										},
										Hide: ast.ObjectFieldHide(1),
										PlusSuper: false,
									},
								},
								Locals: ast.LocalBinds{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
										ast.FodderElement{
											Comment: []string{
												"// loop invariant with defaults applied",
											},
											Kind: ast.FodderKind(2),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p2289,
									FreeVars: ast.Identifiers{
										"end",
										"index",
										"indexable",
										"std",
										"step",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(206),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(219),
											Column: int(8),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "invar",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(204),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(219),
									Column: int(8),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(220),
												Column: int(64),
											},
											End: ast.Location{
												Line: int(220),
												Column: int(65),
											},
										},
									},
								},
								Left: &ast.Index{
									Target: &ast.Var{
										Id: "invar",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"invar",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(220),
													Column: int(50),
												},
												End: ast.Location{
													Line: int(220),
													Column: int(55),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "step",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"invar",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(220),
												Column: int(50),
											},
											End: ast.Location{
												Line: int(220),
												Column: int(60),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2295,
									FreeVars: ast.Identifiers{
										"invar",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(220),
											Column: int(50),
										},
										End: ast.Location{
											Line: int(220),
											Column: int(65),
										},
									},
								},
								Op: ast.BinaryOp(8),
							},
							Left: &ast.Binary{
								Right: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(220),
													Column: int(45),
												},
												End: ast.Location{
													Line: int(220),
													Column: int(46),
												},
											},
										},
									},
									Left: &ast.Index{
										Target: &ast.Var{
											Id: "invar",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"invar",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(220),
														Column: int(32),
													},
													End: ast.Location{
														Line: int(220),
														Column: int(37),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "end",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{
												"invar",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(220),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(220),
													Column: int(41),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"invar",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(220),
												Column: int(32),
											},
											End: ast.Location{
												Line: int(220),
												Column: int(46),
											},
										},
									},
									Op: ast.BinaryOp(8),
								},
								Left: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(220),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(220),
													Column: int(28),
												},
											},
										},
									},
									Left: &ast.Index{
										Target: &ast.Var{
											Id: "invar",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"invar",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(220),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(220),
														Column: int(17),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "index",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{
												"invar",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(220),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(220),
													Column: int(23),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"invar",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(220),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(220),
												Column: int(28),
											},
										},
									},
									Op: ast.BinaryOp(8),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2295,
									FreeVars: ast.Identifiers{
										"invar",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(220),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(220),
											Column: int(46),
										},
									},
								},
								Op: ast.BinaryOp(17),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2295,
								FreeVars: ast.Identifiers{
									"invar",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(220),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(220),
										Column: int(65),
									},
								},
							},
							Op: ast.BinaryOp(17),
						},
						BranchTrue: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(221),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(221),
												Column: int(21),
											},
										},
									},
								},
								Left: &ast.Var{
									Id: "step",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"step",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(221),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(221),
												Column: int(16),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2295,
									FreeVars: ast.Identifiers{
										"step",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(221),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(221),
											Column: int(21),
										},
									},
								},
								Op: ast.BinaryOp(13),
							},
							BranchTrue: &ast.Conditional{
								Cond: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(222),
															Column: int(39),
														},
														End: ast.Location{
															Line: int(222),
															Column: int(42),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isArray",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2295,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(222),
														Column: int(39),
													},
													End: ast.Location{
														Line: int(222),
														Column: int(50),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "indexable",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2337,
															FreeVars: ast.Identifiers{
																"indexable",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(51),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(60),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{
												"indexable",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(222),
													Column: int(39),
												},
												End: ast.Location{
													Line: int(222),
													Column: int(61),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(222),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(222),
															Column: int(15),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isString",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2295,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(222),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(222),
														Column: int(24),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "indexable",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2348,
															FreeVars: ast.Identifiers{
																"indexable",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(34),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{
												"indexable",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(222),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(222),
													Column: int(35),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"indexable",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(222),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(222),
												Column: int(61),
											},
										},
									},
									Op: ast.BinaryOp(18),
								},
								BranchTrue: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Body: &ast.Function{
												ParenLeftFodder: ast.Fodder{},
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.Binary{
															Right: &ast.Index{
																Target: &ast.Var{
																	Id: "invar",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(224),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(224),
																				Column: int(42),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "length",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2362,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(224),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(224),
																			Column: int(49),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "cur",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2362,
																	FreeVars: ast.Identifiers{
																		"cur",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(224),
																			Column: int(30),
																		},
																		End: ast.Location{
																			Line: int(224),
																			Column: int(33),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2362,
																FreeVars: ast.Identifiers{
																	"cur",
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(224),
																		Column: int(30),
																	},
																	End: ast.Location{
																		Line: int(224),
																		Column: int(49),
																	},
																},
															},
															Op: ast.BinaryOp(8),
														},
														Left: &ast.Binary{
															Right: &ast.Index{
																Target: &ast.Var{
																	Id: "invar",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(224),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(224),
																				Column: int(22),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "end",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2362,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(224),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(224),
																			Column: int(26),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "cur",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2362,
																	FreeVars: ast.Identifiers{
																		"cur",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(224),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(224),
																			Column: int(13),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2362,
																FreeVars: ast.Identifiers{
																	"cur",
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(224),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(224),
																		Column: int(26),
																	},
																},
															},
															Op: ast.BinaryOp(8),
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2362,
															FreeVars: ast.Identifiers{
																"cur",
																"invar",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(224),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(224),
																	Column: int(49),
																},
															},
														},
														Op: ast.BinaryOp(18),
													},
													BranchTrue: &ast.Var{
														Id: "slice",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p2362,
															FreeVars: ast.Identifiers{
																"slice",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(225),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(225),
																	Column: int(14),
																},
															},
														},
													},
													BranchFalse: &ast.Apply{
														Target: &ast.Var{
															Id: "build",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p2362,
																FreeVars: ast.Identifiers{
																	"build",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(227),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(227),
																		Column: int(14),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "string",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(228),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(228),
																							Column: int(36),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Index{
																				Target: &ast.Var{
																					Id: "invar",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"invar",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(228),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(228),
																								Column: int(19),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "type",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{
																						"invar",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(228),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(228),
																							Column: int(24),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p2388,
																				FreeVars: ast.Identifiers{
																					"invar",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(228),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(228),
																						Column: int(36),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		BranchTrue: &ast.Binary{
																			Right: &ast.Index{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "invar",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"invar",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(229),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(229),
																									Column: int(26),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "indexable",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p2388,
																						FreeVars: ast.Identifiers{
																							"invar",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(229),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(229),
																								Column: int(36),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "cur",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p2388,
																						FreeVars: ast.Identifiers{
																							"cur",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(229),
																								Column: int(37),
																							},
																							End: ast.Location{
																								Line: int(229),
																								Column: int(40),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"invar",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(229),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(229),
																							Column: int(41),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "slice",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																						},
																					},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{
																						"slice",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(229),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(229),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p2388,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"invar",
																					"slice",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(229),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(229),
																						Column: int(41),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		BranchFalse: &ast.Binary{
																			Right: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "invar",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"invar",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(231),
																												Column: int(22),
																											},
																											End: ast.Location{
																												Line: int(231),
																												Column: int(27),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "indexable",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p2417,
																									FreeVars: ast.Identifiers{
																										"invar",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(231),
																											Column: int(22),
																										},
																										End: ast.Location{
																											Line: int(231),
																											Column: int(37),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "cur",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p2417,
																									FreeVars: ast.Identifiers{
																										"cur",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(231),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(231),
																											Column: int(41),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p2417,
																								FreeVars: ast.Identifiers{
																									"cur",
																									"invar",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(231),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(231),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"invar",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(231),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(231),
																							Column: int(43),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			Left: &ast.Var{
																				Id: "slice",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																						},
																					},
																					Ctx: p2388,
																					FreeVars: ast.Identifiers{
																						"slice",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(231),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(231),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p2388,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"invar",
																					"slice",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(231),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(231),
																						Column: int(43),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p2388,
																			FreeVars: ast.Identifiers{
																				"cur",
																				"invar",
																				"slice",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(228),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(231),
																					Column: int(43),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Index{
																			Target: &ast.Var{
																				Id: "invar",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"invar",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(232),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(232),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "step",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p2388,
																				FreeVars: ast.Identifiers{
																					"invar",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(232),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(232),
																						Column: int(27),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "cur",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p2388,
																				FreeVars: ast.Identifiers{
																					"cur",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(232),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(232),
																						Column: int(14),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p2388,
																			FreeVars: ast.Identifiers{
																				"cur",
																				"invar",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(232),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(232),
																					Column: int(27),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2362,
															FreeVars: ast.Identifiers{
																"build",
																"cur",
																"invar",
																"slice",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(227),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(233),
																	Column: int(10),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p2362,
														FreeVars: ast.Identifiers{
															"build",
															"cur",
															"invar",
															"slice",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(224),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(233),
																Column: int(10),
															},
														},
													},
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "slice",
														CommaFodder: ast.Fodder{},
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(223),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(223),
																Column: int(22),
															},
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "cur",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(223),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(223),
																Column: int(27),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: p2446,
													FreeVars: ast.Identifiers{
														"build",
														"invar",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(223),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(233),
															Column: int(10),
														},
													},
												},
												TrailingComma: false,
											},
											EqFodder: nil,
											Variable: "build",
											CloseFodder: nil,
											Fun: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Body: &ast.Apply{
										Target: &ast.Var{
											Id: "build",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p2295,
												FreeVars: ast.Identifiers{
													"build",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(234),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(234),
														Column: int(10),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.LiteralString{
																Value: "string",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2456,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(234),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(234),
																			Column: int(36),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Left: &ast.Index{
																Target: &ast.Var{
																	Id: "invar",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(234),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(234),
																				Column: int(19),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "type",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2456,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(234),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(234),
																			Column: int(24),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2456,
																FreeVars: ast.Identifiers{
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(234),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(234),
																		Column: int(36),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														BranchTrue: &ast.LiteralString{
															Value: "",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2456,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(234),
																		Column: int(42),
																	},
																	End: ast.Location{
																		Line: int(234),
																		Column: int(44),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														BranchFalse: &ast.Array{
															Elements: nil,
															CloseFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2456,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(234),
																		Column: int(50),
																	},
																	End: ast.Location{
																		Line: int(234),
																		Column: int(52),
																	},
																},
															},
															TrailingComma: false,
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2456,
															FreeVars: ast.Identifiers{
																"invar",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(234),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(234),
																	Column: int(52),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Index{
														Target: &ast.Var{
															Id: "invar",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(234),
																		Column: int(54),
																	},
																	End: ast.Location{
																		Line: int(234),
																		Column: int(59),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "index",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2456,
															FreeVars: ast.Identifiers{
																"invar",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(234),
																	Column: int(54),
																},
																End: ast.Location{
																	Line: int(234),
																	Column: int(65),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2295,
											FreeVars: ast.Identifiers{
												"build",
												"invar",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(234),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(234),
													Column: int(66),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p2295,
										FreeVars: ast.Identifiers{
											"invar",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(223),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(234),
												Column: int(66),
											},
										},
									},
								},
								BranchFalse: &ast.Error{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "$std",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "mod",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: nil,
											LeftBracketFodder: nil,
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														Value: "std.slice accepts a string or an array, but got: %s",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2295,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(64),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(117),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(222),
																			Column: int(120),
																		},
																		End: ast.Location{
																			Line: int(222),
																			Column: int(123),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "type",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2295,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(222),
																		Column: int(120),
																	},
																	End: ast.Location{
																		Line: int(222),
																		Column: int(128),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "indexable",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p2491,
																			FreeVars: ast.Identifiers{
																				"indexable",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(222),
																					Column: int(129),
																				},
																				End: ast.Location{
																					Line: int(222),
																					Column: int(138),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2295,
															FreeVars: ast.Identifiers{
																"indexable",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(120),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(139),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: nil,
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
												"indexable",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(222),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(222),
													Column: int(139),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
											"indexable",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(222),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(234),
												Column: int(66),
											},
										},
									},
								},
								ThenFodder: nil,
								ElseFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"indexable",
										"invar",
										"std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "$std",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "mod",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: nil,
										LeftBracketFodder: nil,
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "got %s but step must be greater than 0",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2295,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(221),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(221),
																Column: int(64),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "step",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2295,
														FreeVars: ast.Identifiers{
															"step",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(221),
																Column: int(67),
															},
															End: ast.Location{
																Line: int(221),
																Column: int(71),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: nil,
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
											"step",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(221),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(221),
												Column: int(71),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"step",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(221),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(234),
											Column: int(66),
										},
									},
								},
							},
							ThenFodder: nil,
							ElseFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"indexable",
									"invar",
									"std",
									"step",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "mod",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "got [%s:%s:%s] but negative index, end, and steps are not supported",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2295,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(220),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(220),
															Column: int(137),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															Target: &ast.Var{
																Id: "invar",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(220),
																			Column: int(141),
																		},
																		End: ast.Location{
																			Line: int(220),
																			Column: int(146),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "index",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2526,
																FreeVars: ast.Identifiers{
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(220),
																		Column: int(141),
																	},
																	End: ast.Location{
																		Line: int(220),
																		Column: int(152),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															Target: &ast.Var{
																Id: "invar",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(220),
																			Column: int(154),
																		},
																		End: ast.Location{
																			Line: int(220),
																			Column: int(159),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "end",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2526,
																FreeVars: ast.Identifiers{
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(220),
																		Column: int(154),
																	},
																	End: ast.Location{
																		Line: int(220),
																		Column: int(163),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															Target: &ast.Var{
																Id: "invar",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(220),
																			Column: int(165),
																		},
																		End: ast.Location{
																			Line: int(220),
																			Column: int(170),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "step",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2526,
																FreeVars: ast.Identifiers{
																	"invar",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(220),
																		Column: int(165),
																	},
																	End: ast.Location{
																		Line: int(220),
																		Column: int(175),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												CloseFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2295,
													FreeVars: ast.Identifiers{
														"invar",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(220),
															Column: int(140),
														},
														End: ast.Location{
															Line: int(220),
															Column: int(176),
														},
													},
												},
												TrailingComma: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"invar",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(220),
											Column: int(68),
										},
										End: ast.Location{
											Line: int(220),
											Column: int(176),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"invar",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(220),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(234),
										Column: int(66),
									},
								},
							},
						},
						ThenFodder: nil,
						ElseFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
								"indexable",
								"invar",
								"std",
								"step",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2295,
						FreeVars: ast.Identifiers{
							"$std",
							"end",
							"index",
							"indexable",
							"std",
							"step",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(204),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(234),
								Column: int(66),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indexable",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(203),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(203),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "index",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(203),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(203),
								Column: int(25),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "end",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(203),
								Column: int(27),
							},
							End: ast.Location{
								Line: int(203),
								Column: int(30),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "step",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(203),
								Column: int(32),
							},
							End: ast.Location{
								Line: int(203),
								Column: int(36),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(203),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(234),
					Column: int(66),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "member",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(237),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(237),
											Column: int(11),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isArray",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(237),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(237),
										Column: int(19),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2558,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(237),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(237),
													Column: int(23),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2554,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(237),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(237),
									Column: int(24),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Binary{
						Right: &ast.LiteralNumber{
							OriginalString: "0",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(238),
										Column: int(27),
									},
									End: ast.Location{
										Line: int(238),
										Column: int(28),
									},
								},
							},
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(238),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(238),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "count",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2554,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(238),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(238),
											Column: int(16),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2572,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(238),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(238),
														Column: int(20),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "x",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2572,
												FreeVars: ast.Identifiers{
													"x",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(238),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(238),
														Column: int(23),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
									"x",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(238),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(238),
										Column: int(24),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2554,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(238),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(238),
									Column: int(28),
								},
							},
						},
						Op: ast.BinaryOp(7),
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(239),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(239),
												Column: int(16),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2554,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(239),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(239),
											Column: int(25),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2587,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(239),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(239),
														Column: int(29),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(239),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(239),
										Column: int(30),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2554,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(240),
											Column: int(44),
										},
										End: ast.Location{
											Line: int(240),
											Column: int(45),
										},
									},
								},
							},
							Left: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(240),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(240),
													Column: int(10),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2554,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(240),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(240),
												Column: int(17),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(240),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(240),
																	Column: int(21),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "findSubstr",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2605,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(240),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(240),
																Column: int(32),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "x",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2609,
																	FreeVars: ast.Identifiers{
																		"x",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(240),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(240),
																			Column: int(34),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2609,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(240),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(240),
																			Column: int(39),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2605,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
														"x",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(240),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(240),
															Column: int(40),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2554,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
										"x",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(240),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(240),
											Column: int(41),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
									"x",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(240),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(240),
										Column: int(45),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchFalse: &ast.Error{
							Expr: &ast.LiteralString{
								Value: "std.member first argument must be an array or a string",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2554,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(241),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(241),
											Column: int(72),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2554,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(241),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(241),
										Column: int(72),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2554,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(239),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(241),
									Column: int(72),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2554,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(237),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(241),
								Column: int(72),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(236),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(236),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(236),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(236),
								Column: int(16),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(236),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(241),
					Column: int(72),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "count",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(243),
										Column: int(19),
									},
									End: ast.Location{
										Line: int(243),
										Column: int(22),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "length",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2632,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(243),
									Column: int(19),
								},
								End: ast.Location{
									Line: int(243),
									Column: int(29),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(243),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(243),
														Column: int(33),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "filter",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2640,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(243),
													Column: int(30),
												},
												End: ast.Location{
													Line: int(243),
													Column: int(40),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													ParenLeftFodder: ast.Fodder{},
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Binary{
														Right: &ast.Var{
															Id: "x",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2646,
																FreeVars: ast.Identifiers{
																	"x",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(243),
																		Column: int(58),
																	},
																	End: ast.Location{
																		Line: int(243),
																		Column: int(59),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "v",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2646,
																FreeVars: ast.Identifiers{
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(243),
																		Column: int(53),
																	},
																	End: ast.Location{
																		Line: int(243),
																		Column: int(54),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2646,
															FreeVars: ast.Identifiers{
																"v",
																"x",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(243),
																	Column: int(53),
																},
																End: ast.Location{
																	Line: int(243),
																	Column: int(59),
																},
															},
														},
														Op: ast.BinaryOp(12),
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "v",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(243),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(243),
																	Column: int(51),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2652,
														FreeVars: ast.Identifiers{
															"x",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(243),
																Column: int(41),
															},
															End: ast.Location{
																Line: int(243),
																Column: int(59),
															},
														},
													},
													TrailingComma: false,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2652,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(243),
																Column: int(61),
															},
															End: ast.Location{
																Line: int(243),
																Column: int(64),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
											"x",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(243),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(243),
												Column: int(65),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p2632,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(243),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(243),
								Column: int(66),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(243),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(243),
								Column: int(12),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(243),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(243),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(243),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(243),
					Column: int(66),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "mod",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(246),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(246),
												Column: int(30),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2669,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(246),
											Column: int(27),
										},
										End: ast.Location{
											Line: int(246),
											Column: int(39),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2673,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(246),
														Column: int(40),
													},
													End: ast.Location{
														Line: int(246),
														Column: int(41),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(246),
										Column: int(27),
									},
									End: ast.Location{
										Line: int(246),
										Column: int(42),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(246),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(246),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2669,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(246),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(246),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2684,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(246),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(246),
														Column: int(22),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(246),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(246),
										Column: int(23),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2669,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(246),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(246),
									Column: int(42),
								},
							},
						},
						Op: ast.BinaryOp(17),
					},
					BranchTrue: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(247),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(247),
											Column: int(10),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "modulo",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(247),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(247),
										Column: int(17),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "a",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2697,
											FreeVars: ast.Identifiers{
												"a",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(247),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(247),
													Column: int(19),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "b",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2697,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(247),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(247),
													Column: int(22),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2669,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(247),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(247),
									Column: int(23),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(248),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(248),
												Column: int(16),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2669,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(248),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(248),
											Column: int(25),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2711,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(248),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(248),
														Column: int(27),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(248),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(248),
										Column: int(28),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(249),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(249),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "format",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2669,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(249),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(249),
											Column: int(17),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2723,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(249),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(249),
														Column: int(19),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2723,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(249),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(249),
														Column: int(22),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(249),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(249),
										Column: int(23),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.LiteralString{
									Value: ".",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2669,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(251),
												Column: int(91),
											},
											End: ast.Location{
												Line: int(251),
												Column: int(94),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								Left: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(251),
															Column: int(77),
														},
														End: ast.Location{
															Line: int(251),
															Column: int(80),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "type",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2669,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(251),
														Column: int(77),
													},
													End: ast.Location{
														Line: int(251),
														Column: int(85),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "b",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p2740,
															FreeVars: ast.Identifiers{
																"b",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(251),
																	Column: int(86),
																},
																End: ast.Location{
																	Line: int(251),
																	Column: int(87),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2669,
											FreeVars: ast.Identifiers{
												"b",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(251),
													Column: int(77),
												},
												End: ast.Location{
													Line: int(251),
													Column: int(88),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.Binary{
										Right: &ast.LiteralString{
											Value: " and ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2669,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(251),
														Column: int(67),
													},
													End: ast.Location{
														Line: int(251),
														Column: int(74),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Left: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(251),
																	Column: int(53),
																},
																End: ast.Location{
																	Line: int(251),
																	Column: int(56),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "type",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2669,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(251),
																Column: int(53),
															},
															End: ast.Location{
																Line: int(251),
																Column: int(61),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2754,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(251),
																			Column: int(62),
																		},
																		End: ast.Location{
																			Line: int(251),
																			Column: int(63),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2669,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(251),
															Column: int(53),
														},
														End: ast.Location{
															Line: int(251),
															Column: int(64),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.LiteralString{
												Value: "Operator % cannot be used on types ",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2669,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(251),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(251),
															Column: int(50),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2669,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(251),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(251),
														Column: int(64),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2669,
											FreeVars: ast.Identifiers{
												"a",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(251),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(251),
													Column: int(74),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2669,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(251),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(251),
												Column: int(88),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2669,
									FreeVars: ast.Identifiers{
										"a",
										"b",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(251),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(251),
											Column: int(94),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p2669,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(251),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(251),
										Column: int(94),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2669,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(248),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(251),
									Column: int(94),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2669,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(246),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(251),
								Column: int(94),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(245),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(245),
								Column: int(8),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(245),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(245),
								Column: int(11),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(245),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(251),
					Column: int(94),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "map",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(254),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(254),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isFunction",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(254),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(254),
											Column: int(23),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "func",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2784,
												FreeVars: ast.Identifiers{
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(254),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(254),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2780,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(254),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(254),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2780,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(254),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(254),
									Column: int(29),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(255),
													Column: int(61),
												},
												End: ast.Location{
													Line: int(255),
													Column: int(64),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2780,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(255),
												Column: int(61),
											},
											End: ast.Location{
												Line: int(255),
												Column: int(69),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2798,
													FreeVars: ast.Identifiers{
														"func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(255),
															Column: int(70),
														},
														End: ast.Location{
															Line: int(255),
															Column: int(74),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(255),
											Column: int(61),
										},
										End: ast.Location{
											Line: int(255),
											Column: int(75),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.map first param must be function, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(255),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(255),
											Column: int(58),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2780,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(255),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(255),
										Column: int(75),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p2780,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(255),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(255),
									Column: int(76),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(256),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(256),
														Column: int(38),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isString",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2780,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(256),
													Column: int(35),
												},
												End: ast.Location{
													Line: int(256),
													Column: int(47),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2816,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(256),
																Column: int(48),
															},
															End: ast.Location{
																Line: int(256),
																Column: int(51),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2780,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(256),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(256),
												Column: int(52),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(256),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(256),
											Column: int(52),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							Left: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(256),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(256),
														Column: int(17),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isArray",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2780,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(256),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(256),
													Column: int(25),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2829,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(256),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(256),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2780,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(256),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(256),
												Column: int(30),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(256),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(256),
											Column: int(30),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2780,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(256),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(256),
										Column: int(52),
									},
								},
							},
							Op: ast.BinaryOp(17),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(257),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(257),
														Column: int(71),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2780,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(257),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(257),
													Column: int(76),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2844,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(257),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(257),
																Column: int(80),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2780,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(257),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(257),
												Column: int(81),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.map second param must be array / string, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2780,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(257),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(257),
												Column: int(65),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(257),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(257),
											Column: int(81),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p2780,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(257),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(257),
										Column: int(82),
									},
								},
							},
						},
						BranchFalse: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(259),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(259),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "makeArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2780,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(259),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(259),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(259),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(259),
																Column: int(24),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2864,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(259),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(259),
															Column: int(31),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2868,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(259),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(259),
																		Column: int(35),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2864,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(259),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(259),
														Column: int(36),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											ParenLeftFodder: ast.Fodder{},
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Var{
													Id: "func",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2874,
														FreeVars: ast.Identifiers{
															"func",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(259),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(259),
																Column: int(54),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																Target: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p2879,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(259),
																				Column: int(55),
																			},
																			End: ast.Location{
																				Line: int(259),
																				Column: int(58),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p2879,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(259),
																				Column: int(59),
																			},
																			End: ast.Location{
																				Line: int(259),
																				Column: int(60),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p2879,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(259),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(259),
																			Column: int(61),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2874,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(259),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(259),
															Column: int(62),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(259),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(259),
															Column: int(48),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2864,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(259),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(259),
														Column: int(62),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2780,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(259),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(259),
										Column: int(63),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2780,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(256),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(259),
									Column: int(63),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2780,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(254),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(259),
								Column: int(63),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(253),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(253),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(253),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(253),
								Column: int(16),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(253),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(259),
					Column: int(63),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "mapWithIndex",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(262),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(262),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isFunction",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(262),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(262),
											Column: int(23),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "func",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2908,
												FreeVars: ast.Identifiers{
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(262),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(262),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2904,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(262),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(262),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2904,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(262),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(262),
									Column: int(29),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(263),
													Column: int(70),
												},
												End: ast.Location{
													Line: int(263),
													Column: int(73),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2904,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(263),
												Column: int(70),
											},
											End: ast.Location{
												Line: int(263),
												Column: int(78),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2922,
													FreeVars: ast.Identifiers{
														"func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(263),
															Column: int(79),
														},
														End: ast.Location{
															Line: int(263),
															Column: int(83),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(263),
											Column: int(70),
										},
										End: ast.Location{
											Line: int(263),
											Column: int(84),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.mapWithIndex first param must be function, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(263),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(263),
											Column: int(67),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2904,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(263),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(263),
										Column: int(84),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p2904,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(263),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(263),
									Column: int(85),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(264),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(264),
														Column: int(38),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isString",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2904,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(264),
													Column: int(35),
												},
												End: ast.Location{
													Line: int(264),
													Column: int(47),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2940,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(264),
																Column: int(48),
															},
															End: ast.Location{
																Line: int(264),
																Column: int(51),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2904,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(264),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(264),
												Column: int(52),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(264),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(264),
											Column: int(52),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							Left: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(264),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(264),
														Column: int(17),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isArray",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2904,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(264),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(264),
													Column: int(25),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2953,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(264),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(264),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2904,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(264),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(264),
												Column: int(30),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(264),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(264),
											Column: int(30),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2904,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(264),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(264),
										Column: int(52),
									},
								},
							},
							Op: ast.BinaryOp(17),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(265),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(265),
														Column: int(71),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p2904,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(265),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(265),
													Column: int(76),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2968,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(265),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(265),
																Column: int(80),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2904,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(265),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(265),
												Column: int(81),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.mapWithIndex second param must be array, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p2904,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(265),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(265),
												Column: int(65),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(265),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(265),
											Column: int(81),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p2904,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(265),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(265),
										Column: int(82),
									},
								},
							},
						},
						BranchFalse: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(267),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(267),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "makeArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p2904,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(267),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(267),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(267),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(267),
																Column: int(24),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2988,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(267),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(267),
															Column: int(31),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p2992,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(267),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(267),
																		Column: int(35),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2988,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(267),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(267),
														Column: int(36),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											ParenLeftFodder: ast.Fodder{},
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Var{
													Id: "func",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p2998,
														FreeVars: ast.Identifiers{
															"func",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(267),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(267),
																Column: int(54),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3002,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(267),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(267),
																			Column: int(56),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																Target: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3002,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(267),
																				Column: int(58),
																			},
																			End: ast.Location{
																				Line: int(267),
																				Column: int(61),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3002,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(267),
																				Column: int(62),
																			},
																			End: ast.Location{
																				Line: int(267),
																				Column: int(63),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3002,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(267),
																			Column: int(58),
																		},
																		End: ast.Location{
																			Line: int(267),
																			Column: int(64),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p2998,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(267),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(267),
															Column: int(65),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(267),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(267),
															Column: int(48),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p2988,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(267),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(267),
														Column: int(65),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p2904,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(267),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(267),
										Column: int(66),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p2904,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(264),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(267),
									Column: int(66),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p2904,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(262),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(267),
								Column: int(66),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(261),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(261),
								Column: int(20),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(261),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(261),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(261),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(267),
					Column: int(66),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "mapWithKey",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(270),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(270),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isFunction",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3030,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(270),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(270),
											Column: int(23),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "func",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3034,
												FreeVars: ast.Identifiers{
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(270),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(270),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3030,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(270),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(270),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3030,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(270),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(270),
									Column: int(29),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(271),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(271),
													Column: int(71),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3030,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(271),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(271),
												Column: int(76),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3048,
													FreeVars: ast.Identifiers{
														"func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(271),
															Column: int(77),
														},
														End: ast.Location{
															Line: int(271),
															Column: int(81),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3030,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(271),
											Column: int(68),
										},
										End: ast.Location{
											Line: int(271),
											Column: int(82),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.mapWithKey first param must be function, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3030,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(271),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(271),
											Column: int(65),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3030,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(271),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(271),
										Column: int(82),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p3030,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(271),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(271),
									Column: int(83),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(272),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(272),
													Column: int(17),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isObject",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3030,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(272),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(272),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "obj",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3065,
													FreeVars: ast.Identifiers{
														"obj",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(272),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(272),
															Column: int(30),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3030,
									FreeVars: ast.Identifiers{
										"obj",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(272),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(272),
											Column: int(31),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3030,
								FreeVars: ast.Identifiers{
									"obj",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(272),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(272),
										Column: int(31),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(273),
														Column: int(67),
													},
													End: ast.Location{
														Line: int(273),
														Column: int(70),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3030,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(273),
													Column: int(67),
												},
												End: ast.Location{
													Line: int(273),
													Column: int(75),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "obj",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3079,
														FreeVars: ast.Identifiers{
															"obj",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(273),
																Column: int(76),
															},
															End: ast.Location{
																Line: int(273),
																Column: int(79),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3030,
										FreeVars: ast.Identifiers{
											"obj",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(273),
												Column: int(67),
											},
											End: ast.Location{
												Line: int(273),
												Column: int(80),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.mapWithKey second param must be object, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3030,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(273),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(273),
												Column: int(64),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3030,
									FreeVars: ast.Identifiers{
										"obj",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(273),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(273),
											Column: int(80),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p3030,
								FreeVars: ast.Identifiers{
									"obj",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(273),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(273),
										Column: int(81),
									},
								},
							},
						},
						BranchFalse: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "$objectFlatMerge",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "$std",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "flatMap",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: nil,
												LeftBracketFodder: nil,
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: nil,
															ParenRightFodder: nil,
															Body: &ast.Array{
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.DesugaredObject{
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Name: &ast.Var{
																						Id: "k",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3030,
																							FreeVars: ast.Identifiers{
																								"k",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(275),
																									Column: int(10),
																								},
																								End: ast.Location{
																									Line: int(275),
																									Column: int(11),
																								},
																							},
																						},
																					},
																					Body: &ast.Apply{
																						Target: &ast.Var{
																							Id: "func",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3109,
																								FreeVars: ast.Identifiers{
																									"func",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(275),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(275),
																										Column: int(18),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "k",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p3113,
																											FreeVars: ast.Identifiers{
																												"k",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(275),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(275),
																													Column: int(20),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "obj",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p3113,
																												FreeVars: ast.Identifiers{
																													"obj",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(275),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(275),
																														Column: int(25),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "k",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p3113,
																												FreeVars: ast.Identifiers{
																													"k",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(275),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(275),
																														Column: int(27),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p3113,
																											FreeVars: ast.Identifiers{
																												"k",
																												"obj",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(275),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(275),
																													Column: int(28),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3109,
																							FreeVars: ast.Identifiers{
																								"func",
																								"k",
																								"obj",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(275),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(275),
																									Column: int(29),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(275),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(275),
																							Column: int(29),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																			},
																			Locals: ast.LocalBinds{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																					},
																				},
																				Ctx: p3030,
																				FreeVars: ast.Identifiers{
																					"func",
																					"k",
																					"obj",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(275),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(275),
																						Column: int(62),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																CloseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"func",
																		"k",
																		"obj",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																TrailingComma: false,
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "k",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"func",
																	"obj",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(275),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(275),
																				Column: int(42),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "objectFields",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3030,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(275),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(275),
																			Column: int(55),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "obj",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3135,
																				FreeVars: ast.Identifiers{
																					"obj",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(275),
																						Column: int(56),
																					},
																					End: ast.Location{
																						Line: int(275),
																						Column: int(59),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3030,
																FreeVars: ast.Identifiers{
																	"obj",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(275),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(275),
																		Column: int(60),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: nil,
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
													"func",
													"obj",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(275),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(275),
														Column: int(62),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"func",
									"obj",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(275),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(275),
										Column: int(62),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3030,
							FreeVars: ast.Identifiers{
								"$std",
								"func",
								"obj",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(272),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(275),
									Column: int(62),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p3030,
						FreeVars: ast.Identifiers{
							"$std",
							"func",
							"obj",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(270),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(275),
								Column: int(62),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(269),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(269),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "obj",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(269),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(269),
								Column: int(23),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(269),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(275),
					Column: int(62),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "flatMap",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(278),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(278),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isFunction",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(278),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(278),
											Column: int(23),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "func",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3160,
												FreeVars: ast.Identifiers{
													"func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(278),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(278),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3156,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(278),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(278),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3156,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(278),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(278),
									Column: int(29),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(279),
													Column: int(65),
												},
												End: ast.Location{
													Line: int(279),
													Column: int(68),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3156,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(279),
												Column: int(65),
											},
											End: ast.Location{
												Line: int(279),
												Column: int(73),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3174,
													FreeVars: ast.Identifiers{
														"func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(279),
															Column: int(74),
														},
														End: ast.Location{
															Line: int(279),
															Column: int(78),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(279),
											Column: int(65),
										},
										End: ast.Location{
											Line: int(279),
											Column: int(79),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.flatMap first param must be function, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(279),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(279),
											Column: int(62),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3156,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(279),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(279),
										Column: int(79),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p3156,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(279),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(279),
									Column: int(80),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(280),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(280),
												Column: int(16),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(280),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(280),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3190,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(280),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(280),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3156,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(280),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(280),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(281),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(281),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "flattenArrays",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(281),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(281),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(281),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(281),
																Column: int(28),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "makeArray",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3206,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(281),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(281),
															Column: int(38),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(281),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(281),
																				Column: int(42),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "length",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3214,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(281),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(281),
																			Column: int(49),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3218,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(281),
																						Column: int(50),
																					},
																					End: ast.Location{
																						Line: int(281),
																						Column: int(53),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3214,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(281),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(281),
																		Column: int(54),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Var{
																	Id: "func",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3224,
																		FreeVars: ast.Identifiers{
																			"func",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(281),
																				Column: int(68),
																			},
																			End: ast.Location{
																				Line: int(281),
																				Column: int(72),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Index{
																				Target: &ast.Var{
																					Id: "arr",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3229,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(281),
																								Column: int(73),
																							},
																							End: ast.Location{
																								Line: int(281),
																								Column: int(76),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3229,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(281),
																								Column: int(77),
																							},
																							End: ast.Location{
																								Line: int(281),
																								Column: int(78),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3229,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(281),
																							Column: int(73),
																						},
																						End: ast.Location{
																							Line: int(281),
																							Column: int(79),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3224,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"func",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(281),
																			Column: int(68),
																		},
																		End: ast.Location{
																			Line: int(281),
																			Column: int(80),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(281),
																			Column: int(65),
																		},
																		End: ast.Location{
																			Line: int(281),
																			Column: int(66),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3214,
																FreeVars: ast.Identifiers{
																	"arr",
																	"func",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(281),
																		Column: int(56),
																	},
																	End: ast.Location{
																		Line: int(281),
																		Column: int(80),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3206,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(281),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(281),
														Column: int(81),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3156,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(281),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(281),
										Column: int(82),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(282),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(282),
													Column: int(16),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3156,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(282),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(282),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3248,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(282),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(282),
															Column: int(29),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(282),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(282),
											Column: int(30),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(283),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(283),
													Column: int(10),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "join",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3156,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(283),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(283),
												Column: int(15),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3260,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(283),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(283),
															Column: int(18),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(283),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(283),
																	Column: int(23),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "makeArray",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3260,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(283),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(283),
																Column: int(33),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(37),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "length",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3273,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(44),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "arr",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3277,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(283),
																							Column: int(45),
																						},
																						End: ast.Location{
																							Line: int(283),
																							Column: int(48),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3273,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(283),
																			Column: int(34),
																		},
																		End: ast.Location{
																			Line: int(283),
																			Column: int(49),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																ParenLeftFodder: ast.Fodder{},
																ParenRightFodder: ast.Fodder{},
																Body: &ast.Apply{
																	Target: &ast.Var{
																		Id: "func",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3283,
																			FreeVars: ast.Identifiers{
																				"func",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(63),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(67),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					Target: &ast.Var{
																						Id: "arr",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3288,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(283),
																									Column: int(68),
																								},
																								End: ast.Location{
																									Line: int(283),
																									Column: int(71),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3288,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(283),
																									Column: int(72),
																								},
																								End: ast.Location{
																									Line: int(283),
																									Column: int(73),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3288,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(283),
																								Column: int(68),
																							},
																							End: ast.Location{
																								Line: int(283),
																								Column: int(74),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3283,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"func",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(63),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(75),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "i",
																		CommaFodder: nil,
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(60),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(61),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3273,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"func",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(283),
																			Column: int(51),
																		},
																		End: ast.Location{
																			Line: int(283),
																			Column: int(75),
																		},
																	},
																},
																TrailingComma: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3260,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(283),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(283),
															Column: int(76),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"arr",
										"func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(283),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(283),
											Column: int(77),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(284),
															Column: int(75),
														},
														End: ast.Location{
															Line: int(284),
															Column: int(78),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "type",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3156,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(284),
														Column: int(75),
													},
													End: ast.Location{
														Line: int(284),
														Column: int(83),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p3308,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(284),
																	Column: int(84),
																},
																End: ast.Location{
																	Line: int(284),
																	Column: int(87),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3156,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(284),
													Column: int(75),
												},
												End: ast.Location{
													Line: int(284),
													Column: int(88),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "std.flatMap second param must be array / string, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3156,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(284),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(284),
													Column: int(72),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3156,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(284),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(284),
												Column: int(88),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3156,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(284),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(284),
											Column: int(89),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3156,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(282),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(284),
										Column: int(89),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3156,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(280),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(284),
									Column: int(89),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p3156,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(278),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(284),
								Column: int(89),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(277),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(277),
								Column: int(15),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(277),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(277),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(277),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(284),
					Column: int(89),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "join",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(288),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(288),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(288),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(288),
															Column: int(25),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3339,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(288),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(288),
																		Column: int(29),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3335,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(288),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(288),
														Column: int(30),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3335,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(288),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(288),
														Column: int(11),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3335,
											FreeVars: ast.Identifiers{
												"arr",
												"i",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(288),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(288),
													Column: int(30),
												},
											},
										},
										Op: ast.BinaryOp(8),
									},
									BranchTrue: &ast.Var{
										Id: "running",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p3335,
											FreeVars: ast.Identifiers{
												"running",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(289),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(289),
													Column: int(16),
												},
											},
										},
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralNull{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3335,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(290),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(290),
															Column: int(29),
														},
													},
												},
											},
											Left: &ast.Index{
												Target: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(290),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(290),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.Var{
													Id: "i",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"i",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(290),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(290),
																Column: int(20),
															},
														},
													},
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(290),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(290),
															Column: int(21),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3335,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(290),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(290),
														Column: int(29),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.Apply{
											Target: &ast.Var{
												Id: "aux",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"aux",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(291),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(291),
															Column: int(12),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3364,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(291),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(291),
																		Column: int(16),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3364,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(291),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(291),
																			Column: int(23),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3364,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(291),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(291),
																			Column: int(19),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3364,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(291),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(291),
																		Column: int(23),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "first",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3364,
																FreeVars: ast.Identifiers{
																	"first",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(291),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(291),
																		Column: int(30),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "running",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3364,
																FreeVars: ast.Identifiers{
																	"running",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(291),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(291),
																		Column: int(39),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3335,
												FreeVars: ast.Identifiers{
													"arr",
													"aux",
													"first",
													"i",
													"running",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(291),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(291),
														Column: int(40),
													},
												},
											},
											TrailingComma: false,
											TailStrict: true,
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(292),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(292),
																		Column: int(38),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "type",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p3335,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(292),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(292),
																	Column: int(43),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "sep",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3386,
																		FreeVars: ast.Identifiers{
																			"sep",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(292),
																				Column: int(44),
																			},
																			End: ast.Location{
																				Line: int(292),
																				Column: int(47),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"sep",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(292),
																Column: int(35),
															},
															End: ast.Location{
																Line: int(292),
																Column: int(48),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(292),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(292),
																		Column: int(18),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "type",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p3335,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(292),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(292),
																	Column: int(23),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "arr",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3398,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(292),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(292),
																					Column: int(27),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3398,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(292),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(292),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3398,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(292),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(292),
																				Column: int(30),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"arr",
															"i",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(292),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(292),
																Column: int(31),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
														"sep",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(292),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(292),
															Column: int(48),
														},
													},
												},
												Op: ast.BinaryOp(13),
											},
											BranchTrue: &ast.Error{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "mod",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralString{
																	Value: "expected %s but arr[%d] was %s ",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3335,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(293),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(293),
																				Column: int(48),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Array{
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(293),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(293),
																									Column: int(55),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "type",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3421,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(293),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(293),
																								Column: int(60),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "sep",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3425,
																									FreeVars: ast.Identifiers{
																										"sep",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(293),
																											Column: int(61),
																										},
																										End: ast.Location{
																											Line: int(293),
																											Column: int(64),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3421,
																					FreeVars: ast.Identifiers{
																						"sep",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(293),
																							Column: int(52),
																						},
																						End: ast.Location{
																							Line: int(293),
																							Column: int(65),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3421,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(293),
																							Column: int(67),
																						},
																						End: ast.Location{
																							Line: int(293),
																							Column: int(68),
																						},
																					},
																				},
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(293),
																									Column: int(70),
																								},
																								End: ast.Location{
																									Line: int(293),
																									Column: int(73),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "type",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3421,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(293),
																								Column: int(70),
																							},
																							End: ast.Location{
																								Line: int(293),
																								Column: int(78),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								Target: &ast.Var{
																									Id: "arr",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p3439,
																										FreeVars: ast.Identifiers{
																											"arr",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(293),
																												Column: int(79),
																											},
																											End: ast.Location{
																												Line: int(293),
																												Column: int(82),
																											},
																										},
																									},
																								},
																								Index: &ast.Var{
																									Id: "i",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p3439,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(293),
																												Column: int(83),
																											},
																											End: ast.Location{
																												Line: int(293),
																												Column: int(84),
																											},
																										},
																									},
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3439,
																									FreeVars: ast.Identifiers{
																										"arr",
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(293),
																											Column: int(79),
																										},
																										End: ast.Location{
																											Line: int(293),
																											Column: int(85),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3421,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(293),
																							Column: int(70),
																						},
																						End: ast.Location{
																							Line: int(293),
																							Column: int(86),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	CloseFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3335,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"sep",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(293),
																				Column: int(51),
																			},
																			End: ast.Location{
																				Line: int(293),
																				Column: int(87),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"arr",
															"i",
															"sep",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(293),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(293),
																Column: int(87),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"$std",
														"arr",
														"i",
														"sep",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(293),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(293),
															Column: int(87),
														},
													},
												},
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Var{
													Id: "first",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"first",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(294),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(294),
																Column: int(20),
															},
														},
													},
												},
												BranchTrue: &ast.Apply{
													Target: &ast.Var{
														Id: "aux",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p3335,
															FreeVars: ast.Identifiers{
																"aux",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(295),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(295),
																	Column: int(12),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3458,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(295),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(295),
																				Column: int(16),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3458,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(295),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(295),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3458,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(295),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(295),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3458,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(295),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(295),
																				Column: int(23),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3458,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(295),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(295),
																				Column: int(30),
																			},
																		},
																	},
																	Value: false,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.Index{
																		Target: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3458,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(295),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(295),
																						Column: int(45),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3458,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(295),
																						Column: int(46),
																					},
																					End: ast.Location{
																						Line: int(295),
																						Column: int(47),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3458,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(295),
																					Column: int(42),
																				},
																				End: ast.Location{
																					Line: int(295),
																					Column: int(48),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "running",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3458,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(295),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(295),
																					Column: int(39),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3458,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"running",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(295),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(295),
																				Column: int(48),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"running",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(295),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(295),
																Column: int(49),
															},
														},
													},
													TrailingComma: false,
													TailStrict: true,
												},
												BranchFalse: &ast.Apply{
													Target: &ast.Var{
														Id: "aux",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p3335,
															FreeVars: ast.Identifiers{
																"aux",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(297),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(297),
																	Column: int(12),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3483,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(297),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(297),
																				Column: int(16),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3483,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(297),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(297),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3483,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(297),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(297),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3483,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(297),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(297),
																				Column: int(23),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3483,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(297),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(297),
																				Column: int(30),
																			},
																		},
																	},
																	Value: false,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.Index{
																		Target: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3483,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(297),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(297),
																						Column: int(51),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3483,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(297),
																						Column: int(52),
																					},
																					End: ast.Location{
																						Line: int(297),
																						Column: int(53),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3483,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(297),
																					Column: int(48),
																				},
																				End: ast.Location{
																					Line: int(297),
																					Column: int(54),
																				},
																			},
																		},
																	},
																	Left: &ast.Binary{
																		Right: &ast.Var{
																			Id: "sep",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3483,
																				FreeVars: ast.Identifiers{
																					"sep",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(297),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(297),
																						Column: int(45),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "running",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3483,
																				FreeVars: ast.Identifiers{
																					"running",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(297),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(297),
																						Column: int(39),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3483,
																			FreeVars: ast.Identifiers{
																				"running",
																				"sep",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(297),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(297),
																					Column: int(45),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3483,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"running",
																			"sep",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(297),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(297),
																				Column: int(54),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3335,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"running",
															"sep",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(297),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(297),
																Column: int(55),
															},
														},
													},
													TrailingComma: false,
													TailStrict: true,
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3335,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"first",
														"i",
														"running",
														"sep",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(294),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(297),
															Column: int(55),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3335,
												FreeVars: ast.Identifiers{
													"$std",
													"arr",
													"aux",
													"first",
													"i",
													"running",
													"sep",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(292),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(297),
														Column: int(55),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3335,
											FreeVars: ast.Identifiers{
												"$std",
												"arr",
												"aux",
												"first",
												"i",
												"running",
												"sep",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(290),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(297),
													Column: int(55),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p3335,
										FreeVars: ast.Identifiers{
											"$std",
											"arr",
											"aux",
											"first",
											"i",
											"running",
											"sep",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(288),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(297),
												Column: int(55),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(287),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(18),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(287),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(21),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "first",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(287),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(28),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(287),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(37),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p3516,
									FreeVars: ast.Identifiers{
										"$std",
										"aux",
										"sep",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(287),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(297),
											Column: int(55),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(298),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(298),
													Column: int(12),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isArray",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(298),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(298),
												Column: int(20),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3529,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(298),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(298),
															Column: int(24),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3525,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(298),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(298),
											Column: int(25),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3525,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(298),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(298),
										Column: int(25),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(299),
														Column: int(61),
													},
													End: ast.Location{
														Line: int(299),
														Column: int(64),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3525,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(299),
													Column: int(61),
												},
												End: ast.Location{
													Line: int(299),
													Column: int(69),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3543,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(299),
																Column: int(70),
															},
															End: ast.Location{
																Line: int(299),
																Column: int(73),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(299),
												Column: int(61),
											},
											End: ast.Location{
												Line: int(299),
												Column: int(74),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "join second parameter should be array, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(299),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(299),
												Column: int(58),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3525,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(299),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(299),
											Column: int(74),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p3525,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(299),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(299),
										Column: int(74),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(300),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(300),
													Column: int(16),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(300),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(300),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "sep",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3559,
													FreeVars: ast.Identifiers{
														"sep",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(300),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(300),
															Column: int(29),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3525,
									FreeVars: ast.Identifiers{
										"sep",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(300),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(300),
											Column: int(30),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Var{
									Id: "aux",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"aux",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(301),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(301),
												Column: int(10),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3568,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(301),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(301),
															Column: int(14),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3568,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(301),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(301),
															Column: int(17),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3568,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(301),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(301),
															Column: int(23),
														},
													},
												},
												Value: true,
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3568,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(301),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(301),
															Column: int(27),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3525,
									FreeVars: ast.Identifiers{
										"arr",
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(301),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(301),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(302),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(302),
														Column: int(16),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isArray",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3525,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(302),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(302),
													Column: int(24),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "sep",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3583,
														FreeVars: ast.Identifiers{
															"sep",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(302),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(302),
																Column: int(28),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"sep",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(302),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(302),
												Column: int(29),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Apply{
									Target: &ast.Var{
										Id: "aux",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p3525,
											FreeVars: ast.Identifiers{
												"aux",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(303),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(303),
													Column: int(10),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3592,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(303),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(303),
																Column: int(14),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													OriginalString: "0",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3592,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(303),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(303),
																Column: int(17),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3592,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(303),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(303),
																Column: int(23),
															},
														},
													},
													Value: true,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Array{
													Elements: nil,
													CloseFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3592,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(303),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(303),
																Column: int(27),
															},
														},
													},
													TrailingComma: false,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(303),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(303),
												Column: int(28),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchFalse: &ast.Error{
									Expr: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(305),
																Column: int(70),
															},
															End: ast.Location{
																Line: int(305),
																Column: int(73),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "type",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3525,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(305),
															Column: int(70),
														},
														End: ast.Location{
															Line: int(305),
															Column: int(78),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "sep",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3608,
																FreeVars: ast.Identifiers{
																	"sep",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(305),
																		Column: int(79),
																	},
																	End: ast.Location{
																		Line: int(305),
																		Column: int(82),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3525,
												FreeVars: ast.Identifiers{
													"sep",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(305),
														Column: int(70),
													},
													End: ast.Location{
														Line: int(305),
														Column: int(83),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.LiteralString{
											Value: "join first parameter should be string or array, got ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3525,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(305),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(305),
														Column: int(67),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3525,
											FreeVars: ast.Identifiers{
												"sep",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(305),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(305),
													Column: int(83),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p3525,
										FreeVars: ast.Identifiers{
											"sep",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(305),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(305),
												Column: int(83),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3525,
									FreeVars: ast.Identifiers{
										"arr",
										"aux",
										"sep",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(302),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(305),
											Column: int(83),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3525,
								FreeVars: ast.Identifiers{
									"arr",
									"aux",
									"sep",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(300),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(305),
										Column: int(83),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p3525,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"sep",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(298),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(305),
									Column: int(83),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p3525,
						FreeVars: ast.Identifiers{
							"$std",
							"arr",
							"sep",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(287),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(305),
								Column: int(83),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "sep",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(286),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(286),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(286),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(286),
								Column: int(16),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(286),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(305),
					Column: int(83),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "lines",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(308),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(308),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "join",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3634,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(308),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(308),
									Column: int(13),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: "\n",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3638,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(308),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(308),
												Column: int(18),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Binary{
									Right: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3643,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(308),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(308),
																Column: int(29),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3638,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(308),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(308),
													Column: int(30),
												},
											},
										},
										TrailingComma: false,
									},
									Left: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3638,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(308),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(308),
													Column: int(23),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p3638,
										FreeVars: ast.Identifiers{
											"arr",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(308),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(308),
												Column: int(30),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p3634,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(308),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(308),
								Column: int(31),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(307),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(307),
								Column: int(12),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(307),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(308),
					Column: int(31),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "deepJoin",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(311),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(311),
											Column: int(11),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isString",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3658,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(311),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(311),
										Column: int(20),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3662,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(311),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(311),
													Column: int(24),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3658,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(311),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(311),
									Column: int(25),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Var{
						Id: "arr",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p3658,
							FreeVars: ast.Identifiers{
								"arr",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(312),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(312),
									Column: int(10),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(313),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(313),
												Column: int(16),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3658,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(313),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(313),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3677,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(313),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(313),
														Column: int(28),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3658,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(313),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(313),
										Column: int(29),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(314),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(314),
												Column: int(10),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "join",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p3658,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(314),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(314),
											Column: int(15),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3689,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(314),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(314),
														Column: int(18),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "$std",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "flatMap",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: nil,
												LeftBracketFodder: nil,
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: nil,
															ParenRightFodder: nil,
															Body: &ast.Array{
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(314),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(314),
																								Column: int(24),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "deepJoin",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3705,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(314),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(314),
																							Column: int(33),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "x",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3709,
																								FreeVars: ast.Identifiers{
																									"x",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(314),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(314),
																										Column: int(35),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3705,
																				FreeVars: ast.Identifiers{
																					"std",
																					"x",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(314),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(314),
																						Column: int(36),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																CloseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																		"x",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																TrailingComma: false,
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "x",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3689,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(314),
																		Column: int(46),
																	},
																	End: ast.Location{
																		Line: int(314),
																		Column: int(49),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: nil,
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(314),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(314),
														Column: int(50),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p3658,
								FreeVars: ast.Identifiers{
									"$std",
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(314),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(314),
										Column: int(51),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "mod",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "Expected string or array, got %s",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3658,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(316),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(316),
															Column: int(47),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(316),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(316),
																	Column: int(53),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "type",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3658,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(316),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(316),
																Column: int(58),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3736,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(316),
																			Column: int(59),
																		},
																		End: ast.Location{
																			Line: int(316),
																			Column: int(62),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3658,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(316),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(316),
															Column: int(63),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(316),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(316),
											Column: int(63),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p3658,
								FreeVars: ast.Identifiers{
									"$std",
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(316),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(316),
										Column: int(63),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p3658,
							FreeVars: ast.Identifiers{
								"$std",
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(313),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(316),
									Column: int(63),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p3658,
						FreeVars: ast.Identifiers{
							"$std",
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(311),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(316),
								Column: int(63),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(310),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(310),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(310),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(316),
					Column: int(63),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "format",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(326),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(326),
																Column: int(21),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3761,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(326),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(326),
															Column: int(28),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3765,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(326),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(326),
																		Column: int(32),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3761,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(326),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(326),
														Column: int(33),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3761,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(326),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(326),
														Column: int(15),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p3761,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(326),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(326),
													Column: int(33),
												},
											},
										},
										Op: ast.BinaryOp(9),
									},
									BranchTrue: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Index{
													Target: &ast.Var{
														Id: "str",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p3775,
															FreeVars: ast.Identifiers{
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(327),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(327),
																	Column: int(20),
																},
															},
														},
													},
													Index: &ast.Var{
														Id: "i",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p3775,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(327),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(327),
																	Column: int(22),
																},
															},
														},
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3775,
														FreeVars: ast.Identifiers{
															"i",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(327),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(327),
																Column: int(23),
															},
														},
													},
												},
												EqFodder: ast.Fodder{},
												Variable: "c",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(327),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(327),
														Column: int(23),
													},
												},
											},
										},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralString{
													Value: "(",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3761,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(328),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(328),
																Column: int(18),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												Left: &ast.Var{
													Id: "c",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3761,
														FreeVars: ast.Identifiers{
															"c",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(328),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(328),
																Column: int(11),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p3761,
													FreeVars: ast.Identifiers{
														"c",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(328),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(328),
															Column: int(18),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Body: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(330),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(330),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "length",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3796,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(330),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(330),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3800,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(330),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(330),
																									Column: int(33),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3796,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(330),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(330),
																					Column: int(34),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	Left: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3796,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(330),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(330),
																					Column: int(15),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3796,
																		FreeVars: ast.Identifiers{
																			"j",
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(330),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(330),
																				Column: int(34),
																			},
																		},
																	},
																	Op: ast.BinaryOp(8),
																},
																BranchTrue: &ast.Error{
																	Expr: &ast.LiteralString{
																		Value: "Truncated format code.",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3796,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(331),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(331),
																					Column: int(43),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		Ctx: p3796,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(331),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(331),
																				Column: int(43),
																			},
																		},
																	},
																},
																BranchFalse: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Index{
																				Target: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3813,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(333),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(333),
																								Column: int(26),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "j",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3813,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(333),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(333),
																								Column: int(28),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3813,
																					FreeVars: ast.Identifiers{
																						"j",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(333),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(333),
																							Column: int(29),
																						},
																					},
																				},
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "c",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(333),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(333),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: ")",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3796,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(334),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(334),
																							Column: int(24),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3796,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(334),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(334),
																							Column: int(17),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3796,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(334),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(334),
																						Column: int(24),
																					},
																				},
																			},
																			Op: ast.BinaryOp(13),
																		},
																		BranchTrue: &ast.Apply{
																			Target: &ast.Var{
																				Id: "consume",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(14),
																						},
																					},
																					Ctx: p3796,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(335),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(335),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3830,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(335),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(335),
																										Column: int(26),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3830,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(335),
																											Column: int(32),
																										},
																										End: ast.Location{
																											Line: int(335),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3830,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(335),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(335),
																											Column: int(29),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3830,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(335),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(335),
																										Column: int(33),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.Var{
																								Id: "c",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3830,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(335),
																											Column: int(39),
																										},
																										End: ast.Location{
																											Line: int(335),
																											Column: int(40),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p3830,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(335),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(335),
																											Column: int(36),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3830,
																								FreeVars: ast.Identifiers{
																									"c",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(335),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(335),
																										Column: int(40),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3796,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(335),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(335),
																						Column: int(41),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		BranchFalse: &ast.DesugaredObject{
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Name: &ast.LiteralString{
																						Value: "i",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Body: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3849,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(25),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "j",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3849,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(21),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3849,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(337),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(337),
																									Column: int(25),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(337),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(337),
																							Column: int(25),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																				ast.DesugaredObjectField{
																					Name: &ast.LiteralString{
																						Value: "v",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Body: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3849,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(337),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(337),
																									Column: int(31),
																								},
																							},
																						},
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(337),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(337),
																							Column: int(31),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																			},
																			Locals: ast.LocalBinds{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(14),
																					},
																				},
																				Ctx: p3796,
																				FreeVars: ast.Identifiers{
																					"j",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(337),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(337),
																						Column: int(33),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(12),
																				},
																			},
																			Ctx: p3796,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(334),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(337),
																					Column: int(33),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		Ctx: p3796,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(333),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(337),
																				Column: int(33),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p3796,
																	FreeVars: ast.Identifiers{
																		"consume",
																		"j",
																		"std",
																		"str",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(330),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(337),
																			Column: int(33),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(329),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(329),
																			Column: int(26),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "j",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(329),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(329),
																			Column: int(29),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "v",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(329),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(329),
																			Column: int(32),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: p3867,
																FreeVars: ast.Identifiers{
																	"consume",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(329),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(337),
																		Column: int(33),
																	},
																},
															},
															TrailingComma: false,
														},
														EqFodder: nil,
														Variable: "consume",
														CloseFodder: nil,
														Fun: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Body: &ast.Apply{
													Target: &ast.Var{
														Id: "consume",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p3761,
															FreeVars: ast.Identifiers{
																"consume",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(338),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(338),
																	Column: int(16),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "str",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3875,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(338),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(338),
																				Column: int(20),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3875,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(338),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(338),
																					Column: int(27),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3875,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(338),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(338),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3875,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(338),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(338),
																				Column: int(27),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralString{
																	Value: "",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3875,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(338),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(338),
																				Column: int(31),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p3761,
														FreeVars: ast.Identifiers{
															"consume",
															"i",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(338),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(338),
																Column: int(32),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p3761,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(329),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(338),
															Column: int(32),
														},
													},
												},
											},
											BranchFalse: &ast.DesugaredObject{
												Asserts: ast.Nodes{},
												Fields: ast.DesugaredObjectFields{
													ast.DesugaredObjectField{
														Name: &ast.LiteralString{
															Value: "i",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														Body: &ast.Var{
															Id: "i",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3890,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(340),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(340),
																		Column: int(15),
																	},
																},
															},
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(340),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(340),
																Column: int(15),
															},
														},
														Hide: ast.ObjectFieldHide(1),
														PlusSuper: false,
													},
													ast.DesugaredObjectField{
														Name: &ast.LiteralString{
															Value: "v",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														Body: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3890,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(340),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(340),
																		Column: int(24),
																	},
																},
															},
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(340),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(340),
																Column: int(24),
															},
														},
														Hide: ast.ObjectFieldHide(1),
														PlusSuper: false,
													},
												},
												Locals: ast.LocalBinds{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p3761,
													FreeVars: ast.Identifiers{
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(340),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(340),
															Column: int(26),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p3761,
												FreeVars: ast.Identifiers{
													"c",
													"i",
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(328),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(340),
														Column: int(26),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p3761,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(327),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(340),
													Column: int(26),
												},
											},
										},
									},
									BranchFalse: &ast.Error{
										Expr: &ast.LiteralString{
											Value: "Truncated format code.",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p3761,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(326),
														Column: int(36),
													},
													End: ast.Location{
														Line: int(326),
														Column: int(60),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(326),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(340),
													Column: int(26),
												},
											},
										},
									},
									ThenFodder: nil,
									ElseFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"i",
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "str",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(325),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(325),
												Column: int(36),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(325),
												Column: int(38),
											},
											End: ast.Location{
												Line: int(325),
												Column: int(39),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p3905,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(325),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(340),
											Column: int(26),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "try_parse_mapping_key",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Body: &ast.Function{
													ParenLeftFodder: ast.Fodder{},
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(344),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(344),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "length",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3920,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(344),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(344),
																				Column: int(30),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3924,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(344),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(344),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3920,
																	FreeVars: ast.Identifiers{
																		"std",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(344),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(344),
																			Column: int(35),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															Left: &ast.Var{
																Id: "j",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3920,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(344),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(344),
																			Column: int(17),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p3920,
																FreeVars: ast.Identifiers{
																	"j",
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(344),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(344),
																		Column: int(35),
																	},
																},
															},
															Op: ast.BinaryOp(9),
														},
														BranchTrue: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Body: &ast.Index{
																		Target: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3934,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(345),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(345),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "j",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3934,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(345),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(345),
																						Column: int(24),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3934,
																			FreeVars: ast.Identifiers{
																				"j",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(345),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(345),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	EqFodder: ast.Fodder{},
																	Variable: "c",
																	CloseFodder: ast.Fodder{},
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(345),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(345),
																			Column: int(25),
																		},
																	},
																},
															},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralString{
																		Value: "#",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(20),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Left: &ast.Var{
																		Id: "c",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(13),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3920,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(346),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(346),
																				Column: int(20),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.Apply{
																	Target: &ast.Var{
																		Id: "consume",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"consume",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(347),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(347),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3951,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(347),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(347),
																								Column: int(22),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3951,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(347),
																									Column: int(28),
																								},
																								End: ast.Location{
																									Line: int(347),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "j",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3951,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(347),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(347),
																									Column: int(25),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3951,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(347),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(347),
																								Column: int(29),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.DesugaredObject{
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "alt",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.LiteralBoolean{
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p3963,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(347),
																												Column: int(40),
																											},
																											End: ast.Location{
																												Line: int(347),
																												Column: int(44),
																											},
																										},
																									},
																									Value: true,
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(347),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(347),
																										Column: int(44),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																						},
																						Locals: ast.LocalBinds{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3951,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(347),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(347),
																									Column: int(46),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3951,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(347),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(347),
																									Column: int(32),
																								},
																							},
																						},
																					},
																					OpFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3951,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(347),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(347),
																								Column: int(46),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3920,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(347),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(347),
																				Column: int(47),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																BranchFalse: &ast.Conditional{
																	Cond: &ast.Binary{
																		Right: &ast.LiteralString{
																			Value: "0",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(348),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(348),
																						Column: int(25),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		Left: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(348),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(348),
																						Column: int(18),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(348),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(348),
																					Column: int(25),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	BranchTrue: &ast.Apply{
																		Target: &ast.Var{
																			Id: "consume",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{
																					"consume",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(349),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(349),
																						Column: int(18),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3980,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(349),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(349),
																									Column: int(22),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3980,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(349),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(349),
																										Column: int(29),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "j",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3980,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(349),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(349),
																										Column: int(25),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3980,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(349),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(349),
																									Column: int(29),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.DesugaredObject{
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Name: &ast.LiteralString{
																										Value: "zero",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Body: &ast.LiteralBoolean{
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p3992,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(349),
																													Column: int(41),
																												},
																												End: ast.Location{
																													Line: int(349),
																													Column: int(45),
																												},
																											},
																										},
																										Value: true,
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(349),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(349),
																											Column: int(45),
																										},
																									},
																									Hide: ast.ObjectFieldHide(1),
																									PlusSuper: false,
																								},
																							},
																							Locals: ast.LocalBinds{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3980,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(349),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(349),
																										Column: int(47),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "v",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p3980,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(349),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(349),
																										Column: int(32),
																									},
																								},
																							},
																						},
																						OpFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3980,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(349),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(349),
																									Column: int(47),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(349),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(349),
																					Column: int(48),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	BranchFalse: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "-",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(350),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(350),
																							Column: int(25),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(350),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(350),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(350),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(350),
																						Column: int(25),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		BranchTrue: &ast.Apply{
																			Target: &ast.Var{
																				Id: "consume",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(351),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(351),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4009,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(351),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(351),
																										Column: int(22),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4009,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(351),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(351),
																											Column: int(29),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4009,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(351),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(351),
																											Column: int(25),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4009,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(351),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(351),
																										Column: int(29),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.DesugaredObject{
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Name: &ast.LiteralString{
																											Value: "left",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Body: &ast.LiteralBoolean{
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4021,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(351),
																														Column: int(41),
																													},
																													End: ast.Location{
																														Line: int(351),
																														Column: int(45),
																													},
																												},
																											},
																											Value: true,
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(351),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(351),
																												Column: int(45),
																											},
																										},
																										Hide: ast.ObjectFieldHide(1),
																										PlusSuper: false,
																									},
																								},
																								Locals: ast.LocalBinds{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4009,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(351),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(351),
																											Column: int(47),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4009,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(351),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(351),
																											Column: int(32),
																										},
																									},
																								},
																							},
																							OpFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4009,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(351),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(351),
																										Column: int(47),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(351),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(351),
																						Column: int(48),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		BranchFalse: &ast.Conditional{
																			Cond: &ast.Binary{
																				Right: &ast.LiteralString{
																					Value: " ",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(352),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(352),
																								Column: int(25),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Left: &ast.Var{
																					Id: "c",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(352),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(352),
																								Column: int(18),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(352),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(352),
																							Column: int(25),
																						},
																					},
																				},
																				Op: ast.BinaryOp(12),
																			},
																			BranchTrue: &ast.Apply{
																				Target: &ast.Var{
																					Id: "consume",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																							},
																						},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{
																							"consume",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(353),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(353),
																								Column: int(18),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4038,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(353),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(353),
																											Column: int(22),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4038,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(353),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(353),
																												Column: int(29),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "j",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4038,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(353),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(353),
																												Column: int(25),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4038,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(353),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(353),
																											Column: int(29),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.DesugaredObject{
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "blank",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.LiteralBoolean{
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4050,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(353),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(353),
																															Column: int(46),
																														},
																													},
																												},
																												Value: true,
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(353),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(353),
																													Column: int(46),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																									},
																									Locals: ast.LocalBinds{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4038,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(353),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(353),
																												Column: int(48),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "v",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4038,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(353),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(353),
																												Column: int(32),
																											},
																										},
																									},
																								},
																								OpFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4038,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(353),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(353),
																											Column: int(48),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(353),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(353),
																							Column: int(49),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			BranchFalse: &ast.Conditional{
																				Cond: &ast.Binary{
																					Right: &ast.LiteralString{
																						Value: "+",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3920,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(354),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(354),
																									Column: int(25),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Left: &ast.Var{
																						Id: "c",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p3920,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(354),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(354),
																									Column: int(18),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(354),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(354),
																								Column: int(25),
																							},
																						},
																					},
																					Op: ast.BinaryOp(12),
																				},
																				BranchTrue: &ast.Apply{
																					Target: &ast.Var{
																						Id: "consume",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																								},
																							},
																							Ctx: p3920,
																							FreeVars: ast.Identifiers{
																								"consume",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(355),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(355),
																									Column: int(18),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "str",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4067,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(355),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(355),
																												Column: int(22),
																											},
																										},
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "1",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4067,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(355),
																													Column: int(28),
																												},
																												End: ast.Location{
																													Line: int(355),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "j",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4067,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(355),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(355),
																													Column: int(25),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4067,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(355),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(355),
																												Column: int(29),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.DesugaredObject{
																										Asserts: ast.Nodes{},
																										Fields: ast.DesugaredObjectFields{
																											ast.DesugaredObjectField{
																												Name: &ast.LiteralString{
																													Value: "plus",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Body: &ast.LiteralBoolean{
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4079,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(355),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(355),
																																Column: int(45),
																															},
																														},
																													},
																													Value: true,
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(355),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(355),
																														Column: int(45),
																													},
																												},
																												Hide: ast.ObjectFieldHide(1),
																												PlusSuper: false,
																											},
																										},
																										Locals: ast.LocalBinds{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4067,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(355),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(355),
																													Column: int(47),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4067,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(355),
																													Column: int(31),
																												},
																												End: ast.Location{
																													Line: int(355),
																													Column: int(32),
																												},
																											},
																										},
																									},
																									OpFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4067,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(355),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(355),
																												Column: int(47),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(355),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(355),
																								Column: int(48),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				BranchFalse: &ast.DesugaredObject{
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Name: &ast.LiteralString{
																								Value: "i",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Body: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4088,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(357),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(357),
																											Column: int(17),
																										},
																									},
																								},
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(357),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(357),
																									Column: int(17),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																						ast.DesugaredObjectField{
																							Name: &ast.LiteralString{
																								Value: "v",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Body: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4088,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(357),
																											Column: int(22),
																										},
																										End: ast.Location{
																											Line: int(357),
																											Column: int(23),
																										},
																									},
																								},
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(357),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(357),
																									Column: int(23),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																					},
																					Locals: ast.LocalBinds{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																							},
																						},
																						Ctx: p3920,
																						FreeVars: ast.Identifiers{
																							"j",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(357),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(357),
																								Column: int(25),
																							},
																						},
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p3920,
																					FreeVars: ast.Identifiers{
																						"c",
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(354),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(357),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p3920,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(352),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(357),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(350),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(357),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p3920,
																		FreeVars: ast.Identifiers{
																			"c",
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(348),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(357),
																				Column: int(25),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p3920,
																	FreeVars: ast.Identifiers{
																		"c",
																		"consume",
																		"j",
																		"str",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(357),
																			Column: int(25),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p3920,
																FreeVars: ast.Identifiers{
																	"consume",
																	"j",
																	"str",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(345),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(357),
																		Column: int(25),
																	},
																},
															},
														},
														BranchFalse: &ast.Error{
															Expr: &ast.LiteralString{
																Value: "Truncated format code.",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p3920,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(344),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(344),
																			Column: int(62),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(344),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(357),
																		Column: int(25),
																	},
																},
															},
														},
														ThenFodder: nil,
														ElseFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"consume",
																"j",
																"std",
																"str",
																"v",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "str",
															CommaFodder: ast.Fodder{},
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(343),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(343),
																	Column: int(24),
																},
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "j",
															CommaFodder: ast.Fodder{},
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(343),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(343),
																	Column: int(27),
																},
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "v",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(343),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(343),
																	Column: int(30),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: p4112,
														FreeVars: ast.Identifiers{
															"consume",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(343),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(357),
																Column: int(25),
															},
														},
													},
													TrailingComma: false,
												},
												EqFodder: nil,
												Variable: "consume",
												CloseFodder: nil,
												Fun: nil,
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Var{
												Id: "consume",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p4117,
													FreeVars: ast.Identifiers{
														"consume",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(358),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(358),
															Column: int(14),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4121,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(358),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(358),
																		Column: int(18),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "i",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4121,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(358),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(358),
																		Column: int(21),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.DesugaredObject{
															Asserts: ast.Nodes{},
															Fields: ast.DesugaredObjectFields{
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "alt",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4129,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(358),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(358),
																					Column: int(35),
																				},
																			},
																		},
																		Value: false,
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(358),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(358),
																			Column: int(35),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "zero",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4129,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(358),
																					Column: int(43),
																				},
																				End: ast.Location{
																					Line: int(358),
																					Column: int(48),
																				},
																			},
																		},
																		Value: false,
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(358),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(358),
																			Column: int(48),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "left",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4129,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(358),
																					Column: int(56),
																				},
																				End: ast.Location{
																					Line: int(358),
																					Column: int(61),
																				},
																			},
																		},
																		Value: false,
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(358),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(358),
																			Column: int(61),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "blank",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4129,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(358),
																					Column: int(70),
																				},
																				End: ast.Location{
																					Line: int(358),
																					Column: int(75),
																				},
																			},
																		},
																		Value: false,
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(358),
																			Column: int(63),
																		},
																		End: ast.Location{
																			Line: int(358),
																			Column: int(75),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "plus",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4129,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(358),
																					Column: int(83),
																				},
																				End: ast.Location{
																					Line: int(358),
																					Column: int(88),
																				},
																			},
																		},
																		Value: false,
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(358),
																			Column: int(77),
																		},
																		End: ast.Location{
																			Line: int(358),
																			Column: int(88),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
															},
															Locals: ast.LocalBinds{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4121,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(358),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(358),
																		Column: int(90),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p4117,
												FreeVars: ast.Identifiers{
													"consume",
													"i",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(358),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(358),
														Column: int(91),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p4117,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(343),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(358),
													Column: int(91),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "str",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(342),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(342),
													Column: int(31),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(342),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(342),
													Column: int(34),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p4142,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(342),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(358),
												Column: int(91),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "try_parse_cflags",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Body: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Binary{
													Right: &ast.LiteralString{
														Value: "*",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4151,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(361),
																	Column: int(43),
																},
																End: ast.Location{
																	Line: int(361),
																	Column: int(46),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													Left: &ast.Index{
														Target: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4151,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(361),
																		Column: int(33),
																	},
																	End: ast.Location{
																		Line: int(361),
																		Column: int(36),
																	},
																},
															},
														},
														Index: &ast.Var{
															Id: "i",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4151,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(361),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(361),
																		Column: int(38),
																	},
																},
															},
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4151,
															FreeVars: ast.Identifiers{
																"i",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(361),
																	Column: int(33),
																},
																End: ast.Location{
																	Line: int(361),
																	Column: int(39),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p4151,
														FreeVars: ast.Identifiers{
															"i",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(361),
																Column: int(33),
															},
															End: ast.Location{
																Line: int(361),
																Column: int(46),
															},
														},
													},
													Op: ast.BinaryOp(12),
												},
												Left: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(361),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(361),
																			Column: int(17),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "length",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4151,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(361),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(361),
																		Column: int(24),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4168,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(361),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(361),
																					Column: int(28),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4151,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(361),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(361),
																	Column: int(29),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Var{
														Id: "i",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4151,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(361),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(361),
																	Column: int(11),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p4151,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(361),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(361),
																Column: int(29),
															},
														},
													},
													Op: ast.BinaryOp(9),
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p4151,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(361),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(361),
															Column: int(46),
														},
													},
												},
												Op: ast.BinaryOp(17),
											},
											BranchTrue: &ast.DesugaredObject{
												Asserts: ast.Nodes{},
												Fields: ast.DesugaredObjectFields{
													ast.DesugaredObjectField{
														Name: &ast.LiteralString{
															Value: "i",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														Body: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4180,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(362),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(362),
																			Column: int(19),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4180,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(362),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(362),
																			Column: int(15),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4180,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(362),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(362),
																		Column: int(19),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(362),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(362),
																Column: int(19),
															},
														},
														Hide: ast.ObjectFieldHide(1),
														PlusSuper: false,
													},
													ast.DesugaredObjectField{
														Name: &ast.LiteralString{
															Value: "v",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														Body: &ast.LiteralString{
															Value: "*",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4180,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(362),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(362),
																		Column: int(27),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(362),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(362),
																Column: int(27),
															},
														},
														Hide: ast.ObjectFieldHide(1),
														PlusSuper: false,
													},
												},
												Locals: ast.LocalBinds{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p4151,
													FreeVars: ast.Identifiers{
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(362),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(362),
															Column: int(29),
														},
													},
												},
											},
											BranchFalse: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Body: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(365),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(365),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "length",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4198,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(365),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(365),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4202,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(365),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(365),
																									Column: int(36),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4198,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(365),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(365),
																					Column: int(37),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	Left: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4198,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(365),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(365),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4198,
																		FreeVars: ast.Identifiers{
																			"j",
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(365),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(365),
																				Column: int(37),
																			},
																		},
																	},
																	Op: ast.BinaryOp(9),
																},
																BranchTrue: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Index{
																				Target: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4212,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(366),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(366),
																								Column: int(24),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "j",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4212,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(366),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(366),
																								Column: int(26),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4212,
																					FreeVars: ast.Identifiers{
																						"j",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(366),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(366),
																							Column: int(27),
																						},
																					},
																				},
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "c",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(366),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(366),
																					Column: int(27),
																				},
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "0",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(367),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(367),
																							Column: int(22),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(367),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(367),
																							Column: int(15),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4198,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(367),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(367),
																						Column: int(22),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		BranchTrue: &ast.Apply{
																			Target: &ast.Var{
																				Id: "consume",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																						},
																					},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(368),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(368),
																							Column: int(20),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4229,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(368),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(368),
																										Column: int(24),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4229,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(368),
																											Column: int(30),
																										},
																										End: ast.Location{
																											Line: int(368),
																											Column: int(31),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4229,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(368),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(368),
																											Column: int(27),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4229,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(368),
																										Column: int(26),
																									},
																									End: ast.Location{
																										Line: int(368),
																										Column: int(31),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "0",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4229,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(368),
																											Column: int(42),
																										},
																										End: ast.Location{
																											Line: int(368),
																											Column: int(43),
																										},
																									},
																								},
																							},
																							Left: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "10",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4229,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(368),
																												Column: int(37),
																											},
																											End: ast.Location{
																												Line: int(368),
																												Column: int(39),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "v",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4229,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(368),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(368),
																												Column: int(34),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4229,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(368),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(368),
																											Column: int(39),
																										},
																									},
																								},
																								Op: ast.BinaryOp(0),
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4229,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(368),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(368),
																										Column: int(43),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4198,
																				FreeVars: ast.Identifiers{
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(368),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(368),
																						Column: int(44),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		BranchFalse: &ast.Conditional{
																			Cond: &ast.Binary{
																				Right: &ast.LiteralString{
																					Value: "1",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(369),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(369),
																								Column: int(27),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Left: &ast.Var{
																					Id: "c",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(369),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(369),
																								Column: int(20),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(369),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(369),
																							Column: int(27),
																						},
																					},
																				},
																				Op: ast.BinaryOp(12),
																			},
																			BranchTrue: &ast.Apply{
																				Target: &ast.Var{
																					Id: "consume",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(12),
																							},
																						},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{
																							"consume",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(370),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(370),
																								Column: int(20),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4257,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(370),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(370),
																											Column: int(24),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4257,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(370),
																												Column: int(30),
																											},
																											End: ast.Location{
																												Line: int(370),
																												Column: int(31),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "j",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4257,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(370),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(370),
																												Column: int(27),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4257,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(370),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(370),
																											Column: int(31),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4257,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(370),
																												Column: int(42),
																											},
																											End: ast.Location{
																												Line: int(370),
																												Column: int(43),
																											},
																										},
																									},
																								},
																								Left: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "10",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4257,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(370),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(370),
																													Column: int(39),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4257,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(370),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(370),
																													Column: int(34),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4257,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(370),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(370),
																												Column: int(39),
																											},
																										},
																									},
																									Op: ast.BinaryOp(0),
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4257,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(370),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(370),
																											Column: int(43),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(370),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(370),
																							Column: int(44),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			BranchFalse: &ast.Conditional{
																				Cond: &ast.Binary{
																					Right: &ast.LiteralString{
																						Value: "2",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(371),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(371),
																									Column: int(27),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Left: &ast.Var{
																						Id: "c",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(371),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(371),
																									Column: int(20),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(371),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(371),
																								Column: int(27),
																							},
																						},
																					},
																					Op: ast.BinaryOp(12),
																				},
																				BranchTrue: &ast.Apply{
																					Target: &ast.Var{
																						Id: "consume",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(12),
																								},
																							},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{
																								"consume",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(372),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(372),
																									Column: int(20),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "str",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4285,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(372),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(372),
																												Column: int(24),
																											},
																										},
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "1",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4285,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(372),
																													Column: int(30),
																												},
																												End: ast.Location{
																													Line: int(372),
																													Column: int(31),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "j",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4285,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(372),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(372),
																													Column: int(27),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4285,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(372),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(372),
																												Column: int(31),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "2",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4285,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(372),
																													Column: int(42),
																												},
																												End: ast.Location{
																													Line: int(372),
																													Column: int(43),
																												},
																											},
																										},
																									},
																									Left: &ast.Binary{
																										Right: &ast.LiteralNumber{
																											OriginalString: "10",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4285,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(372),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(372),
																														Column: int(39),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "v",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4285,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(372),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(372),
																														Column: int(34),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4285,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(372),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(372),
																													Column: int(39),
																												},
																											},
																										},
																										Op: ast.BinaryOp(0),
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4285,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(372),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(372),
																												Column: int(43),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(372),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(372),
																								Column: int(44),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				BranchFalse: &ast.Conditional{
																					Cond: &ast.Binary{
																						Right: &ast.LiteralString{
																							Value: "3",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(373),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(373),
																										Column: int(27),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Left: &ast.Var{
																							Id: "c",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(373),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(373),
																										Column: int(20),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(373),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(373),
																									Column: int(27),
																								},
																							},
																						},
																						Op: ast.BinaryOp(12),
																					},
																					BranchTrue: &ast.Apply{
																						Target: &ast.Var{
																							Id: "consume",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(12),
																									},
																								},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{
																									"consume",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(374),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(374),
																										Column: int(20),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "str",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4313,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(374),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(374),
																													Column: int(24),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.LiteralNumber{
																											OriginalString: "1",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4313,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(374),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(374),
																														Column: int(31),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "j",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4313,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(374),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(374),
																														Column: int(27),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4313,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(374),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(374),
																													Column: int(31),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.LiteralNumber{
																											OriginalString: "3",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4313,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(374),
																														Column: int(42),
																													},
																													End: ast.Location{
																														Line: int(374),
																														Column: int(43),
																													},
																												},
																											},
																										},
																										Left: &ast.Binary{
																											Right: &ast.LiteralNumber{
																												OriginalString: "10",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4313,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(374),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(374),
																															Column: int(39),
																														},
																													},
																												},
																											},
																											Left: &ast.Var{
																												Id: "v",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4313,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(374),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(374),
																															Column: int(34),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4313,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(374),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(374),
																														Column: int(39),
																													},
																												},
																											},
																											Op: ast.BinaryOp(0),
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4313,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(374),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(374),
																													Column: int(43),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{
																								"consume",
																								"j",
																								"str",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(374),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(374),
																									Column: int(44),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					BranchFalse: &ast.Conditional{
																						Cond: &ast.Binary{
																							Right: &ast.LiteralString{
																								Value: "4",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(375),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(375),
																											Column: int(27),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Left: &ast.Var{
																								Id: "c",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(375),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(375),
																											Column: int(20),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(375),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(375),
																										Column: int(27),
																									},
																								},
																							},
																							Op: ast.BinaryOp(12),
																						},
																						BranchTrue: &ast.Apply{
																							Target: &ast.Var{
																								Id: "consume",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(12),
																										},
																									},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{
																										"consume",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(376),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(376),
																											Column: int(20),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "str",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4341,
																												FreeVars: ast.Identifiers{
																													"str",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(376),
																														Column: int(21),
																													},
																													End: ast.Location{
																														Line: int(376),
																														Column: int(24),
																													},
																												},
																											},
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.LiteralNumber{
																												OriginalString: "1",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4341,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(376),
																															Column: int(30),
																														},
																														End: ast.Location{
																															Line: int(376),
																															Column: int(31),
																														},
																													},
																												},
																											},
																											Left: &ast.Var{
																												Id: "j",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4341,
																													FreeVars: ast.Identifiers{
																														"j",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(376),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(376),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4341,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(376),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(376),
																														Column: int(31),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.LiteralNumber{
																												OriginalString: "4",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4341,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(376),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(376),
																															Column: int(43),
																														},
																													},
																												},
																											},
																											Left: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "10",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4341,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(376),
																																Column: int(37),
																															},
																															End: ast.Location{
																																Line: int(376),
																																Column: int(39),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4341,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(376),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(376),
																																Column: int(34),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4341,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(376),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(376),
																															Column: int(39),
																														},
																													},
																												},
																												Op: ast.BinaryOp(0),
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4341,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(376),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(376),
																														Column: int(43),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{
																									"consume",
																									"j",
																									"str",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(376),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(376),
																										Column: int(44),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						BranchFalse: &ast.Conditional{
																							Cond: &ast.Binary{
																								Right: &ast.LiteralString{
																									Value: "5",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(377),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(377),
																												Column: int(27),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Left: &ast.Var{
																									Id: "c",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(377),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(377),
																												Column: int(20),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(377),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(377),
																											Column: int(27),
																										},
																									},
																								},
																								Op: ast.BinaryOp(12),
																							},
																							BranchTrue: &ast.Apply{
																								Target: &ast.Var{
																									Id: "consume",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(12),
																											},
																										},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{
																											"consume",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(378),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(378),
																												Column: int(20),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "str",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4369,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(378),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(378),
																															Column: int(24),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "1",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4369,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(378),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(378),
																																Column: int(31),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "j",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4369,
																														FreeVars: ast.Identifiers{
																															"j",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(378),
																																Column: int(26),
																															},
																															End: ast.Location{
																																Line: int(378),
																																Column: int(27),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4369,
																													FreeVars: ast.Identifiers{
																														"j",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(378),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(378),
																															Column: int(31),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "5",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4369,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(378),
																																Column: int(42),
																															},
																															End: ast.Location{
																																Line: int(378),
																																Column: int(43),
																															},
																														},
																													},
																												},
																												Left: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "10",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4369,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(378),
																																	Column: int(37),
																																},
																																End: ast.Location{
																																	Line: int(378),
																																	Column: int(39),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "v",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4369,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(378),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(378),
																																	Column: int(34),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4369,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(378),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(378),
																																Column: int(39),
																															},
																														},
																													},
																													Op: ast.BinaryOp(0),
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4369,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(378),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(378),
																															Column: int(43),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{
																										"consume",
																										"j",
																										"str",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(378),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(378),
																											Column: int(44),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							BranchFalse: &ast.Conditional{
																								Cond: &ast.Binary{
																									Right: &ast.LiteralString{
																										Value: "6",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(379),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(379),
																													Column: int(27),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Left: &ast.Var{
																										Id: "c",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(379),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(379),
																													Column: int(20),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(379),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(379),
																												Column: int(27),
																											},
																										},
																									},
																									Op: ast.BinaryOp(12),
																								},
																								BranchTrue: &ast.Apply{
																									Target: &ast.Var{
																										Id: "consume",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(12),
																												},
																											},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{
																												"consume",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(380),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(380),
																													Column: int(20),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "str",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4397,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(380),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(380),
																																Column: int(24),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "1",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4397,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(380),
																																	Column: int(30),
																																},
																																End: ast.Location{
																																	Line: int(380),
																																	Column: int(31),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "j",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4397,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(380),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(380),
																																	Column: int(27),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4397,
																														FreeVars: ast.Identifiers{
																															"j",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(380),
																																Column: int(26),
																															},
																															End: ast.Location{
																																Line: int(380),
																																Column: int(31),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "6",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4397,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(380),
																																	Column: int(42),
																																},
																																End: ast.Location{
																																	Line: int(380),
																																	Column: int(43),
																																},
																															},
																														},
																													},
																													Left: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "10",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4397,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(380),
																																		Column: int(37),
																																	},
																																	End: ast.Location{
																																		Line: int(380),
																																		Column: int(39),
																																	},
																																},
																															},
																														},
																														Left: &ast.Var{
																															Id: "v",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4397,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(380),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(380),
																																		Column: int(34),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4397,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(380),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(380),
																																	Column: int(39),
																																},
																															},
																														},
																														Op: ast.BinaryOp(0),
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4397,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(380),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(380),
																																Column: int(43),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{
																											"consume",
																											"j",
																											"str",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(380),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(380),
																												Column: int(44),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								BranchFalse: &ast.Conditional{
																									Cond: &ast.Binary{
																										Right: &ast.LiteralString{
																											Value: "7",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(381),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(381),
																														Column: int(27),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Left: &ast.Var{
																											Id: "c",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(381),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(381),
																														Column: int(20),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(381),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(381),
																													Column: int(27),
																												},
																											},
																										},
																										Op: ast.BinaryOp(12),
																									},
																									BranchTrue: &ast.Apply{
																										Target: &ast.Var{
																											Id: "consume",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(12),
																													},
																												},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{
																													"consume",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(382),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(382),
																														Column: int(20),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "str",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4425,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(382),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(382),
																																	Column: int(24),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "1",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4425,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(382),
																																		Column: int(30),
																																	},
																																	End: ast.Location{
																																		Line: int(382),
																																		Column: int(31),
																																	},
																																},
																															},
																														},
																														Left: &ast.Var{
																															Id: "j",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4425,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(382),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(382),
																																		Column: int(27),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4425,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(382),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(382),
																																	Column: int(31),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "7",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4425,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(382),
																																		Column: int(42),
																																	},
																																	End: ast.Location{
																																		Line: int(382),
																																		Column: int(43),
																																	},
																																},
																															},
																														},
																														Left: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "10",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4425,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(382),
																																			Column: int(37),
																																		},
																																		End: ast.Location{
																																			Line: int(382),
																																			Column: int(39),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Var{
																																Id: "v",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4425,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(382),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(382),
																																			Column: int(34),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4425,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(382),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(382),
																																		Column: int(39),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(0),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4425,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(382),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(382),
																																	Column: int(43),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{
																												"consume",
																												"j",
																												"str",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(382),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(382),
																													Column: int(44),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									BranchFalse: &ast.Conditional{
																										Cond: &ast.Binary{
																											Right: &ast.LiteralString{
																												Value: "8",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(383),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(383),
																															Column: int(27),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Left: &ast.Var{
																												Id: "c",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(383),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(383),
																															Column: int(20),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(383),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(383),
																														Column: int(27),
																													},
																												},
																											},
																											Op: ast.BinaryOp(12),
																										},
																										BranchTrue: &ast.Apply{
																											Target: &ast.Var{
																												Id: "consume",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																														},
																													},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{
																														"consume",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(384),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(384),
																															Column: int(20),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "str",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4453,
																																FreeVars: ast.Identifiers{
																																	"str",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(384),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(384),
																																		Column: int(24),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "1",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4453,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(384),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(384),
																																			Column: int(31),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Var{
																																Id: "j",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4453,
																																	FreeVars: ast.Identifiers{
																																		"j",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(384),
																																			Column: int(26),
																																		},
																																		End: ast.Location{
																																			Line: int(384),
																																			Column: int(27),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4453,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(384),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(384),
																																		Column: int(31),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "8",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4453,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(384),
																																			Column: int(42),
																																		},
																																		End: ast.Location{
																																			Line: int(384),
																																			Column: int(43),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "10",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4453,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(384),
																																				Column: int(37),
																																			},
																																			End: ast.Location{
																																				Line: int(384),
																																				Column: int(39),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Var{
																																	Id: "v",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4453,
																																		FreeVars: ast.Identifiers{
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(384),
																																				Column: int(33),
																																			},
																																			End: ast.Location{
																																				Line: int(384),
																																				Column: int(34),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4453,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(384),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(384),
																																			Column: int(39),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(0),
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4453,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(384),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(384),
																																		Column: int(43),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{
																													"consume",
																													"j",
																													"str",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(384),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(384),
																														Column: int(44),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										BranchFalse: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.LiteralString{
																													Value: "9",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4198,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(385),
																																Column: int(24),
																															},
																															End: ast.Location{
																																Line: int(385),
																																Column: int(27),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Left: &ast.Var{
																													Id: "c",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4198,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(385),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(385),
																																Column: int(20),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(385),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(385),
																															Column: int(27),
																														},
																													},
																												},
																												Op: ast.BinaryOp(12),
																											},
																											BranchTrue: &ast.Apply{
																												Target: &ast.Var{
																													Id: "consume",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																															},
																														},
																														Ctx: p4198,
																														FreeVars: ast.Identifiers{
																															"consume",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(386),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(386),
																																Column: int(20),
																															},
																														},
																													},
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "str",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4481,
																																	FreeVars: ast.Identifiers{
																																		"str",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(386),
																																			Column: int(21),
																																		},
																																		End: ast.Location{
																																			Line: int(386),
																																			Column: int(24),
																																		},
																																	},
																																},
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "1",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4481,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(386),
																																				Column: int(30),
																																			},
																																			End: ast.Location{
																																				Line: int(386),
																																				Column: int(31),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Var{
																																	Id: "j",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4481,
																																		FreeVars: ast.Identifiers{
																																			"j",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(386),
																																				Column: int(26),
																																			},
																																			End: ast.Location{
																																				Line: int(386),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4481,
																																	FreeVars: ast.Identifiers{
																																		"j",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(386),
																																			Column: int(26),
																																		},
																																		End: ast.Location{
																																			Line: int(386),
																																			Column: int(31),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "9",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4481,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(386),
																																				Column: int(42),
																																			},
																																			End: ast.Location{
																																				Line: int(386),
																																				Column: int(43),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Binary{
																																	Right: &ast.LiteralNumber{
																																		OriginalString: "10",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p4481,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(386),
																																					Column: int(37),
																																				},
																																				End: ast.Location{
																																					Line: int(386),
																																					Column: int(39),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Var{
																																		Id: "v",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p4481,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(386),
																																					Column: int(33),
																																				},
																																				End: ast.Location{
																																					Line: int(386),
																																					Column: int(34),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p4481,
																																		FreeVars: ast.Identifiers{
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(386),
																																				Column: int(33),
																																			},
																																			End: ast.Location{
																																				Line: int(386),
																																				Column: int(39),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(0),
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4481,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(386),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(386),
																																			Column: int(43),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{
																														"consume",
																														"j",
																														"str",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(386),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(386),
																															Column: int(44),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											BranchFalse: &ast.DesugaredObject{
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "i",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Var{
																															Id: "j",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4501,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(388),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(388),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(388),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(388),
																																Column: int(19),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "v",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Var{
																															Id: "v",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4501,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(388),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(388),
																																		Column: int(25),
																																	},
																																},
																															},
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(388),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(388),
																																Column: int(25),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																												},
																												Locals: ast.LocalBinds{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																														},
																													},
																													Ctx: p4198,
																													FreeVars: ast.Identifiers{
																														"j",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(388),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(388),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(10),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4198,
																												FreeVars: ast.Identifiers{
																													"c",
																													"consume",
																													"j",
																													"str",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(385),
																														Column: int(16),
																													},
																													End: ast.Location{
																														Line: int(388),
																														Column: int(27),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4198,
																											FreeVars: ast.Identifiers{
																												"c",
																												"consume",
																												"j",
																												"str",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(383),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(388),
																													Column: int(27),
																												},
																											},
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4198,
																										FreeVars: ast.Identifiers{
																											"c",
																											"consume",
																											"j",
																											"str",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(381),
																												Column: int(16),
																											},
																											End: ast.Location{
																												Line: int(388),
																												Column: int(27),
																											},
																										},
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4198,
																									FreeVars: ast.Identifiers{
																										"c",
																										"consume",
																										"j",
																										"str",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(379),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(388),
																											Column: int(27),
																										},
																									},
																								},
																							},
																							ThenFodder: ast.Fodder{},
																							ElseFodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4198,
																								FreeVars: ast.Identifiers{
																									"c",
																									"consume",
																									"j",
																									"str",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(377),
																										Column: int(16),
																									},
																									End: ast.Location{
																										Line: int(388),
																										Column: int(27),
																									},
																								},
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4198,
																							FreeVars: ast.Identifiers{
																								"c",
																								"consume",
																								"j",
																								"str",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(375),
																									Column: int(16),
																								},
																								End: ast.Location{
																									Line: int(388),
																									Column: int(27),
																								},
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4198,
																						FreeVars: ast.Identifiers{
																							"c",
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(373),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(388),
																								Column: int(27),
																							},
																						},
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4198,
																					FreeVars: ast.Identifiers{
																						"c",
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(371),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(388),
																							Column: int(27),
																						},
																					},
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4198,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(369),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(388),
																						Column: int(27),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p4198,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(367),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(388),
																					Column: int(27),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		Ctx: p4198,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(366),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(388),
																				Column: int(27),
																			},
																		},
																	},
																},
																BranchFalse: &ast.Error{
																	Expr: &ast.LiteralString{
																		Value: "Truncated format code.",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4198,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(365),
																					Column: int(40),
																				},
																				End: ast.Location{
																					Line: int(365),
																					Column: int(64),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(365),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(388),
																				Column: int(27),
																			},
																		},
																	},
																},
																ThenFodder: nil,
																ElseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"consume",
																		"j",
																		"std",
																		"str",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(364),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(364),
																			Column: int(26),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "j",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(364),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(364),
																			Column: int(29),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "v",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(364),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(364),
																			Column: int(32),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: p4535,
																FreeVars: ast.Identifiers{
																	"consume",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(364),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(388),
																		Column: int(27),
																	},
																},
															},
															TrailingComma: false,
														},
														EqFodder: nil,
														Variable: "consume",
														CloseFodder: nil,
														Fun: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Body: &ast.Apply{
													Target: &ast.Var{
														Id: "consume",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p4151,
															FreeVars: ast.Identifiers{
																"consume",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(389),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(389),
																	Column: int(16),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "str",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4543,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(389),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(389),
																				Column: int(20),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4543,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(389),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(389),
																				Column: int(23),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNumber{
																	OriginalString: "0",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4543,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(389),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(389),
																				Column: int(26),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p4151,
														FreeVars: ast.Identifiers{
															"consume",
															"i",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(389),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(389),
																Column: int(27),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p4151,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(364),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(389),
															Column: int(27),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p4151,
												FreeVars: ast.Identifiers{
													"i",
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(361),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(389),
														Column: int(27),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "str",
												CommaFodder: ast.Fodder{},
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(360),
														Column: int(33),
													},
													End: ast.Location{
														Line: int(360),
														Column: int(36),
													},
												},
											},
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(360),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(360),
														Column: int(39),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: p4555,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(360),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(389),
													Column: int(27),
												},
											},
										},
										TrailingComma: false,
									},
									EqFodder: nil,
									Variable: "try_parse_field_width",
									CloseFodder: nil,
									Fun: nil,
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							Body: &ast.Local{
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: nil,
										Body: &ast.Function{
											ParenLeftFodder: ast.Fodder{},
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(392),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(392),
																			Column: int(21),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "length",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4567,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(392),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(392),
																		Column: int(28),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4571,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(392),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(392),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4567,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(392),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(392),
																	Column: int(33),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Var{
														Id: "i",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4567,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(392),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(392),
																	Column: int(15),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p4567,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(392),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(392),
																Column: int(33),
															},
														},
													},
													Op: ast.BinaryOp(9),
												},
												BranchTrue: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Index{
																Target: &ast.Var{
																	Id: "str",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4581,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(393),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(393),
																				Column: int(20),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4581,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(393),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(393),
																				Column: int(22),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4581,
																	FreeVars: ast.Identifiers{
																		"i",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(393),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(393),
																			Column: int(23),
																		},
																	},
																},
															},
															EqFodder: ast.Fodder{},
															Variable: "c",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(393),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(393),
																	Column: int(23),
																},
															},
														},
													},
													Body: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.LiteralString{
																Value: ".",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4567,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(394),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(394),
																			Column: int(18),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Left: &ast.Var{
																Id: "c",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4567,
																	FreeVars: ast.Identifiers{
																		"c",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(394),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(394),
																			Column: int(11),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4567,
																FreeVars: ast.Identifiers{
																	"c",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(394),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(394),
																		Column: int(18),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														BranchTrue: &ast.Apply{
															Target: &ast.Var{
																Id: "try_parse_field_width",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p4567,
																	FreeVars: ast.Identifiers{
																		"try_parse_field_width",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(395),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(395),
																			Column: int(30),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4598,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(395),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(395),
																						Column: int(34),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4598,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(395),
																							Column: int(40),
																						},
																						End: ast.Location{
																							Line: int(395),
																							Column: int(41),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4598,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(395),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(395),
																							Column: int(37),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4598,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(395),
																						Column: int(36),
																					},
																					End: ast.Location{
																						Line: int(395),
																						Column: int(41),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4567,
																FreeVars: ast.Identifiers{
																	"i",
																	"str",
																	"try_parse_field_width",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(395),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(395),
																		Column: int(42),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														BranchFalse: &ast.DesugaredObject{
															Asserts: ast.Nodes{},
															Fields: ast.DesugaredObjectFields{
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "i",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4610,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(397),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(397),
																					Column: int(15),
																				},
																			},
																		},
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(397),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(397),
																			Column: int(15),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
																ast.DesugaredObjectField{
																	Name: &ast.LiteralString{
																		Value: "v",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Body: &ast.LiteralNull{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4610,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(397),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(397),
																					Column: int(24),
																				},
																			},
																		},
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(397),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(397),
																			Column: int(24),
																		},
																	},
																	Hide: ast.ObjectFieldHide(1),
																	PlusSuper: false,
																},
															},
															Locals: ast.LocalBinds{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p4567,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(397),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(397),
																		Column: int(26),
																	},
																},
															},
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															Ctx: p4567,
															FreeVars: ast.Identifiers{
																"c",
																"i",
																"str",
																"try_parse_field_width",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(394),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(397),
																	Column: int(26),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p4567,
														FreeVars: ast.Identifiers{
															"i",
															"str",
															"try_parse_field_width",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(393),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(397),
																Column: int(26),
															},
														},
													},
												},
												BranchFalse: &ast.Error{
													Expr: &ast.LiteralString{
														Value: "Truncated format code.",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4567,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(392),
																	Column: int(36),
																},
																End: ast.Location{
																	Line: int(392),
																	Column: int(60),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(392),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(397),
																Column: int(26),
															},
														},
													},
												},
												ThenFodder: nil,
												ElseFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
														"try_parse_field_width",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "str",
													CommaFodder: ast.Fodder{},
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(391),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(391),
															Column: int(34),
														},
													},
												},
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(391),
															Column: int(36),
														},
														End: ast.Location{
															Line: int(391),
															Column: int(37),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: p4625,
												FreeVars: ast.Identifiers{
													"std",
													"try_parse_field_width",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(391),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(397),
														Column: int(26),
													},
												},
											},
											TrailingComma: false,
										},
										EqFodder: nil,
										Variable: "try_parse_precision",
										CloseFodder: nil,
										Fun: nil,
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Body: &ast.Function{
												ParenLeftFodder: ast.Fodder{},
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(401),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(401),
																				Column: int(21),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "length",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4637,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(401),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(401),
																			Column: int(28),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4641,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(401),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(401),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4637,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(401),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(401),
																		Column: int(33),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														Left: &ast.Var{
															Id: "i",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4637,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(401),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(401),
																		Column: int(15),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p4637,
															FreeVars: ast.Identifiers{
																"i",
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(401),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(401),
																	Column: int(33),
																},
															},
														},
														Op: ast.BinaryOp(9),
													},
													BranchTrue: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Body: &ast.Index{
																	Target: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4651,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(402),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(402),
																					Column: int(20),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4651,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(402),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(402),
																					Column: int(22),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4651,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(402),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(402),
																				Column: int(23),
																			},
																		},
																	},
																},
																EqFodder: ast.Fodder{},
																Variable: "c",
																CloseFodder: ast.Fodder{},
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(402),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(402),
																		Column: int(23),
																	},
																},
															},
														},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Binary{
																	Right: &ast.LiteralString{
																		Value: "L",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4637,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(403),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(403),
																					Column: int(42),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Left: &ast.Var{
																		Id: "c",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4637,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(403),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(403),
																					Column: int(35),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4637,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(403),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(403),
																				Column: int(42),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																Left: &ast.Binary{
																	Right: &ast.Binary{
																		Right: &ast.LiteralString{
																			Value: "l",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4637,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(403),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(403),
																						Column: int(30),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		Left: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4637,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(403),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(403),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4637,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(403),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(403),
																					Column: int(30),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	Left: &ast.Binary{
																		Right: &ast.LiteralString{
																			Value: "h",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4637,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(403),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(403),
																						Column: int(18),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		Left: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4637,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(403),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(403),
																						Column: int(11),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4637,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(403),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(403),
																					Column: int(18),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4637,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(403),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(403),
																				Column: int(30),
																			},
																		},
																	},
																	Op: ast.BinaryOp(18),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4637,
																	FreeVars: ast.Identifiers{
																		"c",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(403),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(403),
																			Column: int(42),
																		},
																	},
																},
																Op: ast.BinaryOp(18),
															},
															BranchTrue: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4637,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(404),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(404),
																				Column: int(14),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p4637,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(404),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(404),
																				Column: int(10),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4637,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(404),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(404),
																			Column: int(14),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															BranchFalse: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p4637,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(406),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(406),
																			Column: int(10),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																Ctx: p4637,
																FreeVars: ast.Identifiers{
																	"c",
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(403),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(406),
																		Column: int(10),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															Ctx: p4637,
															FreeVars: ast.Identifiers{
																"i",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(402),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(406),
																	Column: int(10),
																},
															},
														},
													},
													BranchFalse: &ast.Error{
														Expr: &ast.LiteralString{
															Value: "Truncated format code.",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4637,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(401),
																		Column: int(36),
																	},
																	End: ast.Location{
																		Line: int(401),
																		Column: int(60),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(401),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(406),
																	Column: int(10),
																},
															},
														},
													},
													ThenFodder: nil,
													ElseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "str",
														CommaFodder: ast.Fodder{},
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(400),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(400),
																Column: int(40),
															},
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "i",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(400),
																Column: int(42),
															},
															End: ast.Location{
																Line: int(400),
																Column: int(43),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: p4694,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(400),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(406),
															Column: int(10),
														},
													},
												},
												TrailingComma: false,
											},
											EqFodder: nil,
											Variable: "try_parse_length_modifier",
											CloseFodder: nil,
											Fun: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Body: &ast.Function{
													ParenLeftFodder: ast.Fodder{},
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(409),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(409),
																					Column: int(21),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "length",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4706,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(409),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(409),
																				Column: int(28),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4710,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(409),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(409),
																							Column: int(32),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4706,
																	FreeVars: ast.Identifiers{
																		"std",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(409),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(409),
																			Column: int(33),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															Left: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4706,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(409),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(409),
																			Column: int(15),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p4706,
																FreeVars: ast.Identifiers{
																	"i",
																	"std",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(409),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(409),
																		Column: int(33),
																	},
																},
															},
															Op: ast.BinaryOp(9),
														},
														BranchTrue: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Body: &ast.Index{
																		Target: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4720,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(410),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(410),
																						Column: int(20),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4720,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(410),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(410),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4720,
																			FreeVars: ast.Identifiers{
																				"i",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(410),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(410),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	EqFodder: ast.Fodder{},
																	Variable: "c",
																	CloseFodder: ast.Fodder{},
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(410),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(410),
																			Column: int(23),
																		},
																	},
																},
															},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Binary{
																		Right: &ast.LiteralString{
																			Value: "u",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(411),
																						Column: int(39),
																					},
																					End: ast.Location{
																						Line: int(411),
																						Column: int(42),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		Left: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(411),
																						Column: int(34),
																					},
																					End: ast.Location{
																						Line: int(411),
																						Column: int(35),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4706,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(411),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(411),
																					Column: int(42),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	Left: &ast.Binary{
																		Right: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "i",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(411),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(411),
																							Column: int(30),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(411),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(411),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(411),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(411),
																						Column: int(30),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		Left: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "d",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(411),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(411),
																							Column: int(18),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(411),
																							Column: int(10),
																						},
																						End: ast.Location{
																							Line: int(411),
																							Column: int(11),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(411),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(411),
																						Column: int(18),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4706,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(411),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(411),
																					Column: int(30),
																				},
																			},
																		},
																		Op: ast.BinaryOp(18),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4706,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(411),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(411),
																				Column: int(42),
																			},
																		},
																	},
																	Op: ast.BinaryOp(18),
																},
																BranchTrue: &ast.DesugaredObject{
																	Asserts: ast.Nodes{},
																	Fields: ast.DesugaredObjectFields{
																		ast.DesugaredObjectField{
																			Name: &ast.LiteralString{
																				Value: "i",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Body: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "1",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4750,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(412),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(412),
																								Column: int(19),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4750,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(412),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(412),
																								Column: int(15),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4750,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(412),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(412),
																							Column: int(19),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(412),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(412),
																					Column: int(19),
																				},
																			},
																			Hide: ast.ObjectFieldHide(1),
																			PlusSuper: false,
																		},
																		ast.DesugaredObjectField{
																			Name: &ast.LiteralString{
																				Value: "v",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Body: &ast.LiteralString{
																				Value: "d",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4750,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(412),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(412),
																							Column: int(27),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(412),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(412),
																					Column: int(27),
																				},
																			},
																			Hide: ast.ObjectFieldHide(1),
																			PlusSuper: false,
																		},
																		ast.DesugaredObjectField{
																			Name: &ast.LiteralString{
																				Value: "caps",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Body: &ast.LiteralBoolean{
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4750,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(412),
																							Column: int(35),
																						},
																						End: ast.Location{
																							Line: int(412),
																							Column: int(40),
																						},
																					},
																				},
																				Value: false,
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(412),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(412),
																					Column: int(40),
																				},
																			},
																			Hide: ast.ObjectFieldHide(1),
																			PlusSuper: false,
																		},
																	},
																	Locals: ast.LocalBinds{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p4706,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(412),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(412),
																				Column: int(42),
																			},
																		},
																	},
																},
																BranchFalse: &ast.Conditional{
																	Cond: &ast.Binary{
																		Right: &ast.LiteralString{
																			Value: "o",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(413),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(413),
																						Column: int(23),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		Left: &ast.Var{
																			Id: "c",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(413),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(413),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4706,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(413),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(413),
																					Column: int(23),
																				},
																			},
																		},
																		Op: ast.BinaryOp(12),
																	},
																	BranchTrue: &ast.DesugaredObject{
																		Asserts: ast.Nodes{},
																		Fields: ast.DesugaredObjectFields{
																			ast.DesugaredObjectField{
																				Name: &ast.LiteralString{
																					Value: "i",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Body: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4771,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(414),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(414),
																									Column: int(19),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4771,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(414),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(414),
																									Column: int(15),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4771,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(414),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(414),
																								Column: int(19),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(414),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(414),
																						Column: int(19),
																					},
																				},
																				Hide: ast.ObjectFieldHide(1),
																				PlusSuper: false,
																			},
																			ast.DesugaredObjectField{
																				Name: &ast.LiteralString{
																					Value: "v",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Body: &ast.LiteralString{
																					Value: "o",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4771,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(414),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(414),
																								Column: int(27),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(414),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(414),
																						Column: int(27),
																					},
																				},
																				Hide: ast.ObjectFieldHide(1),
																				PlusSuper: false,
																			},
																			ast.DesugaredObjectField{
																				Name: &ast.LiteralString{
																					Value: "caps",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Body: &ast.LiteralBoolean{
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4771,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(414),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(414),
																								Column: int(40),
																							},
																						},
																					},
																					Value: false,
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(414),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(414),
																						Column: int(40),
																					},
																				},
																				Hide: ast.ObjectFieldHide(1),
																				PlusSuper: false,
																			},
																		},
																		Locals: ast.LocalBinds{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p4706,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(414),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(414),
																					Column: int(42),
																				},
																			},
																		},
																	},
																	BranchFalse: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "x",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(415),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(415),
																							Column: int(23),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(415),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(415),
																							Column: int(16),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(415),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(415),
																						Column: int(23),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		BranchTrue: &ast.DesugaredObject{
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Name: &ast.LiteralString{
																						Value: "i",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Body: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4792,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(416),
																										Column: int(18),
																									},
																									End: ast.Location{
																										Line: int(416),
																										Column: int(19),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4792,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(416),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(416),
																										Column: int(15),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4792,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(416),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(416),
																									Column: int(19),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(416),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(416),
																							Column: int(19),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																				ast.DesugaredObjectField{
																					Name: &ast.LiteralString{
																						Value: "v",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Body: &ast.LiteralString{
																						Value: "x",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4792,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(416),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(416),
																									Column: int(27),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(416),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(416),
																							Column: int(27),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																				ast.DesugaredObjectField{
																					Name: &ast.LiteralString{
																						Value: "caps",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Body: &ast.LiteralBoolean{
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4792,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(416),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(416),
																									Column: int(40),
																								},
																							},
																						},
																						Value: false,
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(416),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(416),
																							Column: int(40),
																						},
																					},
																					Hide: ast.ObjectFieldHide(1),
																					PlusSuper: false,
																				},
																			},
																			Locals: ast.LocalBinds{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(416),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(416),
																						Column: int(42),
																					},
																				},
																			},
																		},
																		BranchFalse: &ast.Conditional{
																			Cond: &ast.Binary{
																				Right: &ast.LiteralString{
																					Value: "X",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4706,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(417),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(417),
																								Column: int(23),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				Left: &ast.Var{
																					Id: "c",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4706,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(417),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(417),
																								Column: int(16),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(417),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(417),
																							Column: int(23),
																						},
																					},
																				},
																				Op: ast.BinaryOp(12),
																			},
																			BranchTrue: &ast.DesugaredObject{
																				Asserts: ast.Nodes{},
																				Fields: ast.DesugaredObjectFields{
																					ast.DesugaredObjectField{
																						Name: &ast.LiteralString{
																							Value: "i",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Body: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4813,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(418),
																											Column: int(18),
																										},
																										End: ast.Location{
																											Line: int(418),
																											Column: int(19),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4813,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(418),
																											Column: int(14),
																										},
																										End: ast.Location{
																											Line: int(418),
																											Column: int(15),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4813,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(418),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(418),
																										Column: int(19),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(418),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(418),
																								Column: int(19),
																							},
																						},
																						Hide: ast.ObjectFieldHide(1),
																						PlusSuper: false,
																					},
																					ast.DesugaredObjectField{
																						Name: &ast.LiteralString{
																							Value: "v",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Body: &ast.LiteralString{
																							Value: "x",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4813,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(418),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(418),
																										Column: int(27),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(418),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(418),
																								Column: int(27),
																							},
																						},
																						Hide: ast.ObjectFieldHide(1),
																						PlusSuper: false,
																					},
																					ast.DesugaredObjectField{
																						Name: &ast.LiteralString{
																							Value: "caps",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Body: &ast.LiteralBoolean{
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4813,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(418),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(418),
																										Column: int(39),
																									},
																								},
																							},
																							Value: true,
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(418),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(418),
																								Column: int(39),
																							},
																						},
																						Hide: ast.ObjectFieldHide(1),
																						PlusSuper: false,
																					},
																				},
																				Locals: ast.LocalBinds{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(418),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(418),
																							Column: int(41),
																						},
																					},
																				},
																			},
																			BranchFalse: &ast.Conditional{
																				Cond: &ast.Binary{
																					Right: &ast.LiteralString{
																						Value: "e",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4706,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(419),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(419),
																									Column: int(23),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					Left: &ast.Var{
																						Id: "c",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4706,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(419),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(419),
																									Column: int(16),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4706,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(419),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(419),
																								Column: int(23),
																							},
																						},
																					},
																					Op: ast.BinaryOp(12),
																				},
																				BranchTrue: &ast.DesugaredObject{
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Name: &ast.LiteralString{
																								Value: "i",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Body: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4834,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(420),
																												Column: int(18),
																											},
																											End: ast.Location{
																												Line: int(420),
																												Column: int(19),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "i",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4834,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(420),
																												Column: int(14),
																											},
																											End: ast.Location{
																												Line: int(420),
																												Column: int(15),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4834,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(420),
																											Column: int(14),
																										},
																										End: ast.Location{
																											Line: int(420),
																											Column: int(19),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(420),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(420),
																									Column: int(19),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																						ast.DesugaredObjectField{
																							Name: &ast.LiteralString{
																								Value: "v",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Body: &ast.LiteralString{
																								Value: "e",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4834,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(420),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(420),
																											Column: int(27),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(420),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(420),
																									Column: int(27),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																						ast.DesugaredObjectField{
																							Name: &ast.LiteralString{
																								Value: "caps",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Body: &ast.LiteralBoolean{
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4834,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(420),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(420),
																											Column: int(40),
																										},
																									},
																								},
																								Value: false,
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(420),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(420),
																									Column: int(40),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																					},
																					Locals: ast.LocalBinds{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: p4706,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(420),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(420),
																								Column: int(42),
																							},
																						},
																					},
																				},
																				BranchFalse: &ast.Conditional{
																					Cond: &ast.Binary{
																						Right: &ast.LiteralString{
																							Value: "E",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4706,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(421),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(421),
																										Column: int(23),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Left: &ast.Var{
																							Id: "c",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4706,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(421),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(421),
																										Column: int(16),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4706,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(421),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(421),
																									Column: int(23),
																								},
																							},
																						},
																						Op: ast.BinaryOp(12),
																					},
																					BranchTrue: &ast.DesugaredObject{
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "i",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "1",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4855,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(422),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(422),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "i",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4855,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(422),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(422),
																													Column: int(15),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4855,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(422),
																												Column: int(14),
																											},
																											End: ast.Location{
																												Line: int(422),
																												Column: int(19),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(422),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(422),
																										Column: int(19),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "v",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.LiteralString{
																									Value: "e",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4855,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(422),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(422),
																												Column: int(27),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(422),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(422),
																										Column: int(27),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "caps",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.LiteralBoolean{
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4855,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(422),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(422),
																												Column: int(39),
																											},
																										},
																									},
																									Value: true,
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(422),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(422),
																										Column: int(39),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																						},
																						Locals: ast.LocalBinds{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																								},
																							},
																							Ctx: p4706,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(422),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(422),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					BranchFalse: &ast.Conditional{
																						Cond: &ast.Binary{
																							Right: &ast.LiteralString{
																								Value: "f",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4706,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(423),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(423),
																											Column: int(23),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Left: &ast.Var{
																								Id: "c",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4706,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(423),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(423),
																											Column: int(16),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4706,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(423),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(423),
																										Column: int(23),
																									},
																								},
																							},
																							Op: ast.BinaryOp(12),
																						},
																						BranchTrue: &ast.DesugaredObject{
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Name: &ast.LiteralString{
																										Value: "i",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Body: &ast.Binary{
																										Right: &ast.LiteralNumber{
																											OriginalString: "1",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4876,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(424),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(424),
																														Column: int(19),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "i",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4876,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(424),
																														Column: int(14),
																													},
																													End: ast.Location{
																														Line: int(424),
																														Column: int(15),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4876,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(424),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(424),
																													Column: int(19),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(424),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(424),
																											Column: int(19),
																										},
																									},
																									Hide: ast.ObjectFieldHide(1),
																									PlusSuper: false,
																								},
																								ast.DesugaredObjectField{
																									Name: &ast.LiteralString{
																										Value: "v",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Body: &ast.LiteralString{
																										Value: "f",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4876,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(424),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(424),
																													Column: int(27),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(424),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(424),
																											Column: int(27),
																										},
																									},
																									Hide: ast.ObjectFieldHide(1),
																									PlusSuper: false,
																								},
																								ast.DesugaredObjectField{
																									Name: &ast.LiteralString{
																										Value: "caps",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Body: &ast.LiteralBoolean{
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4876,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(424),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(424),
																													Column: int(40),
																												},
																											},
																										},
																										Value: false,
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(424),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(424),
																											Column: int(40),
																										},
																									},
																									Hide: ast.ObjectFieldHide(1),
																									PlusSuper: false,
																								},
																							},
																							Locals: ast.LocalBinds{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(8),
																									},
																								},
																								Ctx: p4706,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(424),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(424),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						BranchFalse: &ast.Conditional{
																							Cond: &ast.Binary{
																								Right: &ast.LiteralString{
																									Value: "F",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4706,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(425),
																												Column: int(20),
																											},
																											End: ast.Location{
																												Line: int(425),
																												Column: int(23),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Left: &ast.Var{
																									Id: "c",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4706,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(425),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(425),
																												Column: int(16),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4706,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(425),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(425),
																											Column: int(23),
																										},
																									},
																								},
																								Op: ast.BinaryOp(12),
																							},
																							BranchTrue: &ast.DesugaredObject{
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Name: &ast.LiteralString{
																											Value: "i",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Body: &ast.Binary{
																											Right: &ast.LiteralNumber{
																												OriginalString: "1",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4897,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(426),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(426),
																															Column: int(19),
																														},
																													},
																												},
																											},
																											Left: &ast.Var{
																												Id: "i",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4897,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(426),
																															Column: int(14),
																														},
																														End: ast.Location{
																															Line: int(426),
																															Column: int(15),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4897,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(426),
																														Column: int(14),
																													},
																													End: ast.Location{
																														Line: int(426),
																														Column: int(19),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(426),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(426),
																												Column: int(19),
																											},
																										},
																										Hide: ast.ObjectFieldHide(1),
																										PlusSuper: false,
																									},
																									ast.DesugaredObjectField{
																										Name: &ast.LiteralString{
																											Value: "v",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Body: &ast.LiteralString{
																											Value: "f",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4897,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(426),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(426),
																														Column: int(27),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(426),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(426),
																												Column: int(27),
																											},
																										},
																										Hide: ast.ObjectFieldHide(1),
																										PlusSuper: false,
																									},
																									ast.DesugaredObjectField{
																										Name: &ast.LiteralString{
																											Value: "caps",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Body: &ast.LiteralBoolean{
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4897,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(426),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(426),
																														Column: int(39),
																													},
																												},
																											},
																											Value: true,
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(426),
																												Column: int(29),
																											},
																											End: ast.Location{
																												Line: int(426),
																												Column: int(39),
																											},
																										},
																										Hide: ast.ObjectFieldHide(1),
																										PlusSuper: false,
																									},
																								},
																								Locals: ast.LocalBinds{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(8),
																										},
																									},
																									Ctx: p4706,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(426),
																											Column: int(9),
																										},
																										End: ast.Location{
																											Line: int(426),
																											Column: int(41),
																										},
																									},
																								},
																							},
																							BranchFalse: &ast.Conditional{
																								Cond: &ast.Binary{
																									Right: &ast.LiteralString{
																										Value: "g",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4706,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(427),
																													Column: int(20),
																												},
																												End: ast.Location{
																													Line: int(427),
																													Column: int(23),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Left: &ast.Var{
																										Id: "c",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4706,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(427),
																													Column: int(15),
																												},
																												End: ast.Location{
																													Line: int(427),
																													Column: int(16),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4706,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(427),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(427),
																												Column: int(23),
																											},
																										},
																									},
																									Op: ast.BinaryOp(12),
																								},
																								BranchTrue: &ast.DesugaredObject{
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "i",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "1",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4918,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(428),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(428),
																																Column: int(19),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "i",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4918,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(428),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(428),
																																Column: int(15),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4918,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(428),
																															Column: int(14),
																														},
																														End: ast.Location{
																															Line: int(428),
																															Column: int(19),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(428),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(428),
																													Column: int(19),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "v",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.LiteralString{
																												Value: "g",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4918,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(428),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(428),
																															Column: int(27),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(428),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(428),
																													Column: int(27),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "caps",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.LiteralBoolean{
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4918,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(428),
																															Column: int(35),
																														},
																														End: ast.Location{
																															Line: int(428),
																															Column: int(40),
																														},
																													},
																												},
																												Value: false,
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(428),
																													Column: int(29),
																												},
																												End: ast.Location{
																													Line: int(428),
																													Column: int(40),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																									},
																									Locals: ast.LocalBinds{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(8),
																											},
																										},
																										Ctx: p4706,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(428),
																												Column: int(9),
																											},
																											End: ast.Location{
																												Line: int(428),
																												Column: int(42),
																											},
																										},
																									},
																								},
																								BranchFalse: &ast.Conditional{
																									Cond: &ast.Binary{
																										Right: &ast.LiteralString{
																											Value: "G",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4706,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(429),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(429),
																														Column: int(23),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										Left: &ast.Var{
																											Id: "c",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4706,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(429),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(429),
																														Column: int(16),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4706,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(429),
																													Column: int(15),
																												},
																												End: ast.Location{
																													Line: int(429),
																													Column: int(23),
																												},
																											},
																										},
																										Op: ast.BinaryOp(12),
																									},
																									BranchTrue: &ast.DesugaredObject{
																										Asserts: ast.Nodes{},
																										Fields: ast.DesugaredObjectFields{
																											ast.DesugaredObjectField{
																												Name: &ast.LiteralString{
																													Value: "i",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Body: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "1",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4939,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(430),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(430),
																																	Column: int(19),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "i",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4939,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(430),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(430),
																																	Column: int(15),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4939,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(430),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(430),
																																Column: int(19),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(430),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(430),
																														Column: int(19),
																													},
																												},
																												Hide: ast.ObjectFieldHide(1),
																												PlusSuper: false,
																											},
																											ast.DesugaredObjectField{
																												Name: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Body: &ast.LiteralString{
																													Value: "g",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4939,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(430),
																																Column: int(24),
																															},
																															End: ast.Location{
																																Line: int(430),
																																Column: int(27),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(430),
																														Column: int(21),
																													},
																													End: ast.Location{
																														Line: int(430),
																														Column: int(27),
																													},
																												},
																												Hide: ast.ObjectFieldHide(1),
																												PlusSuper: false,
																											},
																											ast.DesugaredObjectField{
																												Name: &ast.LiteralString{
																													Value: "caps",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Body: &ast.LiteralBoolean{
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4939,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(430),
																																Column: int(35),
																															},
																															End: ast.Location{
																																Line: int(430),
																																Column: int(39),
																															},
																														},
																													},
																													Value: true,
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(430),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(430),
																														Column: int(39),
																													},
																												},
																												Hide: ast.ObjectFieldHide(1),
																												PlusSuper: false,
																											},
																										},
																										Locals: ast.LocalBinds{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(8),
																												},
																											},
																											Ctx: p4706,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(430),
																													Column: int(9),
																												},
																												End: ast.Location{
																													Line: int(430),
																													Column: int(41),
																												},
																											},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										Cond: &ast.Binary{
																											Right: &ast.LiteralString{
																												Value: "c",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4706,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(431),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(431),
																															Column: int(23),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Left: &ast.Var{
																												Id: "c",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4706,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(431),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(431),
																															Column: int(16),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4706,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(431),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(431),
																														Column: int(23),
																													},
																												},
																											},
																											Op: ast.BinaryOp(12),
																										},
																										BranchTrue: &ast.DesugaredObject{
																											Asserts: ast.Nodes{},
																											Fields: ast.DesugaredObjectFields{
																												ast.DesugaredObjectField{
																													Name: &ast.LiteralString{
																														Value: "i",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Body: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "1",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4960,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(432),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(432),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														Left: &ast.Var{
																															Id: "i",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4960,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(432),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(432),
																																		Column: int(15),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4960,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(432),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(432),
																																	Column: int(19),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(432),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(432),
																															Column: int(19),
																														},
																													},
																													Hide: ast.ObjectFieldHide(1),
																													PlusSuper: false,
																												},
																												ast.DesugaredObjectField{
																													Name: &ast.LiteralString{
																														Value: "v",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Body: &ast.LiteralString{
																														Value: "c",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4960,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(432),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(432),
																																	Column: int(27),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(432),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(432),
																															Column: int(27),
																														},
																													},
																													Hide: ast.ObjectFieldHide(1),
																													PlusSuper: false,
																												},
																												ast.DesugaredObjectField{
																													Name: &ast.LiteralString{
																														Value: "caps",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Body: &ast.LiteralBoolean{
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4960,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(432),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(432),
																																	Column: int(40),
																																},
																															},
																														},
																														Value: false,
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(432),
																															Column: int(29),
																														},
																														End: ast.Location{
																															Line: int(432),
																															Column: int(40),
																														},
																													},
																													Hide: ast.ObjectFieldHide(1),
																													PlusSuper: false,
																												},
																											},
																											Locals: ast.LocalBinds{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																													},
																												},
																												Ctx: p4706,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(432),
																														Column: int(9),
																													},
																													End: ast.Location{
																														Line: int(432),
																														Column: int(42),
																													},
																												},
																											},
																										},
																										BranchFalse: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.LiteralString{
																													Value: "s",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4706,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(433),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(433),
																																Column: int(23),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												Left: &ast.Var{
																													Id: "c",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4706,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(433),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(433),
																																Column: int(16),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4706,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(433),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(433),
																															Column: int(23),
																														},
																													},
																												},
																												Op: ast.BinaryOp(12),
																											},
																											BranchTrue: &ast.DesugaredObject{
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "i",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "1",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4981,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(434),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(434),
																																			Column: int(19),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Var{
																																Id: "i",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p4981,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(434),
																																			Column: int(14),
																																		},
																																		End: ast.Location{
																																			Line: int(434),
																																			Column: int(15),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4981,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(434),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(434),
																																		Column: int(19),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(434),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(434),
																																Column: int(19),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "v",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.LiteralString{
																															Value: "s",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4981,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(434),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(434),
																																		Column: int(27),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(434),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(434),
																																Column: int(27),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "caps",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.LiteralBoolean{
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4981,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(434),
																																		Column: int(35),
																																	},
																																	End: ast.Location{
																																		Line: int(434),
																																		Column: int(40),
																																	},
																																},
																															},
																															Value: false,
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(434),
																																Column: int(29),
																															},
																															End: ast.Location{
																																Line: int(434),
																																Column: int(40),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																												},
																												Locals: ast.LocalBinds{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													Ctx: p4706,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(434),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(434),
																															Column: int(42),
																														},
																													},
																												},
																											},
																											BranchFalse: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.LiteralString{
																														Value: "%",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4706,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(435),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(435),
																																	Column: int(23),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Left: &ast.Var{
																														Id: "c",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4706,
																															FreeVars: ast.Identifiers{
																																"c",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(435),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(435),
																																	Column: int(16),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p4706,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(435),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(435),
																																Column: int(23),
																															},
																														},
																													},
																													Op: ast.BinaryOp(12),
																												},
																												BranchTrue: &ast.DesugaredObject{
																													Asserts: ast.Nodes{},
																													Fields: ast.DesugaredObjectFields{
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "i",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "1",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5002,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(436),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(436),
																																				Column: int(19),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Var{
																																	Id: "i",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5002,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(436),
																																				Column: int(14),
																																			},
																																			End: ast.Location{
																																				Line: int(436),
																																				Column: int(15),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5002,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(436),
																																			Column: int(14),
																																		},
																																		End: ast.Location{
																																			Line: int(436),
																																			Column: int(19),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(436),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(436),
																																	Column: int(19),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "v",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.LiteralString{
																																Value: "%",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5002,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(436),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(436),
																																			Column: int(27),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(436),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(436),
																																	Column: int(27),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "caps",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.LiteralBoolean{
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5002,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(436),
																																			Column: int(35),
																																		},
																																		End: ast.Location{
																																			Line: int(436),
																																			Column: int(40),
																																		},
																																	},
																																},
																																Value: false,
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(436),
																																	Column: int(29),
																																},
																																End: ast.Location{
																																	Line: int(436),
																																	Column: int(40),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																													},
																													Locals: ast.LocalBinds{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p4706,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(436),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(436),
																																Column: int(42),
																															},
																														},
																													},
																												},
																												BranchFalse: &ast.Error{
																													Expr: &ast.Binary{
																														Right: &ast.Var{
																															Id: "c",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4706,
																																FreeVars: ast.Identifiers{
																																	"c",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(438),
																																		Column: int(50),
																																	},
																																	End: ast.Location{
																																		Line: int(438),
																																		Column: int(51),
																																	},
																																},
																															},
																														},
																														Left: &ast.LiteralString{
																															Value: "Unrecognised conversion type: ",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p4706,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(438),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(438),
																																		Column: int(47),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p4706,
																															FreeVars: ast.Identifiers{
																																"c",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(438),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(438),
																																	Column: int(51),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p4706,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(438),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(438),
																																Column: int(51),
																															},
																														},
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p4706,
																													FreeVars: ast.Identifiers{
																														"c",
																														"i",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(435),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(438),
																															Column: int(51),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p4706,
																												FreeVars: ast.Identifiers{
																													"c",
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(433),
																														Column: int(12),
																													},
																													End: ast.Location{
																														Line: int(438),
																														Column: int(51),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p4706,
																											FreeVars: ast.Identifiers{
																												"c",
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(431),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(438),
																													Column: int(51),
																												},
																											},
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p4706,
																										FreeVars: ast.Identifiers{
																											"c",
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(429),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(438),
																												Column: int(51),
																											},
																										},
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p4706,
																									FreeVars: ast.Identifiers{
																										"c",
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(427),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(438),
																											Column: int(51),
																										},
																									},
																								},
																							},
																							ThenFodder: ast.Fodder{},
																							ElseFodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p4706,
																								FreeVars: ast.Identifiers{
																									"c",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(425),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(438),
																										Column: int(51),
																									},
																								},
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p4706,
																							FreeVars: ast.Identifiers{
																								"c",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(423),
																									Column: int(12),
																								},
																								End: ast.Location{
																									Line: int(438),
																									Column: int(51),
																								},
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p4706,
																						FreeVars: ast.Identifiers{
																							"c",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(421),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(438),
																								Column: int(51),
																							},
																						},
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p4706,
																					FreeVars: ast.Identifiers{
																						"c",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(419),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(438),
																							Column: int(51),
																						},
																					},
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p4706,
																				FreeVars: ast.Identifiers{
																					"c",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(417),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(438),
																						Column: int(51),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p4706,
																			FreeVars: ast.Identifiers{
																				"c",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(415),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(438),
																					Column: int(51),
																				},
																			},
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p4706,
																		FreeVars: ast.Identifiers{
																			"c",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(413),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(438),
																				Column: int(51),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: p4706,
																	FreeVars: ast.Identifiers{
																		"c",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(411),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(438),
																			Column: int(51),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																Ctx: p4706,
																FreeVars: ast.Identifiers{
																	"i",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(410),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(438),
																		Column: int(51),
																	},
																},
															},
														},
														BranchFalse: &ast.Error{
															Expr: &ast.LiteralString{
																Value: "Truncated format code.",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p4706,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(409),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(409),
																			Column: int(60),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(409),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(438),
																		Column: int(51),
																	},
																},
															},
														},
														ThenFodder: nil,
														ElseFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"i",
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "str",
															CommaFodder: ast.Fodder{},
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(408),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(408),
																	Column: int(30),
																},
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "i",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(408),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(408),
																	Column: int(33),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: p5053,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(408),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(438),
																Column: int(51),
															},
														},
													},
													TrailingComma: false,
												},
												EqFodder: nil,
												Variable: "parse_conv_type",
												CloseFodder: nil,
												Fun: nil,
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Body: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(443),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(443),
																						Column: int(21),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "length",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p5065,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(443),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(443),
																					Column: int(28),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5069,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(443),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(443),
																								Column: int(32),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5065,
																		FreeVars: ast.Identifiers{
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(443),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(443),
																				Column: int(33),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5065,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(443),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(443),
																				Column: int(15),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p5065,
																	FreeVars: ast.Identifiers{
																		"i",
																		"std",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(443),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(443),
																			Column: int(33),
																		},
																	},
																},
																Op: ast.BinaryOp(9),
															},
															BranchTrue: &ast.Local{
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: ast.Fodder{},
																		Body: &ast.Apply{
																			Target: &ast.Var{
																				Id: "try_parse_mapping_key",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5079,
																					FreeVars: ast.Identifiers{
																						"try_parse_mapping_key",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(444),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(444),
																							Column: int(41),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5083,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(444),
																										Column: int(42),
																									},
																									End: ast.Location{
																										Line: int(444),
																										Column: int(45),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5083,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(444),
																										Column: int(47),
																									},
																									End: ast.Location{
																										Line: int(444),
																										Column: int(48),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5079,
																				FreeVars: ast.Identifiers{
																					"i",
																					"str",
																					"try_parse_mapping_key",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(444),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(444),
																						Column: int(49),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		EqFodder: ast.Fodder{},
																		Variable: "mkey",
																		CloseFodder: ast.Fodder{},
																		Fun: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(444),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(444),
																				Column: int(49),
																			},
																		},
																	},
																},
																Body: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				Target: &ast.Var{
																					Id: "try_parse_cflags",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5092,
																						FreeVars: ast.Identifiers{
																							"try_parse_cflags",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(445),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(445),
																								Column: int(38),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5096,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(445),
																											Column: int(39),
																										},
																										End: ast.Location{
																											Line: int(445),
																											Column: int(42),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								Target: &ast.Var{
																									Id: "mkey",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"mkey",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(445),
																												Column: int(44),
																											},
																											End: ast.Location{
																												Line: int(445),
																												Column: int(48),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "i",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5096,
																									FreeVars: ast.Identifiers{
																										"mkey",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(445),
																											Column: int(44),
																										},
																										End: ast.Location{
																											Line: int(445),
																											Column: int(50),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5092,
																					FreeVars: ast.Identifiers{
																						"mkey",
																						"str",
																						"try_parse_cflags",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(445),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(445),
																							Column: int(51),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "cflags",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(445),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(445),
																					Column: int(51),
																				},
																			},
																		},
																	},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Body: &ast.Apply{
																					Target: &ast.Var{
																						Id: "try_parse_field_width",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p5108,
																							FreeVars: ast.Identifiers{
																								"try_parse_field_width",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(446),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(446),
																									Column: int(39),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "str",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5112,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(446),
																												Column: int(40),
																											},
																											End: ast.Location{
																												Line: int(446),
																												Column: int(43),
																											},
																										},
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Index{
																									Target: &ast.Var{
																										Id: "cflags",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"cflags",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(446),
																													Column: int(45),
																												},
																												End: ast.Location{
																													Line: int(446),
																													Column: int(51),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "i",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5112,
																										FreeVars: ast.Identifiers{
																											"cflags",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(446),
																												Column: int(45),
																											},
																											End: ast.Location{
																												Line: int(446),
																												Column: int(53),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5108,
																						FreeVars: ast.Identifiers{
																							"cflags",
																							"str",
																							"try_parse_field_width",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(446),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(446),
																								Column: int(54),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				EqFodder: ast.Fodder{},
																				Variable: "fw",
																				CloseFodder: ast.Fodder{},
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(446),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(446),
																						Column: int(54),
																					},
																				},
																			},
																		},
																		Body: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						Target: &ast.Var{
																							Id: "try_parse_precision",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5124,
																								FreeVars: ast.Identifiers{
																									"try_parse_precision",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(447),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(447),
																										Column: int(39),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "str",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5128,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(447),
																													Column: int(40),
																												},
																												End: ast.Location{
																													Line: int(447),
																													Column: int(43),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "fw",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"fw",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(447),
																														Column: int(45),
																													},
																													End: ast.Location{
																														Line: int(447),
																														Column: int(47),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "i",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5128,
																											FreeVars: ast.Identifiers{
																												"fw",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(447),
																													Column: int(45),
																												},
																												End: ast.Location{
																													Line: int(447),
																													Column: int(49),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p5124,
																							FreeVars: ast.Identifiers{
																								"fw",
																								"str",
																								"try_parse_precision",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(447),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(447),
																									Column: int(50),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "prec",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(447),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(447),
																							Column: int(50),
																						},
																					},
																				},
																			},
																			Body: &ast.Local{
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: ast.Fodder{},
																						Body: &ast.Apply{
																							Target: &ast.Var{
																								Id: "try_parse_length_modifier",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5140,
																									FreeVars: ast.Identifiers{
																										"try_parse_length_modifier",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(448),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(448),
																											Column: int(48),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "str",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5144,
																												FreeVars: ast.Identifiers{
																													"str",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(448),
																														Column: int(49),
																													},
																													End: ast.Location{
																														Line: int(448),
																														Column: int(52),
																													},
																												},
																											},
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Index{
																											Target: &ast.Var{
																												Id: "prec",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"prec",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(448),
																															Column: int(54),
																														},
																														End: ast.Location{
																															Line: int(448),
																															Column: int(58),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "i",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5144,
																												FreeVars: ast.Identifiers{
																													"prec",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(448),
																														Column: int(54),
																													},
																													End: ast.Location{
																														Line: int(448),
																														Column: int(60),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5140,
																								FreeVars: ast.Identifiers{
																									"prec",
																									"str",
																									"try_parse_length_modifier",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(448),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(448),
																										Column: int(61),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						EqFodder: ast.Fodder{},
																						Variable: "len_mod",
																						CloseFodder: ast.Fodder{},
																						Fun: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(448),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(448),
																								Column: int(61),
																							},
																						},
																					},
																				},
																				Body: &ast.Local{
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Body: &ast.Apply{
																								Target: &ast.Var{
																									Id: "parse_conv_type",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5156,
																										FreeVars: ast.Identifiers{
																											"parse_conv_type",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(449),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(449),
																												Column: int(36),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "str",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5160,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(449),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(449),
																															Column: int(40),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "len_mod",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5160,
																													FreeVars: ast.Identifiers{
																														"len_mod",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(449),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(449),
																															Column: int(49),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5156,
																									FreeVars: ast.Identifiers{
																										"len_mod",
																										"parse_conv_type",
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(449),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(449),
																											Column: int(50),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							EqFodder: ast.Fodder{},
																							Variable: "ctype",
																							CloseFodder: ast.Fodder{},
																							Fun: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(449),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(449),
																									Column: int(50),
																								},
																							},
																						},
																					},
																					Body: &ast.DesugaredObject{
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "i",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.Index{
																									Target: &ast.Var{
																										Id: "ctype",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"ctype",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(451),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(451),
																													Column: int(17),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "i",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5172,
																										FreeVars: ast.Identifiers{
																											"ctype",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(451),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(451),
																												Column: int(19),
																											},
																										},
																									},
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(451),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(451),
																										Column: int(19),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																							ast.DesugaredObjectField{
																								Name: &ast.LiteralString{
																									Value: "code",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Body: &ast.DesugaredObject{
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "mkey",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "mkey",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"mkey",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(453),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(453),
																																Column: int(21),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"mkey",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(453),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(453),
																															Column: int(23),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(453),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(453),
																													Column: int(23),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "cflags",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "cflags",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"cflags",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(454),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(454),
																																Column: int(25),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"cflags",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(454),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(454),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(454),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(454),
																													Column: int(27),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "fw",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "fw",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"fw",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(455),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(455),
																																Column: int(17),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"fw",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(455),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(455),
																															Column: int(19),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(455),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(455),
																													Column: int(19),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "prec",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "prec",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"prec",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(456),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(456),
																																Column: int(21),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"prec",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(456),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(456),
																															Column: int(23),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(456),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(456),
																													Column: int(23),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "ctype",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "ctype",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"ctype",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(457),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(457),
																																Column: int(23),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "v",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"ctype",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(457),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(457),
																															Column: int(25),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(457),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(457),
																													Column: int(25),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																										ast.DesugaredObjectField{
																											Name: &ast.LiteralString{
																												Value: "caps",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Body: &ast.Index{
																												Target: &ast.Var{
																													Id: "ctype",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"ctype",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(458),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(458),
																																Column: int(22),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "caps",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"ctype",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(458),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(458),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(458),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(458),
																													Column: int(27),
																												},
																											},
																											Hide: ast.ObjectFieldHide(1),
																											PlusSuper: false,
																										},
																									},
																									Locals: ast.LocalBinds{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5172,
																										FreeVars: ast.Identifiers{
																											"cflags",
																											"ctype",
																											"fw",
																											"mkey",
																											"prec",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(452),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(459),
																												Column: int(10),
																											},
																										},
																									},
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(452),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(459),
																										Column: int(10),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																						},
																						Locals: ast.LocalBinds{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																								},
																							},
																							Ctx: p5065,
																							FreeVars: ast.Identifiers{
																								"cflags",
																								"ctype",
																								"fw",
																								"mkey",
																								"prec",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(450),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(460),
																									Column: int(8),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						Ctx: p5065,
																						FreeVars: ast.Identifiers{
																							"cflags",
																							"fw",
																							"len_mod",
																							"mkey",
																							"parse_conv_type",
																							"prec",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(449),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(460),
																								Column: int(8),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																						},
																					},
																					Ctx: p5065,
																					FreeVars: ast.Identifiers{
																						"cflags",
																						"fw",
																						"mkey",
																						"parse_conv_type",
																						"prec",
																						"str",
																						"try_parse_length_modifier",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(448),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(460),
																							Column: int(8),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																					},
																				},
																				Ctx: p5065,
																				FreeVars: ast.Identifiers{
																					"cflags",
																					"fw",
																					"mkey",
																					"parse_conv_type",
																					"str",
																					"try_parse_length_modifier",
																					"try_parse_precision",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(447),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(460),
																						Column: int(8),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			Ctx: p5065,
																			FreeVars: ast.Identifiers{
																				"cflags",
																				"mkey",
																				"parse_conv_type",
																				"str",
																				"try_parse_field_width",
																				"try_parse_length_modifier",
																				"try_parse_precision",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(446),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(460),
																					Column: int(8),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																			},
																		},
																		Ctx: p5065,
																		FreeVars: ast.Identifiers{
																			"mkey",
																			"parse_conv_type",
																			"str",
																			"try_parse_cflags",
																			"try_parse_field_width",
																			"try_parse_length_modifier",
																			"try_parse_precision",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(445),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(460),
																				Column: int(8),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: p5065,
																	FreeVars: ast.Identifiers{
																		"i",
																		"parse_conv_type",
																		"str",
																		"try_parse_cflags",
																		"try_parse_field_width",
																		"try_parse_length_modifier",
																		"try_parse_mapping_key",
																		"try_parse_precision",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(444),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(460),
																			Column: int(8),
																		},
																	},
																},
															},
															BranchFalse: &ast.Error{
																Expr: &ast.LiteralString{
																	Value: "Truncated format code.",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5065,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(443),
																				Column: int(36),
																			},
																			End: ast.Location{
																				Line: int(443),
																				Column: int(60),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(443),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(460),
																			Column: int(8),
																		},
																	},
																},
															},
															ThenFodder: nil,
															ElseFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"i",
																	"parse_conv_type",
																	"std",
																	"str",
																	"try_parse_cflags",
																	"try_parse_field_width",
																	"try_parse_length_modifier",
																	"try_parse_mapping_key",
																	"try_parse_precision",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "str",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(442),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(442),
																		Column: int(25),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(442),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(442),
																		Column: int(28),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: p5233,
															FreeVars: ast.Identifiers{
																"parse_conv_type",
																"std",
																"try_parse_cflags",
																"try_parse_field_width",
																"try_parse_length_modifier",
																"try_parse_mapping_key",
																"try_parse_precision",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(442),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(460),
																	Column: int(8),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: nil,
													Variable: "parse_code",
													CloseFodder: nil,
													Fun: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Body: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(464),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(464),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "length",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5245,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(464),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(464),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "str",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p5249,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(464),
																									Column: int(26),
																								},
																								End: ast.Location{
																									Line: int(464),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p5245,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(464),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(464),
																					Column: int(30),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p5245,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(464),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(464),
																					Column: int(11),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5245,
																		FreeVars: ast.Identifiers{
																			"i",
																			"std",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(464),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(464),
																				Column: int(30),
																			},
																		},
																	},
																	Op: ast.BinaryOp(8),
																},
																BranchTrue: &ast.Binary{
																	Right: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "cur",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5259,
																						FreeVars: ast.Identifiers{
																							"cur",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(465),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(465),
																								Column: int(19),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p5245,
																			FreeVars: ast.Identifiers{
																				"cur",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(465),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(465),
																					Column: int(20),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	Left: &ast.Var{
																		Id: "out",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p5245,
																			FreeVars: ast.Identifiers{
																				"out",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(465),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(465),
																					Column: int(12),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5245,
																		FreeVars: ast.Identifiers{
																			"cur",
																			"out",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(465),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(465),
																				Column: int(20),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																BranchFalse: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Index{
																				Target: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5270,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(467),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(467),
																								Column: int(22),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5270,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(467),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(467),
																								Column: int(24),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5270,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(467),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(467),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "c",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(467),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(467),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.LiteralString{
																				Value: "%",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5245,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(468),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(468),
																							Column: int(20),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			Left: &ast.Var{
																				Id: "c",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5245,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(468),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(468),
																							Column: int(13),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5245,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(468),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(468),
																						Column: int(20),
																					},
																				},
																			},
																			Op: ast.BinaryOp(12),
																		},
																		BranchTrue: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						Target: &ast.Var{
																							Id: "parse_code",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5285,
																								FreeVars: ast.Identifiers{
																									"parse_code",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(469),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(469),
																										Column: int(31),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "str",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5289,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(469),
																													Column: int(32),
																												},
																												End: ast.Location{
																													Line: int(469),
																													Column: int(35),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.LiteralNumber{
																											OriginalString: "1",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5289,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(469),
																														Column: int(41),
																													},
																													End: ast.Location{
																														Line: int(469),
																														Column: int(42),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "i",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5289,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(469),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(469),
																														Column: int(38),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5289,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(469),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(469),
																													Column: int(42),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p5285,
																							FreeVars: ast.Identifiers{
																								"i",
																								"parse_code",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(469),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(469),
																									Column: int(43),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "r",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(469),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(469),
																							Column: int(43),
																						},
																					},
																				},
																			},
																			Body: &ast.Apply{
																				Target: &ast.Var{
																					Id: "parse_codes",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																							},
																						},
																						Ctx: p5245,
																						FreeVars: ast.Identifiers{
																							"parse_codes",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(470),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(470),
																								Column: int(22),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5303,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(470),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(470),
																											Column: int(26),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								Target: &ast.Var{
																									Id: "r",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"r",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(470),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(470),
																												Column: int(29),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "i",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5303,
																									FreeVars: ast.Identifiers{
																										"r",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(470),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(470),
																											Column: int(31),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.Array{
																									Elements: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "cur",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5314,
																													FreeVars: ast.Identifiers{
																														"cur",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(470),
																															Column: int(40),
																														},
																														End: ast.Location{
																															Line: int(470),
																															Column: int(43),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "r",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"r",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(470),
																																Column: int(45),
																															},
																															End: ast.Location{
																																Line: int(470),
																																Column: int(46),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "code",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5314,
																													FreeVars: ast.Identifiers{
																														"r",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(470),
																															Column: int(45),
																														},
																														End: ast.Location{
																															Line: int(470),
																															Column: int(51),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									CloseFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5303,
																										FreeVars: ast.Identifiers{
																											"cur",
																											"r",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(470),
																												Column: int(39),
																											},
																											End: ast.Location{
																												Line: int(470),
																												Column: int(52),
																											},
																										},
																									},
																									TrailingComma: false,
																								},
																								Left: &ast.Var{
																									Id: "out",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5303,
																										FreeVars: ast.Identifiers{
																											"out",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(470),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(470),
																												Column: int(36),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5303,
																									FreeVars: ast.Identifiers{
																										"cur",
																										"out",
																										"r",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(470),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(470),
																											Column: int(52),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralString{
																								Value: "",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5303,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(470),
																											Column: int(54),
																										},
																										End: ast.Location{
																											Line: int(470),
																											Column: int(56),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5245,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"out",
																						"parse_codes",
																						"r",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(470),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(470),
																							Column: int(57),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: true,
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p5245,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"i",
																					"out",
																					"parse_code",
																					"parse_codes",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(469),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(470),
																						Column: int(57),
																					},
																				},
																			},
																		},
																		BranchFalse: &ast.Apply{
																			Target: &ast.Var{
																				Id: "parse_codes",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p5245,
																					FreeVars: ast.Identifiers{
																						"parse_codes",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(472),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(472),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5335,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(472),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(472),
																										Column: int(26),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5335,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(32),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5335,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(29),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5335,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(472),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(472),
																										Column: int(33),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "out",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5335,
																								FreeVars: ast.Identifiers{
																									"out",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(472),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(472),
																										Column: int(38),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.Var{
																								Id: "c",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5335,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(47),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "cur",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5335,
																									FreeVars: ast.Identifiers{
																										"cur",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(43),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5335,
																								FreeVars: ast.Identifiers{
																									"c",
																									"cur",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(472),
																										Column: int(40),
																									},
																									End: ast.Location{
																										Line: int(472),
																										Column: int(47),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5245,
																				FreeVars: ast.Identifiers{
																					"c",
																					"cur",
																					"i",
																					"out",
																					"parse_codes",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(472),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(472),
																						Column: int(48),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: true,
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p5245,
																			FreeVars: ast.Identifiers{
																				"c",
																				"cur",
																				"i",
																				"out",
																				"parse_code",
																				"parse_codes",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(468),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(472),
																					Column: int(48),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p5245,
																		FreeVars: ast.Identifiers{
																			"cur",
																			"i",
																			"out",
																			"parse_code",
																			"parse_codes",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(467),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(472),
																				Column: int(48),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: p5245,
																	FreeVars: ast.Identifiers{
																		"cur",
																		"i",
																		"out",
																		"parse_code",
																		"parse_codes",
																		"std",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(464),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(472),
																			Column: int(48),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(463),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(463),
																			Column: int(26),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(463),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(463),
																			Column: int(29),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "out",
																	CommaFodder: ast.Fodder{},
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(463),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(463),
																			Column: int(34),
																		},
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "cur",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(463),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(463),
																			Column: int(39),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: p5360,
																FreeVars: ast.Identifiers{
																	"parse_code",
																	"parse_codes",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(463),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(472),
																		Column: int(48),
																	},
																},
															},
															TrailingComma: false,
														},
														EqFodder: nil,
														Variable: "parse_codes",
														CloseFodder: nil,
														Fun: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Var{
																	Id: "parse_codes",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p5366,
																		FreeVars: ast.Identifiers{
																			"parse_codes",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(474),
																				Column: int(19),
																			},
																			End: ast.Location{
																				Line: int(474),
																				Column: int(30),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5370,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(474),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(474),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNumber{
																				OriginalString: "0",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5370,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(474),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(474),
																							Column: int(37),
																						},
																					},
																				},
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				Elements: nil,
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5370,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(474),
																							Column: int(39),
																						},
																						End: ast.Location{
																							Line: int(474),
																							Column: int(41),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralString{
																				Value: "",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5370,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(474),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(474),
																							Column: int(45),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p5366,
																	FreeVars: ast.Identifiers{
																		"parse_codes",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(474),
																			Column: int(19),
																		},
																		End: ast.Location{
																			Line: int(474),
																			Column: int(46),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															EqFodder: ast.Fodder{},
															Variable: "codes",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(474),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(474),
																	Column: int(46),
																},
															},
														},
													},
													Body: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: nil,
																Body: &ast.Function{
																	ParenLeftFodder: ast.Fodder{},
																	ParenRightFodder: ast.Fodder{},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: nil,
																				Body: &ast.Function{
																					ParenLeftFodder: ast.Fodder{},
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Conditional{
																						Cond: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "0",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5385,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(484),
																											Column: int(17),
																										},
																										End: ast.Location{
																											Line: int(484),
																											Column: int(18),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "w",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5385,
																									FreeVars: ast.Identifiers{
																										"w",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(484),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(484),
																											Column: int(13),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5385,
																								FreeVars: ast.Identifiers{
																									"w",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(484),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(484),
																										Column: int(18),
																									},
																								},
																							},
																							Op: ast.BinaryOp(10),
																						},
																						BranchTrue: &ast.Var{
																							Id: "v",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p5385,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(485),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(485),
																										Column: int(12),
																									},
																								},
																							},
																						},
																						BranchFalse: &ast.Apply{
																							Target: &ast.Var{
																								Id: "aux",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									Ctx: p5385,
																									FreeVars: ast.Identifiers{
																										"aux",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(487),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(487),
																											Column: int(14),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.LiteralNumber{
																												OriginalString: "1",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5399,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(487),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(487),
																															Column: int(20),
																														},
																													},
																												},
																											},
																											Left: &ast.Var{
																												Id: "w",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5399,
																													FreeVars: ast.Identifiers{
																														"w",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(487),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(487),
																															Column: int(16),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5399,
																												FreeVars: ast.Identifiers{
																													"w",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(487),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(487),
																														Column: int(20),
																													},
																												},
																											},
																											Op: ast.BinaryOp(4),
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.Var{
																												Id: "s",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5399,
																													FreeVars: ast.Identifiers{
																														"s",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(487),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(487),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											Left: &ast.Var{
																												Id: "v",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5399,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(487),
																															Column: int(22),
																														},
																														End: ast.Location{
																															Line: int(487),
																															Column: int(23),
																														},
																													},
																												},
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5399,
																												FreeVars: ast.Identifiers{
																													"s",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(487),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(487),
																														Column: int(27),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5385,
																								FreeVars: ast.Identifiers{
																									"aux",
																									"s",
																									"v",
																									"w",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(487),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(487),
																										Column: int(28),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																								},
																							},
																							Ctx: p5385,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"s",
																								"v",
																								"w",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(484),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(487),
																									Column: int(28),
																								},
																							},
																						},
																					},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "w",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(483),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(483),
																									Column: int(18),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "v",
																							CommaFodder: nil,
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(483),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(483),
																									Column: int(21),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: p5414,
																						FreeVars: ast.Identifiers{
																							"aux",
																							"s",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(483),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(487),
																								Column: int(28),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				EqFodder: nil,
																				Variable: "aux",
																				CloseFodder: nil,
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Body: &ast.Apply{
																			Target: &ast.Var{
																				Id: "aux",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																						},
																					},
																					Ctx: p5419,
																					FreeVars: ast.Identifiers{
																						"aux",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(488),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(488),
																							Column: int(10),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "w",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5423,
																								FreeVars: ast.Identifiers{
																									"w",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(488),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(488),
																										Column: int(12),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p5423,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(488),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(488),
																										Column: int(16),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5419,
																				FreeVars: ast.Identifiers{
																					"aux",
																					"w",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(488),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(488),
																						Column: int(17),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			Ctx: p5419,
																			FreeVars: ast.Identifiers{
																				"s",
																				"w",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(483),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(488),
																					Column: int(17),
																				},
																			},
																		},
																	},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "w",
																			CommaFodder: ast.Fodder{},
																			EqFodder: nil,
																			DefaultArg: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(482),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(482),
																					Column: int(20),
																				},
																			},
																		},
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "s",
																			CommaFodder: nil,
																			EqFodder: nil,
																			DefaultArg: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(482),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(482),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: p5430,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(482),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(488),
																				Column: int(17),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																EqFodder: nil,
																Variable: "padding",
																CloseFodder: nil,
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Body: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: nil,
																	Body: &ast.Function{
																		ParenLeftFodder: ast.Fodder{},
																		ParenRightFodder: ast.Fodder{},
																		Body: &ast.Binary{
																			Right: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5436,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(492),
																							Column: int(41),
																						},
																						End: ast.Location{
																							Line: int(492),
																							Column: int(44),
																						},
																					},
																				},
																			},
																			Left: &ast.Apply{
																				Target: &ast.Var{
																					Id: "padding",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						Ctx: p5436,
																						FreeVars: ast.Identifiers{
																							"padding",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(492),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(492),
																								Column: int(14),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(492),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(492),
																														Column: int(22),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "length",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5449,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(492),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(492),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "str",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5453,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(492),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(492),
																																Column: int(33),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5449,
																										FreeVars: ast.Identifiers{
																											"std",
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(492),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(492),
																												Column: int(34),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								Left: &ast.Var{
																									Id: "w",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5449,
																										FreeVars: ast.Identifiers{
																											"w",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(492),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(492),
																												Column: int(16),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5449,
																									FreeVars: ast.Identifiers{
																										"std",
																										"str",
																										"w",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(492),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(492),
																											Column: int(34),
																										},
																									},
																								},
																								Op: ast.BinaryOp(4),
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "s",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p5449,
																									FreeVars: ast.Identifiers{
																										"s",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(492),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(492),
																											Column: int(37),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5436,
																					FreeVars: ast.Identifiers{
																						"padding",
																						"s",
																						"std",
																						"str",
																						"w",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(492),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(492),
																							Column: int(38),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p5436,
																				FreeVars: ast.Identifiers{
																					"padding",
																					"s",
																					"std",
																					"str",
																					"w",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(492),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(492),
																						Column: int(44),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "str",
																				CommaFodder: ast.Fodder{},
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(491),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(491),
																						Column: int(23),
																					},
																				},
																			},
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "w",
																				CommaFodder: ast.Fodder{},
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(491),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(491),
																						Column: int(26),
																					},
																				},
																			},
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "s",
																				CommaFodder: nil,
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(491),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(491),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: p5464,
																			FreeVars: ast.Identifiers{
																				"padding",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(491),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(492),
																					Column: int(44),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	EqFodder: nil,
																	Variable: "pad_left",
																	CloseFodder: nil,
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Body: &ast.Local{
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: nil,
																		Body: &ast.Function{
																			ParenLeftFodder: ast.Fodder{},
																			ParenRightFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				Right: &ast.Apply{
																					Target: &ast.Var{
																						Id: "padding",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p5472,
																							FreeVars: ast.Identifiers{
																								"padding",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(496),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(496),
																									Column: int(20),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(496),
																															Column: int(25),
																														},
																														End: ast.Location{
																															Line: int(496),
																															Column: int(28),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "length",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5481,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(496),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(496),
																														Column: int(35),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "str",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5485,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(496),
																																	Column: int(36),
																																},
																																End: ast.Location{
																																	Line: int(496),
																																	Column: int(39),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5481,
																											FreeVars: ast.Identifiers{
																												"std",
																												"str",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(496),
																													Column: int(25),
																												},
																												End: ast.Location{
																													Line: int(496),
																													Column: int(40),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									Left: &ast.Var{
																										Id: "w",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5481,
																											FreeVars: ast.Identifiers{
																												"w",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(496),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(496),
																													Column: int(22),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5481,
																										FreeVars: ast.Identifiers{
																											"std",
																											"str",
																											"w",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(496),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(496),
																												Column: int(40),
																											},
																										},
																									},
																									Op: ast.BinaryOp(4),
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "s",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5481,
																										FreeVars: ast.Identifiers{
																											"s",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(496),
																												Column: int(42),
																											},
																											End: ast.Location{
																												Line: int(496),
																												Column: int(43),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p5472,
																						FreeVars: ast.Identifiers{
																							"padding",
																							"s",
																							"std",
																							"str",
																							"w",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(496),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(496),
																								Column: int(44),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				Left: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						Ctx: p5472,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(496),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(496),
																								Column: int(10),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p5472,
																					FreeVars: ast.Identifiers{
																						"padding",
																						"s",
																						"std",
																						"str",
																						"w",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(496),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(496),
																							Column: int(44),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "str",
																					CommaFodder: ast.Fodder{},
																					EqFodder: nil,
																					DefaultArg: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(495),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(495),
																							Column: int(24),
																						},
																					},
																				},
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "w",
																					CommaFodder: ast.Fodder{},
																					EqFodder: nil,
																					DefaultArg: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(495),
																							Column: int(26),
																						},
																						End: ast.Location{
																							Line: int(495),
																							Column: int(27),
																						},
																					},
																				},
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "s",
																					CommaFodder: nil,
																					EqFodder: nil,
																					DefaultArg: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(495),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(495),
																							Column: int(30),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: p5499,
																				FreeVars: ast.Identifiers{
																					"padding",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(495),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(496),
																						Column: int(44),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		EqFodder: nil,
																		Variable: "pad_right",
																		CloseFodder: nil,
																		Fun: nil,
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																Body: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: nil,
																			Body: &ast.Function{
																				ParenLeftFodder: ast.Fodder{},
																				ParenRightFodder: ast.Fodder{},
																				Body: &ast.Local{
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Body: &ast.Conditional{
																								Cond: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "0",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5509,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(514),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(514),
																													Column: int(20),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "mag",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5509,
																											FreeVars: ast.Identifiers{
																												"mag",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(514),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(514),
																													Column: int(15),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5509,
																										FreeVars: ast.Identifiers{
																											"mag",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(514),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(514),
																												Column: int(20),
																											},
																										},
																									},
																									Op: ast.BinaryOp(12),
																								},
																								BranchTrue: &ast.LiteralString{
																									Value: "0",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										Ctx: p5509,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(515),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(515),
																												Column: int(14),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								BranchFalse: &ast.Local{
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: nil,
																											Body: &ast.Function{
																												ParenLeftFodder: ast.Fodder{},
																												ParenRightFodder: ast.Fodder{},
																												Body: &ast.Conditional{
																													Cond: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "0",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5521,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(518),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(518),
																																		Column: int(22),
																																	},
																																},
																															},
																														},
																														Left: &ast.Var{
																															Id: "n",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5521,
																																FreeVars: ast.Identifiers{
																																	"n",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(518),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(518),
																																		Column: int(17),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5521,
																															FreeVars: ast.Identifiers{
																																"n",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(518),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(518),
																																	Column: int(22),
																																},
																															},
																														},
																														Op: ast.BinaryOp(12),
																													},
																													BranchTrue: &ast.Var{
																														Id: "zero_prefix",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(14),
																																},
																															},
																															Ctx: p5521,
																															FreeVars: ast.Identifiers{
																																"zero_prefix",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(26),
																																},
																															},
																														},
																													},
																													BranchFalse: &ast.Binary{
																														Right: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "$std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "mod",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: nil,
																																LeftBracketFodder: nil,
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "n",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5521,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(521),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(521),
																																						Column: int(45),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "radix",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5521,
																																				FreeVars: ast.Identifiers{
																																					"radix",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(521),
																																						Column: int(48),
																																					},
																																					End: ast.Location{
																																						Line: int(521),
																																						Column: int(53),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: nil,
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																	"n",
																																	"radix",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(521),
																																		Column: int(44),
																																	},
																																	End: ast.Location{
																																		Line: int(521),
																																		Column: int(53),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														Left: &ast.Apply{
																															Target: &ast.Var{
																																Id: "aux",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(14),
																																		},
																																	},
																																	Ctx: p5521,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(521),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(521),
																																			Column: int(18),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Apply{
																																			Target: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "std",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(521),
																																								Column: int(19),
																																							},
																																							End: ast.Location{
																																								Line: int(521),
																																								Column: int(22),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "floor",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p5551,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(521),
																																							Column: int(19),
																																						},
																																						End: ast.Location{
																																							Line: int(521),
																																							Column: int(28),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Binary{
																																							Right: &ast.Var{
																																								Id: "radix",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p5556,
																																									FreeVars: ast.Identifiers{
																																										"radix",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(521),
																																											Column: int(33),
																																										},
																																										End: ast.Location{
																																											Line: int(521),
																																											Column: int(38),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.Var{
																																								Id: "n",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p5556,
																																									FreeVars: ast.Identifiers{
																																										"n",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(521),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(521),
																																											Column: int(30),
																																										},
																																									},
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p5556,
																																								FreeVars: ast.Identifiers{
																																									"n",
																																									"radix",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(521),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(521),
																																										Column: int(38),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(1),
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5551,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																					"radix",
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(521),
																																						Column: int(19),
																																					},
																																					End: ast.Location{
																																						Line: int(521),
																																						Column: int(39),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5521,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"n",
																																	"radix",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(521),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(521),
																																		Column: int(40),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5521,
																															FreeVars: ast.Identifiers{
																																"$std",
																																"aux",
																																"n",
																																"radix",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(521),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(521),
																																	Column: int(54),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													ThenFodder: ast.Fodder{},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																															},
																														},
																														Ctx: p5521,
																														FreeVars: ast.Identifiers{
																															"$std",
																															"aux",
																															"n",
																															"radix",
																															"std",
																															"zero_prefix",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(521),
																																Column: int(54),
																															},
																														},
																													},
																												},
																												Parameters: []ast.Parameter{
																													ast.Parameter{
																														NameFodder: ast.Fodder{},
																														Name: "n",
																														CommaFodder: nil,
																														EqFodder: nil,
																														DefaultArg: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(517),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(517),
																																Column: int(22),
																															},
																														},
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: p5568,
																													FreeVars: ast.Identifiers{
																														"$std",
																														"aux",
																														"radix",
																														"std",
																														"zero_prefix",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(517),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(521),
																															Column: int(54),
																														},
																													},
																												},
																												TrailingComma: false,
																											},
																											EqFodder: nil,
																											Variable: "aux",
																											CloseFodder: nil,
																											Fun: nil,
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																									},
																									Body: &ast.Apply{
																										Target: &ast.Var{
																											Id: "aux",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																													},
																												},
																												Ctx: p5509,
																												FreeVars: ast.Identifiers{
																													"aux",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(522),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(522),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "mag",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5576,
																															FreeVars: ast.Identifiers{
																																"mag",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(522),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(522),
																																	Column: int(18),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5509,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"mag",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(522),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(522),
																													Column: int(19),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										Ctx: p5509,
																										FreeVars: ast.Identifiers{
																											"$std",
																											"mag",
																											"radix",
																											"std",
																											"zero_prefix",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(517),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(522),
																												Column: int(19),
																											},
																										},
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(8),
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(8),
																										},
																									},
																									Ctx: p5509,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"mag",
																										"radix",
																										"std",
																										"zero_prefix",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(514),
																											Column: int(9),
																										},
																										End: ast.Location{
																											Line: int(522),
																											Column: int(19),
																										},
																									},
																								},
																							},
																							EqFodder: ast.Fodder{},
																							Variable: "dec",
																							CloseFodder: ast.Fodder{},
																							Fun: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(513),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(522),
																									Column: int(19),
																								},
																							},
																						},
																					},
																					Body: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Body: &ast.Binary{
																									Right: &ast.Conditional{
																										Cond: &ast.Binary{
																											Right: &ast.Var{
																												Id: "plus",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5590,
																													FreeVars: ast.Identifiers{
																														"plus",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(523),
																															Column: int(50),
																														},
																														End: ast.Location{
																															Line: int(523),
																															Column: int(54),
																														},
																													},
																												},
																											},
																											Left: &ast.Binary{
																												Right: &ast.Var{
																													Id: "blank",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5590,
																														FreeVars: ast.Identifiers{
																															"blank",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(523),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(523),
																																Column: int(46),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "neg",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5590,
																														FreeVars: ast.Identifiers{
																															"neg",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(523),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(523),
																																Column: int(37),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5590,
																													FreeVars: ast.Identifiers{
																														"blank",
																														"neg",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(523),
																															Column: int(34),
																														},
																														End: ast.Location{
																															Line: int(523),
																															Column: int(46),
																														},
																													},
																												},
																												Op: ast.BinaryOp(18),
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5590,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"neg",
																													"plus",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(523),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(523),
																														Column: int(54),
																													},
																												},
																											},
																											Op: ast.BinaryOp(18),
																										},
																										BranchTrue: &ast.LiteralNumber{
																											OriginalString: "1",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5590,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(523),
																														Column: int(60),
																													},
																													End: ast.Location{
																														Line: int(523),
																														Column: int(61),
																													},
																												},
																											},
																										},
																										BranchFalse: &ast.LiteralNumber{
																											OriginalString: "0",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5590,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(523),
																														Column: int(67),
																													},
																													End: ast.Location{
																														Line: int(523),
																														Column: int(68),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5590,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"neg",
																												"plus",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(523),
																													Column: int(31),
																												},
																												End: ast.Location{
																													Line: int(523),
																													Column: int(68),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "min_chars",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5590,
																											FreeVars: ast.Identifiers{
																												"min_chars",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(523),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(523),
																													Column: int(27),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5590,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"min_chars",
																											"neg",
																											"plus",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(523),
																												Column: int(18),
																											},
																											End: ast.Location{
																												Line: int(523),
																												Column: int(69),
																											},
																										},
																									},
																									Op: ast.BinaryOp(4),
																								},
																								EqFodder: ast.Fodder{},
																								Variable: "zp",
																								CloseFodder: ast.Fodder{},
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(523),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(523),
																										Column: int(69),
																									},
																								},
																							},
																						},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(524),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(524),
																															Column: int(22),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "max",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5612,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(524),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(524),
																														Column: int(26),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "zp",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5616,
																															FreeVars: ast.Identifiers{
																																"zp",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(524),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(524),
																																	Column: int(29),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "min_digits",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5616,
																															FreeVars: ast.Identifiers{
																																"min_digits",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(524),
																																	Column: int(31),
																																},
																																End: ast.Location{
																																	Line: int(524),
																																	Column: int(41),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5612,
																											FreeVars: ast.Identifiers{
																												"min_digits",
																												"std",
																												"zp",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(524),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(524),
																													Column: int(42),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									EqFodder: ast.Fodder{},
																									Variable: "zp2",
																									CloseFodder: ast.Fodder{},
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(524),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(524),
																											Column: int(42),
																										},
																									},
																								},
																							},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											Target: &ast.Var{
																												Id: "pad_left",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5625,
																													FreeVars: ast.Identifiers{
																														"pad_left",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(525),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(525),
																															Column: int(28),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "dec",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5629,
																																FreeVars: ast.Identifiers{
																																	"dec",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(525),
																																		Column: int(29),
																																	},
																																	End: ast.Location{
																																		Line: int(525),
																																		Column: int(32),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "zp2",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5629,
																																FreeVars: ast.Identifiers{
																																	"zp2",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(525),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(525),
																																		Column: int(37),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															Value: "0",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5629,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(525),
																																		Column: int(39),
																																	},
																																	End: ast.Location{
																																		Line: int(525),
																																		Column: int(42),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5625,
																												FreeVars: ast.Identifiers{
																													"dec",
																													"pad_left",
																													"zp2",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(525),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(525),
																														Column: int(43),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										EqFodder: ast.Fodder{},
																										Variable: "dec2",
																										CloseFodder: ast.Fodder{},
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(525),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(525),
																												Column: int(43),
																											},
																										},
																									},
																								},
																								Body: &ast.Binary{
																									Right: &ast.Var{
																										Id: "dec2",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5637,
																											FreeVars: ast.Identifiers{
																												"dec2",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(526),
																													Column: int(80),
																												},
																												End: ast.Location{
																													Line: int(526),
																													Column: int(84),
																												},
																											},
																										},
																									},
																									Left: &ast.Conditional{
																										Cond: &ast.Var{
																											Id: "neg",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5637,
																												FreeVars: ast.Identifiers{
																													"neg",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(526),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(526),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										BranchTrue: &ast.LiteralString{
																											Value: "-",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5637,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(526),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(526),
																														Column: int(23),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										BranchFalse: &ast.Conditional{
																											Cond: &ast.Var{
																												Id: "plus",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5637,
																													FreeVars: ast.Identifiers{
																														"plus",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(526),
																															Column: int(32),
																														},
																														End: ast.Location{
																															Line: int(526),
																															Column: int(36),
																														},
																													},
																												},
																											},
																											BranchTrue: &ast.LiteralString{
																												Value: "+",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5637,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(526),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(526),
																															Column: int(45),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											BranchFalse: &ast.Conditional{
																												Cond: &ast.Var{
																													Id: "blank",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5637,
																														FreeVars: ast.Identifiers{
																															"blank",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(526),
																																Column: int(54),
																															},
																															End: ast.Location{
																																Line: int(526),
																																Column: int(59),
																															},
																														},
																													},
																												},
																												BranchTrue: &ast.LiteralString{
																													Value: " ",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5637,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(526),
																																Column: int(65),
																															},
																															End: ast.Location{
																																Line: int(526),
																																Column: int(68),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												BranchFalse: &ast.LiteralString{
																													Value: "",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5637,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(526),
																																Column: int(74),
																															},
																															End: ast.Location{
																																Line: int(526),
																																Column: int(76),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5637,
																													FreeVars: ast.Identifiers{
																														"blank",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(526),
																															Column: int(51),
																														},
																														End: ast.Location{
																															Line: int(526),
																															Column: int(76),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5637,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"plus",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(526),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(526),
																														Column: int(76),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5637,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"neg",
																												"plus",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(526),
																													Column: int(8),
																												},
																												End: ast.Location{
																													Line: int(526),
																													Column: int(76),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5637,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"dec2",
																											"neg",
																											"plus",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(526),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(526),
																												Column: int(84),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									Ctx: p5637,
																									FreeVars: ast.Identifiers{
																										"blank",
																										"dec",
																										"neg",
																										"pad_left",
																										"plus",
																										"zp2",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(525),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(526),
																											Column: int(84),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								Ctx: p5637,
																								FreeVars: ast.Identifiers{
																									"blank",
																									"dec",
																									"min_digits",
																									"neg",
																									"pad_left",
																									"plus",
																									"std",
																									"zp",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(524),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(526),
																										Column: int(84),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																								},
																							},
																							Ctx: p5637,
																							FreeVars: ast.Identifiers{
																								"blank",
																								"dec",
																								"min_chars",
																								"min_digits",
																								"neg",
																								"pad_left",
																								"plus",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(523),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(526),
																									Column: int(84),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																							ast.FodderElement{
																								Comment: []string{
																									"// dec is the minimal string needed to represent the number as text.",
																								},
																								Kind: ast.FodderKind(2),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						Ctx: p5637,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"blank",
																							"mag",
																							"min_chars",
																							"min_digits",
																							"neg",
																							"pad_left",
																							"plus",
																							"radix",
																							"std",
																							"zero_prefix",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(513),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(526),
																								Column: int(84),
																							},
																						},
																					},
																				},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "neg",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(25),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "mag",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(30),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "min_chars",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(41),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "min_digits",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(53),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "blank",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(60),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "plus",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(62),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(66),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "radix",
																						CommaFodder: ast.Fodder{},
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(68),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(73),
																							},
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "zero_prefix",
																						CommaFodder: nil,
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(511),
																								Column: int(75),
																							},
																							End: ast.Location{
																								Line: int(511),
																								Column: int(86),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: p5666,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"pad_left",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(511),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(526),
																							Column: int(84),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			EqFodder: nil,
																			Variable: "render_int",
																			CloseFodder: nil,
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: nil,
																				Body: &ast.Function{
																					ParenLeftFodder: ast.Fodder{},
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Body: &ast.Binary{
																									Right: &ast.Conditional{
																										Cond: &ast.Var{
																											Id: "capitals",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5676,
																												FreeVars: ast.Identifiers{
																													"capitals",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(531),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(531),
																														Column: int(37),
																													},
																												},
																											},
																										},
																										BranchTrue: &ast.Array{
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "A",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(44),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(47),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "B",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(49),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(52),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "C",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(54),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(57),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "D",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(59),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(62),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "E",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(64),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(67),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "F",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5681,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(531),
																																	Column: int(69),
																																},
																																End: ast.Location{
																																	Line: int(531),
																																	Column: int(72),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: nil,
																												},
																											},
																											CloseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5676,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(531),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(531),
																														Column: int(73),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										BranchFalse: &ast.Array{
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "a",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(30),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(33),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "b",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(38),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "c",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(40),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(43),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "d",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(45),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(48),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "e",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(50),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(53),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "f",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5690,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(55),
																																},
																																End: ast.Location{
																																	Line: int(532),
																																	Column: int(58),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: nil,
																												},
																											},
																											CloseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5676,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(532),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(532),
																														Column: int(59),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(23),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5676,
																											FreeVars: ast.Identifiers{
																												"capitals",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(531),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(532),
																													Column: int(59),
																												},
																											},
																										},
																									},
																									Left: &ast.Array{
																										Elements: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "0",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(25),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(26),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "1",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(28),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(29),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "2",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(31),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(32),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "3",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(35),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "4",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(37),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(38),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "5",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(40),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(41),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "6",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(43),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(44),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "7",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(46),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(47),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "8",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(49),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(50),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													OriginalString: "9",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5701,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(530),
																																Column: int(52),
																															},
																															End: ast.Location{
																																Line: int(530),
																																Column: int(53),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										CloseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5676,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(530),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(530),
																													Column: int(54),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									OpFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(23),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p5676,
																										FreeVars: ast.Identifiers{
																											"capitals",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(530),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(532),
																												Column: int(59),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								EqFodder: ast.Fodder{},
																								Variable: "numerals",
																								CloseFodder: ast.Fodder{},
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(530),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(532),
																										Column: int(59),
																									},
																								},
																							},
																						},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(533),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(533),
																															Column: int(21),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "abs",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p5720,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(533),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(533),
																														Column: int(25),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "n__",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5724,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(533),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(533),
																																	Column: int(29),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p5720,
																											FreeVars: ast.Identifiers{
																												"n__",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(533),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(533),
																													Column: int(30),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									EqFodder: ast.Fodder{},
																									Variable: "n_",
																									CloseFodder: ast.Fodder{},
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(533),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(533),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: nil,
																										Body: &ast.Function{
																											ParenLeftFodder: ast.Fodder{},
																											ParenRightFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5733,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(535),
																																	Column: int(17),
																																},
																																End: ast.Location{
																																	Line: int(535),
																																	Column: int(18),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "n",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5733,
																															FreeVars: ast.Identifiers{
																																"n",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(535),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(535),
																																	Column: int(13),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5733,
																														FreeVars: ast.Identifiers{
																															"n",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(535),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(535),
																																Column: int(18),
																															},
																														},
																													},
																													Op: ast.BinaryOp(12),
																												},
																												BranchTrue: &ast.LiteralString{
																													Value: "",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																															},
																														},
																														Ctx: p5733,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(536),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(536),
																																Column: int(13),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												BranchFalse: &ast.Binary{
																													Right: &ast.Index{
																														Target: &ast.Var{
																															Id: "numerals",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5733,
																																FreeVars: ast.Identifiers{
																																	"numerals",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(538),
																																		Column: int(36),
																																	},
																																	End: ast.Location{
																																		Line: int(538),
																																		Column: int(44),
																																	},
																																},
																															},
																														},
																														Index: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "$std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "mod",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: nil,
																																LeftBracketFodder: nil,
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "n",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5733,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(538),
																																						Column: int(45),
																																					},
																																					End: ast.Location{
																																						Line: int(538),
																																						Column: int(46),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralNumber{
																																			OriginalString: "16",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5733,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(538),
																																						Column: int(49),
																																					},
																																					End: ast.Location{
																																						Line: int(538),
																																						Column: int(51),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: nil,
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																	"n",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(538),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(538),
																																		Column: int(51),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5733,
																															FreeVars: ast.Identifiers{
																																"$std",
																																"n",
																																"numerals",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(538),
																																	Column: int(36),
																																},
																																End: ast.Location{
																																	Line: int(538),
																																	Column: int(52),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Var{
																															Id: "aux",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																	},
																																},
																																Ctx: p5733,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(538),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(538),
																																		Column: int(14),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(538),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(538),
																																							Column: int(18),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "floor",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5765,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(538),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(538),
																																						Column: int(24),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Binary{
																																						Right: &ast.LiteralNumber{
																																							OriginalString: "16",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p5770,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(538),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(538),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																						},
																																						Left: &ast.Var{
																																							Id: "n",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p5770,
																																								FreeVars: ast.Identifiers{
																																									"n",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(538),
																																										Column: int(25),
																																									},
																																									End: ast.Location{
																																										Line: int(538),
																																										Column: int(26),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p5770,
																																							FreeVars: ast.Identifiers{
																																								"n",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(538),
																																									Column: int(25),
																																								},
																																								End: ast.Location{
																																									Line: int(538),
																																									Column: int(31),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(1),
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5765,
																																			FreeVars: ast.Identifiers{
																																				"n",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(538),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(538),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5733,
																															FreeVars: ast.Identifiers{
																																"aux",
																																"n",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(538),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(538),
																																	Column: int(33),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5733,
																														FreeVars: ast.Identifiers{
																															"$std",
																															"aux",
																															"n",
																															"numerals",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(538),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(538),
																																Column: int(52),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													Ctx: p5733,
																													FreeVars: ast.Identifiers{
																														"$std",
																														"aux",
																														"n",
																														"numerals",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(535),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(538),
																															Column: int(52),
																														},
																													},
																												},
																											},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "n",
																													CommaFodder: nil,
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(534),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(534),
																															Column: int(18),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: p5781,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"aux",
																													"numerals",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(534),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(538),
																														Column: int(52),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										EqFodder: nil,
																										Variable: "aux",
																										CloseFodder: nil,
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																								},
																								Body: &ast.Local{
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5788,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(539),
																																	Column: int(39),
																																},
																																End: ast.Location{
																																	Line: int(539),
																																	Column: int(40),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(539),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(539),
																																			Column: int(25),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "floor",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5788,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(539),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(539),
																																		Column: int(31),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "n_",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5797,
																																			FreeVars: ast.Identifiers{
																																				"n_",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(539),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(539),
																																					Column: int(34),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5788,
																															FreeVars: ast.Identifiers{
																																"n_",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(539),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(539),
																																	Column: int(35),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5788,
																														FreeVars: ast.Identifiers{
																															"n_",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(539),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(539),
																																Column: int(40),
																															},
																														},
																													},
																													Op: ast.BinaryOp(12),
																												},
																												BranchTrue: &ast.LiteralString{
																													Value: "0",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5788,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(539),
																																Column: int(46),
																															},
																															End: ast.Location{
																																Line: int(539),
																																Column: int(49),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												BranchFalse: &ast.Apply{
																													Target: &ast.Var{
																														Id: "aux",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5788,
																															FreeVars: ast.Identifiers{
																																"aux",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(539),
																																	Column: int(55),
																																},
																																End: ast.Location{
																																	Line: int(539),
																																	Column: int(58),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Apply{
																																	Target: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "std",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(539),
																																						Column: int(59),
																																					},
																																					End: ast.Location{
																																						Line: int(539),
																																						Column: int(62),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "floor",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5811,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(539),
																																					Column: int(59),
																																				},
																																				End: ast.Location{
																																					Line: int(539),
																																					Column: int(68),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "n_",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p5815,
																																						FreeVars: ast.Identifiers{
																																							"n_",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(539),
																																								Column: int(69),
																																							},
																																							End: ast.Location{
																																								Line: int(539),
																																								Column: int(71),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5811,
																																		FreeVars: ast.Identifiers{
																																			"n_",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(539),
																																				Column: int(59),
																																			},
																																			End: ast.Location{
																																				Line: int(539),
																																				Column: int(72),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5788,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"n_",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(539),
																																Column: int(55),
																															},
																															End: ast.Location{
																																Line: int(539),
																																Column: int(73),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5788,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"n_",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(539),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(539),
																															Column: int(73),
																														},
																													},
																												},
																											},
																											EqFodder: ast.Fodder{},
																											Variable: "hex",
																											CloseFodder: ast.Fodder{},
																											Fun: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(539),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(539),
																													Column: int(73),
																												},
																											},
																										},
																									},
																									Body: &ast.Local{
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Body: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5824,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(540),
																																	Column: int(25),
																																},
																																End: ast.Location{
																																	Line: int(540),
																																	Column: int(26),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "n__",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5824,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(540),
																																	Column: int(19),
																																},
																																End: ast.Location{
																																	Line: int(540),
																																	Column: int(22),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5824,
																														FreeVars: ast.Identifiers{
																															"n__",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(540),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(540),
																																Column: int(26),
																															},
																														},
																													},
																													Op: ast.BinaryOp(9),
																												},
																												EqFodder: ast.Fodder{},
																												Variable: "neg",
																												CloseFodder: ast.Fodder{},
																												Fun: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(540),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(540),
																														Column: int(26),
																													},
																												},
																											},
																										},
																										Body: &ast.Local{
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Body: &ast.Binary{
																														Right: &ast.Conditional{
																															Cond: &ast.Var{
																																Id: "add_zerox",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5833,
																																	FreeVars: ast.Identifiers{
																																		"add_zerox",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(542),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(542),
																																			Column: int(33),
																																		},
																																	},
																																},
																															},
																															BranchTrue: &ast.LiteralNumber{
																																OriginalString: "2",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5833,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(542),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(542),
																																			Column: int(40),
																																		},
																																	},
																																},
																															},
																															BranchFalse: &ast.LiteralNumber{
																																OriginalString: "0",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5833,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(542),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(542),
																																			Column: int(47),
																																		},
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															ElseFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5833,
																																FreeVars: ast.Identifiers{
																																	"add_zerox",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(542),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(542),
																																		Column: int(47),
																																	},
																																},
																															},
																														},
																														Left: &ast.Binary{
																															Right: &ast.Conditional{
																																Cond: &ast.Binary{
																																	Right: &ast.Var{
																																		Id: "plus",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5833,
																																			FreeVars: ast.Identifiers{
																																				"plus",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(541),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(541),
																																					Column: int(54),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Binary{
																																		Right: &ast.Var{
																																			Id: "blank",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5833,
																																				FreeVars: ast.Identifiers{
																																					"blank",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(541),
																																						Column: int(41),
																																					},
																																					End: ast.Location{
																																						Line: int(541),
																																						Column: int(46),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "neg",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5833,
																																				FreeVars: ast.Identifiers{
																																					"neg",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(541),
																																						Column: int(34),
																																					},
																																					End: ast.Location{
																																						Line: int(541),
																																						Column: int(37),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5833,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																				"neg",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(541),
																																					Column: int(34),
																																				},
																																				End: ast.Location{
																																					Line: int(541),
																																					Column: int(46),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(18),
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5833,
																																		FreeVars: ast.Identifiers{
																																			"blank",
																																			"neg",
																																			"plus",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(541),
																																				Column: int(34),
																																			},
																																			End: ast.Location{
																																				Line: int(541),
																																				Column: int(54),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(18),
																																},
																																BranchTrue: &ast.LiteralNumber{
																																	OriginalString: "1",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5833,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(541),
																																				Column: int(60),
																																			},
																																			End: ast.Location{
																																				Line: int(541),
																																				Column: int(61),
																																			},
																																		},
																																	},
																																},
																																BranchFalse: &ast.LiteralNumber{
																																	OriginalString: "0",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5833,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(541),
																																				Column: int(67),
																																			},
																																			End: ast.Location{
																																				Line: int(541),
																																				Column: int(68),
																																			},
																																		},
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																ElseFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5833,
																																	FreeVars: ast.Identifiers{
																																		"blank",
																																		"neg",
																																		"plus",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(541),
																																			Column: int(31),
																																		},
																																		End: ast.Location{
																																			Line: int(541),
																																			Column: int(68),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Var{
																																Id: "min_chars",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5833,
																																	FreeVars: ast.Identifiers{
																																		"min_chars",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(541),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(541),
																																			Column: int(27),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5833,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"min_chars",
																																	"neg",
																																	"plus",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(541),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(541),
																																		Column: int(69),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(4),
																														},
																														OpFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(17),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5833,
																															FreeVars: ast.Identifiers{
																																"add_zerox",
																																"blank",
																																"min_chars",
																																"neg",
																																"plus",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(541),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(542),
																																	Column: int(48),
																																},
																															},
																														},
																														Op: ast.BinaryOp(4),
																													},
																													EqFodder: ast.Fodder{},
																													Variable: "zp",
																													CloseFodder: ast.Fodder{},
																													Fun: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(541),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(542),
																															Column: int(48),
																														},
																													},
																												},
																											},
																											Body: &ast.Local{
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Body: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(543),
																																				Column: int(19),
																																			},
																																			End: ast.Location{
																																				Line: int(543),
																																				Column: int(22),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "max",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5865,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(543),
																																			Column: int(19),
																																		},
																																		End: ast.Location{
																																			Line: int(543),
																																			Column: int(26),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "zp",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5869,
																																				FreeVars: ast.Identifiers{
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(543),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(543),
																																						Column: int(29),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "min_digits",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5869,
																																				FreeVars: ast.Identifiers{
																																					"min_digits",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(543),
																																						Column: int(31),
																																					},
																																					End: ast.Location{
																																						Line: int(543),
																																						Column: int(41),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5865,
																																FreeVars: ast.Identifiers{
																																	"min_digits",
																																	"std",
																																	"zp",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(543),
																																		Column: int(19),
																																	},
																																	End: ast.Location{
																																		Line: int(543),
																																		Column: int(42),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														EqFodder: ast.Fodder{},
																														Variable: "zp2",
																														CloseFodder: ast.Fodder{},
																														Fun: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(543),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(543),
																																Column: int(42),
																															},
																														},
																													},
																												},
																												Body: &ast.Local{
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Body: &ast.Binary{
																																Right: &ast.Apply{
																																	Target: &ast.Var{
																																		Id: "pad_left",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5879,
																																			FreeVars: ast.Identifiers{
																																				"pad_left",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(545),
																																					Column: int(22),
																																				},
																																				End: ast.Location{
																																					Line: int(545),
																																					Column: int(30),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "hex",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p5883,
																																						FreeVars: ast.Identifiers{
																																							"hex",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(545),
																																								Column: int(31),
																																							},
																																							End: ast.Location{
																																								Line: int(545),
																																								Column: int(34),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "zp2",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p5883,
																																						FreeVars: ast.Identifiers{
																																							"zp2",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(545),
																																								Column: int(36),
																																							},
																																							End: ast.Location{
																																								Line: int(545),
																																								Column: int(39),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralString{
																																					Value: "0",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p5883,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(545),
																																								Column: int(41),
																																							},
																																							End: ast.Location{
																																								Line: int(545),
																																								Column: int(44),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5879,
																																		FreeVars: ast.Identifiers{
																																			"hex",
																																			"pad_left",
																																			"zp2",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(545),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(545),
																																				Column: int(45),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																Left: &ast.Conditional{
																																	Cond: &ast.Var{
																																		Id: "add_zerox",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5879,
																																			FreeVars: ast.Identifiers{
																																				"add_zerox",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(544),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(544),
																																					Column: int(33),
																																				},
																																			},
																																		},
																																	},
																																	BranchTrue: &ast.Conditional{
																																		Cond: &ast.Var{
																																			Id: "capitals",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5879,
																																				FreeVars: ast.Identifiers{
																																					"capitals",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(544),
																																						Column: int(43),
																																					},
																																					End: ast.Location{
																																						Line: int(544),
																																						Column: int(51),
																																					},
																																				},
																																			},
																																		},
																																		BranchTrue: &ast.LiteralString{
																																			Value: "0X",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5879,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(544),
																																						Column: int(57),
																																					},
																																					End: ast.Location{
																																						Line: int(544),
																																						Column: int(61),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		BranchFalse: &ast.LiteralString{
																																			Value: "0x",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p5879,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(544),
																																						Column: int(67),
																																					},
																																					End: ast.Location{
																																						Line: int(544),
																																						Column: int(71),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5879,
																																			FreeVars: ast.Identifiers{
																																				"capitals",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(544),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(544),
																																					Column: int(71),
																																				},
																																			},
																																		},
																																	},
																																	BranchFalse: &ast.LiteralString{
																																		Value: "",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5879,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(544),
																																					Column: int(78),
																																				},
																																				End: ast.Location{
																																					Line: int(544),
																																					Column: int(80),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5879,
																																		FreeVars: ast.Identifiers{
																																			"add_zerox",
																																			"capitals",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(544),
																																				Column: int(21),
																																			},
																																			End: ast.Location{
																																				Line: int(544),
																																				Column: int(80),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(19),
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5879,
																																	FreeVars: ast.Identifiers{
																																		"add_zerox",
																																		"capitals",
																																		"hex",
																																		"pad_left",
																																		"zp2",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(544),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(545),
																																			Column: int(45),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															EqFodder: ast.Fodder{},
																															Variable: "hex2",
																															CloseFodder: ast.Fodder{},
																															Fun: nil,
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(544),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(545),
																																	Column: int(45),
																																},
																															},
																														},
																													},
																													Body: &ast.Binary{
																														Right: &ast.Var{
																															Id: "hex2",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5904,
																																FreeVars: ast.Identifiers{
																																	"hex2",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(546),
																																		Column: int(80),
																																	},
																																	End: ast.Location{
																																		Line: int(546),
																																		Column: int(84),
																																	},
																																},
																															},
																														},
																														Left: &ast.Conditional{
																															Cond: &ast.Var{
																																Id: "neg",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5904,
																																	FreeVars: ast.Identifiers{
																																		"neg",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(546),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(546),
																																			Column: int(14),
																																		},
																																	},
																																},
																															},
																															BranchTrue: &ast.LiteralString{
																																Value: "-",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5904,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(546),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(546),
																																			Column: int(23),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															BranchFalse: &ast.Conditional{
																																Cond: &ast.Var{
																																	Id: "plus",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5904,
																																		FreeVars: ast.Identifiers{
																																			"plus",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(546),
																																				Column: int(32),
																																			},
																																			End: ast.Location{
																																				Line: int(546),
																																				Column: int(36),
																																			},
																																		},
																																	},
																																},
																																BranchTrue: &ast.LiteralString{
																																	Value: "+",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5904,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(546),
																																				Column: int(42),
																																			},
																																			End: ast.Location{
																																				Line: int(546),
																																				Column: int(45),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																BranchFalse: &ast.Conditional{
																																	Cond: &ast.Var{
																																		Id: "blank",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5904,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(546),
																																					Column: int(54),
																																				},
																																				End: ast.Location{
																																					Line: int(546),
																																					Column: int(59),
																																				},
																																			},
																																		},
																																	},
																																	BranchTrue: &ast.LiteralString{
																																		Value: " ",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5904,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(546),
																																					Column: int(65),
																																				},
																																				End: ast.Location{
																																					Line: int(546),
																																					Column: int(68),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	BranchFalse: &ast.LiteralString{
																																		Value: "",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5904,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(546),
																																					Column: int(74),
																																				},
																																				End: ast.Location{
																																					Line: int(546),
																																					Column: int(76),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5904,
																																		FreeVars: ast.Identifiers{
																																			"blank",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(546),
																																				Column: int(51),
																																			},
																																			End: ast.Location{
																																				Line: int(546),
																																				Column: int(76),
																																			},
																																		},
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																ElseFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p5904,
																																	FreeVars: ast.Identifiers{
																																		"blank",
																																		"plus",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(546),
																																			Column: int(29),
																																		},
																																		End: ast.Location{
																																			Line: int(546),
																																			Column: int(76),
																																		},
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															ElseFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5904,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"neg",
																																	"plus",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(546),
																																		Column: int(8),
																																	},
																																	End: ast.Location{
																																		Line: int(546),
																																		Column: int(76),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5904,
																															FreeVars: ast.Identifiers{
																																"blank",
																																"hex2",
																																"neg",
																																"plus",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(546),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(546),
																																	Column: int(84),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														Ctx: p5904,
																														FreeVars: ast.Identifiers{
																															"add_zerox",
																															"blank",
																															"capitals",
																															"hex",
																															"neg",
																															"pad_left",
																															"plus",
																															"zp2",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(544),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(546),
																																Column: int(84),
																															},
																														},
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p5904,
																													FreeVars: ast.Identifiers{
																														"add_zerox",
																														"blank",
																														"capitals",
																														"hex",
																														"min_digits",
																														"neg",
																														"pad_left",
																														"plus",
																														"std",
																														"zp",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(543),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(546),
																															Column: int(84),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p5904,
																												FreeVars: ast.Identifiers{
																													"add_zerox",
																													"blank",
																													"capitals",
																													"hex",
																													"min_chars",
																													"min_digits",
																													"neg",
																													"pad_left",
																													"plus",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(541),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(546),
																														Column: int(84),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											Ctx: p5904,
																											FreeVars: ast.Identifiers{
																												"add_zerox",
																												"blank",
																												"capitals",
																												"hex",
																												"min_chars",
																												"min_digits",
																												"n__",
																												"pad_left",
																												"plus",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(540),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(546),
																													Column: int(84),
																												},
																											},
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										Ctx: p5904,
																										FreeVars: ast.Identifiers{
																											"add_zerox",
																											"aux",
																											"blank",
																											"capitals",
																											"min_chars",
																											"min_digits",
																											"n_",
																											"n__",
																											"pad_left",
																											"plus",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(539),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(546),
																												Column: int(84),
																											},
																										},
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									Ctx: p5904,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"add_zerox",
																										"blank",
																										"capitals",
																										"min_chars",
																										"min_digits",
																										"n_",
																										"n__",
																										"numerals",
																										"pad_left",
																										"plus",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(534),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(546),
																											Column: int(84),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								Ctx: p5904,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"add_zerox",
																									"blank",
																									"capitals",
																									"min_chars",
																									"min_digits",
																									"n__",
																									"numerals",
																									"pad_left",
																									"plus",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(533),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(546),
																										Column: int(84),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																								},
																							},
																							Ctx: p5904,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"add_zerox",
																								"blank",
																								"capitals",
																								"min_chars",
																								"min_digits",
																								"n__",
																								"pad_left",
																								"plus",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(530),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(546),
																									Column: int(84),
																								},
																							},
																						},
																					},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "n__",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(25),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "min_chars",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(36),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "min_digits",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(48),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "blank",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(55),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "plus",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(61),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "add_zerox",
																							CommaFodder: ast.Fodder{},
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(63),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(72),
																								},
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "capitals",
																							CommaFodder: nil,
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(529),
																									Column: int(74),
																								},
																								End: ast.Location{
																									Line: int(529),
																									Column: int(82),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: p5940,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"pad_left",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(529),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(546),
																								Column: int(84),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				EqFodder: nil,
																				Variable: "render_hex",
																				CloseFodder: nil,
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Body: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: nil,
																					Body: &ast.Function{
																						ParenLeftFodder: ast.Fodder{},
																						ParenRightFodder: ast.Fodder{},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: nil,
																									Body: &ast.Function{
																										ParenLeftFodder: ast.Fodder{},
																										ParenRightFodder: ast.Fodder{},
																										Body: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "0",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5951,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(550),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(550),
																																Column: int(17),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "i",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5951,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(550),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(550),
																																Column: int(13),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p5951,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(550),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(550),
																															Column: int(17),
																														},
																													},
																												},
																												Op: ast.BinaryOp(9),
																											},
																											BranchTrue: &ast.LiteralString{
																												Value: "",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(10),
																														},
																													},
																													Ctx: p5951,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(551),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(551),
																															Column: int(13),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											BranchFalse: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.LiteralString{
																														Value: "0",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5951,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(553),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(553),
																																	Column: int(27),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Left: &ast.Index{
																														Target: &ast.Var{
																															Id: "str",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5951,
																																FreeVars: ast.Identifiers{
																																	"str",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(553),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(553),
																																		Column: int(17),
																																	},
																																},
																															},
																														},
																														Index: &ast.Var{
																															Id: "i",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p5951,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(553),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(553),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5951,
																															FreeVars: ast.Identifiers{
																																"i",
																																"str",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(553),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(553),
																																	Column: int(20),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5951,
																														FreeVars: ast.Identifiers{
																															"i",
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(553),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(553),
																																Column: int(27),
																															},
																														},
																													},
																													Op: ast.BinaryOp(12),
																												},
																												BranchTrue: &ast.Apply{
																													Target: &ast.Var{
																														Id: "aux",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(12),
																																},
																															},
																															Ctx: p5951,
																															FreeVars: ast.Identifiers{
																																"aux",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(554),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(554),
																																	Column: int(16),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "str",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5973,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(554),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(554),
																																				Column: int(20),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	Right: &ast.LiteralNumber{
																																		OriginalString: "1",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5973,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(554),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(554),
																																					Column: int(27),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Var{
																																		Id: "i",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5973,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(554),
																																					Column: int(22),
																																				},
																																				End: ast.Location{
																																					Line: int(554),
																																					Column: int(23),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5973,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(554),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(554),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(4),
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5951,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"i",
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(554),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(554),
																																Column: int(28),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												BranchFalse: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(12),
																																	},
																																},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(556),
																																		Column: int(13),
																																	},
																																	End: ast.Location{
																																		Line: int(556),
																																		Column: int(16),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "substr",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p5951,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(556),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(556),
																																	Column: int(23),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "str",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5990,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(556),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(556),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.LiteralNumber{
																																	OriginalString: "0",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5990,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(556),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(556),
																																				Column: int(30),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	Right: &ast.LiteralNumber{
																																		OriginalString: "1",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5990,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(556),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(556),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Var{
																																		Id: "i",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p5990,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(556),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(556),
																																					Column: int(33),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p5990,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(556),
																																				Column: int(32),
																																			},
																																			End: ast.Location{
																																				Line: int(556),
																																				Column: int(37),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(3),
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p5951,
																														FreeVars: ast.Identifiers{
																															"i",
																															"std",
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(556),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(556),
																																Column: int(38),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(10),
																														},
																													},
																													Ctx: p5951,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"i",
																														"std",
																														"str",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(553),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(556),
																															Column: int(38),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(8),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																													},
																												},
																												Ctx: p5951,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"i",
																													"std",
																													"str",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(550),
																														Column: int(9),
																													},
																													End: ast.Location{
																														Line: int(556),
																														Column: int(38),
																													},
																												},
																											},
																										},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "str",
																												CommaFodder: ast.Fodder{},
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(549),
																														Column: int(17),
																													},
																													End: ast.Location{
																														Line: int(549),
																														Column: int(20),
																													},
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "i",
																												CommaFodder: nil,
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(549),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(549),
																														Column: int(23),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: p6006,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(549),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(556),
																													Column: int(38),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									EqFodder: nil,
																									Variable: "aux",
																									CloseFodder: nil,
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							Body: &ast.Apply{
																								Target: &ast.Var{
																									Id: "aux",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										Ctx: p6011,
																										FreeVars: ast.Identifiers{
																											"aux",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(557),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(557),
																												Column: int(10),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "str",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p6015,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(557),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(557),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "1",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6015,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(557),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(557),
																																Column: int(35),
																															},
																														},
																													},
																												},
																												Left: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(557),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(557),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "length",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6015,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(557),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(557),
																																	Column: int(26),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "str",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6027,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(557),
																																				Column: int(27),
																																			},
																																			End: ast.Location{
																																				Line: int(557),
																																				Column: int(30),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6015,
																														FreeVars: ast.Identifiers{
																															"std",
																															"str",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(557),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(557),
																																Column: int(31),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p6015,
																													FreeVars: ast.Identifiers{
																														"std",
																														"str",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(557),
																															Column: int(16),
																														},
																														End: ast.Location{
																															Line: int(557),
																															Column: int(35),
																														},
																													},
																												},
																												Op: ast.BinaryOp(4),
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p6011,
																									FreeVars: ast.Identifiers{
																										"aux",
																										"std",
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(557),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(557),
																											Column: int(36),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								Ctx: p6011,
																								FreeVars: ast.Identifiers{
																									"std",
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(549),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(557),
																										Column: int(36),
																									},
																								},
																							},
																						},
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: ast.Fodder{},
																								Name: "str",
																								CommaFodder: nil,
																								EqFodder: nil,
																								DefaultArg: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(548),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(548),
																										Column: int(34),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: p6035,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(548),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(557),
																									Column: int(36),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					EqFodder: nil,
																					Variable: "strip_trailing_zero",
																					CloseFodder: nil,
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Body: &ast.Local{
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: nil,
																						Body: &ast.Function{
																							ParenLeftFodder: ast.Fodder{},
																							ParenRightFodder: ast.Fodder{},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "std",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(561),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(561),
																																Column: int(21),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "abs",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p6047,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(561),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(561),
																															Column: int(25),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "n__",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6051,
																																FreeVars: ast.Identifiers{
																																	"n__",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(561),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(561),
																																		Column: int(29),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p6047,
																												FreeVars: ast.Identifiers{
																													"n__",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(561),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(561),
																														Column: int(30),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										EqFodder: ast.Fodder{},
																										Variable: "n_",
																										CloseFodder: ast.Fodder{},
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(561),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(561),
																												Column: int(30),
																											},
																										},
																									},
																								},
																								Body: &ast.Local{
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Body: &ast.Apply{
																												Target: &ast.Index{
																													Target: &ast.Var{
																														Id: "std",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(562),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(562),
																																	Column: int(24),
																																},
																															},
																														},
																													},
																													Index: &ast.LiteralString{
																														Value: "floor",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													RightBracketFodder: ast.Fodder{},
																													LeftBracketFodder: ast.Fodder{},
																													Id: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6061,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(562),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(562),
																																Column: int(30),
																															},
																														},
																													},
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "n_",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6065,
																																	FreeVars: ast.Identifiers{
																																		"n_",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(562),
																																			Column: int(31),
																																		},
																																		End: ast.Location{
																																			Line: int(562),
																																			Column: int(33),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p6061,
																													FreeVars: ast.Identifiers{
																														"n_",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(562),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(562),
																															Column: int(34),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											EqFodder: ast.Fodder{},
																											Variable: "whole",
																											CloseFodder: ast.Fodder{},
																											Fun: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(562),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(562),
																													Column: int(34),
																												},
																											},
																										},
																									},
																									Body: &ast.Local{
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Body: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(567),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(567),
																																		Column: int(30),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "pow",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6075,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(567),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(567),
																																	Column: int(34),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.LiteralNumber{
																																	OriginalString: "10",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6079,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(567),
																																				Column: int(35),
																																			},
																																			End: ast.Location{
																																				Line: int(567),
																																				Column: int(37),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "prec",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6079,
																																		FreeVars: ast.Identifiers{
																																			"prec",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(567),
																																				Column: int(39),
																																			},
																																			End: ast.Location{
																																				Line: int(567),
																																				Column: int(43),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6075,
																														FreeVars: ast.Identifiers{
																															"prec",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(567),
																																Column: int(27),
																															},
																															End: ast.Location{
																																Line: int(567),
																																Column: int(44),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												EqFodder: ast.Fodder{},
																												Variable: "denominator",
																												CloseFodder: ast.Fodder{},
																												Fun: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(567),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(567),
																														Column: int(44),
																													},
																												},
																											},
																										},
																										Body: &ast.Local{
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Body: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "0.5",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6087,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(568),
																																		Column: int(53),
																																	},
																																	End: ast.Location{
																																		Line: int(568),
																																		Column: int(56),
																																	},
																																},
																															},
																														},
																														Left: &ast.Binary{
																															Right: &ast.Var{
																																Id: "denominator",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6087,
																																	FreeVars: ast.Identifiers{
																																		"denominator",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(568),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(568),
																																			Column: int(50),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(25),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(28),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "abs",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6087,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(568),
																																				Column: int(25),
																																			},
																																			End: ast.Location{
																																				Line: int(568),
																																				Column: int(32),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "n_",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6099,
																																					FreeVars: ast.Identifiers{
																																						"n_",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(568),
																																							Column: int(33),
																																						},
																																						End: ast.Location{
																																							Line: int(568),
																																							Column: int(35),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6087,
																																	FreeVars: ast.Identifiers{
																																		"n_",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(568),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(568),
																																			Column: int(36),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6087,
																																FreeVars: ast.Identifiers{
																																	"denominator",
																																	"n_",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(568),
																																		Column: int(25),
																																	},
																																	End: ast.Location{
																																		Line: int(568),
																																		Column: int(50),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(0),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6087,
																															FreeVars: ast.Identifiers{
																																"denominator",
																																"n_",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(568),
																																	Column: int(25),
																																},
																																End: ast.Location{
																																	Line: int(568),
																																	Column: int(56),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													EqFodder: ast.Fodder{},
																													Variable: "numerator",
																													CloseFodder: ast.Fodder{},
																													Fun: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(568),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(568),
																															Column: int(56),
																														},
																													},
																												},
																											},
																											Body: &ast.Local{
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Body: &ast.Binary{
																															Right: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(569),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(569),
																																					Column: int(39),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "floor",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6112,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(569),
																																				Column: int(36),
																																			},
																																			End: ast.Location{
																																				Line: int(569),
																																				Column: int(45),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Binary{
																																				Right: &ast.Var{
																																					Id: "denominator",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6117,
																																						FreeVars: ast.Identifiers{
																																							"denominator",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(569),
																																								Column: int(58),
																																							},
																																							End: ast.Location{
																																								Line: int(569),
																																								Column: int(69),
																																							},
																																						},
																																					},
																																				},
																																				Left: &ast.Var{
																																					Id: "numerator",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6117,
																																						FreeVars: ast.Identifiers{
																																							"numerator",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(569),
																																								Column: int(46),
																																							},
																																							End: ast.Location{
																																								Line: int(569),
																																								Column: int(55),
																																							},
																																						},
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6117,
																																					FreeVars: ast.Identifiers{
																																						"denominator",
																																						"numerator",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(569),
																																							Column: int(46),
																																						},
																																						End: ast.Location{
																																							Line: int(569),
																																							Column: int(69),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(1),
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6112,
																																	FreeVars: ast.Identifiers{
																																		"denominator",
																																		"numerator",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(569),
																																			Column: int(36),
																																		},
																																		End: ast.Location{
																																			Line: int(569),
																																			Column: int(70),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Left: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(569),
																																					Column: int(21),
																																				},
																																				End: ast.Location{
																																					Line: int(569),
																																					Column: int(24),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "sign",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6112,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(569),
																																				Column: int(21),
																																			},
																																			End: ast.Location{
																																				Line: int(569),
																																				Column: int(29),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "n_",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6131,
																																					FreeVars: ast.Identifiers{
																																						"n_",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(569),
																																							Column: int(30),
																																						},
																																						End: ast.Location{
																																							Line: int(569),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6112,
																																	FreeVars: ast.Identifiers{
																																		"n_",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(569),
																																			Column: int(21),
																																		},
																																		End: ast.Location{
																																			Line: int(569),
																																			Column: int(33),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6112,
																																FreeVars: ast.Identifiers{
																																	"denominator",
																																	"n_",
																																	"numerator",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(569),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(569),
																																		Column: int(70),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(0),
																														},
																														EqFodder: ast.Fodder{},
																														Variable: "whole",
																														CloseFodder: ast.Fodder{},
																														Fun: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(569),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(569),
																																Column: int(70),
																															},
																														},
																													},
																												},
																												Body: &ast.Local{
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Body: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "$std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"$std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "mod",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: nil,
																																	LeftBracketFodder: nil,
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: nil,
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Apply{
																																				Target: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "std",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(570),
																																									Column: int(20),
																																								},
																																								End: ast.Location{
																																									Line: int(570),
																																									Column: int(23),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "floor",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6149,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(570),
																																								Column: int(20),
																																							},
																																							End: ast.Location{
																																								Line: int(570),
																																								Column: int(29),
																																							},
																																						},
																																					},
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "numerator",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6153,
																																									FreeVars: ast.Identifiers{
																																										"numerator",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(570),
																																											Column: int(30),
																																										},
																																										End: ast.Location{
																																											Line: int(570),
																																											Column: int(39),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6149,
																																					FreeVars: ast.Identifiers{
																																						"numerator",
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(570),
																																							Column: int(20),
																																						},
																																						End: ast.Location{
																																							Line: int(570),
																																							Column: int(40),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "denominator",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6149,
																																					FreeVars: ast.Identifiers{
																																						"denominator",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(570),
																																							Column: int(43),
																																						},
																																						End: ast.Location{
																																							Line: int(570),
																																							Column: int(54),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: nil,
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																		"denominator",
																																		"numerator",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(570),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(570),
																																			Column: int(54),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															EqFodder: ast.Fodder{},
																															Variable: "frac",
																															CloseFodder: ast.Fodder{},
																															Fun: nil,
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(570),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(570),
																																	Column: int(54),
																																},
																															},
																														},
																													},
																													Body: &ast.Local{
																														Binds: ast.LocalBinds{
																															ast.LocalBind{
																																VarFodder: ast.Fodder{},
																																Body: &ast.Conditional{
																																	Cond: &ast.Binary{
																																		Right: &ast.Unary{
																																			Expr: &ast.Var{
																																				Id: "ensure_pt",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6165,
																																					FreeVars: ast.Identifiers{
																																						"ensure_pt",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(571),
																																							Column: int(41),
																																						},
																																						End: ast.Location{
																																							Line: int(571),
																																							Column: int(50),
																																						},
																																					},
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6165,
																																				FreeVars: ast.Identifiers{
																																					"ensure_pt",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(571),
																																						Column: int(40),
																																					},
																																					End: ast.Location{
																																						Line: int(571),
																																						Column: int(50),
																																					},
																																				},
																																			},
																																			Op: ast.UnaryOp(0),
																																		},
																																		Left: &ast.Binary{
																																			Right: &ast.LiteralNumber{
																																				OriginalString: "0",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6165,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(571),
																																							Column: int(35),
																																						},
																																						End: ast.Location{
																																							Line: int(571),
																																							Column: int(36),
																																						},
																																					},
																																				},
																																			},
																																			Left: &ast.Var{
																																				Id: "prec",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6165,
																																					FreeVars: ast.Identifiers{
																																						"prec",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(571),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(571),
																																							Column: int(31),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6165,
																																				FreeVars: ast.Identifiers{
																																					"prec",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(571),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(571),
																																						Column: int(36),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(12),
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6165,
																																			FreeVars: ast.Identifiers{
																																				"ensure_pt",
																																				"prec",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(571),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(571),
																																					Column: int(50),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(17),
																																	},
																																	BranchTrue: &ast.LiteralNumber{
																																		OriginalString: "0",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6165,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(571),
																																					Column: int(56),
																																				},
																																				End: ast.Location{
																																					Line: int(571),
																																					Column: int(57),
																																				},
																																			},
																																		},
																																	},
																																	BranchFalse: &ast.LiteralNumber{
																																		OriginalString: "1",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6165,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(571),
																																					Column: int(63),
																																				},
																																				End: ast.Location{
																																					Line: int(571),
																																					Column: int(64),
																																				},
																																			},
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6165,
																																		FreeVars: ast.Identifiers{
																																			"ensure_pt",
																																			"prec",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(571),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(571),
																																				Column: int(64),
																																			},
																																		},
																																	},
																																},
																																EqFodder: ast.Fodder{},
																																Variable: "dot_size",
																																CloseFodder: ast.Fodder{},
																																Fun: nil,
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(571),
																																		Column: int(13),
																																	},
																																	End: ast.Location{
																																		Line: int(571),
																																		Column: int(64),
																																	},
																																},
																															},
																														},
																														Body: &ast.Local{
																															Binds: ast.LocalBinds{
																																ast.LocalBind{
																																	VarFodder: ast.Fodder{},
																																	Body: &ast.Binary{
																																		Right: &ast.Var{
																																			Id: "dot_size",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6181,
																																				FreeVars: ast.Identifiers{
																																					"dot_size",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(572),
																																						Column: int(36),
																																					},
																																					End: ast.Location{
																																						Line: int(572),
																																						Column: int(44),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Binary{
																																			Right: &ast.Var{
																																				Id: "prec",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6181,
																																					FreeVars: ast.Identifiers{
																																						"prec",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(572),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(572),
																																							Column: int(33),
																																						},
																																					},
																																				},
																																			},
																																			Left: &ast.Var{
																																				Id: "zero_pad",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6181,
																																					FreeVars: ast.Identifiers{
																																						"zero_pad",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(572),
																																							Column: int(18),
																																						},
																																						End: ast.Location{
																																							Line: int(572),
																																							Column: int(26),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6181,
																																				FreeVars: ast.Identifiers{
																																					"prec",
																																					"zero_pad",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(572),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(572),
																																						Column: int(33),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(4),
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6181,
																																			FreeVars: ast.Identifiers{
																																				"dot_size",
																																				"prec",
																																				"zero_pad",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(572),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(572),
																																					Column: int(44),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(4),
																																	},
																																	EqFodder: ast.Fodder{},
																																	Variable: "zp",
																																	CloseFodder: ast.Fodder{},
																																	Fun: nil,
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(572),
																																			Column: int(13),
																																		},
																																		End: ast.Location{
																																			Line: int(572),
																																			Column: int(44),
																																		},
																																	},
																																},
																															},
																															Body: &ast.Local{
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Body: &ast.Apply{
																																			Target: &ast.Var{
																																				Id: "render_int",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6194,
																																					FreeVars: ast.Identifiers{
																																						"render_int",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(573),
																																							Column: int(19),
																																						},
																																						End: ast.Location{
																																							Line: int(573),
																																							Column: int(29),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Binary{
																																							Right: &ast.LiteralNumber{
																																								OriginalString: "0",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6199,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(573),
																																											Column: int(36),
																																										},
																																										End: ast.Location{
																																											Line: int(573),
																																											Column: int(37),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.Var{
																																								Id: "n__",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6199,
																																									FreeVars: ast.Identifiers{
																																										"n__",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(573),
																																											Column: int(30),
																																										},
																																										End: ast.Location{
																																											Line: int(573),
																																											Column: int(33),
																																										},
																																									},
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{
																																									"n__",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(30),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(37),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(9),
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "whole",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{
																																									"whole",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(39),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(44),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "zp",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{
																																									"zp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(46),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(48),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							OriginalString: "0",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(50),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(51),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "blank",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{
																																									"blank",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(53),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(58),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "plus",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{
																																									"plus",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(60),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(64),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							OriginalString: "10",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(66),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(68),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralString{
																																							Value: "",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6199,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(573),
																																										Column: int(70),
																																									},
																																									End: ast.Location{
																																										Line: int(573),
																																										Column: int(72),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6194,
																																				FreeVars: ast.Identifiers{
																																					"blank",
																																					"n__",
																																					"plus",
																																					"render_int",
																																					"whole",
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(573),
																																						Column: int(19),
																																					},
																																					End: ast.Location{
																																						Line: int(573),
																																						Column: int(73),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		EqFodder: ast.Fodder{},
																																		Variable: "str",
																																		CloseFodder: ast.Fodder{},
																																		Fun: nil,
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(573),
																																				Column: int(13),
																																			},
																																			End: ast.Location{
																																				Line: int(573),
																																				Column: int(73),
																																			},
																																		},
																																	},
																																},
																																Body: &ast.Conditional{
																																	Cond: &ast.Binary{
																																		Right: &ast.LiteralNumber{
																																			OriginalString: "0",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(574),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(574),
																																						Column: int(19),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "prec",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"prec",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(574),
																																						Column: int(10),
																																					},
																																					End: ast.Location{
																																						Line: int(574),
																																						Column: int(14),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6218,
																																			FreeVars: ast.Identifiers{
																																				"prec",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(574),
																																					Column: int(10),
																																				},
																																				End: ast.Location{
																																					Line: int(574),
																																					Column: int(19),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(12),
																																	},
																																	BranchTrue: &ast.Binary{
																																		Right: &ast.Conditional{
																																			Cond: &ast.Var{
																																				Id: "ensure_pt",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{
																																						"ensure_pt",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(575),
																																							Column: int(18),
																																						},
																																						End: ast.Location{
																																							Line: int(575),
																																							Column: int(27),
																																						},
																																					},
																																				},
																																			},
																																			BranchTrue: &ast.LiteralString{
																																				Value: ".",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(575),
																																							Column: int(33),
																																						},
																																						End: ast.Location{
																																							Line: int(575),
																																							Column: int(36),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			BranchFalse: &ast.LiteralString{
																																				Value: "",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(575),
																																							Column: int(42),
																																						},
																																						End: ast.Location{
																																							Line: int(575),
																																							Column: int(44),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			ThenFodder: ast.Fodder{},
																																			ElseFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"ensure_pt",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(575),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(575),
																																						Column: int(44),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "str",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(8),
																																					},
																																				},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"str",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(575),
																																						Column: int(9),
																																					},
																																					End: ast.Location{
																																						Line: int(575),
																																						Column: int(12),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6218,
																																			FreeVars: ast.Identifiers{
																																				"ensure_pt",
																																				"str",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(575),
																																					Column: int(9),
																																				},
																																				End: ast.Location{
																																					Line: int(575),
																																					Column: int(44),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(3),
																																	},
																																	BranchFalse: &ast.Conditional{
																																		Cond: &ast.Binary{
																																			Right: &ast.Binary{
																																				Right: &ast.LiteralNumber{
																																					OriginalString: "0",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6218,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(577),
																																								Column: int(31),
																																							},
																																							End: ast.Location{
																																								Line: int(577),
																																								Column: int(32),
																																							},
																																						},
																																					},
																																				},
																																				Left: &ast.Var{
																																					Id: "frac",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6218,
																																						FreeVars: ast.Identifiers{
																																							"frac",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(577),
																																								Column: int(24),
																																							},
																																							End: ast.Location{
																																								Line: int(577),
																																								Column: int(28),
																																							},
																																						},
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{
																																						"frac",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(577),
																																							Column: int(24),
																																						},
																																						End: ast.Location{
																																							Line: int(577),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(7),
																																			},
																																			Left: &ast.Var{
																																				Id: "trailing",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{
																																						"trailing",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(577),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(577),
																																							Column: int(20),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"frac",
																																					"trailing",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(577),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(577),
																																						Column: int(32),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(18),
																																		},
																																		BranchTrue: &ast.Local{
																																			Binds: ast.LocalBinds{
																																				ast.LocalBind{
																																					VarFodder: ast.Fodder{},
																																					Body: &ast.Apply{
																																						Target: &ast.Var{
																																							Id: "render_int",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6247,
																																								FreeVars: ast.Identifiers{
																																									"render_int",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(578),
																																										Column: int(28),
																																									},
																																									End: ast.Location{
																																										Line: int(578),
																																										Column: int(38),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralBoolean{
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(39),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(44),
																																												},
																																											},
																																										},
																																										Value: false,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "frac",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{
																																												"frac",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(46),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(50),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "prec",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{
																																												"prec",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(52),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(56),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralNumber{
																																										OriginalString: "0",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(58),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(59),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralBoolean{
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(61),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(66),
																																												},
																																											},
																																										},
																																										Value: false,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralBoolean{
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(68),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(73),
																																												},
																																											},
																																										},
																																										Value: false,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralNumber{
																																										OriginalString: "10",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(75),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(77),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralString{
																																										Value: "",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6251,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(578),
																																													Column: int(79),
																																												},
																																												End: ast.Location{
																																													Line: int(578),
																																													Column: int(81),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6247,
																																							FreeVars: ast.Identifiers{
																																								"frac",
																																								"prec",
																																								"render_int",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(578),
																																									Column: int(28),
																																								},
																																								End: ast.Location{
																																									Line: int(578),
																																									Column: int(82),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					EqFodder: ast.Fodder{},
																																					Variable: "frac_str",
																																					CloseFodder: ast.Fodder{},
																																					Fun: nil,
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(578),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(578),
																																							Column: int(82),
																																						},
																																					},
																																				},
																																			},
																																			Body: &ast.Binary{
																																				Right: &ast.Conditional{
																																					Cond: &ast.Unary{
																																						Expr: &ast.Var{
																																							Id: "trailing",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6218,
																																								FreeVars: ast.Identifiers{
																																									"trailing",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(579),
																																										Column: int(27),
																																									},
																																									End: ast.Location{
																																										Line: int(579),
																																										Column: int(35),
																																									},
																																								},
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6218,
																																							FreeVars: ast.Identifiers{
																																								"trailing",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(579),
																																									Column: int(26),
																																								},
																																								End: ast.Location{
																																									Line: int(579),
																																									Column: int(35),
																																								},
																																							},
																																						},
																																						Op: ast.UnaryOp(0),
																																					},
																																					BranchTrue: &ast.Apply{
																																						Target: &ast.Var{
																																							Id: "strip_trailing_zero",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6218,
																																								FreeVars: ast.Identifiers{
																																									"strip_trailing_zero",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(579),
																																										Column: int(41),
																																									},
																																									End: ast.Location{
																																										Line: int(579),
																																										Column: int(60),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "frac_str",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6273,
																																											FreeVars: ast.Identifiers{
																																												"frac_str",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(579),
																																													Column: int(61),
																																												},
																																												End: ast.Location{
																																													Line: int(579),
																																													Column: int(69),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6218,
																																							FreeVars: ast.Identifiers{
																																								"frac_str",
																																								"strip_trailing_zero",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(579),
																																									Column: int(41),
																																								},
																																								End: ast.Location{
																																									Line: int(579),
																																									Column: int(70),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					BranchFalse: &ast.Var{
																																						Id: "frac_str",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6218,
																																							FreeVars: ast.Identifiers{
																																								"frac_str",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(579),
																																									Column: int(76),
																																								},
																																								End: ast.Location{
																																									Line: int(579),
																																									Column: int(84),
																																								},
																																							},
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6218,
																																						FreeVars: ast.Identifiers{
																																							"frac_str",
																																							"strip_trailing_zero",
																																							"trailing",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(579),
																																								Column: int(23),
																																							},
																																							End: ast.Location{
																																								Line: int(579),
																																								Column: int(84),
																																							},
																																						},
																																					},
																																				},
																																				Left: &ast.Binary{
																																					Right: &ast.LiteralString{
																																						Value: ".",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6218,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(579),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(579),
																																									Column: int(20),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Left: &ast.Var{
																																						Id: "str",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p6218,
																																							FreeVars: ast.Identifiers{
																																								"str",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(579),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(579),
																																									Column: int(14),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6218,
																																						FreeVars: ast.Identifiers{
																																							"str",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(579),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(579),
																																								Column: int(20),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6218,
																																					FreeVars: ast.Identifiers{
																																						"frac_str",
																																						"str",
																																						"strip_trailing_zero",
																																						"trailing",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(579),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(579),
																																							Column: int(84),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																					},
																																				},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"frac",
																																					"prec",
																																					"render_int",
																																					"str",
																																					"strip_trailing_zero",
																																					"trailing",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(578),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(579),
																																						Column: int(84),
																																					},
																																				},
																																			},
																																		},
																																		BranchFalse: &ast.Var{
																																			Id: "str",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																					},
																																				},
																																				Ctx: p6218,
																																				FreeVars: ast.Identifiers{
																																					"str",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(581),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(581),
																																						Column: int(14),
																																					},
																																				},
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(8),
																																				},
																																			},
																																			Ctx: p6218,
																																			FreeVars: ast.Identifiers{
																																				"frac",
																																				"prec",
																																				"render_int",
																																				"str",
																																				"strip_trailing_zero",
																																				"trailing",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(577),
																																					Column: int(9),
																																				},
																																				End: ast.Location{
																																					Line: int(581),
																																					Column: int(14),
																																				},
																																			},
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(6),
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(6),
																																			},
																																		},
																																		Ctx: p6218,
																																		FreeVars: ast.Identifiers{
																																			"ensure_pt",
																																			"frac",
																																			"prec",
																																			"render_int",
																																			"str",
																																			"strip_trailing_zero",
																																			"trailing",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(574),
																																				Column: int(7),
																																			},
																																			End: ast.Location{
																																				Line: int(581),
																																				Column: int(14),
																																			},
																																		},
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(6),
																																		},
																																	},
																																	Ctx: p6218,
																																	FreeVars: ast.Identifiers{
																																		"blank",
																																		"ensure_pt",
																																		"frac",
																																		"n__",
																																		"plus",
																																		"prec",
																																		"render_int",
																																		"strip_trailing_zero",
																																		"trailing",
																																		"whole",
																																		"zp",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(573),
																																			Column: int(7),
																																		},
																																		End: ast.Location{
																																			Line: int(581),
																																			Column: int(14),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(6),
																																	},
																																},
																																Ctx: p6218,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"dot_size",
																																	"ensure_pt",
																																	"frac",
																																	"n__",
																																	"plus",
																																	"prec",
																																	"render_int",
																																	"strip_trailing_zero",
																																	"trailing",
																																	"whole",
																																	"zero_pad",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(572),
																																		Column: int(7),
																																	},
																																	End: ast.Location{
																																		Line: int(581),
																																		Column: int(14),
																																	},
																																},
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(6),
																																},
																															},
																															Ctx: p6218,
																															FreeVars: ast.Identifiers{
																																"blank",
																																"ensure_pt",
																																"frac",
																																"n__",
																																"plus",
																																"prec",
																																"render_int",
																																"strip_trailing_zero",
																																"trailing",
																																"whole",
																																"zero_pad",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(571),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(581),
																																	Column: int(14),
																																},
																															},
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														Ctx: p6218,
																														FreeVars: ast.Identifiers{
																															"$std",
																															"blank",
																															"denominator",
																															"ensure_pt",
																															"n__",
																															"numerator",
																															"plus",
																															"prec",
																															"render_int",
																															"std",
																															"strip_trailing_zero",
																															"trailing",
																															"whole",
																															"zero_pad",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(570),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(581),
																																Column: int(14),
																															},
																														},
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p6218,
																													FreeVars: ast.Identifiers{
																														"$std",
																														"blank",
																														"denominator",
																														"ensure_pt",
																														"n_",
																														"n__",
																														"numerator",
																														"plus",
																														"prec",
																														"render_int",
																														"std",
																														"strip_trailing_zero",
																														"trailing",
																														"zero_pad",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(569),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(581),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p6218,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"blank",
																													"denominator",
																													"ensure_pt",
																													"n_",
																													"n__",
																													"plus",
																													"prec",
																													"render_int",
																													"std",
																													"strip_trailing_zero",
																													"trailing",
																													"zero_pad",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(568),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(581),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																												ast.FodderElement{
																													Comment: []string{
																														"// Represent the rounded number as an integer * 1/10**prec.",
																													},
																													Kind: ast.FodderKind(2),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																												ast.FodderElement{
																													Comment: []string{
																														"// Note that it can also be equal to 10**prec and we'll need to carry",
																													},
																													Kind: ast.FodderKind(2),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																												ast.FodderElement{
																													Comment: []string{
																														"// over to the wholes.  We operate on the absolute numbers, so that we",
																													},
																													Kind: ast.FodderKind(2),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																												ast.FodderElement{
																													Comment: []string{
																														"// don't have trouble with the rounding direction.",
																													},
																													Kind: ast.FodderKind(2),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											Ctx: p6218,
																											FreeVars: ast.Identifiers{
																												"$std",
																												"blank",
																												"ensure_pt",
																												"n_",
																												"n__",
																												"plus",
																												"prec",
																												"render_int",
																												"std",
																												"strip_trailing_zero",
																												"trailing",
																												"zero_pad",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(567),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(581),
																													Column: int(14),
																												},
																											},
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										Ctx: p6218,
																										FreeVars: ast.Identifiers{
																											"$std",
																											"blank",
																											"ensure_pt",
																											"n_",
																											"n__",
																											"plus",
																											"prec",
																											"render_int",
																											"std",
																											"strip_trailing_zero",
																											"trailing",
																											"zero_pad",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(562),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(581),
																												Column: int(14),
																											},
																										},
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									Ctx: p6218,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"blank",
																										"ensure_pt",
																										"n__",
																										"plus",
																										"prec",
																										"render_int",
																										"std",
																										"strip_trailing_zero",
																										"trailing",
																										"zero_pad",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(561),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(581),
																											Column: int(14),
																										},
																									},
																								},
																							},
																							Parameters: []ast.Parameter{
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "n__",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(31),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "zero_pad",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(41),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "blank",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(43),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(48),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "plus",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(50),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(54),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "ensure_pt",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(56),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(65),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "trailing",
																									CommaFodder: ast.Fodder{},
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(67),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(75),
																										},
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "prec",
																									CommaFodder: nil,
																									EqFodder: nil,
																									DefaultArg: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(560),
																											Column: int(77),
																										},
																										End: ast.Location{
																											Line: int(560),
																											Column: int(81),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: p6320,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"render_int",
																									"std",
																									"strip_trailing_zero",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(560),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(581),
																										Column: int(14),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						EqFodder: nil,
																						Variable: "render_float_dec",
																						CloseFodder: nil,
																						Fun: nil,
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				Body: &ast.Local{
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: nil,
																							Body: &ast.Function{
																								ParenLeftFodder: ast.Fodder{},
																								ParenRightFodder: ast.Fodder{},
																								Body: &ast.Local{
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6330,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(585),
																																	Column: int(34),
																																},
																																End: ast.Location{
																																	Line: int(585),
																																	Column: int(35),
																																},
																															},
																														},
																													},
																													Left: &ast.Var{
																														Id: "n__",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6330,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(585),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(585),
																																	Column: int(30),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6330,
																														FreeVars: ast.Identifiers{
																															"n__",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(585),
																																Column: int(27),
																															},
																															End: ast.Location{
																																Line: int(585),
																																Column: int(35),
																															},
																														},
																													},
																													Op: ast.BinaryOp(12),
																												},
																												BranchTrue: &ast.LiteralNumber{
																													OriginalString: "0",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6330,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(585),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(585),
																																Column: int(42),
																															},
																														},
																													},
																												},
																												BranchFalse: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(585),
																																		Column: int(48),
																																	},
																																	End: ast.Location{
																																		Line: int(585),
																																		Column: int(51),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "floor",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6330,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(585),
																																	Column: int(48),
																																},
																																End: ast.Location{
																																	Line: int(585),
																																	Column: int(57),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	Right: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(585),
																																							Column: int(82),
																																						},
																																						End: ast.Location{
																																							Line: int(585),
																																							Column: int(85),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "log",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6348,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(585),
																																						Column: int(82),
																																					},
																																					End: ast.Location{
																																						Line: int(585),
																																						Column: int(89),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.LiteralNumber{
																																						OriginalString: "10",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6352,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(585),
																																									Column: int(90),
																																								},
																																								End: ast.Location{
																																									Line: int(585),
																																									Column: int(92),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6348,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(585),
																																					Column: int(82),
																																				},
																																				End: ast.Location{
																																					Line: int(585),
																																					Column: int(93),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	Left: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(585),
																																							Column: int(58),
																																						},
																																						End: ast.Location{
																																							Line: int(585),
																																							Column: int(61),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "log",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6348,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(585),
																																						Column: int(58),
																																					},
																																					End: ast.Location{
																																						Line: int(585),
																																						Column: int(65),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(585),
																																											Column: int(66),
																																										},
																																										End: ast.Location{
																																											Line: int(585),
																																											Column: int(69),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "abs",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6366,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(585),
																																										Column: int(66),
																																									},
																																									End: ast.Location{
																																										Line: int(585),
																																										Column: int(73),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "n__",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6370,
																																											FreeVars: ast.Identifiers{
																																												"n__",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(585),
																																													Column: int(74),
																																												},
																																												End: ast.Location{
																																													Line: int(585),
																																													Column: int(77),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6366,
																																							FreeVars: ast.Identifiers{
																																								"n__",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(585),
																																									Column: int(66),
																																								},
																																								End: ast.Location{
																																									Line: int(585),
																																									Column: int(78),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6348,
																																			FreeVars: ast.Identifiers{
																																				"n__",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(585),
																																					Column: int(58),
																																				},
																																				End: ast.Location{
																																					Line: int(585),
																																					Column: int(79),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6348,
																																		FreeVars: ast.Identifiers{
																																			"n__",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(585),
																																				Column: int(58),
																																			},
																																			End: ast.Location{
																																				Line: int(585),
																																				Column: int(93),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(1),
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6330,
																														FreeVars: ast.Identifiers{
																															"n__",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(585),
																																Column: int(48),
																															},
																															End: ast.Location{
																																Line: int(585),
																																Column: int(94),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p6330,
																													FreeVars: ast.Identifiers{
																														"n__",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(585),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(585),
																															Column: int(94),
																														},
																													},
																												},
																											},
																											EqFodder: ast.Fodder{},
																											Variable: "exponent",
																											CloseFodder: ast.Fodder{},
																											Fun: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(585),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(585),
																													Column: int(94),
																												},
																											},
																										},
																									},
																									Body: &ast.Local{
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Body: &ast.Binary{
																													Right: &ast.Apply{
																														Target: &ast.Var{
																															Id: "render_int",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6382,
																																FreeVars: ast.Identifiers{
																																	"render_int",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(587),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(587),
																																		Column: int(32),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		Right: &ast.LiteralNumber{
																																			OriginalString: "0",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6387,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(587),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(587),
																																						Column: int(45),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "exponent",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6387,
																																				FreeVars: ast.Identifiers{
																																					"exponent",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(587),
																																						Column: int(33),
																																					},
																																					End: ast.Location{
																																						Line: int(587),
																																						Column: int(41),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{
																																				"exponent",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(33),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(45),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(9),
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(587),
																																							Column: int(47),
																																						},
																																						End: ast.Location{
																																							Line: int(587),
																																							Column: int(50),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "abs",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6387,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(587),
																																						Column: int(47),
																																					},
																																					End: ast.Location{
																																						Line: int(587),
																																						Column: int(54),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "exponent",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6399,
																																							FreeVars: ast.Identifiers{
																																								"exponent",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(587),
																																									Column: int(55),
																																								},
																																								End: ast.Location{
																																									Line: int(587),
																																									Column: int(63),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{
																																				"exponent",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(47),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(64),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		OriginalString: "3",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(66),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(67),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		OriginalString: "0",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(69),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(70),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralBoolean{
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(72),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(77),
																																				},
																																			},
																																		},
																																		Value: false,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralBoolean{
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(79),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(83),
																																				},
																																			},
																																		},
																																		Value: true,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		OriginalString: "10",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(85),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(87),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralString{
																																		Value: "",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6387,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(89),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(91),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6382,
																															FreeVars: ast.Identifiers{
																																"exponent",
																																"render_int",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(587),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(587),
																																	Column: int(92),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													Left: &ast.Conditional{
																														Cond: &ast.Var{
																															Id: "caps",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6382,
																																FreeVars: ast.Identifiers{
																																	"caps",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(586),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(586),
																																		Column: int(28),
																																	},
																																},
																															},
																														},
																														BranchTrue: &ast.LiteralString{
																															Value: "E",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6382,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(586),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(586),
																																		Column: int(37),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														BranchFalse: &ast.LiteralString{
																															Value: "e",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6382,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(586),
																																		Column: int(43),
																																	},
																																	End: ast.Location{
																																		Line: int(586),
																																		Column: int(46),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6382,
																															FreeVars: ast.Identifiers{
																																"caps",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(586),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(586),
																																	Column: int(46),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(19),
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6382,
																														FreeVars: ast.Identifiers{
																															"caps",
																															"exponent",
																															"render_int",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(586),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(587),
																																Column: int(92),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												EqFodder: ast.Fodder{},
																												Variable: "suff",
																												CloseFodder: ast.Fodder{},
																												Fun: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(586),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(587),
																														Column: int(92),
																													},
																												},
																											},
																										},
																										Body: &ast.Local{
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Body: &ast.Conditional{
																														Cond: &ast.Binary{
																															Right: &ast.Unary{
																																Expr: &ast.LiteralNumber{
																																	OriginalString: "324",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6423,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(588),
																																				Column: int(40),
																																			},
																																			End: ast.Location{
																																				Line: int(588),
																																				Column: int(43),
																																			},
																																		},
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(588),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(588),
																																			Column: int(43),
																																		},
																																	},
																																},
																																Op: ast.UnaryOp(3),
																															},
																															Left: &ast.Var{
																																Id: "exponent",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(588),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(588),
																																			Column: int(35),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6423,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(588),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(588),
																																		Column: int(43),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(12),
																														},
																														BranchTrue: &ast.Binary{
																															Right: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(591),
																																					Column: int(20),
																																				},
																																				End: ast.Location{
																																					Line: int(591),
																																					Column: int(23),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "pow",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6423,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(591),
																																				Column: int(20),
																																			},
																																			End: ast.Location{
																																				Line: int(591),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.LiteralNumber{
																																				OriginalString: "10",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6436,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(591),
																																							Column: int(28),
																																						},
																																						End: ast.Location{
																																							Line: int(591),
																																							Column: int(30),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: ast.Fodder{},
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Binary{
																																				Right: &ast.LiteralNumber{
																																					OriginalString: "1",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6436,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(43),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(44),
																																							},
																																						},
																																					},
																																				},
																																				Left: &ast.Var{
																																					Id: "exponent",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6436,
																																						FreeVars: ast.Identifiers{
																																							"exponent",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(32),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(40),
																																							},
																																						},
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6436,
																																					FreeVars: ast.Identifiers{
																																						"exponent",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(591),
																																							Column: int(32),
																																						},
																																						End: ast.Location{
																																							Line: int(591),
																																							Column: int(44),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(591),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(591),
																																			Column: int(45),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Left: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "10",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6423,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(591),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(591),
																																				Column: int(17),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Var{
																																	Id: "n__",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																			ast.FodderElement{
																																				Comment: []string{
																																					"// Avoid a rounding error where std.pow(10, -324) is 0",
																																				},
																																				Kind: ast.FodderKind(2),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																			ast.FodderElement{
																																				Comment: []string{
																																					"// -324 is the smallest exponent possible.",
																																				},
																																				Kind: ast.FodderKind(2),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																		},
																																		Ctx: p6423,
																																		FreeVars: ast.Identifiers{
																																			"n__",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(591),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(591),
																																				Column: int(12),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{
																																		"n__",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(591),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(591),
																																			Column: int(17),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(0),
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6423,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																	"n__",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(591),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(591),
																																		Column: int(45),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(1),
																														},
																														BranchFalse: &ast.Binary{
																															Right: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(593),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(593),
																																					Column: int(18),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "pow",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6423,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(593),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(593),
																																				Column: int(22),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.LiteralNumber{
																																				OriginalString: "10",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6461,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(593),
																																							Column: int(23),
																																						},
																																						End: ast.Location{
																																							Line: int(593),
																																							Column: int(25),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: ast.Fodder{},
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "exponent",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6461,
																																					FreeVars: ast.Identifiers{
																																						"exponent",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(593),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(593),
																																							Column: int(35),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(593),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(593),
																																			Column: int(36),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Left: &ast.Var{
																																Id: "n__",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																		},
																																	},
																																	Ctx: p6423,
																																	FreeVars: ast.Identifiers{
																																		"n__",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(593),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(593),
																																			Column: int(12),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6423,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																	"n__",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(593),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(593),
																																		Column: int(36),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(1),
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6423,
																															FreeVars: ast.Identifiers{
																																"exponent",
																																"n__",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(588),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(593),
																																	Column: int(36),
																																},
																															},
																														},
																													},
																													EqFodder: ast.Fodder{},
																													Variable: "mantissa",
																													CloseFodder: ast.Fodder{},
																													Fun: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(588),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(593),
																															Column: int(36),
																														},
																													},
																												},
																											},
																											Body: &ast.Local{
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Body: &ast.Binary{
																															Right: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(594),
																																					Column: int(30),
																																				},
																																				End: ast.Location{
																																					Line: int(594),
																																					Column: int(33),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "length",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6479,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(594),
																																				Column: int(30),
																																			},
																																			End: ast.Location{
																																				Line: int(594),
																																				Column: int(40),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "suff",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6483,
																																					FreeVars: ast.Identifiers{
																																						"suff",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(594),
																																							Column: int(41),
																																						},
																																						End: ast.Location{
																																							Line: int(594),
																																							Column: int(45),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6479,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																		"suff",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(594),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(594),
																																			Column: int(46),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Left: &ast.Var{
																																Id: "zero_pad",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6479,
																																	FreeVars: ast.Identifiers{
																																		"zero_pad",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(594),
																																			Column: int(19),
																																		},
																																		End: ast.Location{
																																			Line: int(594),
																																			Column: int(27),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6479,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"suff",
																																	"zero_pad",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(594),
																																		Column: int(19),
																																	},
																																	End: ast.Location{
																																		Line: int(594),
																																		Column: int(46),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(4),
																														},
																														EqFodder: ast.Fodder{},
																														Variable: "zp2",
																														CloseFodder: ast.Fodder{},
																														Fun: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(594),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(594),
																																Column: int(46),
																															},
																														},
																													},
																												},
																												Body: &ast.Binary{
																													Right: &ast.Var{
																														Id: "suff",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6491,
																															FreeVars: ast.Identifiers{
																																"suff",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(595),
																																	Column: int(81),
																																},
																																End: ast.Location{
																																	Line: int(595),
																																	Column: int(85),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Var{
																															Id: "render_float_dec",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(6),
																																	},
																																},
																																Ctx: p6491,
																																FreeVars: ast.Identifiers{
																																	"render_float_dec",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(595),
																																		Column: int(7),
																																	},
																																	End: ast.Location{
																																		Line: int(595),
																																		Column: int(23),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "mantissa",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"mantissa",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "zp2",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"zp2",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(34),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "blank",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(39),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(44),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "plus",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"plus",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(46),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(50),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "ensure_pt",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"ensure_pt",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(52),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(61),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "trailing",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"trailing",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(63),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(71),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "prec",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6499,
																																			FreeVars: ast.Identifiers{
																																				"prec",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(595),
																																					Column: int(73),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(77),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6491,
																															FreeVars: ast.Identifiers{
																																"blank",
																																"ensure_pt",
																																"mantissa",
																																"plus",
																																"prec",
																																"render_float_dec",
																																"trailing",
																																"zp2",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(595),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(595),
																																	Column: int(78),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6491,
																														FreeVars: ast.Identifiers{
																															"blank",
																															"ensure_pt",
																															"mantissa",
																															"plus",
																															"prec",
																															"render_float_dec",
																															"suff",
																															"trailing",
																															"zp2",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(595),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(595),
																																Column: int(85),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p6491,
																													FreeVars: ast.Identifiers{
																														"blank",
																														"ensure_pt",
																														"mantissa",
																														"plus",
																														"prec",
																														"render_float_dec",
																														"std",
																														"suff",
																														"trailing",
																														"zero_pad",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(594),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(595),
																															Column: int(85),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p6491,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"ensure_pt",
																													"exponent",
																													"n__",
																													"plus",
																													"prec",
																													"render_float_dec",
																													"std",
																													"suff",
																													"trailing",
																													"zero_pad",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(588),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(595),
																														Column: int(85),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											Ctx: p6491,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"caps",
																												"ensure_pt",
																												"exponent",
																												"n__",
																												"plus",
																												"prec",
																												"render_float_dec",
																												"render_int",
																												"std",
																												"trailing",
																												"zero_pad",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(586),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(595),
																													Column: int(85),
																												},
																											},
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										Ctx: p6491,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"caps",
																											"ensure_pt",
																											"n__",
																											"plus",
																											"prec",
																											"render_float_dec",
																											"render_int",
																											"std",
																											"trailing",
																											"zero_pad",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(585),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(595),
																												Column: int(85),
																											},
																										},
																									},
																								},
																								Parameters: []ast.Parameter{
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "n__",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(31),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "zero_pad",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(41),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "blank",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(43),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(48),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "plus",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(50),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(54),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "ensure_pt",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(56),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(65),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "trailing",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(67),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(75),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "caps",
																										CommaFodder: ast.Fodder{},
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(77),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(81),
																											},
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "prec",
																										CommaFodder: nil,
																										EqFodder: nil,
																										DefaultArg: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(584),
																												Column: int(83),
																											},
																											End: ast.Location{
																												Line: int(584),
																												Column: int(87),
																											},
																										},
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: p6524,
																									FreeVars: ast.Identifiers{
																										"render_float_dec",
																										"render_int",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(584),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(595),
																											Column: int(85),
																										},
																									},
																								},
																								TrailingComma: false,
																							},
																							EqFodder: nil,
																							Variable: "render_float_sci",
																							CloseFodder: nil,
																							Fun: nil,
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					Body: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: nil,
																								Body: &ast.Function{
																									ParenLeftFodder: ast.Fodder{},
																									ParenRightFodder: ast.Fodder{},
																									Body: &ast.Local{
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Body: &ast.Index{
																													Target: &ast.Var{
																														Id: "code",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"code",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(599),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(599),
																																	Column: int(26),
																																},
																															},
																														},
																													},
																													Index: &ast.LiteralString{
																														Value: "cflags",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													RightBracketFodder: ast.Fodder{},
																													LeftBracketFodder: ast.Fodder{},
																													Id: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p6535,
																														FreeVars: ast.Identifiers{
																															"code",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(599),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(599),
																																Column: int(33),
																															},
																														},
																													},
																												},
																												EqFodder: ast.Fodder{},
																												Variable: "cflags",
																												CloseFodder: ast.Fodder{},
																												Fun: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(599),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(599),
																														Column: int(33),
																													},
																												},
																											},
																										},
																										Body: &ast.Local{
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Body: &ast.Conditional{
																														Cond: &ast.Binary{
																															Right: &ast.LiteralNull{
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6542,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(600),
																																			Column: int(41),
																																		},
																																		End: ast.Location{
																																			Line: int(600),
																																			Column: int(45),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Var{
																																Id: "prec_or_null",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6542,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(600),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(600),
																																			Column: int(37),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6542,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(600),
																																		Column: int(25),
																																	},
																																	End: ast.Location{
																																		Line: int(600),
																																		Column: int(45),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(13),
																														},
																														BranchTrue: &ast.Var{
																															Id: "prec_or_null",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6542,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(600),
																																		Column: int(51),
																																	},
																																	End: ast.Location{
																																		Line: int(600),
																																		Column: int(63),
																																	},
																																},
																															},
																														},
																														BranchFalse: &ast.LiteralNumber{
																															OriginalString: "6",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6542,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(600),
																																		Column: int(69),
																																	},
																																	End: ast.Location{
																																		Line: int(600),
																																		Column: int(70),
																																	},
																																},
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p6542,
																															FreeVars: ast.Identifiers{
																																"prec_or_null",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(600),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(600),
																																	Column: int(70),
																																},
																															},
																														},
																													},
																													EqFodder: ast.Fodder{},
																													Variable: "fpprec",
																													CloseFodder: ast.Fodder{},
																													Fun: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(600),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(600),
																															Column: int(70),
																														},
																													},
																												},
																											},
																											Body: &ast.Local{
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Body: &ast.Conditional{
																															Cond: &ast.Binary{
																																Right: &ast.LiteralNull{
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6555,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(601),
																																				Column: int(40),
																																			},
																																			End: ast.Location{
																																				Line: int(601),
																																				Column: int(44),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Var{
																																	Id: "prec_or_null",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6555,
																																		FreeVars: ast.Identifiers{
																																			"prec_or_null",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(601),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(601),
																																				Column: int(36),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6555,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(601),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(601),
																																			Column: int(44),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(13),
																															},
																															BranchTrue: &ast.Var{
																																Id: "prec_or_null",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6555,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(601),
																																			Column: int(50),
																																		},
																																		End: ast.Location{
																																			Line: int(601),
																																			Column: int(62),
																																		},
																																	},
																																},
																															},
																															BranchFalse: &ast.LiteralNumber{
																																OriginalString: "0",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6555,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(601),
																																			Column: int(68),
																																		},
																																		End: ast.Location{
																																			Line: int(601),
																																			Column: int(69),
																																		},
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															ElseFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6555,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(601),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(601),
																																		Column: int(69),
																																	},
																																},
																															},
																														},
																														EqFodder: ast.Fodder{},
																														Variable: "iprec",
																														CloseFodder: ast.Fodder{},
																														Fun: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(601),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(601),
																																Column: int(69),
																															},
																														},
																													},
																												},
																												Body: &ast.Local{
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Body: &ast.Conditional{
																																Cond: &ast.Binary{
																																	Right: &ast.Unary{
																																		Expr: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "cflags",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(602),
																																							Column: int(37),
																																						},
																																						End: ast.Location{
																																							Line: int(602),
																																							Column: int(43),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "left",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6572,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(602),
																																						Column: int(37),
																																					},
																																					End: ast.Location{
																																						Line: int(602),
																																						Column: int(48),
																																					},
																																				},
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6572,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(602),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(602),
																																					Column: int(48),
																																				},
																																			},
																																		},
																																		Op: ast.UnaryOp(0),
																																	},
																																	Left: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "cflags",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(602),
																																						Column: int(21),
																																					},
																																					End: ast.Location{
																																						Line: int(602),
																																						Column: int(27),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "zero",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6572,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(602),
																																					Column: int(21),
																																				},
																																				End: ast.Location{
																																					Line: int(602),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6572,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(602),
																																				Column: int(21),
																																			},
																																			End: ast.Location{
																																				Line: int(602),
																																				Column: int(48),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(17),
																																},
																																BranchTrue: &ast.Var{
																																	Id: "fw",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6572,
																																		FreeVars: ast.Identifiers{
																																			"fw",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(602),
																																				Column: int(54),
																																			},
																																			End: ast.Location{
																																				Line: int(602),
																																				Column: int(56),
																																			},
																																		},
																																	},
																																},
																																BranchFalse: &ast.LiteralNumber{
																																	OriginalString: "0",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6572,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(602),
																																				Column: int(62),
																																			},
																																			End: ast.Location{
																																				Line: int(602),
																																				Column: int(63),
																																			},
																																		},
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																ElseFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6572,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"fw",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(602),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(602),
																																			Column: int(63),
																																		},
																																	},
																																},
																															},
																															EqFodder: ast.Fodder{},
																															Variable: "zp",
																															CloseFodder: ast.Fodder{},
																															Fun: nil,
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(602),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(602),
																																	Column: int(63),
																																},
																															},
																														},
																													},
																													Body: &ast.Conditional{
																														Cond: &ast.Binary{
																															Right: &ast.LiteralString{
																																Value: "s",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(603),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(603),
																																			Column: int(27),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Left: &ast.Index{
																																Target: &ast.Var{
																																	Id: "code",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(603),
																																				Column: int(10),
																																			},
																																			End: ast.Location{
																																				Line: int(603),
																																				Column: int(14),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "ctype",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(603),
																																			Column: int(10),
																																		},
																																		End: ast.Location{
																																			Line: int(603),
																																			Column: int(20),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6588,
																																FreeVars: ast.Identifiers{
																																	"code",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(603),
																																		Column: int(10),
																																	},
																																	End: ast.Location{
																																		Line: int(603),
																																		Column: int(27),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(12),
																														},
																														BranchTrue: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																		},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(604),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(604),
																																				Column: int(12),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "toString",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(604),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(604),
																																			Column: int(21),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "val",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6604,
																																				FreeVars: ast.Identifiers{
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(604),
																																						Column: int(22),
																																					},
																																					End: ast.Location{
																																						Line: int(604),
																																						Column: int(25),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6588,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"val",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(604),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(604),
																																		Column: int(26),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														BranchFalse: &ast.Conditional{
																															Cond: &ast.Binary{
																																Right: &ast.LiteralString{
																																	Value: "d",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(605),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(605),
																																				Column: int(32),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																Left: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "code",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(605),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(605),
																																					Column: int(19),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "ctype",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(605),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(605),
																																				Column: int(25),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(605),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(605),
																																			Column: int(32),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(12),
																															},
																															BranchTrue: &ast.Conditional{
																																Cond: &ast.Binary{
																																	Right: &ast.LiteralString{
																																		Value: "number",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(606),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(606),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	Left: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(606),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(606),
																																							Column: int(15),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "type",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(606),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(606),
																																						Column: int(20),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "val",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6627,
																																							FreeVars: ast.Identifiers{
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(606),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(606),
																																									Column: int(24),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																				"val",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(606),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(606),
																																					Column: int(25),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																			"val",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(606),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(606),
																																				Column: int(37),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(13),
																																},
																																BranchTrue: &ast.Error{
																																	Expr: &ast.Binary{
																																		Right: &ast.Apply{
																																			Target: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "std",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(608),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(608),
																																								Column: int(37),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "type",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(608),
																																							Column: int(34),
																																						},
																																						End: ast.Location{
																																							Line: int(608),
																																							Column: int(42),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "val",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6641,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(608),
																																										Column: int(43),
																																									},
																																									End: ast.Location{
																																										Line: int(608),
																																										Column: int(46),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(608),
																																						Column: int(34),
																																					},
																																					End: ast.Location{
																																						Line: int(608),
																																						Column: int(47),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		Left: &ast.Binary{
																																			Right: &ast.LiteralString{
																																				Value: ", got ",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(608),
																																							Column: int(23),
																																						},
																																						End: ast.Location{
																																							Line: int(608),
																																							Column: int(31),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			Left: &ast.Binary{
																																				Right: &ast.Var{
																																					Id: "i",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(608),
																																								Column: int(19),
																																							},
																																							End: ast.Location{
																																								Line: int(608),
																																								Column: int(20),
																																							},
																																						},
																																					},
																																				},
																																				Left: &ast.LiteralString{
																																					Value: "Format required number at ",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(607),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(607),
																																								Column: int(45),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				OpFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(16),
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(607),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(608),
																																							Column: int(20),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(607),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(608),
																																						Column: int(31),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(3),
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"std",
																																				"val",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(607),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(608),
																																					Column: int(47),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(3),
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																			},
																																		},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																			"std",
																																			"val",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(607),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(608),
																																				Column: int(47),
																																			},
																																		},
																																	},
																																},
																																BranchFalse: &ast.Apply{
																																	Target: &ast.Var{
																																		Id: "render_int",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																				},
																																			},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"render_int",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(610),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(610),
																																					Column: int(21),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					Right: &ast.Unary{
																																						Expr: &ast.LiteralNumber{
																																							OriginalString: "1",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6664,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(610),
																																										Column: int(30),
																																									},
																																									End: ast.Location{
																																										Line: int(610),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6664,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(610),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(610),
																																									Column: int(31),
																																								},
																																							},
																																						},
																																						Op: ast.UnaryOp(3),
																																					},
																																					Left: &ast.Var{
																																						Id: "val",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6664,
																																							FreeVars: ast.Identifiers{
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(610),
																																									Column: int(22),
																																								},
																																								End: ast.Location{
																																									Line: int(610),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(22),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(31),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(10),
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Apply{
																																					Target: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "std",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(610),
																																										Column: int(33),
																																									},
																																									End: ast.Location{
																																										Line: int(610),
																																										Column: int(36),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "floor",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6664,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(610),
																																									Column: int(33),
																																								},
																																								End: ast.Location{
																																									Line: int(610),
																																									Column: int(42),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Apply{
																																									Target: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "std",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(610),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(610),
																																														Column: int(46),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "abs",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6680,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(610),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(610),
																																													Column: int(50),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													Id: "val",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p6684,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(610),
																																																Column: int(51),
																																															},
																																															End: ast.Location{
																																																Line: int(610),
																																																Column: int(54),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6680,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(610),
																																												Column: int(43),
																																											},
																																											End: ast.Location{
																																												Line: int(610),
																																												Column: int(55),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(33),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(56),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "zp",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"zp",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(58),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(60),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "iprec",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"iprec",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(62),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(67),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "cflags",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(610),
																																									Column: int(69),
																																								},
																																								End: ast.Location{
																																									Line: int(610),
																																									Column: int(75),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "blank",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(69),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(81),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "cflags",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(610),
																																									Column: int(83),
																																								},
																																								End: ast.Location{
																																									Line: int(610),
																																									Column: int(89),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "plus",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(83),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(94),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralNumber{
																																					OriginalString: "10",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(96),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(98),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralString{
																																					Value: "",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6664,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(610),
																																								Column: int(100),
																																							},
																																							End: ast.Location{
																																								Line: int(610),
																																								Column: int(102),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"iprec",
																																			"render_int",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(610),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(610),
																																				Column: int(103),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																ThenFodder: ast.Fodder{},
																																ElseFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(8),
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																		},
																																	},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"i",
																																		"iprec",
																																		"render_int",
																																		"std",
																																		"val",
																																		"zp",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(606),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(610),
																																			Column: int(103),
																																		},
																																	},
																																},
																															},
																															BranchFalse: &ast.Conditional{
																																Cond: &ast.Binary{
																																	Right: &ast.LiteralString{
																																		Value: "o",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(611),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(611),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	Left: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "code",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(611),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(611),
																																						Column: int(19),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "ctype",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(611),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(611),
																																					Column: int(25),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(611),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(611),
																																				Column: int(32),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(12),
																																},
																																BranchTrue: &ast.Conditional{
																																	Cond: &ast.Binary{
																																		Right: &ast.LiteralString{
																																			Value: "number",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(612),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(612),
																																						Column: int(37),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		Left: &ast.Apply{
																																			Target: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "std",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(612),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(612),
																																								Column: int(15),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "type",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(612),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(612),
																																							Column: int(20),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "val",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6728,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(612),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(612),
																																										Column: int(24),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(612),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(612),
																																						Column: int(25),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																				"val",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(612),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(612),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(13),
																																	},
																																	BranchTrue: &ast.Error{
																																		Expr: &ast.Binary{
																																			Right: &ast.Apply{
																																				Target: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "std",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(614),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(614),
																																									Column: int(37),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "type",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(614),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(614),
																																								Column: int(42),
																																							},
																																						},
																																					},
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "val",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6742,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(614),
																																											Column: int(43),
																																										},
																																										End: ast.Location{
																																											Line: int(614),
																																											Column: int(46),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(614),
																																							Column: int(34),
																																						},
																																						End: ast.Location{
																																							Line: int(614),
																																							Column: int(47),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																			},
																																			Left: &ast.Binary{
																																				Right: &ast.LiteralString{
																																					Value: ", got ",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(614),
																																								Column: int(23),
																																							},
																																							End: ast.Location{
																																								Line: int(614),
																																								Column: int(31),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Left: &ast.Binary{
																																					Right: &ast.Var{
																																						Id: "i",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(614),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(614),
																																									Column: int(20),
																																								},
																																							},
																																						},
																																					},
																																					Left: &ast.LiteralString{
																																						Value: "Format required number at ",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(613),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(613),
																																									Column: int(45),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					OpFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(16),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(613),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(614),
																																								Column: int(20),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(613),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(614),
																																							Column: int(31),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																					"std",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(613),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(614),
																																						Column: int(47),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(3),
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																				},
																																			},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"std",
																																				"val",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(613),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(614),
																																					Column: int(47),
																																				},
																																			},
																																		},
																																	},
																																	BranchFalse: &ast.Local{
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					Cond: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "cflags",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(616),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(616),
																																										Column: int(40),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "alt",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6764,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(616),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(616),
																																									Column: int(44),
																																								},
																																							},
																																						},
																																					},
																																					BranchTrue: &ast.LiteralString{
																																						Value: "0",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6764,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(616),
																																									Column: int(50),
																																								},
																																								End: ast.Location{
																																									Line: int(616),
																																									Column: int(53),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					BranchFalse: &ast.LiteralString{
																																						Value: "",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6764,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(616),
																																									Column: int(59),
																																								},
																																								End: ast.Location{
																																									Line: int(616),
																																									Column: int(61),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6764,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(616),
																																								Column: int(31),
																																							},
																																							End: ast.Location{
																																								Line: int(616),
																																								Column: int(61),
																																							},
																																						},
																																					},
																																				},
																																				EqFodder: ast.Fodder{},
																																				Variable: "zero_prefix",
																																				CloseFodder: ast.Fodder{},
																																				Fun: nil,
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(616),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(616),
																																						Column: int(61),
																																					},
																																				},
																																			},
																																		},
																																		Body: &ast.Apply{
																																			Target: &ast.Var{
																																				Id: "render_int",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																						},
																																					},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"render_int",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(617),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(617),
																																							Column: int(21),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Binary{
																																							Right: &ast.Unary{
																																								Expr: &ast.LiteralNumber{
																																									OriginalString: "1",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6777,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(617),
																																												Column: int(30),
																																											},
																																											End: ast.Location{
																																												Line: int(617),
																																												Column: int(31),
																																											},
																																										},
																																									},
																																								},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6777,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																								Op: ast.UnaryOp(3),
																																							},
																																							Left: &ast.Var{
																																								Id: "val",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6777,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(25),
																																										},
																																									},
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(10),
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(617),
																																												Column: int(33),
																																											},
																																											End: ast.Location{
																																												Line: int(617),
																																												Column: int(36),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "floor",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6777,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(33),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(42),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Apply{
																																											Target: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "std",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(617),
																																																Column: int(43),
																																															},
																																															End: ast.Location{
																																																Line: int(617),
																																																Column: int(46),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "abs",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p6793,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(617),
																																															Column: int(43),
																																														},
																																														End: ast.Location{
																																															Line: int(617),
																																															Column: int(50),
																																														},
																																													},
																																												},
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															Id: "val",
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p6797,
																																																FreeVars: ast.Identifiers{
																																																	"val",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(617),
																																																		Column: int(51),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(617),
																																																		Column: int(54),
																																																	},
																																																},
																																															},
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6793,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(617),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(617),
																																														Column: int(55),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(33),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(56),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "zp",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"zp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(58),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(60),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "iprec",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"iprec",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(62),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(67),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "cflags",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(69),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(75),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "blank",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(69),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(81),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "cflags",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(83),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(89),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "plus",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(83),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(94),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							OriginalString: "8",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(96),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(97),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "zero_prefix",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6777,
																																								FreeVars: ast.Identifiers{
																																									"zero_prefix",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(617),
																																										Column: int(99),
																																									},
																																									End: ast.Location{
																																										Line: int(617),
																																										Column: int(110),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"iprec",
																																					"render_int",
																																					"std",
																																					"val",
																																					"zero_prefix",
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(617),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(617),
																																						Column: int(111),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																				},
																																			},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"iprec",
																																				"render_int",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(616),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(617),
																																					Column: int(111),
																																				},
																																			},
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																		},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"i",
																																			"iprec",
																																			"render_int",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(612),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(617),
																																				Column: int(111),
																																			},
																																		},
																																	},
																																},
																																BranchFalse: &ast.Conditional{
																																	Cond: &ast.Binary{
																																		Right: &ast.LiteralString{
																																			Value: "x",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(618),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(618),
																																						Column: int(32),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		Left: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "code",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(618),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(618),
																																							Column: int(19),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "ctype",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(618),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(618),
																																						Column: int(25),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(618),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(618),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(12),
																																	},
																																	BranchTrue: &ast.Conditional{
																																		Cond: &ast.Binary{
																																			Right: &ast.LiteralString{
																																				Value: "number",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(619),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(619),
																																							Column: int(37),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			Left: &ast.Apply{
																																				Target: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "std",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(619),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(619),
																																									Column: int(15),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "type",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(619),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(619),
																																								Column: int(20),
																																							},
																																						},
																																					},
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "val",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6844,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(619),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(619),
																																											Column: int(24),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(619),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(619),
																																							Column: int(25),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(619),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(619),
																																						Column: int(37),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(13),
																																		},
																																		BranchTrue: &ast.Error{
																																			Expr: &ast.Binary{
																																				Right: &ast.Apply{
																																					Target: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "std",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(621),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(621),
																																										Column: int(37),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "type",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(621),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(621),
																																									Column: int(42),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "val",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6858,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(621),
																																												Column: int(43),
																																											},
																																											End: ast.Location{
																																												Line: int(621),
																																												Column: int(46),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(621),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(621),
																																								Column: int(47),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																				},
																																				Left: &ast.Binary{
																																					Right: &ast.LiteralString{
																																						Value: ", got ",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(621),
																																									Column: int(23),
																																								},
																																								End: ast.Location{
																																									Line: int(621),
																																									Column: int(31),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Left: &ast.Binary{
																																						Right: &ast.Var{
																																							Id: "i",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(621),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(621),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																						},
																																						Left: &ast.LiteralString{
																																							Value: "Format required number at ",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(620),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(620),
																																										Column: int(45),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						OpFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(16),
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(620),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(621),
																																									Column: int(20),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(620),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(621),
																																								Column: int(31),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																						"std",
																																						"val",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(620),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(621),
																																							Column: int(47),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																					},
																																				},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																					"std",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(620),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(621),
																																						Column: int(47),
																																					},
																																				},
																																			},
																																		},
																																		BranchFalse: &ast.Apply{
																																			Target: &ast.Var{
																																				Id: "render_hex",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																						},
																																					},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"render_hex",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(623),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(623),
																																							Column: int(21),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(623),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(623),
																																												Column: int(25),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "floor",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6883,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(623),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(623),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "val",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6887,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(623),
																																														Column: int(32),
																																													},
																																													End: ast.Location{
																																														Line: int(623),
																																														Column: int(35),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(623),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(623),
																																										Column: int(36),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "zp",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(21),
																																									},
																																								},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"zp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(624),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(624),
																																										Column: int(24),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							Id: "iprec",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(21),
																																									},
																																								},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"iprec",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(625),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(625),
																																										Column: int(27),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "cflags",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(626),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(626),
																																											Column: int(28),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "blank",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(626),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(626),
																																										Column: int(34),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "cflags",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(627),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(627),
																																											Column: int(28),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "plus",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(627),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(627),
																																										Column: int(33),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "cflags",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(628),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(628),
																																											Column: int(28),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "alt",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(628),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(628),
																																										Column: int(32),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "code",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(629),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(629),
																																											Column: int(26),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "caps",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6883,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(629),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(629),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"code",
																																					"iprec",
																																					"render_hex",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(623),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(629),
																																						Column: int(32),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(8),
																																				},
																																			},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"code",
																																				"i",
																																				"iprec",
																																				"render_hex",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(619),
																																					Column: int(9),
																																				},
																																				End: ast.Location{
																																					Line: int(629),
																																					Column: int(32),
																																				},
																																			},
																																		},
																																	},
																																	BranchFalse: &ast.Conditional{
																																		Cond: &ast.Binary{
																																			Right: &ast.LiteralString{
																																				Value: "f",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(630),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(630),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			Left: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "code",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(630),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(630),
																																								Column: int(19),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "ctype",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(630),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(630),
																																							Column: int(25),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(630),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(630),
																																						Column: int(32),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(12),
																																		},
																																		BranchTrue: &ast.Conditional{
																																			Cond: &ast.Binary{
																																				Right: &ast.LiteralString{
																																					Value: "number",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(631),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(631),
																																								Column: int(37),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Left: &ast.Apply{
																																					Target: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "std",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(631),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(631),
																																										Column: int(15),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "type",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(631),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(631),
																																									Column: int(20),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "val",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6944,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(631),
																																												Column: int(21),
																																											},
																																											End: ast.Location{
																																												Line: int(631),
																																												Column: int(24),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(631),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(631),
																																								Column: int(25),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(631),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(631),
																																							Column: int(37),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(13),
																																			},
																																			BranchTrue: &ast.Error{
																																				Expr: &ast.Binary{
																																					Right: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(633),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(633),
																																											Column: int(37),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "type",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(633),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(633),
																																										Column: int(42),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "val",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6958,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(633),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(633),
																																													Column: int(46),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(633),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(633),
																																									Column: int(47),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					Left: &ast.Binary{
																																						Right: &ast.LiteralString{
																																							Value: ", got ",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(633),
																																										Column: int(23),
																																									},
																																									End: ast.Location{
																																										Line: int(633),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						Left: &ast.Binary{
																																							Right: &ast.Var{
																																								Id: "i",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(633),
																																											Column: int(19),
																																										},
																																										End: ast.Location{
																																											Line: int(633),
																																											Column: int(20),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								Value: "Format required number at ",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(632),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(632),
																																											Column: int(45),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							OpFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(16),
																																								},
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(632),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(633),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(632),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(633),
																																									Column: int(31),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(632),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(633),
																																								Column: int(47),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																						},
																																					},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																						"std",
																																						"val",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(632),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(633),
																																							Column: int(47),
																																						},
																																					},
																																				},
																																			},
																																			BranchFalse: &ast.Apply{
																																				Target: &ast.Var{
																																					Id: "render_float_dec",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(10),
																																							},
																																						},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"render_float_dec",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(635),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(635),
																																								Column: int(27),
																																							},
																																						},
																																					},
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "val",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(635),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(635),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "zp",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																										},
																																									},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"zp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(636),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(636),
																																											Column: int(30),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "cflags",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(637),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(637),
																																												Column: int(34),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "blank",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(637),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(637),
																																											Column: int(40),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "cflags",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(638),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(638),
																																												Column: int(34),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "plus",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(638),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(638),
																																											Column: int(39),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "cflags",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(639),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(639),
																																												Column: int(34),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "alt",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(639),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(639),
																																											Column: int(38),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.LiteralBoolean{
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																										},
																																									},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(640),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(640),
																																											Column: int(32),
																																										},
																																									},
																																								},
																																								Value: true,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "fpprec",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																										},
																																									},
																																									Ctx: p6979,
																																									FreeVars: ast.Identifiers{
																																										"fpprec",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(641),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(641),
																																											Column: int(34),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"fpprec",
																																						"render_float_dec",
																																						"val",
																																						"zp",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(635),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(641),
																																							Column: int(35),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																			},
																																			ThenFodder: ast.Fodder{},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(8),
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(8),
																																					},
																																				},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"fpprec",
																																					"i",
																																					"render_float_dec",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(631),
																																						Column: int(9),
																																					},
																																					End: ast.Location{
																																						Line: int(641),
																																						Column: int(35),
																																					},
																																				},
																																			},
																																		},
																																		BranchFalse: &ast.Conditional{
																																			Cond: &ast.Binary{
																																				Right: &ast.LiteralString{
																																					Value: "e",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(642),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(642),
																																								Column: int(32),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Left: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "code",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(642),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(642),
																																									Column: int(19),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "ctype",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(642),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(642),
																																								Column: int(25),
																																							},
																																						},
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(642),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(642),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(12),
																																			},
																																			BranchTrue: &ast.Conditional{
																																				Cond: &ast.Binary{
																																					Right: &ast.LiteralString{
																																						Value: "number",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(643),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(643),
																																									Column: int(37),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Left: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(643),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(643),
																																											Column: int(15),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "type",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(643),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(643),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "val",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7031,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(643),
																																													Column: int(21),
																																												},
																																												End: ast.Location{
																																													Line: int(643),
																																													Column: int(24),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(643),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(643),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(643),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(643),
																																								Column: int(37),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(13),
																																				},
																																				BranchTrue: &ast.Error{
																																					Expr: &ast.Binary{
																																						Right: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(645),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(645),
																																												Column: int(37),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "type",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(645),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(645),
																																											Column: int(42),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "val",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7045,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(645),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(645),
																																														Column: int(46),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(645),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(645),
																																										Column: int(47),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						Left: &ast.Binary{
																																							Right: &ast.LiteralString{
																																								Value: ", got ",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(645),
																																											Column: int(23),
																																										},
																																										End: ast.Location{
																																											Line: int(645),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							Left: &ast.Binary{
																																								Right: &ast.Var{
																																									Id: "i",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(645),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(645),
																																												Column: int(20),
																																											},
																																										},
																																									},
																																								},
																																								Left: &ast.LiteralString{
																																									Value: "Format required number at ",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(644),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(644),
																																												Column: int(45),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								OpFodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(16),
																																									},
																																								},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(644),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(645),
																																											Column: int(20),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(644),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(645),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(644),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(645),
																																									Column: int(47),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(10),
																																							},
																																						},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(644),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(645),
																																								Column: int(47),
																																							},
																																						},
																																					},
																																				},
																																				BranchFalse: &ast.Apply{
																																					Target: &ast.Var{
																																						Id: "render_float_sci",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"render_float_sci",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(647),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(647),
																																									Column: int(27),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "val",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(647),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(647),
																																												Column: int(31),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "zp",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"zp",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(648),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(648),
																																												Column: int(30),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "cflags",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(649),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(649),
																																													Column: int(34),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "blank",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(649),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(649),
																																												Column: int(40),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "cflags",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(650),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(650),
																																													Column: int(34),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "plus",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(650),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(650),
																																												Column: int(39),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "cflags",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(651),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(651),
																																													Column: int(34),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "alt",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(651),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(651),
																																												Column: int(38),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.LiteralBoolean{
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(652),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(652),
																																												Column: int(32),
																																											},
																																										},
																																									},
																																									Value: true,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "code",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(653),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(653),
																																													Column: int(32),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "caps",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(653),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(653),
																																												Column: int(37),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "fpprec",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																											},
																																										},
																																										Ctx: p7066,
																																										FreeVars: ast.Identifiers{
																																											"fpprec",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(654),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(654),
																																												Column: int(34),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																							"code",
																																							"fpprec",
																																							"render_float_sci",
																																							"val",
																																							"zp",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(647),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(654),
																																								Column: int(35),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																				},
																																				ThenFodder: ast.Fodder{},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(8),
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(8),
																																						},
																																					},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"code",
																																						"fpprec",
																																						"i",
																																						"render_float_sci",
																																						"std",
																																						"val",
																																						"zp",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(643),
																																							Column: int(9),
																																						},
																																						End: ast.Location{
																																							Line: int(654),
																																							Column: int(35),
																																						},
																																					},
																																				},
																																			},
																																			BranchFalse: &ast.Conditional{
																																				Cond: &ast.Binary{
																																					Right: &ast.LiteralString{
																																						Value: "g",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(655),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(655),
																																									Column: int(32),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Left: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "code",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(655),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(655),
																																										Column: int(19),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "ctype",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(655),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(655),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(655),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(655),
																																								Column: int(32),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(12),
																																				},
																																				BranchTrue: &ast.Conditional{
																																					Cond: &ast.Binary{
																																						Right: &ast.LiteralString{
																																							Value: "number",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(656),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(656),
																																										Column: int(37),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						Left: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(656),
																																												Column: int(12),
																																											},
																																											End: ast.Location{
																																												Line: int(656),
																																												Column: int(15),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "type",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(656),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(656),
																																											Column: int(20),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "val",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7124,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(656),
																																														Column: int(21),
																																													},
																																													End: ast.Location{
																																														Line: int(656),
																																														Column: int(24),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(656),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(656),
																																										Column: int(25),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(656),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(656),
																																									Column: int(37),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(13),
																																					},
																																					BranchTrue: &ast.Error{
																																						Expr: &ast.Binary{
																																							Right: &ast.Apply{
																																								Target: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "std",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(658),
																																													Column: int(34),
																																												},
																																												End: ast.Location{
																																													Line: int(658),
																																													Column: int(37),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "type",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(658),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(658),
																																												Column: int(42),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "val",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7138,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(658),
																																															Column: int(43),
																																														},
																																														End: ast.Location{
																																															Line: int(658),
																																															Column: int(46),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(658),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(658),
																																											Column: int(47),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							Left: &ast.Binary{
																																								Right: &ast.LiteralString{
																																									Value: ", got ",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(658),
																																												Column: int(23),
																																											},
																																											End: ast.Location{
																																												Line: int(658),
																																												Column: int(31),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.Var{
																																										Id: "i",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(658),
																																													Column: int(19),
																																												},
																																												End: ast.Location{
																																													Line: int(658),
																																													Column: int(20),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.LiteralString{
																																										Value: "Format required number at ",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(657),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(657),
																																													Column: int(45),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									OpFodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(16),
																																										},
																																									},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(657),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(658),
																																												Column: int(20),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(657),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(658),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(657),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(658),
																																										Column: int(47),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(657),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(658),
																																									Column: int(47),
																																								},
																																							},
																																						},
																																					},
																																					BranchFalse: &ast.Local{
																																						Binds: ast.LocalBinds{
																																							ast.LocalBind{
																																								VarFodder: ast.Fodder{},
																																								Body: &ast.Apply{
																																									Target: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "std",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(660),
																																														Column: int(28),
																																													},
																																													End: ast.Location{
																																														Line: int(660),
																																														Column: int(31),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "floor",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7160,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(660),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(660),
																																													Column: int(37),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													Right: &ast.Apply{
																																														Target: &ast.Index{
																																															Target: &ast.Var{
																																																Id: "std",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{
																																																		"std",
																																																	},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(660),
																																																			Column: int(62),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(660),
																																																			Column: int(65),
																																																		},
																																																	},
																																																},
																																															},
																																															Index: &ast.LiteralString{
																																																Value: "log",
																																																BlockIndent: "",
																																																BlockTermIndent: "",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: nil,
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{},
																																																	LocRange: ast.LocationRange{
																																																		File: nil,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																	},
																																																},
																																																Kind: ast.LiteralStringKind(1),
																																															},
																																															RightBracketFodder: ast.Fodder{},
																																															LeftBracketFodder: ast.Fodder{},
																																															Id: nil,
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7169,
																																																FreeVars: ast.Identifiers{
																																																	"std",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(660),
																																																		Column: int(62),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(660),
																																																		Column: int(69),
																																																	},
																																																},
																																															},
																																														},
																																														FodderLeft: ast.Fodder{},
																																														Arguments: ast.Arguments{
																																															Positional: []ast.CommaSeparatedExpr{
																																																ast.CommaSeparatedExpr{
																																																	Expr: &ast.LiteralNumber{
																																																		OriginalString: "10",
																																																		NodeBase: ast.NodeBase{
																																																			Fodder: ast.Fodder{},
																																																			Ctx: p7173,
																																																			FreeVars: ast.Identifiers{},
																																																			LocRange: ast.LocationRange{
																																																				File: p8,
																																																				FileName: "",
																																																				Begin: ast.Location{
																																																					Line: int(660),
																																																					Column: int(70),
																																																				},
																																																				End: ast.Location{
																																																					Line: int(660),
																																																					Column: int(72),
																																																				},
																																																			},
																																																		},
																																																	},
																																																	CommaFodder: nil,
																																																},
																																															},
																																															Named: nil,
																																														},
																																														FodderRight: ast.Fodder{},
																																														TailStrictFodder: nil,
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7169,
																																															FreeVars: ast.Identifiers{
																																																"std",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(660),
																																																	Column: int(62),
																																																},
																																																End: ast.Location{
																																																	Line: int(660),
																																																	Column: int(73),
																																																},
																																															},
																																														},
																																														TrailingComma: false,
																																														TailStrict: false,
																																													},
																																													Left: &ast.Apply{
																																														Target: &ast.Index{
																																															Target: &ast.Var{
																																																Id: "std",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{
																																																		"std",
																																																	},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(660),
																																																			Column: int(38),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(660),
																																																			Column: int(41),
																																																		},
																																																	},
																																																},
																																															},
																																															Index: &ast.LiteralString{
																																																Value: "log",
																																																BlockIndent: "",
																																																BlockTermIndent: "",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: nil,
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{},
																																																	LocRange: ast.LocationRange{
																																																		File: nil,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																	},
																																																},
																																																Kind: ast.LiteralStringKind(1),
																																															},
																																															RightBracketFodder: ast.Fodder{},
																																															LeftBracketFodder: ast.Fodder{},
																																															Id: nil,
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7169,
																																																FreeVars: ast.Identifiers{
																																																	"std",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(660),
																																																		Column: int(38),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(660),
																																																		Column: int(45),
																																																	},
																																																},
																																															},
																																														},
																																														FodderLeft: ast.Fodder{},
																																														Arguments: ast.Arguments{
																																															Positional: []ast.CommaSeparatedExpr{
																																																ast.CommaSeparatedExpr{
																																																	Expr: &ast.Apply{
																																																		Target: &ast.Index{
																																																			Target: &ast.Var{
																																																				Id: "std",
																																																				NodeBase: ast.NodeBase{
																																																					Fodder: ast.Fodder{},
																																																					Ctx: nil,
																																																					FreeVars: ast.Identifiers{
																																																						"std",
																																																					},
																																																					LocRange: ast.LocationRange{
																																																						File: p8,
																																																						FileName: "",
																																																						Begin: ast.Location{
																																																							Line: int(660),
																																																							Column: int(46),
																																																						},
																																																						End: ast.Location{
																																																							Line: int(660),
																																																							Column: int(49),
																																																						},
																																																					},
																																																				},
																																																			},
																																																			Index: &ast.LiteralString{
																																																				Value: "abs",
																																																				BlockIndent: "",
																																																				BlockTermIndent: "",
																																																				NodeBase: ast.NodeBase{
																																																					Fodder: nil,
																																																					Ctx: nil,
																																																					FreeVars: ast.Identifiers{},
																																																					LocRange: ast.LocationRange{
																																																						File: nil,
																																																						FileName: "",
																																																						Begin: ast.Location{
																																																							Line: int(0),
																																																							Column: int(0),
																																																						},
																																																						End: ast.Location{
																																																							Line: int(0),
																																																							Column: int(0),
																																																						},
																																																					},
																																																				},
																																																				Kind: ast.LiteralStringKind(1),
																																																			},
																																																			RightBracketFodder: ast.Fodder{},
																																																			LeftBracketFodder: ast.Fodder{},
																																																			Id: nil,
																																																			NodeBase: ast.NodeBase{
																																																				Fodder: ast.Fodder{},
																																																				Ctx: p7187,
																																																				FreeVars: ast.Identifiers{
																																																					"std",
																																																				},
																																																				LocRange: ast.LocationRange{
																																																					File: p8,
																																																					FileName: "",
																																																					Begin: ast.Location{
																																																						Line: int(660),
																																																						Column: int(46),
																																																					},
																																																					End: ast.Location{
																																																						Line: int(660),
																																																						Column: int(53),
																																																					},
																																																				},
																																																			},
																																																		},
																																																		FodderLeft: ast.Fodder{},
																																																		Arguments: ast.Arguments{
																																																			Positional: []ast.CommaSeparatedExpr{
																																																				ast.CommaSeparatedExpr{
																																																					Expr: &ast.Var{
																																																						Id: "val",
																																																						NodeBase: ast.NodeBase{
																																																							Fodder: ast.Fodder{},
																																																							Ctx: p7191,
																																																							FreeVars: ast.Identifiers{
																																																								"val",
																																																							},
																																																							LocRange: ast.LocationRange{
																																																								File: p8,
																																																								FileName: "",
																																																								Begin: ast.Location{
																																																									Line: int(660),
																																																									Column: int(54),
																																																								},
																																																								End: ast.Location{
																																																									Line: int(660),
																																																									Column: int(57),
																																																								},
																																																							},
																																																						},
																																																					},
																																																					CommaFodder: nil,
																																																				},
																																																			},
																																																			Named: nil,
																																																		},
																																																		FodderRight: ast.Fodder{},
																																																		TailStrictFodder: nil,
																																																		NodeBase: ast.NodeBase{
																																																			Fodder: ast.Fodder{},
																																																			Ctx: p7187,
																																																			FreeVars: ast.Identifiers{
																																																				"std",
																																																				"val",
																																																			},
																																																			LocRange: ast.LocationRange{
																																																				File: p8,
																																																				FileName: "",
																																																				Begin: ast.Location{
																																																					Line: int(660),
																																																					Column: int(46),
																																																				},
																																																				End: ast.Location{
																																																					Line: int(660),
																																																					Column: int(58),
																																																				},
																																																			},
																																																		},
																																																		TrailingComma: false,
																																																		TailStrict: false,
																																																	},
																																																	CommaFodder: nil,
																																																},
																																															},
																																															Named: nil,
																																														},
																																														FodderRight: ast.Fodder{},
																																														TailStrictFodder: nil,
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7169,
																																															FreeVars: ast.Identifiers{
																																																"std",
																																																"val",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(660),
																																																	Column: int(38),
																																																},
																																																End: ast.Location{
																																																	Line: int(660),
																																																	Column: int(59),
																																																},
																																															},
																																														},
																																														TrailingComma: false,
																																														TailStrict: false,
																																													},
																																													OpFodder: ast.Fodder{},
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7169,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																															"val",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(660),
																																																Column: int(38),
																																															},
																																															End: ast.Location{
																																																Line: int(660),
																																																Column: int(73),
																																															},
																																														},
																																													},
																																													Op: ast.BinaryOp(1),
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7160,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(660),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(660),
																																												Column: int(74),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								EqFodder: ast.Fodder{},
																																								Variable: "exponent",
																																								CloseFodder: ast.Fodder{},
																																								Fun: nil,
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(660),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(660),
																																										Column: int(74),
																																									},
																																								},
																																							},
																																						},
																																						Body: &ast.Conditional{
																																							Cond: &ast.Binary{
																																								Right: &ast.Binary{
																																									Right: &ast.Var{
																																										Id: "fpprec",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"fpprec",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(661),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(661),
																																													Column: int(49),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Var{
																																										Id: "exponent",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"exponent",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(661),
																																													Column: int(31),
																																												},
																																												End: ast.Location{
																																													Line: int(661),
																																													Column: int(39),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"exponent",
																																											"fpprec",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(661),
																																												Column: int(31),
																																											},
																																											End: ast.Location{
																																												Line: int(661),
																																												Column: int(49),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(8),
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.Unary{
																																										Expr: &ast.LiteralNumber{
																																											OriginalString: "4",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6588,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(661),
																																														Column: int(26),
																																													},
																																													End: ast.Location{
																																														Line: int(661),
																																														Column: int(27),
																																													},
																																												},
																																											},
																																										},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(661),
																																													Column: int(25),
																																												},
																																												End: ast.Location{
																																													Line: int(661),
																																													Column: int(27),
																																												},
																																											},
																																										},
																																										Op: ast.UnaryOp(3),
																																									},
																																									Left: &ast.Var{
																																										Id: "exponent",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"exponent",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(661),
																																													Column: int(14),
																																												},
																																												End: ast.Location{
																																													Line: int(661),
																																													Column: int(22),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"exponent",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(661),
																																												Column: int(14),
																																											},
																																											End: ast.Location{
																																												Line: int(661),
																																												Column: int(27),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(9),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"exponent",
																																										"fpprec",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(661),
																																											Column: int(14),
																																										},
																																										End: ast.Location{
																																											Line: int(661),
																																											Column: int(49),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(18),
																																							},
																																							BranchTrue: &ast.Apply{
																																								Target: &ast.Var{
																																									Id: "render_float_sci",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																											},
																																										},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"render_float_sci",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(662),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(662),
																																												Column: int(29),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "val",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(662),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(662),
																																															Column: int(33),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "zp",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{
																																														ast.FodderElement{
																																															Comment: []string{},
																																															Kind: ast.FodderKind(0),
																																															Blanks: int(0),
																																															Indent: int(29),
																																														},
																																													},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"zp",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(663),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(663),
																																															Column: int(32),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "cflags",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(664),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(664),
																																																Column: int(36),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "blank",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(664),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(664),
																																															Column: int(42),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "cflags",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(665),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(665),
																																																Column: int(36),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "plus",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(665),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(665),
																																															Column: int(41),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "cflags",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(666),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(666),
																																																Column: int(36),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "alt",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(666),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(666),
																																															Column: int(40),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "cflags",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(667),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(667),
																																																Column: int(36),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "alt",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(667),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(667),
																																															Column: int(40),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "code",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"code",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(668),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(668),
																																																Column: int(34),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "caps",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"code",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(668),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(668),
																																															Column: int(39),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Binary{
																																												Right: &ast.LiteralNumber{
																																													OriginalString: "1",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7218,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(669),
																																																Column: int(39),
																																															},
																																															End: ast.Location{
																																																Line: int(669),
																																																Column: int(40),
																																															},
																																														},
																																													},
																																												},
																																												Left: &ast.Var{
																																													Id: "fpprec",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: p7218,
																																														FreeVars: ast.Identifiers{
																																															"fpprec",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(669),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(669),
																																																Column: int(36),
																																															},
																																														},
																																													},
																																												},
																																												OpFodder: ast.Fodder{},
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7218,
																																													FreeVars: ast.Identifiers{
																																														"fpprec",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(669),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(669),
																																															Column: int(40),
																																														},
																																													},
																																												},
																																												Op: ast.BinaryOp(4),
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																										"code",
																																										"fpprec",
																																										"render_float_sci",
																																										"val",
																																										"zp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(662),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(669),
																																											Column: int(41),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							BranchFalse: &ast.Local{
																																								Binds: ast.LocalBinds{
																																									ast.LocalBind{
																																										VarFodder: ast.Fodder{},
																																										Body: &ast.Apply{
																																											Target: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "std",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(671),
																																																Column: int(38),
																																															},
																																															End: ast.Location{
																																																Line: int(671),
																																																Column: int(41),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "max",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7267,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(671),
																																															Column: int(38),
																																														},
																																														End: ast.Location{
																																															Line: int(671),
																																															Column: int(45),
																																														},
																																													},
																																												},
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.LiteralNumber{
																																															OriginalString: "1",
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7271,
																																																FreeVars: ast.Identifiers{},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(671),
																																																		Column: int(46),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(671),
																																																		Column: int(47),
																																																	},
																																																},
																																															},
																																														},
																																														CommaFodder: ast.Fodder{},
																																													},
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Binary{
																																															Right: &ast.LiteralNumber{
																																																OriginalString: "1",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p7271,
																																																	FreeVars: ast.Identifiers{},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(671),
																																																			Column: int(60),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(671),
																																																			Column: int(61),
																																																		},
																																																	},
																																																},
																																															},
																																															Left: &ast.Var{
																																																Id: "exponent",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p7271,
																																																	FreeVars: ast.Identifiers{
																																																		"exponent",
																																																	},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(671),
																																																			Column: int(49),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(671),
																																																			Column: int(57),
																																																		},
																																																	},
																																																},
																																															},
																																															OpFodder: ast.Fodder{},
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7271,
																																																FreeVars: ast.Identifiers{
																																																	"exponent",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(671),
																																																		Column: int(49),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(671),
																																																		Column: int(61),
																																																	},
																																																},
																																															},
																																															Op: ast.BinaryOp(3),
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7267,
																																												FreeVars: ast.Identifiers{
																																													"exponent",
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(671),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(671),
																																														Column: int(62),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																										},
																																										EqFodder: ast.Fodder{},
																																										Variable: "digits_before_pt",
																																										CloseFodder: ast.Fodder{},
																																										Fun: nil,
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(671),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(671),
																																												Column: int(62),
																																											},
																																										},
																																									},
																																								},
																																								Body: &ast.Apply{
																																									Target: &ast.Var{
																																										Id: "render_float_dec",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(12),
																																												},
																																											},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"render_float_dec",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(672),
																																													Column: int(13),
																																												},
																																												End: ast.Location{
																																													Line: int(672),
																																													Column: int(29),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													Id: "val",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(672),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(672),
																																																Column: int(33),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													Id: "zp",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Comment: []string{},
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																															},
																																														},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"zp",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(673),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(673),
																																																Column: int(32),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "cflags",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Comment: []string{},
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(674),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(674),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "blank",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(674),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(674),
																																																Column: int(42),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "cflags",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Comment: []string{},
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(675),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(675),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "plus",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(675),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(675),
																																																Column: int(41),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "cflags",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Comment: []string{},
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(676),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(676),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "alt",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(676),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(676),
																																																Column: int(40),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "cflags",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Comment: []string{},
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(677),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(677),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "alt",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(677),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(677),
																																																Column: int(40),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													Right: &ast.Var{
																																														Id: "digits_before_pt",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7284,
																																															FreeVars: ast.Identifiers{
																																																"digits_before_pt",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(678),
																																																	Column: int(39),
																																																},
																																																End: ast.Location{
																																																	Line: int(678),
																																																	Column: int(55),
																																																},
																																															},
																																														},
																																													},
																																													Left: &ast.Var{
																																														Id: "fpprec",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Comment: []string{},
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																},
																																															},
																																															Ctx: p7284,
																																															FreeVars: ast.Identifiers{
																																																"fpprec",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(678),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(678),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													OpFodder: ast.Fodder{},
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7284,
																																														FreeVars: ast.Identifiers{
																																															"digits_before_pt",
																																															"fpprec",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(678),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(678),
																																																Column: int(55),
																																															},
																																														},
																																													},
																																													Op: ast.BinaryOp(4),
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																											"digits_before_pt",
																																											"fpprec",
																																											"render_float_dec",
																																											"val",
																																											"zp",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(672),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(678),
																																												Column: int(56),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(12),
																																										},
																																									},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																										"exponent",
																																										"fpprec",
																																										"render_float_dec",
																																										"std",
																																										"val",
																																										"zp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(671),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(678),
																																											Column: int(56),
																																										},
																																									},
																																								},
																																							},
																																							ThenFodder: ast.Fodder{},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(10),
																																									},
																																								},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																									"code",
																																									"exponent",
																																									"fpprec",
																																									"render_float_dec",
																																									"render_float_sci",
																																									"std",
																																									"val",
																																									"zp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(661),
																																										Column: int(11),
																																									},
																																									End: ast.Location{
																																										Line: int(678),
																																										Column: int(56),
																																									},
																																								},
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																								"code",
																																								"fpprec",
																																								"render_float_dec",
																																								"render_float_sci",
																																								"std",
																																								"val",
																																								"zp",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(660),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(678),
																																									Column: int(56),
																																								},
																																							},
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(8),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(8),
																																							},
																																						},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																							"code",
																																							"fpprec",
																																							"i",
																																							"render_float_dec",
																																							"render_float_sci",
																																							"std",
																																							"val",
																																							"zp",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(656),
																																								Column: int(9),
																																							},
																																							End: ast.Location{
																																								Line: int(678),
																																								Column: int(56),
																																							},
																																						},
																																					},
																																				},
																																				BranchFalse: &ast.Conditional{
																																					Cond: &ast.Binary{
																																						Right: &ast.LiteralString{
																																							Value: "c",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(679),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(679),
																																										Column: int(32),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						Left: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "code",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(679),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(679),
																																											Column: int(19),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "ctype",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(679),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(679),
																																										Column: int(25),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(679),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(679),
																																									Column: int(32),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(12),
																																					},
																																					BranchTrue: &ast.Conditional{
																																						Cond: &ast.Binary{
																																							Right: &ast.LiteralString{
																																								Value: "number",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(680),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(680),
																																											Column: int(37),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							Left: &ast.Apply{
																																								Target: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "std",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(680),
																																													Column: int(12),
																																												},
																																												End: ast.Location{
																																													Line: int(680),
																																													Column: int(15),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "type",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(680),
																																												Column: int(12),
																																											},
																																											End: ast.Location{
																																												Line: int(680),
																																												Column: int(20),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "val",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7351,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(680),
																																															Column: int(21),
																																														},
																																														End: ast.Location{
																																															Line: int(680),
																																															Column: int(24),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(680),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(680),
																																											Column: int(25),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(680),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(680),
																																										Column: int(37),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(12),
																																						},
																																						BranchTrue: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(10),
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(681),
																																												Column: int(11),
																																											},
																																											End: ast.Location{
																																												Line: int(681),
																																												Column: int(14),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "char",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(681),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(681),
																																											Column: int(19),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "val",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7364,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(681),
																																														Column: int(20),
																																													},
																																													End: ast.Location{
																																														Line: int(681),
																																														Column: int(23),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(681),
																																										Column: int(11),
																																									},
																																									End: ast.Location{
																																										Line: int(681),
																																										Column: int(24),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						BranchFalse: &ast.Conditional{
																																							Cond: &ast.Binary{
																																								Right: &ast.LiteralString{
																																									Value: "string",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(682),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(682),
																																												Column: int(42),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								Left: &ast.Apply{
																																									Target: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "std",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(682),
																																														Column: int(17),
																																													},
																																													End: ast.Location{
																																														Line: int(682),
																																														Column: int(20),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "type",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(682),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(682),
																																													Column: int(25),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													Id: "val",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7378,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(682),
																																																Column: int(26),
																																															},
																																															End: ast.Location{
																																																Line: int(682),
																																																Column: int(29),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(682),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(682),
																																												Column: int(30),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(682),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(682),
																																											Column: int(42),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(12),
																																							},
																																							BranchTrue: &ast.Conditional{
																																								Cond: &ast.Binary{
																																									Right: &ast.LiteralNumber{
																																										OriginalString: "1",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(683),
																																													Column: int(33),
																																												},
																																												End: ast.Location{
																																													Line: int(683),
																																													Column: int(34),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Apply{
																																										Target: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "std",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(683),
																																															Column: int(14),
																																														},
																																														End: ast.Location{
																																															Line: int(683),
																																															Column: int(17),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.LiteralString{
																																												Value: "length",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: nil,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6588,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(683),
																																														Column: int(14),
																																													},
																																													End: ast.Location{
																																														Line: int(683),
																																														Column: int(24),
																																													},
																																												},
																																											},
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														Id: "val",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7393,
																																															FreeVars: ast.Identifiers{
																																																"val",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(683),
																																																	Column: int(25),
																																																},
																																																End: ast.Location{
																																																	Line: int(683),
																																																	Column: int(28),
																																																},
																																															},
																																														},
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(683),
																																													Column: int(14),
																																												},
																																												End: ast.Location{
																																													Line: int(683),
																																													Column: int(29),
																																												},
																																											},
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(683),
																																												Column: int(14),
																																											},
																																											End: ast.Location{
																																												Line: int(683),
																																												Column: int(34),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(12),
																																								},
																																								BranchTrue: &ast.Var{
																																									Id: "val",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																											},
																																										},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(684),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(684),
																																												Column: int(16),
																																											},
																																										},
																																									},
																																								},
																																								BranchFalse: &ast.Error{
																																									Expr: &ast.Binary{
																																										Right: &ast.Apply{
																																											Target: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "std",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(686),
																																																Column: int(56),
																																															},
																																															End: ast.Location{
																																																Line: int(686),
																																																Column: int(59),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "length",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p6588,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(686),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(686),
																																															Column: int(66),
																																														},
																																													},
																																												},
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															Id: "val",
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7410,
																																																FreeVars: ast.Identifiers{
																																																	"val",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(686),
																																																		Column: int(67),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(686),
																																																		Column: int(70),
																																																	},
																																																},
																																															},
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6588,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(686),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(686),
																																														Column: int(71),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																										},
																																										Left: &ast.LiteralString{
																																											Value: "%c expected 1-sized string got: ",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6588,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(686),
																																														Column: int(19),
																																													},
																																													End: ast.Location{
																																														Line: int(686),
																																														Column: int(53),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(686),
																																													Column: int(19),
																																												},
																																												End: ast.Location{
																																													Line: int(686),
																																													Column: int(71),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																											},
																																										},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(686),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(686),
																																												Column: int(71),
																																											},
																																										},
																																									},
																																								},
																																								ThenFodder: ast.Fodder{},
																																								ElseFodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(10),
																																									},
																																								},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(10),
																																										},
																																									},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(683),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(686),
																																											Column: int(71),
																																										},
																																									},
																																								},
																																							},
																																							BranchFalse: &ast.Error{
																																								Expr: &ast.Binary{
																																									Right: &ast.Apply{
																																										Target: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "std",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(688),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(688),
																																															Column: int(59),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.LiteralString{
																																												Value: "type",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: nil,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p6588,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(688),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(688),
																																														Column: int(64),
																																													},
																																												},
																																											},
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														Id: "val",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7430,
																																															FreeVars: ast.Identifiers{
																																																"val",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(688),
																																																	Column: int(65),
																																																},
																																																End: ast.Location{
																																																	Line: int(688),
																																																	Column: int(68),
																																																},
																																															},
																																														},
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(688),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(688),
																																													Column: int(69),
																																												},
																																											},
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																									},
																																									Left: &ast.LiteralString{
																																										Value: "%c expected number / string, got: ",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p6588,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(688),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(688),
																																													Column: int(53),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p6588,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(688),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(688),
																																												Column: int(69),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(10),
																																										},
																																									},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(688),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(688),
																																											Column: int(69),
																																										},
																																									},
																																								},
																																							},
																																							ThenFodder: ast.Fodder{},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																								},
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(682),
																																										Column: int(14),
																																									},
																																									End: ast.Location{
																																										Line: int(688),
																																										Column: int(69),
																																									},
																																								},
																																							},
																																						},
																																						ThenFodder: ast.Fodder{},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(8),
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																								},
																																							},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(680),
																																									Column: int(9),
																																								},
																																								End: ast.Location{
																																									Line: int(688),
																																									Column: int(69),
																																								},
																																							},
																																						},
																																					},
																																					BranchFalse: &ast.Error{
																																						Expr: &ast.Binary{
																																							Right: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "code",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(690),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(690),
																																												Column: int(38),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "ctype",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(690),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(690),
																																											Column: int(44),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								Value: "Unknown code: ",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p6588,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(690),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(690),
																																											Column: int(31),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p6588,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(690),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(690),
																																										Column: int(44),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																								},
																																							},
																																							Ctx: p6588,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(690),
																																									Column: int(9),
																																								},
																																								End: ast.Location{
																																									Line: int(690),
																																									Column: int(44),
																																								},
																																							},
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(6),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p6588,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																							"std",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(679),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(690),
																																								Column: int(44),
																																							},
																																						},
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(6),
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p6588,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"code",
																																						"fpprec",
																																						"i",
																																						"render_float_dec",
																																						"render_float_sci",
																																						"std",
																																						"val",
																																						"zp",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(655),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(690),
																																							Column: int(44),
																																						},
																																					},
																																				},
																																			},
																																			ThenFodder: ast.Fodder{},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(6),
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p6588,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"code",
																																					"fpprec",
																																					"i",
																																					"render_float_dec",
																																					"render_float_sci",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(642),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(690),
																																						Column: int(44),
																																					},
																																				},
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(6),
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p6588,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"code",
																																				"fpprec",
																																				"i",
																																				"render_float_dec",
																																				"render_float_sci",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(630),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(690),
																																					Column: int(44),
																																				},
																																			},
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(6),
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p6588,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"code",
																																			"fpprec",
																																			"i",
																																			"iprec",
																																			"render_float_dec",
																																			"render_float_sci",
																																			"render_hex",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(618),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(690),
																																				Column: int(44),
																																			},
																																		},
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																ElseFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(6),
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p6588,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"code",
																																		"fpprec",
																																		"i",
																																		"iprec",
																																		"render_float_dec",
																																		"render_float_sci",
																																		"render_hex",
																																		"render_int",
																																		"std",
																																		"val",
																																		"zp",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(611),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(690),
																																			Column: int(44),
																																		},
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															ElseFodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(6),
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p6588,
																																FreeVars: ast.Identifiers{
																																	"cflags",
																																	"code",
																																	"fpprec",
																																	"i",
																																	"iprec",
																																	"render_float_dec",
																																	"render_float_sci",
																																	"render_hex",
																																	"render_int",
																																	"std",
																																	"val",
																																	"zp",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(605),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(690),
																																		Column: int(44),
																																	},
																																},
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(6),
																																},
																															},
																															Ctx: p6588,
																															FreeVars: ast.Identifiers{
																																"cflags",
																																"code",
																																"fpprec",
																																"i",
																																"iprec",
																																"render_float_dec",
																																"render_float_sci",
																																"render_hex",
																																"render_int",
																																"std",
																																"val",
																																"zp",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(603),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(690),
																																	Column: int(44),
																																},
																															},
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														Ctx: p6588,
																														FreeVars: ast.Identifiers{
																															"cflags",
																															"code",
																															"fpprec",
																															"fw",
																															"i",
																															"iprec",
																															"render_float_dec",
																															"render_float_sci",
																															"render_hex",
																															"render_int",
																															"std",
																															"val",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(602),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(690),
																																Column: int(44),
																															},
																														},
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p6588,
																													FreeVars: ast.Identifiers{
																														"cflags",
																														"code",
																														"fpprec",
																														"fw",
																														"i",
																														"prec_or_null",
																														"render_float_dec",
																														"render_float_sci",
																														"render_hex",
																														"render_int",
																														"std",
																														"val",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(601),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(690),
																															Column: int(44),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p6588,
																												FreeVars: ast.Identifiers{
																													"cflags",
																													"code",
																													"fw",
																													"i",
																													"prec_or_null",
																													"render_float_dec",
																													"render_float_sci",
																													"render_hex",
																													"render_int",
																													"std",
																													"val",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(600),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(690),
																														Column: int(44),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											Ctx: p6588,
																											FreeVars: ast.Identifiers{
																												"code",
																												"fw",
																												"i",
																												"prec_or_null",
																												"render_float_dec",
																												"render_float_sci",
																												"render_hex",
																												"render_int",
																												"std",
																												"val",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(599),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(690),
																													Column: int(44),
																												},
																											},
																										},
																									},
																									Parameters: []ast.Parameter{
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "val",
																											CommaFodder: ast.Fodder{},
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(598),
																													Column: int(23),
																												},
																												End: ast.Location{
																													Line: int(598),
																													Column: int(26),
																												},
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "code",
																											CommaFodder: ast.Fodder{},
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(598),
																													Column: int(28),
																												},
																												End: ast.Location{
																													Line: int(598),
																													Column: int(32),
																												},
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "fw",
																											CommaFodder: ast.Fodder{},
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(598),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(598),
																													Column: int(36),
																												},
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "prec_or_null",
																											CommaFodder: ast.Fodder{},
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(598),
																													Column: int(38),
																												},
																												End: ast.Location{
																													Line: int(598),
																													Column: int(50),
																												},
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "i",
																											CommaFodder: nil,
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(598),
																													Column: int(52),
																												},
																												End: ast.Location{
																													Line: int(598),
																													Column: int(53),
																												},
																											},
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: p7479,
																										FreeVars: ast.Identifiers{
																											"render_float_dec",
																											"render_float_sci",
																											"render_hex",
																											"render_int",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(598),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(690),
																												Column: int(44),
																											},
																										},
																									},
																									TrailingComma: false,
																								},
																								EqFodder: nil,
																								Variable: "format_code",
																								CloseFodder: nil,
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																						},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: nil,
																									Body: &ast.Function{
																										ParenLeftFodder: ast.Fodder{},
																										ParenRightFodder: ast.Fodder{},
																										Body: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(694),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(694),
																																		Column: int(18),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "length",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(694),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(694),
																																	Column: int(25),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "codes",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p7495,
																																		FreeVars: ast.Identifiers{
																																			"codes",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(694),
																																				Column: int(26),
																																			},
																																			End: ast.Location{
																																				Line: int(694),
																																				Column: int(31),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(694),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(694),
																																Column: int(32),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												Left: &ast.Var{
																													Id: "i",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(694),
																																Column: int(10),
																															},
																															End: ast.Location{
																																Line: int(694),
																																Column: int(11),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p7491,
																													FreeVars: ast.Identifiers{
																														"codes",
																														"i",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(694),
																															Column: int(10),
																														},
																														End: ast.Location{
																															Line: int(694),
																															Column: int(32),
																														},
																													},
																												},
																												Op: ast.BinaryOp(8),
																											},
																											BranchTrue: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(695),
																																			Column: int(16),
																																		},
																																		End: ast.Location{
																																			Line: int(695),
																																			Column: int(19),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "length",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(695),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(695),
																																		Column: int(26),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "arr",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7511,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(695),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(695),
																																					Column: int(30),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(695),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(695),
																																	Column: int(31),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													Left: &ast.Var{
																														Id: "j",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(695),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(695),
																																	Column: int(13),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"j",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(695),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(695),
																																Column: int(31),
																															},
																														},
																													},
																													Op: ast.BinaryOp(9),
																												},
																												BranchTrue: &ast.Error{
																													Expr: &ast.Binary{
																														Right: &ast.Var{
																															Id: "j",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(696),
																																		Column: int(84),
																																	},
																																	End: ast.Location{
																																		Line: int(696),
																																		Column: int(85),
																																	},
																																},
																															},
																														},
																														Left: &ast.Binary{
																															Right: &ast.LiteralString{
																																Value: ", expected ",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p7491,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(696),
																																			Column: int(68),
																																		},
																																		End: ast.Location{
																																			Line: int(696),
																																			Column: int(81),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Left: &ast.Binary{
																																Right: &ast.Apply{
																																	Target: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "std",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(696),
																																						Column: int(50),
																																					},
																																					End: ast.Location{
																																						Line: int(696),
																																						Column: int(53),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "length",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7491,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(696),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(696),
																																					Column: int(60),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "arr",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7532,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(696),
																																								Column: int(61),
																																							},
																																							End: ast.Location{
																																								Line: int(696),
																																								Column: int(64),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p7491,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(696),
																																				Column: int(50),
																																			},
																																			End: ast.Location{
																																				Line: int(696),
																																				Column: int(65),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																Left: &ast.LiteralString{
																																	Value: "Too many values to format: ",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p7491,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(696),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(696),
																																				Column: int(47),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p7491,
																																	FreeVars: ast.Identifiers{
																																		"arr",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(696),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(696),
																																			Column: int(65),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"arr",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(696),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(696),
																																		Column: int(81),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"j",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(696),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(696),
																																	Column: int(85),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																															},
																														},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"j",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(696),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(696),
																																Column: int(86),
																															},
																														},
																													},
																												},
																												BranchFalse: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																															},
																														},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(698),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(698),
																																Column: int(12),
																															},
																														},
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													Ctx: p7491,
																													FreeVars: ast.Identifiers{
																														"arr",
																														"j",
																														"std",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(695),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(698),
																															Column: int(12),
																														},
																													},
																												},
																											},
																											BranchFalse: &ast.Local{
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Body: &ast.Index{
																															Target: &ast.Var{
																																Id: "codes",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p7551,
																																	FreeVars: ast.Identifiers{
																																		"codes",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(700),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(700),
																																			Column: int(27),
																																		},
																																	},
																																},
																															},
																															Index: &ast.Var{
																																Id: "i",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p7551,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(700),
																																			Column: int(28),
																																		},
																																		End: ast.Location{
																																			Line: int(700),
																																			Column: int(29),
																																		},
																																	},
																																},
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7551,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																	"i",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(700),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(700),
																																		Column: int(30),
																																	},
																																},
																															},
																														},
																														EqFodder: ast.Fodder{},
																														Variable: "code",
																														CloseFodder: ast.Fodder{},
																														Fun: nil,
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(700),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(700),
																																Column: int(30),
																															},
																														},
																													},
																												},
																												Body: &ast.Conditional{
																													Cond: &ast.Binary{
																														Right: &ast.LiteralString{
																															Value: "string",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(701),
																																		Column: int(30),
																																	},
																																	End: ast.Location{
																																		Line: int(701),
																																		Column: int(38),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Left: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(701),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(701),
																																				Column: int(15),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "type",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p7491,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(701),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(701),
																																			Column: int(20),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "code",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7567,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(701),
																																						Column: int(21),
																																					},
																																					End: ast.Location{
																																						Line: int(701),
																																						Column: int(25),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(701),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(701),
																																		Column: int(26),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"code",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(701),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(701),
																																	Column: int(38),
																																},
																															},
																														},
																														Op: ast.BinaryOp(12),
																													},
																													BranchTrue: &ast.Apply{
																														Target: &ast.Var{
																															Id: "format_codes_arr",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																	},
																																},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"format_codes_arr",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(702),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(702),
																																		Column: int(27),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "codes",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7577,
																																			FreeVars: ast.Identifiers{
																																				"codes",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(702),
																																					Column: int(28),
																																				},
																																				End: ast.Location{
																																					Line: int(702),
																																					Column: int(33),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "arr",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7577,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(702),
																																					Column: int(35),
																																				},
																																				End: ast.Location{
																																					Line: int(702),
																																					Column: int(38),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		Right: &ast.LiteralNumber{
																																			OriginalString: "1",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7577,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(702),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(702),
																																						Column: int(45),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "i",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7577,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(702),
																																						Column: int(40),
																																					},
																																					End: ast.Location{
																																						Line: int(702),
																																						Column: int(41),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7577,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(702),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(702),
																																					Column: int(45),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(3),
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "j",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7577,
																																			FreeVars: ast.Identifiers{
																																				"j",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(702),
																																					Column: int(47),
																																				},
																																				End: ast.Location{
																																					Line: int(702),
																																					Column: int(48),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		Right: &ast.Var{
																																			Id: "code",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7577,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(702),
																																						Column: int(54),
																																					},
																																					End: ast.Location{
																																						Line: int(702),
																																						Column: int(58),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Var{
																																			Id: "v",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7577,
																																				FreeVars: ast.Identifiers{
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(702),
																																						Column: int(50),
																																					},
																																					End: ast.Location{
																																						Line: int(702),
																																						Column: int(51),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7577,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(702),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(702),
																																					Column: int(58),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(3),
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"code",
																																"codes",
																																"format_codes_arr",
																																"i",
																																"j",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(702),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(702),
																																	Column: int(59),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: true,
																													},
																													BranchFalse: &ast.Local{
																														Binds: ast.LocalBinds{
																															ast.LocalBind{
																																VarFodder: ast.Fodder{},
																																Body: &ast.Conditional{
																																	Cond: &ast.Binary{
																																		Right: &ast.LiteralString{
																																			Value: "*",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7600,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(704),
																																						Column: int(37),
																																					},
																																					End: ast.Location{
																																						Line: int(704),
																																						Column: int(40),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		Left: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "code",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(704),
																																							Column: int(26),
																																						},
																																						End: ast.Location{
																																							Line: int(704),
																																							Column: int(30),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "fw",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7600,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(704),
																																						Column: int(26),
																																					},
																																					End: ast.Location{
																																						Line: int(704),
																																						Column: int(33),
																																					},
																																				},
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7600,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(704),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(704),
																																					Column: int(40),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(12),
																																	},
																																	BranchTrue: &ast.DesugaredObject{
																																		Asserts: ast.Nodes{},
																																		Fields: ast.DesugaredObjectFields{
																																			ast.DesugaredObjectField{
																																				Name: &ast.LiteralString{
																																					Value: "j",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Body: &ast.Binary{
																																					Right: &ast.LiteralNumber{
																																						OriginalString: "1",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(705),
																																									Column: int(20),
																																								},
																																								End: ast.Location{
																																									Line: int(705),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					Left: &ast.Var{
																																						Id: "j",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"j",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(705),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(705),
																																									Column: int(17),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7612,
																																						FreeVars: ast.Identifiers{
																																							"j",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(705),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(705),
																																								Column: int(21),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(705),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(705),
																																						Column: int(21),
																																					},
																																				},
																																				Hide: ast.ObjectFieldHide(1),
																																				PlusSuper: false,
																																			},
																																			ast.DesugaredObjectField{
																																				Name: &ast.LiteralString{
																																					Value: "fw",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Body: &ast.Conditional{
																																					Cond: &ast.Binary{
																																						Right: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(706),
																																												Column: int(25),
																																											},
																																											End: ast.Location{
																																												Line: int(706),
																																												Column: int(28),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "length",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7612,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(706),
																																											Column: int(25),
																																										},
																																										End: ast.Location{
																																											Line: int(706),
																																											Column: int(35),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "arr",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7627,
																																												FreeVars: ast.Identifiers{
																																													"arr",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(706),
																																														Column: int(36),
																																													},
																																													End: ast.Location{
																																														Line: int(706),
																																														Column: int(39),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(706),
																																										Column: int(25),
																																									},
																																									End: ast.Location{
																																										Line: int(706),
																																										Column: int(40),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						Left: &ast.Var{
																																							Id: "j",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"j",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(706),
																																										Column: int(20),
																																									},
																																									End: ast.Location{
																																										Line: int(706),
																																										Column: int(21),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(706),
																																									Column: int(20),
																																								},
																																								End: ast.Location{
																																									Line: int(706),
																																									Column: int(40),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(8),
																																					},
																																					BranchTrue: &ast.Error{
																																						Expr: &ast.Binary{
																																							Right: &ast.Var{
																																								Id: "j",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7612,
																																									FreeVars: ast.Identifiers{
																																										"j",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(707),
																																											Column: int(99),
																																										},
																																										End: ast.Location{
																																											Line: int(707),
																																											Column: int(100),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.Binary{
																																								Right: &ast.LiteralString{
																																									Value: ", expected at least ",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7612,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(707),
																																												Column: int(74),
																																											},
																																											End: ast.Location{
																																												Line: int(707),
																																												Column: int(96),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.Apply{
																																										Target: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "std",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(707),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(707),
																																															Column: int(59),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.LiteralString{
																																												Value: "length",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: nil,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7612,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(707),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(707),
																																														Column: int(66),
																																													},
																																												},
																																											},
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														Id: "arr",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p7648,
																																															FreeVars: ast.Identifiers{
																																																"arr",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(707),
																																																	Column: int(67),
																																																},
																																																End: ast.Location{
																																																	Line: int(707),
																																																	Column: int(70),
																																																},
																																															},
																																														},
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7612,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(707),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(707),
																																													Column: int(71),
																																												},
																																											},
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																									},
																																									Left: &ast.LiteralString{
																																										Value: "Not enough values to format: ",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7612,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(707),
																																													Column: int(22),
																																												},
																																												End: ast.Location{
																																													Line: int(707),
																																													Column: int(53),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7612,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(707),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(707),
																																												Column: int(71),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7612,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(707),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(707),
																																											Column: int(96),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"j",
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(707),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(707),
																																										Column: int(100),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																								},
																																							},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(707),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(707),
																																									Column: int(101),
																																								},
																																							},
																																						},
																																					},
																																					BranchFalse: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "arr",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																									},
																																								},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(709),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(709),
																																										Column: int(18),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.Var{
																																							Id: "j",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"j",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(709),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(709),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(709),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(709),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7612,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j",
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(706),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(709),
																																								Column: int(21),
																																							},
																																						},
																																					},
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(706),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(709),
																																						Column: int(21),
																																					},
																																				},
																																				Hide: ast.ObjectFieldHide(1),
																																				PlusSuper: false,
																																			},
																																		},
																																		Locals: ast.LocalBinds{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7600,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"j",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(704),
																																					Column: int(46),
																																				},
																																				End: ast.Location{
																																					Line: int(710),
																																					Column: int(12),
																																				},
																																			},
																																		},
																																	},
																																	BranchFalse: &ast.DesugaredObject{
																																		Asserts: ast.Nodes{},
																																		Fields: ast.DesugaredObjectFields{
																																			ast.DesugaredObjectField{
																																				Name: &ast.LiteralString{
																																					Value: "j",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Body: &ast.Var{
																																					Id: "j",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7671,
																																						FreeVars: ast.Identifiers{
																																							"j",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(711),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(711),
																																								Column: int(17),
																																							},
																																						},
																																					},
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(711),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(711),
																																						Column: int(17),
																																					},
																																				},
																																				Hide: ast.ObjectFieldHide(1),
																																				PlusSuper: false,
																																			},
																																			ast.DesugaredObjectField{
																																				Name: &ast.LiteralString{
																																					Value: "fw",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Body: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "code",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(712),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(712),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "fw",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7671,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(712),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(712),
																																								Column: int(24),
																																							},
																																						},
																																					},
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(712),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(712),
																																						Column: int(24),
																																					},
																																				},
																																				Hide: ast.ObjectFieldHide(1),
																																				PlusSuper: false,
																																			},
																																		},
																																		Locals: ast.LocalBinds{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7600,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"j",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(710),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(713),
																																					Column: int(12),
																																				},
																																			},
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	ElseFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p7600,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"code",
																																			"j",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(704),
																																				Column: int(23),
																																			},
																																			End: ast.Location{
																																				Line: int(713),
																																				Column: int(12),
																																			},
																																		},
																																	},
																																},
																																EqFodder: ast.Fodder{},
																																Variable: "tmp",
																																CloseFodder: ast.Fodder{},
																																Fun: nil,
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(704),
																																		Column: int(17),
																																	},
																																	End: ast.Location{
																																		Line: int(713),
																																		Column: int(12),
																																	},
																																},
																															},
																														},
																														Body: &ast.Local{
																															Binds: ast.LocalBinds{
																																ast.LocalBind{
																																	VarFodder: ast.Fodder{},
																																	Body: &ast.Conditional{
																																		Cond: &ast.Binary{
																																			Right: &ast.LiteralString{
																																				Value: "*",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p7686,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(714),
																																							Column: int(40),
																																						},
																																						End: ast.Location{
																																							Line: int(714),
																																							Column: int(43),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			Left: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "code",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(714),
																																								Column: int(27),
																																							},
																																							End: ast.Location{
																																								Line: int(714),
																																								Column: int(31),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "prec",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p7686,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(714),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(714),
																																							Column: int(36),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7686,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(714),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(714),
																																						Column: int(43),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(12),
																																		},
																																		BranchTrue: &ast.DesugaredObject{
																																			Asserts: ast.Nodes{},
																																			Fields: ast.DesugaredObjectFields{
																																				ast.DesugaredObjectField{
																																					Name: &ast.LiteralString{
																																						Value: "j",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Body: &ast.Binary{
																																						Right: &ast.LiteralNumber{
																																							OriginalString: "1",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7698,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(715),
																																										Column: int(24),
																																									},
																																									End: ast.Location{
																																										Line: int(715),
																																										Column: int(25),
																																									},
																																								},
																																							},
																																						},
																																						Left: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "tmp",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(715),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(715),
																																											Column: int(19),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "j",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7698,
																																								FreeVars: ast.Identifiers{
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(715),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(715),
																																										Column: int(21),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7698,
																																							FreeVars: ast.Identifiers{
																																								"tmp",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(715),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(715),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(715),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(715),
																																							Column: int(25),
																																						},
																																					},
																																					Hide: ast.ObjectFieldHide(1),
																																					PlusSuper: false,
																																				},
																																				ast.DesugaredObjectField{
																																					Name: &ast.LiteralString{
																																						Value: "prec",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Body: &ast.Conditional{
																																						Cond: &ast.Binary{
																																							Right: &ast.Apply{
																																								Target: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "std",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(716),
																																													Column: int(31),
																																												},
																																												End: ast.Location{
																																													Line: int(716),
																																													Column: int(34),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "length",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7698,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(716),
																																												Column: int(31),
																																											},
																																											End: ast.Location{
																																												Line: int(716),
																																												Column: int(41),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "arr",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7716,
																																													FreeVars: ast.Identifiers{
																																														"arr",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(716),
																																															Column: int(42),
																																														},
																																														End: ast.Location{
																																															Line: int(716),
																																															Column: int(45),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7698,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(716),
																																											Column: int(31),
																																										},
																																										End: ast.Location{
																																											Line: int(716),
																																											Column: int(46),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							Left: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "tmp",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(716),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(716),
																																												Column: int(25),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "j",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7698,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(716),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(716),
																																											Column: int(27),
																																										},
																																									},
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7698,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(716),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(716),
																																										Column: int(46),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(8),
																																						},
																																						BranchTrue: &ast.Error{
																																							Expr: &ast.Binary{
																																								Right: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "tmp",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"tmp",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(717),
																																													Column: int(99),
																																												},
																																												End: ast.Location{
																																													Line: int(717),
																																													Column: int(102),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "j",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7698,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(717),
																																												Column: int(99),
																																											},
																																											End: ast.Location{
																																												Line: int(717),
																																												Column: int(104),
																																											},
																																										},
																																									},
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.LiteralString{
																																										Value: ", expected at least ",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7698,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(717),
																																													Column: int(74),
																																												},
																																												End: ast.Location{
																																													Line: int(717),
																																													Column: int(96),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									Left: &ast.Binary{
																																										Right: &ast.Apply{
																																											Target: &ast.Index{
																																												Target: &ast.Var{
																																													Id: "std",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(717),
																																																Column: int(56),
																																															},
																																															End: ast.Location{
																																																Line: int(717),
																																																Column: int(59),
																																															},
																																														},
																																													},
																																												},
																																												Index: &ast.LiteralString{
																																													Value: "length",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: nil,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												LeftBracketFodder: ast.Fodder{},
																																												Id: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p7698,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(717),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(717),
																																															Column: int(66),
																																														},
																																													},
																																												},
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															Id: "arr",
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p7743,
																																																FreeVars: ast.Identifiers{
																																																	"arr",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(717),
																																																		Column: int(67),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(717),
																																																		Column: int(70),
																																																	},
																																																},
																																															},
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7698,
																																												FreeVars: ast.Identifiers{
																																													"arr",
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(717),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(717),
																																														Column: int(71),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																										},
																																										Left: &ast.LiteralString{
																																											Value: "Not enough values to format: ",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7698,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(717),
																																														Column: int(22),
																																													},
																																													End: ast.Location{
																																														Line: int(717),
																																														Column: int(53),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7698,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(717),
																																													Column: int(22),
																																												},
																																												End: ast.Location{
																																													Line: int(717),
																																													Column: int(71),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7698,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(717),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(717),
																																												Column: int(96),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7698,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																										"tmp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(717),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(717),
																																											Column: int(104),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																									},
																																								},
																																								Ctx: p7698,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(717),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(717),
																																										Column: int(105),
																																									},
																																								},
																																							},
																																						},
																																						BranchFalse: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "arr",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																										},
																																									},
																																									Ctx: p7698,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(719),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(719),
																																											Column: int(18),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "tmp",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(719),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(719),
																																												Column: int(22),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "j",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7698,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(719),
																																											Column: int(19),
																																										},
																																										End: ast.Location{
																																											Line: int(719),
																																											Column: int(24),
																																										},
																																									},
																																								},
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7698,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(719),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(719),
																																										Column: int(25),
																																									},
																																								},
																																							},
																																						},
																																						ThenFodder: ast.Fodder{},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7698,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"std",
																																								"tmp",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(716),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(719),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(716),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(719),
																																							Column: int(25),
																																						},
																																					},
																																					Hide: ast.ObjectFieldHide(1),
																																					PlusSuper: false,
																																				},
																																			},
																																			Locals: ast.LocalBinds{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7686,
																																				FreeVars: ast.Identifiers{
																																					"arr",
																																					"std",
																																					"tmp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(714),
																																						Column: int(49),
																																					},
																																					End: ast.Location{
																																						Line: int(720),
																																						Column: int(12),
																																					},
																																				},
																																			},
																																		},
																																		BranchFalse: &ast.DesugaredObject{
																																			Asserts: ast.Nodes{},
																																			Fields: ast.DesugaredObjectFields{
																																				ast.DesugaredObjectField{
																																					Name: &ast.LiteralString{
																																						Value: "j",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Body: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "tmp",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(721),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(721),
																																										Column: int(19),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "j",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7772,
																																							FreeVars: ast.Identifiers{
																																								"tmp",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(721),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(721),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(721),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(721),
																																							Column: int(21),
																																						},
																																					},
																																					Hide: ast.ObjectFieldHide(1),
																																					PlusSuper: false,
																																				},
																																				ast.DesugaredObjectField{
																																					Name: &ast.LiteralString{
																																						Value: "prec",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Body: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "code",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(722),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(722),
																																										Column: int(23),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "prec",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7772,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(722),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(722),
																																									Column: int(28),
																																								},
																																							},
																																						},
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(722),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(722),
																																							Column: int(28),
																																						},
																																					},
																																					Hide: ast.ObjectFieldHide(1),
																																					PlusSuper: false,
																																				},
																																			},
																																			Locals: ast.LocalBinds{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7686,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"tmp",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(720),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(723),
																																						Column: int(12),
																																					},
																																				},
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p7686,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"code",
																																				"std",
																																				"tmp",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(714),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(723),
																																					Column: int(12),
																																				},
																																			},
																																		},
																																	},
																																	EqFodder: ast.Fodder{},
																																	Variable: "tmp2",
																																	CloseFodder: ast.Fodder{},
																																	Fun: nil,
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(714),
																																			Column: int(17),
																																		},
																																		End: ast.Location{
																																			Line: int(723),
																																			Column: int(12),
																																		},
																																	},
																																},
																															},
																															Body: &ast.Local{
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Body: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "tmp2",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"tmp2",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(724),
																																							Column: int(22),
																																						},
																																						End: ast.Location{
																																							Line: int(724),
																																							Column: int(26),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "j",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p7788,
																																				FreeVars: ast.Identifiers{
																																					"tmp2",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(724),
																																						Column: int(22),
																																					},
																																					End: ast.Location{
																																						Line: int(724),
																																						Column: int(28),
																																					},
																																				},
																																			},
																																		},
																																		EqFodder: ast.Fodder{},
																																		Variable: "j2",
																																		CloseFodder: ast.Fodder{},
																																		Fun: nil,
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(724),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(724),
																																				Column: int(28),
																																			},
																																		},
																																	},
																																},
																																Body: &ast.Local{
																																	Binds: ast.LocalBinds{
																																		ast.LocalBind{
																																			VarFodder: ast.Fodder{},
																																			Body: &ast.Conditional{
																																				Cond: &ast.Binary{
																																					Right: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(726),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(726),
																																											Column: int(24),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "length",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7799,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(726),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(726),
																																										Column: int(31),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "arr",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7803,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(726),
																																													Column: int(32),
																																												},
																																												End: ast.Location{
																																													Line: int(726),
																																													Column: int(35),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7799,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(726),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(726),
																																									Column: int(36),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					Left: &ast.Var{
																																						Id: "j2",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7799,
																																							FreeVars: ast.Identifiers{
																																								"j2",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(726),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(726),
																																									Column: int(18),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7799,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(726),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(726),
																																								Column: int(36),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(9),
																																				},
																																				BranchTrue: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "arr",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																								},
																																							},
																																							Ctx: p7799,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(727),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(727),
																																									Column: int(18),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.Var{
																																						Id: "j2",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7799,
																																							FreeVars: ast.Identifiers{
																																								"j2",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(727),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(727),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7799,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(727),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(727),
																																								Column: int(22),
																																							},
																																						},
																																					},
																																				},
																																				BranchFalse: &ast.Error{
																																					Expr: &ast.Binary{
																																						Right: &ast.Var{
																																							Id: "j2",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7799,
																																								FreeVars: ast.Identifiers{
																																									"j2",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(729),
																																										Column: int(100),
																																									},
																																									End: ast.Location{
																																										Line: int(729),
																																										Column: int(102),
																																									},
																																								},
																																							},
																																						},
																																						Left: &ast.Binary{
																																							Right: &ast.LiteralString{
																																								Value: ", expected more than ",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7799,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(729),
																																											Column: int(74),
																																										},
																																										End: ast.Location{
																																											Line: int(729),
																																											Column: int(97),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							Left: &ast.Binary{
																																								Right: &ast.Apply{
																																									Target: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "std",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(729),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(729),
																																														Column: int(59),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "length",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7799,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(729),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(729),
																																													Column: int(66),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													Id: "arr",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p7831,
																																														FreeVars: ast.Identifiers{
																																															"arr",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(729),
																																																Column: int(67),
																																															},
																																															End: ast.Location{
																																																Line: int(729),
																																																Column: int(70),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7799,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(729),
																																												Column: int(56),
																																											},
																																											End: ast.Location{
																																												Line: int(729),
																																												Column: int(71),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								Left: &ast.LiteralString{
																																									Value: "Not enough values to format: ",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7799,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(729),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(729),
																																												Column: int(53),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7799,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(729),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(729),
																																											Column: int(71),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7799,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(729),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(729),
																																										Column: int(97),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7799,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j2",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(729),
																																									Column: int(22),
																																								},
																																								End: ast.Location{
																																									Line: int(729),
																																									Column: int(102),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																							},
																																						},
																																						Ctx: p7799,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(729),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(729),
																																								Column: int(103),
																																							},
																																						},
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																						},
																																					},
																																					Ctx: p7799,
																																					FreeVars: ast.Identifiers{
																																						"arr",
																																						"j2",
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(726),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(729),
																																							Column: int(103),
																																						},
																																					},
																																				},
																																			},
																																			EqFodder: ast.Fodder{},
																																			Variable: "val",
																																			CloseFodder: ast.Fodder{},
																																			Fun: nil,
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(725),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(729),
																																					Column: int(103),
																																				},
																																			},
																																		},
																																	},
																																	Body: &ast.Local{
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					Cond: &ast.Binary{
																																						Right: &ast.LiteralString{
																																							Value: "%",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7848,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(731),
																																										Column: int(30),
																																									},
																																									End: ast.Location{
																																										Line: int(731),
																																										Column: int(33),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						Left: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "code",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(731),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(731),
																																											Column: int(20),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "ctype",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7848,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(731),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(731),
																																										Column: int(26),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7848,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(731),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(731),
																																									Column: int(33),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(12),
																																					},
																																					BranchTrue: &ast.LiteralString{
																																						Value: "%",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																								},
																																							},
																																							Ctx: p7848,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(732),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(732),
																																									Column: int(18),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					BranchFalse: &ast.Apply{
																																						Target: &ast.Var{
																																							Id: "format_code",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																									},
																																								},
																																								Ctx: p7848,
																																								FreeVars: ast.Identifiers{
																																									"format_code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(734),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(734),
																																										Column: int(26),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "val",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7863,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(734),
																																													Column: int(27),
																																												},
																																												End: ast.Location{
																																													Line: int(734),
																																													Column: int(30),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "code",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7863,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(734),
																																													Column: int(32),
																																												},
																																												End: ast.Location{
																																													Line: int(734),
																																													Column: int(36),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "tmp",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(734),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(734),
																																														Column: int(41),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "fw",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7863,
																																											FreeVars: ast.Identifiers{
																																												"tmp",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(734),
																																													Column: int(38),
																																												},
																																												End: ast.Location{
																																													Line: int(734),
																																													Column: int(44),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "tmp2",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"tmp2",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(734),
																																														Column: int(46),
																																													},
																																													End: ast.Location{
																																														Line: int(734),
																																														Column: int(50),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "prec",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7863,
																																											FreeVars: ast.Identifiers{
																																												"tmp2",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(734),
																																													Column: int(46),
																																												},
																																												End: ast.Location{
																																													Line: int(734),
																																													Column: int(55),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "j2",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7863,
																																											FreeVars: ast.Identifiers{
																																												"j2",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(734),
																																													Column: int(57),
																																												},
																																												End: ast.Location{
																																													Line: int(734),
																																													Column: int(59),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p7848,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"format_code",
																																								"j2",
																																								"tmp",
																																								"tmp2",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(734),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(734),
																																									Column: int(60),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																							},
																																						},
																																						Ctx: p7848,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																							"format_code",
																																							"j2",
																																							"tmp",
																																							"tmp2",
																																							"val",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(731),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(734),
																																								Column: int(60),
																																							},
																																						},
																																					},
																																				},
																																				EqFodder: ast.Fodder{},
																																				Variable: "s",
																																				CloseFodder: ast.Fodder{},
																																				Fun: nil,
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(730),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(734),
																																						Column: int(60),
																																					},
																																				},
																																			},
																																		},
																																		Body: &ast.Local{
																																			Binds: ast.LocalBinds{
																																				ast.LocalBind{
																																					VarFodder: ast.Fodder{},
																																					Body: &ast.Conditional{
																																						Cond: &ast.Index{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "code",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(736),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(736),
																																												Column: int(20),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "cflags",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(736),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(736),
																																											Column: int(27),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "left",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7893,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(736),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(736),
																																										Column: int(32),
																																									},
																																								},
																																							},
																																						},
																																						BranchTrue: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "pad_right",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																										},
																																									},
																																									Ctx: p7893,
																																									FreeVars: ast.Identifiers{
																																										"pad_right",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(737),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(737),
																																											Column: int(24),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "s",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7901,
																																												FreeVars: ast.Identifiers{
																																													"s",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(737),
																																														Column: int(25),
																																													},
																																													End: ast.Location{
																																														Line: int(737),
																																														Column: int(26),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "tmp",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"tmp",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(737),
																																															Column: int(28),
																																														},
																																														End: ast.Location{
																																															Line: int(737),
																																															Column: int(31),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.LiteralString{
																																												Value: "fw",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: nil,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7901,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(737),
																																														Column: int(28),
																																													},
																																													End: ast.Location{
																																														Line: int(737),
																																														Column: int(34),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralString{
																																											Value: " ",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7901,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(737),
																																														Column: int(36),
																																													},
																																													End: ast.Location{
																																														Line: int(737),
																																														Column: int(39),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7893,
																																								FreeVars: ast.Identifiers{
																																									"pad_right",
																																									"s",
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(737),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(737),
																																										Column: int(40),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						BranchFalse: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "pad_left",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																										},
																																									},
																																									Ctx: p7893,
																																									FreeVars: ast.Identifiers{
																																										"pad_left",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(739),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(739),
																																											Column: int(23),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "s",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7916,
																																												FreeVars: ast.Identifiers{
																																													"s",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(739),
																																														Column: int(24),
																																													},
																																													End: ast.Location{
																																														Line: int(739),
																																														Column: int(25),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "tmp",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"tmp",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(739),
																																															Column: int(27),
																																														},
																																														End: ast.Location{
																																															Line: int(739),
																																															Column: int(30),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.LiteralString{
																																												Value: "fw",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: nil,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7916,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(739),
																																														Column: int(27),
																																													},
																																													End: ast.Location{
																																														Line: int(739),
																																														Column: int(33),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralString{
																																											Value: " ",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p7916,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(739),
																																														Column: int(35),
																																													},
																																													End: ast.Location{
																																														Line: int(739),
																																														Column: int(38),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p7893,
																																								FreeVars: ast.Identifiers{
																																									"pad_left",
																																									"s",
																																									"tmp",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(739),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(739),
																																										Column: int(39),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						ThenFodder: ast.Fodder{},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																								},
																																							},
																																							Ctx: p7893,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"pad_left",
																																								"pad_right",
																																								"s",
																																								"tmp",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(736),
																																									Column: int(13),
																																								},
																																								End: ast.Location{
																																									Line: int(739),
																																									Column: int(39),
																																								},
																																							},
																																						},
																																					},
																																					EqFodder: ast.Fodder{},
																																					Variable: "s_padded",
																																					CloseFodder: ast.Fodder{},
																																					Fun: nil,
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(735),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(739),
																																							Column: int(39),
																																						},
																																					},
																																				},
																																			},
																																			Body: &ast.Local{
																																				Binds: ast.LocalBinds{
																																					ast.LocalBind{
																																						VarFodder: ast.Fodder{},
																																						Body: &ast.Conditional{
																																							Cond: &ast.Binary{
																																								Right: &ast.LiteralString{
																																									Value: "%",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7933,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(741),
																																												Column: int(30),
																																											},
																																											End: ast.Location{
																																												Line: int(741),
																																												Column: int(33),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								Left: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "code",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(741),
																																													Column: int(16),
																																												},
																																												End: ast.Location{
																																													Line: int(741),
																																													Column: int(20),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "ctype",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7933,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(741),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(741),
																																												Column: int(26),
																																											},
																																										},
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7933,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(741),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(741),
																																											Column: int(33),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(12),
																																							},
																																							BranchTrue: &ast.Var{
																																								Id: "j2",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																										},
																																									},
																																									Ctx: p7933,
																																									FreeVars: ast.Identifiers{
																																										"j2",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(742),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(742),
																																											Column: int(17),
																																										},
																																									},
																																								},
																																							},
																																							BranchFalse: &ast.Binary{
																																								Right: &ast.LiteralNumber{
																																									OriginalString: "1",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7933,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(744),
																																												Column: int(20),
																																											},
																																											End: ast.Location{
																																												Line: int(744),
																																												Column: int(21),
																																											},
																																										},
																																									},
																																								},
																																								Left: &ast.Var{
																																									Id: "j2",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																											},
																																										},
																																										Ctx: p7933,
																																										FreeVars: ast.Identifiers{
																																											"j2",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(744),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(744),
																																												Column: int(17),
																																											},
																																										},
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p7933,
																																									FreeVars: ast.Identifiers{
																																										"j2",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(744),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(744),
																																											Column: int(21),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							ThenFodder: ast.Fodder{},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																								},
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(12),
																																									},
																																								},
																																								Ctx: p7933,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"j2",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(741),
																																										Column: int(13),
																																									},
																																									End: ast.Location{
																																										Line: int(744),
																																										Column: int(21),
																																									},
																																								},
																																							},
																																						},
																																						EqFodder: ast.Fodder{},
																																						Variable: "j3",
																																						CloseFodder: ast.Fodder{},
																																						Fun: nil,
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(740),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(744),
																																								Column: int(21),
																																							},
																																						},
																																					},
																																				},
																																				Body: &ast.Apply{
																																					Target: &ast.Var{
																																						Id: "format_codes_arr",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p7491,
																																							FreeVars: ast.Identifiers{
																																								"format_codes_arr",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(745),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(745),
																																									Column: int(27),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "codes",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7958,
																																										FreeVars: ast.Identifiers{
																																											"codes",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(745),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(745),
																																												Column: int(33),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "arr",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7958,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(745),
																																												Column: int(35),
																																											},
																																											End: ast.Location{
																																												Line: int(745),
																																												Column: int(38),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									Right: &ast.LiteralNumber{
																																										OriginalString: "1",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7958,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(745),
																																													Column: int(44),
																																												},
																																												End: ast.Location{
																																													Line: int(745),
																																													Column: int(45),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Var{
																																										Id: "i",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7958,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(745),
																																													Column: int(40),
																																												},
																																												End: ast.Location{
																																													Line: int(745),
																																													Column: int(41),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7958,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(745),
																																												Column: int(40),
																																											},
																																											End: ast.Location{
																																												Line: int(745),
																																												Column: int(45),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "j3",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7958,
																																										FreeVars: ast.Identifiers{
																																											"j3",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(745),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(745),
																																												Column: int(49),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									Right: &ast.Var{
																																										Id: "s_padded",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7958,
																																											FreeVars: ast.Identifiers{
																																												"s_padded",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(745),
																																													Column: int(55),
																																												},
																																												End: ast.Location{
																																													Line: int(745),
																																													Column: int(63),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Var{
																																										Id: "v",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p7958,
																																											FreeVars: ast.Identifiers{
																																												"v",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(745),
																																													Column: int(51),
																																												},
																																												End: ast.Location{
																																													Line: int(745),
																																													Column: int(52),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p7958,
																																										FreeVars: ast.Identifiers{
																																											"s_padded",
																																											"v",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(745),
																																												Column: int(51),
																																											},
																																											End: ast.Location{
																																												Line: int(745),
																																												Column: int(63),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p7491,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"codes",
																																							"format_codes_arr",
																																							"i",
																																							"j3",
																																							"s_padded",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(745),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(745),
																																								Column: int(64),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: true,
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																						},
																																					},
																																					Ctx: p7491,
																																					FreeVars: ast.Identifiers{
																																						"arr",
																																						"code",
																																						"codes",
																																						"format_codes_arr",
																																						"i",
																																						"j2",
																																						"s_padded",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(740),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(745),
																																							Column: int(64),
																																						},
																																					},
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																					},
																																				},
																																				Ctx: p7491,
																																				FreeVars: ast.Identifiers{
																																					"arr",
																																					"code",
																																					"codes",
																																					"format_codes_arr",
																																					"i",
																																					"j2",
																																					"pad_left",
																																					"pad_right",
																																					"s",
																																					"tmp",
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(735),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(745),
																																						Column: int(64),
																																					},
																																				},
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																				},
																																			},
																																			Ctx: p7491,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"code",
																																				"codes",
																																				"format_code",
																																				"format_codes_arr",
																																				"i",
																																				"j2",
																																				"pad_left",
																																				"pad_right",
																																				"tmp",
																																				"tmp2",
																																				"v",
																																				"val",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(730),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(745),
																																					Column: int(64),
																																				},
																																			},
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																			},
																																		},
																																		Ctx: p7491,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"code",
																																			"codes",
																																			"format_code",
																																			"format_codes_arr",
																																			"i",
																																			"j2",
																																			"pad_left",
																																			"pad_right",
																																			"std",
																																			"tmp",
																																			"tmp2",
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(725),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(745),
																																				Column: int(64),
																																			},
																																		},
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																		},
																																	},
																																	Ctx: p7491,
																																	FreeVars: ast.Identifiers{
																																		"arr",
																																		"code",
																																		"codes",
																																		"format_code",
																																		"format_codes_arr",
																																		"i",
																																		"pad_left",
																																		"pad_right",
																																		"std",
																																		"tmp",
																																		"tmp2",
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(724),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(745),
																																			Column: int(64),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																	},
																																},
																																Ctx: p7491,
																																FreeVars: ast.Identifiers{
																																	"arr",
																																	"code",
																																	"codes",
																																	"format_code",
																																	"format_codes_arr",
																																	"i",
																																	"pad_left",
																																	"pad_right",
																																	"std",
																																	"tmp",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(714),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(745),
																																		Column: int(64),
																																	},
																																},
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(10),
																																},
																															},
																															Ctx: p7491,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"code",
																																"codes",
																																"format_code",
																																"format_codes_arr",
																																"i",
																																"j",
																																"pad_left",
																																"pad_right",
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(704),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(745),
																																	Column: int(64),
																																},
																															},
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p7491,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"code",
																															"codes",
																															"format_code",
																															"format_codes_arr",
																															"i",
																															"j",
																															"pad_left",
																															"pad_right",
																															"std",
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(701),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(745),
																																Column: int(64),
																															},
																														},
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													Ctx: p7491,
																													FreeVars: ast.Identifiers{
																														"arr",
																														"codes",
																														"format_code",
																														"format_codes_arr",
																														"i",
																														"j",
																														"pad_left",
																														"pad_right",
																														"std",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(700),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(745),
																															Column: int(64),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p7491,
																												FreeVars: ast.Identifiers{
																													"arr",
																													"codes",
																													"format_code",
																													"format_codes_arr",
																													"i",
																													"j",
																													"pad_left",
																													"pad_right",
																													"std",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(694),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(745),
																														Column: int(64),
																													},
																												},
																											},
																										},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "codes",
																												CommaFodder: ast.Fodder{},
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(693),
																														Column: int(28),
																													},
																													End: ast.Location{
																														Line: int(693),
																														Column: int(33),
																													},
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "arr",
																												CommaFodder: ast.Fodder{},
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(693),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(693),
																														Column: int(38),
																													},
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "i",
																												CommaFodder: ast.Fodder{},
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(693),
																														Column: int(40),
																													},
																													End: ast.Location{
																														Line: int(693),
																														Column: int(41),
																													},
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "j",
																												CommaFodder: ast.Fodder{},
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(693),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(693),
																														Column: int(44),
																													},
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "v",
																												CommaFodder: nil,
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(693),
																														Column: int(46),
																													},
																													End: ast.Location{
																														Line: int(693),
																														Column: int(47),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: p7999,
																											FreeVars: ast.Identifiers{
																												"format_code",
																												"format_codes_arr",
																												"pad_left",
																												"pad_right",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(693),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(745),
																													Column: int(64),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									EqFodder: nil,
																									Variable: "format_codes_arr",
																									CloseFodder: nil,
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: nil,
																										Body: &ast.Function{
																											ParenLeftFodder: ast.Fodder{},
																											ParenRightFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(749),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(749),
																																			Column: int(18),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "length",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8011,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(749),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(749),
																																		Column: int(25),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "codes",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p8015,
																																			FreeVars: ast.Identifiers{
																																				"codes",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(749),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(749),
																																					Column: int(31),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8011,
																															FreeVars: ast.Identifiers{
																																"codes",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(749),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(749),
																																	Column: int(32),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													Left: &ast.Var{
																														Id: "i",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8011,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(749),
																																	Column: int(10),
																																},
																																End: ast.Location{
																																	Line: int(749),
																																	Column: int(11),
																																},
																															},
																														},
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p8011,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"i",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(749),
																																Column: int(10),
																															},
																															End: ast.Location{
																																Line: int(749),
																																Column: int(32),
																															},
																														},
																													},
																													Op: ast.BinaryOp(8),
																												},
																												BranchTrue: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p8011,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(750),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(750),
																																Column: int(10),
																															},
																														},
																													},
																												},
																												BranchFalse: &ast.Local{
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Body: &ast.Index{
																																Target: &ast.Var{
																																	Id: "codes",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p8028,
																																		FreeVars: ast.Identifiers{
																																			"codes",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(752),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(752),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.Var{
																																	Id: "i",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p8028,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(752),
																																				Column: int(28),
																																			},
																																			End: ast.Location{
																																				Line: int(752),
																																				Column: int(29),
																																			},
																																		},
																																	},
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p8028,
																																	FreeVars: ast.Identifiers{
																																		"codes",
																																		"i",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(752),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(752),
																																			Column: int(30),
																																		},
																																	},
																																},
																															},
																															EqFodder: ast.Fodder{},
																															Variable: "code",
																															CloseFodder: ast.Fodder{},
																															Fun: nil,
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(752),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(752),
																																	Column: int(30),
																																},
																															},
																														},
																													},
																													Body: &ast.Conditional{
																														Cond: &ast.Binary{
																															Right: &ast.LiteralString{
																																Value: "string",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p8011,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(753),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(753),
																																			Column: int(38),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Left: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(753),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(753),
																																					Column: int(15),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "type",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p8011,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(753),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(753),
																																				Column: int(20),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "code",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8044,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(753),
																																							Column: int(21),
																																						},
																																						End: ast.Location{
																																							Line: int(753),
																																							Column: int(25),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p8011,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(753),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(753),
																																			Column: int(26),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8011,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(753),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(753),
																																		Column: int(38),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(12),
																														},
																														BranchTrue: &ast.Apply{
																															Target: &ast.Var{
																																Id: "format_codes_obj",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																		},
																																	},
																																	Ctx: p8011,
																																	FreeVars: ast.Identifiers{
																																		"format_codes_obj",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(754),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(754),
																																			Column: int(27),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "codes",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8054,
																																				FreeVars: ast.Identifiers{
																																					"codes",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(754),
																																						Column: int(28),
																																					},
																																					End: ast.Location{
																																						Line: int(754),
																																						Column: int(33),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "obj",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8054,
																																				FreeVars: ast.Identifiers{
																																					"obj",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(754),
																																						Column: int(35),
																																					},
																																					End: ast.Location{
																																						Line: int(754),
																																						Column: int(38),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			Right: &ast.LiteralNumber{
																																				OriginalString: "1",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8054,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(754),
																																							Column: int(44),
																																						},
																																						End: ast.Location{
																																							Line: int(754),
																																							Column: int(45),
																																						},
																																					},
																																				},
																																			},
																																			Left: &ast.Var{
																																				Id: "i",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8054,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(754),
																																							Column: int(40),
																																						},
																																						End: ast.Location{
																																							Line: int(754),
																																							Column: int(41),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8054,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(754),
																																						Column: int(40),
																																					},
																																					End: ast.Location{
																																						Line: int(754),
																																						Column: int(45),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(3),
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			Right: &ast.Var{
																																				Id: "code",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8054,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(754),
																																							Column: int(51),
																																						},
																																						End: ast.Location{
																																							Line: int(754),
																																							Column: int(55),
																																						},
																																					},
																																				},
																																			},
																																			Left: &ast.Var{
																																				Id: "v",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8054,
																																					FreeVars: ast.Identifiers{
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(754),
																																							Column: int(47),
																																						},
																																						End: ast.Location{
																																							Line: int(754),
																																							Column: int(48),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8054,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(754),
																																						Column: int(47),
																																					},
																																					End: ast.Location{
																																						Line: int(754),
																																						Column: int(55),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(3),
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8011,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"codes",
																																	"format_codes_obj",
																																	"i",
																																	"obj",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(754),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(754),
																																		Column: int(56),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: true,
																														},
																														BranchFalse: &ast.Local{
																															Binds: ast.LocalBinds{
																																ast.LocalBind{
																																	VarFodder: ast.Fodder{},
																																	Body: &ast.Conditional{
																																		Cond: &ast.Binary{
																																			Right: &ast.LiteralNull{
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8075,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(757),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(757),
																																							Column: int(33),
																																						},
																																					},
																																				},
																																			},
																																			Left: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "code",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(757),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(757),
																																								Column: int(20),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "mkey",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8075,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(757),
																																							Column: int(16),
																																						},
																																						End: ast.Location{
																																							Line: int(757),
																																							Column: int(25),
																																						},
																																					},
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8075,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(757),
																																						Column: int(16),
																																					},
																																					End: ast.Location{
																																						Line: int(757),
																																						Column: int(33),
																																					},
																																				},
																																			},
																																			Op: ast.BinaryOp(12),
																																		},
																																		BranchTrue: &ast.Error{
																																			Expr: &ast.LiteralString{
																																				Value: "Mapping keys required.",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8075,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(758),
																																							Column: int(21),
																																						},
																																						End: ast.Location{
																																							Line: int(758),
																																							Column: int(45),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(14),
																																					},
																																				},
																																				Ctx: p8075,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(758),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(758),
																																						Column: int(45),
																																					},
																																				},
																																			},
																																		},
																																		BranchFalse: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "code",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(14),
																																						},
																																					},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(760),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(760),
																																							Column: int(19),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "mkey",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p8075,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(760),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(760),
																																						Column: int(24),
																																					},
																																				},
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(12),
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(12),
																																				},
																																			},
																																			Ctx: p8075,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(757),
																																					Column: int(13),
																																				},
																																				End: ast.Location{
																																					Line: int(760),
																																					Column: int(24),
																																				},
																																			},
																																		},
																																	},
																																	EqFodder: ast.Fodder{},
																																	Variable: "f",
																																	CloseFodder: ast.Fodder{},
																																	Fun: nil,
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(756),
																																			Column: int(17),
																																		},
																																		End: ast.Location{
																																			Line: int(760),
																																			Column: int(24),
																																		},
																																	},
																																},
																															},
																															Body: &ast.Local{
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Body: &ast.Conditional{
																																			Cond: &ast.Binary{
																																				Right: &ast.LiteralString{
																																					Value: "*",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8099,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(762),
																																								Column: int(27),
																																							},
																																							End: ast.Location{
																																								Line: int(762),
																																								Column: int(30),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				Left: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "code",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(762),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(762),
																																									Column: int(20),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "fw",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8099,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(762),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(762),
																																								Column: int(23),
																																							},
																																						},
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8099,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(762),
																																							Column: int(16),
																																						},
																																						End: ast.Location{
																																							Line: int(762),
																																							Column: int(30),
																																						},
																																					},
																																				},
																																				Op: ast.BinaryOp(12),
																																			},
																																			BranchTrue: &ast.Error{
																																				Expr: &ast.LiteralString{
																																					Value: "Cannot use * field width with object.",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8099,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(763),
																																								Column: int(21),
																																							},
																																							End: ast.Location{
																																								Line: int(763),
																																								Column: int(60),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(14),
																																						},
																																					},
																																					Ctx: p8099,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(763),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(763),
																																							Column: int(60),
																																						},
																																					},
																																				},
																																			},
																																			BranchFalse: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "code",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																							},
																																						},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(765),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(765),
																																								Column: int(19),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "fw",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p8099,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(765),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(765),
																																							Column: int(22),
																																						},
																																					},
																																				},
																																			},
																																			ThenFodder: ast.Fodder{},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(12),
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																					},
																																				},
																																				Ctx: p8099,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(762),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(765),
																																						Column: int(22),
																																					},
																																				},
																																			},
																																		},
																																		EqFodder: ast.Fodder{},
																																		Variable: "fw",
																																		CloseFodder: ast.Fodder{},
																																		Fun: nil,
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(761),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(765),
																																				Column: int(22),
																																			},
																																		},
																																	},
																																},
																																Body: &ast.Local{
																																	Binds: ast.LocalBinds{
																																		ast.LocalBind{
																																			VarFodder: ast.Fodder{},
																																			Body: &ast.Conditional{
																																				Cond: &ast.Binary{
																																					Right: &ast.LiteralString{
																																						Value: "*",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p8123,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(767),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(767),
																																									Column: int(32),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					Left: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "code",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(767),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(767),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "prec",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p8123,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(767),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(767),
																																									Column: int(25),
																																								},
																																							},
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8123,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(767),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(767),
																																								Column: int(32),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(12),
																																				},
																																				BranchTrue: &ast.Error{
																																					Expr: &ast.LiteralString{
																																						Value: "Cannot use * precision with object.",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p8123,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(768),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(768),
																																									Column: int(58),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																							},
																																						},
																																						Ctx: p8123,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(768),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(768),
																																								Column: int(58),
																																							},
																																						},
																																					},
																																				},
																																				BranchFalse: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "code",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																								},
																																							},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(770),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(770),
																																									Column: int(19),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "prec",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8123,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(770),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(770),
																																								Column: int(24),
																																							},
																																						},
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																						},
																																					},
																																					Ctx: p8123,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(767),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(770),
																																							Column: int(24),
																																						},
																																					},
																																				},
																																			},
																																			EqFodder: ast.Fodder{},
																																			Variable: "prec",
																																			CloseFodder: ast.Fodder{},
																																			Fun: nil,
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(766),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(770),
																																					Column: int(24),
																																				},
																																			},
																																		},
																																	},
																																	Body: &ast.Local{
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					Cond: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(772),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(772),
																																											Column: int(19),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "objectHasAll",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p8150,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(772),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(772),
																																										Column: int(32),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "obj",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8154,
																																											FreeVars: ast.Identifiers{
																																												"obj",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(772),
																																													Column: int(33),
																																												},
																																												End: ast.Location{
																																													Line: int(772),
																																													Column: int(36),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "f",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8154,
																																											FreeVars: ast.Identifiers{
																																												"f",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(772),
																																													Column: int(38),
																																												},
																																												End: ast.Location{
																																													Line: int(772),
																																													Column: int(39),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p8150,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																								"obj",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(772),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(772),
																																									Column: int(40),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					BranchTrue: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "obj",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																									},
																																								},
																																								Ctx: p8150,
																																								FreeVars: ast.Identifiers{
																																									"obj",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(773),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(773),
																																										Column: int(18),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.Var{
																																							Id: "f",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p8150,
																																								FreeVars: ast.Identifiers{
																																									"f",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(773),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(773),
																																										Column: int(20),
																																									},
																																								},
																																							},
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p8150,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																								"obj",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(773),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(773),
																																									Column: int(21),
																																								},
																																							},
																																						},
																																					},
																																					BranchFalse: &ast.Error{
																																						Expr: &ast.Binary{
																																							Right: &ast.Var{
																																								Id: "f",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8150,
																																									FreeVars: ast.Identifiers{
																																										"f",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(775),
																																											Column: int(41),
																																										},
																																										End: ast.Location{
																																											Line: int(775),
																																											Column: int(42),
																																										},
																																									},
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								Value: "No such field: ",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8150,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(775),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(775),
																																											Column: int(38),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p8150,
																																								FreeVars: ast.Identifiers{
																																									"f",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(775),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(775),
																																										Column: int(42),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																								},
																																							},
																																							Ctx: p8150,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(775),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(775),
																																									Column: int(42),
																																								},
																																							},
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																							},
																																						},
																																						Ctx: p8150,
																																						FreeVars: ast.Identifiers{
																																							"f",
																																							"obj",
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(772),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(775),
																																								Column: int(42),
																																							},
																																						},
																																					},
																																				},
																																				EqFodder: ast.Fodder{},
																																				Variable: "val",
																																				CloseFodder: ast.Fodder{},
																																				Fun: nil,
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(771),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(775),
																																						Column: int(42),
																																					},
																																				},
																																			},
																																		},
																																		Body: &ast.Local{
																																			Binds: ast.LocalBinds{
																																				ast.LocalBind{
																																					VarFodder: ast.Fodder{},
																																					Body: &ast.Conditional{
																																						Cond: &ast.Binary{
																																							Right: &ast.LiteralString{
																																								Value: "%",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8182,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(777),
																																											Column: int(30),
																																										},
																																										End: ast.Location{
																																											Line: int(777),
																																											Column: int(33),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							Left: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "code",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(777),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(777),
																																												Column: int(20),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "ctype",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8182,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(777),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(777),
																																											Column: int(26),
																																										},
																																									},
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p8182,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(777),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(777),
																																										Column: int(33),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(12),
																																						},
																																						BranchTrue: &ast.LiteralString{
																																							Value: "%",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																									},
																																								},
																																								Ctx: p8182,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(778),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(778),
																																										Column: int(18),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						BranchFalse: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "format_code",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Comment: []string{},
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																										},
																																									},
																																									Ctx: p8182,
																																									FreeVars: ast.Identifiers{
																																										"format_code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(780),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(780),
																																											Column: int(26),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "val",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p8197,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(780),
																																														Column: int(27),
																																													},
																																													End: ast.Location{
																																														Line: int(780),
																																														Column: int(30),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "code",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p8197,
																																												FreeVars: ast.Identifiers{
																																													"code",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(780),
																																														Column: int(32),
																																													},
																																													End: ast.Location{
																																														Line: int(780),
																																														Column: int(36),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "fw",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p8197,
																																												FreeVars: ast.Identifiers{
																																													"fw",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(780),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(780),
																																														Column: int(40),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "prec",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p8197,
																																												FreeVars: ast.Identifiers{
																																													"prec",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(780),
																																														Column: int(42),
																																													},
																																													End: ast.Location{
																																														Line: int(780),
																																														Column: int(46),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "f",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p8197,
																																												FreeVars: ast.Identifiers{
																																													"f",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(780),
																																														Column: int(48),
																																													},
																																													End: ast.Location{
																																														Line: int(780),
																																														Column: int(49),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p8182,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"f",
																																									"format_code",
																																									"fw",
																																									"prec",
																																									"val",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(780),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(780),
																																										Column: int(50),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						ThenFodder: ast.Fodder{},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																							},
																																						},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																								},
																																							},
																																							Ctx: p8182,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"f",
																																								"format_code",
																																								"fw",
																																								"prec",
																																								"val",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(777),
																																									Column: int(13),
																																								},
																																								End: ast.Location{
																																									Line: int(780),
																																									Column: int(50),
																																								},
																																							},
																																						},
																																					},
																																					EqFodder: ast.Fodder{},
																																					Variable: "s",
																																					CloseFodder: ast.Fodder{},
																																					Fun: nil,
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(776),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(780),
																																							Column: int(50),
																																						},
																																					},
																																				},
																																			},
																																			Body: &ast.Local{
																																				Binds: ast.LocalBinds{
																																					ast.LocalBind{
																																						VarFodder: ast.Fodder{},
																																						Body: &ast.Conditional{
																																							Cond: &ast.Index{
																																								Target: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "code",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(782),
																																													Column: int(16),
																																												},
																																												End: ast.Location{
																																													Line: int(782),
																																													Column: int(20),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "cflags",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(782),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(782),
																																												Column: int(27),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "left",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8221,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(782),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(782),
																																											Column: int(32),
																																										},
																																									},
																																								},
																																							},
																																							BranchTrue: &ast.Apply{
																																								Target: &ast.Var{
																																									Id: "pad_right",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																											},
																																										},
																																										Ctx: p8221,
																																										FreeVars: ast.Identifiers{
																																											"pad_right",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(783),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(783),
																																												Column: int(24),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "s",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8229,
																																													FreeVars: ast.Identifiers{
																																														"s",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(783),
																																															Column: int(25),
																																														},
																																														End: ast.Location{
																																															Line: int(783),
																																															Column: int(26),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "fw",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8229,
																																													FreeVars: ast.Identifiers{
																																														"fw",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(783),
																																															Column: int(28),
																																														},
																																														End: ast.Location{
																																															Line: int(783),
																																															Column: int(30),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.LiteralString{
																																												Value: " ",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8229,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(783),
																																															Column: int(32),
																																														},
																																														End: ast.Location{
																																															Line: int(783),
																																															Column: int(35),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8221,
																																									FreeVars: ast.Identifiers{
																																										"fw",
																																										"pad_right",
																																										"s",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(783),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(783),
																																											Column: int(36),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							BranchFalse: &ast.Apply{
																																								Target: &ast.Var{
																																									Id: "pad_left",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Comment: []string{},
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																											},
																																										},
																																										Ctx: p8221,
																																										FreeVars: ast.Identifiers{
																																											"pad_left",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(785),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(785),
																																												Column: int(23),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "s",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8241,
																																													FreeVars: ast.Identifiers{
																																														"s",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(785),
																																															Column: int(24),
																																														},
																																														End: ast.Location{
																																															Line: int(785),
																																															Column: int(25),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "fw",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8241,
																																													FreeVars: ast.Identifiers{
																																														"fw",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(785),
																																															Column: int(27),
																																														},
																																														End: ast.Location{
																																															Line: int(785),
																																															Column: int(29),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.LiteralString{
																																												Value: " ",
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p8241,
																																													FreeVars: ast.Identifiers{},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(785),
																																															Column: int(31),
																																														},
																																														End: ast.Location{
																																															Line: int(785),
																																															Column: int(34),
																																														},
																																													},
																																												},
																																												Kind: ast.LiteralStringKind(1),
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p8221,
																																									FreeVars: ast.Identifiers{
																																										"fw",
																																										"pad_left",
																																										"s",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(785),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(785),
																																											Column: int(35),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							ThenFodder: ast.Fodder{},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																								},
																																							},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Comment: []string{},
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(12),
																																									},
																																								},
																																								Ctx: p8221,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"fw",
																																									"pad_left",
																																									"pad_right",
																																									"s",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(782),
																																										Column: int(13),
																																									},
																																									End: ast.Location{
																																										Line: int(785),
																																										Column: int(35),
																																									},
																																								},
																																							},
																																						},
																																						EqFodder: ast.Fodder{},
																																						Variable: "s_padded",
																																						CloseFodder: ast.Fodder{},
																																						Fun: nil,
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(781),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(785),
																																								Column: int(35),
																																							},
																																						},
																																					},
																																				},
																																				Body: &ast.Apply{
																																					Target: &ast.Var{
																																						Id: "format_codes_obj",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Comment: []string{},
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																								},
																																							},
																																							Ctx: p8011,
																																							FreeVars: ast.Identifiers{
																																								"format_codes_obj",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(786),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(786),
																																									Column: int(27),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "codes",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p8256,
																																										FreeVars: ast.Identifiers{
																																											"codes",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(786),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(786),
																																												Column: int(33),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "obj",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p8256,
																																										FreeVars: ast.Identifiers{
																																											"obj",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(786),
																																												Column: int(35),
																																											},
																																											End: ast.Location{
																																												Line: int(786),
																																												Column: int(38),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									Right: &ast.LiteralNumber{
																																										OriginalString: "1",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8256,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(786),
																																													Column: int(44),
																																												},
																																												End: ast.Location{
																																													Line: int(786),
																																													Column: int(45),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Var{
																																										Id: "i",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8256,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(786),
																																													Column: int(40),
																																												},
																																												End: ast.Location{
																																													Line: int(786),
																																													Column: int(41),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p8256,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(786),
																																												Column: int(40),
																																											},
																																											End: ast.Location{
																																												Line: int(786),
																																												Column: int(45),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									Right: &ast.Var{
																																										Id: "s_padded",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8256,
																																											FreeVars: ast.Identifiers{
																																												"s_padded",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(786),
																																													Column: int(51),
																																												},
																																												End: ast.Location{
																																													Line: int(786),
																																													Column: int(59),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Var{
																																										Id: "v",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p8256,
																																											FreeVars: ast.Identifiers{
																																												"v",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(786),
																																													Column: int(47),
																																												},
																																												End: ast.Location{
																																													Line: int(786),
																																													Column: int(48),
																																												},
																																											},
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p8256,
																																										FreeVars: ast.Identifiers{
																																											"s_padded",
																																											"v",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(786),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(786),
																																												Column: int(59),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p8011,
																																						FreeVars: ast.Identifiers{
																																							"codes",
																																							"format_codes_obj",
																																							"i",
																																							"obj",
																																							"s_padded",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(786),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(786),
																																								Column: int(60),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: true,
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																						},
																																					},
																																					Ctx: p8011,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																						"codes",
																																						"format_codes_obj",
																																						"fw",
																																						"i",
																																						"obj",
																																						"pad_left",
																																						"pad_right",
																																						"s",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(781),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(786),
																																							Column: int(60),
																																						},
																																					},
																																				},
																																			},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																					},
																																				},
																																				Ctx: p8011,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"codes",
																																					"f",
																																					"format_code",
																																					"format_codes_obj",
																																					"fw",
																																					"i",
																																					"obj",
																																					"pad_left",
																																					"pad_right",
																																					"prec",
																																					"v",
																																					"val",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(776),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(786),
																																						Column: int(60),
																																					},
																																				},
																																			},
																																		},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																				},
																																			},
																																			Ctx: p8011,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"codes",
																																				"f",
																																				"format_code",
																																				"format_codes_obj",
																																				"fw",
																																				"i",
																																				"obj",
																																				"pad_left",
																																				"pad_right",
																																				"prec",
																																				"std",
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(771),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(786),
																																					Column: int(60),
																																				},
																																			},
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Comment: []string{},
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																			},
																																		},
																																		Ctx: p8011,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																			"codes",
																																			"f",
																																			"format_code",
																																			"format_codes_obj",
																																			"fw",
																																			"i",
																																			"obj",
																																			"pad_left",
																																			"pad_right",
																																			"std",
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(766),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(786),
																																				Column: int(60),
																																			},
																																		},
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																		},
																																	},
																																	Ctx: p8011,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																		"codes",
																																		"f",
																																		"format_code",
																																		"format_codes_obj",
																																		"i",
																																		"obj",
																																		"pad_left",
																																		"pad_right",
																																		"std",
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(761),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(786),
																																			Column: int(60),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																	},
																																},
																																Ctx: p8011,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"codes",
																																	"format_code",
																																	"format_codes_obj",
																																	"i",
																																	"obj",
																																	"pad_left",
																																	"pad_right",
																																	"std",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(756),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(786),
																																		Column: int(60),
																																	},
																																},
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(8),
																																},
																															},
																															Ctx: p8011,
																															FreeVars: ast.Identifiers{
																																"code",
																																"codes",
																																"format_code",
																																"format_codes_obj",
																																"i",
																																"obj",
																																"pad_left",
																																"pad_right",
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(753),
																																	Column: int(9),
																																},
																																End: ast.Location{
																																	Line: int(786),
																																	Column: int(60),
																																},
																															},
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p8011,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"format_code",
																															"format_codes_obj",
																															"i",
																															"obj",
																															"pad_left",
																															"pad_right",
																															"std",
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(752),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(786),
																																Column: int(60),
																															},
																														},
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p8011,
																													FreeVars: ast.Identifiers{
																														"codes",
																														"format_code",
																														"format_codes_obj",
																														"i",
																														"obj",
																														"pad_left",
																														"pad_right",
																														"std",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(749),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(786),
																															Column: int(60),
																														},
																													},
																												},
																											},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "codes",
																													CommaFodder: ast.Fodder{},
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(748),
																															Column: int(28),
																														},
																														End: ast.Location{
																															Line: int(748),
																															Column: int(33),
																														},
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "obj",
																													CommaFodder: ast.Fodder{},
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(748),
																															Column: int(35),
																														},
																														End: ast.Location{
																															Line: int(748),
																															Column: int(38),
																														},
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "i",
																													CommaFodder: ast.Fodder{},
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(748),
																															Column: int(40),
																														},
																														End: ast.Location{
																															Line: int(748),
																															Column: int(41),
																														},
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "v",
																													CommaFodder: nil,
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(748),
																															Column: int(43),
																														},
																														End: ast.Location{
																															Line: int(748),
																															Column: int(44),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: p8293,
																												FreeVars: ast.Identifiers{
																													"format_code",
																													"format_codes_obj",
																													"pad_left",
																													"pad_right",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(748),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(786),
																														Column: int(60),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										EqFodder: nil,
																										Variable: "format_codes_obj",
																										CloseFodder: nil,
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																								},
																								Body: &ast.Conditional{
																									Cond: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(788),
																															Column: int(8),
																														},
																														End: ast.Location{
																															Line: int(788),
																															Column: int(11),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "isArray",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p8301,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(788),
																														Column: int(8),
																													},
																													End: ast.Location{
																														Line: int(788),
																														Column: int(19),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "vals",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8305,
																															FreeVars: ast.Identifiers{
																																"vals",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(788),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(788),
																																	Column: int(24),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p8301,
																											FreeVars: ast.Identifiers{
																												"std",
																												"vals",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(788),
																													Column: int(8),
																												},
																												End: ast.Location{
																													Line: int(788),
																													Column: int(25),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									BranchTrue: &ast.Apply{
																										Target: &ast.Var{
																											Id: "format_codes_arr",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												Ctx: p8301,
																												FreeVars: ast.Identifiers{
																													"format_codes_arr",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(789),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(789),
																														Column: int(23),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "codes",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8314,
																															FreeVars: ast.Identifiers{
																																"codes",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(789),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(789),
																																	Column: int(29),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "vals",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8314,
																															FreeVars: ast.Identifiers{
																																"vals",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(789),
																																	Column: int(31),
																																},
																																End: ast.Location{
																																	Line: int(789),
																																	Column: int(35),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8314,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(789),
																																	Column: int(37),
																																},
																																End: ast.Location{
																																	Line: int(789),
																																	Column: int(38),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8314,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(789),
																																	Column: int(40),
																																},
																																End: ast.Location{
																																	Line: int(789),
																																	Column: int(41),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p8314,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(789),
																																	Column: int(43),
																																},
																																End: ast.Location{
																																	Line: int(789),
																																	Column: int(45),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p8301,
																											FreeVars: ast.Identifiers{
																												"codes",
																												"format_codes_arr",
																												"vals",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(789),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(789),
																													Column: int(46),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									BranchFalse: &ast.Conditional{
																										Cond: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "std",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(790),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(790),
																																Column: int(16),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "isObject",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p8301,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(790),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(790),
																															Column: int(25),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "vals",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8331,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(790),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(790),
																																		Column: int(30),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p8301,
																												FreeVars: ast.Identifiers{
																													"std",
																													"vals",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(790),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(790),
																														Column: int(31),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										BranchTrue: &ast.Apply{
																											Target: &ast.Var{
																												Id: "format_codes_obj",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p8301,
																													FreeVars: ast.Identifiers{
																														"format_codes_obj",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(791),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(791),
																															Column: int(23),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "codes",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8340,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(791),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(791),
																																		Column: int(29),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "vals",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8340,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(791),
																																		Column: int(31),
																																	},
																																	End: ast.Location{
																																		Line: int(791),
																																		Column: int(35),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															OriginalString: "0",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8340,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(791),
																																		Column: int(37),
																																	},
																																	End: ast.Location{
																																		Line: int(791),
																																		Column: int(38),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															Value: "",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8340,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(791),
																																		Column: int(40),
																																	},
																																	End: ast.Location{
																																		Line: int(791),
																																		Column: int(42),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p8301,
																												FreeVars: ast.Identifiers{
																													"codes",
																													"format_codes_obj",
																													"vals",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(791),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(791),
																														Column: int(43),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										BranchFalse: &ast.Apply{
																											Target: &ast.Var{
																												Id: "format_codes_arr",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													Ctx: p8301,
																													FreeVars: ast.Identifiers{
																														"format_codes_arr",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(793),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(793),
																															Column: int(23),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "codes",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8353,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(793),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(793),
																																		Column: int(29),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Array{
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "vals",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p8358,
																																			FreeVars: ast.Identifiers{
																																				"vals",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(793),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(793),
																																					Column: int(36),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															CloseFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8353,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(793),
																																		Column: int(31),
																																	},
																																	End: ast.Location{
																																		Line: int(793),
																																		Column: int(37),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															OriginalString: "0",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8353,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(793),
																																		Column: int(39),
																																	},
																																	End: ast.Location{
																																		Line: int(793),
																																		Column: int(40),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															OriginalString: "0",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8353,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(793),
																																		Column: int(42),
																																	},
																																	End: ast.Location{
																																		Line: int(793),
																																		Column: int(43),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															Value: "",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p8353,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(793),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(793),
																																		Column: int(47),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p8301,
																												FreeVars: ast.Identifiers{
																													"codes",
																													"format_codes_arr",
																													"vals",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(793),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(793),
																														Column: int(48),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(4),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p8301,
																											FreeVars: ast.Identifiers{
																												"codes",
																												"format_codes_arr",
																												"format_codes_obj",
																												"std",
																												"vals",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(790),
																													Column: int(10),
																												},
																												End: ast.Location{
																													Line: int(793),
																													Column: int(48),
																												},
																											},
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(4),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(1),
																												Indent: int(4),
																											},
																										},
																										Ctx: p8301,
																										FreeVars: ast.Identifiers{
																											"codes",
																											"format_codes_arr",
																											"format_codes_obj",
																											"std",
																											"vals",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(788),
																												Column: int(5),
																											},
																											End: ast.Location{
																												Line: int(793),
																												Column: int(48),
																											},
																										},
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(1),
																											Indent: int(4),
																										},
																										ast.FodderElement{
																											Comment: []string{
																												"// Render a parsed format string with an object of values.",
																											},
																											Kind: ast.FodderKind(2),
																											Blanks: int(0),
																											Indent: int(4),
																										},
																									},
																									Ctx: p8301,
																									FreeVars: ast.Identifiers{
																										"codes",
																										"format_code",
																										"format_codes_arr",
																										"pad_left",
																										"pad_right",
																										"std",
																										"vals",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(748),
																											Column: int(5),
																										},
																										End: ast.Location{
																											Line: int(793),
																											Column: int(48),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(1),
																										Indent: int(4),
																									},
																									ast.FodderElement{
																										Comment: []string{
																											"// Render a parsed format string with an array of values.",
																										},
																										Kind: ast.FodderKind(2),
																										Blanks: int(0),
																										Indent: int(4),
																									},
																								},
																								Ctx: p8301,
																								FreeVars: ast.Identifiers{
																									"codes",
																									"format_code",
																									"pad_left",
																									"pad_right",
																									"std",
																									"vals",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(693),
																										Column: int(5),
																									},
																									End: ast.Location{
																										Line: int(793),
																										Column: int(48),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(1),
																									Indent: int(4),
																								},
																								ast.FodderElement{
																									Comment: []string{
																										"// Render a value with an arbitrary format code.",
																									},
																									Kind: ast.FodderKind(2),
																									Blanks: int(0),
																									Indent: int(4),
																								},
																							},
																							Ctx: p8301,
																							FreeVars: ast.Identifiers{
																								"codes",
																								"pad_left",
																								"pad_right",
																								"render_float_dec",
																								"render_float_sci",
																								"render_hex",
																								"render_int",
																								"std",
																								"vals",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(598),
																									Column: int(5),
																								},
																								End: ast.Location{
																									Line: int(793),
																									Column: int(48),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(1),
																								Indent: int(4),
																							},
																							ast.FodderElement{
																								Comment: []string{
																									"// Render floating point in scientific form",
																								},
																								Kind: ast.FodderKind(2),
																								Blanks: int(0),
																								Indent: int(4),
																							},
																						},
																						Ctx: p8301,
																						FreeVars: ast.Identifiers{
																							"codes",
																							"pad_left",
																							"pad_right",
																							"render_float_dec",
																							"render_hex",
																							"render_int",
																							"std",
																							"vals",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(584),
																								Column: int(5),
																							},
																							End: ast.Location{
																								Line: int(793),
																								Column: int(48),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(1),
																							Indent: int(4),
																						},
																						ast.FodderElement{
																							Comment: []string{
																								"// Render floating point in decimal form",
																							},
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(4),
																						},
																					},
																					Ctx: p8301,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"codes",
																						"pad_left",
																						"pad_right",
																						"render_hex",
																						"render_int",
																						"std",
																						"strip_trailing_zero",
																						"vals",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(560),
																							Column: int(5),
																						},
																						End: ast.Location{
																							Line: int(793),
																							Column: int(48),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(1),
																						Indent: int(4),
																					},
																				},
																				Ctx: p8301,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"codes",
																					"pad_left",
																					"pad_right",
																					"render_hex",
																					"render_int",
																					"std",
																					"vals",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(548),
																						Column: int(5),
																					},
																					End: ast.Location{
																						Line: int(793),
																						Column: int(48),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(1),
																					Indent: int(4),
																				},
																				ast.FodderElement{
																					Comment: []string{
																						"// Render an integer in hexadecimal.",
																					},
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(4),
																				},
																			},
																			Ctx: p8301,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"codes",
																				"pad_left",
																				"pad_right",
																				"render_int",
																				"std",
																				"vals",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(529),
																					Column: int(5),
																				},
																				End: ast.Location{
																					Line: int(793),
																					Column: int(48),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(1),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// Render a sign & magnitude integer (radix ranges from decimal to binary).",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// neg should be a boolean, and when true indicates that we should render a negative number.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// mag must always be a whole number >= 0, it's the magnitude of the integer to render",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// min_chars must be a whole number >= 0",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"//   It is the field width, i.e. std.length() of the result should be >= min_chars",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// min_digits must be a whole number >= 0. It's the number of zeroes to pad with.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// blank must be a boolean, if true adds an additional ' ' in front of a positive number, so",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// that it is aligned with negative numbers with the same number of digits.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// plus must be a boolean, if true adds a '+' in front of a positive number, so that it is",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// aligned with negative numbers with the same number of digits.  This takes precedence over",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// blank, if both are true.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// radix must be a whole number >1 and <= 10.  It is the base of the system of numerals.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																			ast.FodderElement{
																				Comment: []string{
																					"// zero_prefix is a string prefixed before the sign to all numbers that are not 0.",
																				},
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																			},
																		},
																		Ctx: p8301,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"codes",
																			"pad_left",
																			"pad_right",
																			"std",
																			"vals",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(511),
																				Column: int(5),
																			},
																			End: ast.Location{
																				Line: int(793),
																				Column: int(48),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(1),
																			Indent: int(4),
																		},
																		ast.FodderElement{
																			Comment: []string{
																				"// Add s to the right of str so that its length is at least w.",
																			},
																			Kind: ast.FodderKind(2),
																			Blanks: int(0),
																			Indent: int(4),
																		},
																	},
																	Ctx: p8301,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"codes",
																		"pad_left",
																		"padding",
																		"std",
																		"vals",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(495),
																			Column: int(5),
																		},
																		End: ast.Location{
																			Line: int(793),
																			Column: int(48),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(1),
																		Indent: int(4),
																	},
																	ast.FodderElement{
																		Comment: []string{
																			"// Add s to the left of str so that its length is at least w.",
																		},
																		Kind: ast.FodderKind(2),
																		Blanks: int(0),
																		Indent: int(4),
																	},
																},
																Ctx: p8301,
																FreeVars: ast.Identifiers{
																	"$std",
																	"codes",
																	"padding",
																	"std",
																	"vals",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(491),
																		Column: int(5),
																	},
																	End: ast.Location{
																		Line: int(793),
																		Column: int(48),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(2),
																	Indent: int(4),
																},
																ast.FodderElement{
																	Comment: []string{
																		"///////////////////////",
																	},
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																},
																ast.FodderElement{
																	Comment: []string{
																		"// Format the values //",
																	},
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																},
																ast.FodderElement{
																	Comment: []string{
																		"///////////////////////",
																	},
																	Kind: ast.FodderKind(2),
																	Blanks: int(1),
																	Indent: int(4),
																},
																ast.FodderElement{
																	Comment: []string{
																		"// Useful utilities",
																	},
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																},
															},
															Ctx: p8301,
															FreeVars: ast.Identifiers{
																"$std",
																"codes",
																"std",
																"vals",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(482),
																	Column: int(5),
																},
																End: ast.Location{
																	Line: int(793),
																	Column: int(48),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(1),
																Indent: int(4),
															},
														},
														Ctx: p8301,
														FreeVars: ast.Identifiers{
															"$std",
															"parse_codes",
															"std",
															"str",
															"vals",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(474),
																Column: int(5),
															},
															End: ast.Location{
																Line: int(793),
																Column: int(48),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(1),
															Indent: int(4),
														},
														ast.FodderElement{
															Comment: []string{
																"// Parse a format string (containing none or more % format tags).",
															},
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
														},
													},
													Ctx: p8301,
													FreeVars: ast.Identifiers{
														"$std",
														"parse_code",
														"std",
														"str",
														"vals",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(463),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(793),
															Column: int(48),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(2),
														Indent: int(4),
													},
													ast.FodderElement{
														Comment: []string{
															"// Parsed initial %, now the rest.",
														},
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p8301,
												FreeVars: ast.Identifiers{
													"$std",
													"parse_conv_type",
													"std",
													"str",
													"try_parse_cflags",
													"try_parse_field_width",
													"try_parse_length_modifier",
													"try_parse_mapping_key",
													"try_parse_precision",
													"vals",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(442),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(793),
														Column: int(48),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(1),
													Indent: int(4),
												},
											},
											Ctx: p8301,
											FreeVars: ast.Identifiers{
												"$std",
												"std",
												"str",
												"try_parse_cflags",
												"try_parse_field_width",
												"try_parse_length_modifier",
												"try_parse_mapping_key",
												"try_parse_precision",
												"vals",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(408),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(793),
													Column: int(48),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(1),
												Indent: int(4),
											},
											ast.FodderElement{
												Comment: []string{
													"// Ignored, if it exists.",
												},
												Kind: ast.FodderKind(2),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p8301,
										FreeVars: ast.Identifiers{
											"$std",
											"std",
											"str",
											"try_parse_cflags",
											"try_parse_field_width",
											"try_parse_mapping_key",
											"try_parse_precision",
											"vals",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(400),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(793),
												Column: int(48),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(4),
										},
									},
									Ctx: p8301,
									FreeVars: ast.Identifiers{
										"$std",
										"std",
										"str",
										"try_parse_cflags",
										"try_parse_field_width",
										"try_parse_mapping_key",
										"vals",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(391),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(793),
											Column: int(48),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
									},
								},
								Ctx: p8301,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"str",
									"try_parse_cflags",
									"try_parse_mapping_key",
									"vals",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(360),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(793),
										Column: int(48),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
								},
							},
							Ctx: p8301,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"str",
								"try_parse_mapping_key",
								"vals",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(342),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(793),
									Column: int(48),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(1),
								Indent: int(4),
							},
							ast.FodderElement{
								Comment: []string{
									"/////////////////////////////",
								},
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
							},
							ast.FodderElement{
								Comment: []string{
									"// Parse the mini-language //",
								},
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
							},
							ast.FodderElement{
								Comment: []string{
									"/////////////////////////////",
								},
								Kind: ast.FodderKind(2),
								Blanks: int(1),
								Indent: int(4),
							},
						},
						Ctx: p8301,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"str",
							"vals",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(325),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(793),
								Column: int(48),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(319),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(319),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "vals",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(319),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(319),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(319),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(793),
					Column: int(48),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "foldr",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.LiteralNumber{
											OriginalString: "0",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8451,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(797),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(797),
														Column: int(17),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "idx",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8451,
												FreeVars: ast.Identifiers{
													"idx",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(797),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(797),
														Column: int(13),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8451,
											FreeVars: ast.Identifiers{
												"idx",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(797),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(797),
													Column: int(17),
												},
											},
										},
										Op: ast.BinaryOp(9),
									},
									BranchTrue: &ast.Var{
										Id: "running",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p8451,
											FreeVars: ast.Identifiers{
												"running",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(798),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(798),
													Column: int(16),
												},
											},
										},
									},
									BranchFalse: &ast.Apply{
										Target: &ast.Var{
											Id: "aux",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p8451,
												FreeVars: ast.Identifiers{
													"aux",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(800),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(800),
														Column: int(12),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "func",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8464,
															FreeVars: ast.Identifiers{
																"func",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(800),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(800),
																	Column: int(17),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8464,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(800),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(800),
																	Column: int(22),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Var{
															Id: "func",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8464,
																FreeVars: ast.Identifiers{
																	"func",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(800),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(800),
																		Column: int(28),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p8474,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(800),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(800),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "idx",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p8474,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(800),
																						Column: int(33),
																					},
																					End: ast.Location{
																						Line: int(800),
																						Column: int(36),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p8474,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"idx",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(800),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(800),
																					Column: int(37),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "running",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p8474,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(800),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(800),
																					Column: int(46),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8464,
															FreeVars: ast.Identifiers{
																"arr",
																"func",
																"idx",
																"running",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(800),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(800),
																	Column: int(47),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "1",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8464,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(800),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(800),
																		Column: int(56),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "idx",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8464,
																FreeVars: ast.Identifiers{
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(800),
																		Column: int(49),
																	},
																	End: ast.Location{
																		Line: int(800),
																		Column: int(52),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8464,
															FreeVars: ast.Identifiers{
																"idx",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(800),
																	Column: int(49),
																},
																End: ast.Location{
																	Line: int(800),
																	Column: int(56),
																},
															},
														},
														Op: ast.BinaryOp(4),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8451,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"func",
												"idx",
												"running",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(800),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(800),
													Column: int(57),
												},
											},
										},
										TrailingComma: false,
										TailStrict: true,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p8451,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
											"func",
											"idx",
											"running",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(797),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(800),
												Column: int(57),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "func",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(796),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(796),
												Column: int(19),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(796),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(796),
												Column: int(24),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(796),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(796),
												Column: int(33),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "idx",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(796),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(796),
												Column: int(38),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p8492,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(796),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(800),
											Column: int(57),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p8497,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(801),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(801),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "func",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8501,
											FreeVars: ast.Identifiers{
												"func",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(801),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(801),
													Column: int(13),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8501,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(801),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(801),
													Column: int(18),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "init",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8501,
											FreeVars: ast.Identifiers{
												"init",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(801),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(801),
													Column: int(24),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										Right: &ast.LiteralNumber{
											OriginalString: "1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8501,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(801),
														Column: int(44),
													},
													End: ast.Location{
														Line: int(801),
														Column: int(45),
													},
												},
											},
										},
										Left: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(801),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(801),
																Column: int(29),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8501,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(801),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(801),
															Column: int(36),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8517,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(801),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(801),
																		Column: int(40),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8501,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(801),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(801),
														Column: int(41),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8501,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(801),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(801),
													Column: int(45),
												},
											},
										},
										Op: ast.BinaryOp(4),
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8497,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"func",
								"init",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(801),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(801),
									Column: int(46),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8497,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"init",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(796),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(801),
								Column: int(46),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(795),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(795),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(795),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(795),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "init",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(795),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(795),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(795),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(801),
					Column: int(46),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "foldl",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(805),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(805),
																Column: int(20),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8538,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(805),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(805),
															Column: int(27),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "arr",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8542,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(805),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(805),
																		Column: int(31),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8538,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(805),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(805),
														Column: int(32),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "idx",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8538,
												FreeVars: ast.Identifiers{
													"idx",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(805),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(805),
														Column: int(13),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8538,
											FreeVars: ast.Identifiers{
												"arr",
												"idx",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(805),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(805),
													Column: int(32),
												},
											},
										},
										Op: ast.BinaryOp(8),
									},
									BranchTrue: &ast.Var{
										Id: "running",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p8538,
											FreeVars: ast.Identifiers{
												"running",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(806),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(806),
													Column: int(16),
												},
											},
										},
									},
									BranchFalse: &ast.Apply{
										Target: &ast.Var{
											Id: "aux",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p8538,
												FreeVars: ast.Identifiers{
													"aux",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(808),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(808),
														Column: int(12),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "func",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8557,
															FreeVars: ast.Identifiers{
																"func",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(808),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(808),
																	Column: int(17),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8557,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(808),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(808),
																	Column: int(22),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Var{
															Id: "func",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8557,
																FreeVars: ast.Identifiers{
																	"func",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(808),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(808),
																		Column: int(28),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "running",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p8566,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(808),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(808),
																					Column: int(36),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p8566,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(808),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(808),
																						Column: int(41),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "idx",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p8566,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(808),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(808),
																						Column: int(45),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p8566,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"idx",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(808),
																					Column: int(38),
																				},
																				End: ast.Location{
																					Line: int(808),
																					Column: int(46),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8557,
															FreeVars: ast.Identifiers{
																"arr",
																"func",
																"idx",
																"running",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(808),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(808),
																	Column: int(47),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "1",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8557,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(808),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(808),
																		Column: int(56),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "idx",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p8557,
																FreeVars: ast.Identifiers{
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(808),
																		Column: int(49),
																	},
																	End: ast.Location{
																		Line: int(808),
																		Column: int(52),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8557,
															FreeVars: ast.Identifiers{
																"idx",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(808),
																	Column: int(49),
																},
																End: ast.Location{
																	Line: int(808),
																	Column: int(56),
																},
															},
														},
														Op: ast.BinaryOp(3),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8538,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"func",
												"idx",
												"running",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(808),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(808),
													Column: int(57),
												},
											},
										},
										TrailingComma: false,
										TailStrict: true,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p8538,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
											"func",
											"idx",
											"running",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(805),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(808),
												Column: int(57),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "func",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(804),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(804),
												Column: int(19),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(804),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(804),
												Column: int(24),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(804),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(804),
												Column: int(33),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "idx",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(804),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(804),
												Column: int(38),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p8585,
									FreeVars: ast.Identifiers{
										"aux",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(804),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(808),
											Column: int(57),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p8590,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(809),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(809),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "func",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8594,
											FreeVars: ast.Identifiers{
												"func",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(809),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(809),
													Column: int(13),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8594,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(809),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(809),
													Column: int(18),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "init",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8594,
											FreeVars: ast.Identifiers{
												"init",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(809),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(809),
													Column: int(24),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8594,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(809),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(809),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8590,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"func",
								"init",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(809),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(809),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8590,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"init",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(804),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(809),
								Column: int(28),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(803),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(803),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(803),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(803),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "init",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(803),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(803),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(803),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(809),
					Column: int(28),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "filterMap",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(813),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(813),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isFunction",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(813),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(813),
											Column: int(23),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "filter_func",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8619,
												FreeVars: ast.Identifiers{
													"filter_func",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(813),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(813),
														Column: int(35),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8615,
								FreeVars: ast.Identifiers{
									"filter_func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(813),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(813),
										Column: int(36),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8615,
							FreeVars: ast.Identifiers{
								"filter_func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(813),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(813),
									Column: int(36),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(814),
													Column: int(67),
												},
												End: ast.Location{
													Line: int(814),
													Column: int(70),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(814),
												Column: int(67),
											},
											End: ast.Location{
												Line: int(814),
												Column: int(75),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "filter_func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8633,
													FreeVars: ast.Identifiers{
														"filter_func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(814),
															Column: int(76),
														},
														End: ast.Location{
															Line: int(814),
															Column: int(87),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"filter_func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(814),
											Column: int(67),
										},
										End: ast.Location{
											Line: int(814),
											Column: int(88),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.filterMap first param must be function, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(814),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(814),
											Column: int(64),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8615,
								FreeVars: ast.Identifiers{
									"filter_func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(814),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(814),
										Column: int(88),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8615,
							FreeVars: ast.Identifiers{
								"filter_func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(814),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(814),
									Column: int(89),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(815),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(815),
													Column: int(17),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isFunction",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(815),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(815),
												Column: int(28),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "map_func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8650,
													FreeVars: ast.Identifiers{
														"map_func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(815),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(815),
															Column: int(37),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"map_func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(815),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(815),
											Column: int(38),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8615,
								FreeVars: ast.Identifiers{
									"map_func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(815),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(815),
										Column: int(38),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(816),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(816),
														Column: int(71),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8615,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(816),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(816),
													Column: int(76),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "map_func",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p8664,
														FreeVars: ast.Identifiers{
															"map_func",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(816),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(816),
																Column: int(85),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"map_func",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(816),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(816),
												Column: int(86),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.filterMap second param must be function, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(816),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(816),
												Column: int(65),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"map_func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(816),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(816),
											Column: int(86),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p8615,
								FreeVars: ast.Identifiers{
									"map_func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(816),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(816),
										Column: int(87),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(817),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(817),
														Column: int(17),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isArray",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8615,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(817),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(817),
													Column: int(25),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p8681,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(817),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(817),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(817),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(817),
												Column: int(30),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(817),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(817),
											Column: int(30),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							BranchTrue: &ast.Error{
								Expr: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(818),
															Column: int(64),
														},
														End: ast.Location{
															Line: int(818),
															Column: int(67),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "type",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8615,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(818),
														Column: int(64),
													},
													End: ast.Location{
														Line: int(818),
														Column: int(72),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p8695,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(818),
																	Column: int(73),
																},
																End: ast.Location{
																	Line: int(818),
																	Column: int(76),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8615,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(818),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(818),
													Column: int(77),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.LiteralString{
										Value: "std.filterMap third param must be array, got ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8615,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(818),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(818),
													Column: int(61),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(818),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(818),
												Column: int(77),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(818),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(818),
											Column: int(78),
										},
									},
								},
							},
							BranchFalse: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(820),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(820),
													Column: int(10),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "map",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8615,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(820),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(820),
												Column: int(14),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "map_func",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8711,
													FreeVars: ast.Identifiers{
														"map_func",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(820),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(820),
															Column: int(23),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(820),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(820),
																	Column: int(28),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "filter",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p8711,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(820),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(820),
																Column: int(35),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "filter_func",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p8721,
																	FreeVars: ast.Identifiers{
																		"filter_func",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(820),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(820),
																			Column: int(47),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p8721,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(820),
																			Column: int(49),
																		},
																		End: ast.Location{
																			Line: int(820),
																			Column: int(52),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8711,
													FreeVars: ast.Identifiers{
														"arr",
														"filter_func",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(820),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(820),
															Column: int(53),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8615,
									FreeVars: ast.Identifiers{
										"arr",
										"filter_func",
										"map_func",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(820),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(820),
											Column: int(54),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8615,
								FreeVars: ast.Identifiers{
									"arr",
									"filter_func",
									"map_func",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(817),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(820),
										Column: int(54),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8615,
							FreeVars: ast.Identifiers{
								"arr",
								"filter_func",
								"map_func",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(815),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(820),
									Column: int(54),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8615,
						FreeVars: ast.Identifiers{
							"arr",
							"filter_func",
							"map_func",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(813),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(820),
								Column: int(54),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "filter_func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(812),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(812),
								Column: int(24),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "map_func",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(812),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(812),
								Column: int(34),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(812),
								Column: int(36),
							},
							End: ast.Location{
								Line: int(812),
								Column: int(39),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(812),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(820),
					Column: int(54),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "assertEqual",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Var{
							Id: "b",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8741,
								FreeVars: ast.Identifiers{
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(823),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(823),
										Column: int(14),
									},
								},
							},
						},
						Left: &ast.Var{
							Id: "a",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8741,
								FreeVars: ast.Identifiers{
									"a",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(823),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(823),
										Column: int(9),
									},
								},
							},
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8741,
							FreeVars: ast.Identifiers{
								"a",
								"b",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(823),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(823),
									Column: int(14),
								},
							},
						},
						Op: ast.BinaryOp(12),
					},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8741,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(824),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(824),
									Column: int(11),
								},
							},
						},
						Value: true,
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Var{
								Id: "b",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8741,
									FreeVars: ast.Identifiers{
										"b",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(826),
											Column: int(49),
										},
										End: ast.Location{
											Line: int(826),
											Column: int(50),
										},
									},
								},
							},
							Left: &ast.Binary{
								Right: &ast.LiteralString{
									Value: " != ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8741,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(826),
												Column: int(40),
											},
											End: ast.Location{
												Line: int(826),
												Column: int(46),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								Left: &ast.Binary{
									Right: &ast.Var{
										Id: "a",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8741,
											FreeVars: ast.Identifiers{
												"a",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(826),
													Column: int(36),
												},
												End: ast.Location{
													Line: int(826),
													Column: int(37),
												},
											},
										},
									},
									Left: &ast.LiteralString{
										Value: "Assertion failed. ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8741,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(826),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(826),
													Column: int(33),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8741,
										FreeVars: ast.Identifiers{
											"a",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(826),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(826),
												Column: int(37),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8741,
									FreeVars: ast.Identifiers{
										"a",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(826),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(826),
											Column: int(46),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8741,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(826),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(826),
										Column: int(50),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8741,
							FreeVars: ast.Identifiers{
								"a",
								"b",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(826),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(826),
									Column: int(50),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8741,
						FreeVars: ast.Identifiers{
							"a",
							"b",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(823),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(826),
								Column: int(50),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(822),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(822),
								Column: int(16),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(822),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(822),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(822),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(826),
					Column: int(50),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "abs",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(829),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(829),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(829),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(829),
											Column: int(21),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "n",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8780,
												FreeVars: ast.Identifiers{
													"n",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(829),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(829),
														Column: int(23),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8776,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(829),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(829),
										Column: int(24),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8776,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(829),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(829),
									Column: int(24),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(830),
													Column: int(47),
												},
												End: ast.Location{
													Line: int(830),
													Column: int(50),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8776,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(830),
												Column: int(47),
											},
											End: ast.Location{
												Line: int(830),
												Column: int(55),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "n",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8794,
													FreeVars: ast.Identifiers{
														"n",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(830),
															Column: int(56),
														},
														End: ast.Location{
															Line: int(830),
															Column: int(57),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{
										"n",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(830),
											Column: int(47),
										},
										End: ast.Location{
											Line: int(830),
											Column: int(58),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.abs expected number, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(830),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(830),
											Column: int(44),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8776,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(830),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(830),
										Column: int(58),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8776,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(830),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(830),
									Column: int(58),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(832),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(832),
											Column: int(15),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "n",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{
										"n",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(832),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(832),
											Column: int(11),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8776,
								FreeVars: ast.Identifiers{
									"n",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(832),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(832),
										Column: int(15),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchTrue: &ast.Var{
							Id: "n",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8776,
								FreeVars: ast.Identifiers{
									"n",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(832),
										Column: int(21),
									},
									End: ast.Location{
										Line: int(832),
										Column: int(22),
									},
								},
							},
						},
						BranchFalse: &ast.Unary{
							Expr: &ast.Var{
								Id: "n",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8776,
									FreeVars: ast.Identifiers{
										"n",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(832),
											Column: int(29),
										},
										End: ast.Location{
											Line: int(832),
											Column: int(30),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8776,
								FreeVars: ast.Identifiers{
									"n",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(832),
										Column: int(28),
									},
									End: ast.Location{
										Line: int(832),
										Column: int(30),
									},
								},
							},
							Op: ast.UnaryOp(3),
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8776,
							FreeVars: ast.Identifiers{
								"n",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(832),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(832),
									Column: int(30),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8776,
						FreeVars: ast.Identifiers{
							"n",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(829),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(832),
								Column: int(30),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "n",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(828),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(828),
								Column: int(8),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(828),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(832),
					Column: int(30),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "sign",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(835),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(835),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(835),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(835),
											Column: int(21),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "n",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8833,
												FreeVars: ast.Identifiers{
													"n",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(835),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(835),
														Column: int(23),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8829,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(835),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(835),
										Column: int(24),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8829,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(835),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(835),
									Column: int(24),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(836),
													Column: int(48),
												},
												End: ast.Location{
													Line: int(836),
													Column: int(51),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8829,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(836),
												Column: int(48),
											},
											End: ast.Location{
												Line: int(836),
												Column: int(56),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "n",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8847,
													FreeVars: ast.Identifiers{
														"n",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(836),
															Column: int(57),
														},
														End: ast.Location{
															Line: int(836),
															Column: int(58),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{
										"n",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(836),
											Column: int(48),
										},
										End: ast.Location{
											Line: int(836),
											Column: int(59),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.sign expected number, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(836),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(836),
											Column: int(45),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8829,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(836),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(836),
										Column: int(59),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8829,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(836),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(836),
									Column: int(59),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(838),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(838),
											Column: int(15),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "n",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{
										"n",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(838),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(838),
											Column: int(11),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8829,
								FreeVars: ast.Identifiers{
									"n",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(838),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(838),
										Column: int(15),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchTrue: &ast.LiteralNumber{
							OriginalString: "1",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(8),
									},
								},
								Ctx: p8829,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(839),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(839),
										Column: int(10),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8829,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(840),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(840),
												Column: int(20),
											},
										},
									},
								},
								Left: &ast.Var{
									Id: "n",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8829,
										FreeVars: ast.Identifiers{
											"n",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(840),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(840),
												Column: int(16),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{
										"n",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(840),
											Column: int(15),
										},
										End: ast.Location{
											Line: int(840),
											Column: int(20),
										},
									},
								},
								Op: ast.BinaryOp(9),
							},
							BranchTrue: &ast.Unary{
								Expr: &ast.LiteralNumber{
									OriginalString: "1",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8829,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(841),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(841),
												Column: int(11),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(8),
										},
									},
									Ctx: p8829,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(841),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(841),
											Column: int(11),
										},
									},
								},
								Op: ast.UnaryOp(3),
							},
							BranchFalse: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8829,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(842),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(842),
											Column: int(13),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8829,
								FreeVars: ast.Identifiers{
									"n",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(840),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(842),
										Column: int(13),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(6),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8829,
							FreeVars: ast.Identifiers{
								"n",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(838),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(842),
									Column: int(13),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8829,
						FreeVars: ast.Identifiers{
							"n",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(835),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(842),
								Column: int(13),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "n",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(834),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(834),
								Column: int(9),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(834),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(842),
					Column: int(13),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "max",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(845),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(845),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(845),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(845),
											Column: int(21),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8895,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(845),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(845),
														Column: int(23),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8891,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(845),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(845),
										Column: int(24),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8891,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(845),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(845),
									Column: int(24),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(846),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(846),
													Column: int(62),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(846),
												Column: int(59),
											},
											End: ast.Location{
												Line: int(846),
												Column: int(67),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "a",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8909,
													FreeVars: ast.Identifiers{
														"a",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(846),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(846),
															Column: int(69),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(846),
											Column: int(59),
										},
										End: ast.Location{
											Line: int(846),
											Column: int(70),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.max first param expected number, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(846),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(846),
											Column: int(56),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8891,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(846),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(846),
										Column: int(70),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8891,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(846),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(846),
									Column: int(70),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(847),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(847),
													Column: int(17),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isNumber",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(847),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(847),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "b",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8926,
													FreeVars: ast.Identifiers{
														"b",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(847),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(847),
															Column: int(28),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(847),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(847),
											Column: int(29),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8891,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(847),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(847),
										Column: int(29),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(848),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(848),
														Column: int(63),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8891,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(848),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(848),
													Column: int(68),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "b",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p8940,
														FreeVars: ast.Identifiers{
															"b",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(848),
																Column: int(69),
															},
															End: ast.Location{
																Line: int(848),
																Column: int(70),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(848),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(848),
												Column: int(71),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.max second param expected number, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(848),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(848),
												Column: int(57),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(848),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(848),
											Column: int(71),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p8891,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(848),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(848),
										Column: int(71),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.Var{
									Id: "b",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{
											"b",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(850),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(850),
												Column: int(15),
											},
										},
									},
								},
								Left: &ast.Var{
									Id: "a",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8891,
										FreeVars: ast.Identifiers{
											"a",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(850),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(850),
												Column: int(11),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"a",
										"b",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(850),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(850),
											Column: int(15),
										},
									},
								},
								Op: ast.BinaryOp(7),
							},
							BranchTrue: &ast.Var{
								Id: "a",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"a",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(850),
											Column: int(21),
										},
										End: ast.Location{
											Line: int(850),
											Column: int(22),
										},
									},
								},
							},
							BranchFalse: &ast.Var{
								Id: "b",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8891,
									FreeVars: ast.Identifiers{
										"b",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(850),
											Column: int(28),
										},
										End: ast.Location{
											Line: int(850),
											Column: int(29),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p8891,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(850),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(850),
										Column: int(29),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8891,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(847),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(850),
									Column: int(29),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8891,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(845),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(850),
								Column: int(29),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(844),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(844),
								Column: int(8),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(844),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(844),
								Column: int(11),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(844),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(850),
					Column: int(29),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "min",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(853),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(853),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isNumber",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(853),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(853),
											Column: int(21),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p8980,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(853),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(853),
														Column: int(23),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8976,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(853),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(853),
										Column: int(24),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8976,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(853),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(853),
									Column: int(24),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(854),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(854),
													Column: int(62),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(854),
												Column: int(59),
											},
											End: ast.Location{
												Line: int(854),
												Column: int(67),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "a",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p8994,
													FreeVars: ast.Identifiers{
														"a",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(854),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(854),
															Column: int(69),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(854),
											Column: int(59),
										},
										End: ast.Location{
											Line: int(854),
											Column: int(70),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "std.min first param expected number, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(854),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(854),
											Column: int(56),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8976,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(854),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(854),
										Column: int(70),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p8976,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(854),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(854),
									Column: int(70),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(855),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(855),
													Column: int(17),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isNumber",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(855),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(855),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "b",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9011,
													FreeVars: ast.Identifiers{
														"b",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(855),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(855),
															Column: int(28),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(855),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(855),
											Column: int(29),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p8976,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(855),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(855),
										Column: int(29),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(856),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(856),
														Column: int(63),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p8976,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(856),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(856),
													Column: int(68),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "b",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9025,
														FreeVars: ast.Identifiers{
															"b",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(856),
																Column: int(69),
															},
															End: ast.Location{
																Line: int(856),
																Column: int(70),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(856),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(856),
												Column: int(71),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "std.min second param expected number, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(856),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(856),
												Column: int(57),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(856),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(856),
											Column: int(71),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p8976,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(856),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(856),
										Column: int(71),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.Var{
									Id: "b",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{
											"b",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(858),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(858),
												Column: int(15),
											},
										},
									},
								},
								Left: &ast.Var{
									Id: "a",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p8976,
										FreeVars: ast.Identifiers{
											"a",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(858),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(858),
												Column: int(11),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"a",
										"b",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(858),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(858),
											Column: int(15),
										},
									},
								},
								Op: ast.BinaryOp(9),
							},
							BranchTrue: &ast.Var{
								Id: "a",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"a",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(858),
											Column: int(21),
										},
										End: ast.Location{
											Line: int(858),
											Column: int(22),
										},
									},
								},
							},
							BranchFalse: &ast.Var{
								Id: "b",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p8976,
									FreeVars: ast.Identifiers{
										"b",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(858),
											Column: int(28),
										},
										End: ast.Location{
											Line: int(858),
											Column: int(29),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p8976,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(858),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(858),
										Column: int(29),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p8976,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(855),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(858),
									Column: int(29),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p8976,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(853),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(858),
								Column: int(29),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(852),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(852),
								Column: int(8),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(852),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(852),
								Column: int(11),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(852),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(858),
					Column: int(29),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "clamp",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Var{
							Id: "minVal",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9057,
								FreeVars: ast.Identifiers{
									"minVal",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(861),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(861),
										Column: int(18),
									},
								},
							},
						},
						Left: &ast.Var{
							Id: "x",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9057,
								FreeVars: ast.Identifiers{
									"x",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(861),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(861),
										Column: int(9),
									},
								},
							},
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p9057,
							FreeVars: ast.Identifiers{
								"minVal",
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(861),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(861),
									Column: int(18),
								},
							},
						},
						Op: ast.BinaryOp(9),
					},
					BranchTrue: &ast.Var{
						Id: "minVal",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p9057,
							FreeVars: ast.Identifiers{
								"minVal",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(861),
									Column: int(24),
								},
								End: ast.Location{
									Line: int(861),
									Column: int(30),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Var{
								Id: "maxVal",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p9057,
									FreeVars: ast.Identifiers{
										"maxVal",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(862),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(862),
											Column: int(23),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "x",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p9057,
									FreeVars: ast.Identifiers{
										"x",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(862),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(862),
											Column: int(14),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9057,
								FreeVars: ast.Identifiers{
									"maxVal",
									"x",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(862),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(862),
										Column: int(23),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						BranchTrue: &ast.Var{
							Id: "maxVal",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9057,
								FreeVars: ast.Identifiers{
									"maxVal",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(862),
										Column: int(29),
									},
									End: ast.Location{
										Line: int(862),
										Column: int(35),
									},
								},
							},
						},
						BranchFalse: &ast.Var{
							Id: "x",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9057,
								FreeVars: ast.Identifiers{
									"x",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(863),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(863),
										Column: int(11),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p9057,
							FreeVars: ast.Identifiers{
								"maxVal",
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(862),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(863),
									Column: int(11),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p9057,
						FreeVars: ast.Identifiers{
							"maxVal",
							"minVal",
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(861),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(863),
								Column: int(11),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(860),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(860),
								Column: int(10),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "minVal",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(860),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(860),
								Column: int(18),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "maxVal",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(860),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(860),
								Column: int(26),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(860),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(863),
					Column: int(11),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "flattenArrays",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(866),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(866),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "foldl",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p9089,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(866),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(866),
									Column: int(14),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										Right: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9095,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(866),
														Column: int(34),
													},
													End: ast.Location{
														Line: int(866),
														Column: int(35),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9095,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(866),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(866),
														Column: int(31),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9095,
											FreeVars: ast.Identifiers{
												"a",
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(866),
													Column: int(30),
												},
												End: ast.Location{
													Line: int(866),
													Column: int(35),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "a",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(866),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(866),
													Column: int(25),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "b",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(866),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(866),
													Column: int(28),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9101,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(866),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(866),
												Column: int(35),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "arrs",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9101,
										FreeVars: ast.Identifiers{
											"arrs",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(866),
												Column: int(37),
											},
											End: ast.Location{
												Line: int(866),
												Column: int(41),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Array{
									Elements: nil,
									CloseFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9101,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(866),
												Column: int(43),
											},
											End: ast.Location{
												Line: int(866),
												Column: int(45),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p9089,
						FreeVars: ast.Identifiers{
							"arrs",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(866),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(866),
								Column: int(46),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arrs",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(865),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(865),
								Column: int(21),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(865),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(866),
					Column: int(46),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestIni",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(870),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(870),
														Column: int(10),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "join",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9119,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(870),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(870),
													Column: int(15),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Array{
													Elements: nil,
													CloseFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9123,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(870),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(870),
																Column: int(18),
															},
														},
													},
													TrailingComma: false,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "flatMap",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Function{
																	ParenLeftFodder: nil,
																	ParenRightFodder: nil,
																	Body: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Local{
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Body: &ast.Index{
																								Target: &ast.Var{
																									Id: "body",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9138,
																										FreeVars: ast.Identifiers{
																											"body",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(871),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(871),
																												Column: int(37),
																											},
																										},
																									},
																								},
																								Index: &ast.Var{
																									Id: "k",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9138,
																										FreeVars: ast.Identifiers{
																											"k",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(871),
																												Column: int(38),
																											},
																											End: ast.Location{
																												Line: int(871),
																												Column: int(39),
																											},
																										},
																									},
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9138,
																									FreeVars: ast.Identifiers{
																										"body",
																										"k",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(871),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(871),
																											Column: int(40),
																										},
																									},
																								},
																							},
																							EqFodder: ast.Fodder{},
																							Variable: "value_or_values",
																							CloseFodder: ast.Fodder{},
																							Fun: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(871),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(871),
																									Column: int(40),
																								},
																							},
																						},
																					},
																					Body: &ast.Conditional{
																						Cond: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(872),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(872),
																												Column: int(15),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "isArray",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9149,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(872),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(872),
																											Column: int(23),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "value_or_values",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p9153,
																												FreeVars: ast.Identifiers{
																													"value_or_values",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(872),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(872),
																														Column: int(39),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9149,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value_or_values",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(872),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(872),
																										Column: int(40),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						BranchTrue: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "$std",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"$std",
																										},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "flatMap",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: nil,
																								LeftBracketFodder: nil,
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"$std",
																									},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							FodderLeft: nil,
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Function{
																											ParenLeftFodder: nil,
																											ParenRightFodder: nil,
																											Body: &ast.Array{
																												Elements: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "$std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "mod",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: nil,
																																LeftBracketFodder: nil,
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralString{
																																			Value: "%s = %s",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p9174,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(873),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(873),
																																						Column: int(21),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Array{
																																			Elements: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "k",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p9178,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(873),
																																									Column: int(25),
																																								},
																																								End: ast.Location{
																																									Line: int(873),
																																									Column: int(26),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: ast.Fodder{},
																																				},
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "value",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p9178,
																																							FreeVars: ast.Identifiers{
																																								"value",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(873),
																																									Column: int(28),
																																								},
																																								End: ast.Location{
																																									Line: int(873),
																																									Column: int(33),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			CloseFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p9174,
																																				FreeVars: ast.Identifiers{
																																					"k",
																																					"value",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(873),
																																						Column: int(24),
																																					},
																																					End: ast.Location{
																																						Line: int(873),
																																						Column: int(34),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: nil,
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																	"k",
																																	"value",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(873),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(873),
																																		Column: int(34),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														CommaFodder: nil,
																													},
																												},
																												CloseFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																														"k",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												TrailingComma: false,
																											},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: nil,
																													Name: "value",
																													CommaFodder: nil,
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"k",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "value_or_values",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p9149,
																												FreeVars: ast.Identifiers{
																													"value_or_values",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(873),
																														Column: int(48),
																													},
																													End: ast.Location{
																														Line: int(873),
																														Column: int(63),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: nil,
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"k",
																									"value_or_values",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(873),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(873),
																										Column: int(64),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						BranchFalse: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "$std",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "mod",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: nil,
																											LeftBracketFodder: nil,
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																										},
																										FodderLeft: nil,
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "%s = %s",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9200,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(875),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(875),
																																	Column: int(21),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: nil,
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Array{
																														Elements: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "k",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p9204,
																																		FreeVars: ast.Identifiers{
																																			"k",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(875),
																																				Column: int(25),
																																			},
																																			End: ast.Location{
																																				Line: int(875),
																																				Column: int(26),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "value_or_values",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p9204,
																																		FreeVars: ast.Identifiers{
																																			"value_or_values",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(875),
																																				Column: int(28),
																																			},
																																			End: ast.Location{
																																				Line: int(875),
																																				Column: int(43),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														CloseFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9200,
																															FreeVars: ast.Identifiers{
																																"k",
																																"value_or_values",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(875),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(875),
																																	Column: int(44),
																																},
																															},
																														},
																														TrailingComma: false,
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: nil,
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$std",
																												"k",
																												"value_or_values",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(875),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(875),
																													Column: int(44),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p9149,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"k",
																									"value_or_values",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(875),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(875),
																										Column: int(45),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																								},
																							},
																							Ctx: p9149,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"k",
																								"std",
																								"value_or_values",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(872),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(875),
																									Column: int(45),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: p9149,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"body",
																							"k",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(871),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(875),
																								Column: int(45),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"body",
																				"k",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: nil,
																			Name: "k",
																			CommaFodder: nil,
																			EqFodder: nil,
																			DefaultArg: nil,
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"body",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(877),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(877),
																						Column: int(21),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "objectFields",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9123,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(877),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(877),
																					Column: int(34),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "body",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9228,
																						FreeVars: ast.Identifiers{
																							"body",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(877),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(877),
																								Column: int(39),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9123,
																		FreeVars: ast.Identifiers{
																			"body",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(877),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(877),
																				Column: int(40),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"body",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(870),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(878),
																Column: int(8),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9119,
										FreeVars: ast.Identifiers{
											"$std",
											"body",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(870),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(878),
												Column: int(9),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "body",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(869),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(869),
												Column: int(26),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p9234,
									FreeVars: ast.Identifiers{
										"$std",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(869),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(878),
											Column: int(9),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "body_lines",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Var{
												Id: "body_lines",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9242,
													FreeVars: ast.Identifiers{
														"body_lines",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(880),
															Column: int(62),
														},
														End: ast.Location{
															Line: int(880),
															Column: int(72),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "sbody",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9246,
																FreeVars: ast.Identifiers{
																	"sbody",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(880),
																		Column: int(73),
																	},
																	End: ast.Location{
																		Line: int(880),
																		Column: int(78),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9242,
												FreeVars: ast.Identifiers{
													"body_lines",
													"sbody",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(880),
														Column: int(62),
													},
													End: ast.Location{
														Line: int(880),
														Column: int(79),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Array{
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "$std",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "mod",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: nil,
															LeftBracketFodder: nil,
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		Value: "[%s]",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9259,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(880),
																					Column: int(42),
																				},
																				End: ast.Location{
																					Line: int(880),
																					Column: int(48),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "sname",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9263,
																						FreeVars: ast.Identifiers{
																							"sname",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(880),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(880),
																								Column: int(57),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9259,
																			FreeVars: ast.Identifiers{
																				"sname",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(880),
																					Column: int(51),
																				},
																				End: ast.Location{
																					Line: int(880),
																					Column: int(58),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: nil,
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"sname",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(880),
																	Column: int(42),
																},
																End: ast.Location{
																	Line: int(880),
																	Column: int(58),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9242,
												FreeVars: ast.Identifiers{
													"$std",
													"sname",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(880),
														Column: int(41),
													},
													End: ast.Location{
														Line: int(880),
														Column: int(59),
													},
												},
											},
											TrailingComma: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9242,
											FreeVars: ast.Identifiers{
												"$std",
												"body_lines",
												"sbody",
												"sname",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(880),
													Column: int(41),
												},
												End: ast.Location{
													Line: int(880),
													Column: int(79),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "sname",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(880),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(880),
													Column: int(30),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "sbody",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(880),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(880),
													Column: int(37),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p9270,
										FreeVars: ast.Identifiers{
											"$std",
											"body_lines",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(880),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(880),
												Column: int(79),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "section_lines",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							ast.LocalBind{
								VarFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(10),
									},
								},
								Body: &ast.Conditional{
									Cond: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(881),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(881),
															Column: int(29),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "objectHas",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9279,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(881),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(881),
														Column: int(39),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "ini",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9283,
															FreeVars: ast.Identifiers{
																"ini",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(881),
																	Column: int(40),
																},
																End: ast.Location{
																	Line: int(881),
																	Column: int(43),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														Value: "main",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9283,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(881),
																	Column: int(45),
																},
																End: ast.Location{
																	Line: int(881),
																	Column: int(51),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9279,
											FreeVars: ast.Identifiers{
												"ini",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(881),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(881),
													Column: int(52),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchTrue: &ast.Apply{
										Target: &ast.Var{
											Id: "body_lines",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9279,
												FreeVars: ast.Identifiers{
													"body_lines",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(881),
														Column: int(58),
													},
													End: ast.Location{
														Line: int(881),
														Column: int(68),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Index{
														Target: &ast.Var{
															Id: "ini",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"ini",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(881),
																		Column: int(69),
																	},
																	End: ast.Location{
																		Line: int(881),
																		Column: int(72),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "main",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9295,
															FreeVars: ast.Identifiers{
																"ini",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(881),
																	Column: int(69),
																},
																End: ast.Location{
																	Line: int(881),
																	Column: int(77),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9279,
											FreeVars: ast.Identifiers{
												"body_lines",
												"ini",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(881),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(881),
													Column: int(78),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchFalse: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9279,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(881),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(881),
													Column: int(86),
												},
											},
										},
										TrailingComma: false,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9279,
										FreeVars: ast.Identifiers{
											"body_lines",
											"ini",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(881),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(881),
												Column: int(86),
											},
										},
									},
								},
								EqFodder: ast.Fodder{},
								Variable: "main_body",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(881),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(881),
										Column: int(86),
									},
								},
							},
							ast.LocalBind{
								VarFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(10),
									},
								},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "$std",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "flatMap",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: nil,
										LeftBracketFodder: nil,
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													ParenLeftFodder: nil,
													ParenRightFodder: nil,
													Body: &ast.Array{
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Var{
																		Id: "section_lines",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			Ctx: p9314,
																			FreeVars: ast.Identifiers{
																				"section_lines",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(883),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(883),
																					Column: int(20),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "k",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9318,
																						FreeVars: ast.Identifiers{
																							"k",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(883),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(883),
																								Column: int(22),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "ini",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"ini",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(883),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(883),
																										Column: int(27),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "sections",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9318,
																							FreeVars: ast.Identifiers{
																								"ini",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(883),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(883),
																									Column: int(36),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "k",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9318,
																							FreeVars: ast.Identifiers{
																								"k",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(883),
																									Column: int(37),
																								},
																								End: ast.Location{
																									Line: int(883),
																									Column: int(38),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9318,
																						FreeVars: ast.Identifiers{
																							"ini",
																							"k",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(883),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(883),
																								Column: int(39),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9314,
																		FreeVars: ast.Identifiers{
																			"ini",
																			"k",
																			"section_lines",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(883),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(883),
																				Column: int(40),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														CloseFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"ini",
																"k",
																"section_lines",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														TrailingComma: false,
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: nil,
															Name: "k",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"ini",
															"section_lines",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(884),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(884),
																		Column: int(19),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "objectFields",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9338,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(884),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(884),
																	Column: int(32),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "ini",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"ini",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(884),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(884),
																					Column: int(36),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "sections",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9345,
																		FreeVars: ast.Identifiers{
																			"ini",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(884),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(884),
																				Column: int(45),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9338,
														FreeVars: ast.Identifiers{
															"ini",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(884),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(884),
																Column: int(46),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: nil,
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
											"ini",
											"section_lines",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(882),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(885),
												Column: int(6),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "all_sections",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(882),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(885),
										Column: int(6),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(886),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(886),
												Column: int(8),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "join",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p9355,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(886),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(886),
											Column: int(13),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "\n",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9359,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(886),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(886),
														Column: int(18),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Binary{
											Right: &ast.Array{
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9364,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(886),
																		Column: int(67),
																	},
																	End: ast.Location{
																		Line: int(886),
																		Column: int(69),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: nil,
													},
												},
												CloseFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9359,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(886),
															Column: int(66),
														},
														End: ast.Location{
															Line: int(886),
															Column: int(70),
														},
													},
												},
												TrailingComma: false,
											},
											Left: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(886),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(886),
																		Column: int(35),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "flattenArrays",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9359,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(886),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(886),
																	Column: int(49),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "all_sections",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9374,
																		FreeVars: ast.Identifiers{
																			"all_sections",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(886),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(886),
																				Column: int(62),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9359,
														FreeVars: ast.Identifiers{
															"all_sections",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(886),
																Column: int(32),
															},
															End: ast.Location{
																Line: int(886),
																Column: int(63),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Var{
													Id: "main_body",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9359,
														FreeVars: ast.Identifiers{
															"main_body",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(886),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(886),
																Column: int(29),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9359,
													FreeVars: ast.Identifiers{
														"all_sections",
														"main_body",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(886),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(886),
															Column: int(63),
														},
													},
												},
												Op: ast.BinaryOp(3),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9359,
												FreeVars: ast.Identifiers{
													"all_sections",
													"main_body",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(886),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(886),
														Column: int(70),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p9355,
								FreeVars: ast.Identifiers{
									"all_sections",
									"main_body",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(886),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(886),
										Column: int(71),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
								},
							},
							Ctx: p9355,
							FreeVars: ast.Identifiers{
								"$std",
								"body_lines",
								"ini",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(880),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(886),
									Column: int(71),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p9355,
						FreeVars: ast.Identifiers{
							"$std",
							"ini",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(869),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(886),
								Column: int(71),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "ini",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(868),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(868),
								Column: int(18),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(868),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(886),
					Column: int(71),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestToml",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(888),
										Column: int(25),
									},
									End: ast.Location{
										Line: int(888),
										Column: int(28),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "manifestTomlEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p9395,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(888),
									Column: int(25),
								},
								End: ast.Location{
									Line: int(888),
									Column: int(43),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "value",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9399,
										FreeVars: ast.Identifiers{
											"value",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(888),
												Column: int(44),
											},
											End: ast.Location{
												Line: int(888),
												Column: int(49),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: "  ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9399,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(888),
												Column: int(51),
											},
											End: ast.Location{
												Line: int(888),
												Column: int(55),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p9395,
						FreeVars: ast.Identifiers{
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(888),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(888),
								Column: int(56),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(888),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(888),
								Column: int(21),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(888),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(888),
					Column: int(56),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestTomlEx",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Body: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(892),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(892),
												Column: int(29),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "escapeStringJson",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p9414,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(892),
											Column: int(26),
										},
										End: ast.Location{
											Line: int(892),
											Column: int(46),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "escapeStringToml",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(892),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(892),
									Column: int(46),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(894),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(894),
																	Column: int(33),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "set",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9424,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(894),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(894),
																Column: int(37),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(894),
																					Column: int(38),
																				},
																				End: ast.Location{
																					Line: int(894),
																					Column: int(41),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "stringChars",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9432,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(894),
																				Column: int(38),
																			},
																			End: ast.Location{
																				Line: int(894),
																				Column: int(53),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralString{
																				Value: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p9436,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(894),
																							Column: int(54),
																						},
																						End: ast.Location{
																							Line: int(894),
																							Column: int(120),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9432,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(894),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(894),
																			Column: int(121),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9424,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(894),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(894),
															Column: int(122),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "bare_allowed",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(894),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(894),
													Column: int(122),
												},
											},
										},
									},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Var{
												Id: "bare_allowed",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9442,
													FreeVars: ast.Identifiers{
														"bare_allowed",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(895),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(895),
															Column: int(85),
														},
													},
												},
											},
											Left: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(895),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(895),
																	Column: int(15),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "setUnion",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9442,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(895),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(895),
																Column: int(24),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(895),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(895),
																					Column: int(28),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "set",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9456,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(895),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(895),
																				Column: int(32),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(895),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(895),
																									Column: int(36),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "stringChars",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9464,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(895),
																								Column: int(33),
																							},
																							End: ast.Location{
																								Line: int(895),
																								Column: int(48),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "key",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9468,
																									FreeVars: ast.Identifiers{
																										"key",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(895),
																											Column: int(49),
																										},
																										End: ast.Location{
																											Line: int(895),
																											Column: int(52),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p9464,
																					FreeVars: ast.Identifiers{
																						"key",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(895),
																							Column: int(33),
																						},
																						End: ast.Location{
																							Line: int(895),
																							Column: int(53),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9456,
																	FreeVars: ast.Identifiers{
																		"key",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(895),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(895),
																			Column: int(54),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "bare_allowed",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9456,
																	FreeVars: ast.Identifiers{
																		"bare_allowed",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(895),
																			Column: int(56),
																		},
																		End: ast.Location{
																			Line: int(895),
																			Column: int(68),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9442,
													FreeVars: ast.Identifiers{
														"bare_allowed",
														"key",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(895),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(895),
															Column: int(69),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9442,
												FreeVars: ast.Identifiers{
													"bare_allowed",
													"key",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(895),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(895),
														Column: int(85),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.Var{
											Id: "key",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9442,
												FreeVars: ast.Identifiers{
													"key",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(895),
														Column: int(91),
													},
													End: ast.Location{
														Line: int(895),
														Column: int(94),
													},
												},
											},
										},
										BranchFalse: &ast.Apply{
											Target: &ast.Var{
												Id: "escapeStringToml",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9442,
													FreeVars: ast.Identifiers{
														"escapeStringToml",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(895),
															Column: int(100),
														},
														End: ast.Location{
															Line: int(895),
															Column: int(116),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "key",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9483,
																FreeVars: ast.Identifiers{
																	"key",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(895),
																		Column: int(117),
																	},
																	End: ast.Location{
																		Line: int(895),
																		Column: int(120),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9442,
												FreeVars: ast.Identifiers{
													"escapeStringToml",
													"key",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(895),
														Column: int(100),
													},
													End: ast.Location{
														Line: int(895),
														Column: int(121),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p9442,
											FreeVars: ast.Identifiers{
												"bare_allowed",
												"escapeStringToml",
												"key",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(895),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(895),
													Column: int(121),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p9442,
										FreeVars: ast.Identifiers{
											"escapeStringToml",
											"key",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(894),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(895),
												Column: int(121),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "key",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(893),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(893),
												Column: int(24),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p9491,
									FreeVars: ast.Identifiers{
										"escapeStringToml",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(893),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(895),
											Column: int(121),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "escapeKeyToml",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(896),
															Column: int(64),
														},
														End: ast.Location{
															Line: int(896),
															Column: int(67),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "all",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9500,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(896),
														Column: int(64),
													},
													End: ast.Location{
														Line: int(896),
														Column: int(71),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(896),
																			Column: int(72),
																		},
																		End: ast.Location{
																			Line: int(896),
																			Column: int(75),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "map",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9508,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(896),
																		Column: int(72),
																	},
																	End: ast.Location{
																		Line: int(896),
																		Column: int(79),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(896),
																						Column: int(80),
																					},
																					End: ast.Location{
																						Line: int(896),
																						Column: int(83),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "isObject",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9515,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(896),
																					Column: int(80),
																				},
																				End: ast.Location{
																					Line: int(896),
																					Column: int(92),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9515,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(896),
																					Column: int(94),
																				},
																				End: ast.Location{
																					Line: int(896),
																					Column: int(95),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9508,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(896),
																	Column: int(72),
																},
																End: ast.Location{
																	Line: int(896),
																	Column: int(96),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9500,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(896),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(896),
													Column: int(97),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.Binary{
										Right: &ast.Binary{
											Right: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9500,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(896),
															Column: int(59),
														},
														End: ast.Location{
															Line: int(896),
															Column: int(60),
														},
													},
												},
											},
											Left: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(896),
																	Column: int(43),
																},
																End: ast.Location{
																	Line: int(896),
																	Column: int(46),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9500,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(896),
																Column: int(43),
															},
															End: ast.Location{
																Line: int(896),
																Column: int(53),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "v",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9532,
																	FreeVars: ast.Identifiers{
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(896),
																			Column: int(54),
																		},
																		End: ast.Location{
																			Line: int(896),
																			Column: int(55),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9500,
													FreeVars: ast.Identifiers{
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(896),
															Column: int(43),
														},
														End: ast.Location{
															Line: int(896),
															Column: int(56),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9500,
												FreeVars: ast.Identifiers{
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(896),
														Column: int(43),
													},
													End: ast.Location{
														Line: int(896),
														Column: int(60),
													},
												},
											},
											Op: ast.BinaryOp(7),
										},
										Left: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(896),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(896),
																Column: int(28),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "isArray",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9500,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(896),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(896),
															Column: int(36),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "v",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9544,
																FreeVars: ast.Identifiers{
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(896),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(896),
																		Column: int(38),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9500,
												FreeVars: ast.Identifiers{
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(896),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(896),
														Column: int(39),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9500,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(896),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(896),
													Column: int(60),
												},
											},
										},
										Op: ast.BinaryOp(17),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9500,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(896),
												Column: int(25),
											},
											End: ast.Location{
												Line: int(896),
												Column: int(97),
											},
										},
									},
									Op: ast.BinaryOp(17),
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(896),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(896),
												Column: int(21),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p9550,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(896),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(896),
											Column: int(97),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "isTableArray",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Var{
											Id: "isTableArray",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9556,
												FreeVars: ast.Identifiers{
													"isTableArray",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(897),
														Column: int(41),
													},
													End: ast.Location{
														Line: int(897),
														Column: int(53),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9560,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(897),
																	Column: int(54),
																},
																End: ast.Location{
																	Line: int(897),
																	Column: int(55),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9556,
											FreeVars: ast.Identifiers{
												"isTableArray",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(897),
													Column: int(41),
												},
												End: ast.Location{
													Line: int(897),
													Column: int(56),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(897),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(897),
															Column: int(25),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isObject",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9556,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(897),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(897),
														Column: int(34),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9571,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(897),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(897),
																	Column: int(36),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9556,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(897),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(897),
													Column: int(37),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p9556,
										FreeVars: ast.Identifiers{
											"isTableArray",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(897),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(897),
												Column: int(56),
											},
										},
									},
									Op: ast.BinaryOp(18),
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(897),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(897),
												Column: int(18),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p9576,
									FreeVars: ast.Identifiers{
										"isTableArray",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(897),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(897),
											Column: int(56),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "isSection",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9582,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(899),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(899),
														Column: int(21),
													},
												},
											},
											Value: true,
										},
										Left: &ast.Var{
											Id: "v",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9582,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(899),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(899),
														Column: int(13),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9582,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(899),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(899),
													Column: int(21),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.LiteralString{
										Value: "true",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
												},
											},
											Ctx: p9582,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(900),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(900),
													Column: int(17),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9582,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(901),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(901),
															Column: int(27),
														},
													},
												},
												Value: false,
											},
											Left: &ast.Var{
												Id: "v",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9582,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(901),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(901),
															Column: int(18),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9582,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(901),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(901),
														Column: int(27),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "false",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p9582,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(902),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(902),
														Column: int(18),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(903),
																Column: int(22),
															},
															End: ast.Location{
																Line: int(903),
																Column: int(26),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "v",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(903),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(903),
																Column: int(18),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9582,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(903),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(903),
															Column: int(26),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.Error{
												Expr: &ast.Binary{
													Right: &ast.Var{
														Id: "indexedPath",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"indexedPath",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(904),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(904),
																	Column: int(61),
																},
															},
														},
													},
													Left: &ast.LiteralString{
														Value: "Tried to manifest \"null\" at ",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(904),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(904),
																	Column: int(47),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{
															"indexedPath",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(904),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(904),
																Column: int(61),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
														},
													},
													Ctx: p9582,
													FreeVars: ast.Identifiers{
														"indexedPath",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(904),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(904),
															Column: int(61),
														},
													},
												},
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(905),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(905),
																		Column: int(20),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "isNumber",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(905),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(905),
																	Column: int(29),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "v",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9619,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(905),
																				Column: int(30),
																			},
																			End: ast.Location{
																				Line: int(905),
																				Column: int(31),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(905),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(905),
																Column: int(32),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												BranchTrue: &ast.Binary{
													Right: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(906),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(906),
																	Column: int(17),
																},
															},
														},
													},
													Left: &ast.LiteralString{
														Value: "",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																},
															},
															Ctx: p9582,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(906),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(906),
																	Column: int(13),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(906),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(906),
																Column: int(17),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												BranchFalse: &ast.Conditional{
													Cond: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(907),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(907),
																			Column: int(20),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "isString",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9582,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(907),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(907),
																		Column: int(29),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9637,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(907),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(907),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(907),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(907),
																	Column: int(32),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchTrue: &ast.Apply{
														Target: &ast.Var{
															Id: "escapeStringToml",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p9582,
																FreeVars: ast.Identifiers{
																	"escapeStringToml",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(908),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(908),
																		Column: int(27),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9646,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(908),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(908),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"escapeStringToml",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(908),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(908),
																	Column: int(30),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchFalse: &ast.Conditional{
														Cond: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(909),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(909),
																				Column: int(20),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "isFunction",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9582,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(909),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(909),
																			Column: int(31),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "v",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p9658,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(909),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(909),
																						Column: int(33),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9582,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(909),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(909),
																		Column: int(34),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														BranchTrue: &ast.Error{
															Expr: &ast.Binary{
																Right: &ast.Var{
																	Id: "indexedPath",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"indexedPath",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(910),
																				Column: int(52),
																			},
																			End: ast.Location{
																				Line: int(910),
																				Column: int(63),
																			},
																		},
																	},
																},
																Left: &ast.LiteralString{
																	Value: "Tried to manifest function at ",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(910),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(910),
																				Column: int(49),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9582,
																	FreeVars: ast.Identifiers{
																		"indexedPath",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(910),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(910),
																			Column: int(63),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p9582,
																FreeVars: ast.Identifiers{
																	"indexedPath",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(910),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(910),
																		Column: int(63),
																	},
																},
															},
														},
														BranchFalse: &ast.Conditional{
															Cond: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(911),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(911),
																					Column: int(20),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "isArray",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(911),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(911),
																				Column: int(28),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "v",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p9678,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(911),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(911),
																							Column: int(30),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9582,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(911),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(911),
																			Column: int(31),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															BranchTrue: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9582,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(912),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(912),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	Left: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(912),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(912),
																							Column: int(17),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "length",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p9582,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(912),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(912),
																						Column: int(24),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9692,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(912),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(912),
																									Column: int(26),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9582,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(912),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(912),
																					Column: int(27),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(912),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(912),
																				Column: int(32),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.LiteralString{
																	Value: "[]",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(913),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(913),
																				Column: int(17),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																BranchFalse: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(915),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(915),
																									Column: int(30),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "range",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9705,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(915),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(915),
																								Column: int(36),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralNumber{
																								OriginalString: "0",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9709,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(915),
																											Column: int(37),
																										},
																										End: ast.Location{
																											Line: int(915),
																											Column: int(38),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9709,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(915),
																												Column: int(56),
																											},
																											End: ast.Location{
																												Line: int(915),
																												Column: int(57),
																											},
																										},
																									},
																								},
																								Left: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(915),
																														Column: int(40),
																													},
																													End: ast.Location{
																														Line: int(915),
																														Column: int(43),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "length",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p9709,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(915),
																													Column: int(40),
																												},
																												End: ast.Location{
																													Line: int(915),
																													Column: int(50),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9720,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(915),
																																Column: int(51),
																															},
																															End: ast.Location{
																																Line: int(915),
																																Column: int(52),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9709,
																										FreeVars: ast.Identifiers{
																											"std",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(915),
																												Column: int(40),
																											},
																											End: ast.Location{
																												Line: int(915),
																												Column: int(53),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9709,
																									FreeVars: ast.Identifiers{
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(915),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(915),
																											Column: int(57),
																										},
																									},
																								},
																								Op: ast.BinaryOp(4),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p9705,
																					FreeVars: ast.Identifiers{
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(915),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(915),
																							Column: int(58),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "range",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(915),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(915),
																					Column: int(58),
																				},
																			},
																		},
																	},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Body: &ast.Conditional{
																					Cond: &ast.Var{
																						Id: "inline",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9729,
																							FreeVars: ast.Identifiers{
																								"inline",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(916),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(916),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					BranchTrue: &ast.LiteralString{
																						Value: "",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9729,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(916),
																									Column: int(47),
																								},
																								End: ast.Location{
																									Line: int(916),
																									Column: int(49),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					BranchFalse: &ast.Binary{
																						Right: &ast.Var{
																							Id: "indent",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9729,
																								FreeVars: ast.Identifiers{
																									"indent",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(916),
																										Column: int(65),
																									},
																									End: ast.Location{
																										Line: int(916),
																										Column: int(71),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "cindent",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9729,
																								FreeVars: ast.Identifiers{
																									"cindent",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(916),
																										Column: int(55),
																									},
																									End: ast.Location{
																										Line: int(916),
																										Column: int(62),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9729,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"indent",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(916),
																									Column: int(55),
																								},
																								End: ast.Location{
																									Line: int(916),
																									Column: int(71),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9729,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"indent",
																							"inline",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(916),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(916),
																								Column: int(71),
																							},
																						},
																					},
																				},
																				EqFodder: ast.Fodder{},
																				Variable: "new_indent",
																				CloseFodder: ast.Fodder{},
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(916),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(916),
																						Column: int(71),
																					},
																				},
																			},
																		},
																		Body: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Conditional{
																						Cond: &ast.Var{
																							Id: "inline",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9743,
																								FreeVars: ast.Identifiers{
																									"inline",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(917),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(917),
																										Column: int(40),
																									},
																								},
																							},
																						},
																						BranchTrue: &ast.LiteralString{
																							Value: " ",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9743,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(917),
																										Column: int(46),
																									},
																									End: ast.Location{
																										Line: int(917),
																										Column: int(49),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						BranchFalse: &ast.LiteralString{
																							Value: "\n",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9743,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(917),
																										Column: int(55),
																									},
																									End: ast.Location{
																										Line: int(917),
																										Column: int(59),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9743,
																							FreeVars: ast.Identifiers{
																								"inline",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(917),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(917),
																									Column: int(59),
																								},
																							},
																						},
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "separator",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(917),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(917),
																							Column: int(59),
																						},
																					},
																				},
																			},
																			Body: &ast.Local{
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: ast.Fodder{},
																						Body: &ast.Binary{
																							Right: &ast.Array{
																								Elements: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.LiteralString{
																												Value: "]",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p9755,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(924),
																															Column: int(77),
																														},
																														End: ast.Location{
																															Line: int(924),
																															Column: int(80),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Left: &ast.Binary{
																												Right: &ast.Conditional{
																													Cond: &ast.Var{
																														Id: "inline",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9755,
																															FreeVars: ast.Identifiers{
																																"inline",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(924),
																																	Column: int(46),
																																},
																																End: ast.Location{
																																	Line: int(924),
																																	Column: int(52),
																																},
																															},
																														},
																													},
																													BranchTrue: &ast.LiteralString{
																														Value: "",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9755,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(924),
																																	Column: int(58),
																																},
																																End: ast.Location{
																																	Line: int(924),
																																	Column: int(60),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													BranchFalse: &ast.Var{
																														Id: "cindent",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9755,
																															FreeVars: ast.Identifiers{
																																"cindent",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(924),
																																	Column: int(66),
																																},
																																End: ast.Location{
																																	Line: int(924),
																																	Column: int(73),
																																},
																															},
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													ElseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9755,
																														FreeVars: ast.Identifiers{
																															"cindent",
																															"inline",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(924),
																																Column: int(43),
																															},
																															End: ast.Location{
																																Line: int(924),
																																Column: int(73),
																															},
																														},
																													},
																												},
																												Left: &ast.Var{
																													Id: "separator",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9755,
																														FreeVars: ast.Identifiers{
																															"separator",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(924),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(924),
																																Column: int(39),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p9755,
																													FreeVars: ast.Identifiers{
																														"cindent",
																														"inline",
																														"separator",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(924),
																															Column: int(30),
																														},
																														End: ast.Location{
																															Line: int(924),
																															Column: int(74),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p9755,
																												FreeVars: ast.Identifiers{
																													"cindent",
																													"inline",
																													"separator",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(924),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(924),
																														Column: int(80),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: nil,
																									},
																								},
																								CloseFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9768,
																									FreeVars: ast.Identifiers{
																										"cindent",
																										"inline",
																										"separator",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(924),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(924),
																											Column: int(81),
																										},
																									},
																								},
																								TrailingComma: false,
																							},
																							Left: &ast.Binary{
																								Right: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(919),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(919),
																														Column: int(32),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "join",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p9768,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(919),
																													Column: int(29),
																												},
																												End: ast.Location{
																													Line: int(919),
																													Column: int(37),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Array{
																													Elements: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																Right: &ast.Var{
																																	Id: "separator",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p9782,
																																		FreeVars: ast.Identifiers{
																																			"separator",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(919),
																																				Column: int(45),
																																			},
																																			End: ast.Location{
																																				Line: int(919),
																																				Column: int(54),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.LiteralString{
																																	Value: ",",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p9782,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(919),
																																				Column: int(39),
																																			},
																																			End: ast.Location{
																																				Line: int(919),
																																				Column: int(42),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p9782,
																																	FreeVars: ast.Identifiers{
																																		"separator",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(919),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(919),
																																			Column: int(54),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															CommaFodder: nil,
																														},
																													},
																													CloseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9786,
																														FreeVars: ast.Identifiers{
																															"separator",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(919),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(919),
																																Column: int(55),
																															},
																														},
																													},
																													TrailingComma: false,
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "$std",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "flatMap",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: nil,
																														LeftBracketFodder: nil,
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"$std",
																															},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																													},
																													FodderLeft: nil,
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Function{
																																	ParenLeftFodder: nil,
																																	ParenRightFodder: nil,
																																	Body: &ast.Array{
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Array{
																																					Elements: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Binary{
																																								Right: &ast.Apply{
																																									Target: &ast.Var{
																																										Id: "renderValue",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p9803,
																																											FreeVars: ast.Identifiers{
																																												"renderValue",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(921),
																																													Column: int(54),
																																												},
																																												End: ast.Location{
																																													Line: int(921),
																																													Column: int(65),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "v",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p9808,
																																															FreeVars: ast.Identifiers{
																																																"v",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(921),
																																																	Column: int(66),
																																																},
																																																End: ast.Location{
																																																	Line: int(921),
																																																	Column: int(67),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.Var{
																																														Id: "i",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p9808,
																																															FreeVars: ast.Identifiers{
																																																"i",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(921),
																																																	Column: int(68),
																																																},
																																																End: ast.Location{
																																																	Line: int(921),
																																																	Column: int(69),
																																																},
																																															},
																																														},
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p9808,
																																														FreeVars: ast.Identifiers{
																																															"i",
																																															"v",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(921),
																																																Column: int(66),
																																															},
																																															End: ast.Location{
																																																Line: int(921),
																																																Column: int(70),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													Right: &ast.Array{
																																														Elements: []ast.CommaSeparatedExpr{
																																															ast.CommaSeparatedExpr{
																																																Expr: &ast.Var{
																																																	Id: "i",
																																																	NodeBase: ast.NodeBase{
																																																		Fodder: ast.Fodder{},
																																																		Ctx: p9817,
																																																		FreeVars: ast.Identifiers{
																																																			"i",
																																																		},
																																																		LocRange: ast.LocationRange{
																																																			File: p8,
																																																			FileName: "",
																																																			Begin: ast.Location{
																																																				Line: int(921),
																																																				Column: int(87),
																																																			},
																																																			End: ast.Location{
																																																				Line: int(921),
																																																				Column: int(88),
																																																			},
																																																		},
																																																	},
																																																},
																																																CommaFodder: nil,
																																															},
																																														},
																																														CloseFodder: ast.Fodder{},
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p9808,
																																															FreeVars: ast.Identifiers{
																																																"i",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(921),
																																																	Column: int(86),
																																																},
																																																End: ast.Location{
																																																	Line: int(921),
																																																	Column: int(89),
																																																},
																																															},
																																														},
																																														TrailingComma: false,
																																													},
																																													Left: &ast.Var{
																																														Id: "indexedPath",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p9808,
																																															FreeVars: ast.Identifiers{
																																																"indexedPath",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(921),
																																																	Column: int(72),
																																																},
																																																End: ast.Location{
																																																	Line: int(921),
																																																	Column: int(83),
																																																},
																																															},
																																														},
																																													},
																																													OpFodder: ast.Fodder{},
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p9808,
																																														FreeVars: ast.Identifiers{
																																															"i",
																																															"indexedPath",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(921),
																																																Column: int(72),
																																															},
																																															End: ast.Location{
																																																Line: int(921),
																																																Column: int(89),
																																															},
																																														},
																																													},
																																													Op: ast.BinaryOp(3),
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.LiteralBoolean{
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p9808,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(921),
																																																Column: int(91),
																																															},
																																															End: ast.Location{
																																																Line: int(921),
																																																Column: int(95),
																																															},
																																														},
																																													},
																																													Value: true,
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.LiteralString{
																																													Value: "",
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p9808,
																																														FreeVars: ast.Identifiers{},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(921),
																																																Column: int(97),
																																															},
																																															End: ast.Location{
																																																Line: int(921),
																																																Column: int(99),
																																															},
																																														},
																																													},
																																													Kind: ast.LiteralStringKind(1),
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p9803,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																											"indexedPath",
																																											"renderValue",
																																											"v",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(921),
																																												Column: int(54),
																																											},
																																											End: ast.Location{
																																												Line: int(921),
																																												Column: int(100),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								Left: &ast.Var{
																																									Id: "new_indent",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p9803,
																																										FreeVars: ast.Identifiers{
																																											"new_indent",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(921),
																																												Column: int(41),
																																											},
																																											End: ast.Location{
																																												Line: int(921),
																																												Column: int(51),
																																											},
																																										},
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p9803,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																										"indexedPath",
																																										"new_indent",
																																										"renderValue",
																																										"v",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(921),
																																											Column: int(41),
																																										},
																																										End: ast.Location{
																																											Line: int(921),
																																											Column: int(100),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					CloseFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Comment: []string{},
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(39),
																																							},
																																						},
																																						Ctx: p9830,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																							"indexedPath",
																																							"new_indent",
																																							"renderValue",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(921),
																																								Column: int(40),
																																							},
																																							End: ast.Location{
																																								Line: int(921),
																																								Column: int(101),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		CloseFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"indexedPath",
																																				"new_indent",
																																				"renderValue",
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																	},
																																	Parameters: []ast.Parameter{
																																		ast.Parameter{
																																			NameFodder: nil,
																																			Name: "i",
																																			CommaFodder: nil,
																																			EqFodder: nil,
																																			DefaultArg: nil,
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"indexedPath",
																																			"new_indent",
																																			"renderValue",
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																},
																																CommaFodder: nil,
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "range",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p9786,
																																		FreeVars: ast.Identifiers{
																																			"range",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(922),
																																				Column: int(49),
																																			},
																																			End: ast.Location{
																																				Line: int(922),
																																				Column: int(54),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: nil,
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"$std",
																															"indexedPath",
																															"new_indent",
																															"range",
																															"renderValue",
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(920),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(923),
																																Column: int(39),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9768,
																										FreeVars: ast.Identifiers{
																											"$std",
																											"indexedPath",
																											"new_indent",
																											"range",
																											"renderValue",
																											"separator",
																											"std",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(919),
																												Column: int(29),
																											},
																											End: ast.Location{
																												Line: int(923),
																												Column: int(40),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								Left: &ast.Array{
																									Elements: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Var{
																													Id: "separator",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9843,
																														FreeVars: ast.Identifiers{
																															"separator",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(918),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(918),
																																Column: int(43),
																															},
																														},
																													},
																												},
																												Left: &ast.LiteralString{
																													Value: "[",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9843,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(918),
																																Column: int(28),
																															},
																															End: ast.Location{
																																Line: int(918),
																																Column: int(31),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p9843,
																													FreeVars: ast.Identifiers{
																														"separator",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(918),
																															Column: int(28),
																														},
																														End: ast.Location{
																															Line: int(918),
																															Column: int(43),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: nil,
																										},
																									},
																									CloseFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9768,
																										FreeVars: ast.Identifiers{
																											"separator",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(918),
																												Column: int(27),
																											},
																											End: ast.Location{
																												Line: int(918),
																												Column: int(44),
																											},
																										},
																									},
																									TrailingComma: false,
																								},
																								OpFodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(26),
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9768,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"indexedPath",
																										"new_indent",
																										"range",
																										"renderValue",
																										"separator",
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(918),
																											Column: int(27),
																										},
																										End: ast.Location{
																											Line: int(923),
																											Column: int(40),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							OpFodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(26),
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9768,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"cindent",
																									"indexedPath",
																									"inline",
																									"new_indent",
																									"range",
																									"renderValue",
																									"separator",
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(918),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(924),
																										Column: int(81),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						EqFodder: ast.Fodder{},
																						Variable: "lines",
																						CloseFodder: ast.Fodder{},
																						Fun: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(918),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(924),
																								Column: int(81),
																							},
																						},
																					},
																				},
																				Body: &ast.Apply{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "std",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(12),
																									},
																								},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(925),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(925),
																										Column: int(16),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "join",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9582,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(925),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(925),
																									Column: int(21),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.LiteralString{
																									Value: "",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9861,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(925),
																												Column: int(22),
																											},
																											End: ast.Location{
																												Line: int(925),
																												Column: int(24),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "lines",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9861,
																										FreeVars: ast.Identifiers{
																											"lines",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(925),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(925),
																												Column: int(31),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9582,
																						FreeVars: ast.Identifiers{
																							"lines",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(925),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(925),
																								Column: int(32),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																						},
																					},
																					Ctx: p9582,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"cindent",
																						"indexedPath",
																						"inline",
																						"new_indent",
																						"range",
																						"renderValue",
																						"separator",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(918),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(925),
																							Column: int(32),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(12),
																					},
																				},
																				Ctx: p9582,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"cindent",
																					"indexedPath",
																					"inline",
																					"new_indent",
																					"range",
																					"renderValue",
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(917),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(925),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(12),
																				},
																			},
																			Ctx: p9582,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"cindent",
																				"indent",
																				"indexedPath",
																				"inline",
																				"range",
																				"renderValue",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(916),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(925),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"cindent",
																			"indent",
																			"indexedPath",
																			"inline",
																			"renderValue",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(915),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(925),
																				Column: int(32),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p9582,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"cindent",
																		"indent",
																		"indexedPath",
																		"inline",
																		"renderValue",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(912),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(925),
																			Column: int(32),
																		},
																	},
																},
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(926),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(926),
																						Column: int(20),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "isObject",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9582,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(926),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(926),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "v",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9885,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(926),
																								Column: int(30),
																							},
																							End: ast.Location{
																								Line: int(926),
																								Column: int(31),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(926),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(926),
																				Column: int(32),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																BranchTrue: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				Right: &ast.Array{
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralString{
																								Value: " }",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p9894,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(933),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(933),
																											Column: int(32),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							CommaFodder: nil,
																						},
																					},
																					CloseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9895,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(933),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(933),
																								Column: int(33),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				Left: &ast.Binary{
																					Right: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(928),
																											Column: int(27),
																										},
																										End: ast.Location{
																											Line: int(928),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "join",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p9895,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(928),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(928),
																										Column: int(35),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Array{
																										Elements: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralString{
																													Value: ", ",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p9907,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(928),
																																Column: int(37),
																															},
																															End: ast.Location{
																																Line: int(928),
																																Column: int(41),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												CommaFodder: nil,
																											},
																										},
																										CloseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p9908,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(928),
																													Column: int(36),
																												},
																												End: ast.Location{
																													Line: int(928),
																													Column: int(42),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "$std",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "flatMap",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: nil,
																											LeftBracketFodder: nil,
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																										},
																										FodderLeft: nil,
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Function{
																														ParenLeftFodder: nil,
																														ParenRightFodder: nil,
																														Body: &ast.Array{
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Array{
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					Right: &ast.Apply{
																																						Target: &ast.Var{
																																							Id: "renderValue",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p9924,
																																								FreeVars: ast.Identifiers{
																																									"renderValue",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(930),
																																										Column: int(66),
																																									},
																																									End: ast.Location{
																																										Line: int(930),
																																										Column: int(77),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "v",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p9929,
																																												FreeVars: ast.Identifiers{
																																													"v",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(930),
																																														Column: int(78),
																																													},
																																													End: ast.Location{
																																														Line: int(930),
																																														Column: int(79),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.Var{
																																											Id: "k",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p9929,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(930),
																																														Column: int(80),
																																													},
																																													End: ast.Location{
																																														Line: int(930),
																																														Column: int(81),
																																													},
																																												},
																																											},
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p9929,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																												"v",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(930),
																																													Column: int(78),
																																												},
																																												End: ast.Location{
																																													Line: int(930),
																																													Column: int(82),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Binary{
																																										Right: &ast.Array{
																																											Elements: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														Id: "k",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p9938,
																																															FreeVars: ast.Identifiers{
																																																"k",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(930),
																																																	Column: int(99),
																																																},
																																																End: ast.Location{
																																																	Line: int(930),
																																																	Column: int(100),
																																																},
																																															},
																																														},
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											CloseFodder: ast.Fodder{},
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p9929,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(930),
																																														Column: int(98),
																																													},
																																													End: ast.Location{
																																														Line: int(930),
																																														Column: int(101),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																										},
																																										Left: &ast.Var{
																																											Id: "indexedPath",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p9929,
																																												FreeVars: ast.Identifiers{
																																													"indexedPath",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(930),
																																														Column: int(84),
																																													},
																																													End: ast.Location{
																																														Line: int(930),
																																														Column: int(95),
																																													},
																																												},
																																											},
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p9929,
																																											FreeVars: ast.Identifiers{
																																												"indexedPath",
																																												"k",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(930),
																																													Column: int(84),
																																												},
																																												End: ast.Location{
																																													Line: int(930),
																																													Column: int(101),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralBoolean{
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p9929,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(930),
																																													Column: int(103),
																																												},
																																												End: ast.Location{
																																													Line: int(930),
																																													Column: int(107),
																																												},
																																											},
																																										},
																																										Value: true,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralString{
																																										Value: "",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p9929,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(930),
																																													Column: int(109),
																																												},
																																												End: ast.Location{
																																													Line: int(930),
																																													Column: int(111),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p9924,
																																							FreeVars: ast.Identifiers{
																																								"indexedPath",
																																								"k",
																																								"renderValue",
																																								"v",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(930),
																																									Column: int(66),
																																								},
																																								End: ast.Location{
																																									Line: int(930),
																																									Column: int(112),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					Left: &ast.Binary{
																																						Right: &ast.LiteralString{
																																							Value: " = ",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p9924,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(930),
																																										Column: int(58),
																																									},
																																									End: ast.Location{
																																										Line: int(930),
																																										Column: int(63),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						Left: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "escapeKeyToml",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p9924,
																																									FreeVars: ast.Identifiers{
																																										"escapeKeyToml",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(930),
																																											Column: int(39),
																																										},
																																										End: ast.Location{
																																											Line: int(930),
																																											Column: int(52),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "k",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p9954,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(930),
																																														Column: int(53),
																																													},
																																													End: ast.Location{
																																														Line: int(930),
																																														Column: int(54),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p9924,
																																								FreeVars: ast.Identifiers{
																																									"escapeKeyToml",
																																									"k",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(930),
																																										Column: int(39),
																																									},
																																									End: ast.Location{
																																										Line: int(930),
																																										Column: int(55),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p9924,
																																							FreeVars: ast.Identifiers{
																																								"escapeKeyToml",
																																								"k",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(930),
																																									Column: int(39),
																																								},
																																								End: ast.Location{
																																									Line: int(930),
																																									Column: int(63),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					OpFodder: ast.Fodder{},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p9924,
																																						FreeVars: ast.Identifiers{
																																							"escapeKeyToml",
																																							"indexedPath",
																																							"k",
																																							"renderValue",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(930),
																																								Column: int(39),
																																							},
																																							End: ast.Location{
																																								Line: int(930),
																																								Column: int(112),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		CloseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(37),
																																				},
																																			},
																																			Ctx: p9960,
																																			FreeVars: ast.Identifiers{
																																				"escapeKeyToml",
																																				"indexedPath",
																																				"k",
																																				"renderValue",
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(930),
																																					Column: int(38),
																																				},
																																				End: ast.Location{
																																					Line: int(930),
																																					Column: int(113),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															CloseFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"escapeKeyToml",
																																	"indexedPath",
																																	"k",
																																	"renderValue",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														Parameters: []ast.Parameter{
																															ast.Parameter{
																																NameFodder: nil,
																																Name: "k",
																																CommaFodder: nil,
																																EqFodder: nil,
																																DefaultArg: nil,
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"escapeKeyToml",
																																"indexedPath",
																																"renderValue",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														TrailingComma: false,
																													},
																													CommaFodder: nil,
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(931),
																																			Column: int(47),
																																		},
																																		End: ast.Location{
																																			Line: int(931),
																																			Column: int(50),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "objectFields",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p9908,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(931),
																																		Column: int(47),
																																	},
																																	End: ast.Location{
																																		Line: int(931),
																																		Column: int(63),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "v",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p9973,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(931),
																																					Column: int(64),
																																				},
																																				End: ast.Location{
																																					Line: int(931),
																																					Column: int(65),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p9908,
																															FreeVars: ast.Identifiers{
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(931),
																																	Column: int(47),
																																},
																																End: ast.Location{
																																	Line: int(931),
																																	Column: int(66),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: nil,
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$std",
																												"escapeKeyToml",
																												"indexedPath",
																												"renderValue",
																												"std",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(929),
																													Column: int(36),
																												},
																												End: ast.Location{
																													Line: int(932),
																													Column: int(37),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9895,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"escapeKeyToml",
																								"indexedPath",
																								"renderValue",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(928),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(932),
																									Column: int(38),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					Left: &ast.Array{
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.LiteralString{
																									Value: "{ ",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p9981,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(927),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(927),
																												Column: int(30),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								CommaFodder: nil,
																							},
																						},
																						CloseFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9895,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(927),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(927),
																									Column: int(31),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(24),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p9895,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"escapeKeyToml",
																							"indexedPath",
																							"renderValue",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(927),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(932),
																								Column: int(38),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				OpFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(24),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p9895,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"escapeKeyToml",
																						"indexedPath",
																						"renderValue",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(927),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(933),
																							Column: int(33),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "lines",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(927),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(933),
																					Column: int(33),
																				},
																			},
																		},
																	},
																	Body: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(934),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(934),
																							Column: int(14),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "join",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p9582,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(934),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(934),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralString{
																						Value: "",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9995,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(934),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(934),
																									Column: int(22),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "lines",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p9995,
																							FreeVars: ast.Identifiers{
																								"lines",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(934),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(934),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p9582,
																			FreeVars: ast.Identifiers{
																				"lines",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(934),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(934),
																					Column: int(30),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		Ctx: p9582,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"escapeKeyToml",
																			"indexedPath",
																			"renderValue",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(927),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(934),
																				Column: int(30),
																			},
																		},
																	},
																},
																BranchFalse: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p9582,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"escapeKeyToml",
																		"indexedPath",
																		"renderValue",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(926),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(934),
																			Column: int(30),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p9582,
																FreeVars: ast.Identifiers{
																	"$std",
																	"cindent",
																	"escapeKeyToml",
																	"indent",
																	"indexedPath",
																	"inline",
																	"renderValue",
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(911),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(934),
																		Column: int(30),
																	},
																},
															},
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p9582,
															FreeVars: ast.Identifiers{
																"$std",
																"cindent",
																"escapeKeyToml",
																"indent",
																"indexedPath",
																"inline",
																"renderValue",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(909),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(934),
																	Column: int(30),
																},
															},
														},
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p9582,
														FreeVars: ast.Identifiers{
															"$std",
															"cindent",
															"escapeKeyToml",
															"escapeStringToml",
															"indent",
															"indexedPath",
															"inline",
															"renderValue",
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(907),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(934),
																Column: int(30),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p9582,
													FreeVars: ast.Identifiers{
														"$std",
														"cindent",
														"escapeKeyToml",
														"escapeStringToml",
														"indent",
														"indexedPath",
														"inline",
														"renderValue",
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(905),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(934),
															Column: int(30),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p9582,
												FreeVars: ast.Identifiers{
													"$std",
													"cindent",
													"escapeKeyToml",
													"escapeStringToml",
													"indent",
													"indexedPath",
													"inline",
													"renderValue",
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(903),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(934),
														Column: int(30),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p9582,
											FreeVars: ast.Identifiers{
												"$std",
												"cindent",
												"escapeKeyToml",
												"escapeStringToml",
												"indent",
												"indexedPath",
												"inline",
												"renderValue",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(901),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(934),
													Column: int(30),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(8),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p9582,
										FreeVars: ast.Identifiers{
											"$std",
											"cindent",
											"escapeKeyToml",
											"escapeStringToml",
											"indent",
											"indexedPath",
											"inline",
											"renderValue",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(899),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(934),
												Column: int(30),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(898),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(898),
												Column: int(20),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "indexedPath",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(898),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(898),
												Column: int(33),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "inline",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(898),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(898),
												Column: int(41),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(898),
												Column: int(43),
											},
											End: ast.Location{
												Line: int(898),
												Column: int(50),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p10019,
									FreeVars: ast.Identifiers{
										"$std",
										"escapeKeyToml",
										"escapeStringToml",
										"indent",
										"renderValue",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(898),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(934),
											Column: int(30),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "renderValue",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(936),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(936),
																	Column: int(24),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "flattenArrays",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10029,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(936),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(936),
																Column: int(38),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "$std",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "flatMap",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: nil,
																	LeftBracketFodder: nil,
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Function{
																				ParenLeftFodder: nil,
																				ParenRightFodder: nil,
																				Body: &ast.Conditional{
																					Cond: &ast.Unary{
																						Expr: &ast.Apply{
																							Target: &ast.Var{
																								Id: "isSection",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10044,
																									FreeVars: ast.Identifiers{
																										"isSection",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(939),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(939),
																											Column: int(24),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Index{
																											Target: &ast.Var{
																												Id: "v",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10049,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(939),
																															Column: int(25),
																														},
																														End: ast.Location{
																															Line: int(939),
																															Column: int(26),
																														},
																													},
																												},
																											},
																											Index: &ast.Var{
																												Id: "k",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10049,
																													FreeVars: ast.Identifiers{
																														"k",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(939),
																															Column: int(27),
																														},
																														End: ast.Location{
																															Line: int(939),
																															Column: int(28),
																														},
																													},
																												},
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10049,
																												FreeVars: ast.Identifiers{
																													"k",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(939),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(939),
																														Column: int(29),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10044,
																								FreeVars: ast.Identifiers{
																									"isSection",
																									"k",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(939),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(939),
																										Column: int(30),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10044,
																							FreeVars: ast.Identifiers{
																								"isSection",
																								"k",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(939),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(939),
																									Column: int(30),
																								},
																							},
																						},
																						Op: ast.UnaryOp(0),
																					},
																					BranchTrue: &ast.Array{
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Array{
																									Elements: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Apply{
																													Target: &ast.Var{
																														Id: "renderValue",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p10063,
																															FreeVars: ast.Identifiers{
																																"renderValue",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(937),
																																	Column: int(49),
																																},
																																End: ast.Location{
																																	Line: int(937),
																																	Column: int(60),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "v",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p10068,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(937),
																																					Column: int(61),
																																				},
																																				End: ast.Location{
																																					Line: int(937),
																																					Column: int(62),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.Var{
																																		Id: "k",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p10068,
																																			FreeVars: ast.Identifiers{
																																				"k",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(937),
																																					Column: int(63),
																																				},
																																				End: ast.Location{
																																					Line: int(937),
																																					Column: int(64),
																																				},
																																			},
																																		},
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10068,
																																		FreeVars: ast.Identifiers{
																																			"k",
																																			"v",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(937),
																																				Column: int(61),
																																			},
																																			End: ast.Location{
																																				Line: int(937),
																																				Column: int(65),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	Right: &ast.Array{
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "k",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p10077,
																																						FreeVars: ast.Identifiers{
																																							"k",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(937),
																																								Column: int(82),
																																							},
																																							End: ast.Location{
																																								Line: int(937),
																																								Column: int(83),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		CloseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p10068,
																																			FreeVars: ast.Identifiers{
																																				"k",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(937),
																																					Column: int(81),
																																				},
																																				End: ast.Location{
																																					Line: int(937),
																																					Column: int(84),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																	},
																																	Left: &ast.Var{
																																		Id: "indexedPath",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p10068,
																																			FreeVars: ast.Identifiers{
																																				"indexedPath",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(937),
																																					Column: int(67),
																																				},
																																				End: ast.Location{
																																					Line: int(937),
																																					Column: int(78),
																																				},
																																			},
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10068,
																																		FreeVars: ast.Identifiers{
																																			"indexedPath",
																																			"k",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(937),
																																				Column: int(67),
																																			},
																																			End: ast.Location{
																																				Line: int(937),
																																				Column: int(84),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(3),
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.LiteralBoolean{
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10068,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(937),
																																				Column: int(86),
																																			},
																																			End: ast.Location{
																																				Line: int(937),
																																				Column: int(91),
																																			},
																																		},
																																	},
																																	Value: false,
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "cindent",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10068,
																																		FreeVars: ast.Identifiers{
																																			"cindent",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(937),
																																				Column: int(93),
																																			},
																																			End: ast.Location{
																																				Line: int(937),
																																				Column: int(100),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10063,
																														FreeVars: ast.Identifiers{
																															"cindent",
																															"indexedPath",
																															"k",
																															"renderValue",
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(937),
																																Column: int(49),
																															},
																															End: ast.Location{
																																Line: int(937),
																																Column: int(101),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												Left: &ast.Binary{
																													Right: &ast.LiteralString{
																														Value: " = ",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p10063,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(937),
																																	Column: int(41),
																																},
																																End: ast.Location{
																																	Line: int(937),
																																	Column: int(46),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Left: &ast.Binary{
																														Right: &ast.Apply{
																															Target: &ast.Var{
																																Id: "escapeKeyToml",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p10063,
																																	FreeVars: ast.Identifiers{
																																		"escapeKeyToml",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(937),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(937),
																																			Column: int(35),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "k",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p10095,
																																				FreeVars: ast.Identifiers{
																																					"k",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(937),
																																						Column: int(36),
																																					},
																																					End: ast.Location{
																																						Line: int(937),
																																						Column: int(37),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p10063,
																																FreeVars: ast.Identifiers{
																																	"escapeKeyToml",
																																	"k",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(937),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(937),
																																		Column: int(38),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														Left: &ast.Var{
																															Id: "cindent",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p10063,
																																FreeVars: ast.Identifiers{
																																	"cindent",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(937),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(937),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p10063,
																															FreeVars: ast.Identifiers{
																																"cindent",
																																"escapeKeyToml",
																																"k",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(937),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(937),
																																	Column: int(38),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10063,
																														FreeVars: ast.Identifiers{
																															"cindent",
																															"escapeKeyToml",
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(937),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(937),
																																Column: int(46),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10063,
																													FreeVars: ast.Identifiers{
																														"cindent",
																														"escapeKeyToml",
																														"indexedPath",
																														"k",
																														"renderValue",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(937),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(937),
																															Column: int(101),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: nil,
																										},
																									},
																									CloseFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										Ctx: p10104,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"escapeKeyToml",
																											"indexedPath",
																											"k",
																											"renderValue",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(937),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(937),
																												Column: int(102),
																											},
																										},
																									},
																									TrailingComma: false,
																								},
																								CommaFodder: nil,
																							},
																						},
																						CloseFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"escapeKeyToml",
																								"indexedPath",
																								"k",
																								"renderValue",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					BranchFalse: &ast.Array{
																						Elements: nil,
																						CloseFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					ThenFodder: nil,
																					ElseFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"escapeKeyToml",
																							"indexedPath",
																							"isSection",
																							"k",
																							"renderValue",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: nil,
																						Name: "k",
																						CommaFodder: nil,
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"escapeKeyToml",
																						"indexedPath",
																						"isSection",
																						"renderValue",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(938),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(938),
																									Column: int(23),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "objectFields",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10044,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(938),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(938),
																								Column: int(36),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10119,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(938),
																											Column: int(37),
																										},
																										End: ast.Location{
																											Line: int(938),
																											Column: int(38),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10044,
																					FreeVars: ast.Identifiers{
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(938),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(938),
																							Column: int(39),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: nil,
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"cindent",
																		"escapeKeyToml",
																		"indexedPath",
																		"isSection",
																		"renderValue",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(936),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(940),
																			Column: int(10),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10029,
													FreeVars: ast.Identifiers{
														"$std",
														"cindent",
														"escapeKeyToml",
														"indexedPath",
														"isSection",
														"renderValue",
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(936),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(940),
															Column: int(11),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "kvp",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(936),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(940),
													Column: int(11),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "$std",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "flatMap",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: nil,
															LeftBracketFodder: nil,
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Function{
																		ParenLeftFodder: nil,
																		ParenRightFodder: nil,
																		Body: &ast.Conditional{
																			Cond: &ast.Apply{
																				Target: &ast.Var{
																					Id: "isSection",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10138,
																						FreeVars: ast.Identifiers{
																							"isSection",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(949),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(949),
																								Column: int(23),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								Target: &ast.Var{
																									Id: "v",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10143,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(949),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(949),
																												Column: int(25),
																											},
																										},
																									},
																								},
																								Index: &ast.Var{
																									Id: "k",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10143,
																										FreeVars: ast.Identifiers{
																											"k",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(949),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(949),
																												Column: int(27),
																											},
																										},
																									},
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10143,
																									FreeVars: ast.Identifiers{
																										"k",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(949),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(949),
																											Column: int(28),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10138,
																					FreeVars: ast.Identifiers{
																						"isSection",
																						"k",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(949),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(949),
																							Column: int(29),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			BranchTrue: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Conditional{
																							Cond: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(943),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(943),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "isObject",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10157,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(943),
																												Column: int(16),
																											},
																											End: ast.Location{
																												Line: int(943),
																												Column: int(28),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10162,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(943),
																																Column: int(29),
																															},
																															End: ast.Location{
																																Line: int(943),
																																Column: int(30),
																															},
																														},
																													},
																												},
																												Index: &ast.Var{
																													Id: "k",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10162,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(943),
																																Column: int(31),
																															},
																															End: ast.Location{
																																Line: int(943),
																																Column: int(32),
																															},
																														},
																													},
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10162,
																													FreeVars: ast.Identifiers{
																														"k",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(943),
																															Column: int(29),
																														},
																														End: ast.Location{
																															Line: int(943),
																															Column: int(33),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10157,
																									FreeVars: ast.Identifiers{
																										"k",
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(943),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(943),
																											Column: int(34),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							BranchTrue: &ast.Apply{
																								Target: &ast.Var{
																									Id: "renderTable",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(14),
																											},
																										},
																										Ctx: p10157,
																										FreeVars: ast.Identifiers{
																											"renderTable",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(944),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(944),
																												Column: int(26),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(27),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(28),
																															},
																														},
																													},
																												},
																												Index: &ast.Var{
																													Id: "k",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(29),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(30),
																															},
																														},
																													},
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10175,
																													FreeVars: ast.Identifiers{
																														"k",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(944),
																															Column: int(27),
																														},
																														End: ast.Location{
																															Line: int(944),
																															Column: int(31),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Array{
																													Elements: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p10184,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(944),
																																			Column: int(41),
																																		},
																																		End: ast.Location{
																																			Line: int(944),
																																			Column: int(42),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													CloseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(40),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(43),
																															},
																														},
																													},
																													TrailingComma: false,
																												},
																												Left: &ast.Var{
																													Id: "path",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"path",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(37),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10175,
																													FreeVars: ast.Identifiers{
																														"k",
																														"path",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(944),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(944),
																															Column: int(43),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Array{
																													Elements: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p10194,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(944),
																																			Column: int(60),
																																		},
																																		End: ast.Location{
																																			Line: int(944),
																																			Column: int(61),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													CloseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(59),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(62),
																															},
																														},
																													},
																													TrailingComma: false,
																												},
																												Left: &ast.Var{
																													Id: "indexedPath",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10175,
																														FreeVars: ast.Identifiers{
																															"indexedPath",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(45),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(56),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10175,
																													FreeVars: ast.Identifiers{
																														"indexedPath",
																														"k",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(944),
																															Column: int(45),
																														},
																														End: ast.Location{
																															Line: int(944),
																															Column: int(62),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "cindent",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10175,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(944),
																															Column: int(64),
																														},
																														End: ast.Location{
																															Line: int(944),
																															Column: int(71),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10157,
																									FreeVars: ast.Identifiers{
																										"cindent",
																										"indexedPath",
																										"k",
																										"path",
																										"renderTable",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(944),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(944),
																											Column: int(72),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							BranchFalse: &ast.Apply{
																								Target: &ast.Var{
																									Id: "renderTableArray",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(14),
																											},
																										},
																										Ctx: p10157,
																										FreeVars: ast.Identifiers{
																											"renderTableArray",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(946),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(946),
																												Column: int(31),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(32),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(33),
																															},
																														},
																													},
																												},
																												Index: &ast.Var{
																													Id: "k",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(35),
																															},
																														},
																													},
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10210,
																													FreeVars: ast.Identifiers{
																														"k",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(946),
																															Column: int(32),
																														},
																														End: ast.Location{
																															Line: int(946),
																															Column: int(36),
																														},
																													},
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Array{
																													Elements: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p10219,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(946),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(946),
																																			Column: int(47),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													CloseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(45),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(48),
																															},
																														},
																													},
																													TrailingComma: false,
																												},
																												Left: &ast.Var{
																													Id: "path",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"path",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(42),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10210,
																													FreeVars: ast.Identifiers{
																														"k",
																														"path",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(946),
																															Column: int(38),
																														},
																														End: ast.Location{
																															Line: int(946),
																															Column: int(48),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												Right: &ast.Array{
																													Elements: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p10229,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(946),
																																			Column: int(65),
																																		},
																																		End: ast.Location{
																																			Line: int(946),
																																			Column: int(66),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													CloseFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(64),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(67),
																															},
																														},
																													},
																													TrailingComma: false,
																												},
																												Left: &ast.Var{
																													Id: "indexedPath",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10210,
																														FreeVars: ast.Identifiers{
																															"indexedPath",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(946),
																																Column: int(50),
																															},
																															End: ast.Location{
																																Line: int(946),
																																Column: int(61),
																															},
																														},
																													},
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10210,
																													FreeVars: ast.Identifiers{
																														"indexedPath",
																														"k",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(946),
																															Column: int(50),
																														},
																														End: ast.Location{
																															Line: int(946),
																															Column: int(67),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "cindent",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p10210,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(946),
																															Column: int(69),
																														},
																														End: ast.Location{
																															Line: int(946),
																															Column: int(76),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10157,
																									FreeVars: ast.Identifiers{
																										"cindent",
																										"indexedPath",
																										"k",
																										"path",
																										"renderTableArray",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(946),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(946),
																											Column: int(77),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							ThenFodder: ast.Fodder{},
																							ElseFodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(12),
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(12),
																									},
																								},
																								Ctx: p10157,
																								FreeVars: ast.Identifiers{
																									"cindent",
																									"indexedPath",
																									"k",
																									"path",
																									"renderTable",
																									"renderTableArray",
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(943),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(946),
																										Column: int(77),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"indexedPath",
																						"k",
																						"path",
																						"renderTable",
																						"renderTableArray",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			BranchFalse: &ast.Array{
																				Elements: nil,
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			ThenFodder: nil,
																			ElseFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"cindent",
																					"indexedPath",
																					"isSection",
																					"k",
																					"path",
																					"renderTable",
																					"renderTableArray",
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: nil,
																				Name: "k",
																				CommaFodder: nil,
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"cindent",
																				"indexedPath",
																				"isSection",
																				"path",
																				"renderTable",
																				"renderTableArray",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(948),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(948),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "objectFields",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p10138,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(948),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(948),
																						Column: int(36),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10254,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(948),
																									Column: int(37),
																								},
																								End: ast.Location{
																									Line: int(948),
																									Column: int(38),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10138,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(948),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(948),
																					Column: int(39),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: nil,
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"cindent",
																"indexedPath",
																"isSection",
																"path",
																"renderTable",
																"renderTableArray",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(941),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(950),
																	Column: int(10),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Array{
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(941),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(941),
																						Column: int(30),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "join",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10265,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(941),
																					Column: int(27),
																				},
																				End: ast.Location{
																					Line: int(941),
																					Column: int(35),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					Value: "\n",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10269,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(941),
																								Column: int(36),
																							},
																							End: ast.Location{
																								Line: int(941),
																								Column: int(40),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "kvp",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10269,
																						FreeVars: ast.Identifiers{
																							"kvp",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(941),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(941),
																								Column: int(45),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10265,
																		FreeVars: ast.Identifiers{
																			"kvp",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(941),
																				Column: int(27),
																			},
																			End: ast.Location{
																				Line: int(941),
																				Column: int(46),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														CloseFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10138,
															FreeVars: ast.Identifiers{
																"kvp",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(941),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(941),
																	Column: int(47),
																},
															},
														},
														TrailingComma: false,
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10138,
														FreeVars: ast.Identifiers{
															"$std",
															"cindent",
															"indexedPath",
															"isSection",
															"kvp",
															"path",
															"renderTable",
															"renderTableArray",
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(941),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(950),
																Column: int(10),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												EqFodder: ast.Fodder{},
												Variable: "sections",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(941),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(950),
														Column: int(10),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(951),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(951),
																Column: int(12),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "join",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10281,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(951),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(951),
															Column: int(17),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "\n\n",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10285,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(951),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(951),
																		Column: int(24),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "sections",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10285,
																FreeVars: ast.Identifiers{
																	"sections",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(951),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(951),
																		Column: int(34),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10281,
												FreeVars: ast.Identifiers{
													"sections",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(951),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(951),
														Column: int(35),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p10281,
											FreeVars: ast.Identifiers{
												"$std",
												"cindent",
												"indexedPath",
												"isSection",
												"kvp",
												"path",
												"renderTable",
												"renderTableArray",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(941),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(951),
													Column: int(35),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p10281,
										FreeVars: ast.Identifiers{
											"$std",
											"cindent",
											"escapeKeyToml",
											"indexedPath",
											"isSection",
											"path",
											"renderTable",
											"renderTableArray",
											"renderValue",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(936),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(951),
												Column: int(35),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(935),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(935),
												Column: int(28),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(935),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(935),
												Column: int(34),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "indexedPath",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(935),
												Column: int(36),
											},
											End: ast.Location{
												Line: int(935),
												Column: int(47),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(935),
												Column: int(49),
											},
											End: ast.Location{
												Line: int(935),
												Column: int(56),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p10294,
									FreeVars: ast.Identifiers{
										"$std",
										"escapeKeyToml",
										"isSection",
										"renderTable",
										"renderTableArray",
										"renderValue",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(935),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(951),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "renderTableInternal",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Binary{
									Right: &ast.Apply{
										Target: &ast.Var{
											Id: "renderTableInternal",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10300,
												FreeVars: ast.Identifiers{
													"renderTableInternal",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(955),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(955),
														Column: int(30),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10304,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(955),
																	Column: int(31),
																},
																End: ast.Location{
																	Line: int(955),
																	Column: int(32),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "path",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10304,
															FreeVars: ast.Identifiers{
																"path",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(955),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(955),
																	Column: int(38),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "indexedPath",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10304,
															FreeVars: ast.Identifiers{
																"indexedPath",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(955),
																	Column: int(40),
																},
																End: ast.Location{
																	Line: int(955),
																	Column: int(51),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														Right: &ast.Var{
															Id: "indent",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10304,
																FreeVars: ast.Identifiers{
																	"indent",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(955),
																		Column: int(63),
																	},
																	End: ast.Location{
																		Line: int(955),
																		Column: int(69),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "cindent",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10304,
																FreeVars: ast.Identifiers{
																	"cindent",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(955),
																		Column: int(53),
																	},
																	End: ast.Location{
																		Line: int(955),
																		Column: int(60),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10304,
															FreeVars: ast.Identifiers{
																"cindent",
																"indent",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(955),
																	Column: int(53),
																},
																End: ast.Location{
																	Line: int(955),
																	Column: int(69),
																},
															},
														},
														Op: ast.BinaryOp(3),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p10300,
											FreeVars: ast.Identifiers{
												"cindent",
												"indent",
												"indexedPath",
												"path",
												"renderTableInternal",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(955),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(955),
													Column: int(70),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									Left: &ast.Binary{
										Right: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.DesugaredObject{
													Asserts: ast.Nodes{},
													Fields: ast.DesugaredObjectFields{},
													Locals: ast.LocalBinds{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10300,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(954),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(954),
																Column: int(22),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "v",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10300,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(954),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(954),
																Column: int(16),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10300,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(954),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(954),
															Column: int(22),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralString{
												Value: "",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10300,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(954),
															Column: int(28),
														},
														End: ast.Location{
															Line: int(954),
															Column: int(30),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											BranchFalse: &ast.LiteralString{
												Value: "\n",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10300,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(954),
															Column: int(36),
														},
														End: ast.Location{
															Line: int(954),
															Column: int(40),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10300,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(954),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(954),
														Column: int(40),
													},
												},
											},
										},
										Left: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "]",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10300,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(953),
															Column: int(71),
														},
														End: ast.Location{
															Line: int(953),
															Column: int(74),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(953),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(953),
																		Column: int(28),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "join",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10300,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(953),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(953),
																	Column: int(33),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralString{
																	Value: ".",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10338,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(953),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(953),
																				Column: int(37),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(953),
																						Column: int(39),
																					},
																					End: ast.Location{
																						Line: int(953),
																						Column: int(42),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "map",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10338,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(953),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(953),
																					Column: int(46),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "escapeKeyToml",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10347,
																						FreeVars: ast.Identifiers{
																							"escapeKeyToml",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(953),
																								Column: int(47),
																							},
																							End: ast.Location{
																								Line: int(953),
																								Column: int(60),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "path",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10347,
																						FreeVars: ast.Identifiers{
																							"path",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(953),
																								Column: int(62),
																							},
																							End: ast.Location{
																								Line: int(953),
																								Column: int(66),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10338,
																		FreeVars: ast.Identifiers{
																			"escapeKeyToml",
																			"path",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(953),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(953),
																				Column: int(67),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10300,
														FreeVars: ast.Identifiers{
															"escapeKeyToml",
															"path",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(953),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(953),
																Column: int(68),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Binary{
													Right: &ast.LiteralString{
														Value: "[",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10300,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(953),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(953),
																	Column: int(22),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													Left: &ast.Var{
														Id: "cindent",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p10300,
															FreeVars: ast.Identifiers{
																"cindent",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(953),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(953),
																	Column: int(16),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10300,
														FreeVars: ast.Identifiers{
															"cindent",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(953),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(953),
																Column: int(22),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10300,
													FreeVars: ast.Identifiers{
														"cindent",
														"escapeKeyToml",
														"path",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(953),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(953),
															Column: int(68),
														},
													},
												},
												Op: ast.BinaryOp(3),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10300,
												FreeVars: ast.Identifiers{
													"cindent",
													"escapeKeyToml",
													"path",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(953),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(953),
														Column: int(74),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										OpFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p10300,
											FreeVars: ast.Identifiers{
												"cindent",
												"escapeKeyToml",
												"path",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(953),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(954),
													Column: int(41),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									OpFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(8),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10300,
										FreeVars: ast.Identifiers{
											"cindent",
											"escapeKeyToml",
											"indent",
											"indexedPath",
											"path",
											"renderTableInternal",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(953),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(955),
												Column: int(70),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(952),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(952),
												Column: int(20),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(952),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(952),
												Column: int(26),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "indexedPath",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(952),
												Column: int(28),
											},
											End: ast.Location{
												Line: int(952),
												Column: int(39),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(952),
												Column: int(41),
											},
											End: ast.Location{
												Line: int(952),
												Column: int(48),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p10366,
									FreeVars: ast.Identifiers{
										"escapeKeyToml",
										"indent",
										"renderTableInternal",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(952),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(955),
											Column: int(70),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "renderTable",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(957),
																	Column: int(23),
																},
																End: ast.Location{
																	Line: int(957),
																	Column: int(26),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "range",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10376,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(957),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(957),
																Column: int(32),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10380,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(957),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(957),
																			Column: int(34),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10380,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(957),
																				Column: int(52),
																			},
																			End: ast.Location{
																				Line: int(957),
																				Column: int(53),
																			},
																		},
																	},
																},
																Left: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(957),
																						Column: int(36),
																					},
																					End: ast.Location{
																						Line: int(957),
																						Column: int(39),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "length",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10380,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(957),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(957),
																					Column: int(46),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "v",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10391,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(957),
																								Column: int(47),
																							},
																							End: ast.Location{
																								Line: int(957),
																								Column: int(48),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10380,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(957),
																				Column: int(36),
																			},
																			End: ast.Location{
																				Line: int(957),
																				Column: int(49),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10380,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(957),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(957),
																			Column: int(53),
																		},
																	},
																},
																Op: ast.BinaryOp(4),
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10376,
													FreeVars: ast.Identifiers{
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(957),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(957),
															Column: int(54),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "range",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(957),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(957),
													Column: int(54),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "flatMap",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Function{
																	ParenLeftFodder: nil,
																	ParenRightFodder: nil,
																	Body: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.Apply{
																						Target: &ast.Var{
																							Id: "renderTableInternal",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10411,
																								FreeVars: ast.Identifiers{
																									"renderTableInternal",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(961),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(961),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "v",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(35),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "i",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(36),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(37),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10416,
																											FreeVars: ast.Identifiers{
																												"i",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(961),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(961),
																													Column: int(38),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "path",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10416,
																											FreeVars: ast.Identifiers{
																												"path",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(961),
																													Column: int(40),
																												},
																												End: ast.Location{
																													Line: int(961),
																													Column: int(44),
																												},
																											},
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.Array{
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "i",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p10427,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(961),
																																	Column: int(61),
																																},
																																End: ast.Location{
																																	Line: int(961),
																																	Column: int(62),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											CloseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(60),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(63),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										Left: &ast.Var{
																											Id: "indexedPath",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"indexedPath",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(46),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(57),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10416,
																											FreeVars: ast.Identifiers{
																												"i",
																												"indexedPath",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(961),
																													Column: int(46),
																												},
																												End: ast.Location{
																													Line: int(961),
																													Column: int(63),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.Var{
																											Id: "indent",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"indent",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(75),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(81),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "cindent",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10416,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(961),
																														Column: int(65),
																													},
																													End: ast.Location{
																														Line: int(961),
																														Column: int(72),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10416,
																											FreeVars: ast.Identifiers{
																												"cindent",
																												"indent",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(961),
																													Column: int(65),
																												},
																												End: ast.Location{
																													Line: int(961),
																													Column: int(81),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10411,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"i",
																								"indent",
																								"indexedPath",
																								"path",
																								"renderTableInternal",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(961),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(961),
																									Column: int(82),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					Left: &ast.Binary{
																						Right: &ast.Conditional{
																							Cond: &ast.Binary{
																								Right: &ast.DesugaredObject{
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{},
																									Locals: ast.LocalBinds{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10411,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(960),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(960),
																												Column: int(28),
																											},
																										},
																									},
																								},
																								Left: &ast.Index{
																									Target: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10411,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(960),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(960),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									Index: &ast.Var{
																										Id: "i",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10411,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(960),
																													Column: int(20),
																												},
																												End: ast.Location{
																													Line: int(960),
																													Column: int(21),
																												},
																											},
																										},
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10411,
																										FreeVars: ast.Identifiers{
																											"i",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(960),
																												Column: int(18),
																											},
																											End: ast.Location{
																												Line: int(960),
																												Column: int(22),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10411,
																									FreeVars: ast.Identifiers{
																										"i",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(960),
																											Column: int(18),
																										},
																										End: ast.Location{
																											Line: int(960),
																											Column: int(28),
																										},
																									},
																								},
																								Op: ast.BinaryOp(12),
																							},
																							BranchTrue: &ast.LiteralString{
																								Value: "",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10411,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(960),
																											Column: int(34),
																										},
																										End: ast.Location{
																											Line: int(960),
																											Column: int(36),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							BranchFalse: &ast.LiteralString{
																								Value: "\n",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10411,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(960),
																											Column: int(42),
																										},
																										End: ast.Location{
																											Line: int(960),
																											Column: int(46),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							ThenFodder: ast.Fodder{},
																							ElseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10411,
																								FreeVars: ast.Identifiers{
																									"i",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(960),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(960),
																										Column: int(46),
																									},
																								},
																							},
																						},
																						Left: &ast.Binary{
																							Right: &ast.LiteralString{
																								Value: "]]",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10411,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(959),
																											Column: int(75),
																										},
																										End: ast.Location{
																											Line: int(959),
																											Column: int(79),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							Left: &ast.Binary{
																								Right: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(959),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(959),
																														Column: int(32),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "join",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10411,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(959),
																													Column: int(29),
																												},
																												End: ast.Location{
																													Line: int(959),
																													Column: int(37),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralString{
																													Value: ".",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10465,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(959),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(959),
																																Column: int(41),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(959),
																																		Column: int(43),
																																	},
																																	End: ast.Location{
																																		Line: int(959),
																																		Column: int(46),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "map",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p10465,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(959),
																																	Column: int(43),
																																},
																																End: ast.Location{
																																	Line: int(959),
																																	Column: int(50),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "escapeKeyToml",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10474,
																																		FreeVars: ast.Identifiers{
																																			"escapeKeyToml",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(959),
																																				Column: int(51),
																																			},
																																			End: ast.Location{
																																				Line: int(959),
																																				Column: int(64),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "path",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p10474,
																																		FreeVars: ast.Identifiers{
																																			"path",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(959),
																																				Column: int(66),
																																			},
																																			End: ast.Location{
																																				Line: int(959),
																																				Column: int(70),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p10465,
																														FreeVars: ast.Identifiers{
																															"escapeKeyToml",
																															"path",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(959),
																																Column: int(43),
																															},
																															End: ast.Location{
																																Line: int(959),
																																Column: int(71),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10411,
																										FreeVars: ast.Identifiers{
																											"escapeKeyToml",
																											"path",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(959),
																												Column: int(29),
																											},
																											End: ast.Location{
																												Line: int(959),
																												Column: int(72),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								Left: &ast.Binary{
																									Right: &ast.LiteralString{
																										Value: "[[",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10411,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(959),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(959),
																													Column: int(26),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Left: &ast.Var{
																										Id: "cindent",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10411,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(959),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(959),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p10411,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(959),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(959),
																												Column: int(26),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10411,
																									FreeVars: ast.Identifiers{
																										"cindent",
																										"escapeKeyToml",
																										"path",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(959),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(959),
																											Column: int(72),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10411,
																								FreeVars: ast.Identifiers{
																									"cindent",
																									"escapeKeyToml",
																									"path",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(959),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(959),
																										Column: int(79),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						OpFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(11),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10411,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"escapeKeyToml",
																								"i",
																								"path",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(959),
																									Column: int(12),
																								},
																								End: ast.Location{
																									Line: int(960),
																									Column: int(47),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(11),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10411,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"escapeKeyToml",
																							"i",
																							"indent",
																							"indexedPath",
																							"path",
																							"renderTableInternal",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(959),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(961),
																								Column: int(82),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"cindent",
																				"escapeKeyToml",
																				"i",
																				"indent",
																				"indexedPath",
																				"path",
																				"renderTableInternal",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: nil,
																			Name: "i",
																			CommaFodder: nil,
																			EqFodder: nil,
																			DefaultArg: nil,
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"cindent",
																			"escapeKeyToml",
																			"indent",
																			"indexedPath",
																			"path",
																			"renderTableInternal",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "range",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10495,
																		FreeVars: ast.Identifiers{
																			"range",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(962),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(962),
																				Column: int(25),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"cindent",
															"escapeKeyToml",
															"indent",
															"indexedPath",
															"path",
															"range",
															"renderTableInternal",
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(958),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(963),
																Column: int(10),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "sections",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(958),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(963),
														Column: int(10),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(964),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(964),
																Column: int(12),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "join",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10504,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(964),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(964),
															Column: int(17),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "\n\n",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10508,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(964),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(964),
																		Column: int(24),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "sections",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10508,
																FreeVars: ast.Identifiers{
																	"sections",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(964),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(964),
																		Column: int(34),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10504,
												FreeVars: ast.Identifiers{
													"sections",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(964),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(964),
														Column: int(35),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p10504,
											FreeVars: ast.Identifiers{
												"$std",
												"cindent",
												"escapeKeyToml",
												"indent",
												"indexedPath",
												"path",
												"range",
												"renderTableInternal",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(958),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(964),
													Column: int(35),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p10504,
										FreeVars: ast.Identifiers{
											"$std",
											"cindent",
											"escapeKeyToml",
											"indent",
											"indexedPath",
											"path",
											"renderTableInternal",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(957),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(964),
												Column: int(35),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(956),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(956),
												Column: int(25),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(956),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(956),
												Column: int(31),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "indexedPath",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(956),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(956),
												Column: int(44),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(956),
												Column: int(46),
											},
											End: ast.Location{
												Line: int(956),
												Column: int(53),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p10517,
									FreeVars: ast.Identifiers{
										"$std",
										"escapeKeyToml",
										"indent",
										"renderTableInternal",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(956),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(964),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "renderTableArray",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(965),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(965),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isObject",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10525,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(965),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(965),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "value",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10529,
												FreeVars: ast.Identifiers{
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(965),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(965),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p10525,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(965),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(965),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Var{
								Id: "renderTableInternal",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p10525,
									FreeVars: ast.Identifiers{
										"renderTableInternal",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(966),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(966),
											Column: int(26),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "value",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10538,
												FreeVars: ast.Identifiers{
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(966),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(966),
														Column: int(32),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											Elements: nil,
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10538,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(966),
														Column: int(34),
													},
													End: ast.Location{
														Line: int(966),
														Column: int(36),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											Elements: nil,
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10538,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(966),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(966),
														Column: int(40),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10538,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(966),
														Column: int(42),
													},
													End: ast.Location{
														Line: int(966),
														Column: int(44),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p10525,
								FreeVars: ast.Identifiers{
									"renderTableInternal",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(966),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(966),
										Column: int(45),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(968),
														Column: int(51),
													},
													End: ast.Location{
														Line: int(968),
														Column: int(54),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p10525,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(968),
													Column: int(51),
												},
												End: ast.Location{
													Line: int(968),
													Column: int(59),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "value",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10554,
														FreeVars: ast.Identifiers{
															"value",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(968),
																Column: int(60),
															},
															End: ast.Location{
																Line: int(968),
																Column: int(65),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10525,
										FreeVars: ast.Identifiers{
											"std",
											"value",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(968),
												Column: int(51),
											},
											End: ast.Location{
												Line: int(968),
												Column: int(66),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "TOML body must be an object. Got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10525,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(968),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(968),
												Column: int(48),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10525,
									FreeVars: ast.Identifiers{
										"std",
										"value",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(968),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(968),
											Column: int(66),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p10525,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(968),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(968),
										Column: int(66),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p10525,
							FreeVars: ast.Identifiers{
								"renderTableInternal",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(965),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(968),
									Column: int(66),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p10525,
						FreeVars: ast.Identifiers{
							"$std",
							"indent",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(891),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(968),
								Column: int(66),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(890),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(890),
								Column: int(23),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(890),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(890),
								Column: int(31),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(890),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(968),
					Column: int(66),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "escapeStringJson",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(971),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(971),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "toString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10577,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(971),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(971),
												Column: int(29),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str_",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10581,
													FreeVars: ast.Identifiers{
														"str_",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(971),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(971),
															Column: int(34),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10577,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(971),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(971),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "str",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(971),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(971),
									Column: int(35),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "\"",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10590,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(973),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(973),
															Column: int(19),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Var{
												Id: "ch",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10590,
													FreeVars: ast.Identifiers{
														"ch",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(973),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(973),
															Column: int(12),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10590,
												FreeVars: ast.Identifiers{
													"ch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(973),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(973),
														Column: int(19),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "\\\"",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p10590,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(974),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(974),
														Column: int(14),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralString{
													Value: "\\",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10590,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(975),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(975),
																Column: int(25),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												Left: &ast.Var{
													Id: "ch",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10590,
														FreeVars: ast.Identifiers{
															"ch",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(975),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(975),
																Column: int(17),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10590,
													FreeVars: ast.Identifiers{
														"ch",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(975),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(975),
															Column: int(25),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralString{
												Value: "\\\\",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p10590,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(976),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(976),
															Column: int(15),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.LiteralString{
														Value: "\b",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10590,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(977),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(977),
																	Column: int(25),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													Left: &ast.Var{
														Id: "ch",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10590,
															FreeVars: ast.Identifiers{
																"ch",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(977),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(977),
																	Column: int(17),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10590,
														FreeVars: ast.Identifiers{
															"ch",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(977),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(977),
																Column: int(25),
															},
														},
													},
													Op: ast.BinaryOp(12),
												},
												BranchTrue: &ast.LiteralString{
													Value: "\\b",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p10590,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(978),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(978),
																Column: int(14),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												BranchFalse: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.LiteralString{
															Value: "\f",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10590,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(979),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(979),
																		Column: int(25),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														Left: &ast.Var{
															Id: "ch",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10590,
																FreeVars: ast.Identifiers{
																	"ch",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(979),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(979),
																		Column: int(17),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10590,
															FreeVars: ast.Identifiers{
																"ch",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(979),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(979),
																	Column: int(25),
																},
															},
														},
														Op: ast.BinaryOp(12),
													},
													BranchTrue: &ast.LiteralString{
														Value: "\\f",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p10590,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(980),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(980),
																	Column: int(14),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													BranchFalse: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.LiteralString{
																Value: "\n",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10590,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(981),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(981),
																			Column: int(25),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Left: &ast.Var{
																Id: "ch",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10590,
																	FreeVars: ast.Identifiers{
																		"ch",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(981),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(981),
																			Column: int(17),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10590,
																FreeVars: ast.Identifiers{
																	"ch",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(981),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(981),
																		Column: int(25),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														BranchTrue: &ast.LiteralString{
															Value: "\\n",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p10590,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(982),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(982),
																		Column: int(14),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														BranchFalse: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.LiteralString{
																	Value: "\r",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10590,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(983),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(983),
																				Column: int(25),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																Left: &ast.Var{
																	Id: "ch",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10590,
																		FreeVars: ast.Identifiers{
																			"ch",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(983),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(983),
																				Column: int(17),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10590,
																	FreeVars: ast.Identifiers{
																		"ch",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(983),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(983),
																			Column: int(25),
																		},
																	},
																},
																Op: ast.BinaryOp(12),
															},
															BranchTrue: &ast.LiteralString{
																Value: "\\r",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p10590,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(984),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(984),
																			Column: int(14),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralString{
																		Value: "\t",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10590,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(985),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(985),
																					Column: int(25),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Left: &ast.Var{
																		Id: "ch",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p10590,
																			FreeVars: ast.Identifiers{
																				"ch",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(985),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(985),
																					Column: int(17),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10590,
																		FreeVars: ast.Identifiers{
																			"ch",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(985),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(985),
																				Column: int(25),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.LiteralString{
																	Value: "\\t",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p10590,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(986),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(986),
																				Column: int(14),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																BranchFalse: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(988),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(988),
																									Column: int(23),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "codepoint",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10651,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(988),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(988),
																								Column: int(33),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "ch",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10655,
																									FreeVars: ast.Identifiers{
																										"ch",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(988),
																											Column: int(34),
																										},
																										End: ast.Location{
																											Line: int(988),
																											Column: int(36),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10651,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(988),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(988),
																							Column: int(37),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "cp",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(988),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(988),
																					Column: int(37),
																				},
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.Binary{
																				Right: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "159",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10590,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(989),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(989),
																									Column: int(46),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "cp",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10590,
																							FreeVars: ast.Identifiers{
																								"cp",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(989),
																									Column: int(37),
																								},
																								End: ast.Location{
																									Line: int(989),
																									Column: int(39),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10590,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(989),
																								Column: int(37),
																							},
																							End: ast.Location{
																								Line: int(989),
																								Column: int(46),
																							},
																						},
																					},
																					Op: ast.BinaryOp(10),
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "127",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10590,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(989),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(989),
																									Column: int(33),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "cp",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p10590,
																							FreeVars: ast.Identifiers{
																								"cp",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(989),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(989),
																									Column: int(26),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10590,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(989),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(989),
																								Column: int(33),
																							},
																						},
																					},
																					Op: ast.BinaryOp(8),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10590,
																					FreeVars: ast.Identifiers{
																						"cp",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(989),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(989),
																							Column: int(46),
																						},
																					},
																				},
																				Op: ast.BinaryOp(17),
																			},
																			Left: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "32",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10590,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(989),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(989),
																								Column: int(19),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "cp",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p10590,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(989),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(989),
																								Column: int(14),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10590,
																					FreeVars: ast.Identifiers{
																						"cp",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(989),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(989),
																							Column: int(19),
																						},
																					},
																				},
																				Op: ast.BinaryOp(9),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p10590,
																				FreeVars: ast.Identifiers{
																					"cp",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(989),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(989),
																						Column: int(47),
																					},
																				},
																			},
																			Op: ast.BinaryOp(18),
																		},
																		BranchTrue: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "$std",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"$std",
																						},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "mod",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: nil,
																				LeftBracketFodder: nil,
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			FodderLeft: nil,
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "\\u%04x",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p10590,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(990),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(990),
																										Column: int(20),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: nil,
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "cp",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p10690,
																											FreeVars: ast.Identifiers{
																												"cp",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(990),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(990),
																													Column: int(26),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10590,
																								FreeVars: ast.Identifiers{
																									"cp",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(990),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(990),
																										Column: int(27),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: nil,
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"cp",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(990),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(990),
																						Column: int(27),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		BranchFalse: &ast.Var{
																			Id: "ch",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p10590,
																				FreeVars: ast.Identifiers{
																					"ch",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(992),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(992),
																						Column: int(13),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p10590,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"ch",
																				"cp",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(989),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(992),
																					Column: int(13),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p10590,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"ch",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(988),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(992),
																				Column: int(13),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p10590,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"ch",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(985),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(992),
																			Column: int(13),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10590,
																FreeVars: ast.Identifiers{
																	"$std",
																	"ch",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(983),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(992),
																		Column: int(13),
																	},
																},
															},
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10590,
															FreeVars: ast.Identifiers{
																"$std",
																"ch",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(981),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(992),
																	Column: int(13),
																},
															},
														},
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10590,
														FreeVars: ast.Identifiers{
															"$std",
															"ch",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(979),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(992),
																Column: int(13),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10590,
													FreeVars: ast.Identifiers{
														"$std",
														"ch",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(977),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(992),
															Column: int(13),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10590,
												FreeVars: ast.Identifiers{
													"$std",
													"ch",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(975),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(992),
														Column: int(13),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p10590,
											FreeVars: ast.Identifiers{
												"$std",
												"ch",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(973),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(992),
													Column: int(13),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(972),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(972),
													Column: int(19),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p10718,
										FreeVars: ast.Identifiers{
											"$std",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(972),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(992),
												Column: int(13),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "trans",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "\"%s\"",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p10729,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(993),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(993),
														Column: int(11),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(993),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(993),
																Column: int(17),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "join",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10729,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(993),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(993),
															Column: int(22),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10738,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(993),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(993),
																		Column: int(25),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "$std",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "flatMap",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: nil,
																LeftBracketFodder: nil,
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															FodderLeft: nil,
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Function{
																			ParenLeftFodder: nil,
																			ParenRightFodder: nil,
																			Body: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Var{
																								Id: "trans",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10751,
																									FreeVars: ast.Identifiers{
																										"trans",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(993),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(993),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "ch",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10755,
																												FreeVars: ast.Identifiers{
																													"ch",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(993),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(993),
																														Column: int(36),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10751,
																								FreeVars: ast.Identifiers{
																									"ch",
																									"trans",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(993),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(993),
																										Column: int(37),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"trans",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: nil,
																					Name: "ch",
																					CommaFodder: nil,
																					EqFodder: nil,
																					DefaultArg: nil,
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"trans",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		CommaFodder: nil,
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(993),
																								Column: int(48),
																							},
																							End: ast.Location{
																								Line: int(993),
																								Column: int(51),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "stringChars",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10738,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(993),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(993),
																							Column: int(63),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10769,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(993),
																										Column: int(64),
																									},
																									End: ast.Location{
																										Line: int(993),
																										Column: int(67),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p10738,
																				FreeVars: ast.Identifiers{
																					"std",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(993),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(993),
																						Column: int(68),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: nil,
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																	"std",
																	"str",
																	"trans",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(993),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(993),
																		Column: int(69),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10729,
												FreeVars: ast.Identifiers{
													"$std",
													"std",
													"str",
													"trans",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(993),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(993),
														Column: int(70),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"str",
									"trans",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(993),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(993),
										Column: int(70),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p10729,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(972),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(993),
									Column: int(70),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p10729,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"str_",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(971),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(993),
								Column: int(70),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(970),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(970),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(970),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(993),
					Column: int(70),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "escapeStringPython",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(996),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(996),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "escapeStringJson",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p10789,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(996),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(996),
									Column: int(25),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "str",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10793,
										FreeVars: ast.Identifiers{
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(996),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(996),
												Column: int(29),
											},
										},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p10789,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(996),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(996),
								Column: int(30),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(995),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(995),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(995),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(996),
					Column: int(30),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "escapeStringBash",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(999),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(999),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "toString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10807,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(999),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(999),
												Column: int(29),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str_",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10811,
													FreeVars: ast.Identifiers{
														"str_",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(999),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(999),
															Column: int(34),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10807,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(999),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(999),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "str",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(999),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(999),
									Column: int(35),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "'",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10820,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1001),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(1001),
															Column: int(19),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Var{
												Id: "ch",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10820,
													FreeVars: ast.Identifiers{
														"ch",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1001),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1001),
															Column: int(12),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10820,
												FreeVars: ast.Identifiers{
													"ch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1001),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1001),
														Column: int(19),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "'\"'\"'",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p10820,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1002),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1002),
														Column: int(18),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Var{
											Id: "ch",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p10820,
												FreeVars: ast.Identifiers{
													"ch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1004),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1004),
														Column: int(11),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p10820,
											FreeVars: ast.Identifiers{
												"ch",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1001),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1004),
													Column: int(11),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1000),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1000),
													Column: int(19),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p10833,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1000),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1004),
												Column: int(11),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "trans",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "'%s'",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
													},
												},
												Ctx: p10843,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1005),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(1005),
														Column: int(11),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1005),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(1005),
																Column: int(17),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "join",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10843,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1005),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(1005),
															Column: int(22),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10852,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1005),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1005),
																		Column: int(25),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "$std",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "flatMap",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: nil,
																LeftBracketFodder: nil,
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															FodderLeft: nil,
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Function{
																			ParenLeftFodder: nil,
																			ParenRightFodder: nil,
																			Body: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Var{
																								Id: "trans",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p10865,
																									FreeVars: ast.Identifiers{
																										"trans",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1005),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(1005),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "ch",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p10869,
																												FreeVars: ast.Identifiers{
																													"ch",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1005),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(1005),
																														Column: int(36),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10865,
																								FreeVars: ast.Identifiers{
																									"ch",
																									"trans",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1005),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(1005),
																										Column: int(37),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"trans",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: nil,
																					Name: "ch",
																					CommaFodder: nil,
																					EqFodder: nil,
																					DefaultArg: nil,
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"trans",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		CommaFodder: nil,
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1005),
																								Column: int(48),
																							},
																							End: ast.Location{
																								Line: int(1005),
																								Column: int(51),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "stringChars",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p10852,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1005),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(1005),
																							Column: int(63),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p10883,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1005),
																										Column: int(64),
																									},
																									End: ast.Location{
																										Line: int(1005),
																										Column: int(67),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p10852,
																				FreeVars: ast.Identifiers{
																					"std",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1005),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(1005),
																						Column: int(68),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: nil,
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																	"std",
																	"str",
																	"trans",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1005),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(1005),
																		Column: int(69),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10843,
												FreeVars: ast.Identifiers{
													"$std",
													"std",
													"str",
													"trans",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1005),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(1005),
														Column: int(70),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"str",
									"trans",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1005),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1005),
										Column: int(70),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p10843,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1000),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1005),
									Column: int(70),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p10843,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"str_",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(999),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1005),
								Column: int(70),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(998),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(998),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(998),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1005),
					Column: int(70),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "escapeStringDollars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1008),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1008),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "toString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10904,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1008),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1008),
												Column: int(29),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str_",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10908,
													FreeVars: ast.Identifiers{
														"str_",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1008),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1008),
															Column: int(34),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10904,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1008),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1008),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "str",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1008),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1008),
									Column: int(35),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "$",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10917,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1010),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(1010),
															Column: int(19),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Var{
												Id: "ch",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10917,
													FreeVars: ast.Identifiers{
														"ch",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1010),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1010),
															Column: int(12),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10917,
												FreeVars: ast.Identifiers{
													"ch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1010),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1010),
														Column: int(19),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "$$",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p10917,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1011),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1011),
														Column: int(13),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Var{
											Id: "ch",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p10917,
												FreeVars: ast.Identifiers{
													"ch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1013),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1013),
														Column: int(11),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p10917,
											FreeVars: ast.Identifiers{
												"ch",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1010),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1013),
													Column: int(11),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1009),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1009),
													Column: int(19),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p10930,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1009),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1013),
												Column: int(11),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "trans",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1014),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(1014),
												Column: int(8),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "foldl",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10937,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1014),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1014),
											Column: int(14),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											ParenLeftFodder: ast.Fodder{},
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Var{
														Id: "trans",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p10944,
															FreeVars: ast.Identifiers{
																"trans",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1014),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(1014),
																	Column: int(39),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p10948,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1014),
																				Column: int(40),
																			},
																			End: ast.Location{
																				Line: int(1014),
																				Column: int(41),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10944,
														FreeVars: ast.Identifiers{
															"b",
															"trans",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1014),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1014),
																Column: int(42),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Var{
													Id: "a",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p10944,
														FreeVars: ast.Identifiers{
															"a",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1014),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1014),
																Column: int(31),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10944,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"trans",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1014),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1014),
															Column: int(42),
														},
													},
												},
												Op: ast.BinaryOp(3),
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "a",
													CommaFodder: ast.Fodder{},
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1014),
															Column: int(24),
														},
														End: ast.Location{
															Line: int(1014),
															Column: int(25),
														},
													},
												},
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "b",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1014),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1014),
															Column: int(28),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10955,
												FreeVars: ast.Identifiers{
													"trans",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1014),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1014),
														Column: int(42),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1014),
																Column: int(44),
															},
															End: ast.Location{
																Line: int(1014),
																Column: int(47),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "stringChars",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10955,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1014),
															Column: int(44),
														},
														End: ast.Location{
															Line: int(1014),
															Column: int(59),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p10965,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1014),
																		Column: int(60),
																	},
																	End: ast.Location{
																		Line: int(1014),
																		Column: int(63),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10955,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1014),
														Column: int(44),
													},
													End: ast.Location{
														Line: int(1014),
														Column: int(64),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p10955,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1014),
														Column: int(66),
													},
													End: ast.Location{
														Line: int(1014),
														Column: int(68),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p10937,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"trans",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1014),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1014),
										Column: int(69),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p10937,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1009),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1014),
									Column: int(69),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p10937,
						FreeVars: ast.Identifiers{
							"std",
							"str_",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1008),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1014),
								Column: int(69),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1007),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(1007),
								Column: int(27),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1007),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1014),
					Column: int(69),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "escapeStringXML",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1025),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1025),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "toString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p10985,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1025),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1025),
												Column: int(29),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str_",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p10989,
													FreeVars: ast.Identifiers{
														"str_",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1025),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1025),
															Column: int(34),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p10985,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1025),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1025),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "str",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1025),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1025),
									Column: int(35),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1026),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1026),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "join",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p10998,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1026),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1026),
										Column: int(13),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										Value: "",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11002,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1026),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1026),
													Column: int(16),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "$std",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "flatMap",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: nil,
											LeftBracketFodder: nil,
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														ParenLeftFodder: nil,
														ParenRightFodder: nil,
														Body: &ast.Array{
															Elements: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1026),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1026),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "get",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11018,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1026),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(1026),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "xml_escapes",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11022,
																							FreeVars: ast.Identifiers{
																								"xml_escapes",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1026),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(1026),
																									Column: int(38),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "ch",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11022,
																							FreeVars: ast.Identifiers{
																								"ch",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1026),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1026),
																									Column: int(42),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "ch",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11022,
																							FreeVars: ast.Identifiers{
																								"ch",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1026),
																									Column: int(44),
																								},
																								End: ast.Location{
																									Line: int(1026),
																									Column: int(46),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11018,
																			FreeVars: ast.Identifiers{
																				"ch",
																				"std",
																				"xml_escapes",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1026),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(1026),
																					Column: int(47),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															CloseFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"ch",
																	"std",
																	"xml_escapes",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															TrailingComma: false,
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: nil,
																Name: "ch",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
																"xml_escapes",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														TrailingComma: false,
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1026),
																			Column: int(58),
																		},
																		End: ast.Location{
																			Line: int(1026),
																			Column: int(61),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "stringChars",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11002,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1026),
																		Column: int(58),
																	},
																	End: ast.Location{
																		Line: int(1026),
																		Column: int(73),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11040,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1026),
																					Column: int(74),
																				},
																				End: ast.Location{
																					Line: int(1026),
																					Column: int(77),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11002,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1026),
																	Column: int(58),
																},
																End: ast.Location{
																	Line: int(1026),
																	Column: int(78),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: nil,
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
												"std",
												"str",
												"xml_escapes",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1026),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1026),
													Column: int(79),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p10998,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"str",
								"xml_escapes",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1026),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1026),
									Column: int(80),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p10998,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"str_",
							"xml_escapes",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1025),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1026),
								Column: int(80),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1024),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1024),
								Column: int(23),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
						"xml_escapes",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1024),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1026),
					Column: int(80),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestJson",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1028),
										Column: int(25),
									},
									End: ast.Location{
										Line: int(1028),
										Column: int(28),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "manifestJsonEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p11056,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1028),
									Column: int(25),
								},
								End: ast.Location{
									Line: int(1028),
									Column: int(43),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "value",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11060,
										FreeVars: ast.Identifiers{
											"value",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1028),
												Column: int(44),
											},
											End: ast.Location{
												Line: int(1028),
												Column: int(49),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: "    ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11060,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1028),
												Column: int(51),
											},
											End: ast.Location{
												Line: int(1028),
												Column: int(57),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p11056,
						FreeVars: ast.Identifiers{
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1028),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(1028),
								Column: int(58),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1028),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1028),
								Column: int(21),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1028),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1028),
					Column: int(58),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestJsonMinified",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1030),
										Column: int(33),
									},
									End: ast.Location{
										Line: int(1030),
										Column: int(36),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "manifestJsonEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p11073,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1030),
									Column: int(33),
								},
								End: ast.Location{
									Line: int(1030),
									Column: int(51),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "value",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11077,
										FreeVars: ast.Identifiers{
											"value",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1030),
												Column: int(52),
											},
											End: ast.Location{
												Line: int(1030),
												Column: int(57),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: "",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11077,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1030),
												Column: int(59),
											},
											End: ast.Location{
												Line: int(1030),
												Column: int(61),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: "",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11077,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1030),
												Column: int(63),
											},
											End: ast.Location{
												Line: int(1030),
												Column: int(65),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									Value: ":",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p11077,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1030),
												Column: int(67),
											},
											End: ast.Location{
												Line: int(1030),
												Column: int(70),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p11073,
						FreeVars: ast.Identifiers{
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1030),
								Column: int(33),
							},
							End: ast.Location{
								Line: int(1030),
								Column: int(71),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1030),
								Column: int(24),
							},
							End: ast.Location{
								Line: int(1030),
								Column: int(29),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1030),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1030),
					Column: int(71),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestJsonEx",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11093,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1034),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1034),
														Column: int(19),
													},
												},
											},
											Value: true,
										},
										Left: &ast.Var{
											Id: "v",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11093,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1034),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1034),
														Column: int(11),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11093,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1034),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1034),
													Column: int(19),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.LiteralString{
										Value: "true",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p11093,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1035),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1035),
													Column: int(15),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11093,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1036),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1036),
															Column: int(25),
														},
													},
												},
												Value: false,
											},
											Left: &ast.Var{
												Id: "v",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11093,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1036),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1036),
															Column: int(16),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11093,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1036),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1036),
														Column: int(25),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "false",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p11093,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1037),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1037),
														Column: int(16),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11093,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1038),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1038),
																Column: int(24),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "v",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11093,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1038),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1038),
																Column: int(16),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11093,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1038),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1038),
															Column: int(24),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralString{
												Value: "null",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p11093,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1039),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1039),
															Column: int(15),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1040),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1040),
																		Column: int(18),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "isNumber",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11093,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1040),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1040),
																	Column: int(27),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "v",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11124,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1040),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1040),
																				Column: int(29),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11093,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1040),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1040),
																Column: int(30),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												BranchTrue: &ast.Binary{
													Right: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11093,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1041),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(1041),
																	Column: int(15),
																},
															},
														},
													},
													Left: &ast.LiteralString{
														Value: "",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p11093,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1041),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1041),
																	Column: int(11),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11093,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1041),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1041),
																Column: int(15),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												BranchFalse: &ast.Conditional{
													Cond: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1042),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1042),
																			Column: int(18),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "isString",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11093,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1042),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1042),
																		Column: int(27),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11142,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1042),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1042),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11093,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1042),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1042),
																	Column: int(30),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchTrue: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1043),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1043),
																			Column: int(12),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "escapeStringJson",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11093,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1043),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1043),
																		Column: int(29),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11154,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1043),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(1043),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11093,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1043),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1043),
																	Column: int(32),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchFalse: &ast.Conditional{
														Cond: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1044),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1044),
																				Column: int(18),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "isFunction",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11093,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1044),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1044),
																			Column: int(29),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "v",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11166,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1044),
																						Column: int(30),
																					},
																					End: ast.Location{
																						Line: int(1044),
																						Column: int(31),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11093,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1044),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1044),
																		Column: int(32),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														BranchTrue: &ast.Error{
															Expr: &ast.Binary{
																Right: &ast.Var{
																	Id: "path",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{
																			"path",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1045),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(1045),
																				Column: int(54),
																			},
																		},
																	},
																},
																Left: &ast.LiteralString{
																	Value: "Tried to manifest function at ",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1045),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1045),
																				Column: int(47),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11093,
																	FreeVars: ast.Identifiers{
																		"path",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1045),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1045),
																			Column: int(54),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p11093,
																FreeVars: ast.Identifiers{
																	"path",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1045),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1045),
																		Column: int(54),
																	},
																},
															},
														},
														BranchFalse: &ast.Conditional{
															Cond: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1046),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1046),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "isArray",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1046),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1046),
																				Column: int(26),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "v",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11186,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1046),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1046),
																							Column: int(28),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11093,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1046),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1046),
																			Column: int(29),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															BranchTrue: &ast.Local{
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: ast.Fodder{},
																		Body: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1047),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(1047),
																								Column: int(26),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "range",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11196,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1047),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1047),
																							Column: int(32),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralNumber{
																							OriginalString: "0",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11200,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1047),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1047),
																										Column: int(34),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11200,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1047),
																											Column: int(52),
																										},
																										End: ast.Location{
																											Line: int(1047),
																											Column: int(53),
																										},
																									},
																								},
																							},
																							Left: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1047),
																													Column: int(36),
																												},
																												End: ast.Location{
																													Line: int(1047),
																													Column: int(39),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "length",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11200,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1047),
																												Column: int(36),
																											},
																											End: ast.Location{
																												Line: int(1047),
																												Column: int(46),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "v",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p11211,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1047),
																															Column: int(47),
																														},
																														End: ast.Location{
																															Line: int(1047),
																															Column: int(48),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11200,
																									FreeVars: ast.Identifiers{
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1047),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(1047),
																											Column: int(49),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11200,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1047),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(1047),
																										Column: int(53),
																									},
																								},
																							},
																							Op: ast.BinaryOp(4),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11196,
																				FreeVars: ast.Identifiers{
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1047),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1047),
																						Column: int(54),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		EqFodder: ast.Fodder{},
																		Variable: "range",
																		CloseFodder: ast.Fodder{},
																		Fun: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1047),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1047),
																				Column: int(54),
																			},
																		},
																	},
																},
																Body: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				Right: &ast.Var{
																					Id: "indent",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11220,
																						FreeVars: ast.Identifiers{
																							"indent",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1048),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1048),
																								Column: int(44),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "cindent",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11220,
																						FreeVars: ast.Identifiers{
																							"cindent",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1048),
																								Column: int(28),
																							},
																							End: ast.Location{
																								Line: int(1048),
																								Column: int(35),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11220,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"indent",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1048),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1048),
																							Column: int(44),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "new_indent",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1048),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1048),
																					Column: int(44),
																				},
																			},
																		},
																	},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Body: &ast.Binary{
																					Right: &ast.Array{
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.LiteralString{
																										Value: "]",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11232,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1055),
																													Column: int(46),
																												},
																												End: ast.Location{
																													Line: int(1055),
																													Column: int(49),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									Left: &ast.Binary{
																										Right: &ast.Var{
																											Id: "cindent",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11232,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1055),
																														Column: int(36),
																													},
																													End: ast.Location{
																														Line: int(1055),
																														Column: int(43),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "newline",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11232,
																												FreeVars: ast.Identifiers{
																													"newline",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1055),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(1055),
																														Column: int(33),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11232,
																											FreeVars: ast.Identifiers{
																												"cindent",
																												"newline",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1055),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(1055),
																													Column: int(43),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11232,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"newline",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1055),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(1055),
																												Column: int(49),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: nil,
																							},
																						},
																						CloseFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11240,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"newline",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1055),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1055),
																									Column: int(50),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					Left: &ast.Binary{
																						Right: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1050),
																												Column: int(25),
																											},
																											End: ast.Location{
																												Line: int(1050),
																												Column: int(28),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "join",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11240,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1050),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1050),
																											Column: int(33),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Array{
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														Right: &ast.Var{
																															Id: "newline",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p11254,
																																FreeVars: ast.Identifiers{
																																	"newline",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1050),
																																		Column: int(41),
																																	},
																																	End: ast.Location{
																																		Line: int(1050),
																																		Column: int(48),
																																	},
																																},
																															},
																														},
																														Left: &ast.LiteralString{
																															Value: ",",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p11254,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1050),
																																		Column: int(35),
																																	},
																																	End: ast.Location{
																																		Line: int(1050),
																																		Column: int(38),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p11254,
																															FreeVars: ast.Identifiers{
																																"newline",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1050),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(1050),
																																	Column: int(48),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													CommaFodder: nil,
																												},
																											},
																											CloseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11258,
																												FreeVars: ast.Identifiers{
																													"newline",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1050),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(1050),
																														Column: int(49),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "$std",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"$std",
																														},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "flatMap",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: nil,
																												LeftBracketFodder: nil,
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															ParenLeftFodder: nil,
																															ParenRightFodder: nil,
																															Body: &ast.Array{
																																Elements: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Array{
																																			Elements: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Binary{
																																						Right: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "aux",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p11275,
																																									FreeVars: ast.Identifiers{
																																										"aux",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1052),
																																											Column: int(50),
																																										},
																																										End: ast.Location{
																																											Line: int(1052),
																																											Column: int(53),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "v",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p11280,
																																													FreeVars: ast.Identifiers{
																																														"v",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1052),
																																															Column: int(54),
																																														},
																																														End: ast.Location{
																																															Line: int(1052),
																																															Column: int(55),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.Var{
																																												Id: "i",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p11280,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1052),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(1052),
																																															Column: int(57),
																																														},
																																													},
																																												},
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11280,
																																												FreeVars: ast.Identifiers{
																																													"i",
																																													"v",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1052),
																																														Column: int(54),
																																													},
																																													End: ast.Location{
																																														Line: int(1052),
																																														Column: int(58),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Binary{
																																											Right: &ast.Array{
																																												Elements: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															Id: "i",
																																															NodeBase: ast.NodeBase{
																																																Fodder: ast.Fodder{},
																																																Ctx: p11289,
																																																FreeVars: ast.Identifiers{
																																																	"i",
																																																},
																																																LocRange: ast.LocationRange{
																																																	File: p8,
																																																	FileName: "",
																																																	Begin: ast.Location{
																																																		Line: int(1052),
																																																		Column: int(68),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(1052),
																																																		Column: int(69),
																																																	},
																																																},
																																															},
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												CloseFodder: ast.Fodder{},
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p11280,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1052),
																																															Column: int(67),
																																														},
																																														End: ast.Location{
																																															Line: int(1052),
																																															Column: int(70),
																																														},
																																													},
																																												},
																																												TrailingComma: false,
																																											},
																																											Left: &ast.Var{
																																												Id: "path",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p11280,
																																													FreeVars: ast.Identifiers{
																																														"path",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1052),
																																															Column: int(60),
																																														},
																																														End: ast.Location{
																																															Line: int(1052),
																																															Column: int(64),
																																														},
																																													},
																																												},
																																											},
																																											OpFodder: ast.Fodder{},
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11280,
																																												FreeVars: ast.Identifiers{
																																													"i",
																																													"path",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1052),
																																														Column: int(60),
																																													},
																																													End: ast.Location{
																																														Line: int(1052),
																																														Column: int(70),
																																													},
																																												},
																																											},
																																											Op: ast.BinaryOp(3),
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "new_indent",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11280,
																																												FreeVars: ast.Identifiers{
																																													"new_indent",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1052),
																																														Column: int(72),
																																													},
																																													End: ast.Location{
																																														Line: int(1052),
																																														Column: int(82),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p11275,
																																								FreeVars: ast.Identifiers{
																																									"aux",
																																									"i",
																																									"new_indent",
																																									"path",
																																									"v",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1052),
																																										Column: int(50),
																																									},
																																									End: ast.Location{
																																										Line: int(1052),
																																										Column: int(83),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						Left: &ast.Var{
																																							Id: "new_indent",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p11275,
																																								FreeVars: ast.Identifiers{
																																									"new_indent",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1052),
																																										Column: int(37),
																																									},
																																									End: ast.Location{
																																										Line: int(1052),
																																										Column: int(47),
																																									},
																																								},
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p11275,
																																							FreeVars: ast.Identifiers{
																																								"aux",
																																								"i",
																																								"new_indent",
																																								"path",
																																								"v",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1052),
																																									Column: int(37),
																																								},
																																								End: ast.Location{
																																									Line: int(1052),
																																									Column: int(83),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			CloseFodder: ast.Fodder{},
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Comment: []string{},
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(35),
																																					},
																																				},
																																				Ctx: p11302,
																																				FreeVars: ast.Identifiers{
																																					"aux",
																																					"i",
																																					"new_indent",
																																					"path",
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1052),
																																						Column: int(36),
																																					},
																																					End: ast.Location{
																																						Line: int(1052),
																																						Column: int(84),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																CloseFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																		"i",
																																		"new_indent",
																																		"path",
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																TrailingComma: false,
																															},
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "i",
																																	CommaFodder: nil,
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"new_indent",
																																	"path",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "range",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p11258,
																																FreeVars: ast.Identifiers{
																																	"range",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1053),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(1053),
																																		Column: int(50),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: nil,
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"aux",
																													"new_indent",
																													"path",
																													"range",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1051),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(1054),
																														Column: int(35),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11240,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"aux",
																									"new_indent",
																									"newline",
																									"path",
																									"range",
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1050),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1054),
																										Column: int(36),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						Left: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.Var{
																											Id: "newline",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11315,
																												FreeVars: ast.Identifiers{
																													"newline",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1049),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(1049),
																														Column: int(37),
																													},
																												},
																											},
																										},
																										Left: &ast.LiteralString{
																											Value: "[",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11315,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1049),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(1049),
																														Column: int(27),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11315,
																											FreeVars: ast.Identifiers{
																												"newline",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1049),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(1049),
																													Column: int(37),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11240,
																								FreeVars: ast.Identifiers{
																									"newline",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1049),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(1049),
																										Column: int(38),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						OpFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(22),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11240,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"aux",
																								"new_indent",
																								"newline",
																								"path",
																								"range",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1049),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(1054),
																									Column: int(36),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(22),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11240,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"aux",
																							"cindent",
																							"new_indent",
																							"newline",
																							"path",
																							"range",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1049),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(1055),
																								Column: int(50),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				EqFodder: ast.Fodder{},
																				Variable: "lines",
																				CloseFodder: ast.Fodder{},
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1049),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1055),
																						Column: int(50),
																					},
																				},
																			},
																		},
																		Body: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1056),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(1056),
																								Column: int(12),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "join",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11093,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1056),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(1056),
																							Column: int(17),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11333,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1056),
																										Column: int(18),
																									},
																									End: ast.Location{
																										Line: int(1056),
																										Column: int(20),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "lines",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11333,
																								FreeVars: ast.Identifiers{
																									"lines",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1056),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(1056),
																										Column: int(27),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11093,
																				FreeVars: ast.Identifiers{
																					"lines",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1056),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1056),
																						Column: int(28),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																				},
																			},
																			Ctx: p11093,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"aux",
																				"cindent",
																				"new_indent",
																				"newline",
																				"path",
																				"range",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1049),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(1056),
																					Column: int(28),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"aux",
																			"cindent",
																			"indent",
																			"newline",
																			"path",
																			"range",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1048),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1056),
																				Column: int(28),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11093,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"aux",
																		"cindent",
																		"indent",
																		"newline",
																		"path",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1047),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1056),
																			Column: int(28),
																		},
																	},
																},
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1057),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1057),
																						Column: int(18),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "isObject",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11093,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1057),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1057),
																					Column: int(27),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "v",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11352,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1057),
																								Column: int(28),
																							},
																							End: ast.Location{
																								Line: int(1057),
																								Column: int(29),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1057),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1057),
																				Column: int(30),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																BranchTrue: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				Right: &ast.Array{
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.LiteralString{
																									Value: "}",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11362,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1065),
																												Column: int(46),
																											},
																											End: ast.Location{
																												Line: int(1065),
																												Column: int(49),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								Left: &ast.Binary{
																									Right: &ast.Var{
																										Id: "cindent",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11362,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1065),
																													Column: int(36),
																												},
																												End: ast.Location{
																													Line: int(1065),
																													Column: int(43),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "newline",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11362,
																											FreeVars: ast.Identifiers{
																												"newline",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1065),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(1065),
																													Column: int(33),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11362,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"newline",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1065),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(1065),
																												Column: int(43),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11362,
																									FreeVars: ast.Identifiers{
																										"cindent",
																										"newline",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1065),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(1065),
																											Column: int(49),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: nil,
																						},
																					},
																					CloseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11370,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"newline",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1065),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1065),
																								Column: int(50),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				Left: &ast.Binary{
																					Right: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1059),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1059),
																											Column: int(28),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "join",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11370,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1059),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1059),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Array{
																										Elements: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													Right: &ast.Var{
																														Id: "newline",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p11384,
																															FreeVars: ast.Identifiers{
																																"newline",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1059),
																																	Column: int(41),
																																},
																																End: ast.Location{
																																	Line: int(1059),
																																	Column: int(48),
																																},
																															},
																														},
																													},
																													Left: &ast.LiteralString{
																														Value: ",",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p11384,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1059),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(1059),
																																	Column: int(38),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p11384,
																														FreeVars: ast.Identifiers{
																															"newline",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1059),
																																Column: int(35),
																															},
																															End: ast.Location{
																																Line: int(1059),
																																Column: int(48),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												CommaFodder: nil,
																											},
																										},
																										CloseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11388,
																											FreeVars: ast.Identifiers{
																												"newline",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1059),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(1059),
																													Column: int(49),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "$std",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "flatMap",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: nil,
																											LeftBracketFodder: nil,
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																										},
																										FodderLeft: nil,
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Function{
																														ParenLeftFodder: nil,
																														ParenRightFodder: nil,
																														Body: &ast.Array{
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Array{
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					Right: &ast.Apply{
																																						Target: &ast.Var{
																																							Id: "aux",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p11405,
																																								FreeVars: ast.Identifiers{
																																									"aux",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1062),
																																										Column: int(39),
																																									},
																																									End: ast.Location{
																																										Line: int(1062),
																																										Column: int(42),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "v",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"v",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(44),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.Var{
																																											Id: "k",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(45),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(46),
																																													},
																																												},
																																											},
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p11410,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																												"v",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1062),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(1062),
																																													Column: int(47),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Binary{
																																										Right: &ast.Array{
																																											Elements: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														Id: "k",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p11419,
																																															FreeVars: ast.Identifiers{
																																																"k",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1062),
																																																	Column: int(57),
																																																},
																																																End: ast.Location{
																																																	Line: int(1062),
																																																	Column: int(58),
																																																},
																																															},
																																														},
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											CloseFodder: ast.Fodder{},
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(59),
																																													},
																																												},
																																											},
																																											TrailingComma: false,
																																										},
																																										Left: &ast.Var{
																																											Id: "path",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"path",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(49),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(53),
																																													},
																																												},
																																											},
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p11410,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																												"path",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1062),
																																													Column: int(49),
																																												},
																																												End: ast.Location{
																																													Line: int(1062),
																																													Column: int(59),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Binary{
																																										Right: &ast.Var{
																																											Id: "indent",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"indent",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(71),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(77),
																																													},
																																												},
																																											},
																																										},
																																										Left: &ast.Var{
																																											Id: "cindent",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p11410,
																																												FreeVars: ast.Identifiers{
																																													"cindent",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1062),
																																														Column: int(61),
																																													},
																																													End: ast.Location{
																																														Line: int(1062),
																																														Column: int(68),
																																													},
																																												},
																																											},
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p11410,
																																											FreeVars: ast.Identifiers{
																																												"cindent",
																																												"indent",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1062),
																																													Column: int(61),
																																												},
																																												End: ast.Location{
																																													Line: int(1062),
																																													Column: int(77),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p11405,
																																							FreeVars: ast.Identifiers{
																																								"aux",
																																								"cindent",
																																								"indent",
																																								"k",
																																								"path",
																																								"v",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1062),
																																									Column: int(39),
																																								},
																																								End: ast.Location{
																																									Line: int(1062),
																																									Column: int(78),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					Left: &ast.Binary{
																																						Right: &ast.Var{
																																							Id: "key_val_sep",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p11405,
																																								FreeVars: ast.Identifiers{
																																									"key_val_sep",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1061),
																																										Column: int(82),
																																									},
																																									End: ast.Location{
																																										Line: int(1061),
																																										Column: int(93),
																																									},
																																								},
																																							},
																																						},
																																						Left: &ast.Binary{
																																							Right: &ast.Apply{
																																								Target: &ast.Index{
																																									Target: &ast.Var{
																																										Id: "std",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1061),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(1061),
																																													Column: int(59),
																																												},
																																											},
																																										},
																																									},
																																									Index: &ast.LiteralString{
																																										Value: "escapeStringJson",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: nil,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									LeftBracketFodder: ast.Fodder{},
																																									Id: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p11405,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1061),
																																												Column: int(56),
																																											},
																																											End: ast.Location{
																																												Line: int(1061),
																																												Column: int(76),
																																											},
																																										},
																																									},
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												Id: "k",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p11444,
																																													FreeVars: ast.Identifiers{
																																														"k",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1061),
																																															Column: int(77),
																																														},
																																														End: ast.Location{
																																															Line: int(1061),
																																															Column: int(78),
																																														},
																																													},
																																												},
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p11405,
																																									FreeVars: ast.Identifiers{
																																										"k",
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1061),
																																											Column: int(56),
																																										},
																																										End: ast.Location{
																																											Line: int(1061),
																																											Column: int(79),
																																										},
																																									},
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																							},
																																							Left: &ast.Binary{
																																								Right: &ast.Var{
																																									Id: "indent",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p11405,
																																										FreeVars: ast.Identifiers{
																																											"indent",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1061),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(1061),
																																												Column: int(53),
																																											},
																																										},
																																									},
																																								},
																																								Left: &ast.Var{
																																									Id: "cindent",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p11405,
																																										FreeVars: ast.Identifiers{
																																											"cindent",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1061),
																																												Column: int(37),
																																											},
																																											End: ast.Location{
																																												Line: int(1061),
																																												Column: int(44),
																																											},
																																										},
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p11405,
																																									FreeVars: ast.Identifiers{
																																										"cindent",
																																										"indent",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1061),
																																											Column: int(37),
																																										},
																																										End: ast.Location{
																																											Line: int(1061),
																																											Column: int(53),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							OpFodder: ast.Fodder{},
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p11405,
																																								FreeVars: ast.Identifiers{
																																									"cindent",
																																									"indent",
																																									"k",
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1061),
																																										Column: int(37),
																																									},
																																									End: ast.Location{
																																										Line: int(1061),
																																										Column: int(79),
																																									},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																						},
																																						OpFodder: ast.Fodder{},
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p11405,
																																							FreeVars: ast.Identifiers{
																																								"cindent",
																																								"indent",
																																								"k",
																																								"key_val_sep",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1061),
																																									Column: int(37),
																																								},
																																								End: ast.Location{
																																									Line: int(1061),
																																									Column: int(93),
																																								},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																					},
																																					OpFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Comment: []string{},
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(36),
																																						},
																																					},
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p11405,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																							"cindent",
																																							"indent",
																																							"k",
																																							"key_val_sep",
																																							"path",
																																							"std",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1061),
																																								Column: int(37),
																																							},
																																							End: ast.Location{
																																								Line: int(1062),
																																								Column: int(78),
																																							},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		CloseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Comment: []string{},
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(35),
																																				},
																																			},
																																			Ctx: p11458,
																																			FreeVars: ast.Identifiers{
																																				"aux",
																																				"cindent",
																																				"indent",
																																				"k",
																																				"key_val_sep",
																																				"path",
																																				"std",
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1061),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(1062),
																																					Column: int(79),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															CloseFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"cindent",
																																	"indent",
																																	"k",
																																	"key_val_sep",
																																	"path",
																																	"std",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														Parameters: []ast.Parameter{
																															ast.Parameter{
																																NameFodder: nil,
																																Name: "k",
																																CommaFodder: nil,
																																EqFodder: nil,
																																DefaultArg: nil,
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"aux",
																																"cindent",
																																"indent",
																																"key_val_sep",
																																"path",
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														TrailingComma: false,
																													},
																													CommaFodder: nil,
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1063),
																																			Column: int(45),
																																		},
																																		End: ast.Location{
																																			Line: int(1063),
																																			Column: int(48),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "objectFields",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p11388,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1063),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(1063),
																																		Column: int(61),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "v",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p11471,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1063),
																																					Column: int(62),
																																				},
																																				End: ast.Location{
																																					Line: int(1063),
																																					Column: int(63),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p11388,
																															FreeVars: ast.Identifiers{
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1063),
																																	Column: int(45),
																																},
																																End: ast.Location{
																																	Line: int(1063),
																																	Column: int(64),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: nil,
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$std",
																												"aux",
																												"cindent",
																												"indent",
																												"key_val_sep",
																												"path",
																												"std",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1060),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(1064),
																													Column: int(35),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11370,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"aux",
																								"cindent",
																								"indent",
																								"key_val_sep",
																								"newline",
																								"path",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1059),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1064),
																									Column: int(36),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					Left: &ast.Array{
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.Var{
																										Id: "newline",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11480,
																											FreeVars: ast.Identifiers{
																												"newline",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1058),
																													Column: int(30),
																												},
																												End: ast.Location{
																													Line: int(1058),
																													Column: int(37),
																												},
																											},
																										},
																									},
																									Left: &ast.LiteralString{
																										Value: "{",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11480,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1058),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(1058),
																													Column: int(27),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11480,
																										FreeVars: ast.Identifiers{
																											"newline",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1058),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(1058),
																												Column: int(37),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: nil,
																							},
																						},
																						CloseFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11370,
																							FreeVars: ast.Identifiers{
																								"newline",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1058),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(1058),
																									Column: int(38),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(22),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11370,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"aux",
																							"cindent",
																							"indent",
																							"key_val_sep",
																							"newline",
																							"path",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1058),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(1064),
																								Column: int(36),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				OpFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(22),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11370,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"aux",
																						"cindent",
																						"indent",
																						"key_val_sep",
																						"newline",
																						"path",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1058),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1065),
																							Column: int(50),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "lines",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1058),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1065),
																					Column: int(50),
																				},
																			},
																		},
																	},
																	Body: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1066),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(1066),
																							Column: int(12),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "join",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11093,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1066),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1066),
																						Column: int(17),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralString{
																						Value: "",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11498,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1066),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(1066),
																									Column: int(20),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "lines",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11498,
																							FreeVars: ast.Identifiers{
																								"lines",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1066),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(1066),
																									Column: int(27),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11093,
																			FreeVars: ast.Identifiers{
																				"lines",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1066),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(1066),
																					Column: int(28),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p11093,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"aux",
																			"cindent",
																			"indent",
																			"key_val_sep",
																			"newline",
																			"path",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1058),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1066),
																				Column: int(28),
																			},
																		},
																	},
																},
																BranchFalse: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11093,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"aux",
																		"cindent",
																		"indent",
																		"key_val_sep",
																		"newline",
																		"path",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1057),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(1066),
																			Column: int(28),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11093,
																FreeVars: ast.Identifiers{
																	"$std",
																	"aux",
																	"cindent",
																	"indent",
																	"key_val_sep",
																	"newline",
																	"path",
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1046),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1066),
																		Column: int(28),
																	},
																},
															},
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11093,
															FreeVars: ast.Identifiers{
																"$std",
																"aux",
																"cindent",
																"indent",
																"key_val_sep",
																"newline",
																"path",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1044),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1066),
																	Column: int(28),
																},
															},
														},
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11093,
														FreeVars: ast.Identifiers{
															"$std",
															"aux",
															"cindent",
															"indent",
															"key_val_sep",
															"newline",
															"path",
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1042),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1066),
																Column: int(28),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11093,
													FreeVars: ast.Identifiers{
														"$std",
														"aux",
														"cindent",
														"indent",
														"key_val_sep",
														"newline",
														"path",
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1040),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1066),
															Column: int(28),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11093,
												FreeVars: ast.Identifiers{
													"$std",
													"aux",
													"cindent",
													"indent",
													"key_val_sep",
													"newline",
													"path",
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1038),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1066),
														Column: int(28),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11093,
											FreeVars: ast.Identifiers{
												"$std",
												"aux",
												"cindent",
												"indent",
												"key_val_sep",
												"newline",
												"path",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1036),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1066),
													Column: int(28),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p11093,
										FreeVars: ast.Identifiers{
											"$std",
											"aux",
											"cindent",
											"indent",
											"key_val_sep",
											"newline",
											"path",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1034),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1066),
												Column: int(28),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1033),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1033),
												Column: int(16),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1033),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1033),
												Column: int(22),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1033),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1033),
												Column: int(31),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p11522,
									FreeVars: ast.Identifiers{
										"$std",
										"aux",
										"indent",
										"key_val_sep",
										"newline",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1033),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1066),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p11527,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1067),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1067),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "value",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11531,
											FreeVars: ast.Identifiers{
												"value",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1067),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1067),
													Column: int(14),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11531,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1067),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1067),
													Column: int(18),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										Value: "",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11531,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1067),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1067),
													Column: int(22),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p11527,
							FreeVars: ast.Identifiers{
								"aux",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1067),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1067),
									Column: int(23),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p11527,
						FreeVars: ast.Identifiers{
							"$std",
							"indent",
							"key_val_sep",
							"newline",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1033),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1067),
								Column: int(23),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1032),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1032),
								Column: int(23),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1032),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(1032),
								Column: int(31),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "newline",
						CommaFodder: ast.Fodder{},
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralString{
							Value: "\n",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p11527,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1032),
										Column: int(41),
									},
									End: ast.Location{
										Line: int(1032),
										Column: int(45),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1032),
								Column: int(33),
							},
							End: ast.Location{
								Line: int(1032),
								Column: int(45),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "key_val_sep",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralString{
							Value: ": ",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p11527,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1032),
										Column: int(59),
									},
									End: ast.Location{
										Line: int(1032),
										Column: int(63),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1032),
								Column: int(47),
							},
							End: ast.Location{
								Line: int(1032),
								Column: int(63),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1032),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1067),
					Column: int(23),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestYamlDoc",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1071),
																Column: int(55),
															},
															End: ast.Location{
																Line: int(1071),
																Column: int(58),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11554,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1071),
															Column: int(55),
														},
														End: ast.Location{
															Line: int(1071),
															Column: int(65),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "strSet",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11558,
																FreeVars: ast.Identifiers{
																	"strSet",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1071),
																		Column: int(66),
																	},
																	End: ast.Location{
																		Line: int(1071),
																		Column: int(72),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11554,
												FreeVars: ast.Identifiers{
													"std",
													"strSet",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1071),
														Column: int(55),
													},
													End: ast.Location{
														Line: int(1071),
														Column: int(73),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1071),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1071),
																Column: int(13),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11554,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1071),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1071),
															Column: int(20),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1071),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1071),
																				Column: int(24),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "setInter",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11573,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1071),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1071),
																			Column: int(33),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "charSet",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11577,
																				FreeVars: ast.Identifiers{
																					"charSet",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1071),
																						Column: int(34),
																					},
																					End: ast.Location{
																						Line: int(1071),
																						Column: int(41),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "strSet",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11577,
																				FreeVars: ast.Identifiers{
																					"strSet",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1071),
																						Column: int(43),
																					},
																					End: ast.Location{
																						Line: int(1071),
																						Column: int(49),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11573,
																FreeVars: ast.Identifiers{
																	"charSet",
																	"std",
																	"strSet",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1071),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1071),
																		Column: int(50),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p11554,
												FreeVars: ast.Identifiers{
													"charSet",
													"std",
													"strSet",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1071),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1071),
														Column: int(51),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11554,
											FreeVars: ast.Identifiers{
												"charSet",
												"std",
												"strSet",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1071),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1071),
													Column: int(73),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.LiteralBoolean{
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p11554,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1072),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1072),
													Column: int(13),
												},
											},
										},
										Value: true,
									},
									BranchFalse: &ast.LiteralBoolean{
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p11554,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1073),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1073),
													Column: int(17),
												},
											},
										},
										Value: false,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p11554,
										FreeVars: ast.Identifiers{
											"charSet",
											"std",
											"strSet",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1071),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1073),
												Column: int(17),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "charSet",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1070),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1070),
												Column: int(28),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "strSet",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1070),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(1070),
												Column: int(36),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p11591,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1070),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1073),
											Column: int(17),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "onlyChars",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Array{
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "true",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																		ast.FodderElement{
																			Comment: []string{
																				"// Boolean types taken from https://yaml.org/type/bool.html",
																			},
																			Kind: ast.FodderKind(2),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1080),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1080),
																			Column: int(15),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "false",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1081),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1081),
																			Column: int(16),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "yes",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1082),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1082),
																			Column: int(14),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "no",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1083),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1083),
																			Column: int(13),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "on",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1084),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1084),
																			Column: int(13),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "off",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1085),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1085),
																			Column: int(14),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "y",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1086),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1086),
																			Column: int(12),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "n",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1087),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1087),
																			Column: int(12),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: ".nan",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																		ast.FodderElement{
																			Comment: []string{
																				"// Numerical words taken from https://yaml.org/type/float.html",
																			},
																			Kind: ast.FodderKind(2),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1089),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1089),
																			Column: int(15),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "-.inf",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1090),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1090),
																			Column: int(16),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "+.inf",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1091),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1091),
																			Column: int(16),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: ".inf",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1092),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1092),
																			Column: int(15),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "null",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1093),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1093),
																			Column: int(15),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "-",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																		ast.FodderElement{
																			Comment: []string{
																				"// Invalid keys that contain no invalid characters",
																			},
																			Kind: ast.FodderKind(2),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1095),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1095),
																			Column: int(12),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "---",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1096),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1096),
																			Column: int(14),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralString{
																Value: "",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p11603,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1097),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1097),
																			Column: int(11),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															CommaFodder: ast.Fodder{},
														},
													},
													CloseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11637,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1078),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(1098),
																Column: int(8),
															},
														},
													},
													TrailingComma: true,
												},
												EqFodder: ast.Fodder{},
												Variable: "reserved",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1078),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1098),
														Column: int(8),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "$std",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "flatMap",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: nil,
															LeftBracketFodder: nil,
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Function{
																		ParenLeftFodder: nil,
																		ParenRightFodder: nil,
																		Body: &ast.Conditional{
																			Cond: &ast.Binary{
																				Right: &ast.Apply{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "std",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1099),
																										Column: int(57),
																									},
																									End: ast.Location{
																										Line: int(1099),
																										Column: int(60),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "asciiLower",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11655,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1099),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(1099),
																									Column: int(71),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "key",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11659,
																										FreeVars: ast.Identifiers{
																											"key",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1099),
																												Column: int(72),
																											},
																											End: ast.Location{
																												Line: int(1099),
																												Column: int(75),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11655,
																						FreeVars: ast.Identifiers{
																							"key",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1099),
																								Column: int(57),
																							},
																							End: ast.Location{
																								Line: int(1099),
																								Column: int(76),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				Left: &ast.Var{
																					Id: "word",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11655,
																						FreeVars: ast.Identifiers{
																							"word",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1099),
																								Column: int(49),
																							},
																							End: ast.Location{
																								Line: int(1099),
																								Column: int(53),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11655,
																					FreeVars: ast.Identifiers{
																						"key",
																						"std",
																						"word",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1099),
																							Column: int(49),
																						},
																						End: ast.Location{
																							Line: int(1099),
																							Column: int(76),
																						},
																					},
																				},
																				Op: ast.BinaryOp(12),
																			},
																			BranchTrue: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "word",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11668,
																								FreeVars: ast.Identifiers{
																									"word",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1099),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1099),
																										Column: int(24),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"word",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			BranchFalse: &ast.Array{
																				Elements: nil,
																				CloseFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			ThenFodder: nil,
																			ElseFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"key",
																					"std",
																					"word",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: nil,
																				Name: "word",
																				CommaFodder: nil,
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"key",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "reserved",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11655,
																			FreeVars: ast.Identifiers{
																				"reserved",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1099),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1099),
																					Column: int(45),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: nil,
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"key",
																"reserved",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1099),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1099),
																	Column: int(77),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													EqFodder: ast.Fodder{},
													Variable: "bad",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1099),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1099),
															Column: int(77),
														},
													},
												},
											},
											Body: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.LiteralNumber{
														OriginalString: "0",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11681,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1100),
																	Column: int(28),
																},
																End: ast.Location{
																	Line: int(1100),
																	Column: int(29),
																},
															},
														},
													},
													Left: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1100),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(1100),
																			Column: int(13),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "length",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11681,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1100),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(1100),
																		Column: int(20),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "bad",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11690,
																			FreeVars: ast.Identifiers{
																				"bad",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1100),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1100),
																					Column: int(24),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11681,
															FreeVars: ast.Identifiers{
																"bad",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1100),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(1100),
																	Column: int(25),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11681,
														FreeVars: ast.Identifiers{
															"bad",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1100),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1100),
																Column: int(29),
															},
														},
													},
													Op: ast.BinaryOp(7),
												},
												BranchTrue: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p11681,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1101),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1101),
																Column: int(13),
															},
														},
													},
													Value: true,
												},
												BranchFalse: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11681,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1102),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1102),
																Column: int(17),
															},
														},
													},
													Value: false,
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p11681,
													FreeVars: ast.Identifiers{
														"bad",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1100),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1102),
															Column: int(17),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p11681,
												FreeVars: ast.Identifiers{
													"$std",
													"key",
													"reserved",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1099),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1102),
														Column: int(17),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
												ast.FodderElement{
													Comment: []string{
														"// NOTE: These values are checked for case insensitively.",
													},
													Kind: ast.FodderKind(2),
													Blanks: int(0),
													Indent: int(6),
												},
												ast.FodderElement{
													Comment: []string{
														"// While this approach results in some false positives, it eliminates",
													},
													Kind: ast.FodderKind(2),
													Blanks: int(0),
													Indent: int(6),
												},
												ast.FodderElement{
													Comment: []string{
														"// the risk of missing a permutation.",
													},
													Kind: ast.FodderKind(2),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p11681,
											FreeVars: ast.Identifiers{
												"$std",
												"key",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1078),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1102),
													Column: int(17),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "key",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1074),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(1074),
													Column: int(25),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p11708,
										FreeVars: ast.Identifiers{
											"$std",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1074),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1102),
												Column: int(17),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "isReserved",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Body: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Binary{
													Right: &ast.Binary{
														Right: &ast.Var{
															Id: "type",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11718,
																FreeVars: ast.Identifiers{
																	"type",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1105),
																		Column: int(78),
																	},
																	End: ast.Location{
																		Line: int(1105),
																		Column: int(82),
																	},
																},
															},
														},
														Left: &ast.LiteralString{
															Value: "-",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11718,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1105),
																		Column: int(72),
																	},
																	End: ast.Location{
																		Line: int(1105),
																		Column: int(75),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11718,
															FreeVars: ast.Identifiers{
																"type",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1105),
																	Column: int(72),
																},
																End: ast.Location{
																	Line: int(1105),
																	Column: int(82),
																},
															},
														},
														Op: ast.BinaryOp(3),
													},
													Left: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1105),
																			Column: int(45),
																		},
																		End: ast.Location{
																			Line: int(1105),
																			Column: int(48),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "substr",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11718,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1105),
																		Column: int(45),
																	},
																	End: ast.Location{
																		Line: int(1105),
																		Column: int(55),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "m_key",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11730,
																			FreeVars: ast.Identifiers{
																				"m_key",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(56),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(61),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11730,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(63),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(64),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "3",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11730,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(66),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(67),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11718,
															FreeVars: ast.Identifiers{
																"m_key",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1105),
																	Column: int(45),
																},
																End: ast.Location{
																	Line: int(1105),
																	Column: int(68),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11718,
														FreeVars: ast.Identifiers{
															"m_key",
															"std",
															"type",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1105),
																Column: int(45),
															},
															End: ast.Location{
																Line: int(1105),
																Column: int(82),
															},
														},
													},
													Op: ast.BinaryOp(12),
												},
												Left: &ast.Binary{
													Right: &ast.Var{
														Id: "type",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11718,
															FreeVars: ast.Identifiers{
																"type",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1105),
																	Column: int(37),
																},
																End: ast.Location{
																	Line: int(1105),
																	Column: int(41),
																},
															},
														},
													},
													Left: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1105),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(1105),
																			Column: int(13),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "substr",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11718,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1105),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(1105),
																		Column: int(20),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "m_key",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11747,
																			FreeVars: ast.Identifiers{
																				"m_key",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(26),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11747,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "2",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11747,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1105),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1105),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p11718,
															FreeVars: ast.Identifiers{
																"m_key",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1105),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(1105),
																	Column: int(33),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p11718,
														FreeVars: ast.Identifiers{
															"m_key",
															"std",
															"type",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1105),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1105),
																Column: int(41),
															},
														},
													},
													Op: ast.BinaryOp(12),
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11718,
													FreeVars: ast.Identifiers{
														"m_key",
														"std",
														"type",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1105),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1105),
															Column: int(82),
														},
													},
												},
												Op: ast.BinaryOp(18),
											},
											BranchTrue: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p11718,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1106),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1106),
															Column: int(13),
														},
													},
												},
												Value: true,
											},
											BranchFalse: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p11718,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1107),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1107),
															Column: int(17),
														},
													},
												},
												Value: false,
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
													ast.FodderElement{
														Comment: []string{
															"// Look for positive or negative numerical types (ex: 0x)",
														},
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p11718,
												FreeVars: ast.Identifiers{
													"m_key",
													"std",
													"type",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1105),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1107),
														Column: int(17),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "m_key",
												CommaFodder: ast.Fodder{},
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1103),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1103),
														Column: int(26),
													},
												},
											},
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "type",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1103),
														Column: int(28),
													},
													End: ast.Location{
														Line: int(1103),
														Column: int(32),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: p11762,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1103),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(1107),
													Column: int(17),
												},
											},
										},
										TrailingComma: false,
									},
									EqFodder: nil,
									Variable: "typeMatch",
									CloseFodder: nil,
									Fun: nil,
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							Body: &ast.Local{
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: nil,
										Body: &ast.Function{
											ParenLeftFodder: ast.Fodder{},
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1124),
																				Column: int(23),
																			},
																			End: ast.Location{
																				Line: int(1124),
																				Column: int(26),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "set",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11774,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1124),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1124),
																			Column: int(30),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1124),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(1124),
																								Column: int(34),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "stringChars",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11782,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1124),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1124),
																							Column: int(46),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-/",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11786,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1124),
																										Column: int(47),
																									},
																									End: ast.Location{
																										Line: int(1124),
																										Column: int(104),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11782,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1124),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1124),
																						Column: int(105),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p11774,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1124),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1124),
																		Column: int(106),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														EqFodder: ast.Fodder{},
														Variable: "letters",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1124),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1124),
																Column: int(106),
															},
														},
													},
												},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1125),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1125),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "set",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11796,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1125),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1125),
																				Column: int(29),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1125),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(1125),
																									Column: int(33),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "stringChars",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11804,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1125),
																								Column: int(30),
																							},
																							End: ast.Location{
																								Line: int(1125),
																								Column: int(45),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralString{
																								Value: "0123456789",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11808,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1125),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(1125),
																											Column: int(58),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11804,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1125),
																							Column: int(30),
																						},
																						End: ast.Location{
																							Line: int(1125),
																							Column: int(59),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p11796,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1125),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1125),
																			Column: int(60),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															EqFodder: ast.Fodder{},
															Variable: "digits",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1125),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1125),
																	Column: int(60),
																},
															},
														},
													},
													Body: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Body: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1126),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1126),
																						Column: int(27),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "set",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11818,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1126),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1126),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1126),
																											Column: int(41),
																										},
																										End: ast.Location{
																											Line: int(1126),
																											Column: int(44),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "stringChars",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11827,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1126),
																										Column: int(41),
																									},
																									End: ast.Location{
																										Line: int(1126),
																										Column: int(56),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.LiteralString{
																										Value: "_-",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11831,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1126),
																													Column: int(57),
																												},
																												End: ast.Location{
																													Line: int(1126),
																													Column: int(61),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11827,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(41),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(62),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					Left: &ast.Var{
																						Id: "digits",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11827,
																							FreeVars: ast.Identifiers{
																								"digits",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(32),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(38),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11827,
																						FreeVars: ast.Identifiers{
																							"digits",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1126),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(1126),
																								Column: int(62),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p11818,
																		FreeVars: ast.Identifiers{
																			"digits",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1126),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1126),
																				Column: int(63),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																EqFodder: ast.Fodder{},
																Variable: "intChars",
																CloseFodder: ast.Fodder{},
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1126),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1126),
																		Column: int(63),
																	},
																},
															},
														},
														Body: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Body: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1127),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1127),
																							Column: int(27),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "set",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11844,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1127),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1127),
																						Column: int(31),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1127),
																												Column: int(43),
																											},
																											End: ast.Location{
																												Line: int(1127),
																												Column: int(46),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "stringChars",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11853,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1127),
																											Column: int(43),
																										},
																										End: ast.Location{
																											Line: int(1127),
																											Column: int(58),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.LiteralString{
																											Value: "b",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11857,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1127),
																														Column: int(59),
																													},
																													End: ast.Location{
																														Line: int(1127),
																														Column: int(62),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11853,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1127),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(1127),
																										Column: int(63),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						Left: &ast.Var{
																							Id: "intChars",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11853,
																								FreeVars: ast.Identifiers{
																									"intChars",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1127),
																										Column: int(32),
																									},
																									End: ast.Location{
																										Line: int(1127),
																										Column: int(40),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11853,
																							FreeVars: ast.Identifiers{
																								"intChars",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1127),
																									Column: int(32),
																								},
																								End: ast.Location{
																									Line: int(1127),
																									Column: int(63),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p11844,
																			FreeVars: ast.Identifiers{
																				"intChars",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1127),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1127),
																					Column: int(64),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	EqFodder: ast.Fodder{},
																	Variable: "binChars",
																	CloseFodder: ast.Fodder{},
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1127),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1127),
																			Column: int(64),
																		},
																	},
																},
															},
															Body: &ast.Local{
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: ast.Fodder{},
																		Body: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1128),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(1128),
																								Column: int(27),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "set",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11870,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1128),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1128),
																							Column: int(31),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1128),
																													Column: int(41),
																												},
																												End: ast.Location{
																													Line: int(1128),
																													Column: int(44),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "stringChars",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11879,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1128),
																												Column: int(41),
																											},
																											End: ast.Location{
																												Line: int(1128),
																												Column: int(56),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.LiteralString{
																												Value: "abcdefx_-",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p11883,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1128),
																															Column: int(57),
																														},
																														End: ast.Location{
																															Line: int(1128),
																															Column: int(68),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11879,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(41),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(69),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							Left: &ast.Var{
																								Id: "digits",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11879,
																									FreeVars: ast.Identifiers{
																										"digits",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(32),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(38),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11879,
																								FreeVars: ast.Identifiers{
																									"digits",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1128),
																										Column: int(32),
																									},
																									End: ast.Location{
																										Line: int(1128),
																										Column: int(69),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p11870,
																				FreeVars: ast.Identifiers{
																					"digits",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1128),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1128),
																						Column: int(70),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		EqFodder: ast.Fodder{},
																		Variable: "hexChars",
																		CloseFodder: ast.Fodder{},
																		Fun: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1128),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1128),
																				Column: int(70),
																			},
																		},
																	},
																},
																Body: &ast.Local{
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1129),
																									Column: int(26),
																								},
																								End: ast.Location{
																									Line: int(1129),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "set",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11896,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1129),
																								Column: int(26),
																							},
																							End: ast.Location{
																								Line: int(1129),
																								Column: int(33),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								Right: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1129),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(1129),
																														Column: int(46),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "stringChars",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11905,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1129),
																													Column: int(43),
																												},
																												End: ast.Location{
																													Line: int(1129),
																													Column: int(58),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralString{
																													Value: "e._-",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p11909,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1129),
																																Column: int(59),
																															},
																															End: ast.Location{
																																Line: int(1129),
																																Column: int(65),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11905,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1129),
																												Column: int(43),
																											},
																											End: ast.Location{
																												Line: int(1129),
																												Column: int(66),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								Left: &ast.Var{
																									Id: "digits",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11905,
																										FreeVars: ast.Identifiers{
																											"digits",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1129),
																												Column: int(34),
																											},
																											End: ast.Location{
																												Line: int(1129),
																												Column: int(40),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11905,
																									FreeVars: ast.Identifiers{
																										"digits",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1129),
																											Column: int(34),
																										},
																										End: ast.Location{
																											Line: int(1129),
																											Column: int(66),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p11896,
																					FreeVars: ast.Identifiers{
																						"digits",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1129),
																							Column: int(26),
																						},
																						End: ast.Location{
																							Line: int(1129),
																							Column: int(67),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			EqFodder: ast.Fodder{},
																			Variable: "floatChars",
																			CloseFodder: ast.Fodder{},
																			Fun: nil,
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1129),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1129),
																					Column: int(67),
																				},
																			},
																		},
																	},
																	Body: &ast.Local{
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Body: &ast.Apply{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "std",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1130),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1130),
																										Column: int(28),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "set",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11922,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1130),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1130),
																									Column: int(32),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									Right: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1130),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(1130),
																															Column: int(45),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "stringChars",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11931,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1130),
																														Column: int(42),
																													},
																													End: ast.Location{
																														Line: int(1130),
																														Column: int(57),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														Value: "-",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p11935,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1130),
																																	Column: int(58),
																																},
																																End: ast.Location{
																																	Line: int(1130),
																																	Column: int(61),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11931,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1130),
																													Column: int(42),
																												},
																												End: ast.Location{
																													Line: int(1130),
																													Column: int(62),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									Left: &ast.Var{
																										Id: "digits",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11931,
																											FreeVars: ast.Identifiers{
																												"digits",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1130),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(1130),
																													Column: int(39),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11931,
																										FreeVars: ast.Identifiers{
																											"digits",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1130),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(1130),
																												Column: int(62),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p11922,
																						FreeVars: ast.Identifiers{
																							"digits",
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1130),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1130),
																								Column: int(63),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				EqFodder: ast.Fodder{},
																				Variable: "dateChars",
																				CloseFodder: ast.Fodder{},
																				Fun: nil,
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1130),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1130),
																						Column: int(63),
																					},
																				},
																			},
																		},
																		Body: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1131),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1131),
																											Column: int(28),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "set",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11948,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1131),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1131),
																										Column: int(32),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										Right: &ast.Var{
																											Id: "floatChars",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11953,
																												FreeVars: ast.Identifiers{
																													"floatChars",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1131),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(1131),
																														Column: int(53),
																													},
																												},
																											},
																										},
																										Left: &ast.Var{
																											Id: "letters",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11953,
																												FreeVars: ast.Identifiers{
																													"letters",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1131),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(1131),
																														Column: int(40),
																													},
																												},
																											},
																										},
																										OpFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11953,
																											FreeVars: ast.Identifiers{
																												"floatChars",
																												"letters",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1131),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(1131),
																													Column: int(53),
																												},
																											},
																										},
																										Op: ast.BinaryOp(3),
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p11948,
																							FreeVars: ast.Identifiers{
																								"floatChars",
																								"letters",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1131),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1131),
																									Column: int(54),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "safeChars",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1131),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(1131),
																							Column: int(54),
																						},
																					},
																				},
																			},
																			Body: &ast.Local{
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: ast.Fodder{},
																						Body: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1132),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(1132),
																												Column: int(24),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "asciiLower",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11966,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1132),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1132),
																											Column: int(35),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "key",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p11970,
																												FreeVars: ast.Identifiers{
																													"key",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1132),
																														Column: int(36),
																													},
																													End: ast.Location{
																														Line: int(1132),
																														Column: int(39),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p11966,
																								FreeVars: ast.Identifiers{
																									"key",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1132),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(1132),
																										Column: int(40),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						EqFodder: ast.Fodder{},
																						Variable: "keyLc",
																						CloseFodder: ast.Fodder{},
																						Fun: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1132),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(1132),
																								Column: int(40),
																							},
																						},
																					},
																				},
																				Body: &ast.Local{
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Body: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1133),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(1133),
																													Column: int(27),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "stringChars",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11980,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1133),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(1133),
																												Column: int(39),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "key",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p11984,
																													FreeVars: ast.Identifiers{
																														"key",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1133),
																															Column: int(40),
																														},
																														End: ast.Location{
																															Line: int(1133),
																															Column: int(43),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p11980,
																									FreeVars: ast.Identifiers{
																										"key",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1133),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(1133),
																											Column: int(44),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							EqFodder: ast.Fodder{},
																							Variable: "keyChars",
																							CloseFodder: ast.Fodder{},
																							Fun: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1133),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(1133),
																									Column: int(44),
																								},
																							},
																						},
																					},
																					Body: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Body: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1134),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(1134),
																														Column: int(25),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "set",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p11994,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1134),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(1134),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "keyChars",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p11998,
																														FreeVars: ast.Identifiers{
																															"keyChars",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1134),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(1134),
																																Column: int(38),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p11994,
																										FreeVars: ast.Identifiers{
																											"keyChars",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1134),
																												Column: int(22),
																											},
																											End: ast.Location{
																												Line: int(1134),
																												Column: int(39),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								EqFodder: ast.Fodder{},
																								Variable: "keySet",
																								CloseFodder: ast.Fodder{},
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1134),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(1134),
																										Column: int(39),
																									},
																								},
																							},
																						},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1135),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(1135),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "set",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12008,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1135),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(1135),
																														Column: int(31),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1135),
																																			Column: int(32),
																																		},
																																		End: ast.Location{
																																			Line: int(1135),
																																			Column: int(35),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "stringChars",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12016,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1135),
																																		Column: int(32),
																																	},
																																	End: ast.Location{
																																		Line: int(1135),
																																		Column: int(47),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "keyLc",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12020,
																																			FreeVars: ast.Identifiers{
																																				"keyLc",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1135),
																																					Column: int(48),
																																				},
																																				End: ast.Location{
																																					Line: int(1135),
																																					Column: int(53),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12016,
																															FreeVars: ast.Identifiers{
																																"keyLc",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1135),
																																	Column: int(32),
																																},
																																End: ast.Location{
																																	Line: int(1135),
																																	Column: int(54),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12008,
																											FreeVars: ast.Identifiers{
																												"keyLc",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1135),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(1135),
																													Column: int(55),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									EqFodder: ast.Fodder{},
																									Variable: "keySetLc",
																									CloseFodder: ast.Fodder{},
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1135),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(1135),
																											Column: int(55),
																										},
																									},
																								},
																							},
																							Body: &ast.Conditional{
																								Cond: &ast.Unary{
																									Expr: &ast.Apply{
																										Target: &ast.Var{
																											Id: "onlyChars",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12028,
																												FreeVars: ast.Identifiers{
																													"onlyChars",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1137),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(1137),
																														Column: int(20),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "safeChars",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12032,
																															FreeVars: ast.Identifiers{
																																"safeChars",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1137),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(1137),
																																	Column: int(30),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "keySet",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12032,
																															FreeVars: ast.Identifiers{
																																"keySet",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1137),
																																	Column: int(32),
																																},
																																End: ast.Location{
																																	Line: int(1137),
																																	Column: int(38),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12028,
																											FreeVars: ast.Identifiers{
																												"keySet",
																												"onlyChars",
																												"safeChars",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1137),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(1137),
																													Column: int(39),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12028,
																										FreeVars: ast.Identifiers{
																											"keySet",
																											"onlyChars",
																											"safeChars",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1137),
																												Column: int(10),
																											},
																											End: ast.Location{
																												Line: int(1137),
																												Column: int(39),
																											},
																										},
																									},
																									Op: ast.UnaryOp(0),
																								},
																								BranchTrue: &ast.LiteralBoolean{
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(8),
																											},
																										},
																										Ctx: p12028,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1138),
																												Column: int(9),
																											},
																											End: ast.Location{
																												Line: int(1138),
																												Column: int(14),
																											},
																										},
																									},
																									Value: false,
																								},
																								BranchFalse: &ast.Conditional{
																									Cond: &ast.Apply{
																										Target: &ast.Var{
																											Id: "isReserved",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12028,
																												FreeVars: ast.Identifiers{
																													"isReserved",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1140),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(1140),
																														Column: int(25),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "key",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12046,
																															FreeVars: ast.Identifiers{
																																"key",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1140),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(1140),
																																	Column: int(29),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12028,
																											FreeVars: ast.Identifiers{
																												"isReserved",
																												"key",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1140),
																													Column: int(15),
																												},
																												End: ast.Location{
																													Line: int(1140),
																													Column: int(30),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									BranchTrue: &ast.LiteralBoolean{
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(8),
																												},
																											},
																											Ctx: p12028,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1141),
																													Column: int(9),
																												},
																												End: ast.Location{
																													Line: int(1141),
																													Column: int(14),
																												},
																											},
																										},
																										Value: false,
																									},
																									BranchFalse: &ast.Conditional{
																										Cond: &ast.Binary{
																											Right: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "2",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1150),
																																Column: int(58),
																															},
																															End: ast.Location{
																																Line: int(1150),
																																Column: int(59),
																															},
																														},
																													},
																												},
																												Left: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1150),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(1150),
																																		Column: int(21),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "length",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1150),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(1150),
																																	Column: int(28),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Apply{
																																	Target: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "std",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1150),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(1150),
																																						Column: int(32),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "findSubstr",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12067,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1150),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(1150),
																																					Column: int(43),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralString{
																																					Value: "-",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12071,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1150),
																																								Column: int(44),
																																							},
																																							End: ast.Location{
																																								Line: int(1150),
																																								Column: int(47),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "key",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12071,
																																						FreeVars: ast.Identifiers{
																																							"key",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1150),
																																								Column: int(49),
																																							},
																																							End: ast.Location{
																																								Line: int(1150),
																																								Column: int(52),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12067,
																																		FreeVars: ast.Identifiers{
																																			"key",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1150),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(1150),
																																				Column: int(53),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"key",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1150),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(1150),
																																Column: int(54),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12028,
																													FreeVars: ast.Identifiers{
																														"key",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1150),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(1150),
																															Column: int(59),
																														},
																													},
																												},
																												Op: ast.BinaryOp(12),
																											},
																											Left: &ast.Apply{
																												Target: &ast.Var{
																													Id: "onlyChars",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"onlyChars",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1149),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1149),
																																Column: int(24),
																															},
																														},
																													},
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "dateChars",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12082,
																																	FreeVars: ast.Identifiers{
																																		"dateChars",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1149),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(1149),
																																			Column: int(34),
																																		},
																																	},
																																},
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "keySet",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12082,
																																	FreeVars: ast.Identifiers{
																																		"keySet",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1149),
																																			Column: int(36),
																																		},
																																		End: ast.Location{
																																			Line: int(1149),
																																			Column: int(42),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12028,
																													FreeVars: ast.Identifiers{
																														"dateChars",
																														"keySet",
																														"onlyChars",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1149),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1149),
																															Column: int(43),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											OpFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(14),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12028,
																												FreeVars: ast.Identifiers{
																													"dateChars",
																													"key",
																													"keySet",
																													"onlyChars",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1149),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(1150),
																														Column: int(59),
																													},
																												},
																											},
																											Op: ast.BinaryOp(17),
																										},
																										BranchTrue: &ast.LiteralBoolean{
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																													},
																												},
																												Ctx: p12028,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1151),
																														Column: int(9),
																													},
																													End: ast.Location{
																														Line: int(1151),
																														Column: int(14),
																													},
																												},
																											},
																											Value: false,
																										},
																										BranchFalse: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "2",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1158),
																																	Column: int(57),
																																},
																																End: ast.Location{
																																	Line: int(1158),
																																	Column: int(58),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1158),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(1158),
																																			Column: int(21),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "length",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1158),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(1158),
																																		Column: int(28),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1158),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(1158),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "findSubstr",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12107,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1158),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(1158),
																																						Column: int(43),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.LiteralString{
																																						Value: "-",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12111,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1158),
																																									Column: int(44),
																																								},
																																								End: ast.Location{
																																									Line: int(1158),
																																									Column: int(47),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					CommaFodder: ast.Fodder{},
																																				},
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "key",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12111,
																																							FreeVars: ast.Identifiers{
																																								"key",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1158),
																																									Column: int(49),
																																								},
																																								End: ast.Location{
																																									Line: int(1158),
																																									Column: int(52),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12107,
																																			FreeVars: ast.Identifiers{
																																				"key",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1158),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(1158),
																																					Column: int(53),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"key",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1158),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(1158),
																																	Column: int(54),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"key",
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1158),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(1158),
																																Column: int(58),
																															},
																														},
																													},
																													Op: ast.BinaryOp(9),
																												},
																												Left: &ast.Apply{
																													Target: &ast.Var{
																														Id: "onlyChars",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"onlyChars",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1157),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1157),
																																	Column: int(24),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "intChars",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12122,
																																		FreeVars: ast.Identifiers{
																																			"intChars",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1157),
																																				Column: int(25),
																																			},
																																			End: ast.Location{
																																				Line: int(1157),
																																				Column: int(33),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "keySetLc",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12122,
																																		FreeVars: ast.Identifiers{
																																			"keySetLc",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1157),
																																				Column: int(35),
																																			},
																																			End: ast.Location{
																																				Line: int(1157),
																																				Column: int(43),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"intChars",
																															"keySetLc",
																															"onlyChars",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1157),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1157),
																																Column: int(44),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(14),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12028,
																													FreeVars: ast.Identifiers{
																														"intChars",
																														"key",
																														"keySetLc",
																														"onlyChars",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1157),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1158),
																															Column: int(58),
																														},
																													},
																												},
																												Op: ast.BinaryOp(17),
																											},
																											BranchTrue: &ast.LiteralBoolean{
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																														},
																													},
																													Ctx: p12028,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1159),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(1159),
																															Column: int(14),
																														},
																													},
																												},
																												Value: false,
																											},
																											BranchFalse: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.Apply{
																														Target: &ast.Var{
																															Id: "typeMatch",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"typeMatch",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1168),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(1168),
																																		Column: int(27),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "key",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12138,
																																			FreeVars: ast.Identifiers{
																																				"key",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1168),
																																					Column: int(28),
																																				},
																																				End: ast.Location{
																																					Line: int(1168),
																																					Column: int(31),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralString{
																																		Value: "0b",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12138,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1168),
																																					Column: int(33),
																																				},
																																				End: ast.Location{
																																					Line: int(1168),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"key",
																																"typeMatch",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1168),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(1168),
																																	Column: int(38),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													Left: &ast.Binary{
																														Right: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "2",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1167),
																																			Column: int(36),
																																		},
																																		End: ast.Location{
																																			Line: int(1167),
																																			Column: int(37),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1167),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1167),
																																					Column: int(21),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "length",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1167),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1167),
																																				Column: int(28),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "key",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12153,
																																					FreeVars: ast.Identifiers{
																																						"key",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1167),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(1167),
																																							Column: int(32),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"key",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1167),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(1167),
																																			Column: int(33),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"key",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1167),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(1167),
																																		Column: int(37),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(7),
																														},
																														Left: &ast.Apply{
																															Target: &ast.Var{
																																Id: "onlyChars",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"onlyChars",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1166),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(1166),
																																			Column: int(24),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "binChars",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12162,
																																				FreeVars: ast.Identifiers{
																																					"binChars",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1166),
																																						Column: int(25),
																																					},
																																					End: ast.Location{
																																						Line: int(1166),
																																						Column: int(33),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "keySetLc",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12162,
																																				FreeVars: ast.Identifiers{
																																					"keySetLc",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1166),
																																						Column: int(35),
																																					},
																																					End: ast.Location{
																																						Line: int(1166),
																																						Column: int(43),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"binChars",
																																	"keySetLc",
																																	"onlyChars",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1166),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(1166),
																																		Column: int(44),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														OpFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(14),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"binChars",
																																"key",
																																"keySetLc",
																																"onlyChars",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1166),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1167),
																																	Column: int(37),
																																},
																															},
																														},
																														Op: ast.BinaryOp(17),
																													},
																													OpFodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(14),
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"binChars",
																															"key",
																															"keySetLc",
																															"onlyChars",
																															"std",
																															"typeMatch",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1166),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1168),
																																Column: int(38),
																															},
																														},
																													},
																													Op: ast.BinaryOp(17),
																												},
																												BranchTrue: &ast.LiteralBoolean{
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																															},
																														},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1169),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(1169),
																																Column: int(14),
																															},
																														},
																													},
																													Value: false,
																												},
																												BranchFalse: &ast.Conditional{
																													Cond: &ast.Binary{
																														Right: &ast.Binary{
																															Right: &ast.LiteralNumber{
																																OriginalString: "2",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1180),
																																			Column: int(59),
																																		},
																																		End: ast.Location{
																																			Line: int(1180),
																																			Column: int(60),
																																		},
																																	},
																																},
																															},
																															Left: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1180),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1180),
																																					Column: int(21),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "length",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1180),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1180),
																																				Column: int(28),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Apply{
																																				Target: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "std",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1180),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(1180),
																																									Column: int(32),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.LiteralString{
																																						Value: "findSubstr",
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																						NodeBase: ast.NodeBase{
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{},
																																							LocRange: ast.LocationRange{
																																								File: nil,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																							},
																																						},
																																						Kind: ast.LiteralStringKind(1),
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12189,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1180),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(1180),
																																								Column: int(43),
																																							},
																																						},
																																					},
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.LiteralString{
																																								Value: "e",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p12193,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1180),
																																											Column: int(44),
																																										},
																																										End: ast.Location{
																																											Line: int(1180),
																																											Column: int(47),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								Id: "keyLc",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p12193,
																																									FreeVars: ast.Identifiers{
																																										"keyLc",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1180),
																																											Column: int(49),
																																										},
																																										End: ast.Location{
																																											Line: int(1180),
																																											Column: int(54),
																																										},
																																									},
																																								},
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12189,
																																					FreeVars: ast.Identifiers{
																																						"keyLc",
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1180),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(1180),
																																							Column: int(55),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"keyLc",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1180),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(1180),
																																			Column: int(56),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"keyLc",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1180),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(1180),
																																		Column: int(60),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(9),
																														},
																														Left: &ast.Binary{
																															Right: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "3",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1179),
																																				Column: int(57),
																																			},
																																			End: ast.Location{
																																				Line: int(1179),
																																				Column: int(58),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Apply{
																																	Target: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "std",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1179),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(1179),
																																						Column: int(21),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "length",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1179),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1179),
																																					Column: int(28),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Apply{
																																					Target: &ast.Index{
																																						Target: &ast.Var{
																																							Id: "std",
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1179),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(1179),
																																										Column: int(32),
																																									},
																																								},
																																							},
																																						},
																																						Index: &ast.LiteralString{
																																							Value: "findSubstr",
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																							NodeBase: ast.NodeBase{
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{},
																																								LocRange: ast.LocationRange{
																																									File: nil,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																								},
																																							},
																																							Kind: ast.LiteralStringKind(1),
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						LeftBracketFodder: ast.Fodder{},
																																						Id: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12214,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1179),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(1179),
																																									Column: int(43),
																																								},
																																							},
																																						},
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.LiteralString{
																																									Value: "-",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p12218,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1179),
																																												Column: int(44),
																																											},
																																											End: ast.Location{
																																												Line: int(1179),
																																												Column: int(47),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									Id: "key",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p12218,
																																										FreeVars: ast.Identifiers{
																																											"key",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1179),
																																												Column: int(49),
																																											},
																																											End: ast.Location{
																																												Line: int(1179),
																																												Column: int(52),
																																											},
																																										},
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12214,
																																						FreeVars: ast.Identifiers{
																																							"key",
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1179),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(1179),
																																								Column: int(53),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"key",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1179),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1179),
																																				Column: int(54),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"key",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1179),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(1179),
																																			Column: int(58),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(9),
																															},
																															Left: &ast.Binary{
																																Right: &ast.Binary{
																																	Right: &ast.LiteralNumber{
																																		OriginalString: "1",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1178),
																																					Column: int(58),
																																				},
																																				End: ast.Location{
																																					Line: int(1178),
																																					Column: int(59),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1178),
																																							Column: int(18),
																																						},
																																						End: ast.Location{
																																							Line: int(1178),
																																							Column: int(21),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "length",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12028,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1178),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(1178),
																																						Column: int(28),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Apply{
																																						Target: &ast.Index{
																																							Target: &ast.Var{
																																								Id: "std",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1178),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(1178),
																																											Column: int(32),
																																										},
																																									},
																																								},
																																							},
																																							Index: &ast.LiteralString{
																																								Value: "findSubstr",
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																								NodeBase: ast.NodeBase{
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{},
																																									LocRange: ast.LocationRange{
																																										File: nil,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																									},
																																								},
																																								Kind: ast.LiteralStringKind(1),
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							LeftBracketFodder: ast.Fodder{},
																																							Id: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p12239,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1178),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(1178),
																																										Column: int(43),
																																									},
																																								},
																																							},
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.LiteralString{
																																										Value: ".",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p12243,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1178),
																																													Column: int(44),
																																												},
																																												End: ast.Location{
																																													Line: int(1178),
																																													Column: int(47),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										Id: "key",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p12243,
																																											FreeVars: ast.Identifiers{
																																												"key",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1178),
																																													Column: int(49),
																																												},
																																												End: ast.Location{
																																													Line: int(1178),
																																													Column: int(52),
																																												},
																																											},
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12239,
																																							FreeVars: ast.Identifiers{
																																								"key",
																																								"std",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1178),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(1178),
																																									Column: int(53),
																																								},
																																							},
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"key",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1178),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1178),
																																					Column: int(54),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"key",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1178),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1178),
																																				Column: int(59),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(12),
																																},
																																Left: &ast.Apply{
																																	Target: &ast.Var{
																																		Id: "onlyChars",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"onlyChars",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1177),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(1177),
																																					Column: int(24),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "floatChars",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12254,
																																						FreeVars: ast.Identifiers{
																																							"floatChars",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1177),
																																								Column: int(25),
																																							},
																																							End: ast.Location{
																																								Line: int(1177),
																																								Column: int(35),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "keySetLc",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12254,
																																						FreeVars: ast.Identifiers{
																																							"keySetLc",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1177),
																																								Column: int(37),
																																							},
																																							End: ast.Location{
																																								Line: int(1177),
																																								Column: int(45),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"floatChars",
																																			"keySetLc",
																																			"onlyChars",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1177),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(1177),
																																				Column: int(46),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																OpFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(14),
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"floatChars",
																																		"key",
																																		"keySetLc",
																																		"onlyChars",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1177),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(1178),
																																			Column: int(59),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(17),
																															},
																															OpFodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(14),
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"floatChars",
																																	"key",
																																	"keySetLc",
																																	"onlyChars",
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1177),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(1179),
																																		Column: int(58),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(17),
																														},
																														OpFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(14),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"floatChars",
																																"key",
																																"keyLc",
																																"keySetLc",
																																"onlyChars",
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1177),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1180),
																																	Column: int(60),
																																},
																															},
																														},
																														Op: ast.BinaryOp(17),
																													},
																													BranchTrue: &ast.LiteralBoolean{
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(8),
																																},
																															},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1181),
																																	Column: int(9),
																																},
																																End: ast.Location{
																																	Line: int(1181),
																																	Column: int(14),
																																},
																															},
																														},
																														Value: false,
																													},
																													BranchFalse: &ast.Conditional{
																														Cond: &ast.Binary{
																															Right: &ast.Apply{
																																Target: &ast.Var{
																																	Id: "typeMatch",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"typeMatch",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1192),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1192),
																																				Column: int(27),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				Id: "key",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12274,
																																					FreeVars: ast.Identifiers{
																																						"key",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1192),
																																							Column: int(28),
																																						},
																																						End: ast.Location{
																																							Line: int(1192),
																																							Column: int(31),
																																						},
																																					},
																																				},
																																			},
																																			CommaFodder: ast.Fodder{},
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.LiteralString{
																																				Value: "0x",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12274,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1192),
																																							Column: int(33),
																																						},
																																						End: ast.Location{
																																							Line: int(1192),
																																							Column: int(37),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"key",
																																		"typeMatch",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1192),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(1192),
																																			Column: int(38),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Left: &ast.Binary{
																																Right: &ast.Binary{
																																	Right: &ast.LiteralNumber{
																																		OriginalString: "2",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1191),
																																					Column: int(41),
																																				},
																																				End: ast.Location{
																																					Line: int(1191),
																																					Column: int(42),
																																				},
																																			},
																																		},
																																	},
																																	Left: &ast.Apply{
																																		Target: &ast.Index{
																																			Target: &ast.Var{
																																				Id: "std",
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1191),
																																							Column: int(18),
																																						},
																																						End: ast.Location{
																																							Line: int(1191),
																																							Column: int(21),
																																						},
																																					},
																																				},
																																			},
																																			Index: &ast.LiteralString{
																																				Value: "length",
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				Kind: ast.LiteralStringKind(1),
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			LeftBracketFodder: ast.Fodder{},
																																			Id: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12028,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1191),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(1191),
																																						Column: int(28),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "keyChars",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12289,
																																							FreeVars: ast.Identifiers{
																																								"keyChars",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1191),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(1191),
																																									Column: int(37),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"keyChars",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1191),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1191),
																																					Column: int(38),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	OpFodder: ast.Fodder{},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"keyChars",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1191),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(1191),
																																				Column: int(42),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(7),
																																},
																																Left: &ast.Binary{
																																	Right: &ast.Binary{
																																		Right: &ast.LiteralNumber{
																																			OriginalString: "2",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12028,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1190),
																																						Column: int(57),
																																					},
																																					End: ast.Location{
																																						Line: int(1190),
																																						Column: int(58),
																																					},
																																				},
																																			},
																																		},
																																		Left: &ast.Apply{
																																			Target: &ast.Index{
																																				Target: &ast.Var{
																																					Id: "std",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1190),
																																								Column: int(18),
																																							},
																																							End: ast.Location{
																																								Line: int(1190),
																																								Column: int(21),
																																							},
																																						},
																																					},
																																				},
																																				Index: &ast.LiteralString{
																																					Value: "length",
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					Kind: ast.LiteralStringKind(1),
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				LeftBracketFodder: ast.Fodder{},
																																				Id: nil,
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12028,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1190),
																																							Column: int(18),
																																						},
																																						End: ast.Location{
																																							Line: int(1190),
																																							Column: int(28),
																																						},
																																					},
																																				},
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							Target: &ast.Index{
																																								Target: &ast.Var{
																																									Id: "std",
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1190),
																																												Column: int(29),
																																											},
																																											End: ast.Location{
																																												Line: int(1190),
																																												Column: int(32),
																																											},
																																										},
																																									},
																																								},
																																								Index: &ast.LiteralString{
																																									Value: "findSubstr",
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																									NodeBase: ast.NodeBase{
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{},
																																										LocRange: ast.LocationRange{
																																											File: nil,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																										},
																																									},
																																									Kind: ast.LiteralStringKind(1),
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								LeftBracketFodder: ast.Fodder{},
																																								Id: nil,
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p12308,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1190),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(1190),
																																											Column: int(43),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralString{
																																											Value: "-",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p12312,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1190),
																																														Column: int(44),
																																													},
																																													End: ast.Location{
																																														Line: int(1190),
																																														Column: int(47),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											Id: "key",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p12312,
																																												FreeVars: ast.Identifiers{
																																													"key",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1190),
																																														Column: int(49),
																																													},
																																													End: ast.Location{
																																														Line: int(1190),
																																														Column: int(52),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p12308,
																																								FreeVars: ast.Identifiers{
																																									"key",
																																									"std",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1190),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(1190),
																																										Column: int(53),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12028,
																																				FreeVars: ast.Identifiers{
																																					"key",
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1190),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(1190),
																																						Column: int(54),
																																					},
																																				},
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																		},
																																		OpFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"key",
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1190),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(1190),
																																					Column: int(58),
																																				},
																																			},
																																		},
																																		Op: ast.BinaryOp(9),
																																	},
																																	Left: &ast.Apply{
																																		Target: &ast.Var{
																																			Id: "onlyChars",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12028,
																																				FreeVars: ast.Identifiers{
																																					"onlyChars",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1189),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(1189),
																																						Column: int(24),
																																					},
																																				},
																																			},
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "hexChars",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12323,
																																							FreeVars: ast.Identifiers{
																																								"hexChars",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1189),
																																									Column: int(25),
																																								},
																																								End: ast.Location{
																																									Line: int(1189),
																																									Column: int(33),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: ast.Fodder{},
																																				},
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						Id: "keySetLc",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p12323,
																																							FreeVars: ast.Identifiers{
																																								"keySetLc",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1189),
																																									Column: int(35),
																																								},
																																								End: ast.Location{
																																									Line: int(1189),
																																									Column: int(43),
																																								},
																																							},
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12028,
																																			FreeVars: ast.Identifiers{
																																				"hexChars",
																																				"keySetLc",
																																				"onlyChars",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1189),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(1189),
																																					Column: int(44),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																	},
																																	OpFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Comment: []string{},
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(14),
																																		},
																																	},
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12028,
																																		FreeVars: ast.Identifiers{
																																			"hexChars",
																																			"key",
																																			"keySetLc",
																																			"onlyChars",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1189),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(1190),
																																				Column: int(58),
																																			},
																																		},
																																	},
																																	Op: ast.BinaryOp(17),
																																},
																																OpFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(14),
																																	},
																																},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12028,
																																	FreeVars: ast.Identifiers{
																																		"hexChars",
																																		"key",
																																		"keyChars",
																																		"keySetLc",
																																		"onlyChars",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1189),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(1191),
																																			Column: int(42),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(17),
																															},
																															OpFodder: ast.Fodder{
																																ast.FodderElement{
																																	Comment: []string{},
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(14),
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{
																																	"hexChars",
																																	"key",
																																	"keyChars",
																																	"keySetLc",
																																	"onlyChars",
																																	"std",
																																	"typeMatch",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1189),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(1192),
																																		Column: int(38),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(17),
																														},
																														BranchTrue: &ast.LiteralBoolean{
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Comment: []string{},
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(8),
																																	},
																																},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1193),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(1193),
																																		Column: int(14),
																																	},
																																},
																															},
																															Value: false,
																														},
																														BranchFalse: &ast.LiteralBoolean{
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12028,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1195),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(1195),
																																		Column: int(16),
																																	},
																																},
																															},
																															Value: true,
																														},
																														ThenFodder: ast.Fodder{},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																															ast.FodderElement{
																																Comment: []string{
																																	"// All checks pass. Key is safe for emission without quotes.",
																																},
																																Kind: ast.FodderKind(2),
																																Blanks: int(0),
																																Indent: int(6),
																															},
																														},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12028,
																															FreeVars: ast.Identifiers{
																																"hexChars",
																																"key",
																																"keyChars",
																																"keySetLc",
																																"onlyChars",
																																"std",
																																"typeMatch",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1189),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(1195),
																																	Column: int(16),
																																},
																															},
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Comment: []string{},
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																														ast.FodderElement{
																															Comment: []string{
																																"/* Check for hexadecimals.  Keys that meet all of the following:",
																																"     - all characters match [0-9a-fx_\\-]",
																																"     - has at most 1 dash",
																																"     - has at least 3 characters",
																																"     - starts with (-)0x",
																																"   are considered hexadecimals.",
																																"*/",
																															},
																															Kind: ast.FodderKind(2),
																															Blanks: int(0),
																															Indent: int(6),
																														},
																													},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12028,
																														FreeVars: ast.Identifiers{
																															"floatChars",
																															"hexChars",
																															"key",
																															"keyChars",
																															"keyLc",
																															"keySetLc",
																															"onlyChars",
																															"std",
																															"typeMatch",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1177),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(1195),
																																Column: int(16),
																															},
																														},
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																													ast.FodderElement{
																														Comment: []string{
																															"/* Check for floats. Keys that meet all of the following:",
																															"     - all characters match [0-9e._\\-]",
																															"     - has at most a single period",
																															"     - has at most two dashes",
																															"     - has at most 1 'e'",
																															"   are considered floats.",
																															"*/",
																														},
																														Kind: ast.FodderKind(2),
																														Blanks: int(0),
																														Indent: int(6),
																													},
																												},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12028,
																													FreeVars: ast.Identifiers{
																														"binChars",
																														"floatChars",
																														"hexChars",
																														"key",
																														"keyChars",
																														"keyLc",
																														"keySetLc",
																														"onlyChars",
																														"std",
																														"typeMatch",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1166),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(1195),
																															Column: int(16),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																												ast.FodderElement{
																													Comment: []string{
																														"/* Check for binary integers.  Keys that meet all of the following:",
																														"     - all characters match [0-9b_\\-]",
																														"     - has at least 3 characters",
																														"     - starts with (-)0b",
																														"   are considered binary integers.",
																														"*/",
																													},
																													Kind: ast.FodderKind(2),
																													Blanks: int(0),
																													Indent: int(6),
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12028,
																												FreeVars: ast.Identifiers{
																													"binChars",
																													"floatChars",
																													"hexChars",
																													"intChars",
																													"key",
																													"keyChars",
																													"keyLc",
																													"keySetLc",
																													"onlyChars",
																													"std",
																													"typeMatch",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1157),
																														Column: int(12),
																													},
																													End: ast.Location{
																														Line: int(1195),
																														Column: int(16),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																											ast.FodderElement{
																												Comment: []string{
																													"/* Check for integers.  Keys that meet all of the following:",
																													"     - all characters match [0-9_\\-]",
																													"     - has at most 1 dash",
																													"   are considered integers.",
																													"*/",
																												},
																												Kind: ast.FodderKind(2),
																												Blanks: int(0),
																												Indent: int(6),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12028,
																											FreeVars: ast.Identifiers{
																												"binChars",
																												"dateChars",
																												"floatChars",
																												"hexChars",
																												"intChars",
																												"key",
																												"keyChars",
																												"keyLc",
																												"keySet",
																												"keySetLc",
																												"onlyChars",
																												"std",
																												"typeMatch",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1149),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(1195),
																													Column: int(16),
																												},
																											},
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																										ast.FodderElement{
																											Comment: []string{
																												"/* Check for timestamp values.  Since spaces and colons are already forbidden,",
																												"   all that could potentially pass is the standard date format (ex MM-DD-YYYY, YYYY-DD-MM, etc).",
																												"   This check is even more conservative: Keys that meet all of the following:",
																												"     - all characters match [0-9\\-]",
																												"     - has exactly 2 dashes",
																												"   are considered dates.",
																												"*/",
																											},
																											Kind: ast.FodderKind(2),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12028,
																										FreeVars: ast.Identifiers{
																											"binChars",
																											"dateChars",
																											"floatChars",
																											"hexChars",
																											"intChars",
																											"isReserved",
																											"key",
																											"keyChars",
																											"keyLc",
																											"keySet",
																											"keySetLc",
																											"onlyChars",
																											"std",
																											"typeMatch",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1140),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(1195),
																												Column: int(16),
																											},
																										},
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																									ast.FodderElement{
																										Comment: []string{
																											"// Check for reserved words",
																										},
																										Kind: ast.FodderKind(2),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																										ast.FodderElement{
																											Comment: []string{
																												"// Check for unsafe characters",
																											},
																											Kind: ast.FodderKind(2),
																											Blanks: int(0),
																											Indent: int(6),
																										},
																									},
																									Ctx: p12028,
																									FreeVars: ast.Identifiers{
																										"binChars",
																										"dateChars",
																										"floatChars",
																										"hexChars",
																										"intChars",
																										"isReserved",
																										"key",
																										"keyChars",
																										"keyLc",
																										"keySet",
																										"keySetLc",
																										"onlyChars",
																										"safeChars",
																										"std",
																										"typeMatch",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1137),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(1195),
																											Column: int(16),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																									},
																								},
																								Ctx: p12028,
																								FreeVars: ast.Identifiers{
																									"binChars",
																									"dateChars",
																									"floatChars",
																									"hexChars",
																									"intChars",
																									"isReserved",
																									"key",
																									"keyChars",
																									"keyLc",
																									"keySet",
																									"onlyChars",
																									"safeChars",
																									"std",
																									"typeMatch",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1135),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(1195),
																										Column: int(16),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																								},
																							},
																							Ctx: p12028,
																							FreeVars: ast.Identifiers{
																								"binChars",
																								"dateChars",
																								"floatChars",
																								"hexChars",
																								"intChars",
																								"isReserved",
																								"key",
																								"keyChars",
																								"keyLc",
																								"onlyChars",
																								"safeChars",
																								"std",
																								"typeMatch",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1134),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(1195),
																									Column: int(16),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																							},
																						},
																						Ctx: p12028,
																						FreeVars: ast.Identifiers{
																							"binChars",
																							"dateChars",
																							"floatChars",
																							"hexChars",
																							"intChars",
																							"isReserved",
																							"key",
																							"keyLc",
																							"onlyChars",
																							"safeChars",
																							"std",
																							"typeMatch",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1133),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(1195),
																								Column: int(16),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																						},
																					},
																					Ctx: p12028,
																					FreeVars: ast.Identifiers{
																						"binChars",
																						"dateChars",
																						"floatChars",
																						"hexChars",
																						"intChars",
																						"isReserved",
																						"key",
																						"onlyChars",
																						"safeChars",
																						"std",
																						"typeMatch",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1132),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(1195),
																							Column: int(16),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																					},
																				},
																				Ctx: p12028,
																				FreeVars: ast.Identifiers{
																					"binChars",
																					"dateChars",
																					"floatChars",
																					"hexChars",
																					"intChars",
																					"isReserved",
																					"key",
																					"letters",
																					"onlyChars",
																					"std",
																					"typeMatch",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1131),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(1195),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			Ctx: p12028,
																			FreeVars: ast.Identifiers{
																				"binChars",
																				"digits",
																				"floatChars",
																				"hexChars",
																				"intChars",
																				"isReserved",
																				"key",
																				"letters",
																				"onlyChars",
																				"std",
																				"typeMatch",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1130),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(1195),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																			},
																		},
																		Ctx: p12028,
																		FreeVars: ast.Identifiers{
																			"binChars",
																			"digits",
																			"hexChars",
																			"intChars",
																			"isReserved",
																			"key",
																			"letters",
																			"onlyChars",
																			"std",
																			"typeMatch",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1129),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(1195),
																				Column: int(16),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	Ctx: p12028,
																	FreeVars: ast.Identifiers{
																		"binChars",
																		"digits",
																		"intChars",
																		"isReserved",
																		"key",
																		"letters",
																		"onlyChars",
																		"std",
																		"typeMatch",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1128),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(1195),
																			Column: int(16),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																Ctx: p12028,
																FreeVars: ast.Identifiers{
																	"digits",
																	"intChars",
																	"isReserved",
																	"key",
																	"letters",
																	"onlyChars",
																	"std",
																	"typeMatch",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1127),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(1195),
																		Column: int(16),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															Ctx: p12028,
															FreeVars: ast.Identifiers{
																"digits",
																"isReserved",
																"key",
																"letters",
																"onlyChars",
																"std",
																"typeMatch",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1126),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(1195),
																	Column: int(16),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p12028,
														FreeVars: ast.Identifiers{
															"isReserved",
															"key",
															"letters",
															"onlyChars",
															"std",
															"typeMatch",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1125),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1195),
																Column: int(16),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
														ast.FodderElement{
															Comment: []string{
																"/*",
																"For a key to be considered safe to emit without quotes, the following must be true",
																"  - All characters must match [a-zA-Z0-9_/\\-]",
																"  - Not match the integer format defined in https://yaml.org/type/int.html",
																"  - Not match the float format defined in https://yaml.org/type/float.html",
																"  - Not match the timestamp format defined in https://yaml.org/type/timestamp.html",
																"  - Not match the boolean format defined in https://yaml.org/type/bool.html",
																"  - Not match the null format defined in https://yaml.org/type/null.html",
																"  - Not match (ignoring case) any reserved words which pass the above tests.",
																"    Reserved words are defined in isReserved() above.",
																"",
																"Since the remaining YAML types require characters outside the set chosen as valid",
																"for the elimination of quotes from the YAML output, the remaining types listed at",
																"https://yaml.org/type/ are by default always quoted.",
																"*/",
															},
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p12028,
													FreeVars: ast.Identifiers{
														"isReserved",
														"key",
														"onlyChars",
														"std",
														"typeMatch",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1124),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1195),
															Column: int(16),
														},
													},
												},
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "key",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1108),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1108),
															Column: int(23),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: p12386,
												FreeVars: ast.Identifiers{
													"isReserved",
													"onlyChars",
													"std",
													"typeMatch",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1108),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1195),
														Column: int(16),
													},
												},
											},
											TrailingComma: false,
										},
										EqFodder: nil,
										Variable: "bareSafe",
										CloseFodder: nil,
										Fun: nil,
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Body: &ast.Function{
												ParenLeftFodder: ast.Fodder{},
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Conditional{
													Cond: &ast.Apply{
														Target: &ast.Var{
															Id: "bareSafe",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p12394,
																FreeVars: ast.Identifiers{
																	"bareSafe",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1197),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(1197),
																		Column: int(18),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "key",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12398,
																			FreeVars: ast.Identifiers{
																				"key",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1197),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(1197),
																					Column: int(22),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p12394,
															FreeVars: ast.Identifiers{
																"bareSafe",
																"key",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1197),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(1197),
																	Column: int(23),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchTrue: &ast.Var{
														Id: "key",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p12394,
															FreeVars: ast.Identifiers{
																"key",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1197),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1197),
																	Column: int(32),
																},
															},
														},
													},
													BranchFalse: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1197),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(1197),
																			Column: int(41),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "escapeStringJson",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p12394,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1197),
																		Column: int(38),
																	},
																	End: ast.Location{
																		Line: int(1197),
																		Column: int(58),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "key",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12411,
																			FreeVars: ast.Identifiers{
																				"key",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1197),
																					Column: int(59),
																				},
																				End: ast.Location{
																					Line: int(1197),
																					Column: int(62),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p12394,
															FreeVars: ast.Identifiers{
																"key",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1197),
																	Column: int(38),
																},
																End: ast.Location{
																	Line: int(1197),
																	Column: int(63),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p12394,
														FreeVars: ast.Identifiers{
															"bareSafe",
															"key",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1197),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1197),
																Column: int(63),
															},
														},
													},
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "key",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1196),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(1196),
																Column: int(28),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: p12417,
													FreeVars: ast.Identifiers{
														"bareSafe",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1196),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1197),
															Column: int(63),
														},
													},
												},
												TrailingComma: false,
											},
											EqFodder: nil,
											Variable: "escapeKeyYaml",
											CloseFodder: nil,
											Fun: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Body: &ast.Function{
													ParenLeftFodder: ast.Fodder{},
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.LiteralBoolean{
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p12425,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1199),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1199),
																			Column: int(19),
																		},
																	},
																},
																Value: true,
															},
															Left: &ast.Var{
																Id: "v",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p12425,
																	FreeVars: ast.Identifiers{
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1199),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(1199),
																			Column: int(11),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p12425,
																FreeVars: ast.Identifiers{
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1199),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(1199),
																		Column: int(19),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														BranchTrue: &ast.LiteralString{
															Value: "true",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p12425,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1200),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1200),
																		Column: int(15),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														BranchFalse: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p12425,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1201),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1201),
																				Column: int(25),
																			},
																		},
																	},
																	Value: false,
																},
																Left: &ast.Var{
																	Id: "v",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p12425,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1201),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1201),
																				Column: int(16),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p12425,
																	FreeVars: ast.Identifiers{
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1201),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1201),
																			Column: int(25),
																		},
																	},
																},
																Op: ast.BinaryOp(12),
															},
															BranchTrue: &ast.LiteralString{
																Value: "false",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																		},
																	},
																	Ctx: p12425,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1202),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1202),
																			Column: int(16),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralNull{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12425,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1203),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1203),
																					Column: int(24),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12425,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1203),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1203),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p12425,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1203),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1203),
																				Column: int(24),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.LiteralString{
																	Value: "null",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																			},
																		},
																		Ctx: p12425,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1204),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1204),
																				Column: int(15),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																BranchFalse: &ast.Conditional{
																	Cond: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1205),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1205),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "isNumber",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1205),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1205),
																						Column: int(27),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12456,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1205),
																									Column: int(28),
																								},
																								End: ast.Location{
																									Line: int(1205),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12425,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1205),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1205),
																					Column: int(30),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	BranchTrue: &ast.Binary{
																		Right: &ast.Var{
																			Id: "v",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1206),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(1206),
																						Column: int(15),
																					},
																				},
																			},
																		},
																		Left: &ast.LiteralString{
																			Value: "",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1206),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1206),
																						Column: int(11),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12425,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1206),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(1206),
																					Column: int(15),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	BranchFalse: &ast.Conditional{
																		Cond: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1207),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(1207),
																								Column: int(18),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "isString",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p12425,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1207),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1207),
																							Column: int(27),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "v",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12474,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1207),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(1207),
																										Column: int(29),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1207),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1207),
																						Column: int(30),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		BranchTrue: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1208),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1208),
																											Column: int(24),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "length",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12484,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1208),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(1208),
																										Column: int(31),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12488,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1208),
																													Column: int(32),
																												},
																												End: ast.Location{
																													Line: int(1208),
																													Column: int(33),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12484,
																							FreeVars: ast.Identifiers{
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1208),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(1208),
																									Column: int(34),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "len",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1208),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1208),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			Body: &ast.Conditional{
																				Cond: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "0",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1209),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(1209),
																									Column: int(20),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "len",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"len",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1209),
																									Column: int(12),
																								},
																								End: ast.Location{
																									Line: int(1209),
																									Column: int(15),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"len",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1209),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(1209),
																								Column: int(20),
																							},
																						},
																					},
																					Op: ast.BinaryOp(12),
																				},
																				BranchTrue: &ast.LiteralString{
																					Value: "\"\"",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																							},
																						},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1210),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(1210),
																								Column: int(15),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				BranchFalse: &ast.Conditional{
																					Cond: &ast.Binary{
																						Right: &ast.LiteralString{
																							Value: "\n",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1211),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(1211),
																										Column: int(35),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1211),
																											Column: int(17),
																										},
																										End: ast.Location{
																											Line: int(1211),
																											Column: int(18),
																										},
																									},
																								},
																							},
																							Index: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12425,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1211),
																												Column: int(25),
																											},
																											End: ast.Location{
																												Line: int(1211),
																												Column: int(26),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "len",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12425,
																										FreeVars: ast.Identifiers{
																											"len",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1211),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(1211),
																												Column: int(22),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"len",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1211),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(1211),
																											Column: int(26),
																										},
																									},
																								},
																								Op: ast.BinaryOp(4),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"len",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1211),
																										Column: int(17),
																									},
																									End: ast.Location{
																										Line: int(1211),
																										Column: int(27),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"len",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1211),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(1211),
																									Column: int(35),
																								},
																							},
																						},
																						Op: ast.BinaryOp(12),
																					},
																					BranchTrue: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Body: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1212),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(1212),
																														Column: int(28),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "split",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12519,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1212),
																													Column: int(25),
																												},
																												End: ast.Location{
																													Line: int(1212),
																													Column: int(34),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12523,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1212),
																																Column: int(35),
																															},
																															End: ast.Location{
																																Line: int(1212),
																																Column: int(36),
																															},
																														},
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralString{
																													Value: "\n",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12523,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1212),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(1212),
																																Column: int(42),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12519,
																										FreeVars: ast.Identifiers{
																											"std",
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1212),
																												Column: int(25),
																											},
																											End: ast.Location{
																												Line: int(1212),
																												Column: int(43),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								EqFodder: ast.Fodder{},
																								Variable: "split",
																								CloseFodder: ast.Fodder{},
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1212),
																										Column: int(17),
																									},
																									End: ast.Location{
																										Line: int(1212),
																										Column: int(43),
																									},
																								},
																							},
																						},
																						Body: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1213),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(1213),
																												Column: int(14),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "join",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1213),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(1213),
																											Column: int(19),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.LiteralString{
																												Value: "  ",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12537,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1213),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(1213),
																															Column: int(41),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											Left: &ast.Binary{
																												Right: &ast.Var{
																													Id: "cindent",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12537,
																														FreeVars: ast.Identifiers{
																															"cindent",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1213),
																																Column: int(27),
																															},
																															End: ast.Location{
																																Line: int(1213),
																																Column: int(34),
																															},
																														},
																													},
																												},
																												Left: &ast.LiteralString{
																													Value: "\n",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12537,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1213),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1213),
																																Column: int(24),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12537,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1213),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1213),
																															Column: int(34),
																														},
																													},
																												},
																												Op: ast.BinaryOp(3),
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12537,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1213),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(1213),
																														Column: int(41),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											Right: &ast.Apply{
																												Target: &ast.Index{
																													Target: &ast.Var{
																														Id: "$std",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"$std",
																															},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																													},
																													Index: &ast.LiteralString{
																														Value: "slice",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													RightBracketFodder: nil,
																													LeftBracketFodder: nil,
																													Id: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"$std",
																														},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																												},
																												FodderLeft: nil,
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "split",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12537,
																																	FreeVars: ast.Identifiers{
																																		"split",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1213),
																																			Column: int(51),
																																		},
																																		End: ast.Location{
																																			Line: int(1213),
																																			Column: int(56),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.LiteralNumber{
																																OriginalString: "0",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12537,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1213),
																																			Column: int(57),
																																		},
																																		End: ast.Location{
																																			Line: int(1213),
																																			Column: int(58),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																Right: &ast.LiteralNumber{
																																	OriginalString: "1",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12537,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1213),
																																				Column: int(79),
																																			},
																																			End: ast.Location{
																																				Line: int(1213),
																																				Column: int(80),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.Apply{
																																	Target: &ast.Index{
																																		Target: &ast.Var{
																																			Id: "std",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1213),
																																						Column: int(59),
																																					},
																																					End: ast.Location{
																																						Line: int(1213),
																																						Column: int(62),
																																					},
																																				},
																																			},
																																		},
																																		Index: &ast.LiteralString{
																																			Value: "length",
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																			NodeBase: ast.NodeBase{
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{},
																																				LocRange: ast.LocationRange{
																																					File: nil,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																				},
																																			},
																																			Kind: ast.LiteralStringKind(1),
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		LeftBracketFodder: ast.Fodder{},
																																		Id: nil,
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12537,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1213),
																																					Column: int(59),
																																				},
																																				End: ast.Location{
																																					Line: int(1213),
																																					Column: int(69),
																																				},
																																			},
																																		},
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "split",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p12565,
																																						FreeVars: ast.Identifiers{
																																							"split",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1213),
																																								Column: int(70),
																																							},
																																							End: ast.Location{
																																								Line: int(1213),
																																								Column: int(75),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12537,
																																		FreeVars: ast.Identifiers{
																																			"split",
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1213),
																																				Column: int(59),
																																			},
																																			End: ast.Location{
																																				Line: int(1213),
																																				Column: int(76),
																																			},
																																		},
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12537,
																																	FreeVars: ast.Identifiers{
																																		"split",
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1213),
																																			Column: int(59),
																																		},
																																		End: ast.Location{
																																			Line: int(1213),
																																			Column: int(80),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(4),
																															},
																															CommaFodder: nil,
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.LiteralNull{
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: nil,
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																														"split",
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1213),
																															Column: int(51),
																														},
																														End: ast.Location{
																															Line: int(1213),
																															Column: int(81),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											Left: &ast.Array{
																												Elements: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															Value: "|",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12574,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1213),
																																		Column: int(44),
																																	},
																																	End: ast.Location{
																																		Line: int(1213),
																																		Column: int(47),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														CommaFodder: nil,
																													},
																												},
																												CloseFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12537,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1213),
																															Column: int(43),
																														},
																														End: ast.Location{
																															Line: int(1213),
																															Column: int(48),
																														},
																													},
																												},
																												TrailingComma: false,
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12537,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"split",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1213),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(1213),
																														Column: int(81),
																													},
																												},
																											},
																											Op: ast.BinaryOp(3),
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"cindent",
																									"split",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1213),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(1213),
																										Column: int(82),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																								},
																							},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"cindent",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1212),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1213),
																									Column: int(82),
																								},
																							},
																						},
																					},
																					BranchFalse: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1215),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(1215),
																											Column: int(14),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "escapeStringJson",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1215),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(1215),
																										Column: int(31),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12588,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1215),
																													Column: int(32),
																												},
																												End: ast.Location{
																													Line: int(1215),
																													Column: int(33),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1215),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1215),
																									Column: int(34),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"cindent",
																							"len",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1211),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(1215),
																								Column: int(34),
																							},
																						},
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p12425,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"cindent",
																						"len",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1209),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(1215),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"cindent",
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1208),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1215),
																						Column: int(34),
																					},
																				},
																			},
																		},
																		BranchFalse: &ast.Conditional{
																			Cond: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1216),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(1216),
																									Column: int(18),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "isFunction",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1216),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(1216),
																								Column: int(29),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "v",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12607,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1216),
																											Column: int(30),
																										},
																										End: ast.Location{
																											Line: int(1216),
																											Column: int(31),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p12425,
																					FreeVars: ast.Identifiers{
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1216),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1216),
																							Column: int(32),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			BranchTrue: &ast.Error{
																				Expr: &ast.Binary{
																					Right: &ast.Var{
																						Id: "path",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"path",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1217),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(1217),
																									Column: int(54),
																								},
																							},
																						},
																					},
																					Left: &ast.LiteralString{
																						Value: "Tried to manifest function at ",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1217),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(1217),
																									Column: int(47),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"path",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1217),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(1217),
																								Column: int(54),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p12425,
																					FreeVars: ast.Identifiers{
																						"path",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1217),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(1217),
																							Column: int(54),
																						},
																					},
																				},
																			},
																			BranchFalse: &ast.Conditional{
																				Cond: &ast.Apply{
																					Target: &ast.Index{
																						Target: &ast.Var{
																							Id: "std",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1218),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(1218),
																										Column: int(18),
																									},
																								},
																							},
																						},
																						Index: &ast.LiteralString{
																							Value: "isArray",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1218),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(1218),
																									Column: int(26),
																								},
																							},
																						},
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									Id: "v",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12627,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1218),
																												Column: int(27),
																											},
																											End: ast.Location{
																												Line: int(1218),
																												Column: int(28),
																											},
																										},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1218),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(1218),
																								Column: int(29),
																							},
																						},
																					},
																					TrailingComma: false,
																					TailStrict: false,
																				},
																				BranchTrue: &ast.Conditional{
																					Cond: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "0",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1219),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1219),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						Left: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1219),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(1219),
																												Column: int(15),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "length",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1219),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(1219),
																											Column: int(22),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "v",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12641,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1219),
																														Column: int(23),
																													},
																													End: ast.Location{
																														Line: int(1219),
																														Column: int(24),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1219),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(1219),
																										Column: int(25),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1219),
																									Column: int(12),
																								},
																								End: ast.Location{
																									Line: int(1219),
																									Column: int(30),
																								},
																							},
																						},
																						Op: ast.BinaryOp(12),
																					},
																					BranchTrue: &ast.LiteralString{
																						Value: "[]",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																								},
																							},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1220),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1220),
																									Column: int(15),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					BranchFalse: &ast.Local{
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: nil,
																								Body: &ast.Function{
																									ParenLeftFodder: ast.Fodder{},
																									ParenRightFodder: ast.Fodder{},
																									Body: &ast.Conditional{
																										Cond: &ast.Binary{
																											Right: &ast.Binary{
																												Right: &ast.LiteralNumber{
																													OriginalString: "0",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12654,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1223),
																																Column: int(58),
																															},
																															End: ast.Location{
																																Line: int(1223),
																																Column: int(59),
																															},
																														},
																													},
																												},
																												Left: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1223),
																																		Column: int(38),
																																	},
																																	End: ast.Location{
																																		Line: int(1223),
																																		Column: int(41),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "length",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12654,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1223),
																																	Column: int(38),
																																},
																																End: ast.Location{
																																	Line: int(1223),
																																	Column: int(48),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "value",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12663,
																																		FreeVars: ast.Identifiers{
																																			"value",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1223),
																																				Column: int(49),
																																			},
																																			End: ast.Location{
																																				Line: int(1223),
																																				Column: int(54),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12654,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1223),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(1223),
																																Column: int(55),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12654,
																													FreeVars: ast.Identifiers{
																														"std",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1223),
																															Column: int(38),
																														},
																														End: ast.Location{
																															Line: int(1223),
																															Column: int(59),
																														},
																													},
																												},
																												Op: ast.BinaryOp(7),
																											},
																											Left: &ast.Apply{
																												Target: &ast.Index{
																													Target: &ast.Var{
																														Id: "std",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1223),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(1223),
																																	Column: int(19),
																																},
																															},
																														},
																													},
																													Index: &ast.LiteralString{
																														Value: "isArray",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													RightBracketFodder: ast.Fodder{},
																													LeftBracketFodder: ast.Fodder{},
																													Id: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12654,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1223),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(1223),
																																Column: int(27),
																															},
																														},
																													},
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "value",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12675,
																																	FreeVars: ast.Identifiers{
																																		"value",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1223),
																																			Column: int(28),
																																		},
																																		End: ast.Location{
																																			Line: int(1223),
																																			Column: int(33),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12654,
																													FreeVars: ast.Identifiers{
																														"std",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1223),
																															Column: int(16),
																														},
																														End: ast.Location{
																															Line: int(1223),
																															Column: int(34),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											OpFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12654,
																												FreeVars: ast.Identifiers{
																													"std",
																													"value",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1223),
																														Column: int(16),
																													},
																													End: ast.Location{
																														Line: int(1223),
																														Column: int(59),
																													},
																												},
																											},
																											Op: ast.BinaryOp(17),
																										},
																										BranchTrue: &ast.DesugaredObject{
																											Asserts: ast.Nodes{},
																											Fields: ast.DesugaredObjectFields{
																												ast.DesugaredObjectField{
																													Name: &ast.LiteralString{
																														Value: "new_indent",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Body: &ast.Binary{
																														Right: &ast.LiteralString{
																															Value: "  ",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12684,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1230),
																																		Column: int(37),
																																	},
																																	End: ast.Location{
																																		Line: int(1230),
																																		Column: int(41),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Left: &ast.Var{
																															Id: "cindent",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12684,
																																FreeVars: ast.Identifiers{
																																	"cindent",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1230),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(1230),
																																		Column: int(34),
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12684,
																															FreeVars: ast.Identifiers{
																																"cindent",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1230),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(1230),
																																	Column: int(41),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1230),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1230),
																															Column: int(41),
																														},
																													},
																													Hide: ast.ObjectFieldHide(1),
																													PlusSuper: false,
																												},
																												ast.DesugaredObjectField{
																													Name: &ast.LiteralString{
																														Value: "space",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													Body: &ast.Binary{
																														Right: &ast.Index{
																															Target: &ast.Self{
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1231),
																																			Column: int(29),
																																		},
																																		End: ast.Location{
																																			Line: int(1231),
																																			Column: int(33),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "new_indent",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12684,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1231),
																																		Column: int(29),
																																	},
																																	End: ast.Location{
																																		Line: int(1231),
																																		Column: int(44),
																																	},
																																},
																															},
																														},
																														Left: &ast.LiteralString{
																															Value: "\n",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12684,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1231),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1231),
																																		Column: int(26),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12684,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1231),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(1231),
																																	Column: int(44),
																																},
																															},
																														},
																														Op: ast.BinaryOp(3),
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1231),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1231),
																															Column: int(44),
																														},
																													},
																													Hide: ast.ObjectFieldHide(1),
																													PlusSuper: false,
																												},
																											},
																											Locals: ast.LocalBinds{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12654,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1223),
																														Column: int(65),
																													},
																													End: ast.Location{
																														Line: int(1232),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										BranchFalse: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12654,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1232),
																																	Column: int(66),
																																},
																																End: ast.Location{
																																	Line: int(1232),
																																	Column: int(67),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1232),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(1232),
																																			Column: int(49),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "length",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12654,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1232),
																																		Column: int(46),
																																	},
																																	End: ast.Location{
																																		Line: int(1232),
																																		Column: int(56),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "value",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12707,
																																			FreeVars: ast.Identifiers{
																																				"value",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1232),
																																					Column: int(57),
																																				},
																																				End: ast.Location{
																																					Line: int(1232),
																																					Column: int(62),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12654,
																															FreeVars: ast.Identifiers{
																																"std",
																																"value",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1232),
																																	Column: int(46),
																																},
																																End: ast.Location{
																																	Line: int(1232),
																																	Column: int(63),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12654,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1232),
																																Column: int(46),
																															},
																															End: ast.Location{
																																Line: int(1232),
																																Column: int(67),
																															},
																														},
																													},
																													Op: ast.BinaryOp(7),
																												},
																												Left: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1232),
																																		Column: int(23),
																																	},
																																	End: ast.Location{
																																		Line: int(1232),
																																		Column: int(26),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "isObject",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12654,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1232),
																																	Column: int(23),
																																},
																																End: ast.Location{
																																	Line: int(1232),
																																	Column: int(35),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "value",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12719,
																																		FreeVars: ast.Identifiers{
																																			"value",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1232),
																																				Column: int(36),
																																			},
																																			End: ast.Location{
																																				Line: int(1232),
																																				Column: int(41),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12654,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1232),
																																Column: int(23),
																															},
																															End: ast.Location{
																																Line: int(1232),
																																Column: int(42),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12654,
																													FreeVars: ast.Identifiers{
																														"std",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1232),
																															Column: int(23),
																														},
																														End: ast.Location{
																															Line: int(1232),
																															Column: int(67),
																														},
																													},
																												},
																												Op: ast.BinaryOp(17),
																											},
																											BranchTrue: &ast.DesugaredObject{
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "new_indent",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Binary{
																															Right: &ast.LiteralString{
																																Value: "  ",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12728,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1233),
																																			Column: int(37),
																																		},
																																		End: ast.Location{
																																			Line: int(1233),
																																			Column: int(41),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Left: &ast.Var{
																																Id: "cindent",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12728,
																																	FreeVars: ast.Identifiers{
																																		"cindent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1233),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(1233),
																																			Column: int(34),
																																		},
																																	},
																																},
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12728,
																																FreeVars: ast.Identifiers{
																																	"cindent",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1233),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(1233),
																																		Column: int(41),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1233),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1233),
																																Column: int(41),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "space",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.LiteralString{
																															Value: " ",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12728,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1237),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1237),
																																		Column: int(25),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1237),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1237),
																																Column: int(25),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																												},
																												Locals: ast.LocalBinds{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12654,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1232),
																															Column: int(73),
																														},
																														End: ast.Location{
																															Line: int(1238),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											BranchFalse: &ast.DesugaredObject{
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "new_indent",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Var{
																															Id: "cindent",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12739,
																																FreeVars: ast.Identifiers{
																																	"cindent",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1240),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(1240),
																																		Column: int(34),
																																	},
																																},
																															},
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1240),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1240),
																																Column: int(34),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "space",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.LiteralString{
																															Value: " ",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12739,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1241),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1241),
																																		Column: int(25),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1241),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1241),
																																Column: int(25),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																												},
																												Locals: ast.LocalBinds{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12654,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1238),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1242),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12654,
																												FreeVars: ast.Identifiers{
																													"cindent",
																													"std",
																													"value",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1232),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(1242),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Comment: []string{},
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(12),
																												},
																											},
																											Ctx: p12654,
																											FreeVars: ast.Identifiers{
																												"cindent",
																												"std",
																												"value",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1223),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(1242),
																													Column: int(14),
																												},
																											},
																										},
																									},
																									Parameters: []ast.Parameter{
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "value",
																											CommaFodder: nil,
																											EqFodder: nil,
																											DefaultArg: nil,
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1222),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(1222),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: p12748,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1222),
																												Column: int(17),
																											},
																											End: ast.Location{
																												Line: int(1242),
																												Column: int(14),
																											},
																										},
																									},
																									TrailingComma: false,
																								},
																								EqFodder: nil,
																								Variable: "params",
																								CloseFodder: nil,
																								Fun: nil,
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																						},
																						Body: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1243),
																															Column: int(25),
																														},
																														End: ast.Location{
																															Line: int(1243),
																															Column: int(28),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "range",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p12757,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1243),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(1243),
																														Column: int(34),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12761,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1243),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(1243),
																																	Column: int(36),
																																},
																															},
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "1",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12761,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1243),
																																		Column: int(54),
																																	},
																																	End: ast.Location{
																																		Line: int(1243),
																																		Column: int(55),
																																	},
																																},
																															},
																														},
																														Left: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1243),
																																				Column: int(38),
																																			},
																																			End: ast.Location{
																																				Line: int(1243),
																																				Column: int(41),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "length",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12761,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1243),
																																			Column: int(38),
																																		},
																																		End: ast.Location{
																																			Line: int(1243),
																																			Column: int(48),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "v",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12772,
																																				FreeVars: ast.Identifiers{
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1243),
																																						Column: int(49),
																																					},
																																					End: ast.Location{
																																						Line: int(1243),
																																						Column: int(50),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12761,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1243),
																																		Column: int(38),
																																	},
																																	End: ast.Location{
																																		Line: int(1243),
																																		Column: int(51),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12761,
																															FreeVars: ast.Identifiers{
																																"std",
																																"v",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1243),
																																	Column: int(38),
																																},
																																End: ast.Location{
																																	Line: int(1243),
																																	Column: int(55),
																																},
																															},
																														},
																														Op: ast.BinaryOp(4),
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12757,
																											FreeVars: ast.Identifiers{
																												"std",
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1243),
																													Column: int(25),
																												},
																												End: ast.Location{
																													Line: int(1243),
																													Column: int(56),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									EqFodder: ast.Fodder{},
																									Variable: "range",
																									CloseFodder: ast.Fodder{},
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1243),
																											Column: int(17),
																										},
																										End: ast.Location{
																											Line: int(1243),
																											Column: int(56),
																										},
																									},
																								},
																							},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "$std",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"$std",
																														},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "flatMap",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: nil,
																												LeftBracketFodder: nil,
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															ParenLeftFodder: nil,
																															ParenRightFodder: nil,
																															Body: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "$std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"$std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "flatMap",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: nil,
																																	LeftBracketFodder: nil,
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: nil,
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Function{
																																				ParenLeftFodder: nil,
																																				ParenRightFodder: nil,
																																				Body: &ast.Array{
																																					Elements: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Binary{
																																								Right: &ast.Apply{
																																									Target: &ast.Var{
																																										Id: "aux",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p12800,
																																											FreeVars: ast.Identifiers{
																																												"aux",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1245),
																																													Column: int(33),
																																												},
																																												End: ast.Location{
																																													Line: int(1245),
																																													Column: int(36),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "v",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p12805,
																																															FreeVars: ast.Identifiers{
																																																"v",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(37),
																																																},
																																																End: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(38),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.Var{
																																														Id: "i",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p12805,
																																															FreeVars: ast.Identifiers{
																																																"i",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(39),
																																																},
																																																End: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(40),
																																																},
																																															},
																																														},
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p12805,
																																														FreeVars: ast.Identifiers{
																																															"i",
																																															"v",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1245),
																																																Column: int(37),
																																															},
																																															End: ast.Location{
																																																Line: int(1245),
																																																Column: int(41),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													Right: &ast.Array{
																																														Elements: []ast.CommaSeparatedExpr{
																																															ast.CommaSeparatedExpr{
																																																Expr: &ast.Var{
																																																	Id: "i",
																																																	NodeBase: ast.NodeBase{
																																																		Fodder: ast.Fodder{},
																																																		Ctx: p12814,
																																																		FreeVars: ast.Identifiers{
																																																			"i",
																																																		},
																																																		LocRange: ast.LocationRange{
																																																			File: p8,
																																																			FileName: "",
																																																			Begin: ast.Location{
																																																				Line: int(1245),
																																																				Column: int(51),
																																																			},
																																																			End: ast.Location{
																																																				Line: int(1245),
																																																				Column: int(52),
																																																			},
																																																		},
																																																	},
																																																},
																																																CommaFodder: nil,
																																															},
																																														},
																																														CloseFodder: ast.Fodder{},
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p12805,
																																															FreeVars: ast.Identifiers{
																																																"i",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(50),
																																																},
																																																End: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(53),
																																																},
																																															},
																																														},
																																														TrailingComma: false,
																																													},
																																													Left: &ast.Var{
																																														Id: "path",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p12805,
																																															FreeVars: ast.Identifiers{
																																																"path",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(43),
																																																},
																																																End: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(47),
																																																},
																																															},
																																														},
																																													},
																																													OpFodder: ast.Fodder{},
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p12805,
																																														FreeVars: ast.Identifiers{
																																															"i",
																																															"path",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1245),
																																																Column: int(43),
																																															},
																																															End: ast.Location{
																																																Line: int(1245),
																																																Column: int(53),
																																															},
																																														},
																																													},
																																													Op: ast.BinaryOp(3),
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "param",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"param",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(55),
																																																},
																																																End: ast.Location{
																																																	Line: int(1245),
																																																	Column: int(60),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "new_indent",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p12805,
																																														FreeVars: ast.Identifiers{
																																															"param",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1245),
																																																Column: int(55),
																																															},
																																															End: ast.Location{
																																																Line: int(1245),
																																																Column: int(71),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p12800,
																																										FreeVars: ast.Identifiers{
																																											"aux",
																																											"i",
																																											"param",
																																											"path",
																																											"v",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1245),
																																												Column: int(33),
																																											},
																																											End: ast.Location{
																																												Line: int(1245),
																																												Column: int(72),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "param",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"param",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1245),
																																														Column: int(19),
																																													},
																																													End: ast.Location{
																																														Line: int(1245),
																																														Column: int(24),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "space",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p12800,
																																											FreeVars: ast.Identifiers{
																																												"param",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1245),
																																													Column: int(19),
																																												},
																																												End: ast.Location{
																																													Line: int(1245),
																																													Column: int(30),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.LiteralString{
																																										Value: "-",
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Comment: []string{},
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(12),
																																												},
																																											},
																																											Ctx: p12800,
																																											FreeVars: ast.Identifiers{},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1245),
																																													Column: int(13),
																																												},
																																												End: ast.Location{
																																													Line: int(1245),
																																													Column: int(16),
																																												},
																																											},
																																										},
																																										Kind: ast.LiteralStringKind(1),
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p12800,
																																										FreeVars: ast.Identifiers{
																																											"param",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1245),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(1245),
																																												Column: int(30),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p12800,
																																									FreeVars: ast.Identifiers{
																																										"aux",
																																										"i",
																																										"param",
																																										"path",
																																										"v",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1245),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(1245),
																																											Column: int(72),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					CloseFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																							"i",
																																							"param",
																																							"path",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																				},
																																				Parameters: []ast.Parameter{
																																					ast.Parameter{
																																						NameFodder: nil,
																																						Name: "param",
																																						CommaFodder: nil,
																																						EqFodder: nil,
																																						DefaultArg: nil,
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"aux",
																																						"i",
																																						"path",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Array{
																																				Elements: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "params",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p12843,
																																									FreeVars: ast.Identifiers{
																																										"params",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1247),
																																											Column: int(27),
																																										},
																																										End: ast.Location{
																																											Line: int(1247),
																																											Column: int(33),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "v",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p12848,
																																													FreeVars: ast.Identifiers{
																																														"v",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1247),
																																															Column: int(34),
																																														},
																																														End: ast.Location{
																																															Line: int(1247),
																																															Column: int(35),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.Var{
																																												Id: "i",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p12848,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1247),
																																															Column: int(36),
																																														},
																																														End: ast.Location{
																																															Line: int(1247),
																																															Column: int(37),
																																														},
																																													},
																																												},
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p12848,
																																												FreeVars: ast.Identifiers{
																																													"i",
																																													"v",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1247),
																																														Column: int(34),
																																													},
																																													End: ast.Location{
																																														Line: int(1247),
																																														Column: int(38),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p12843,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																									"params",
																																									"v",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1247),
																																										Column: int(27),
																																									},
																																									End: ast.Location{
																																										Line: int(1247),
																																										Column: int(39),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				CloseFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p12854,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																						"params",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1247),
																																							Column: int(26),
																																						},
																																						End: ast.Location{
																																							Line: int(1247),
																																							Column: int(40),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: nil,
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																		"aux",
																																		"i",
																																		"params",
																																		"path",
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1244),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(1248),
																																			Column: int(12),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "i",
																																	CommaFodder: nil,
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																	"aux",
																																	"params",
																																	"path",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "range",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12854,
																																FreeVars: ast.Identifiers{
																																	"range",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1246),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1246),
																																		Column: int(27),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: nil,
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"aux",
																													"params",
																													"path",
																													"range",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1244),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(1248),
																														Column: int(12),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										EqFodder: ast.Fodder{},
																										Variable: "parts",
																										CloseFodder: ast.Fodder{},
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1244),
																												Column: int(17),
																											},
																											End: ast.Location{
																												Line: int(1248),
																												Column: int(12),
																											},
																										},
																									},
																								},
																								Body: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																													},
																												},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1249),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(1249),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "join",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12425,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1249),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(1249),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													Right: &ast.Var{
																														Id: "cindent",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12872,
																															FreeVars: ast.Identifiers{
																																"cindent",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1249),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(1249),
																																	Column: int(34),
																																},
																															},
																														},
																													},
																													Left: &ast.LiteralString{
																														Value: "\n",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12872,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1249),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(1249),
																																	Column: int(24),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12872,
																														FreeVars: ast.Identifiers{
																															"cindent",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1249),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1249),
																																Column: int(34),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "parts",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12872,
																														FreeVars: ast.Identifiers{
																															"parts",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1249),
																																Column: int(36),
																															},
																															End: ast.Location{
																																Line: int(1249),
																																Column: int(41),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12425,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"parts",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1249),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(1249),
																												Column: int(42),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"aux",
																										"cindent",
																										"params",
																										"path",
																										"range",
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1244),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(1249),
																											Column: int(42),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"aux",
																									"cindent",
																									"params",
																									"path",
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1243),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(1249),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																								},
																							},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"aux",
																								"cindent",
																								"path",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1222),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1249),
																									Column: int(42),
																								},
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"aux",
																							"cindent",
																							"path",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1219),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(1249),
																								Column: int(42),
																							},
																						},
																					},
																				},
																				BranchFalse: &ast.Conditional{
																					Cond: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1250),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(1250),
																											Column: int(18),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "isObject",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1250),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(1250),
																										Column: int(27),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										Id: "v",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12897,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1250),
																													Column: int(28),
																												},
																												End: ast.Location{
																													Line: int(1250),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1250),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(1250),
																									Column: int(30),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					BranchTrue: &ast.Conditional{
																						Cond: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "0",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1251),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1251),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							Left: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1251),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(1251),
																													Column: int(15),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "length",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12425,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1251),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(1251),
																												Column: int(22),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "v",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12911,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1251),
																															Column: int(23),
																														},
																														End: ast.Location{
																															Line: int(1251),
																															Column: int(24),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1251),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(1251),
																											Column: int(25),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1251),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(1251),
																										Column: int(30),
																									},
																								},
																							},
																							Op: ast.BinaryOp(12),
																						},
																						BranchTrue: &ast.LiteralString{
																							Value: "{}",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1252),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(1252),
																										Column: int(15),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						BranchFalse: &ast.Local{
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: nil,
																									Body: &ast.Function{
																										ParenLeftFodder: ast.Fodder{},
																										ParenRightFodder: ast.Fodder{},
																										Body: &ast.Conditional{
																											Cond: &ast.Binary{
																												Right: &ast.Binary{
																													Right: &ast.LiteralNumber{
																														OriginalString: "0",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12924,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1255),
																																	Column: int(58),
																																},
																																End: ast.Location{
																																	Line: int(1255),
																																	Column: int(59),
																																},
																															},
																														},
																													},
																													Left: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1255),
																																			Column: int(38),
																																		},
																																		End: ast.Location{
																																			Line: int(1255),
																																			Column: int(41),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "length",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12924,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1255),
																																		Column: int(38),
																																	},
																																	End: ast.Location{
																																		Line: int(1255),
																																		Column: int(48),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "value",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12933,
																																			FreeVars: ast.Identifiers{
																																				"value",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1255),
																																					Column: int(49),
																																				},
																																				End: ast.Location{
																																					Line: int(1255),
																																					Column: int(54),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12924,
																															FreeVars: ast.Identifiers{
																																"std",
																																"value",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1255),
																																	Column: int(38),
																																},
																																End: ast.Location{
																																	Line: int(1255),
																																	Column: int(55),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12924,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1255),
																																Column: int(38),
																															},
																															End: ast.Location{
																																Line: int(1255),
																																Column: int(59),
																															},
																														},
																													},
																													Op: ast.BinaryOp(7),
																												},
																												Left: &ast.Apply{
																													Target: &ast.Index{
																														Target: &ast.Var{
																															Id: "std",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1255),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(1255),
																																		Column: int(19),
																																	},
																																},
																															},
																														},
																														Index: &ast.LiteralString{
																															Value: "isArray",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														RightBracketFodder: ast.Fodder{},
																														LeftBracketFodder: ast.Fodder{},
																														Id: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12924,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1255),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(1255),
																																	Column: int(27),
																																},
																															},
																														},
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	Id: "value",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12945,
																																		FreeVars: ast.Identifiers{
																																			"value",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1255),
																																				Column: int(28),
																																			},
																																			End: ast.Location{
																																				Line: int(1255),
																																				Column: int(33),
																																			},
																																		},
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12924,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1255),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(1255),
																																Column: int(34),
																															},
																														},
																													},
																													TrailingComma: false,
																													TailStrict: false,
																												},
																												OpFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12924,
																													FreeVars: ast.Identifiers{
																														"std",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1255),
																															Column: int(16),
																														},
																														End: ast.Location{
																															Line: int(1255),
																															Column: int(59),
																														},
																													},
																												},
																												Op: ast.BinaryOp(17),
																											},
																											BranchTrue: &ast.DesugaredObject{
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "new_indent",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Conditional{
																															Cond: &ast.Var{
																																Id: "indent_array_in_object",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12954,
																																	FreeVars: ast.Identifiers{
																																		"indent_array_in_object",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1262),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(1262),
																																			Column: int(52),
																																		},
																																	},
																																},
																															},
																															BranchTrue: &ast.Binary{
																																Right: &ast.LiteralString{
																																	Value: "  ",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12954,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1262),
																																				Column: int(68),
																																			},
																																			End: ast.Location{
																																				Line: int(1262),
																																				Column: int(72),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																Left: &ast.Var{
																																	Id: "cindent",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p12954,
																																		FreeVars: ast.Identifiers{
																																			"cindent",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1262),
																																				Column: int(58),
																																			},
																																			End: ast.Location{
																																				Line: int(1262),
																																				Column: int(65),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12954,
																																	FreeVars: ast.Identifiers{
																																		"cindent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1262),
																																			Column: int(58),
																																		},
																																		End: ast.Location{
																																			Line: int(1262),
																																			Column: int(72),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															BranchFalse: &ast.Var{
																																Id: "cindent",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12954,
																																	FreeVars: ast.Identifiers{
																																		"cindent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1262),
																																			Column: int(78),
																																		},
																																		End: ast.Location{
																																			Line: int(1262),
																																			Column: int(85),
																																		},
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															ElseFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12954,
																																FreeVars: ast.Identifiers{
																																	"cindent",
																																	"indent_array_in_object",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1262),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(1262),
																																		Column: int(85),
																																	},
																																},
																															},
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1262),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1262),
																																Column: int(85),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																													ast.DesugaredObjectField{
																														Name: &ast.LiteralString{
																															Value: "space",
																															BlockIndent: "",
																															BlockTermIndent: "",
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															Kind: ast.LiteralStringKind(1),
																														},
																														Body: &ast.Binary{
																															Right: &ast.Index{
																																Target: &ast.Self{
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1263),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(1263),
																																				Column: int(33),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "new_indent",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12954,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1263),
																																			Column: int(29),
																																		},
																																		End: ast.Location{
																																			Line: int(1263),
																																			Column: int(44),
																																		},
																																	},
																																},
																															},
																															Left: &ast.LiteralString{
																																Value: "\n",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12954,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1263),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(1263),
																																			Column: int(26),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															OpFodder: ast.Fodder{},
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12954,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1263),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1263),
																																		Column: int(44),
																																	},
																																},
																															},
																															Op: ast.BinaryOp(3),
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1263),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1263),
																																Column: int(44),
																															},
																														},
																														Hide: ast.ObjectFieldHide(1),
																														PlusSuper: false,
																													},
																												},
																												Locals: ast.LocalBinds{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12924,
																													FreeVars: ast.Identifiers{
																														"cindent",
																														"indent_array_in_object",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1255),
																															Column: int(65),
																														},
																														End: ast.Location{
																															Line: int(1264),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											BranchFalse: &ast.Conditional{
																												Cond: &ast.Binary{
																													Right: &ast.Binary{
																														Right: &ast.LiteralNumber{
																															OriginalString: "0",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12924,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1264),
																																		Column: int(66),
																																	},
																																	End: ast.Location{
																																		Line: int(1264),
																																		Column: int(67),
																																	},
																																},
																															},
																														},
																														Left: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1264),
																																				Column: int(46),
																																			},
																																			End: ast.Location{
																																				Line: int(1264),
																																				Column: int(49),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "length",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p12924,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1264),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(1264),
																																			Column: int(56),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "value",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p12983,
																																				FreeVars: ast.Identifiers{
																																					"value",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1264),
																																						Column: int(57),
																																					},
																																					End: ast.Location{
																																						Line: int(1264),
																																						Column: int(62),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12924,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"value",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1264),
																																		Column: int(46),
																																	},
																																	End: ast.Location{
																																		Line: int(1264),
																																		Column: int(63),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														OpFodder: ast.Fodder{},
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12924,
																															FreeVars: ast.Identifiers{
																																"std",
																																"value",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1264),
																																	Column: int(46),
																																},
																																End: ast.Location{
																																	Line: int(1264),
																																	Column: int(67),
																																},
																															},
																														},
																														Op: ast.BinaryOp(7),
																													},
																													Left: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "std",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1264),
																																			Column: int(23),
																																		},
																																		End: ast.Location{
																																			Line: int(1264),
																																			Column: int(26),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "isObject",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p12924,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1264),
																																		Column: int(23),
																																	},
																																	End: ast.Location{
																																		Line: int(1264),
																																		Column: int(35),
																																	},
																																},
																															},
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		Id: "value",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p12995,
																																			FreeVars: ast.Identifiers{
																																				"value",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1264),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(1264),
																																					Column: int(41),
																																				},
																																			},
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p12924,
																															FreeVars: ast.Identifiers{
																																"std",
																																"value",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1264),
																																	Column: int(23),
																																},
																																End: ast.Location{
																																	Line: int(1264),
																																	Column: int(42),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12924,
																														FreeVars: ast.Identifiers{
																															"std",
																															"value",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1264),
																																Column: int(23),
																															},
																															End: ast.Location{
																																Line: int(1264),
																																Column: int(67),
																															},
																														},
																													},
																													Op: ast.BinaryOp(17),
																												},
																												BranchTrue: &ast.DesugaredObject{
																													Asserts: ast.Nodes{},
																													Fields: ast.DesugaredObjectFields{
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "new_indent",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.Binary{
																																Right: &ast.LiteralString{
																																	Value: "  ",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p13004,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1265),
																																				Column: int(37),
																																			},
																																			End: ast.Location{
																																				Line: int(1265),
																																				Column: int(41),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																Left: &ast.Var{
																																	Id: "cindent",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p13004,
																																		FreeVars: ast.Identifiers{
																																			"cindent",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1265),
																																				Column: int(27),
																																			},
																																			End: ast.Location{
																																				Line: int(1265),
																																				Column: int(34),
																																			},
																																		},
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13004,
																																	FreeVars: ast.Identifiers{
																																		"cindent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1265),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(1265),
																																			Column: int(41),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1265),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1265),
																																	Column: int(41),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "space",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.Binary{
																																Right: &ast.Index{
																																	Target: &ast.Self{
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1266),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(1266),
																																					Column: int(33),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "new_indent",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	LeftBracketFodder: ast.Fodder{},
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p13004,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1266),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(1266),
																																				Column: int(44),
																																			},
																																		},
																																	},
																																},
																																Left: &ast.LiteralString{
																																	Value: "\n",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: p13004,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1266),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(1266),
																																				Column: int(26),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																OpFodder: ast.Fodder{},
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13004,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1266),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(1266),
																																			Column: int(44),
																																		},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1266),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1266),
																																	Column: int(44),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																													},
																													Locals: ast.LocalBinds{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12924,
																														FreeVars: ast.Identifiers{
																															"cindent",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1264),
																																Column: int(73),
																															},
																															End: ast.Location{
																																Line: int(1267),
																																Column: int(14),
																															},
																														},
																													},
																												},
																												BranchFalse: &ast.DesugaredObject{
																													Asserts: ast.Nodes{},
																													Fields: ast.DesugaredObjectFields{
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "new_indent",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.Var{
																																Id: "cindent",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13019,
																																	FreeVars: ast.Identifiers{
																																		"cindent",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1269),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(1269),
																																			Column: int(34),
																																		},
																																	},
																																},
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1269),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1269),
																																	Column: int(34),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																														ast.DesugaredObjectField{
																															Name: &ast.LiteralString{
																																Value: "space",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															Body: &ast.LiteralString{
																																Value: " ",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13019,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1270),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(1270),
																																			Column: int(25),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1270),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(1270),
																																	Column: int(25),
																																},
																															},
																															Hide: ast.ObjectFieldHide(1),
																															PlusSuper: false,
																														},
																													},
																													Locals: ast.LocalBinds{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p12924,
																														FreeVars: ast.Identifiers{
																															"cindent",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1267),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1271),
																																Column: int(14),
																															},
																														},
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												ElseFodder: ast.Fodder{},
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p12924,
																													FreeVars: ast.Identifiers{
																														"cindent",
																														"std",
																														"value",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1264),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1271),
																															Column: int(14),
																														},
																													},
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											ElseFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(12),
																													},
																												},
																												Ctx: p12924,
																												FreeVars: ast.Identifiers{
																													"cindent",
																													"indent_array_in_object",
																													"std",
																													"value",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1255),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1271),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "value",
																												CommaFodder: nil,
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1254),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(1254),
																														Column: int(29),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: p13028,
																											FreeVars: ast.Identifiers{
																												"cindent",
																												"indent_array_in_object",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1254),
																													Column: int(17),
																												},
																												End: ast.Location{
																													Line: int(1271),
																													Column: int(14),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									EqFodder: nil,
																									Variable: "params",
																									CloseFodder: nil,
																									Fun: nil,
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							Body: &ast.Local{
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "$std",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"$std",
																														},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "flatMap",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: nil,
																												LeftBracketFodder: nil,
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"$std",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															ParenLeftFodder: nil,
																															ParenRightFodder: nil,
																															Body: &ast.Apply{
																																Target: &ast.Index{
																																	Target: &ast.Var{
																																		Id: "$std",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"$std",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																	},
																																	Index: &ast.LiteralString{
																																		Value: "flatMap",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: nil,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	RightBracketFodder: nil,
																																	LeftBracketFodder: nil,
																																	Id: nil,
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"$std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																},
																																FodderLeft: nil,
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Function{
																																				ParenLeftFodder: nil,
																																				ParenRightFodder: nil,
																																				Body: &ast.Array{
																																					Elements: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Binary{
																																								Right: &ast.Apply{
																																									Target: &ast.Var{
																																										Id: "aux",
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p13053,
																																											FreeVars: ast.Identifiers{
																																												"aux",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1273),
																																													Column: int(102),
																																												},
																																												End: ast.Location{
																																													Line: int(1273),
																																													Column: int(105),
																																												},
																																											},
																																										},
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "v",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p13058,
																																															FreeVars: ast.Identifiers{
																																																"v",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(106),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(107),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.Var{
																																														Id: "k",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p13058,
																																															FreeVars: ast.Identifiers{
																																																"k",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(108),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(109),
																																																},
																																															},
																																														},
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p13058,
																																														FreeVars: ast.Identifiers{
																																															"k",
																																															"v",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1273),
																																																Column: int(106),
																																															},
																																															End: ast.Location{
																																																Line: int(1273),
																																																Column: int(110),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													Right: &ast.Array{
																																														Elements: []ast.CommaSeparatedExpr{
																																															ast.CommaSeparatedExpr{
																																																Expr: &ast.Var{
																																																	Id: "k",
																																																	NodeBase: ast.NodeBase{
																																																		Fodder: ast.Fodder{},
																																																		Ctx: p13067,
																																																		FreeVars: ast.Identifiers{
																																																			"k",
																																																		},
																																																		LocRange: ast.LocationRange{
																																																			File: p8,
																																																			FileName: "",
																																																			Begin: ast.Location{
																																																				Line: int(1273),
																																																				Column: int(120),
																																																			},
																																																			End: ast.Location{
																																																				Line: int(1273),
																																																				Column: int(121),
																																																			},
																																																		},
																																																	},
																																																},
																																																CommaFodder: nil,
																																															},
																																														},
																																														CloseFodder: ast.Fodder{},
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p13058,
																																															FreeVars: ast.Identifiers{
																																																"k",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(119),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(122),
																																																},
																																															},
																																														},
																																														TrailingComma: false,
																																													},
																																													Left: &ast.Var{
																																														Id: "path",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: p13058,
																																															FreeVars: ast.Identifiers{
																																																"path",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(112),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(116),
																																																},
																																															},
																																														},
																																													},
																																													OpFodder: ast.Fodder{},
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p13058,
																																														FreeVars: ast.Identifiers{
																																															"k",
																																															"path",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1273),
																																																Column: int(112),
																																															},
																																															End: ast.Location{
																																																Line: int(1273),
																																																Column: int(122),
																																															},
																																														},
																																													},
																																													Op: ast.BinaryOp(3),
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "param",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"param",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(124),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(129),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "new_indent",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p13058,
																																														FreeVars: ast.Identifiers{
																																															"param",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1273),
																																																Column: int(124),
																																															},
																																															End: ast.Location{
																																																Line: int(1273),
																																																Column: int(140),
																																															},
																																														},
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p13053,
																																										FreeVars: ast.Identifiers{
																																											"aux",
																																											"k",
																																											"param",
																																											"path",
																																											"v",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1273),
																																												Column: int(102),
																																											},
																																											End: ast.Location{
																																												Line: int(1273),
																																												Column: int(141),
																																											},
																																										},
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																								},
																																								Left: &ast.Binary{
																																									Right: &ast.Index{
																																										Target: &ast.Var{
																																											Id: "param",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"param",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1273),
																																														Column: int(88),
																																													},
																																													End: ast.Location{
																																														Line: int(1273),
																																														Column: int(93),
																																													},
																																												},
																																											},
																																										},
																																										Index: &ast.LiteralString{
																																											Value: "space",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: nil,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										LeftBracketFodder: ast.Fodder{},
																																										Id: nil,
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p13053,
																																											FreeVars: ast.Identifiers{
																																												"param",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1273),
																																													Column: int(88),
																																												},
																																												End: ast.Location{
																																													Line: int(1273),
																																													Column: int(99),
																																												},
																																											},
																																										},
																																									},
																																									Left: &ast.Binary{
																																										Right: &ast.LiteralString{
																																											Value: ":",
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p13053,
																																												FreeVars: ast.Identifiers{},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1273),
																																														Column: int(82),
																																													},
																																													End: ast.Location{
																																														Line: int(1273),
																																														Column: int(85),
																																													},
																																												},
																																											},
																																											Kind: ast.LiteralStringKind(1),
																																										},
																																										Left: &ast.Conditional{
																																											Cond: &ast.Var{
																																												Id: "quote_keys",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p13053,
																																													FreeVars: ast.Identifiers{
																																														"quote_keys",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1273),
																																															Column: int(17),
																																														},
																																														End: ast.Location{
																																															Line: int(1273),
																																															Column: int(27),
																																														},
																																													},
																																												},
																																											},
																																											BranchTrue: &ast.Apply{
																																												Target: &ast.Index{
																																													Target: &ast.Var{
																																														Id: "std",
																																														NodeBase: ast.NodeBase{
																																															Fodder: ast.Fodder{},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"std",
																																															},
																																															LocRange: ast.LocationRange{
																																																File: p8,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(33),
																																																},
																																																End: ast.Location{
																																																	Line: int(1273),
																																																	Column: int(36),
																																																},
																																															},
																																														},
																																													},
																																													Index: &ast.LiteralString{
																																														Value: "escapeStringJson",
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																														NodeBase: ast.NodeBase{
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{},
																																															LocRange: ast.LocationRange{
																																																File: nil,
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																															},
																																														},
																																														Kind: ast.LiteralStringKind(1),
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													LeftBracketFodder: ast.Fodder{},
																																													Id: nil,
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p13053,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1273),
																																																Column: int(33),
																																															},
																																															End: ast.Location{
																																																Line: int(1273),
																																																Column: int(53),
																																															},
																																														},
																																													},
																																												},
																																												FodderLeft: ast.Fodder{},
																																												Arguments: ast.Arguments{
																																													Positional: []ast.CommaSeparatedExpr{
																																														ast.CommaSeparatedExpr{
																																															Expr: &ast.Var{
																																																Id: "k",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p13098,
																																																	FreeVars: ast.Identifiers{
																																																		"k",
																																																	},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(1273),
																																																			Column: int(54),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(1273),
																																																			Column: int(55),
																																																		},
																																																	},
																																																},
																																															},
																																															CommaFodder: nil,
																																														},
																																													},
																																													Named: nil,
																																												},
																																												FodderRight: ast.Fodder{},
																																												TailStrictFodder: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p13053,
																																													FreeVars: ast.Identifiers{
																																														"k",
																																														"std",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1273),
																																															Column: int(33),
																																														},
																																														End: ast.Location{
																																															Line: int(1273),
																																															Column: int(56),
																																														},
																																													},
																																												},
																																												TrailingComma: false,
																																												TailStrict: false,
																																											},
																																											BranchFalse: &ast.Apply{
																																												Target: &ast.Var{
																																													Id: "escapeKeyYaml",
																																													NodeBase: ast.NodeBase{
																																														Fodder: ast.Fodder{},
																																														Ctx: p13053,
																																														FreeVars: ast.Identifiers{
																																															"escapeKeyYaml",
																																														},
																																														LocRange: ast.LocationRange{
																																															File: p8,
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(1273),
																																																Column: int(62),
																																															},
																																															End: ast.Location{
																																																Line: int(1273),
																																																Column: int(75),
																																															},
																																														},
																																													},
																																												},
																																												FodderLeft: ast.Fodder{},
																																												Arguments: ast.Arguments{
																																													Positional: []ast.CommaSeparatedExpr{
																																														ast.CommaSeparatedExpr{
																																															Expr: &ast.Var{
																																																Id: "k",
																																																NodeBase: ast.NodeBase{
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p13106,
																																																	FreeVars: ast.Identifiers{
																																																		"k",
																																																	},
																																																	LocRange: ast.LocationRange{
																																																		File: p8,
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(1273),
																																																			Column: int(76),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(1273),
																																																			Column: int(77),
																																																		},
																																																	},
																																																},
																																															},
																																															CommaFodder: nil,
																																														},
																																													},
																																													Named: nil,
																																												},
																																												FodderRight: ast.Fodder{},
																																												TailStrictFodder: nil,
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p13053,
																																													FreeVars: ast.Identifiers{
																																														"escapeKeyYaml",
																																														"k",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1273),
																																															Column: int(62),
																																														},
																																														End: ast.Location{
																																															Line: int(1273),
																																															Column: int(78),
																																														},
																																													},
																																												},
																																												TrailingComma: false,
																																												TailStrict: false,
																																											},
																																											ThenFodder: ast.Fodder{},
																																											ElseFodder: ast.Fodder{},
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p13053,
																																												FreeVars: ast.Identifiers{
																																													"escapeKeyYaml",
																																													"k",
																																													"quote_keys",
																																													"std",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1273),
																																														Column: int(14),
																																													},
																																													End: ast.Location{
																																														Line: int(1273),
																																														Column: int(78),
																																													},
																																												},
																																											},
																																										},
																																										OpFodder: ast.Fodder{},
																																										NodeBase: ast.NodeBase{
																																											Fodder: ast.Fodder{},
																																											Ctx: p13053,
																																											FreeVars: ast.Identifiers{
																																												"escapeKeyYaml",
																																												"k",
																																												"quote_keys",
																																												"std",
																																											},
																																											LocRange: ast.LocationRange{
																																												File: p8,
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(1273),
																																													Column: int(13),
																																												},
																																												End: ast.Location{
																																													Line: int(1273),
																																													Column: int(85),
																																												},
																																											},
																																										},
																																										Op: ast.BinaryOp(3),
																																									},
																																									OpFodder: ast.Fodder{},
																																									NodeBase: ast.NodeBase{
																																										Fodder: ast.Fodder{},
																																										Ctx: p13053,
																																										FreeVars: ast.Identifiers{
																																											"escapeKeyYaml",
																																											"k",
																																											"param",
																																											"quote_keys",
																																											"std",
																																										},
																																										LocRange: ast.LocationRange{
																																											File: p8,
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(1273),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(1273),
																																												Column: int(99),
																																											},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																								},
																																								OpFodder: ast.Fodder{},
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p13053,
																																									FreeVars: ast.Identifiers{
																																										"aux",
																																										"escapeKeyYaml",
																																										"k",
																																										"param",
																																										"path",
																																										"quote_keys",
																																										"std",
																																										"v",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1273),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(1273),
																																											Column: int(141),
																																										},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					CloseFodder: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																							"escapeKeyYaml",
																																							"k",
																																							"param",
																																							"path",
																																							"quote_keys",
																																							"std",
																																							"v",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																					TrailingComma: false,
																																				},
																																				Parameters: []ast.Parameter{
																																					ast.Parameter{
																																						NameFodder: nil,
																																						Name: "param",
																																						CommaFodder: nil,
																																						EqFodder: nil,
																																						DefaultArg: nil,
																																						LocRange: ast.LocationRange{
																																							File: nil,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																						},
																																					},
																																				},
																																				NodeBase: ast.NodeBase{
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"aux",
																																						"escapeKeyYaml",
																																						"k",
																																						"path",
																																						"quote_keys",
																																						"std",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: nil,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Array{
																																				Elements: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							Target: &ast.Var{
																																								Id: "params",
																																								NodeBase: ast.NodeBase{
																																									Fodder: ast.Fodder{},
																																									Ctx: p13120,
																																									FreeVars: ast.Identifiers{
																																										"params",
																																									},
																																									LocRange: ast.LocationRange{
																																										File: p8,
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(1275),
																																											Column: int(27),
																																										},
																																										End: ast.Location{
																																											Line: int(1275),
																																											Column: int(33),
																																										},
																																									},
																																								},
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											Target: &ast.Var{
																																												Id: "v",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p13125,
																																													FreeVars: ast.Identifiers{
																																														"v",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1275),
																																															Column: int(34),
																																														},
																																														End: ast.Location{
																																															Line: int(1275),
																																															Column: int(35),
																																														},
																																													},
																																												},
																																											},
																																											Index: &ast.Var{
																																												Id: "k",
																																												NodeBase: ast.NodeBase{
																																													Fodder: ast.Fodder{},
																																													Ctx: p13125,
																																													FreeVars: ast.Identifiers{
																																														"k",
																																													},
																																													LocRange: ast.LocationRange{
																																														File: p8,
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(1275),
																																															Column: int(36),
																																														},
																																														End: ast.Location{
																																															Line: int(1275),
																																															Column: int(37),
																																														},
																																													},
																																												},
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											LeftBracketFodder: ast.Fodder{},
																																											Id: nil,
																																											NodeBase: ast.NodeBase{
																																												Fodder: ast.Fodder{},
																																												Ctx: p13125,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																													"v",
																																												},
																																												LocRange: ast.LocationRange{
																																													File: p8,
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(1275),
																																														Column: int(34),
																																													},
																																													End: ast.Location{
																																														Line: int(1275),
																																														Column: int(38),
																																													},
																																												},
																																											},
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																							NodeBase: ast.NodeBase{
																																								Fodder: ast.Fodder{},
																																								Ctx: p13120,
																																								FreeVars: ast.Identifiers{
																																									"k",
																																									"params",
																																									"v",
																																								},
																																								LocRange: ast.LocationRange{
																																									File: p8,
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(1275),
																																										Column: int(27),
																																									},
																																									End: ast.Location{
																																										Line: int(1275),
																																										Column: int(39),
																																									},
																																								},
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				CloseFodder: ast.Fodder{},
																																				NodeBase: ast.NodeBase{
																																					Fodder: ast.Fodder{},
																																					Ctx: p13131,
																																					FreeVars: ast.Identifiers{
																																						"k",
																																						"params",
																																						"v",
																																					},
																																					LocRange: ast.LocationRange{
																																						File: p8,
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(1275),
																																							Column: int(26),
																																						},
																																						End: ast.Location{
																																							Line: int(1275),
																																							Column: int(40),
																																						},
																																					},
																																				},
																																				TrailingComma: false,
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																FodderRight: nil,
																																TailStrictFodder: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																		"aux",
																																		"escapeKeyYaml",
																																		"k",
																																		"params",
																																		"path",
																																		"quote_keys",
																																		"std",
																																		"v",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1272),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(1276),
																																			Column: int(12),
																																		},
																																	},
																																},
																																TrailingComma: false,
																																TailStrict: false,
																															},
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "k",
																																	CommaFodder: nil,
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																	"aux",
																																	"escapeKeyYaml",
																																	"params",
																																	"path",
																																	"quote_keys",
																																	"std",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																															TrailingComma: false,
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Apply{
																															Target: &ast.Index{
																																Target: &ast.Var{
																																	Id: "std",
																																	NodeBase: ast.NodeBase{
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																		LocRange: ast.LocationRange{
																																			File: p8,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(1274),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(1274),
																																				Column: int(25),
																																			},
																																		},
																																	},
																																},
																																Index: &ast.LiteralString{
																																	Value: "objectFields",
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																	NodeBase: ast.NodeBase{
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{},
																																		LocRange: ast.LocationRange{
																																			File: nil,
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																		},
																																	},
																																	Kind: ast.LiteralStringKind(1),
																																},
																																RightBracketFodder: ast.Fodder{},
																																LeftBracketFodder: ast.Fodder{},
																																Id: nil,
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13131,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1274),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(1274),
																																			Column: int(38),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "v",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p13144,
																																				FreeVars: ast.Identifiers{
																																					"v",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1274),
																																						Column: int(39),
																																					},
																																					End: ast.Location{
																																						Line: int(1274),
																																						Column: int(40),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p13131,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"v",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1274),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(1274),
																																		Column: int(41),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: nil,
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"aux",
																													"escapeKeyYaml",
																													"params",
																													"path",
																													"quote_keys",
																													"std",
																													"v",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1272),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(1276),
																														Column: int(12),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										EqFodder: ast.Fodder{},
																										Variable: "lines",
																										CloseFodder: ast.Fodder{},
																										Fun: nil,
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1272),
																												Column: int(17),
																											},
																											End: ast.Location{
																												Line: int(1276),
																												Column: int(12),
																											},
																										},
																									},
																								},
																								Body: &ast.Apply{
																									Target: &ast.Index{
																										Target: &ast.Var{
																											Id: "std",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																													},
																												},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1277),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(1277),
																														Column: int(14),
																													},
																												},
																											},
																										},
																										Index: &ast.LiteralString{
																											Value: "join",
																											BlockIndent: "",
																											BlockTermIndent: "",
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											Kind: ast.LiteralStringKind(1),
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p12425,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1277),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(1277),
																													Column: int(19),
																												},
																											},
																										},
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													Right: &ast.Var{
																														Id: "cindent",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p13158,
																															FreeVars: ast.Identifiers{
																																"cindent",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1277),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(1277),
																																	Column: int(34),
																																},
																															},
																														},
																													},
																													Left: &ast.LiteralString{
																														Value: "\n",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p13158,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1277),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(1277),
																																	Column: int(24),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													OpFodder: ast.Fodder{},
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p13158,
																														FreeVars: ast.Identifiers{
																															"cindent",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1277),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1277),
																																Column: int(34),
																															},
																														},
																													},
																													Op: ast.BinaryOp(3),
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													Id: "lines",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p13158,
																														FreeVars: ast.Identifiers{
																															"lines",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1277),
																																Column: int(36),
																															},
																															End: ast.Location{
																																Line: int(1277),
																																Column: int(41),
																															},
																														},
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p12425,
																										FreeVars: ast.Identifiers{
																											"cindent",
																											"lines",
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1277),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(1277),
																												Column: int(42),
																											},
																										},
																									},
																									TrailingComma: false,
																									TailStrict: false,
																								},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									Ctx: p12425,
																									FreeVars: ast.Identifiers{
																										"$std",
																										"aux",
																										"cindent",
																										"escapeKeyYaml",
																										"params",
																										"path",
																										"quote_keys",
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1272),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(1277),
																											Column: int(42),
																										},
																									},
																								},
																							},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Comment: []string{},
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																									},
																								},
																								Ctx: p12425,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"aux",
																									"cindent",
																									"escapeKeyYaml",
																									"indent_array_in_object",
																									"path",
																									"quote_keys",
																									"std",
																									"v",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1254),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(1277),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																								},
																							},
																							Ctx: p12425,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"aux",
																								"cindent",
																								"escapeKeyYaml",
																								"indent_array_in_object",
																								"path",
																								"quote_keys",
																								"std",
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1251),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(1277),
																									Column: int(42),
																								},
																							},
																						},
																					},
																					BranchFalse: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p12425,
																						FreeVars: ast.Identifiers{
																							"$std",
																							"aux",
																							"cindent",
																							"escapeKeyYaml",
																							"indent_array_in_object",
																							"path",
																							"quote_keys",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1250),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(1277),
																								Column: int(42),
																							},
																						},
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p12425,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"aux",
																						"cindent",
																						"escapeKeyYaml",
																						"indent_array_in_object",
																						"path",
																						"quote_keys",
																						"std",
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1218),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(1277),
																							Column: int(42),
																						},
																					},
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p12425,
																				FreeVars: ast.Identifiers{
																					"$std",
																					"aux",
																					"cindent",
																					"escapeKeyYaml",
																					"indent_array_in_object",
																					"path",
																					"quote_keys",
																					"std",
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1216),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(1277),
																						Column: int(42),
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p12425,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"aux",
																				"cindent",
																				"escapeKeyYaml",
																				"indent_array_in_object",
																				"path",
																				"quote_keys",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1207),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(1277),
																					Column: int(42),
																				},
																			},
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p12425,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"aux",
																			"cindent",
																			"escapeKeyYaml",
																			"indent_array_in_object",
																			"path",
																			"quote_keys",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1205),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(1277),
																				Column: int(42),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p12425,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"aux",
																		"cindent",
																		"escapeKeyYaml",
																		"indent_array_in_object",
																		"path",
																		"quote_keys",
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1203),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(1277),
																			Column: int(42),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p12425,
																FreeVars: ast.Identifiers{
																	"$std",
																	"aux",
																	"cindent",
																	"escapeKeyYaml",
																	"indent_array_in_object",
																	"path",
																	"quote_keys",
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1201),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1277),
																		Column: int(42),
																	},
																},
															},
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															Ctx: p12425,
															FreeVars: ast.Identifiers{
																"$std",
																"aux",
																"cindent",
																"escapeKeyYaml",
																"indent_array_in_object",
																"path",
																"quote_keys",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1199),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(1277),
																	Column: int(42),
																},
															},
														},
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "v",
															CommaFodder: ast.Fodder{},
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1198),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1198),
																	Column: int(16),
																},
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "path",
															CommaFodder: ast.Fodder{},
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1198),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1198),
																	Column: int(22),
																},
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "cindent",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1198),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1198),
																	Column: int(31),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: p13190,
														FreeVars: ast.Identifiers{
															"$std",
															"aux",
															"escapeKeyYaml",
															"indent_array_in_object",
															"quote_keys",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1198),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1277),
																Column: int(42),
															},
														},
													},
													TrailingComma: false,
												},
												EqFodder: nil,
												Variable: "aux",
												CloseFodder: nil,
												Fun: nil,
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Var{
												Id: "aux",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(4),
														},
													},
													Ctx: p13195,
													FreeVars: ast.Identifiers{
														"aux",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1278),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(1278),
															Column: int(8),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "value",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13199,
																FreeVars: ast.Identifiers{
																	"value",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1278),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1278),
																		Column: int(14),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Array{
															Elements: nil,
															CloseFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13199,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1278),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(1278),
																		Column: int(18),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13199,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1278),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1278),
																		Column: int(22),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13195,
												FreeVars: ast.Identifiers{
													"aux",
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1278),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(1278),
														Column: int(23),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
												},
											},
											Ctx: p13195,
											FreeVars: ast.Identifiers{
												"$std",
												"escapeKeyYaml",
												"indent_array_in_object",
												"quote_keys",
												"std",
												"value",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1198),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(1278),
													Column: int(23),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p13195,
										FreeVars: ast.Identifiers{
											"$std",
											"bareSafe",
											"indent_array_in_object",
											"quote_keys",
											"std",
											"value",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1196),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(1278),
												Column: int(23),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p13195,
									FreeVars: ast.Identifiers{
										"$std",
										"indent_array_in_object",
										"isReserved",
										"onlyChars",
										"quote_keys",
										"std",
										"typeMatch",
										"value",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1108),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1278),
											Column: int(23),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p13195,
								FreeVars: ast.Identifiers{
									"$std",
									"indent_array_in_object",
									"isReserved",
									"onlyChars",
									"quote_keys",
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1103),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1278),
										Column: int(23),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p13195,
							FreeVars: ast.Identifiers{
								"$std",
								"indent_array_in_object",
								"onlyChars",
								"quote_keys",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1074),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1278),
									Column: int(23),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p13195,
						FreeVars: ast.Identifiers{
							"$std",
							"indent_array_in_object",
							"quote_keys",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1070),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1278),
								Column: int(23),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1069),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1069),
								Column: int(24),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent_array_in_object",
						CommaFodder: ast.Fodder{},
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13195,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1069),
										Column: int(49),
									},
									End: ast.Location{
										Line: int(1069),
										Column: int(54),
									},
								},
							},
							Value: false,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1069),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(1069),
								Column: int(54),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "quote_keys",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13195,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1069),
										Column: int(67),
									},
									End: ast.Location{
										Line: int(1069),
										Column: int(71),
									},
								},
							},
							Value: true,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1069),
								Column: int(56),
							},
							End: ast.Location{
								Line: int(1069),
								Column: int(71),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1069),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1278),
					Column: int(23),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestYamlStream",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1281),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1281),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1281),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1281),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "value",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13233,
												FreeVars: ast.Identifiers{
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1281),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1281),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1281),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1281),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13229,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1281),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1281),
									Column: int(27),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1282),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(1282),
													Column: int(63),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13229,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1282),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(1282),
												Column: int(68),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "value",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13247,
													FreeVars: ast.Identifiers{
														"value",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1282),
															Column: int(69),
														},
														End: ast.Location{
															Line: int(1282),
															Column: int(74),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{
										"std",
										"value",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1282),
											Column: int(60),
										},
										End: ast.Location{
											Line: int(1282),
											Column: int(75),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "manifestYamlStream only takes arrays, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1282),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1282),
											Column: int(57),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1282),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1282),
										Column: int(75),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p13229,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1282),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1282),
									Column: int(75),
								},
							},
						},
					},
					BranchFalse: &ast.Binary{
						Right: &ast.Conditional{
							Cond: &ast.Var{
								Id: "c_document_end",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{
										"c_document_end",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1286),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(1286),
											Column: int(28),
										},
									},
								},
							},
							BranchTrue: &ast.LiteralString{
								Value: "\n...\n",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1286),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(1286),
											Column: int(43),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							BranchFalse: &ast.LiteralString{
								Value: "\n",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1286),
											Column: int(49),
										},
										End: ast.Location{
											Line: int(1286),
											Column: int(53),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{
									"c_document_end",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1286),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1286),
										Column: int(53),
									},
								},
							},
						},
						Left: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1284),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1284),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "join",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13229,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1284),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1284),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "\n---\n",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p13271,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1285),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1285),
															Column: int(18),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "$std",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "flatMap",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: nil,
													LeftBracketFodder: nil,
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																ParenLeftFodder: nil,
																ParenRightFodder: nil,
																Body: &ast.Array{
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				Target: &ast.Index{
																					Target: &ast.Var{
																						Id: "std",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1285),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(1285),
																									Column: int(24),
																								},
																							},
																						},
																					},
																					Index: &ast.LiteralString{
																						Value: "manifestYamlDoc",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p13287,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1285),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(1285),
																								Column: int(40),
																							},
																						},
																					},
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "e",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13291,
																									FreeVars: ast.Identifiers{
																										"e",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1285),
																											Column: int(41),
																										},
																										End: ast.Location{
																											Line: int(1285),
																											Column: int(42),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "indent_array_in_object",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13291,
																									FreeVars: ast.Identifiers{
																										"indent_array_in_object",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1285),
																											Column: int(44),
																										},
																										End: ast.Location{
																											Line: int(1285),
																											Column: int(66),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "quote_keys",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13291,
																									FreeVars: ast.Identifiers{
																										"quote_keys",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1285),
																											Column: int(68),
																										},
																										End: ast.Location{
																											Line: int(1285),
																											Column: int(78),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p13287,
																					FreeVars: ast.Identifiers{
																						"e",
																						"indent_array_in_object",
																						"quote_keys",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1285),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1285),
																							Column: int(79),
																						},
																					},
																				},
																				TrailingComma: false,
																				TailStrict: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	CloseFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"e",
																			"indent_array_in_object",
																			"quote_keys",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: nil,
																		Name: "e",
																		CommaFodder: nil,
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"indent_array_in_object",
																		"quote_keys",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																TrailingComma: false,
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "value",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13271,
																	FreeVars: ast.Identifiers{
																		"value",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1285),
																			Column: int(89),
																		},
																		End: ast.Location{
																			Line: int(1285),
																			Column: int(94),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: nil,
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
														"indent_array_in_object",
														"quote_keys",
														"std",
														"value",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1285),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1285),
															Column: int(95),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13229,
									FreeVars: ast.Identifiers{
										"$std",
										"indent_array_in_object",
										"quote_keys",
										"std",
										"value",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1284),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1286),
											Column: int(8),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "---\n",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p13229,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1284),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1284),
											Column: int(14),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{
									"$std",
									"indent_array_in_object",
									"quote_keys",
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1284),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1286),
										Column: int(8),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13229,
							FreeVars: ast.Identifiers{
								"$std",
								"c_document_end",
								"indent_array_in_object",
								"quote_keys",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1284),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1286),
									Column: int(53),
								},
							},
						},
						Op: ast.BinaryOp(3),
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p13229,
						FreeVars: ast.Identifiers{
							"$std",
							"c_document_end",
							"indent_array_in_object",
							"quote_keys",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1281),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1286),
								Column: int(53),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1280),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1280),
								Column: int(27),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent_array_in_object",
						CommaFodder: ast.Fodder{},
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1280),
										Column: int(52),
									},
									End: ast.Location{
										Line: int(1280),
										Column: int(57),
									},
								},
							},
							Value: false,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1280),
								Column: int(29),
							},
							End: ast.Location{
								Line: int(1280),
								Column: int(57),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c_document_end",
						CommaFodder: ast.Fodder{},
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1280),
										Column: int(74),
									},
									End: ast.Location{
										Line: int(1280),
										Column: int(78),
									},
								},
							},
							Value: true,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1280),
								Column: int(59),
							},
							End: ast.Location{
								Line: int(1280),
								Column: int(78),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "quote_keys",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13229,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1280),
										Column: int(91),
									},
									End: ast.Location{
										Line: int(1280),
										Column: int(95),
									},
								},
							},
							Value: true,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1280),
								Column: int(80),
							},
							End: ast.Location{
								Line: int(1280),
								Column: int(95),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1280),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1286),
					Column: int(53),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestPython",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1290),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1290),
											Column: int(11),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isObject",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13326,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1290),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1290),
										Column: int(20),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "v",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13330,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1290),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1290),
													Column: int(22),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13326,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1290),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1290),
									Column: int(23),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "$std",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "flatMap",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: nil,
										LeftBracketFodder: nil,
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													ParenLeftFodder: nil,
													ParenRightFodder: nil,
													Body: &ast.Array{
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "$std",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "mod",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: nil,
																		LeftBracketFodder: nil,
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					Value: "%s: %s",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: p13354,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1292),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(1292),
																								Column: int(17),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Array{
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1292),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(1292),
																													Column: int(24),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "escapeStringPython",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p13362,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1292),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(1292),
																												Column: int(43),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "k",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p13366,
																													FreeVars: ast.Identifiers{
																														"k",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1292),
																															Column: int(44),
																														},
																														End: ast.Location{
																															Line: int(1292),
																															Column: int(45),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13362,
																									FreeVars: ast.Identifiers{
																										"k",
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1292),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1292),
																											Column: int(46),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "std",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1292),
																													Column: int(48),
																												},
																												End: ast.Location{
																													Line: int(1292),
																													Column: int(51),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "manifestPython",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p13362,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1292),
																												Column: int(48),
																											},
																											End: ast.Location{
																												Line: int(1292),
																												Column: int(66),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "v",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p13378,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1292),
																																Column: int(67),
																															},
																															End: ast.Location{
																																Line: int(1292),
																																Column: int(68),
																															},
																														},
																													},
																												},
																												Index: &ast.Var{
																													Id: "k",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p13378,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1292),
																																Column: int(69),
																															},
																															End: ast.Location{
																																Line: int(1292),
																																Column: int(70),
																															},
																														},
																													},
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p13378,
																													FreeVars: ast.Identifiers{
																														"k",
																														"v",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1292),
																															Column: int(67),
																														},
																														End: ast.Location{
																															Line: int(1292),
																															Column: int(71),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13362,
																									FreeVars: ast.Identifiers{
																										"k",
																										"std",
																										"v",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1292),
																											Column: int(48),
																										},
																										End: ast.Location{
																											Line: int(1292),
																											Column: int(72),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							CommaFodder: nil,
																						},
																					},
																					CloseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p13354,
																						FreeVars: ast.Identifiers{
																							"k",
																							"std",
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1292),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1292),
																								Column: int(73),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: nil,
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"k",
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1292),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1292),
																				Column: int(73),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														CloseFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"k",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														TrailingComma: false,
													},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: nil,
															Name: "k",
															CommaFodder: nil,
															EqFodder: nil,
															DefaultArg: nil,
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1293),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1293),
																		Column: int(21),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "objectFields",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13394,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1293),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1293),
																	Column: int(34),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "v",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13398,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1293),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1293),
																				Column: int(36),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13394,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1293),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1293),
																Column: int(37),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: nil,
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1291),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(1294),
												Column: int(8),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "fields",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1291),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1294),
										Column: int(8),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "{%s}",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p13326,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1295),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1295),
														Column: int(13),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1295),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1295),
																			Column: int(20),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "join",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13418,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1295),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1295),
																		Column: int(25),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		Value: ", ",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13422,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1295),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(1295),
																					Column: int(30),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "fields",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13422,
																			FreeVars: ast.Identifiers{
																				"fields",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1295),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(1295),
																					Column: int(38),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13418,
															FreeVars: ast.Identifiers{
																"fields",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1295),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1295),
																	Column: int(39),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"fields",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1295),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1295),
														Column: int(40),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"fields",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1295),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1295),
										Column: int(40),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p13326,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1291),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1295),
									Column: int(40),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1296),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1296),
												Column: int(16),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13326,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1296),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1296),
											Column: int(24),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "v",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13439,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1296),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(1296),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13326,
								FreeVars: ast.Identifiers{
									"std",
									"v",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1296),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1296),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "[%s]",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p13326,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1297),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1297),
														Column: int(13),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1297),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1297),
																			Column: int(20),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "join",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13458,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1297),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1297),
																		Column: int(25),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		Value: ", ",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13462,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1297),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(1297),
																					Column: int(30),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "$std",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "flatMap",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: nil,
																			LeftBracketFodder: nil,
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Function{
																						ParenLeftFodder: nil,
																						ParenRightFodder: nil,
																						Body: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1297),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(1297),
																															Column: int(36),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "manifestPython",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p13478,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1297),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(1297),
																														Column: int(51),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "v2",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p13482,
																															FreeVars: ast.Identifiers{
																																"v2",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1297),
																																	Column: int(52),
																																},
																																End: ast.Location{
																																	Line: int(1297),
																																	Column: int(54),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p13478,
																											FreeVars: ast.Identifiers{
																												"std",
																												"v2",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1297),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(1297),
																													Column: int(55),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v2",
																								},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: nil,
																								Name: "v2",
																								CommaFodder: nil,
																								EqFodder: nil,
																								DefaultArg: nil,
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13462,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1297),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1297),
																									Column: int(67),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: nil,
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1297),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(1297),
																					Column: int(68),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13458,
															FreeVars: ast.Identifiers{
																"$std",
																"std",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1297),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1297),
																	Column: int(69),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"$std",
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1297),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1297),
														Column: int(70),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"v",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1297),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1297),
										Column: int(70),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1298),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1298),
													Column: int(16),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13326,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1298),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1298),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "v",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13503,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1298),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1298),
															Column: int(27),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13326,
									FreeVars: ast.Identifiers{
										"std",
										"v",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1298),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1298),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "mod",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "%s",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p13326,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1299),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1299),
															Column: int(11),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1299),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1299),
																				Column: int(18),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "escapeStringPython",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13522,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1299),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1299),
																			Column: int(37),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "v",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p13526,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1299),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(1299),
																						Column: int(39),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13522,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1299),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1299),
																		Column: int(40),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												CloseFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13326,
													FreeVars: ast.Identifiers{
														"std",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1299),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(1299),
															Column: int(41),
														},
													},
												},
												TrailingComma: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"std",
										"v",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1299),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1299),
											Column: int(41),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1300),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1300),
														Column: int(16),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isFunction",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13326,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1300),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1300),
													Column: int(27),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "v",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13540,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1300),
																Column: int(28),
															},
															End: ast.Location{
																Line: int(1300),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13326,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1300),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1300),
												Column: int(30),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Error{
									Expr: &ast.LiteralString{
										Value: "cannot manifest function",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13326,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1301),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1301),
													Column: int(39),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p13326,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1301),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1301),
												Column: int(39),
											},
										},
									},
								},
								BranchFalse: &ast.Conditional{
									Cond: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1302),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1302),
															Column: int(16),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isNumber",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1302),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1302),
														Column: int(25),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13555,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1302),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1302),
																	Column: int(27),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13326,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1302),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1302),
													Column: int(28),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchTrue: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1303),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1303),
															Column: int(10),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "toString",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1303),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1303),
														Column: int(19),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13567,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1303),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1303),
																	Column: int(21),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13326,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1303),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1303),
													Column: int(22),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13326,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1304),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(1304),
															Column: int(22),
														},
													},
												},
												Value: true,
											},
											Left: &ast.Var{
												Id: "v",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13326,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1304),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1304),
															Column: int(14),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1304),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1304),
														Column: int(22),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.LiteralString{
											Value: "True",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p13326,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1305),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1305),
														Column: int(13),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13326,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1306),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1306),
																Column: int(23),
															},
														},
													},
													Value: false,
												},
												Left: &ast.Var{
													Id: "v",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13326,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1306),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1306),
																Column: int(14),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13326,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1306),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1306),
															Column: int(23),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.LiteralString{
												Value: "False",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p13326,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1307),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1307),
															Column: int(14),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13326,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1308),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1308),
																	Column: int(22),
																},
															},
														},
													},
													Left: &ast.Var{
														Id: "v",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13326,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1308),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1308),
																	Column: int(14),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13326,
														FreeVars: ast.Identifiers{
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1308),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1308),
																Column: int(22),
															},
														},
													},
													Op: ast.BinaryOp(12),
												},
												BranchTrue: &ast.LiteralString{
													Value: "None",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p13326,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1309),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1309),
																Column: int(13),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												BranchFalse: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13326,
													FreeVars: ast.Identifiers{
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1308),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1309),
															Column: int(13),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13326,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1306),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1309),
														Column: int(13),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13326,
											FreeVars: ast.Identifiers{
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1304),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1309),
													Column: int(13),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13326,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1302),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1309),
												Column: int(13),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13326,
									FreeVars: ast.Identifiers{
										"std",
										"v",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1300),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(1309),
											Column: int(13),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13326,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"v",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1298),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1309),
										Column: int(13),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13326,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"v",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1296),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(1309),
									Column: int(13),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p13326,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"v",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1290),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1309),
								Column: int(13),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1289),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1289),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1289),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1309),
					Column: int(13),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestPythonVars",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "flatMap",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Function{
												ParenLeftFodder: nil,
												ParenRightFodder: nil,
												Body: &ast.Array{
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "$std",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "mod",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: nil,
																	LeftBracketFodder: nil,
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralString{
																				Value: "%s = %s",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p13635,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1312),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1312),
																							Column: int(28),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "k",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13639,
																								FreeVars: ast.Identifiers{
																									"k",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1312),
																										Column: int(32),
																									},
																									End: ast.Location{
																										Line: int(1312),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "std",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1312),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1312),
																												Column: int(38),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "manifestPython",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p13639,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1312),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1312),
																											Column: int(53),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Index{
																											Target: &ast.Var{
																												Id: "conf",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p13650,
																													FreeVars: ast.Identifiers{
																														"conf",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1312),
																															Column: int(54),
																														},
																														End: ast.Location{
																															Line: int(1312),
																															Column: int(58),
																														},
																													},
																												},
																											},
																											Index: &ast.Var{
																												Id: "k",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p13650,
																													FreeVars: ast.Identifiers{
																														"k",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1312),
																															Column: int(59),
																														},
																														End: ast.Location{
																															Line: int(1312),
																															Column: int(60),
																														},
																													},
																												},
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p13650,
																												FreeVars: ast.Identifiers{
																													"conf",
																													"k",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1312),
																														Column: int(54),
																													},
																													End: ast.Location{
																														Line: int(1312),
																														Column: int(61),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13639,
																								FreeVars: ast.Identifiers{
																									"conf",
																									"k",
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1312),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(1312),
																										Column: int(62),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p13635,
																					FreeVars: ast.Identifiers{
																						"conf",
																						"k",
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1312),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1312),
																							Column: int(63),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: nil,
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"conf",
																		"k",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1312),
																			Column: int(19),
																		},
																		End: ast.Location{
																			Line: int(1312),
																			Column: int(63),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													CloseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"conf",
															"k",
															"std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: nil,
														Name: "k",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
														"conf",
														"std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												TrailingComma: false,
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1312),
																	Column: int(73),
																},
																End: ast.Location{
																	Line: int(1312),
																	Column: int(76),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "objectFields",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p13666,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1312),
																Column: int(73),
															},
															End: ast.Location{
																Line: int(1312),
																Column: int(89),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "conf",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13670,
																	FreeVars: ast.Identifiers{
																		"conf",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1312),
																			Column: int(90),
																		},
																		End: ast.Location{
																			Line: int(1312),
																			Column: int(94),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13666,
													FreeVars: ast.Identifiers{
														"conf",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1312),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(1312),
															Column: int(95),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"conf",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1312),
											Column: int(18),
										},
										End: ast.Location{
											Line: int(1312),
											Column: int(96),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "vars",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1312),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1312),
									Column: int(96),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1313),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1313),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "join",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13680,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1313),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1313),
										Column: int(13),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										Value: "\n",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13684,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1313),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1313),
													Column: int(18),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										Right: &ast.Array{
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														Value: "",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13689,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1313),
																	Column: int(28),
																},
																End: ast.Location{
																	Line: int(1313),
																	Column: int(30),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													CommaFodder: nil,
												},
											},
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13684,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1313),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1313),
														Column: int(31),
													},
												},
											},
											TrailingComma: false,
										},
										Left: &ast.Var{
											Id: "vars",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13684,
												FreeVars: ast.Identifiers{
													"vars",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1313),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1313),
														Column: int(24),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13684,
											FreeVars: ast.Identifiers{
												"vars",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1313),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1313),
													Column: int(31),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13680,
							FreeVars: ast.Identifiers{
								"std",
								"vars",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1313),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1313),
									Column: int(32),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p13680,
						FreeVars: ast.Identifiers{
							"$std",
							"conf",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1312),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1313),
								Column: int(32),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "conf",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1311),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1311),
								Column: int(26),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1311),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1313),
					Column: int(32),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "manifestXmlJsonml",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1316),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1316),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p13707,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1316),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1316),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "value",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13711,
												FreeVars: ast.Identifiers{
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1316),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1316),
														Column: int(26),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13707,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1316),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1316),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p13707,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1316),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1316),
									Column: int(27),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "Expected a JSONML value (an array), got %s",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13707,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1317),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1317),
														Column: int(57),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1317),
																Column: int(60),
															},
															End: ast.Location{
																Line: int(1317),
																Column: int(63),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "type",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13707,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1317),
															Column: int(60),
														},
														End: ast.Location{
															Line: int(1317),
															Column: int(68),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "value",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13732,
																FreeVars: ast.Identifiers{
																	"value",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1317),
																		Column: int(69),
																	},
																	End: ast.Location{
																		Line: int(1317),
																		Column: int(74),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13707,
												FreeVars: ast.Identifiers{
													"std",
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1317),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(1317),
														Column: int(75),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1317),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1317),
										Column: int(75),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p13707,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1317),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1317),
									Column: int(75),
								},
							},
						},
					},
					BranchFalse: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1320),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1320),
																Column: int(15),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "isString",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p13747,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1320),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1320),
															Column: int(24),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "v",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13751,
																FreeVars: ast.Identifiers{
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1320),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1320),
																		Column: int(26),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13747,
												FreeVars: ast.Identifiers{
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1320),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1320),
														Column: int(27),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchTrue: &ast.Var{
											Id: "v",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p13747,
												FreeVars: ast.Identifiers{
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1321),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1321),
														Column: int(12),
													},
												},
											},
										},
										BranchFalse: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.Index{
														Target: &ast.Var{
															Id: "v",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13761,
																FreeVars: ast.Identifiers{
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1323),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1323),
																		Column: int(24),
																	},
																},
															},
														},
														Index: &ast.LiteralNumber{
															OriginalString: "0",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13761,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1323),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1323),
																		Column: int(26),
																	},
																},
															},
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p13761,
															FreeVars: ast.Identifiers{
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1323),
																	Column: int(23),
																},
																End: ast.Location{
																	Line: int(1323),
																	Column: int(27),
																},
															},
														},
													},
													EqFodder: ast.Fodder{},
													Variable: "tag",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1323),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1323),
															Column: int(27),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Binary{
															Right: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1324),
																					Column: int(50),
																				},
																				End: ast.Location{
																					Line: int(1324),
																					Column: int(53),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "isObject",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13773,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1324),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(1324),
																				Column: int(62),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Index{
																				Target: &ast.Var{
																					Id: "v",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p13778,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1324),
																								Column: int(63),
																							},
																							End: ast.Location{
																								Line: int(1324),
																								Column: int(64),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralNumber{
																					OriginalString: "1",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p13778,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1324),
																								Column: int(65),
																							},
																							End: ast.Location{
																								Line: int(1324),
																								Column: int(66),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p13778,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1324),
																							Column: int(63),
																						},
																						End: ast.Location{
																							Line: int(1324),
																							Column: int(67),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13773,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1324),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(1324),
																			Column: int(68),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															Left: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13773,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1324),
																				Column: int(45),
																			},
																			End: ast.Location{
																				Line: int(1324),
																				Column: int(46),
																			},
																		},
																	},
																},
																Left: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1324),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1324),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "length",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13773,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1324),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1324),
																					Column: int(39),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "v",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p13793,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1324),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(1324),
																								Column: int(41),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13773,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1324),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1324),
																				Column: int(42),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13773,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1324),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1324),
																			Column: int(46),
																		},
																	},
																},
																Op: ast.BinaryOp(7),
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p13773,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1324),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(1324),
																		Column: int(68),
																	},
																},
															},
															Op: ast.BinaryOp(17),
														},
														EqFodder: ast.Fodder{},
														Variable: "has_attrs",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1324),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1324),
																Column: int(68),
															},
														},
													},
												},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Var{
																	Id: "has_attrs",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13802,
																		FreeVars: ast.Identifiers{
																			"has_attrs",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1325),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1325),
																				Column: int(37),
																			},
																		},
																	},
																},
																BranchTrue: &ast.Index{
																	Target: &ast.Var{
																		Id: "v",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13802,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1325),
																					Column: int(43),
																				},
																				End: ast.Location{
																					Line: int(1325),
																					Column: int(44),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13802,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1325),
																					Column: int(45),
																				},
																				End: ast.Location{
																					Line: int(1325),
																					Column: int(46),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13802,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1325),
																				Column: int(43),
																			},
																			End: ast.Location{
																				Line: int(1325),
																				Column: int(47),
																			},
																		},
																	},
																},
																BranchFalse: &ast.DesugaredObject{
																	Asserts: ast.Nodes{},
																	Fields: ast.DesugaredObjectFields{},
																	Locals: ast.LocalBinds{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13802,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1325),
																				Column: int(53),
																			},
																			End: ast.Location{
																				Line: int(1325),
																				Column: int(55),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13802,
																	FreeVars: ast.Identifiers{
																		"has_attrs",
																		"v",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1325),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1325),
																			Column: int(55),
																		},
																	},
																},
															},
															EqFodder: ast.Fodder{},
															Variable: "attrs",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1325),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1325),
																	Column: int(55),
																},
															},
														},
													},
													Body: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	Cond: &ast.Var{
																		Id: "has_attrs",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13815,
																			FreeVars: ast.Identifiers{
																				"has_attrs",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1326),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1326),
																					Column: int(40),
																				},
																			},
																		},
																	},
																	BranchTrue: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "$std",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "slice",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: nil,
																			LeftBracketFodder: nil,
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13815,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1326),
																									Column: int(46),
																								},
																								End: ast.Location{
																									Line: int(1326),
																									Column: int(47),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNumber{
																						OriginalString: "2",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13815,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1326),
																									Column: int(48),
																								},
																								End: ast.Location{
																									Line: int(1326),
																									Column: int(49),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: nil,
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1326),
																					Column: int(46),
																				},
																				End: ast.Location{
																					Line: int(1326),
																					Column: int(51),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	BranchFalse: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "$std",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "slice",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: nil,
																			LeftBracketFodder: nil,
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "v",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13815,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1326),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(1326),
																									Column: int(58),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13815,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1326),
																									Column: int(59),
																								},
																								End: ast.Location{
																									Line: int(1326),
																									Column: int(60),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: nil,
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"v",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1326),
																					Column: int(57),
																				},
																				End: ast.Location{
																					Line: int(1326),
																					Column: int(62),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13815,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"has_attrs",
																			"v",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1326),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1326),
																				Column: int(62),
																			},
																		},
																	},
																},
																EqFodder: ast.Fodder{},
																Variable: "children",
																CloseFodder: ast.Fodder{},
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1326),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1326),
																		Column: int(62),
																	},
																},
															},
														},
														Body: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Body: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1328),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(1328),
																							Column: int(16),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "join",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p13852,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1328),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1328),
																						Column: int(21),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralString{
																						Value: "",
																						BlockIndent: "",
																						BlockTermIndent: "",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p13856,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1328),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(1328),
																									Column: int(24),
																								},
																							},
																						},
																						Kind: ast.LiteralStringKind(1),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "$std",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"$std",
																									},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "flatMap",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: nil,
																							LeftBracketFodder: nil,
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"$std",
																								},
																								LocRange: ast.LocationRange{
																									File: nil,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																								},
																							},
																						},
																						FodderLeft: nil,
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Function{
																										ParenLeftFodder: nil,
																										ParenRightFodder: nil,
																										Body: &ast.Array{
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														Target: &ast.Index{
																															Target: &ast.Var{
																																Id: "$std",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"$std",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																															},
																															Index: &ast.LiteralString{
																																Value: "mod",
																																BlockIndent: "",
																																BlockTermIndent: "",
																																NodeBase: ast.NodeBase{
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{},
																																	LocRange: ast.LocationRange{
																																		File: nil,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																	},
																																},
																																Kind: ast.LiteralStringKind(1),
																															},
																															RightBracketFodder: nil,
																															LeftBracketFodder: nil,
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"$std",
																																},
																																LocRange: ast.LocationRange{
																																	File: nil,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																},
																															},
																														},
																														FodderLeft: nil,
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralString{
																																		Value: " %s=\"%s\"",
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p13875,
																																			FreeVars: ast.Identifiers{},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1328),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(1328),
																																					Column: int(37),
																																				},
																																			},
																																		},
																																		Kind: ast.LiteralStringKind(1),
																																	},
																																	CommaFodder: nil,
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Array{
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					Id: "k",
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p13879,
																																						FreeVars: ast.Identifiers{
																																							"k",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1328),
																																								Column: int(41),
																																							},
																																							End: ast.Location{
																																								Line: int(1328),
																																								Column: int(42),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					Target: &ast.Var{
																																						Id: "attrs",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p13879,
																																							FreeVars: ast.Identifiers{
																																								"attrs",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1328),
																																									Column: int(44),
																																								},
																																								End: ast.Location{
																																									Line: int(1328),
																																									Column: int(49),
																																								},
																																							},
																																						},
																																					},
																																					Index: &ast.Var{
																																						Id: "k",
																																						NodeBase: ast.NodeBase{
																																							Fodder: ast.Fodder{},
																																							Ctx: p13879,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																							},
																																							LocRange: ast.LocationRange{
																																								File: p8,
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(1328),
																																									Column: int(50),
																																								},
																																								End: ast.Location{
																																									Line: int(1328),
																																									Column: int(51),
																																								},
																																							},
																																						},
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					LeftBracketFodder: ast.Fodder{},
																																					Id: nil,
																																					NodeBase: ast.NodeBase{
																																						Fodder: ast.Fodder{},
																																						Ctx: p13879,
																																						FreeVars: ast.Identifiers{
																																							"attrs",
																																							"k",
																																						},
																																						LocRange: ast.LocationRange{
																																							File: p8,
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(1328),
																																								Column: int(44),
																																							},
																																							End: ast.Location{
																																								Line: int(1328),
																																								Column: int(52),
																																							},
																																						},
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		CloseFodder: ast.Fodder{},
																																		NodeBase: ast.NodeBase{
																																			Fodder: ast.Fodder{},
																																			Ctx: p13875,
																																			FreeVars: ast.Identifiers{
																																				"attrs",
																																				"k",
																																			},
																																			LocRange: ast.LocationRange{
																																				File: p8,
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(1328),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(1328),
																																					Column: int(53),
																																				},
																																			},
																																		},
																																		TrailingComma: false,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														FodderRight: nil,
																														TailStrictFodder: nil,
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"$std",
																																"attrs",
																																"k",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1328),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(1328),
																																	Column: int(53),
																																},
																															},
																														},
																														TrailingComma: false,
																														TailStrict: false,
																													},
																													CommaFodder: nil,
																												},
																											},
																											CloseFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"$std",
																													"attrs",
																													"k",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: nil,
																												Name: "k",
																												CommaFodder: nil,
																												EqFodder: nil,
																												DefaultArg: nil,
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$std",
																												"attrs",
																											},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										TrailingComma: false,
																									},
																									CommaFodder: nil,
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										Target: &ast.Index{
																											Target: &ast.Var{
																												Id: "std",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1328),
																															Column: int(63),
																														},
																														End: ast.Location{
																															Line: int(1328),
																															Column: int(66),
																														},
																													},
																												},
																											},
																											Index: &ast.LiteralString{
																												Value: "objectFields",
																												BlockIndent: "",
																												BlockTermIndent: "",
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												Kind: ast.LiteralStringKind(1),
																											},
																											RightBracketFodder: ast.Fodder{},
																											LeftBracketFodder: ast.Fodder{},
																											Id: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p13856,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1328),
																														Column: int(63),
																													},
																													End: ast.Location{
																														Line: int(1328),
																														Column: int(79),
																													},
																												},
																											},
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														Id: "attrs",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: p13900,
																															FreeVars: ast.Identifiers{
																																"attrs",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1328),
																																	Column: int(80),
																																},
																																End: ast.Location{
																																	Line: int(1328),
																																	Column: int(85),
																																},
																															},
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p13856,
																											FreeVars: ast.Identifiers{
																												"attrs",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1328),
																													Column: int(63),
																												},
																												End: ast.Location{
																													Line: int(1328),
																													Column: int(86),
																												},
																											},
																										},
																										TrailingComma: false,
																										TailStrict: false,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: nil,
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"$std",
																								"attrs",
																								"std",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1328),
																									Column: int(26),
																								},
																								End: ast.Location{
																									Line: int(1328),
																									Column: int(87),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p13852,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"attrs",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1328),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1328),
																					Column: int(88),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	EqFodder: ast.Fodder{},
																	Variable: "attrs_str",
																	CloseFodder: ast.Fodder{},
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1327),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1328),
																			Column: int(88),
																		},
																	},
																},
															},
															Body: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1329),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1329),
																					Column: int(14),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "deepJoin",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p13747,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1329),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1329),
																				Column: int(23),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "<",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(28),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "tag",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{
																									"tag",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(30),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "attrs_str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{
																									"attrs_str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(44),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: ">",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(46),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(49),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							Target: &ast.Index{
																								Target: &ast.Var{
																									Id: "$std",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"$std",
																										},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																								},
																								Index: &ast.LiteralString{
																									Value: "flatMap",
																									BlockIndent: "",
																									BlockTermIndent: "",
																									NodeBase: ast.NodeBase{
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: nil,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																										},
																									},
																									Kind: ast.LiteralStringKind(1),
																								},
																								RightBracketFodder: nil,
																								LeftBracketFodder: nil,
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"$std",
																									},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																							},
																							FodderLeft: nil,
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Function{
																											ParenLeftFodder: nil,
																											ParenRightFodder: nil,
																											Body: &ast.Array{
																												Elements: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Apply{
																															Target: &ast.Var{
																																Id: "aux",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p13934,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1329),
																																			Column: int(52),
																																		},
																																		End: ast.Location{
																																			Line: int(1329),
																																			Column: int(55),
																																		},
																																	},
																																},
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			Id: "x",
																																			NodeBase: ast.NodeBase{
																																				Fodder: ast.Fodder{},
																																				Ctx: p13938,
																																				FreeVars: ast.Identifiers{
																																					"x",
																																				},
																																				LocRange: ast.LocationRange{
																																					File: p8,
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(1329),
																																						Column: int(56),
																																					},
																																					End: ast.Location{
																																						Line: int(1329),
																																						Column: int(57),
																																					},
																																				},
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p13934,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"x",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1329),
																																		Column: int(52),
																																	},
																																	End: ast.Location{
																																		Line: int(1329),
																																		Column: int(58),
																																	},
																																},
																															},
																															TrailingComma: false,
																															TailStrict: false,
																														},
																														CommaFodder: nil,
																													},
																												},
																												CloseFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"x",
																													},
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																												TrailingComma: false,
																											},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: nil,
																													Name: "x",
																													CommaFodder: nil,
																													EqFodder: nil,
																													DefaultArg: nil,
																													LocRange: ast.LocationRange{
																														File: nil,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																													},
																												},
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"aux",
																												},
																												LocRange: ast.LocationRange{
																													File: nil,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																												},
																											},
																											TrailingComma: false,
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "children",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p13916,
																												FreeVars: ast.Identifiers{
																													"children",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1329),
																														Column: int(68),
																													},
																													End: ast.Location{
																														Line: int(1329),
																														Column: int(76),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: nil,
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"$std",
																									"aux",
																									"children",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(51),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(77),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: "</",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(79),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(83),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "tag",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{
																									"tag",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(85),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(88),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							Value: ">",
																							BlockIndent: "",
																							BlockTermIndent: "",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p13916,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1329),
																										Column: int(90),
																									},
																									End: ast.Location{
																										Line: int(1329),
																										Column: int(93),
																									},
																								},
																							},
																							Kind: ast.LiteralStringKind(1),
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p13951,
																					FreeVars: ast.Identifiers{
																						"$std",
																						"attrs_str",
																						"aux",
																						"children",
																						"tag",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1329),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1329),
																							Column: int(94),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p13747,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"attrs_str",
																		"aux",
																		"children",
																		"std",
																		"tag",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1329),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1329),
																			Column: int(95),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p13747,
																FreeVars: ast.Identifiers{
																	"$std",
																	"attrs",
																	"aux",
																	"children",
																	"std",
																	"tag",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1327),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1329),
																		Column: int(95),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																},
															},
															Ctx: p13747,
															FreeVars: ast.Identifiers{
																"$std",
																"attrs",
																"aux",
																"has_attrs",
																"std",
																"tag",
																"v",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1326),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1329),
																	Column: int(95),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p13747,
														FreeVars: ast.Identifiers{
															"$std",
															"aux",
															"has_attrs",
															"std",
															"tag",
															"v",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1325),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1329),
																Column: int(95),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
														},
													},
													Ctx: p13747,
													FreeVars: ast.Identifiers{
														"$std",
														"aux",
														"std",
														"tag",
														"v",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1324),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1329),
															Column: int(95),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p13747,
												FreeVars: ast.Identifiers{
													"$std",
													"aux",
													"std",
													"v",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1323),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1329),
														Column: int(95),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p13747,
											FreeVars: ast.Identifiers{
												"$std",
												"aux",
												"std",
												"v",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1320),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1329),
													Column: int(95),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "v",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1319),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1319),
													Column: int(18),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p13968,
										FreeVars: ast.Identifiers{
											"$std",
											"aux",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1319),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1329),
												Column: int(95),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "aux",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Var{
								Id: "aux",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(6),
										},
									},
									Ctx: p13707,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1331),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1331),
											Column: int(10),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "value",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p13976,
												FreeVars: ast.Identifiers{
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1331),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1331),
														Column: int(16),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p13707,
								FreeVars: ast.Identifiers{
									"aux",
									"value",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1331),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1331),
										Column: int(17),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p13707,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1319),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1331),
									Column: int(17),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p13707,
						FreeVars: ast.Identifiers{
							"$std",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1316),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1331),
								Column: int(17),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1315),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1315),
								Column: int(26),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1315),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1331),
					Column: int(17),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "base64",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1338),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1338),
														Column: int(13),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "isString",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13996,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1338),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1338),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "input",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14000,
														FreeVars: ast.Identifiers{
															"input",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1338),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1338),
																Column: int(28),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13996,
										FreeVars: ast.Identifiers{
											"input",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1338),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1338),
												Column: int(29),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1339),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1339),
														Column: int(12),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "map",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p13996,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1339),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1339),
													Column: int(16),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1339),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1339),
																	Column: int(20),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "codepoint",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14015,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1339),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1339),
																Column: int(30),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "input",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14015,
														FreeVars: ast.Identifiers{
															"input",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1339),
																Column: int(32),
															},
															End: ast.Location{
																Line: int(1339),
																Column: int(37),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p13996,
										FreeVars: ast.Identifiers{
											"input",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1339),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1339),
												Column: int(38),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchFalse: &ast.Var{
									Id: "input",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p13996,
										FreeVars: ast.Identifiers{
											"input",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1341),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1341),
												Column: int(14),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p13996,
									FreeVars: ast.Identifiers{
										"input",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1338),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1341),
											Column: int(14),
										},
									},
								},
							},
							EqFodder: ast.Fodder{},
							Variable: "bytes",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1337),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1341),
									Column: int(14),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1344),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1344),
																	Column: int(18),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1344),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1344),
																Column: int(25),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14040,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1344),
																			Column: int(26),
																		},
																		End: ast.Location{
																			Line: int(1344),
																			Column: int(29),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14036,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1344),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1344),
															Column: int(30),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "i",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14036,
													FreeVars: ast.Identifiers{
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1344),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1344),
															Column: int(11),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14036,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1344),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1344),
														Column: int(30),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										BranchTrue: &ast.Var{
											Id: "r",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p14036,
												FreeVars: ast.Identifiers{
													"r",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1345),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1345),
														Column: int(10),
													},
												},
											},
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1346),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1346),
																		Column: int(27),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1346),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1346),
																	Column: int(34),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14059,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1346),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1346),
																				Column: int(38),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"arr",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1346),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(1346),
																Column: int(39),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Binary{
													Right: &ast.LiteralNumber{
														OriginalString: "1",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1346),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1346),
																	Column: int(20),
																},
															},
														},
													},
													Left: &ast.Var{
														Id: "i",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1346),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1346),
																	Column: int(16),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"i",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1346),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1346),
																Column: int(20),
															},
														},
													},
													Op: ast.BinaryOp(3),
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14036,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1346),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1346),
															Column: int(39),
														},
													},
												},
												Op: ast.BinaryOp(8),
											},
											BranchTrue: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Binary{
															Right: &ast.LiteralString{
																Value: "==",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p14073,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1352),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1352),
																			Column: int(15),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Left: &ast.Binary{
																Right: &ast.Index{
																	Target: &ast.Var{
																		Id: "base64_table",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																				ast.FodderElement{
																					Comment: []string{
																						"// 2 LSB of i",
																					},
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p14073,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1351),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1351),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "4",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14073,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1351),
																						Column: int(40),
																					},
																					End: ast.Location{
																						Line: int(1351),
																						Column: int(41),
																					},
																				},
																			},
																		},
																		Left: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "3",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14073,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1351),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1351),
																							Column: int(35),
																						},
																					},
																				},
																			},
																			Left: &ast.Index{
																				Target: &ast.Var{
																					Id: "arr",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14073,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1351),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1351),
																								Column: int(28),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14073,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1351),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1351),
																								Column: int(30),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14073,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1351),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1351),
																							Column: int(31),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14073,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1351),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1351),
																						Column: int(35),
																					},
																				},
																			},
																			Op: ast.BinaryOp(14),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14073,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1351),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1351),
																					Column: int(41),
																				},
																			},
																		},
																		Op: ast.BinaryOp(5),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14073,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1351),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1351),
																				Column: int(42),
																			},
																		},
																	},
																},
																Left: &ast.Index{
																	Target: &ast.Var{
																		Id: "base64_table",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																				ast.FodderElement{
																					Comment: []string{
																						"// 6 MSB of i",
																					},
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p14073,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1349),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1349),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "2",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14073,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1349),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(1349),
																						Column: int(43),
																					},
																				},
																			},
																		},
																		Left: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "252",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14073,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1349),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1349),
																							Column: int(37),
																						},
																					},
																				},
																			},
																			Left: &ast.Index{
																				Target: &ast.Var{
																					Id: "arr",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14073,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1349),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1349),
																								Column: int(28),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14073,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1349),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1349),
																								Column: int(30),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14073,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1349),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1349),
																							Column: int(31),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14073,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1349),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1349),
																						Column: int(37),
																					},
																				},
																			},
																			Op: ast.BinaryOp(14),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14073,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1349),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1349),
																					Column: int(43),
																				},
																			},
																		},
																		Op: ast.BinaryOp(6),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14073,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1349),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1349),
																				Column: int(44),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14073,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1349),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1351),
																			Column: int(42),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14073,
																FreeVars: ast.Identifiers{
																	"arr",
																	"base64_table",
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1349),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1352),
																		Column: int(15),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														EqFodder: ast.Fodder{},
														Variable: "str",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1347),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1352),
																Column: int(15),
															},
														},
													},
												},
												Body: &ast.Apply{
													Target: &ast.Var{
														Id: "aux",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"aux",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1353),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1353),
																	Column: int(12),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14119,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1353),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1353),
																				Column: int(16),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "3",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14119,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1353),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1353),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14119,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1353),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1353),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14119,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1353),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1353),
																				Column: int(23),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14119,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1353),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1353),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "r",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14119,
																			FreeVars: ast.Identifiers{
																				"r",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1353),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1353),
																					Column: int(26),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14119,
																		FreeVars: ast.Identifiers{
																			"r",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1353),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1353),
																				Column: int(32),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"r",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1353),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1353),
																Column: int(33),
															},
														},
													},
													TrailingComma: false,
													TailStrict: true,
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p14036,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"base64_table",
														"i",
														"r",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1347),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1353),
															Column: int(33),
														},
													},
												},
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1354),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(1354),
																			Column: int(27),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "length",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14036,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1354),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1354),
																		Column: int(34),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "arr",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14145,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1354),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1354),
																					Column: int(38),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"arr",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1354),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1354),
																	Column: int(39),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "2",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14036,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1354),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1354),
																		Column: int(20),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "i",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14036,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1354),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1354),
																		Column: int(16),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1354),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1354),
																	Column: int(20),
																},
															},
														},
														Op: ast.BinaryOp(3),
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"arr",
															"i",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1354),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1354),
																Column: int(39),
															},
														},
													},
													Op: ast.BinaryOp(8),
												},
												BranchTrue: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Binary{
																Right: &ast.LiteralString{
																	Value: "=",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		Ctx: p14159,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1362),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1362),
																				Column: int(14),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																Left: &ast.Binary{
																	Right: &ast.Index{
																		Target: &ast.Var{
																			Id: "base64_table",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																					ast.FodderElement{
																						Comment: []string{
																							"// 4 LSB of i+1",
																						},
																						Kind: ast.FodderKind(2),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p14159,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1361),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1361),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		Index: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "2",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1361),
																							Column: int(45),
																						},
																						End: ast.Location{
																							Line: int(1361),
																							Column: int(46),
																						},
																					},
																				},
																			},
																			Left: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "15",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1361),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1361),
																								Column: int(40),
																							},
																						},
																					},
																				},
																				Left: &ast.Index{
																					Target: &ast.Var{
																						Id: "arr",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1361),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1361),
																									Column: int(28),
																								},
																							},
																						},
																					},
																					Index: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1361),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1361),
																										Column: int(34),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1361),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1361),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1361),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(1361),
																									Column: int(34),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1361),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1361),
																								Column: int(35),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1361),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1361),
																							Column: int(40),
																						},
																					},
																				},
																				Op: ast.BinaryOp(14),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14159,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1361),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1361),
																						Column: int(46),
																					},
																				},
																			},
																			Op: ast.BinaryOp(5),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14159,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1361),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1361),
																					Column: int(47),
																				},
																			},
																		},
																	},
																	Left: &ast.Binary{
																		Right: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_table",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																						ast.FodderElement{
																							Comment: []string{
																								"// 2 LSB of i, 4 MSB of i+1",
																							},
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1359),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1359),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.Binary{
																				Right: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "4",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1359),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1359),
																									Column: int(67),
																								},
																							},
																						},
																					},
																					Left: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "240",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1359),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1359),
																										Column: int(61),
																									},
																								},
																							},
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "arr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14159,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1359),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1359),
																											Column: int(48),
																										},
																									},
																								},
																							},
																							Index: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14159,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1359),
																												Column: int(53),
																											},
																											End: ast.Location{
																												Line: int(1359),
																												Column: int(54),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "i",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14159,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1359),
																												Column: int(49),
																											},
																											End: ast.Location{
																												Line: int(1359),
																												Column: int(50),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14159,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1359),
																											Column: int(49),
																										},
																										End: ast.Location{
																											Line: int(1359),
																											Column: int(54),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1359),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1359),
																										Column: int(55),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1359),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1359),
																									Column: int(61),
																								},
																							},
																						},
																						Op: ast.BinaryOp(14),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1359),
																								Column: int(44),
																							},
																							End: ast.Location{
																								Line: int(1359),
																								Column: int(67),
																							},
																						},
																					},
																					Op: ast.BinaryOp(6),
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "4",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1359),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1359),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					Left: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "3",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1359),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(1359),
																										Column: int(35),
																									},
																								},
																							},
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "arr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14159,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1359),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1359),
																											Column: int(28),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14159,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1359),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1359),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1359),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1359),
																										Column: int(31),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1359),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1359),
																									Column: int(35),
																								},
																							},
																						},
																						Op: ast.BinaryOp(14),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1359),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(1359),
																								Column: int(41),
																							},
																						},
																					},
																					Op: ast.BinaryOp(5),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1359),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1359),
																							Column: int(67),
																						},
																					},
																				},
																				Op: ast.BinaryOp(16),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14159,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1359),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1359),
																						Column: int(68),
																					},
																				},
																			},
																		},
																		Left: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_table",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																						ast.FodderElement{
																							Comment: []string{
																								"// 6 MSB of i",
																							},
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1357),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1357),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "2",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1357),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1357),
																								Column: int(43),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "252",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1357),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1357),
																									Column: int(37),
																								},
																							},
																						},
																					},
																					Left: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1357),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1357),
																										Column: int(28),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14159,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1357),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1357),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14159,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1357),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1357),
																									Column: int(31),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14159,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1357),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1357),
																								Column: int(37),
																							},
																						},
																					},
																					Op: ast.BinaryOp(14),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14159,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1357),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1357),
																							Column: int(43),
																						},
																					},
																				},
																				Op: ast.BinaryOp(6),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14159,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1357),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1357),
																						Column: int(44),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14159,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1357),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1359),
																					Column: int(68),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14159,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1357),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1361),
																				Column: int(47),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14159,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1357),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1362),
																			Column: int(14),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															EqFodder: ast.Fodder{},
															Variable: "str",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1355),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1362),
																	Column: int(14),
																},
															},
														},
													},
													Body: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p14036,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1363),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1363),
																		Column: int(12),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "arr",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14245,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1363),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1363),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "3",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14245,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1363),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1363),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14245,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1363),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1363),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14245,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1363),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1363),
																					Column: int(23),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14245,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1363),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1363),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "r",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14245,
																				FreeVars: ast.Identifiers{
																					"r",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1363),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1363),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14245,
																			FreeVars: ast.Identifiers{
																				"r",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1363),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1363),
																					Column: int(32),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"arr",
																"aux",
																"i",
																"r",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1363),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1363),
																	Column: int(33),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"base64_table",
															"i",
															"r",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1355),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1363),
																Column: int(33),
															},
														},
													},
												},
												BranchFalse: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Binary{
																Right: &ast.Index{
																	Target: &ast.Var{
																		Id: "base64_table",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																				ast.FodderElement{
																					Comment: []string{
																						"// 6 LSB of i+2",
																					},
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p14268,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1373),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1373),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "63",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1373),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(1373),
																						Column: int(40),
																					},
																				},
																			},
																		},
																		Left: &ast.Index{
																			Target: &ast.Var{
																				Id: "arr",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1373),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1373),
																							Column: int(28),
																						},
																					},
																				},
																			},
																			Index: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "2",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1373),
																								Column: int(33),
																							},
																							End: ast.Location{
																								Line: int(1373),
																								Column: int(34),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1373),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1373),
																								Column: int(30),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1373),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(1373),
																							Column: int(34),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1373),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1373),
																						Column: int(35),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14268,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1373),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1373),
																					Column: int(40),
																				},
																			},
																		},
																		Op: ast.BinaryOp(14),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14268,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1373),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1373),
																				Column: int(42),
																			},
																		},
																	},
																},
																Left: &ast.Binary{
																	Right: &ast.Index{
																		Target: &ast.Var{
																			Id: "base64_table",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																					ast.FodderElement{
																						Comment: []string{
																							"// 4 LSB of i+1, 2 MSB of i+2",
																						},
																						Kind: ast.FodderKind(2),
																						Blanks: int(0),
																						Indent: int(10),
																					},
																				},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1371),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1371),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		Index: &ast.Binary{
																			Right: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "6",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1371),
																								Column: int(71),
																							},
																							End: ast.Location{
																								Line: int(1371),
																								Column: int(72),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "192",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1371),
																									Column: int(63),
																								},
																								End: ast.Location{
																									Line: int(1371),
																									Column: int(66),
																								},
																							},
																						},
																					},
																					Left: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1371),
																										Column: int(50),
																									},
																									End: ast.Location{
																										Line: int(1371),
																										Column: int(53),
																									},
																								},
																							},
																						},
																						Index: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "2",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1371),
																											Column: int(58),
																										},
																										End: ast.Location{
																											Line: int(1371),
																											Column: int(59),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1371),
																											Column: int(54),
																										},
																										End: ast.Location{
																											Line: int(1371),
																											Column: int(55),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1371),
																										Column: int(54),
																									},
																									End: ast.Location{
																										Line: int(1371),
																										Column: int(59),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1371),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(1371),
																									Column: int(60),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1371),
																								Column: int(50),
																							},
																							End: ast.Location{
																								Line: int(1371),
																								Column: int(66),
																							},
																						},
																					},
																					Op: ast.BinaryOp(14),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1371),
																							Column: int(49),
																						},
																						End: ast.Location{
																							Line: int(1371),
																							Column: int(72),
																						},
																					},
																				},
																				Op: ast.BinaryOp(6),
																			},
																			Left: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "2",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1371),
																								Column: int(45),
																							},
																							End: ast.Location{
																								Line: int(1371),
																								Column: int(46),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "15",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1371),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1371),
																									Column: int(40),
																								},
																							},
																						},
																					},
																					Left: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1371),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1371),
																										Column: int(28),
																									},
																								},
																							},
																						},
																						Index: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1371),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(1371),
																											Column: int(34),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1371),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1371),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1371),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1371),
																										Column: int(34),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1371),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1371),
																									Column: int(35),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1371),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1371),
																								Column: int(40),
																							},
																						},
																					},
																					Op: ast.BinaryOp(14),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1371),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1371),
																							Column: int(46),
																						},
																					},
																				},
																				Op: ast.BinaryOp(5),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1371),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1371),
																						Column: int(72),
																					},
																				},
																			},
																			Op: ast.BinaryOp(16),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14268,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1371),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1371),
																					Column: int(73),
																				},
																			},
																		},
																	},
																	Left: &ast.Binary{
																		Right: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_table",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																						ast.FodderElement{
																							Comment: []string{
																								"// 2 LSB of i, 4 MSB of i+1",
																							},
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1369),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1369),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.Binary{
																				Right: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "4",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1369),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1369),
																									Column: int(67),
																								},
																							},
																						},
																					},
																					Left: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "240",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1369),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1369),
																										Column: int(61),
																									},
																								},
																							},
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "arr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1369),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1369),
																											Column: int(48),
																										},
																									},
																								},
																							},
																							Index: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14268,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1369),
																												Column: int(53),
																											},
																											End: ast.Location{
																												Line: int(1369),
																												Column: int(54),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "i",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14268,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1369),
																												Column: int(49),
																											},
																											End: ast.Location{
																												Line: int(1369),
																												Column: int(50),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1369),
																											Column: int(49),
																										},
																										End: ast.Location{
																											Line: int(1369),
																											Column: int(54),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1369),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1369),
																										Column: int(55),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1369),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1369),
																									Column: int(61),
																								},
																							},
																						},
																						Op: ast.BinaryOp(14),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1369),
																								Column: int(44),
																							},
																							End: ast.Location{
																								Line: int(1369),
																								Column: int(67),
																							},
																						},
																					},
																					Op: ast.BinaryOp(6),
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "4",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1369),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1369),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					Left: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "3",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1369),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(1369),
																										Column: int(35),
																									},
																								},
																							},
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "arr",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1369),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1369),
																											Column: int(28),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14268,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1369),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1369),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1369),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1369),
																										Column: int(31),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1369),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1369),
																									Column: int(35),
																								},
																							},
																						},
																						Op: ast.BinaryOp(14),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1369),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(1369),
																								Column: int(41),
																							},
																						},
																					},
																					Op: ast.BinaryOp(5),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1369),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1369),
																							Column: int(67),
																						},
																					},
																				},
																				Op: ast.BinaryOp(16),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1369),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1369),
																						Column: int(68),
																					},
																				},
																			},
																		},
																		Left: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_table",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																						ast.FodderElement{
																							Comment: []string{
																								"// 6 MSB of i",
																							},
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																						},
																					},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1367),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1367),
																							Column: int(23),
																						},
																					},
																				},
																			},
																			Index: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "2",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1367),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1367),
																								Column: int(43),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "252",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1367),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1367),
																									Column: int(37),
																								},
																							},
																						},
																					},
																					Left: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1367),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1367),
																										Column: int(28),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14268,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1367),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1367),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14268,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1367),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1367),
																									Column: int(31),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14268,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1367),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1367),
																								Column: int(37),
																							},
																						},
																					},
																					Op: ast.BinaryOp(14),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14268,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1367),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1367),
																							Column: int(43),
																						},
																					},
																				},
																				Op: ast.BinaryOp(6),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14268,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1367),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1367),
																						Column: int(44),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14268,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1367),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1369),
																					Column: int(68),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14268,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1367),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1371),
																				Column: int(73),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14268,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1367),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1373),
																			Column: int(42),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															EqFodder: ast.Fodder{},
															Variable: "str",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1365),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1373),
																	Column: int(42),
																},
															},
														},
													},
													Body: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p14036,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1374),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1374),
																		Column: int(12),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "arr",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14385,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1374),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1374),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "3",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14385,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1374),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1374),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14385,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1374),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1374),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14385,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1374),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1374),
																					Column: int(23),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14385,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1374),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1374),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "r",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14385,
																				FreeVars: ast.Identifiers{
																					"r",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1374),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1374),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14385,
																			FreeVars: ast.Identifiers{
																				"r",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1374),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1374),
																					Column: int(32),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14036,
															FreeVars: ast.Identifiers{
																"arr",
																"aux",
																"i",
																"r",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1374),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1374),
																	Column: int(33),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p14036,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"base64_table",
															"i",
															"r",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1365),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1374),
																Column: int(33),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14036,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"base64_table",
														"i",
														"r",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1354),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1374),
															Column: int(33),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14036,
												FreeVars: ast.Identifiers{
													"arr",
													"aux",
													"base64_table",
													"i",
													"r",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1346),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1374),
														Column: int(33),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p14036,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"base64_table",
												"i",
												"r",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1344),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1374),
													Column: int(33),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "arr",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1343),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1343),
													Column: int(18),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1343),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1343),
													Column: int(21),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "r",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1343),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1343),
													Column: int(24),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p14409,
										FreeVars: ast.Identifiers{
											"aux",
											"base64_table",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1343),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1374),
												Column: int(33),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "aux",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1376),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1376),
															Column: int(23),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "all",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14418,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1376),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1376),
														Column: int(27),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "$std",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "flatMap",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: nil,
															LeftBracketFodder: nil,
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Function{
																		ParenLeftFodder: nil,
																		ParenRightFodder: nil,
																		Body: &ast.Array{
																			Elements: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "256",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14433,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1376),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1376),
																										Column: int(36),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "a",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14433,
																								FreeVars: ast.Identifiers{
																									"a",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1376),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1376),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14433,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1376),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(1376),
																									Column: int(36),
																								},
																							},
																						},
																						Op: ast.BinaryOp(9),
																					},
																					CommaFodder: nil,
																				},
																			},
																			CloseFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: nil,
																				Name: "a",
																				CommaFodder: nil,
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "bytes",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14440,
																			FreeVars: ast.Identifiers{
																				"bytes",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1376),
																					Column: int(46),
																				},
																				End: ast.Location{
																					Line: int(1376),
																					Column: int(51),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: nil,
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"bytes",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1376),
																	Column: int(28),
																},
																End: ast.Location{
																	Line: int(1376),
																	Column: int(52),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14418,
											FreeVars: ast.Identifiers{
												"$std",
												"bytes",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1376),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1376),
													Column: int(53),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									EqFodder: ast.Fodder{},
									Variable: "sanity",
									CloseFodder: ast.Fodder{},
									Fun: nil,
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1376),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1376),
											Column: int(53),
										},
									},
								},
							},
							Body: &ast.Conditional{
								Cond: &ast.Unary{
									Expr: &ast.Var{
										Id: "sanity",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14447,
											FreeVars: ast.Identifiers{
												"sanity",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1377),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1377),
													Column: int(15),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p14447,
										FreeVars: ast.Identifiers{
											"sanity",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1377),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1377),
												Column: int(15),
											},
										},
									},
									Op: ast.UnaryOp(0),
								},
								BranchTrue: &ast.Error{
									Expr: &ast.LiteralString{
										Value: "Can only base64 encode strings / arrays of single bytes.",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14447,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1378),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1378),
													Column: int(71),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p14447,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1378),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1378),
												Column: int(71),
											},
										},
									},
								},
								BranchFalse: &ast.Apply{
									Target: &ast.Var{
										Id: "aux",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p14447,
											FreeVars: ast.Identifiers{
												"aux",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1380),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1380),
													Column: int(10),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "bytes",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14459,
														FreeVars: ast.Identifiers{
															"bytes",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1380),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1380),
																Column: int(16),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													OriginalString: "0",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14459,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1380),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1380),
																Column: int(19),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14459,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1380),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(1380),
																Column: int(23),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p14447,
										FreeVars: ast.Identifiers{
											"aux",
											"bytes",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1380),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1380),
												Column: int(24),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p14447,
									FreeVars: ast.Identifiers{
										"aux",
										"bytes",
										"sanity",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1377),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1380),
											Column: int(24),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
									},
								},
								Ctx: p14447,
								FreeVars: ast.Identifiers{
									"$std",
									"aux",
									"bytes",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1376),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1380),
										Column: int(24),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
								},
							},
							Ctx: p14447,
							FreeVars: ast.Identifiers{
								"$std",
								"base64_table",
								"bytes",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1343),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1380),
									Column: int(24),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p14447,
						FreeVars: ast.Identifiers{
							"$std",
							"base64_table",
							"input",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1337),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1380),
								Column: int(24),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "input",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1336),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1336),
								Column: int(15),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"base64_table",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1336),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1380),
					Column: int(24),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "base64DecodeBytes",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.LiteralNumber{
							OriginalString: "0",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p14480,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1384),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(1384),
										Column: int(32),
									},
								},
							},
						},
						Left: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1384),
																Column: int(8),
															},
															End: ast.Location{
																Line: int(1384),
																Column: int(11),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14480,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1384),
															Column: int(8),
														},
														End: ast.Location{
															Line: int(1384),
															Column: int(18),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14496,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1384),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1384),
																		Column: int(22),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14480,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1384),
														Column: int(8),
													},
													End: ast.Location{
														Line: int(1384),
														Column: int(23),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											OriginalString: "4",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14480,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1384),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(1384),
														Column: int(27),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1384),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1384),
										Column: int(27),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p14480,
							FreeVars: ast.Identifiers{
								"$std",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1384),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1384),
									Column: int(32),
								},
							},
						},
						Op: ast.BinaryOp(13),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "Not a base64 encoded string \"%s\"",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14480,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1385),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1385),
														Column: int(47),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "str",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14480,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1385),
														Column: int(50),
													},
													End: ast.Location{
														Line: int(1385),
														Column: int(53),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1385),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1385),
										Column: int(53),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p14480,
							FreeVars: ast.Identifiers{
								"$std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1385),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1385),
									Column: int(53),
								},
							},
						},
					},
					BranchFalse: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1388),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1388),
																	Column: int(20),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14526,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1388),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1388),
																Column: int(27),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "str",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14530,
																	FreeVars: ast.Identifiers{
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1388),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1388),
																			Column: int(31),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14526,
													FreeVars: ast.Identifiers{
														"std",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1388),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1388),
															Column: int(32),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "i",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14526,
													FreeVars: ast.Identifiers{
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1388),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1388),
															Column: int(13),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14526,
												FreeVars: ast.Identifiers{
													"i",
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1388),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1388),
														Column: int(32),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										BranchTrue: &ast.Var{
											Id: "r",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p14526,
												FreeVars: ast.Identifiers{
													"r",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1389),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1389),
														Column: int(12),
													},
												},
											},
										},
										BranchFalse: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.Array{
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "4",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14546,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1392),
																						Column: int(76),
																					},
																					End: ast.Location{
																						Line: int(1392),
																						Column: int(77),
																					},
																				},
																			},
																		},
																		Left: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_inv",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14546,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1392),
																							Column: int(50),
																						},
																						End: ast.Location{
																							Line: int(1392),
																							Column: int(60),
																						},
																					},
																				},
																			},
																			Index: &ast.Index{
																				Target: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14546,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1392),
																								Column: int(61),
																							},
																							End: ast.Location{
																								Line: int(1392),
																								Column: int(64),
																							},
																						},
																					},
																				},
																				Index: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14546,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1392),
																									Column: int(69),
																								},
																								End: ast.Location{
																									Line: int(1392),
																									Column: int(70),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14546,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1392),
																									Column: int(65),
																								},
																								End: ast.Location{
																									Line: int(1392),
																									Column: int(66),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14546,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1392),
																								Column: int(65),
																							},
																							End: ast.Location{
																								Line: int(1392),
																								Column: int(70),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14546,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1392),
																							Column: int(61),
																						},
																						End: ast.Location{
																							Line: int(1392),
																							Column: int(71),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14546,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1392),
																						Column: int(50),
																					},
																					End: ast.Location{
																						Line: int(1392),
																						Column: int(72),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14546,
																			FreeVars: ast.Identifiers{
																				"base64_inv",
																				"i",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1392),
																					Column: int(50),
																				},
																				End: ast.Location{
																					Line: int(1392),
																					Column: int(77),
																				},
																			},
																		},
																		Op: ast.BinaryOp(6),
																	},
																	Left: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "2",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14546,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1392),
																						Column: int(45),
																					},
																					End: ast.Location{
																						Line: int(1392),
																						Column: int(46),
																					},
																				},
																			},
																		},
																		Left: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_inv",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14546,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1392),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1392),
																							Column: int(33),
																						},
																					},
																				},
																			},
																			Index: &ast.Index{
																				Target: &ast.Var{
																					Id: "str",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14546,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1392),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1392),
																								Column: int(37),
																							},
																						},
																					},
																				},
																				Index: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14546,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1392),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1392),
																								Column: int(39),
																							},
																						},
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14546,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1392),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1392),
																							Column: int(40),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14546,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1392),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1392),
																						Column: int(41),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14546,
																			FreeVars: ast.Identifiers{
																				"base64_inv",
																				"i",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1392),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(1392),
																					Column: int(46),
																				},
																			},
																		},
																		Op: ast.BinaryOp(5),
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14546,
																		FreeVars: ast.Identifiers{
																			"base64_inv",
																			"i",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1392),
																				Column: int(23),
																			},
																			End: ast.Location{
																				Line: int(1392),
																				Column: int(78),
																			},
																		},
																	},
																	Op: ast.BinaryOp(16),
																},
																CommaFodder: nil,
															},
														},
														CloseFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14575,
															FreeVars: ast.Identifiers{
																"base64_inv",
																"i",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1392),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(1392),
																	Column: int(79),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: ast.Fodder{},
													Variable: "n1",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1392),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1392),
															Column: int(79),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.LiteralString{
																	Value: "=",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14582,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1395),
																				Column: int(30),
																			},
																			End: ast.Location{
																				Line: int(1395),
																				Column: int(33),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																Left: &ast.Index{
																	Target: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14582,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1395),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1395),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	Index: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "2",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14582,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1395),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1395),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14582,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1395),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1395),
																						Column: int(21),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14582,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1395),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1395),
																					Column: int(25),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14582,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1395),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1395),
																				Column: int(26),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14582,
																	FreeVars: ast.Identifiers{
																		"i",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1395),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1395),
																			Column: int(33),
																		},
																	},
																},
																Op: ast.BinaryOp(12),
															},
															BranchTrue: &ast.Array{
																Elements: nil,
																CloseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14582,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1395),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(1395),
																			Column: int(41),
																		},
																	},
																},
																TrailingComma: false,
															},
															BranchFalse: &ast.Array{
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "2",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14599,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1396),
																								Column: int(83),
																							},
																							End: ast.Location{
																								Line: int(1396),
																								Column: int(84),
																							},
																						},
																					},
																				},
																				Left: &ast.Index{
																					Target: &ast.Var{
																						Id: "base64_inv",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14599,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1396),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(1396),
																									Column: int(67),
																								},
																							},
																						},
																					},
																					Index: &ast.Index{
																						Target: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14599,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1396),
																										Column: int(68),
																									},
																									End: ast.Location{
																										Line: int(1396),
																										Column: int(71),
																									},
																								},
																							},
																						},
																						Index: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "2",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14599,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1396),
																											Column: int(76),
																										},
																										End: ast.Location{
																											Line: int(1396),
																											Column: int(77),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14599,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1396),
																											Column: int(72),
																										},
																										End: ast.Location{
																											Line: int(1396),
																											Column: int(73),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14599,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1396),
																										Column: int(72),
																									},
																									End: ast.Location{
																										Line: int(1396),
																										Column: int(77),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14599,
																							FreeVars: ast.Identifiers{
																								"i",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1396),
																									Column: int(68),
																								},
																								End: ast.Location{
																									Line: int(1396),
																									Column: int(78),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14599,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1396),
																								Column: int(57),
																							},
																							End: ast.Location{
																								Line: int(1396),
																								Column: int(79),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14599,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1396),
																							Column: int(57),
																						},
																						End: ast.Location{
																							Line: int(1396),
																							Column: int(84),
																						},
																					},
																				},
																				Op: ast.BinaryOp(6),
																			},
																			Left: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "4",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14599,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1396),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(1396),
																								Column: int(53),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "15",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14599,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1396),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1396),
																									Column: int(47),
																								},
																							},
																						},
																					},
																					Left: &ast.Index{
																						Target: &ast.Var{
																							Id: "base64_inv",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14599,
																								FreeVars: ast.Identifiers{
																									"base64_inv",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1396),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1396),
																										Column: int(30),
																									},
																								},
																							},
																						},
																						Index: &ast.Index{
																							Target: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14599,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1396),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(1396),
																											Column: int(34),
																										},
																									},
																								},
																							},
																							Index: &ast.Binary{
																								Right: &ast.LiteralNumber{
																									OriginalString: "1",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14599,
																										FreeVars: ast.Identifiers{},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1396),
																												Column: int(39),
																											},
																											End: ast.Location{
																												Line: int(1396),
																												Column: int(40),
																											},
																										},
																									},
																								},
																								Left: &ast.Var{
																									Id: "i",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14599,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1396),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1396),
																												Column: int(36),
																											},
																										},
																									},
																								},
																								OpFodder: ast.Fodder{},
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14599,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1396),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1396),
																											Column: int(40),
																										},
																									},
																								},
																								Op: ast.BinaryOp(3),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14599,
																								FreeVars: ast.Identifiers{
																									"i",
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1396),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(1396),
																										Column: int(41),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14599,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																								"i",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1396),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1396),
																									Column: int(42),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14599,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1396),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1396),
																								Column: int(47),
																							},
																						},
																					},
																					Op: ast.BinaryOp(14),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14599,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1396),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1396),
																							Column: int(53),
																						},
																					},
																				},
																				Op: ast.BinaryOp(5),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14599,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1396),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(1396),
																						Column: int(85),
																					},
																				},
																			},
																			Op: ast.BinaryOp(16),
																		},
																		CommaFodder: nil,
																	},
																},
																CloseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14582,
																	FreeVars: ast.Identifiers{
																		"base64_inv",
																		"i",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1396),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1396),
																			Column: int(86),
																		},
																	},
																},
																TrailingComma: false,
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(12),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																	},
																},
																Ctx: p14582,
																FreeVars: ast.Identifiers{
																	"base64_inv",
																	"i",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1395),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1396),
																		Column: int(86),
																	},
																},
															},
														},
														EqFodder: ast.Fodder{},
														Variable: "n2",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1394),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1396),
																Column: int(86),
															},
														},
													},
												},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralString{
																		Value: "=",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14643,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1399),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(1399),
																					Column: int(33),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	Left: &ast.Index{
																		Target: &ast.Var{
																			Id: "str",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14643,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1399),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(1399),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		Index: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "3",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14643,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1399),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1399),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14643,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1399),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(1399),
																							Column: int(21),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14643,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1399),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1399),
																						Column: int(25),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14643,
																			FreeVars: ast.Identifiers{
																				"i",
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1399),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1399),
																					Column: int(26),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14643,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1399),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1399),
																				Column: int(33),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.Array{
																	Elements: nil,
																	CloseFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14643,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1399),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1399),
																				Column: int(41),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																BranchFalse: &ast.Array{
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				Right: &ast.Index{
																					Target: &ast.Var{
																						Id: "base64_inv",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14660,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1400),
																									Column: int(55),
																								},
																								End: ast.Location{
																									Line: int(1400),
																									Column: int(65),
																								},
																							},
																						},
																					},
																					Index: &ast.Index{
																						Target: &ast.Var{
																							Id: "str",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14660,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1400),
																										Column: int(66),
																									},
																									End: ast.Location{
																										Line: int(1400),
																										Column: int(69),
																									},
																								},
																							},
																						},
																						Index: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "3",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14660,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1400),
																											Column: int(74),
																										},
																										End: ast.Location{
																											Line: int(1400),
																											Column: int(75),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14660,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1400),
																											Column: int(70),
																										},
																										End: ast.Location{
																											Line: int(1400),
																											Column: int(71),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14660,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1400),
																										Column: int(70),
																									},
																									End: ast.Location{
																										Line: int(1400),
																										Column: int(75),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14660,
																							FreeVars: ast.Identifiers{
																								"i",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1400),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1400),
																									Column: int(76),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14660,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1400),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(1400),
																								Column: int(77),
																							},
																						},
																					},
																				},
																				Left: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "6",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14660,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1400),
																									Column: int(51),
																								},
																								End: ast.Location{
																									Line: int(1400),
																									Column: int(52),
																								},
																							},
																						},
																					},
																					Left: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "3",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14660,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1400),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1400),
																										Column: int(46),
																									},
																								},
																							},
																						},
																						Left: &ast.Index{
																							Target: &ast.Var{
																								Id: "base64_inv",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14660,
																									FreeVars: ast.Identifiers{
																										"base64_inv",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1400),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1400),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							Index: &ast.Index{
																								Target: &ast.Var{
																									Id: "str",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14660,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1400),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(1400),
																												Column: int(34),
																											},
																										},
																									},
																								},
																								Index: &ast.Binary{
																									Right: &ast.LiteralNumber{
																										OriginalString: "2",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p14660,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1400),
																													Column: int(39),
																												},
																												End: ast.Location{
																													Line: int(1400),
																													Column: int(40),
																												},
																											},
																										},
																									},
																									Left: &ast.Var{
																										Id: "i",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p14660,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1400),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(1400),
																													Column: int(36),
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14660,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1400),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1400),
																												Column: int(40),
																											},
																										},
																									},
																									Op: ast.BinaryOp(3),
																								},
																								RightBracketFodder: ast.Fodder{},
																								LeftBracketFodder: ast.Fodder{},
																								Id: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14660,
																									FreeVars: ast.Identifiers{
																										"i",
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1400),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(1400),
																											Column: int(41),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14660,
																								FreeVars: ast.Identifiers{
																									"base64_inv",
																									"i",
																									"str",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1400),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1400),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14660,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																								"i",
																								"str",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1400),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1400),
																									Column: int(46),
																								},
																							},
																						},
																						Op: ast.BinaryOp(14),
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14660,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1400),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(1400),
																								Column: int(52),
																							},
																						},
																					},
																					Op: ast.BinaryOp(5),
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14660,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1400),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1400),
																							Column: int(77),
																						},
																					},
																				},
																				Op: ast.BinaryOp(16),
																			},
																			CommaFodder: nil,
																		},
																	},
																	CloseFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14643,
																		FreeVars: ast.Identifiers{
																			"base64_inv",
																			"i",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1400),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1400),
																				Column: int(78),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(12),
																		},
																	},
																	Ctx: p14643,
																	FreeVars: ast.Identifiers{
																		"base64_inv",
																		"i",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1399),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1400),
																			Column: int(78),
																		},
																	},
																},
															},
															EqFodder: ast.Fodder{},
															Variable: "n3",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1398),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1400),
																	Column: int(78),
																},
															},
														},
													},
													Body: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p14526,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1401),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1401),
																		Column: int(14),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14702,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1401),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1401),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "4",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14702,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1401),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1401),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14702,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1401),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1401),
																						Column: int(21),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14702,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1401),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1401),
																					Column: int(25),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Var{
																			Id: "n3",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14702,
																				FreeVars: ast.Identifiers{
																					"n3",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1401),
																						Column: int(41),
																					},
																					End: ast.Location{
																						Line: int(1401),
																						Column: int(43),
																					},
																				},
																			},
																		},
																		Left: &ast.Binary{
																			Right: &ast.Var{
																				Id: "n2",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14702,
																					FreeVars: ast.Identifiers{
																						"n2",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1401),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1401),
																							Column: int(38),
																						},
																					},
																				},
																			},
																			Left: &ast.Binary{
																				Right: &ast.Var{
																					Id: "n1",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14702,
																						FreeVars: ast.Identifiers{
																							"n1",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1401),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(1401),
																								Column: int(33),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "r",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14702,
																						FreeVars: ast.Identifiers{
																							"r",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1401),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(1401),
																								Column: int(28),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14702,
																					FreeVars: ast.Identifiers{
																						"n1",
																						"r",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1401),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1401),
																							Column: int(33),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14702,
																				FreeVars: ast.Identifiers{
																					"n1",
																					"n2",
																					"r",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1401),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(1401),
																						Column: int(38),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14702,
																			FreeVars: ast.Identifiers{
																				"n1",
																				"n2",
																				"n3",
																				"r",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1401),
																					Column: int(27),
																				},
																				End: ast.Location{
																					Line: int(1401),
																					Column: int(43),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14526,
															FreeVars: ast.Identifiers{
																"aux",
																"i",
																"n1",
																"n2",
																"n3",
																"r",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1401),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1401),
																	Column: int(44),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
															ast.FodderElement{
																Comment: []string{
																	"// 2 LSB of i+2, all 6 bits of i+3",
																},
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p14526,
														FreeVars: ast.Identifiers{
															"aux",
															"base64_inv",
															"i",
															"n1",
															"n2",
															"r",
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1398),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1401),
																Column: int(44),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
														},
														ast.FodderElement{
															Comment: []string{
																"// 4 LSB of i+1, 4MSB of i+2",
															},
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(10),
														},
													},
													Ctx: p14526,
													FreeVars: ast.Identifiers{
														"aux",
														"base64_inv",
														"i",
														"n1",
														"r",
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1394),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1401),
															Column: int(44),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
													ast.FodderElement{
														Comment: []string{
															"// all 6 bits of i, 2 MSB of i+1",
														},
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p14526,
												FreeVars: ast.Identifiers{
													"aux",
													"base64_inv",
													"i",
													"r",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1392),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1401),
														Column: int(44),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p14526,
											FreeVars: ast.Identifiers{
												"aux",
												"base64_inv",
												"i",
												"r",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1388),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1401),
													Column: int(44),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "str",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1387),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1387),
													Column: int(20),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1387),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(1387),
													Column: int(23),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "r",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1387),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(1387),
													Column: int(26),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p14737,
										FreeVars: ast.Identifiers{
											"aux",
											"base64_inv",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1387),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1401),
												Column: int(44),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "aux",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Apply{
							Target: &ast.Var{
								Id: "aux",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p14480,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1402),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1402),
											Column: int(10),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "str",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14745,
												FreeVars: ast.Identifiers{
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1402),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1402),
														Column: int(14),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											OriginalString: "0",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14745,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1402),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1402),
														Column: int(17),
													},
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											Elements: nil,
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14745,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1402),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(1402),
														Column: int(21),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p14480,
								FreeVars: ast.Identifiers{
									"aux",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1402),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1402),
										Column: int(22),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p14480,
							FreeVars: ast.Identifiers{
								"base64_inv",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1387),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1402),
									Column: int(22),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p14480,
						FreeVars: ast.Identifiers{
							"$std",
							"base64_inv",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1384),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1402),
								Column: int(22),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1383),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1383),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"base64_inv",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1383),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1402),
					Column: int(22),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "base64Decode",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1405),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1405),
													Column: int(22),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "base64DecodeBytes",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p14766,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1405),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(1405),
												Column: int(40),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14770,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1405),
															Column: int(41),
														},
														End: ast.Location{
															Line: int(1405),
															Column: int(44),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p14766,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1405),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(1405),
											Column: int(45),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "bytes",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1405),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1405),
									Column: int(45),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1406),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1406),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "join",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p14779,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1406),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1406),
										Column: int(13),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										Value: "",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14783,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1406),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1406),
													Column: int(16),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1406),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(1406),
															Column: int(21),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "map",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14783,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1406),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(1406),
														Column: int(25),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1406),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(1406),
																		Column: int(29),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "char",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14795,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1406),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1406),
																	Column: int(34),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "bytes",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14795,
															FreeVars: ast.Identifiers{
																"bytes",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1406),
																	Column: int(36),
																},
																End: ast.Location{
																	Line: int(1406),
																	Column: int(41),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14783,
											FreeVars: ast.Identifiers{
												"bytes",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1406),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1406),
													Column: int(42),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p14779,
							FreeVars: ast.Identifiers{
								"bytes",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1406),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1406),
									Column: int(43),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p14779,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1405),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1406),
								Column: int(43),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1404),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1404),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1404),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1406),
					Column: int(43),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "reverse",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1409),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1409),
													Column: int(18),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p14814,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1409),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1409),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14818,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1409),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1409),
															Column: int(29),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p14814,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1409),
											Column: int(15),
										},
										End: ast.Location{
											Line: int(1409),
											Column: int(30),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "l",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1409),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1409),
									Column: int(30),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1410),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1410),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "makeArray",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p14827,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1410),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1410),
										Column: int(18),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "l",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14831,
											FreeVars: ast.Identifiers{
												"l",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1410),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1410),
													Column: int(20),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Index{
											Target: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14836,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1410),
															Column: int(34),
														},
														End: ast.Location{
															Line: int(1410),
															Column: int(37),
														},
													},
												},
											},
											Index: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "1",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14836,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1410),
																Column: int(46),
															},
															End: ast.Location{
																Line: int(1410),
																Column: int(47),
															},
														},
													},
												},
												Left: &ast.Binary{
													Right: &ast.Var{
														Id: "i",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14836,
															FreeVars: ast.Identifiers{
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1410),
																	Column: int(42),
																},
																End: ast.Location{
																	Line: int(1410),
																	Column: int(43),
																},
															},
														},
													},
													Left: &ast.Var{
														Id: "l",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14836,
															FreeVars: ast.Identifiers{
																"l",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1410),
																	Column: int(38),
																},
																End: ast.Location{
																	Line: int(1410),
																	Column: int(39),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14836,
														FreeVars: ast.Identifiers{
															"i",
															"l",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1410),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(1410),
																Column: int(43),
															},
														},
													},
													Op: ast.BinaryOp(4),
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14836,
													FreeVars: ast.Identifiers{
														"i",
														"l",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1410),
															Column: int(38),
														},
														End: ast.Location{
															Line: int(1410),
															Column: int(47),
														},
													},
												},
												Op: ast.BinaryOp(4),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14836,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"l",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1410),
														Column: int(34),
													},
													End: ast.Location{
														Line: int(1410),
														Column: int(48),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1410),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1410),
														Column: int(32),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p14831,
											FreeVars: ast.Identifiers{
												"arr",
												"l",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1410),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(1410),
													Column: int(48),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p14827,
							FreeVars: ast.Identifiers{
								"arr",
								"l",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1410),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1410),
									Column: int(49),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p14827,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1409),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1410),
								Column: int(49),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1408),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1408),
								Column: int(14),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1408),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1410),
					Column: int(49),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "sort",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1415),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1415),
																	Column: int(20),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14867,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1415),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1415),
																Column: int(27),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14871,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1415),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1415),
																			Column: int(31),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14867,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1415),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1415),
															Column: int(32),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "l",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1415),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1415),
													Column: int(32),
												},
											},
										},
									},
									Body: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.LiteralNumber{
												OriginalString: "1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14877,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1416),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(1416),
															Column: int(30),
														},
													},
												},
											},
											Left: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1416),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(1416),
																	Column: int(13),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p14877,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1416),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1416),
																Column: int(20),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14886,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1416),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1416),
																			Column: int(24),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p14877,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1416),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1416),
															Column: int(25),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14877,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1416),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1416),
														Column: int(30),
													},
												},
											},
											Op: ast.BinaryOp(10),
										},
										BranchTrue: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p14877,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1417),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1417),
														Column: int(12),
													},
												},
											},
										},
										BranchFalse: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.LiteralNumber{
														OriginalString: "0",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p14896,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1419),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(1419),
																	Column: int(22),
																},
															},
														},
													},
													EqFodder: ast.Fodder{},
													Variable: "pos",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1419),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1419),
															Column: int(22),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Apply{
															Target: &ast.Var{
																Id: "keyF",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14901,
																	FreeVars: ast.Identifiers{
																		"keyF",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1420),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1420),
																			Column: int(27),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			Target: &ast.Var{
																				Id: "arr",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14906,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1420),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1420),
																							Column: int(31),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "pos",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14906,
																					FreeVars: ast.Identifiers{
																						"pos",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1420),
																							Column: int(32),
																						},
																						End: ast.Location{
																							Line: int(1420),
																							Column: int(35),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14906,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"pos",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1420),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1420),
																						Column: int(36),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p14901,
																FreeVars: ast.Identifiers{
																	"arr",
																	"keyF",
																	"pos",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1420),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1420),
																		Column: int(37),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														EqFodder: ast.Fodder{},
														Variable: "pivot",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1420),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1420),
																Column: int(37),
															},
														},
													},
												},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1421),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1421),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "makeArray",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14919,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1421),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1421),
																				Column: int(35),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				Right: &ast.LiteralNumber{
																					OriginalString: "1",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14924,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1421),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(1421),
																								Column: int(41),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "l",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14924,
																						FreeVars: ast.Identifiers{
																							"l",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1421),
																								Column: int(36),
																							},
																							End: ast.Location{
																								Line: int(1421),
																								Column: int(37),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14924,
																					FreeVars: ast.Identifiers{
																						"l",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1421),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1421),
																							Column: int(41),
																						},
																					},
																				},
																				Op: ast.BinaryOp(4),
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Function{
																				ParenLeftFodder: ast.Fodder{},
																				ParenRightFodder: ast.Fodder{},
																				Body: &ast.Conditional{
																					Cond: &ast.Binary{
																						Right: &ast.Var{
																							Id: "pos",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"pos",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(62),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(65),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(59),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14932,
																							FreeVars: ast.Identifiers{
																								"i",
																								"pos",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1421),
																									Column: int(58),
																								},
																								End: ast.Location{
																									Line: int(1421),
																									Column: int(65),
																								},
																							},
																						},
																						Op: ast.BinaryOp(9),
																					},
																					BranchTrue: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(71),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(74),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(75),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(76),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14932,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1421),
																									Column: int(71),
																								},
																								End: ast.Location{
																									Line: int(1421),
																									Column: int(77),
																								},
																							},
																						},
																					},
																					BranchFalse: &ast.Index{
																						Target: &ast.Var{
																							Id: "arr",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(83),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(86),
																									},
																								},
																							},
																						},
																						Index: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14932,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1421),
																											Column: int(91),
																										},
																										End: ast.Location{
																											Line: int(1421),
																											Column: int(92),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14932,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1421),
																											Column: int(87),
																										},
																										End: ast.Location{
																											Line: int(1421),
																											Column: int(88),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14932,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1421),
																										Column: int(87),
																									},
																									End: ast.Location{
																										Line: int(1421),
																										Column: int(92),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14932,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1421),
																									Column: int(83),
																								},
																								End: ast.Location{
																									Line: int(1421),
																									Column: int(93),
																								},
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					ElseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14932,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																							"pos",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1421),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(1421),
																								Column: int(93),
																							},
																						},
																					},
																				},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "i",
																						CommaFodder: nil,
																						EqFodder: nil,
																						DefaultArg: nil,
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1421),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(1421),
																								Column: int(53),
																							},
																						},
																					},
																				},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p14924,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"pos",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1421),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(1421),
																							Column: int(93),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14919,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"l",
																		"pos",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1421),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1421),
																			Column: int(94),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															EqFodder: ast.Fodder{},
															Variable: "rest",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1421),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1421),
																	Column: int(94),
																},
															},
														},
													},
													Body: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Body: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "std",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1422),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1422),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "filter",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14963,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1422),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1422),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Function{
																					ParenLeftFodder: ast.Fodder{},
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Binary{
																						Right: &ast.Var{
																							Id: "pivot",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14969,
																								FreeVars: ast.Identifiers{
																									"pivot",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1422),
																										Column: int(55),
																									},
																									End: ast.Location{
																										Line: int(1422),
																										Column: int(60),
																									},
																								},
																							},
																						},
																						Left: &ast.Apply{
																							Target: &ast.Var{
																								Id: "keyF",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14969,
																									FreeVars: ast.Identifiers{
																										"keyF",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1422),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1422),
																											Column: int(49),
																										},
																									},
																								},
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											Id: "x",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p14976,
																												FreeVars: ast.Identifiers{
																													"x",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1422),
																														Column: int(50),
																													},
																													End: ast.Location{
																														Line: int(1422),
																														Column: int(51),
																													},
																												},
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14969,
																								FreeVars: ast.Identifiers{
																									"keyF",
																									"x",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1422),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1422),
																										Column: int(52),
																									},
																								},
																							},
																							TrailingComma: false,
																							TailStrict: false,
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p14969,
																							FreeVars: ast.Identifiers{
																								"keyF",
																								"pivot",
																								"x",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1422),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1422),
																									Column: int(60),
																								},
																							},
																						},
																						Op: ast.BinaryOp(9),
																					},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "x",
																							CommaFodder: nil,
																							EqFodder: nil,
																							DefaultArg: nil,
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1422),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1422),
																									Column: int(43),
																								},
																							},
																						},
																					},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14981,
																						FreeVars: ast.Identifiers{
																							"keyF",
																							"pivot",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1422),
																								Column: int(33),
																							},
																							End: ast.Location{
																								Line: int(1422),
																								Column: int(60),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "rest",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p14981,
																						FreeVars: ast.Identifiers{
																							"rest",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1422),
																								Column: int(62),
																							},
																							End: ast.Location{
																								Line: int(1422),
																								Column: int(66),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14963,
																		FreeVars: ast.Identifiers{
																			"keyF",
																			"pivot",
																			"rest",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1422),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1422),
																				Column: int(67),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																EqFodder: ast.Fodder{},
																Variable: "left",
																CloseFodder: ast.Fodder{},
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1422),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1422),
																		Column: int(67),
																	},
																},
															},
														},
														Body: &ast.Local{
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Body: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1423),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1423),
																							Column: int(26),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "filter",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p14993,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1423),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1423),
																						Column: int(33),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Function{
																						ParenLeftFodder: ast.Fodder{},
																						ParenRightFodder: ast.Fodder{},
																						Body: &ast.Binary{
																							Right: &ast.Var{
																								Id: "pivot",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14999,
																									FreeVars: ast.Identifiers{
																										"pivot",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1423),
																											Column: int(57),
																										},
																										End: ast.Location{
																											Line: int(1423),
																											Column: int(62),
																										},
																									},
																								},
																							},
																							Left: &ast.Apply{
																								Target: &ast.Var{
																									Id: "keyF",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p14999,
																										FreeVars: ast.Identifiers{
																											"keyF",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1423),
																												Column: int(46),
																											},
																											End: ast.Location{
																												Line: int(1423),
																												Column: int(50),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												Id: "x",
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p15006,
																													FreeVars: ast.Identifiers{
																														"x",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1423),
																															Column: int(51),
																														},
																														End: ast.Location{
																															Line: int(1423),
																															Column: int(52),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p14999,
																									FreeVars: ast.Identifiers{
																										"keyF",
																										"x",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1423),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(1423),
																											Column: int(53),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p14999,
																								FreeVars: ast.Identifiers{
																									"keyF",
																									"pivot",
																									"x",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1423),
																										Column: int(46),
																									},
																									End: ast.Location{
																										Line: int(1423),
																										Column: int(62),
																									},
																								},
																							},
																							Op: ast.BinaryOp(8),
																						},
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: ast.Fodder{},
																								Name: "x",
																								CommaFodder: nil,
																								EqFodder: nil,
																								DefaultArg: nil,
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1423),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(1423),
																										Column: int(44),
																									},
																								},
																							},
																						},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15011,
																							FreeVars: ast.Identifiers{
																								"keyF",
																								"pivot",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1423),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1423),
																									Column: int(62),
																								},
																							},
																						},
																						TrailingComma: false,
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "rest",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15011,
																							FreeVars: ast.Identifiers{
																								"rest",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1423),
																									Column: int(64),
																								},
																								End: ast.Location{
																									Line: int(1423),
																									Column: int(68),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14993,
																			FreeVars: ast.Identifiers{
																				"keyF",
																				"pivot",
																				"rest",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1423),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(1423),
																					Column: int(69),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	EqFodder: ast.Fodder{},
																	Variable: "right",
																	CloseFodder: ast.Fodder{},
																	Fun: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1423),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1423),
																			Column: int(69),
																		},
																	},
																},
															},
															Body: &ast.Binary{
																Right: &ast.Apply{
																	Target: &ast.Var{
																		Id: "quickSort",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14877,
																			FreeVars: ast.Identifiers{
																				"quickSort",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1424),
																					Column: int(46),
																				},
																				End: ast.Location{
																					Line: int(1424),
																					Column: int(55),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "right",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p15022,
																						FreeVars: ast.Identifiers{
																							"right",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1424),
																								Column: int(56),
																							},
																							End: ast.Location{
																								Line: int(1424),
																								Column: int(61),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "keyF",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p15022,
																						FreeVars: ast.Identifiers{
																							"keyF",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1424),
																								Column: int(63),
																							},
																							End: ast.Location{
																								Line: int(1424),
																								Column: int(67),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14877,
																		FreeVars: ast.Identifiers{
																			"keyF",
																			"quickSort",
																			"right",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1424),
																				Column: int(46),
																			},
																			End: ast.Location{
																				Line: int(1424),
																				Column: int(68),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																Left: &ast.Binary{
																	Right: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					Target: &ast.Var{
																						Id: "arr",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15032,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1424),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1424),
																									Column: int(37),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "pos",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15032,
																							FreeVars: ast.Identifiers{
																								"pos",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1424),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1424),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p15032,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"pos",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1424),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1424),
																								Column: int(42),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14877,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"pos",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1424),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1424),
																					Column: int(43),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	Left: &ast.Apply{
																		Target: &ast.Var{
																			Id: "quickSort",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																					},
																				},
																				Ctx: p14877,
																				FreeVars: ast.Identifiers{
																					"quickSort",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1424),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1424),
																						Column: int(18),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "left",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15044,
																							FreeVars: ast.Identifiers{
																								"left",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1424),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(1424),
																									Column: int(23),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "keyF",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15044,
																							FreeVars: ast.Identifiers{
																								"keyF",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1424),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1424),
																									Column: int(29),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p14877,
																			FreeVars: ast.Identifiers{
																				"keyF",
																				"left",
																				"quickSort",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1424),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(1424),
																					Column: int(30),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p14877,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"keyF",
																			"left",
																			"pos",
																			"quickSort",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1424),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1424),
																				Column: int(43),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p14877,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"keyF",
																		"left",
																		"pos",
																		"quickSort",
																		"right",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1424),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1424),
																			Column: int(68),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p14877,
																FreeVars: ast.Identifiers{
																	"arr",
																	"keyF",
																	"left",
																	"pivot",
																	"pos",
																	"quickSort",
																	"rest",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1423),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1424),
																		Column: int(68),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p14877,
															FreeVars: ast.Identifiers{
																"arr",
																"keyF",
																"pivot",
																"pos",
																"quickSort",
																"rest",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1422),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1424),
																	Column: int(68),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p14877,
														FreeVars: ast.Identifiers{
															"arr",
															"keyF",
															"l",
															"pivot",
															"pos",
															"quickSort",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1421),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1424),
																Column: int(68),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p14877,
													FreeVars: ast.Identifiers{
														"arr",
														"keyF",
														"l",
														"pos",
														"quickSort",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1420),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1424),
															Column: int(68),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p14877,
												FreeVars: ast.Identifiers{
													"arr",
													"keyF",
													"l",
													"quickSort",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1419),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1424),
														Column: int(68),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p14877,
											FreeVars: ast.Identifiers{
												"arr",
												"keyF",
												"l",
												"quickSort",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1416),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1424),
													Column: int(68),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p14877,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"quickSort",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1415),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1424),
												Column: int(68),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1414),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1414),
												Column: int(24),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "keyF",
										CommaFodder: nil,
										EqFodder: ast.Fodder{},
										DefaultArg: &ast.Var{
											Id: "id",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p14877,
												FreeVars: ast.Identifiers{
													"id",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1414),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1414),
														Column: int(33),
													},
												},
											},
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1414),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(1414),
												Column: int(33),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p15069,
									FreeVars: ast.Identifiers{
										"id",
										"quickSort",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1414),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1424),
											Column: int(68),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "quickSort",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Body: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1427),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1427),
																		Column: int(21),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15081,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1427),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1427),
																	Column: int(28),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15085,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1427),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1427),
																				Column: int(30),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15081,
														FreeVars: ast.Identifiers{
															"a",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1427),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1427),
																Column: int(31),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "la",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1427),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1427),
														Column: int(31),
													},
												},
											},
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1427),
																		Column: int(38),
																	},
																	End: ast.Location{
																		Line: int(1427),
																		Column: int(41),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15093,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1427),
																	Column: int(38),
																},
																End: ast.Location{
																	Line: int(1427),
																	Column: int(48),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15097,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1427),
																				Column: int(49),
																			},
																			End: ast.Location{
																				Line: int(1427),
																				Column: int(50),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15093,
														FreeVars: ast.Identifiers{
															"b",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1427),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(1427),
																Column: int(51),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "lb",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1427),
														Column: int(33),
													},
													End: ast.Location{
														Line: int(1427),
														Column: int(51),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Body: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Var{
																	Id: "la",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"la",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1429),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1429),
																				Column: int(19),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1429),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(1429),
																				Column: int(13),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15106,
																	FreeVars: ast.Identifiers{
																		"i",
																		"la",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1429),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(1429),
																			Column: int(19),
																		},
																	},
																},
																Op: ast.BinaryOp(12),
															},
															BranchTrue: &ast.Binary{
																Right: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "$std",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "slice",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: nil,
																		LeftBracketFodder: nil,
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "b",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p15106,
																						FreeVars: ast.Identifiers{
																							"b",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1430),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1430),
																								Column: int(21),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "j",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p15106,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1430),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(1430),
																								Column: int(23),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNull{
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNull{
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: nil,
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"b",
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1430),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1430),
																				Column: int(25),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																Left: &ast.Var{
																	Id: "prefix",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"prefix",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1430),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1430),
																				Column: int(17),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15106,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"b",
																		"j",
																		"prefix",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1430),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1430),
																			Column: int(25),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Var{
																		Id: "lb",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"lb",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1431),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1431),
																					Column: int(24),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1431),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(1431),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"j",
																			"lb",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1431),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1431),
																				Column: int(24),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.Binary{
																	Right: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "$std",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"$std",
																					},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "slice",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: nil,
																			LeftBracketFodder: nil,
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "a",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15106,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1432),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1432),
																									Column: int(21),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15106,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1432),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(1432),
																									Column: int(23),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: nil,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: nil,
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																				"a",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1432),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1432),
																					Column: int(25),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	Left: &ast.Var{
																		Id: "prefix",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																				},
																			},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"prefix",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1432),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1432),
																					Column: int(17),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"a",
																			"i",
																			"prefix",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1432),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1432),
																				Column: int(25),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																BranchFalse: &ast.Conditional{
																	Cond: &ast.Binary{
																		Right: &ast.Apply{
																			Target: &ast.Var{
																				Id: "keyF",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15106,
																					FreeVars: ast.Identifiers{
																						"keyF",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1434),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1434),
																							Column: int(32),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							Target: &ast.Var{
																								Id: "b",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15164,
																									FreeVars: ast.Identifiers{
																										"b",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1434),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(1434),
																											Column: int(34),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15164,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1434),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1434),
																											Column: int(36),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15164,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1434),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1434),
																										Column: int(37),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15106,
																				FreeVars: ast.Identifiers{
																					"b",
																					"j",
																					"keyF",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1434),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1434),
																						Column: int(38),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		Left: &ast.Apply{
																			Target: &ast.Var{
																				Id: "keyF",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15106,
																					FreeVars: ast.Identifiers{
																						"keyF",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1434),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(1434),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							Target: &ast.Var{
																								Id: "a",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15176,
																									FreeVars: ast.Identifiers{
																										"a",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1434),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(1434),
																											Column: int(20),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15176,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1434),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1434),
																											Column: int(22),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15176,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1434),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(1434),
																										Column: int(23),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15106,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																					"keyF",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1434),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(1434),
																						Column: int(24),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"a",
																				"b",
																				"i",
																				"j",
																				"keyF",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1434),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(1434),
																					Column: int(38),
																				},
																			},
																		},
																		Op: ast.BinaryOp(10),
																	},
																	BranchTrue: &ast.Apply{
																		Target: &ast.Var{
																			Id: "aux",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(12),
																					},
																				},
																				Ctx: p15106,
																				FreeVars: ast.Identifiers{
																					"aux",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1435),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1435),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15190,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1435),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(1435),
																										Column: int(22),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15190,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1435),
																										Column: int(17),
																									},
																									End: ast.Location{
																										Line: int(1435),
																										Column: int(18),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15190,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1435),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(1435),
																									Column: int(22),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "j",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15190,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1435),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(1435),
																									Column: int(25),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "a",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p15201,
																												FreeVars: ast.Identifiers{
																													"a",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1435),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(1435),
																														Column: int(38),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "i",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p15201,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1435),
																														Column: int(39),
																													},
																													End: ast.Location{
																														Line: int(1435),
																														Column: int(40),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p15201,
																											FreeVars: ast.Identifiers{
																												"a",
																												"i",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1435),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(1435),
																													Column: int(41),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15190,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1435),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(1435),
																										Column: int(42),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						Left: &ast.Var{
																							Id: "prefix",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15190,
																								FreeVars: ast.Identifiers{
																									"prefix",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1435),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(1435),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15190,
																							FreeVars: ast.Identifiers{
																								"a",
																								"i",
																								"prefix",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1435),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(1435),
																									Column: int(42),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"a",
																				"aux",
																				"i",
																				"j",
																				"prefix",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1435),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1435),
																					Column: int(43),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: true,
																	},
																	BranchFalse: &ast.Apply{
																		Target: &ast.Var{
																			Id: "aux",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(12),
																					},
																				},
																				Ctx: p15106,
																				FreeVars: ast.Identifiers{
																					"aux",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1437),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1437),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15217,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1437),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(1437),
																									Column: int(18),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.LiteralNumber{
																							OriginalString: "1",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15217,
																								FreeVars: ast.Identifiers{},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1437),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(1437),
																										Column: int(25),
																									},
																								},
																							},
																						},
																						Left: &ast.Var{
																							Id: "j",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15217,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1437),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1437),
																										Column: int(21),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15217,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1437),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1437),
																									Column: int(25),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						Right: &ast.Array{
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "b",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p15229,
																												FreeVars: ast.Identifiers{
																													"b",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1437),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(1437),
																														Column: int(38),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "j",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p15229,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1437),
																														Column: int(39),
																													},
																													End: ast.Location{
																														Line: int(1437),
																														Column: int(40),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p15229,
																											FreeVars: ast.Identifiers{
																												"b",
																												"j",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1437),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(1437),
																													Column: int(41),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							CloseFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15217,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1437),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(1437),
																										Column: int(42),
																									},
																								},
																							},
																							TrailingComma: false,
																						},
																						Left: &ast.Var{
																							Id: "prefix",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15217,
																								FreeVars: ast.Identifiers{
																									"prefix",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1437),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(1437),
																										Column: int(33),
																									},
																								},
																							},
																						},
																						OpFodder: ast.Fodder{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15217,
																							FreeVars: ast.Identifiers{
																								"b",
																								"j",
																								"prefix",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1437),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(1437),
																									Column: int(42),
																								},
																							},
																						},
																						Op: ast.BinaryOp(3),
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15106,
																			FreeVars: ast.Identifiers{
																				"aux",
																				"b",
																				"i",
																				"j",
																				"prefix",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1437),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1437),
																					Column: int(43),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: true,
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																			},
																		},
																		Ctx: p15106,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"i",
																			"j",
																			"keyF",
																			"prefix",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1434),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1437),
																				Column: int(43),
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15106,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"a",
																		"aux",
																		"b",
																		"i",
																		"j",
																		"keyF",
																		"lb",
																		"prefix",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1431),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(1437),
																			Column: int(43),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: p15106,
																FreeVars: ast.Identifiers{
																	"$std",
																	"a",
																	"aux",
																	"b",
																	"i",
																	"j",
																	"keyF",
																	"la",
																	"lb",
																	"prefix",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1429),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1437),
																		Column: int(43),
																	},
																},
															},
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1428),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1428),
																		Column: int(18),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "j",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1428),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1428),
																		Column: int(21),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "prefix",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1428),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1428),
																		Column: int(29),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: p15248,
															FreeVars: ast.Identifiers{
																"$std",
																"a",
																"aux",
																"b",
																"keyF",
																"la",
																"lb",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1428),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1437),
																	Column: int(43),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: nil,
													Variable: "aux",
													CloseFodder: nil,
													Fun: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Body: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p15253,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1438),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1438),
																Column: int(10),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15257,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1438),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1438),
																			Column: int(12),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15257,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1438),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(1438),
																			Column: int(15),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Array{
																Elements: nil,
																CloseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15257,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1438),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1438),
																			Column: int(19),
																		},
																	},
																},
																TrailingComma: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15253,
													FreeVars: ast.Identifiers{
														"aux",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1438),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1438),
															Column: int(20),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p15253,
												FreeVars: ast.Identifiers{
													"$std",
													"a",
													"b",
													"keyF",
													"la",
													"lb",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1428),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1438),
														Column: int(20),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p15253,
											FreeVars: ast.Identifiers{
												"$std",
												"a",
												"b",
												"keyF",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1427),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1438),
													Column: int(20),
												},
											},
										},
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "a",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1426),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1426),
													Column: int(18),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "b",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1426),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1426),
													Column: int(21),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: p15266,
										FreeVars: ast.Identifiers{
											"$std",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1426),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1438),
												Column: int(20),
											},
										},
									},
									TrailingComma: false,
								},
								EqFodder: nil,
								Variable: "merge",
								CloseFodder: nil,
								Fun: nil,
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1440),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1440),
															Column: int(18),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15275,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1440),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1440),
														Column: int(25),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15279,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1440),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1440),
																	Column: int(29),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15275,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1440),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1440),
													Column: int(30),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									EqFodder: ast.Fodder{},
									Variable: "l",
									CloseFodder: ast.Fodder{},
									Fun: nil,
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1440),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1440),
											Column: int(30),
										},
									},
								},
							},
							Body: &ast.Conditional{
								Cond: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "30",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15285,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1441),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(1441),
													Column: int(29),
												},
											},
										},
									},
									Left: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1441),
															Column: int(8),
														},
														End: ast.Location{
															Line: int(1441),
															Column: int(11),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15285,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1441),
														Column: int(8),
													},
													End: ast.Location{
														Line: int(1441),
														Column: int(18),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15294,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1441),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1441),
																	Column: int(22),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15285,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1441),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(1441),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p15285,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1441),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1441),
												Column: int(29),
											},
										},
									},
									Op: ast.BinaryOp(10),
								},
								BranchTrue: &ast.Apply{
									Target: &ast.Var{
										Id: "quickSort",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p15285,
											FreeVars: ast.Identifiers{
												"quickSort",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1442),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1442),
													Column: int(16),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15304,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1442),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1442),
																Column: int(20),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
										},
										Named: []ast.NamedArgument{
											ast.NamedArgument{
												NameFodder: ast.Fodder{},
												Name: "keyF",
												EqFodder: ast.Fodder{},
												Arg: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15304,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1442),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(1442),
																Column: int(31),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
										},
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p15285,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"quickSort",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1442),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1442),
												Column: int(32),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchFalse: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1444),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1444),
																	Column: int(22),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "floor",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15317,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1444),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(1444),
																Column: int(28),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "2",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15322,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1444),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1444),
																				Column: int(34),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "l",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15322,
																		FreeVars: ast.Identifiers{
																			"l",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1444),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1444),
																				Column: int(30),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15322,
																	FreeVars: ast.Identifiers{
																		"l",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1444),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1444),
																			Column: int(34),
																		},
																	},
																},
																Op: ast.BinaryOp(1),
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15317,
													FreeVars: ast.Identifiers{
														"l",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1444),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(1444),
															Column: int(35),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "mid",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1444),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1444),
													Column: int(35),
												},
											},
										},
									},
									Body: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "slice",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15337,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1445),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1445),
																				Column: int(23),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "mid",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15337,
																		FreeVars: ast.Identifiers{
																			"mid",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1445),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1445),
																				Column: int(28),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"arr",
															"mid",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1445),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1445),
																Column: int(29),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "left",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1445),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1445),
														Column: int(29),
													},
												},
											},
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "$std",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "slice",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: nil,
														LeftBracketFodder: nil,
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15352,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1445),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1445),
																				Column: int(42),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "mid",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15352,
																		FreeVars: ast.Identifiers{
																			"mid",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1445),
																				Column: int(43),
																			},
																			End: ast.Location{
																				Line: int(1445),
																				Column: int(46),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: nil,
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
															"arr",
															"mid",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1445),
																Column: int(39),
															},
															End: ast.Location{
																Line: int(1445),
																Column: int(48),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "right",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1445),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1445),
														Column: int(48),
													},
												},
											},
										},
										Body: &ast.Apply{
											Target: &ast.Var{
												Id: "merge",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
														},
													},
													Ctx: p15285,
													FreeVars: ast.Identifiers{
														"merge",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1446),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1446),
															Column: int(12),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1446),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1446),
																				Column: int(16),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "sort",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15369,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1446),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1446),
																			Column: int(21),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "left",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15373,
																				FreeVars: ast.Identifiers{
																					"left",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1446),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1446),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
																Named: []ast.NamedArgument{
																	ast.NamedArgument{
																		NameFodder: ast.Fodder{},
																		Name: "keyF",
																		EqFodder: ast.Fodder{},
																		Arg: &ast.Var{
																			Id: "keyF",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15373,
																				FreeVars: ast.Identifiers{
																					"keyF",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1446),
																						Column: int(33),
																					},
																					End: ast.Location{
																						Line: int(1446),
																						Column: int(37),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15369,
																FreeVars: ast.Identifiers{
																	"keyF",
																	"left",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1446),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1446),
																		Column: int(38),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1446),
																				Column: int(40),
																			},
																			End: ast.Location{
																				Line: int(1446),
																				Column: int(43),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "sort",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15369,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1446),
																			Column: int(40),
																		},
																		End: ast.Location{
																			Line: int(1446),
																			Column: int(48),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "right",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15387,
																				FreeVars: ast.Identifiers{
																					"right",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1446),
																						Column: int(49),
																					},
																					End: ast.Location{
																						Line: int(1446),
																						Column: int(54),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
																Named: []ast.NamedArgument{
																	ast.NamedArgument{
																		NameFodder: ast.Fodder{},
																		Name: "keyF",
																		EqFodder: ast.Fodder{},
																		Arg: &ast.Var{
																			Id: "keyF",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15387,
																				FreeVars: ast.Identifiers{
																					"keyF",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1446),
																						Column: int(61),
																					},
																					End: ast.Location{
																						Line: int(1446),
																						Column: int(65),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15369,
																FreeVars: ast.Identifiers{
																	"keyF",
																	"right",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1446),
																		Column: int(40),
																	},
																	End: ast.Location{
																		Line: int(1446),
																		Column: int(66),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15285,
												FreeVars: ast.Identifiers{
													"keyF",
													"left",
													"merge",
													"right",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1446),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1446),
														Column: int(67),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: p15285,
											FreeVars: ast.Identifiers{
												"$std",
												"arr",
												"keyF",
												"merge",
												"mid",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1445),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1446),
													Column: int(67),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p15285,
										FreeVars: ast.Identifiers{
											"$std",
											"arr",
											"keyF",
											"l",
											"merge",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1444),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1446),
												Column: int(67),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p15285,
									FreeVars: ast.Identifiers{
										"$std",
										"arr",
										"keyF",
										"l",
										"merge",
										"quickSort",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1441),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1446),
											Column: int(67),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
									},
								},
								Ctx: p15285,
								FreeVars: ast.Identifiers{
									"$std",
									"arr",
									"keyF",
									"merge",
									"quickSort",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1440),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1446),
										Column: int(67),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
								},
							},
							Ctx: p15285,
							FreeVars: ast.Identifiers{
								"$std",
								"arr",
								"keyF",
								"quickSort",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1426),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1446),
									Column: int(67),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p15285,
						FreeVars: ast.Identifiers{
							"$std",
							"arr",
							"id",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1414),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1446),
								Column: int(67),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1413),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1413),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15285,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1413),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(1413),
										Column: int(20),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1413),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1413),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1413),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1446),
					Column: int(67),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "uniq",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.LiteralNumber{
											OriginalString: "0",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15419,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1450),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1450),
														Column: int(28),
													},
												},
											},
										},
										Left: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1450),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1450),
																Column: int(13),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15419,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1450),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1450),
															Column: int(20),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "a",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15428,
																FreeVars: ast.Identifiers{
																	"a",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1450),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1450),
																		Column: int(22),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15419,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1450),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1450),
														Column: int(23),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15419,
											FreeVars: ast.Identifiers{
												"a",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1450),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1450),
													Column: int(28),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "b",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15435,
														FreeVars: ast.Identifiers{
															"b",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1451),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1451),
																Column: int(11),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p15419,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1451),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1451),
													Column: int(12),
												},
											},
										},
										TrailingComma: false,
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15419,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1452),
																Column: int(45),
															},
															End: ast.Location{
																Line: int(1452),
																Column: int(49),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15446,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1452),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(1452),
																			Column: int(51),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15419,
													FreeVars: ast.Identifiers{
														"b",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1452),
															Column: int(45),
														},
														End: ast.Location{
															Line: int(1452),
															Column: int(52),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Apply{
												Target: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15419,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1452),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1452),
																Column: int(19),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																Target: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15455,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1452),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1452),
																				Column: int(21),
																			},
																		},
																	},
																},
																Index: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15455,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1452),
																					Column: int(38),
																				},
																				End: ast.Location{
																					Line: int(1452),
																					Column: int(39),
																				},
																			},
																		},
																	},
																	Left: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1452),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(1452),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "length",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15455,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1452),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1452),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "a",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15467,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1452),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(1452),
																									Column: int(34),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15455,
																			FreeVars: ast.Identifiers{
																				"a",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1452),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1452),
																					Column: int(35),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15455,
																		FreeVars: ast.Identifiers{
																			"a",
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1452),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1452),
																				Column: int(39),
																			},
																		},
																	},
																	Op: ast.BinaryOp(4),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15455,
																	FreeVars: ast.Identifiers{
																		"a",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1452),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1452),
																			Column: int(40),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15419,
													FreeVars: ast.Identifiers{
														"a",
														"keyF",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1452),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1452),
															Column: int(41),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15419,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"keyF",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1452),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1452),
														Column: int(52),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p15419,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1453),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1453),
														Column: int(10),
													},
												},
											},
										},
										BranchFalse: &ast.Binary{
											Right: &ast.Array{
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15481,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1455),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(1455),
																		Column: int(15),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												CloseFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15419,
													FreeVars: ast.Identifiers{
														"b",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1455),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1455),
															Column: int(16),
														},
													},
												},
												TrailingComma: false,
											},
											Left: &ast.Var{
												Id: "a",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p15419,
													FreeVars: ast.Identifiers{
														"a",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1455),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1455),
															Column: int(10),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15419,
												FreeVars: ast.Identifiers{
													"a",
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1455),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1455),
														Column: int(16),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15419,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"keyF",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1452),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1455),
													Column: int(16),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p15419,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1450),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1455),
												Column: int(16),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1449),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1449),
												Column: int(14),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1449),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1449),
												Column: int(17),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p15494,
									FreeVars: ast.Identifiers{
										"keyF",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1449),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1455),
											Column: int(16),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "f",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1456),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1456),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "foldl",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15502,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1456),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1456),
										Column: int(14),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "f",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15506,
											FreeVars: ast.Identifiers{
												"f",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1456),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1456),
													Column: int(16),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15506,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1456),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1456),
													Column: int(21),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15506,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1456),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1456),
													Column: int(25),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p15502,
							FreeVars: ast.Identifiers{
								"arr",
								"f",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1456),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1456),
									Column: int(26),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p15502,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1449),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1456),
								Column: int(26),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1448),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1448),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15502,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1448),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(1448),
										Column: int(20),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1448),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1448),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1448),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1456),
					Column: int(26),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "set",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1459),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1459),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "uniq",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p15526,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1459),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1459),
									Column: int(13),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1459),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(1459),
														Column: int(17),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "sort",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15534,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1459),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1459),
													Column: int(22),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15538,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1459),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1459),
																Column: int(26),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15538,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1459),
																Column: int(28),
															},
															End: ast.Location{
																Line: int(1459),
																Column: int(32),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p15534,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1459),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(1459),
												Column: int(33),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "keyF",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p15534,
										FreeVars: ast.Identifiers{
											"keyF",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1459),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(1459),
												Column: int(39),
											},
										},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p15526,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1459),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1459),
								Column: int(40),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1458),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1458),
								Column: int(10),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15526,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1458),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(1458),
										Column: int(19),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1458),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1458),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1458),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1459),
					Column: int(40),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "setMember",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralNumber{
						OriginalString: "0",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p15554,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1463),
									Column: int(48),
								},
								End: ast.Location{
									Line: int(1463),
									Column: int(49),
								},
							},
						},
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
										ast.FodderElement{
											Comment: []string{
												"// TODO(dcunnin): Binary chop for O(log n) complexity",
											},
											Kind: ast.FodderKind(2),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1463),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1463),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "length",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15554,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1463),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1463),
										Column: int(15),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1463),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(1463),
															Column: int(19),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "setInter",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15569,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1463),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1463),
														Column: int(28),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Array{
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "x",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15575,
																		FreeVars: ast.Identifiers{
																			"x",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1463),
																				Column: int(30),
																			},
																			End: ast.Location{
																				Line: int(1463),
																				Column: int(31),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														CloseFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15577,
															FreeVars: ast.Identifiers{
																"x",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1463),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1463),
																	Column: int(32),
																},
															},
														},
														TrailingComma: false,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "arr",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15577,
															FreeVars: ast.Identifiers{
																"arr",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1463),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(1463),
																	Column: int(37),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "keyF",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15577,
															FreeVars: ast.Identifiers{
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1463),
																	Column: int(39),
																},
																End: ast.Location{
																	Line: int(1463),
																	Column: int(43),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15569,
											FreeVars: ast.Identifiers{
												"arr",
												"keyF",
												"std",
												"x",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1463),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1463),
													Column: int(44),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p15554,
							FreeVars: ast.Identifiers{
								"arr",
								"keyF",
								"std",
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1463),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1463),
									Column: int(45),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p15554,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1463),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1463),
								Column: int(49),
							},
						},
					},
					Op: ast.BinaryOp(7),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1461),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1461),
								Column: int(14),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1461),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1461),
								Column: int(19),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15554,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1461),
										Column: int(26),
									},
									End: ast.Location{
										Line: int(1461),
										Column: int(28),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1461),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1461),
								Column: int(28),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1461),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1463),
					Column: int(49),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "setUnion",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1468),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1468),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15602,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1468),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1468),
															Column: int(25),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "a",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15606,
																FreeVars: ast.Identifiers{
																	"a",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1468),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(1468),
																		Column: int(27),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1468),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1468),
														Column: int(28),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1468),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1468),
														Column: int(11),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15602,
											FreeVars: ast.Identifiers{
												"a",
												"i",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1468),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1468),
													Column: int(28),
												},
											},
										},
										Op: ast.BinaryOp(8),
									},
									BranchTrue: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "$std",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "slice",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: nil,
												LeftBracketFodder: nil,
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1469),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1469),
																		Column: int(16),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "j",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"j",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1469),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1469),
																		Column: int(18),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: nil,
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
													"b",
													"j",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1469),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1469),
														Column: int(20),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "acc",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"acc",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1469),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1469),
														Column: int(12),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15602,
											FreeVars: ast.Identifiers{
												"$std",
												"acc",
												"b",
												"j",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1469),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1469),
													Column: int(20),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1470),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1470),
																	Column: int(23),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15602,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1470),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1470),
																Column: int(30),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15641,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1470),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1470),
																			Column: int(32),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15602,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1470),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1470),
															Column: int(33),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "j",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15602,
													FreeVars: ast.Identifiers{
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1470),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1470),
															Column: int(16),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1470),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1470),
														Column: int(33),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										BranchTrue: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "$std",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "slice",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: nil,
													LeftBracketFodder: nil,
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1471),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1471),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1471),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1471),
																			Column: int(18),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: nil,
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
														"a",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1471),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1471),
															Column: int(20),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "acc",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p15602,
													FreeVars: ast.Identifiers{
														"acc",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1471),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1471),
															Column: int(12),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"$std",
													"a",
													"acc",
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1471),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1471),
														Column: int(20),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										BranchFalse: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.Apply{
														Target: &ast.Var{
															Id: "keyF",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15670,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1473),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1473),
																		Column: int(24),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "a",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15675,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1473),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1473),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15675,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1473),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(1473),
																						Column: int(28),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15675,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1473),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1473),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15670,
															FreeVars: ast.Identifiers{
																"a",
																"i",
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1473),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1473),
																	Column: int(30),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													EqFodder: ast.Fodder{},
													Variable: "ak",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1473),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1473),
															Column: int(30),
														},
													},
												},
											},
											Body: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Apply{
															Target: &ast.Var{
																Id: "keyF",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15685,
																	FreeVars: ast.Identifiers{
																		"keyF",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1474),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1474),
																			Column: int(24),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			Target: &ast.Var{
																				Id: "b",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15690,
																					FreeVars: ast.Identifiers{
																						"b",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1474),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1474),
																							Column: int(26),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "j",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15690,
																					FreeVars: ast.Identifiers{
																						"j",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1474),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1474),
																							Column: int(28),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15690,
																				FreeVars: ast.Identifiers{
																					"b",
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1474),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1474),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15685,
																FreeVars: ast.Identifiers{
																	"b",
																	"j",
																	"keyF",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1474),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1474),
																		Column: int(30),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														EqFodder: ast.Fodder{},
														Variable: "bk",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1474),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1474),
																Column: int(30),
															},
														},
													},
												},
												Body: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.Var{
															Id: "bk",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"bk",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1475),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1475),
																		Column: int(20),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "ak",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"ak",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1475),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1475),
																		Column: int(14),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15602,
															FreeVars: ast.Identifiers{
																"ak",
																"bk",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1475),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1475),
																	Column: int(20),
																},
															},
														},
														Op: ast.BinaryOp(12),
													},
													BranchTrue: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1476),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1476),
																		Column: int(14),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "a",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15709,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1476),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1476),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "b",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15709,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1476),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1476),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "1",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(26),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15709,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1476),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1476),
																					Column: int(26),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "1",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(33),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "j",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15709,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1476),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1476),
																					Column: int(33),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Array{
																			Elements: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Index{
																						Target: &ast.Var{
																							Id: "a",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15728,
																								FreeVars: ast.Identifiers{
																									"a",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1476),
																										Column: int(42),
																									},
																									End: ast.Location{
																										Line: int(1476),
																										Column: int(43),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15728,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1476),
																										Column: int(44),
																									},
																									End: ast.Location{
																										Line: int(1476),
																										Column: int(45),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p15728,
																							FreeVars: ast.Identifiers{
																								"a",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1476),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1476),
																									Column: int(46),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			CloseFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(41),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(47),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		Left: &ast.Var{
																			Id: "acc",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15709,
																				FreeVars: ast.Identifiers{
																					"acc",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1476),
																						Column: int(35),
																					},
																					End: ast.Location{
																						Line: int(1476),
																						Column: int(38),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15709,
																			FreeVars: ast.Identifiers{
																				"a",
																				"acc",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1476),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1476),
																					Column: int(47),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15602,
															FreeVars: ast.Identifiers{
																"a",
																"acc",
																"aux",
																"b",
																"i",
																"j",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1476),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1476),
																	Column: int(48),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													BranchFalse: &ast.Conditional{
														Cond: &ast.Binary{
															Right: &ast.Var{
																Id: "bk",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"bk",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1477),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1477),
																			Column: int(24),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "ak",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"ak",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1477),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1477),
																			Column: int(19),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"ak",
																	"bk",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1477),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1477),
																		Column: int(24),
																	},
																},
															},
															Op: ast.BinaryOp(9),
														},
														BranchTrue: &ast.Apply{
															Target: &ast.Var{
																Id: "aux",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1478),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1478),
																			Column: int(14),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "a",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15751,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1478),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1478),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "b",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15751,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1478),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1478),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15751,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1478),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1478),
																							Column: int(26),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15751,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1478),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1478),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15751,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1478),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1478),
																						Column: int(26),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "j",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15751,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1478),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1478),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							Target: &ast.Var{
																								Id: "a",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15767,
																									FreeVars: ast.Identifiers{
																										"a",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1478),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(1478),
																											Column: int(39),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15767,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1478),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(1478),
																											Column: int(41),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15767,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1478),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1478),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15751,
																					FreeVars: ast.Identifiers{
																						"a",
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1478),
																							Column: int(37),
																						},
																						End: ast.Location{
																							Line: int(1478),
																							Column: int(43),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			Left: &ast.Var{
																				Id: "acc",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15751,
																					FreeVars: ast.Identifiers{
																						"acc",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1478),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1478),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15751,
																				FreeVars: ast.Identifiers{
																					"a",
																					"acc",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1478),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1478),
																						Column: int(43),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"aux",
																	"b",
																	"i",
																	"j",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1478),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1478),
																		Column: int(44),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: true,
														},
														BranchFalse: &ast.Apply{
															Target: &ast.Var{
																Id: "aux",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p15602,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1480),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1480),
																			Column: int(14),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "a",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15783,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1480),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1480),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "b",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15783,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1480),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1480),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15783,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1480),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1480),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15783,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1480),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1480),
																							Column: int(29),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "j",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15783,
																					FreeVars: ast.Identifiers{
																						"j",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1480),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1480),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15783,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1480),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1480),
																						Column: int(29),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.Array{
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							Target: &ast.Var{
																								Id: "b",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15799,
																									FreeVars: ast.Identifiers{
																										"b",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1480),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(1480),
																											Column: int(39),
																										},
																									},
																								},
																							},
																							Index: &ast.Var{
																								Id: "j",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p15799,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1480),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(1480),
																											Column: int(41),
																										},
																									},
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p15799,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1480),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1480),
																										Column: int(42),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				CloseFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15783,
																					FreeVars: ast.Identifiers{
																						"b",
																						"j",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1480),
																							Column: int(37),
																						},
																						End: ast.Location{
																							Line: int(1480),
																							Column: int(43),
																						},
																					},
																				},
																				TrailingComma: false,
																			},
																			Left: &ast.Var{
																				Id: "acc",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15783,
																					FreeVars: ast.Identifiers{
																						"acc",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1480),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1480),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15783,
																				FreeVars: ast.Identifiers{
																					"acc",
																					"b",
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1480),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1480),
																						Column: int(43),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15602,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"aux",
																	"b",
																	"i",
																	"j",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1480),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1480),
																		Column: int(44),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: true,
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15602,
															FreeVars: ast.Identifiers{
																"a",
																"acc",
																"ak",
																"aux",
																"b",
																"bk",
																"i",
																"j",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1477),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(1480),
																	Column: int(44),
																},
															},
														},
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p15602,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"ak",
															"aux",
															"b",
															"bk",
															"i",
															"j",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1475),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1480),
																Column: int(44),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p15602,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"ak",
														"aux",
														"b",
														"i",
														"j",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1474),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1480),
															Column: int(44),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p15602,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1473),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1480),
														Column: int(44),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15602,
											FreeVars: ast.Identifiers{
												"$std",
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1470),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1480),
													Column: int(44),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p15602,
										FreeVars: ast.Identifiers{
											"$std",
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1468),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1480),
												Column: int(44),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1467),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1467),
												Column: int(16),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1467),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1467),
												Column: int(19),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1467),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1467),
												Column: int(22),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1467),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1467),
												Column: int(25),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1467),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1467),
												Column: int(30),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p15824,
									FreeVars: ast.Identifiers{
										"$std",
										"aux",
										"keyF",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1467),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1480),
											Column: int(44),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p15829,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1481),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1481),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "a",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15833,
											FreeVars: ast.Identifiers{
												"a",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1481),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1481),
													Column: int(10),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "b",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15833,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1481),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1481),
													Column: int(13),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15833,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1481),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1481),
													Column: int(16),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15833,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1481),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1481),
													Column: int(19),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15833,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1481),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1481),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p15829,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1481),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1481),
									Column: int(24),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
							ast.FodderElement{
								Comment: []string{
									"// NOTE: order matters, values in `a` win",
								},
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p15829,
						FreeVars: ast.Identifiers{
							"$std",
							"a",
							"b",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1467),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1481),
								Column: int(24),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1465),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1465),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1465),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1465),
								Column: int(16),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p15829,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1465),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(1465),
										Column: int(25),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1465),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1465),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1465),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1481),
					Column: int(24),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "setInter",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1485),
																	Column: int(37),
																},
																End: ast.Location{
																	Line: int(1485),
																	Column: int(40),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1485),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(1485),
																Column: int(47),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15865,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1485),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(1485),
																			Column: int(49),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1485),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(1485),
															Column: int(50),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "j",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1485),
															Column: int(32),
														},
														End: ast.Location{
															Line: int(1485),
															Column: int(33),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15861,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1485),
														Column: int(32),
													},
													End: ast.Location{
														Line: int(1485),
														Column: int(50),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										Left: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1485),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1485),
																	Column: int(18),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1485),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1485),
																Column: int(25),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15880,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1485),
																			Column: int(26),
																		},
																		End: ast.Location{
																			Line: int(1485),
																			Column: int(27),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1485),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1485),
															Column: int(28),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "i",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1485),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1485),
															Column: int(11),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15861,
												FreeVars: ast.Identifiers{
													"a",
													"i",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1485),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1485),
														Column: int(28),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p15861,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"i",
												"j",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1485),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1485),
													Column: int(50),
												},
											},
										},
										Op: ast.BinaryOp(18),
									},
									BranchTrue: &ast.Var{
										Id: "acc",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p15861,
											FreeVars: ast.Identifiers{
												"acc",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1486),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1486),
													Column: int(12),
												},
											},
										},
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1488),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1488),
																Column: int(30),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																Target: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15898,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1488),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1488),
																				Column: int(32),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "j",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15898,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1488),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1488),
																				Column: int(34),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15898,
																	FreeVars: ast.Identifiers{
																		"b",
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1488),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1488),
																			Column: int(35),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"b",
														"j",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1488),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1488),
															Column: int(36),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Apply{
												Target: &ast.Var{
													Id: "keyF",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1488),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1488),
																Column: int(16),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																Target: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15910,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1488),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1488),
																				Column: int(18),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15910,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1488),
																				Column: int(19),
																			},
																			End: ast.Location{
																				Line: int(1488),
																				Column: int(20),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15910,
																	FreeVars: ast.Identifiers{
																		"a",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1488),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1488),
																			Column: int(21),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"a",
														"i",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1488),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1488),
															Column: int(22),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15861,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"i",
													"j",
													"keyF",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1488),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1488),
														Column: int(36),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										BranchTrue: &ast.Apply{
											Target: &ast.Var{
												Id: "aux",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
														},
													},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"aux",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1489),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1489),
															Column: int(14),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "a",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15923,
																FreeVars: ast.Identifiers{
																	"a",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1489),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1489),
																		Column: int(16),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15923,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1489),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1489),
																		Column: int(19),
																	},
																},
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(26),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(22),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15923,
																FreeVars: ast.Identifiers{
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1489),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1489),
																		Column: int(26),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(33),
																		},
																	},
																},
															},
															Left: &ast.Var{
																Id: "j",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(29),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15923,
																FreeVars: ast.Identifiers{
																	"j",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1489),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(1489),
																		Column: int(33),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.Array{
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			Target: &ast.Var{
																				Id: "a",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15942,
																					FreeVars: ast.Identifiers{
																						"a",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1489),
																							Column: int(42),
																						},
																						End: ast.Location{
																							Line: int(1489),
																							Column: int(43),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p15942,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1489),
																							Column: int(44),
																						},
																						End: ast.Location{
																							Line: int(1489),
																							Column: int(45),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p15942,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1489),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(1489),
																						Column: int(46),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																CloseFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{
																		"a",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(41),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(47),
																		},
																	},
																},
																TrailingComma: false,
															},
															Left: &ast.Var{
																Id: "acc",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15923,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1489),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1489),
																			Column: int(38),
																		},
																	},
																},
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p15923,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"i",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1489),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(1489),
																		Column: int(47),
																	},
																},
															},
															Op: ast.BinaryOp(3),
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15861,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1489),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1489),
														Column: int(48),
													},
												},
											},
											TrailingComma: false,
											TailStrict: true,
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Var{
														Id: "keyF",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15861,
															FreeVars: ast.Identifiers{
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1490),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(1490),
																	Column: int(34),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "b",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15960,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1490),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1490),
																					Column: int(36),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15960,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1490),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1490),
																					Column: int(38),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15960,
																		FreeVars: ast.Identifiers{
																			"b",
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1490),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1490),
																				Column: int(39),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"b",
															"j",
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1490),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1490),
																Column: int(40),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "keyF",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p15861,
															FreeVars: ast.Identifiers{
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1490),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1490),
																	Column: int(21),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "a",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15972,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1490),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1490),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p15972,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1490),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1490),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15972,
																		FreeVars: ast.Identifiers{
																			"a",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1490),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1490),
																				Column: int(26),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"a",
															"i",
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1490),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1490),
																Column: int(27),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"i",
														"j",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1490),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1490),
															Column: int(40),
														},
													},
												},
												Op: ast.BinaryOp(9),
											},
											BranchTrue: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1491),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1491),
																Column: int(14),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15985,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1491),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1491),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15985,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1491),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1491),
																			Column: int(19),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15985,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1491),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1491),
																				Column: int(26),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p15985,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1491),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1491),
																				Column: int(22),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15985,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1491),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1491),
																			Column: int(26),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "j",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15985,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1491),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1491),
																			Column: int(29),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "acc",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p15985,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1491),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1491),
																			Column: int(34),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1491),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1491),
															Column: int(35),
														},
													},
												},
												TrailingComma: false,
												TailStrict: true,
											},
											BranchFalse: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p15861,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1493),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1493),
																Column: int(14),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16005,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1493),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1493),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16005,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1493),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1493),
																			Column: int(19),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16005,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1493),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1493),
																			Column: int(22),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16005,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1493),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1493),
																				Column: int(29),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "j",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16005,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1493),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1493),
																				Column: int(25),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16005,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1493),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(1493),
																			Column: int(29),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "acc",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16005,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1493),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1493),
																			Column: int(34),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p15861,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1493),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1493),
															Column: int(35),
														},
													},
												},
												TrailingComma: false,
												TailStrict: true,
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p15861,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1490),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(1493),
														Column: int(35),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p15861,
											FreeVars: ast.Identifiers{
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1488),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1493),
													Column: int(35),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p15861,
										FreeVars: ast.Identifiers{
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1485),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1493),
												Column: int(35),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1484),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1484),
												Column: int(16),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1484),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1484),
												Column: int(19),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1484),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1484),
												Column: int(22),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1484),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1484),
												Column: int(25),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1484),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1484),
												Column: int(30),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p16028,
									FreeVars: ast.Identifiers{
										"aux",
										"keyF",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1484),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1493),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p16033,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1494),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1494),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "a",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16037,
											FreeVars: ast.Identifiers{
												"a",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1494),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1494),
													Column: int(10),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "b",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16037,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1494),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1494),
													Column: int(13),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16037,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1494),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1494),
													Column: int(16),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16037,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1494),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1494),
													Column: int(19),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16037,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1494),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1494),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16033,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1494),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1494),
									Column: int(24),
								},
							},
						},
						TrailingComma: false,
						TailStrict: true,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p16033,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1484),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1494),
								Column: int(24),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1483),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1483),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1483),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1483),
								Column: int(16),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16033,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1483),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(1483),
										Column: int(25),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1483),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1483),
								Column: int(25),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1483),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1494),
					Column: int(24),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "setDiff",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1498),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1498),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1498),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1498),
															Column: int(25),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "a",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16067,
																FreeVars: ast.Identifiers{
																	"a",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1498),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(1498),
																		Column: int(27),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16063,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1498),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1498),
														Column: int(28),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										Left: &ast.Var{
											Id: "i",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16063,
												FreeVars: ast.Identifiers{
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1498),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1498),
														Column: int(11),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16063,
											FreeVars: ast.Identifiers{
												"a",
												"i",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1498),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1498),
													Column: int(28),
												},
											},
										},
										Op: ast.BinaryOp(8),
									},
									BranchTrue: &ast.Var{
										Id: "acc",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p16063,
											FreeVars: ast.Identifiers{
												"acc",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1499),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1499),
													Column: int(12),
												},
											},
										},
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1500),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1500),
																	Column: int(23),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1500),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1500),
																Column: int(30),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16086,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1500),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1500),
																			Column: int(32),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1500),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1500),
															Column: int(33),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "j",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1500),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1500),
															Column: int(16),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16063,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1500),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1500),
														Column: int(33),
													},
												},
											},
											Op: ast.BinaryOp(8),
										},
										BranchTrue: &ast.Binary{
											Right: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "$std",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "slice",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: nil,
													LeftBracketFodder: nil,
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16063,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1501),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1501),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "i",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16063,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1501),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1501),
																			Column: int(18),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: nil,
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
														"a",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1501),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1501),
															Column: int(20),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											Left: &ast.Var{
												Id: "acc",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"acc",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1501),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1501),
															Column: int(12),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16063,
												FreeVars: ast.Identifiers{
													"$std",
													"a",
													"acc",
													"i",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1501),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1501),
														Column: int(20),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Apply{
													Target: &ast.Var{
														Id: "keyF",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1503),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1503),
																	Column: int(30),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "b",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16119,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1503),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1503),
																					Column: int(32),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16119,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1503),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1503),
																					Column: int(34),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16119,
																		FreeVars: ast.Identifiers{
																			"b",
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1503),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1503),
																				Column: int(35),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"b",
															"j",
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1503),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1503),
																Column: int(36),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												Left: &ast.Apply{
													Target: &ast.Var{
														Id: "keyF",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1503),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1503),
																	Column: int(16),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	Target: &ast.Var{
																		Id: "a",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16131,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1503),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(1503),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	Index: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16131,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1503),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(1503),
																					Column: int(20),
																				},
																			},
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16131,
																		FreeVars: ast.Identifiers{
																			"a",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1503),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1503),
																				Column: int(21),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"a",
															"i",
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1503),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1503),
																Column: int(22),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"i",
														"j",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1503),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1503),
															Column: int(36),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											BranchTrue: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1504),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1504),
																Column: int(14),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16144,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1504),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1504),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16144,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1504),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1504),
																			Column: int(19),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16144,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1504),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1504),
																				Column: int(26),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16144,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1504),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1504),
																				Column: int(22),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16144,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1504),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1504),
																			Column: int(26),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																Right: &ast.LiteralNumber{
																	OriginalString: "1",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16144,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1504),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(1504),
																				Column: int(33),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "j",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16144,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1504),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1504),
																				Column: int(29),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16144,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1504),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1504),
																			Column: int(33),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "acc",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16144,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1504),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1504),
																			Column: int(38),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1504),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1504),
															Column: int(39),
														},
													},
												},
												TrailingComma: false,
												TailStrict: true,
											},
											BranchFalse: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Var{
															Id: "keyF",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16063,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1505),
																		Column: int(30),
																	},
																	End: ast.Location{
																		Line: int(1505),
																		Column: int(34),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "b",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16169,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1505),
																						Column: int(35),
																					},
																					End: ast.Location{
																						Line: int(1505),
																						Column: int(36),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "j",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16169,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1505),
																						Column: int(37),
																					},
																					End: ast.Location{
																						Line: int(1505),
																						Column: int(38),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16169,
																			FreeVars: ast.Identifiers{
																				"b",
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1505),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1505),
																					Column: int(39),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"b",
																"j",
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1505),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(1505),
																	Column: int(40),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Apply{
														Target: &ast.Var{
															Id: "keyF",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16063,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1505),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1505),
																		Column: int(21),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		Target: &ast.Var{
																			Id: "a",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16181,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1505),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1505),
																						Column: int(23),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16181,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1505),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1505),
																						Column: int(25),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16181,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1505),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1505),
																					Column: int(26),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"a",
																"i",
																"keyF",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1505),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1505),
																	Column: int(27),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"a",
															"b",
															"i",
															"j",
															"keyF",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1505),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1505),
																Column: int(40),
															},
														},
													},
													Op: ast.BinaryOp(9),
												},
												BranchTrue: &ast.Apply{
													Target: &ast.Var{
														Id: "aux",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																},
															},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"aux",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1506),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1506),
																	Column: int(14),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16194,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1506),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1506),
																				Column: int(16),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16194,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1506),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1506),
																				Column: int(19),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16194,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1506),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1506),
																					Column: int(26),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "i",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16194,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1506),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1506),
																					Column: int(22),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16194,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1506),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1506),
																				Column: int(26),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "j",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16194,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1506),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1506),
																				Column: int(29),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					Target: &ast.Var{
																						Id: "a",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16210,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1506),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1506),
																									Column: int(39),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16210,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1506),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1506),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p16210,
																						FreeVars: ast.Identifiers{
																							"a",
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1506),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1506),
																								Column: int(42),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16194,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1506),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1506),
																					Column: int(43),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	Left: &ast.Var{
																		Id: "acc",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16194,
																			FreeVars: ast.Identifiers{
																				"acc",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1506),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1506),
																					Column: int(34),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16194,
																		FreeVars: ast.Identifiers{
																			"a",
																			"acc",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1506),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1506),
																				Column: int(43),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"aux",
															"b",
															"i",
															"j",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1506),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1506),
																Column: int(44),
															},
														},
													},
													TrailingComma: false,
													TailStrict: true,
												},
												BranchFalse: &ast.Apply{
													Target: &ast.Var{
														Id: "aux",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																},
															},
															Ctx: p16063,
															FreeVars: ast.Identifiers{
																"aux",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1508),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1508),
																	Column: int(14),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16226,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1508),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1508),
																				Column: int(16),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16226,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1508),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1508),
																				Column: int(19),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16226,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1508),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1508),
																				Column: int(22),
																			},
																		},
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "1",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16226,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1508),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1508),
																					Column: int(29),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "j",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16226,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1508),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1508),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16226,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1508),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1508),
																				Column: int(29),
																			},
																		},
																	},
																	Op: ast.BinaryOp(3),
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "acc",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16226,
																		FreeVars: ast.Identifiers{
																			"acc",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1508),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1508),
																				Column: int(34),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16063,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"aux",
															"b",
															"i",
															"j",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1508),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1508),
																Column: int(35),
															},
														},
													},
													TrailingComma: false,
													TailStrict: true,
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16063,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
														"keyF",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1505),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(1508),
															Column: int(35),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p16063,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1503),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1508),
														Column: int(35),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16063,
											FreeVars: ast.Identifiers{
												"$std",
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1500),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1508),
													Column: int(35),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p16063,
										FreeVars: ast.Identifiers{
											"$std",
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1498),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1508),
												Column: int(35),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1497),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1497),
												Column: int(16),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1497),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1497),
												Column: int(19),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1497),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1497),
												Column: int(22),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										CommaFodder: ast.Fodder{},
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1497),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1497),
												Column: int(25),
											},
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1497),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1497),
												Column: int(30),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p16251,
									FreeVars: ast.Identifiers{
										"$std",
										"aux",
										"keyF",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1497),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1508),
											Column: int(35),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "aux",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Var{
							Id: "aux",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p16256,
								FreeVars: ast.Identifiers{
									"aux",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1509),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1509),
										Column: int(8),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "a",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16260,
											FreeVars: ast.Identifiers{
												"a",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1509),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1509),
													Column: int(10),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "b",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16260,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1509),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1509),
													Column: int(13),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16260,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1509),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1509),
													Column: int(16),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										OriginalString: "0",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16260,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1509),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1509),
													Column: int(19),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16260,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1509),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1509),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16256,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1509),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1509),
									Column: int(24),
								},
							},
						},
						TrailingComma: false,
						TailStrict: true,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p16256,
						FreeVars: ast.Identifiers{
							"$std",
							"a",
							"b",
							"keyF",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1497),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1509),
								Column: int(24),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1496),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1496),
								Column: int(12),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1496),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1496),
								Column: int(15),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							Id: "id",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16256,
								FreeVars: ast.Identifiers{
									"id",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1496),
										Column: int(22),
									},
									End: ast.Location{
										Line: int(1496),
										Column: int(24),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1496),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(1496),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"id",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1496),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1509),
					Column: int(24),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "mergePatch",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1512),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1512),
											Column: int(11),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isObject",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16282,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1512),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1512),
										Column: int(20),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "patch",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16286,
											FreeVars: ast.Identifiers{
												"patch",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1512),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1512),
													Column: int(26),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16282,
							FreeVars: ast.Identifiers{
								"patch",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1512),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1512),
									Column: int(27),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1514),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1514),
															Column: int(15),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "isObject",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16297,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1514),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1514),
														Column: int(24),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "target",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16301,
															FreeVars: ast.Identifiers{
																"target",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1514),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(1514),
																	Column: int(31),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16297,
											FreeVars: ast.Identifiers{
												"std",
												"target",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1514),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1514),
													Column: int(32),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchTrue: &ast.Var{
										Id: "target",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16297,
											FreeVars: ast.Identifiers{
												"target",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1514),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(1514),
													Column: int(44),
												},
											},
										},
									},
									BranchFalse: &ast.DesugaredObject{
										Asserts: ast.Nodes{},
										Fields: ast.DesugaredObjectFields{},
										Locals: ast.LocalBinds{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16297,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1514),
													Column: int(50),
												},
												End: ast.Location{
													Line: int(1514),
													Column: int(52),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p16297,
										FreeVars: ast.Identifiers{
											"std",
											"target",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1514),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1514),
												Column: int(52),
											},
										},
									},
								},
								EqFodder: ast.Fodder{},
								Variable: "target_object",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1513),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1514),
										Column: int(52),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Body: &ast.Conditional{
										Cond: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1517),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1517),
																Column: int(15),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "isObject",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16317,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1517),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1517),
															Column: int(24),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "target_object",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16321,
																FreeVars: ast.Identifiers{
																	"target_object",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1517),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1517),
																		Column: int(38),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16317,
												FreeVars: ast.Identifiers{
													"std",
													"target_object",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1517),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1517),
														Column: int(39),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchTrue: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1517),
																Column: int(45),
															},
															End: ast.Location{
																Line: int(1517),
																Column: int(48),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "objectFields",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16317,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1517),
															Column: int(45),
														},
														End: ast.Location{
															Line: int(1517),
															Column: int(61),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "target_object",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16332,
																FreeVars: ast.Identifiers{
																	"target_object",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1517),
																		Column: int(62),
																	},
																	End: ast.Location{
																		Line: int(1517),
																		Column: int(75),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16317,
												FreeVars: ast.Identifiers{
													"std",
													"target_object",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1517),
														Column: int(45),
													},
													End: ast.Location{
														Line: int(1517),
														Column: int(76),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchFalse: &ast.Array{
											Elements: nil,
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16317,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1517),
														Column: int(82),
													},
													End: ast.Location{
														Line: int(1517),
														Column: int(84),
													},
												},
											},
											TrailingComma: false,
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p16317,
											FreeVars: ast.Identifiers{
												"std",
												"target_object",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1517),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1517),
													Column: int(84),
												},
											},
										},
									},
									EqFodder: ast.Fodder{},
									Variable: "target_fields",
									CloseFodder: ast.Fodder{},
									Fun: nil,
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1516),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1517),
											Column: int(84),
										},
									},
								},
							},
							Body: &ast.Local{
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "$std",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "flatMap",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: nil,
												LeftBracketFodder: nil,
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: nil,
															ParenRightFodder: nil,
															Body: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.LiteralNull{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16351,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1519),
																					Column: int(78),
																				},
																				End: ast.Location{
																					Line: int(1519),
																					Column: int(82),
																				},
																			},
																		},
																	},
																	Left: &ast.Index{
																		Target: &ast.Var{
																			Id: "patch",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16351,
																				FreeVars: ast.Identifiers{
																					"patch",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1519),
																						Column: int(66),
																					},
																					End: ast.Location{
																						Line: int(1519),
																						Column: int(71),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "k",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16351,
																				FreeVars: ast.Identifiers{
																					"k",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1519),
																						Column: int(72),
																					},
																					End: ast.Location{
																						Line: int(1519),
																						Column: int(73),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16351,
																			FreeVars: ast.Identifiers{
																				"k",
																				"patch",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1519),
																					Column: int(66),
																				},
																				End: ast.Location{
																					Line: int(1519),
																					Column: int(74),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16351,
																		FreeVars: ast.Identifiers{
																			"k",
																			"patch",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1519),
																				Column: int(66),
																			},
																			End: ast.Location{
																				Line: int(1519),
																				Column: int(82),
																			},
																		},
																	},
																	Op: ast.BinaryOp(12),
																},
																BranchTrue: &ast.Array{
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "k",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p16362,
																					FreeVars: ast.Identifiers{
																						"k",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1519),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1519),
																							Column: int(29),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	CloseFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"k",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																BranchFalse: &ast.Array{
																	Elements: nil,
																	CloseFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	TrailingComma: false,
																},
																ThenFodder: nil,
																ElseFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"k",
																		"patch",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "k",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"patch",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1519),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1519),
																				Column: int(42),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "objectFields",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16351,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1519),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(1519),
																			Column: int(55),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "patch",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16377,
																				FreeVars: ast.Identifiers{
																					"patch",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1519),
																						Column: int(56),
																					},
																					End: ast.Location{
																						Line: int(1519),
																						Column: int(61),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16351,
																FreeVars: ast.Identifiers{
																	"patch",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1519),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(1519),
																		Column: int(62),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: nil,
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
													"patch",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1519),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1519),
														Column: int(83),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										EqFodder: ast.Fodder{},
										Variable: "null_fields",
										CloseFodder: ast.Fodder{},
										Fun: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1519),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1519),
												Column: int(83),
											},
										},
									},
								},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1520),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(1520),
																	Column: int(30),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "setUnion",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16388,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1520),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(1520),
																Column: int(39),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "target_fields",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16392,
																	FreeVars: ast.Identifiers{
																		"target_fields",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1520),
																			Column: int(40),
																		},
																		End: ast.Location{
																			Line: int(1520),
																			Column: int(53),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1520),
																					Column: int(55),
																				},
																				End: ast.Location{
																					Line: int(1520),
																					Column: int(58),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "objectFields",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16392,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1520),
																				Column: int(55),
																			},
																			End: ast.Location{
																				Line: int(1520),
																				Column: int(71),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "patch",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p16402,
																					FreeVars: ast.Identifiers{
																						"patch",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1520),
																							Column: int(72),
																						},
																						End: ast.Location{
																							Line: int(1520),
																							Column: int(77),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16392,
																	FreeVars: ast.Identifiers{
																		"patch",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1520),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(1520),
																			Column: int(78),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16388,
													FreeVars: ast.Identifiers{
														"patch",
														"std",
														"target_fields",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1520),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1520),
															Column: int(79),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "both_fields",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1520),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1520),
													Column: int(79),
												},
											},
										},
									},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "$std",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$std",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "$objectFlatMerge",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: nil,
											LeftBracketFodder: nil,
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "$std",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "flatMap",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: nil,
															LeftBracketFodder: nil,
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Function{
																		ParenLeftFodder: nil,
																		ParenRightFodder: nil,
																		Body: &ast.Array{
																			Elements: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.DesugaredObject{
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Name: &ast.Var{
																									Id: "k",
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p16282,
																										FreeVars: ast.Identifiers{
																											"k",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1523),
																												Column: int(10),
																											},
																											End: ast.Location{
																												Line: int(1523),
																												Column: int(11),
																											},
																										},
																									},
																								},
																								Body: &ast.Conditional{
																									Cond: &ast.Unary{
																										Expr: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "std",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1524),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1524),
																																Column: int(18),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "objectHas",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p16434,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1524),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1524),
																															Column: int(28),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "patch",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16438,
																																FreeVars: ast.Identifiers{
																																	"patch",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1524),
																																		Column: int(29),
																																	},
																																	End: ast.Location{
																																		Line: int(1524),
																																		Column: int(34),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															Id: "k",
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16438,
																																FreeVars: ast.Identifiers{
																																	"k",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1524),
																																		Column: int(36),
																																	},
																																	End: ast.Location{
																																		Line: int(1524),
																																		Column: int(37),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1524),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(1524),
																														Column: int(38),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: false,
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p16434,
																											FreeVars: ast.Identifiers{
																												"k",
																												"patch",
																												"std",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1524),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(1524),
																													Column: int(38),
																												},
																											},
																										},
																										Op: ast.UnaryOp(0),
																									},
																									BranchTrue: &ast.Index{
																										Target: &ast.Var{
																											Id: "target_object",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Comment: []string{},
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(12),
																													},
																												},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"target_object",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1525),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1525),
																														Column: int(26),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "k",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"k",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1525),
																														Column: int(27),
																													},
																													End: ast.Location{
																														Line: int(1525),
																														Column: int(28),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p16434,
																											FreeVars: ast.Identifiers{
																												"k",
																												"target_object",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1525),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(1525),
																													Column: int(29),
																												},
																											},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										Cond: &ast.Unary{
																											Expr: &ast.Apply{
																												Target: &ast.Index{
																													Target: &ast.Var{
																														Id: "std",
																														NodeBase: ast.NodeBase{
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																															LocRange: ast.LocationRange{
																																File: p8,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(1526),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(1526),
																																	Column: int(23),
																																},
																															},
																														},
																													},
																													Index: &ast.LiteralString{
																														Value: "objectHas",
																														BlockIndent: "",
																														BlockTermIndent: "",
																														NodeBase: ast.NodeBase{
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{},
																															LocRange: ast.LocationRange{
																																File: nil,
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																															},
																														},
																														Kind: ast.LiteralStringKind(1),
																													},
																													RightBracketFodder: ast.Fodder{},
																													LeftBracketFodder: ast.Fodder{},
																													Id: nil,
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p16434,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1526),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1526),
																																Column: int(33),
																															},
																														},
																													},
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "target_object",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16461,
																																	FreeVars: ast.Identifiers{
																																		"target_object",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1526),
																																			Column: int(34),
																																		},
																																		End: ast.Location{
																																			Line: int(1526),
																																			Column: int(47),
																																		},
																																	},
																																},
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16461,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1526),
																																			Column: int(49),
																																		},
																																		End: ast.Location{
																																			Line: int(1526),
																																			Column: int(50),
																																		},
																																	},
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p16434,
																													FreeVars: ast.Identifiers{
																														"k",
																														"std",
																														"target_object",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1526),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1526),
																															Column: int(51),
																														},
																													},
																												},
																												TrailingComma: false,
																												TailStrict: false,
																											},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"k",
																													"std",
																													"target_object",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1526),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(1526),
																														Column: int(51),
																													},
																												},
																											},
																											Op: ast.UnaryOp(0),
																										},
																										BranchTrue: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "std",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																															},
																														},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1527),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(1527),
																																Column: int(16),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "mergePatch",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p16434,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1527),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(1527),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNull{
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16476,
																																FreeVars: ast.Identifiers{},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1527),
																																		Column: int(28),
																																	},
																																	End: ast.Location{
																																		Line: int(1527),
																																		Column: int(32),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															Target: &ast.Var{
																																Id: "patch",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16476,
																																	FreeVars: ast.Identifiers{
																																		"patch",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1527),
																																			Column: int(34),
																																		},
																																		End: ast.Location{
																																			Line: int(1527),
																																			Column: int(39),
																																		},
																																	},
																																},
																															},
																															Index: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16476,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1527),
																																			Column: int(40),
																																		},
																																		End: ast.Location{
																																			Line: int(1527),
																																			Column: int(41),
																																		},
																																	},
																																},
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16476,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"patch",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1527),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(1527),
																																		Column: int(42),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1527),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1527),
																														Column: int(43),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: true,
																										},
																										BranchFalse: &ast.Apply{
																											Target: &ast.Index{
																												Target: &ast.Var{
																													Id: "std",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Comment: []string{},
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																															},
																														},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1529),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(1529),
																																Column: int(16),
																															},
																														},
																													},
																												},
																												Index: &ast.LiteralString{
																													Value: "mergePatch",
																													BlockIndent: "",
																													BlockTermIndent: "",
																													NodeBase: ast.NodeBase{
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{},
																														LocRange: ast.LocationRange{
																															File: nil,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																														},
																													},
																													Kind: ast.LiteralStringKind(1),
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p16434,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1529),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(1529),
																															Column: int(27),
																														},
																													},
																												},
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															Target: &ast.Var{
																																Id: "target_object",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16494,
																																	FreeVars: ast.Identifiers{
																																		"target_object",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1529),
																																			Column: int(28),
																																		},
																																		End: ast.Location{
																																			Line: int(1529),
																																			Column: int(41),
																																		},
																																	},
																																},
																															},
																															Index: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16494,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1529),
																																			Column: int(42),
																																		},
																																		End: ast.Location{
																																			Line: int(1529),
																																			Column: int(43),
																																		},
																																	},
																																},
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16494,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"target_object",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1529),
																																		Column: int(28),
																																	},
																																	End: ast.Location{
																																		Line: int(1529),
																																		Column: int(44),
																																	},
																																},
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															Target: &ast.Var{
																																Id: "patch",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16494,
																																	FreeVars: ast.Identifiers{
																																		"patch",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1529),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(1529),
																																			Column: int(51),
																																		},
																																	},
																																},
																															},
																															Index: &ast.Var{
																																Id: "k",
																																NodeBase: ast.NodeBase{
																																	Fodder: ast.Fodder{},
																																	Ctx: p16494,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																	LocRange: ast.LocationRange{
																																		File: p8,
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(1529),
																																			Column: int(52),
																																		},
																																		End: ast.Location{
																																			Line: int(1529),
																																			Column: int(53),
																																		},
																																	},
																																},
																															},
																															RightBracketFodder: ast.Fodder{},
																															LeftBracketFodder: ast.Fodder{},
																															Id: nil,
																															NodeBase: ast.NodeBase{
																																Fodder: ast.Fodder{},
																																Ctx: p16494,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"patch",
																																},
																																LocRange: ast.LocationRange{
																																	File: p8,
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(1529),
																																		Column: int(46),
																																	},
																																	End: ast.Location{
																																		Line: int(1529),
																																		Column: int(54),
																																	},
																																},
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: ast.Fodder{},
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p16434,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																													"target_object",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1529),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1529),
																														Column: int(55),
																													},
																												},
																											},
																											TrailingComma: false,
																											TailStrict: true,
																										},
																										ThenFodder: ast.Fodder{},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p16434,
																											FreeVars: ast.Identifiers{
																												"k",
																												"patch",
																												"std",
																												"target_object",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1526),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(1529),
																													Column: int(55),
																												},
																											},
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Comment: []string{},
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																										},
																									},
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Comment: []string{},
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																											},
																										},
																										Ctx: p16434,
																										FreeVars: ast.Identifiers{
																											"k",
																											"patch",
																											"std",
																											"target_object",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1524),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(1529),
																												Column: int(55),
																											},
																										},
																									},
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1523),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(1529),
																										Column: int(55),
																									},
																								},
																								Hide: ast.ObjectFieldHide(1),
																								PlusSuper: false,
																							},
																						},
																						Locals: ast.LocalBinds{},
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Comment: []string{},
																									Kind: ast.FodderKind(0),
																									Blanks: int(1),
																									Indent: int(6),
																								},
																							},
																							Ctx: p16282,
																							FreeVars: ast.Identifiers{
																								"k",
																								"patch",
																								"std",
																								"target_object",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1522),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(1531),
																									Column: int(8),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			CloseFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"k",
																					"patch",
																					"std",
																					"target_object",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			TrailingComma: false,
																		},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: nil,
																				Name: "k",
																				CommaFodder: nil,
																				EqFodder: nil,
																				DefaultArg: nil,
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"patch",
																				"std",
																				"target_object",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		Target: &ast.Index{
																			Target: &ast.Var{
																				Id: "std",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1530),
																							Column: int(18),
																						},
																						End: ast.Location{
																							Line: int(1530),
																							Column: int(21),
																						},
																					},
																				},
																			},
																			Index: &ast.LiteralString{
																				Value: "setDiff",
																				BlockIndent: "",
																				BlockTermIndent: "",
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																				Kind: ast.LiteralStringKind(1),
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16282,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1530),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1530),
																						Column: int(29),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "both_fields",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16524,
																							FreeVars: ast.Identifiers{
																								"both_fields",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1530),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(1530),
																									Column: int(41),
																								},
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						Id: "null_fields",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16524,
																							FreeVars: ast.Identifiers{
																								"null_fields",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1530),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(1530),
																									Column: int(54),
																								},
																							},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16282,
																			FreeVars: ast.Identifiers{
																				"both_fields",
																				"null_fields",
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1530),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1530),
																					Column: int(55),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: nil,
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
																"both_fields",
																"null_fields",
																"patch",
																"std",
																"target_object",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1522),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(1531),
																	Column: int(8),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: nil,
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
												"both_fields",
												"null_fields",
												"patch",
												"std",
												"target_object",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1522),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1531),
													Column: int(8),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p16282,
										FreeVars: ast.Identifiers{
											"$std",
											"null_fields",
											"patch",
											"std",
											"target_fields",
											"target_object",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1520),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1531),
												Column: int(8),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(6),
										},
									},
									Ctx: p16282,
									FreeVars: ast.Identifiers{
										"$std",
										"patch",
										"std",
										"target_fields",
										"target_object",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1519),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1531),
											Column: int(8),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(6),
									},
								},
								Ctx: p16282,
								FreeVars: ast.Identifiers{
									"$std",
									"patch",
									"std",
									"target_object",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1516),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1531),
										Column: int(8),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p16282,
							FreeVars: ast.Identifiers{
								"$std",
								"patch",
								"std",
								"target",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1513),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1531),
									Column: int(8),
								},
							},
						},
					},
					BranchFalse: &ast.Var{
						Id: "patch",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p16282,
							FreeVars: ast.Identifiers{
								"patch",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1533),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1533),
									Column: int(12),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p16282,
						FreeVars: ast.Identifiers{
							"$std",
							"patch",
							"std",
							"target",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1512),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1533),
								Column: int(12),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "target",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1511),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1511),
								Column: int(20),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "patch",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1511),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1511),
								Column: int(27),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1511),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1533),
					Column: int(12),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "get",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1536),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1536),
											Column: int(11),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "objectHasEx",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16555,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1536),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1536),
										Column: int(23),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "o",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16559,
											FreeVars: ast.Identifiers{
												"o",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1536),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(1536),
													Column: int(25),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "f",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16559,
											FreeVars: ast.Identifiers{
												"f",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1536),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(1536),
													Column: int(28),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "inc_hidden",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16559,
											FreeVars: ast.Identifiers{
												"inc_hidden",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1536),
													Column: int(30),
												},
												End: ast.Location{
													Line: int(1536),
													Column: int(40),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16555,
							FreeVars: ast.Identifiers{
								"f",
								"inc_hidden",
								"o",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1536),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1536),
									Column: int(41),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Index{
						Target: &ast.Var{
							Id: "o",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16555,
								FreeVars: ast.Identifiers{
									"o",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1536),
										Column: int(47),
									},
									End: ast.Location{
										Line: int(1536),
										Column: int(48),
									},
								},
							},
						},
						Index: &ast.Var{
							Id: "f",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16555,
								FreeVars: ast.Identifiers{
									"f",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1536),
										Column: int(49),
									},
									End: ast.Location{
										Line: int(1536),
										Column: int(50),
									},
								},
							},
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16555,
							FreeVars: ast.Identifiers{
								"f",
								"o",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1536),
									Column: int(47),
								},
								End: ast.Location{
									Line: int(1536),
									Column: int(51),
								},
							},
						},
					},
					BranchFalse: &ast.Var{
						Id: "default",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16555,
							FreeVars: ast.Identifiers{
								"default",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1536),
									Column: int(57),
								},
								End: ast.Location{
									Line: int(1536),
									Column: int(64),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p16555,
						FreeVars: ast.Identifiers{
							"default",
							"f",
							"inc_hidden",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1536),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1536),
								Column: int(64),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1535),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1535),
								Column: int(8),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1535),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1535),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "default",
						CommaFodder: ast.Fodder{},
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralNull{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16555,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1535),
										Column: int(21),
									},
									End: ast.Location{
										Line: int(1535),
										Column: int(25),
									},
								},
							},
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1535),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1535),
								Column: int(25),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "inc_hidden",
						CommaFodder: nil,
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p16555,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1535),
										Column: int(38),
									},
									End: ast.Location{
										Line: int(1535),
										Column: int(42),
									},
								},
							},
							Value: true,
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1535),
								Column: int(27),
							},
							End: ast.Location{
								Line: int(1535),
								Column: int(42),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1535),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1536),
					Column: int(64),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectFields",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1539),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1539),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "objectFieldsEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16588,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1539),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1539),
									Column: int(23),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "o",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16592,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1539),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1539),
												Column: int(25),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16592,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1539),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1539),
												Column: int(32),
											},
										},
									},
									Value: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p16588,
						FreeVars: ast.Identifiers{
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1539),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1539),
								Column: int(33),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1538),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1538),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1538),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1539),
					Column: int(33),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectFieldsAll",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1542),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1542),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "objectFieldsEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16606,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1542),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1542),
									Column: int(23),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "o",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16610,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1542),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1542),
												Column: int(25),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16610,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1542),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1542),
												Column: int(31),
											},
										},
									},
									Value: true,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p16606,
						FreeVars: ast.Identifiers{
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1542),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1542),
								Column: int(32),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1541),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1541),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1541),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1542),
					Column: int(32),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectHas",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1545),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1545),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "objectHasEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16624,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1545),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1545),
									Column: int(20),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "o",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16628,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1545),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1545),
												Column: int(22),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "f",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16628,
										FreeVars: ast.Identifiers{
											"f",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1545),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1545),
												Column: int(25),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16628,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1545),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1545),
												Column: int(32),
											},
										},
									},
									Value: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p16624,
						FreeVars: ast.Identifiers{
							"f",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1545),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1545),
								Column: int(33),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1544),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1544),
								Column: int(14),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1544),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1544),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1544),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1545),
					Column: int(33),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectHasAll",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1548),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1548),
										Column: int(8),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "objectHasEx",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p16644,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1548),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1548),
									Column: int(20),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "o",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16648,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1548),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1548),
												Column: int(22),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "f",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16648,
										FreeVars: ast.Identifiers{
											"f",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1548),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1548),
												Column: int(25),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16648,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1548),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1548),
												Column: int(31),
											},
										},
									},
									Value: true,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p16644,
						FreeVars: ast.Identifiers{
							"f",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1548),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1548),
								Column: int(32),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1547),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1547),
								Column: int(17),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1547),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1547),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1547),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1548),
					Column: int(32),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectValues",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "$std",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "flatMap",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: nil,
						LeftBracketFodder: nil,
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					FodderLeft: nil,
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: nil,
									ParenRightFodder: nil,
									Body: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Index{
													Target: &ast.Var{
														Id: "o",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16670,
															FreeVars: ast.Identifiers{
																"o",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1551),
																	Column: int(6),
																},
																End: ast.Location{
																	Line: int(1551),
																	Column: int(7),
																},
															},
														},
													},
													Index: &ast.Var{
														Id: "k",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16670,
															FreeVars: ast.Identifiers{
																"k",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1551),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(1551),
																	Column: int(9),
																},
															},
														},
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16670,
														FreeVars: ast.Identifiers{
															"k",
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1551),
																Column: int(6),
															},
															End: ast.Location{
																Line: int(1551),
																Column: int(10),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"k",
												"o",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										TrailingComma: false,
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: nil,
											Name: "k",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1551),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1551),
														Column: int(23),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "objectFields",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16683,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1551),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1551),
													Column: int(36),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "o",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16687,
														FreeVars: ast.Identifiers{
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1551),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(1551),
																Column: int(38),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16683,
										FreeVars: ast.Identifiers{
											"o",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1551),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(1551),
												Column: int(39),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: nil,
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1551),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1551),
								Column: int(40),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1550),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1550),
								Column: int(17),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1550),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1551),
					Column: int(40),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectValuesAll",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "$std",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "flatMap",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: nil,
						LeftBracketFodder: nil,
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					FodderLeft: nil,
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: nil,
									ParenRightFodder: nil,
									Body: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Index{
													Target: &ast.Var{
														Id: "o",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16707,
															FreeVars: ast.Identifiers{
																"o",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1554),
																	Column: int(6),
																},
																End: ast.Location{
																	Line: int(1554),
																	Column: int(7),
																},
															},
														},
													},
													Index: &ast.Var{
														Id: "k",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16707,
															FreeVars: ast.Identifiers{
																"k",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1554),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(1554),
																	Column: int(9),
																},
															},
														},
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16707,
														FreeVars: ast.Identifiers{
															"k",
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1554),
																Column: int(6),
															},
															End: ast.Location{
																Line: int(1554),
																Column: int(10),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"k",
												"o",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										TrailingComma: false,
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: nil,
											Name: "k",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1554),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1554),
														Column: int(23),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "objectFieldsAll",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16720,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1554),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1554),
													Column: int(39),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "o",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16724,
														FreeVars: ast.Identifiers{
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1554),
																Column: int(40),
															},
															End: ast.Location{
																Line: int(1554),
																Column: int(41),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16720,
										FreeVars: ast.Identifiers{
											"o",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1554),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(1554),
												Column: int(42),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: nil,
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1554),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1554),
								Column: int(43),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1553),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1553),
								Column: int(20),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1553),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1554),
					Column: int(43),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectKeysValues",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "$std",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "flatMap",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: nil,
						LeftBracketFodder: nil,
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					FodderLeft: nil,
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: nil,
									ParenRightFodder: nil,
									Body: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.DesugaredObject{
													Asserts: ast.Nodes{},
													Fields: ast.DesugaredObjectFields{
														ast.DesugaredObjectField{
															Name: &ast.LiteralString{
																Value: "key",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Body: &ast.Var{
																Id: "k",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16746,
																	FreeVars: ast.Identifiers{
																		"k",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1557),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1557),
																			Column: int(14),
																		},
																	},
																},
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1557),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(1557),
																	Column: int(14),
																},
															},
															Hide: ast.ObjectFieldHide(1),
															PlusSuper: false,
														},
														ast.DesugaredObjectField{
															Name: &ast.LiteralString{
																Value: "value",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Body: &ast.Index{
																Target: &ast.Var{
																	Id: "o",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16746,
																		FreeVars: ast.Identifiers{
																			"o",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1557),
																				Column: int(23),
																			},
																			End: ast.Location{
																				Line: int(1557),
																				Column: int(24),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "k",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16746,
																		FreeVars: ast.Identifiers{
																			"k",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1557),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1557),
																				Column: int(26),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16746,
																	FreeVars: ast.Identifiers{
																		"k",
																		"o",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1557),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1557),
																			Column: int(27),
																		},
																	},
																},
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1557),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(1557),
																	Column: int(27),
																},
															},
															Hide: ast.ObjectFieldHide(1),
															PlusSuper: false,
														},
													},
													Locals: ast.LocalBinds{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16755,
														FreeVars: ast.Identifiers{
															"k",
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1557),
																Column: int(6),
															},
															End: ast.Location{
																Line: int(1557),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"k",
												"o",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										TrailingComma: false,
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: nil,
											Name: "k",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1557),
														Column: int(39),
													},
													End: ast.Location{
														Line: int(1557),
														Column: int(42),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "objectFields",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16765,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1557),
													Column: int(39),
												},
												End: ast.Location{
													Line: int(1557),
													Column: int(55),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "o",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16769,
														FreeVars: ast.Identifiers{
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1557),
																Column: int(56),
															},
															End: ast.Location{
																Line: int(1557),
																Column: int(57),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16765,
										FreeVars: ast.Identifiers{
											"o",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1557),
												Column: int(39),
											},
											End: ast.Location{
												Line: int(1557),
												Column: int(58),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: nil,
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1557),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1557),
								Column: int(59),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1556),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(1556),
								Column: int(21),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1556),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1557),
					Column: int(59),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "objectKeysValuesAll",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "$std",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
								},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "flatMap",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: nil,
						LeftBracketFodder: nil,
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					FodderLeft: nil,
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: nil,
									ParenRightFodder: nil,
									Body: &ast.Array{
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.DesugaredObject{
													Asserts: ast.Nodes{},
													Fields: ast.DesugaredObjectFields{
														ast.DesugaredObjectField{
															Name: &ast.LiteralString{
																Value: "key",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Body: &ast.Var{
																Id: "k",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16791,
																	FreeVars: ast.Identifiers{
																		"k",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1560),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1560),
																			Column: int(14),
																		},
																	},
																},
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1560),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(1560),
																	Column: int(14),
																},
															},
															Hide: ast.ObjectFieldHide(1),
															PlusSuper: false,
														},
														ast.DesugaredObjectField{
															Name: &ast.LiteralString{
																Value: "value",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															Body: &ast.Index{
																Target: &ast.Var{
																	Id: "o",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16791,
																		FreeVars: ast.Identifiers{
																			"o",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1560),
																				Column: int(23),
																			},
																			End: ast.Location{
																				Line: int(1560),
																				Column: int(24),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "k",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16791,
																		FreeVars: ast.Identifiers{
																			"k",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1560),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1560),
																				Column: int(26),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16791,
																	FreeVars: ast.Identifiers{
																		"k",
																		"o",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1560),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1560),
																			Column: int(27),
																		},
																	},
																},
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1560),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(1560),
																	Column: int(27),
																},
															},
															Hide: ast.ObjectFieldHide(1),
															PlusSuper: false,
														},
													},
													Locals: ast.LocalBinds{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16800,
														FreeVars: ast.Identifiers{
															"k",
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1560),
																Column: int(6),
															},
															End: ast.Location{
																Line: int(1560),
																Column: int(29),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										CloseFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"k",
												"o",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										TrailingComma: false,
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: nil,
											Name: "k",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"o",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: nil,
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1560),
														Column: int(39),
													},
													End: ast.Location{
														Line: int(1560),
														Column: int(42),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "objectFieldsAll",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16810,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1560),
													Column: int(39),
												},
												End: ast.Location{
													Line: int(1560),
													Column: int(58),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "o",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16814,
														FreeVars: ast.Identifiers{
															"o",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1560),
																Column: int(59),
															},
															End: ast.Location{
																Line: int(1560),
																Column: int(60),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16810,
										FreeVars: ast.Identifiers{
											"o",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1560),
												Column: int(39),
											},
											End: ast.Location{
												Line: int(1560),
												Column: int(61),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: nil,
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"o",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1560),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1560),
								Column: int(62),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1559),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(1559),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1559),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1560),
					Column: int(62),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "equals",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1563),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1563),
													Column: int(19),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16829,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1563),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1563),
												Column: int(24),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "a",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16833,
													FreeVars: ast.Identifiers{
														"a",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1563),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(1563),
															Column: int(26),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p16829,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1563),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(1563),
											Column: int(27),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "ta",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1563),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1563),
									Column: int(27),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1564),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1564),
														Column: int(19),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16843,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1564),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1564),
													Column: int(24),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "b",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16847,
														FreeVars: ast.Identifiers{
															"b",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1564),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(1564),
																Column: int(26),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16843,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1564),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1564),
												Column: int(27),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "tb",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1564),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1564),
										Column: int(27),
									},
								},
							},
						},
						Body: &ast.Conditional{
							Cond: &ast.Unary{
								Expr: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1565),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1565),
														Column: int(12),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "primitiveEquals",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1565),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1565),
													Column: int(28),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "ta",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16861,
														FreeVars: ast.Identifiers{
															"ta",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1565),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(1565),
																Column: int(31),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "tb",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16861,
														FreeVars: ast.Identifiers{
															"tb",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1565),
																Column: int(33),
															},
															End: ast.Location{
																Line: int(1565),
																Column: int(35),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16857,
										FreeVars: ast.Identifiers{
											"std",
											"ta",
											"tb",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1565),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1565),
												Column: int(36),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p16857,
									FreeVars: ast.Identifiers{
										"std",
										"ta",
										"tb",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1565),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1565),
											Column: int(36),
										},
									},
								},
								Op: ast.UnaryOp(0),
							},
							BranchTrue: &ast.LiteralBoolean{
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p16857,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1566),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1566),
											Column: int(12),
										},
									},
								},
								Value: false,
							},
							BranchFalse: &ast.Conditional{
								Cond: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1568),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1568),
														Column: int(13),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "primitiveEquals",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1568),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1568),
													Column: int(29),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "ta",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16878,
														FreeVars: ast.Identifiers{
															"ta",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1568),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1568),
																Column: int(32),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "array",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16878,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1568),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1568),
																Column: int(41),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16857,
										FreeVars: ast.Identifiers{
											"std",
											"ta",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1568),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1568),
												Column: int(42),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								BranchTrue: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Body: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1569),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1569),
																	Column: int(23),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16889,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1569),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1569),
																Column: int(30),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16893,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1569),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1569),
																			Column: int(32),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16889,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1569),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1569),
															Column: int(33),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											EqFodder: ast.Fodder{},
											Variable: "la",
											CloseFodder: ast.Fodder{},
											Fun: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1569),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1569),
													Column: int(33),
												},
											},
										},
									},
									Body: &ast.Conditional{
										Cond: &ast.Unary{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1570),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1570),
																	Column: int(16),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "primitiveEquals",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16857,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1570),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1570),
																Column: int(32),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "la",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16906,
																	FreeVars: ast.Identifiers{
																		"la",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1570),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(1570),
																			Column: int(35),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1570),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1570),
																					Column: int(40),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "length",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16906,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1570),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(1570),
																				Column: int(47),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "b",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p16916,
																					FreeVars: ast.Identifiers{
																						"b",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1570),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(1570),
																							Column: int(49),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16906,
																	FreeVars: ast.Identifiers{
																		"b",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1570),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(1570),
																			Column: int(50),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16857,
													FreeVars: ast.Identifiers{
														"b",
														"la",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1570),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1570),
															Column: int(51),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16857,
												FreeVars: ast.Identifiers{
													"b",
													"la",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1570),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1570),
														Column: int(51),
													},
												},
											},
											Op: ast.UnaryOp(0),
										},
										BranchTrue: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p16857,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1571),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1571),
														Column: int(16),
													},
												},
											},
											Value: false,
										},
										BranchFalse: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Body: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Var{
																	Id: "la",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16929,
																		FreeVars: ast.Identifiers{
																			"la",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1574),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1574),
																				Column: int(23),
																			},
																		},
																	},
																},
																Left: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16929,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1574),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1574),
																				Column: int(17),
																			},
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16929,
																	FreeVars: ast.Identifiers{
																		"i",
																		"la",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1574),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1574),
																			Column: int(23),
																		},
																	},
																},
																Op: ast.BinaryOp(8),
															},
															BranchTrue: &ast.LiteralBoolean{
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(14),
																		},
																	},
																	Ctx: p16929,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1575),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1575),
																			Column: int(19),
																		},
																	},
																},
																Value: true,
															},
															BranchFalse: &ast.Conditional{
																Cond: &ast.Binary{
																	Right: &ast.Index{
																		Target: &ast.Var{
																			Id: "b",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16929,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1576),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1576),
																						Column: int(30),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16929,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1576),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1576),
																						Column: int(32),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16929,
																			FreeVars: ast.Identifiers{
																				"b",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1576),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1576),
																					Column: int(33),
																				},
																			},
																		},
																	},
																	Left: &ast.Index{
																		Target: &ast.Var{
																			Id: "a",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16929,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1576),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1576),
																						Column: int(22),
																					},
																				},
																			},
																		},
																		Index: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p16929,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1576),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1576),
																						Column: int(24),
																					},
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p16929,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1576),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1576),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16929,
																		FreeVars: ast.Identifiers{
																			"a",
																			"b",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1576),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1576),
																				Column: int(33),
																			},
																		},
																	},
																	Op: ast.BinaryOp(13),
																},
																BranchTrue: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(14),
																			},
																		},
																		Ctx: p16929,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1577),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1577),
																				Column: int(20),
																			},
																		},
																	},
																	Value: false,
																},
																BranchFalse: &ast.Apply{
																	Target: &ast.Var{
																		Id: "aux",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(14),
																				},
																			},
																			Ctx: p16929,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1579),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1579),
																					Column: int(18),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "a",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p16959,
																						FreeVars: ast.Identifiers{
																							"a",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1579),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(1579),
																								Column: int(20),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "b",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p16959,
																						FreeVars: ast.Identifiers{
																							"b",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1579),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(1579),
																								Column: int(23),
																							},
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					Right: &ast.LiteralNumber{
																						OriginalString: "1",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16959,
																							FreeVars: ast.Identifiers{},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1579),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(1579),
																									Column: int(30),
																								},
																							},
																						},
																					},
																					Left: &ast.Var{
																						Id: "i",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p16959,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1579),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1579),
																									Column: int(26),
																								},
																							},
																						},
																					},
																					OpFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p16959,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1579),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1579),
																								Column: int(30),
																							},
																						},
																					},
																					Op: ast.BinaryOp(3),
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p16929,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1579),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1579),
																				Column: int(31),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: true,
																},
																ThenFodder: ast.Fodder{},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16929,
																	FreeVars: ast.Identifiers{
																		"a",
																		"aux",
																		"b",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1576),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1579),
																			Column: int(31),
																		},
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(12),
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																	},
																},
																Ctx: p16929,
																FreeVars: ast.Identifiers{
																	"a",
																	"aux",
																	"b",
																	"i",
																	"la",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1574),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1579),
																		Column: int(31),
																	},
																},
															},
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "a",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1573),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1573),
																		Column: int(22),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "b",
																CommaFodder: ast.Fodder{},
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1573),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1573),
																		Column: int(25),
																	},
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1573),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(1573),
																		Column: int(28),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: p16975,
															FreeVars: ast.Identifiers{
																"aux",
																"la",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1573),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1579),
																	Column: int(31),
																},
															},
														},
														TrailingComma: false,
													},
													EqFodder: nil,
													Variable: "aux",
													CloseFodder: nil,
													Fun: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Body: &ast.Apply{
												Target: &ast.Var{
													Id: "aux",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p16857,
														FreeVars: ast.Identifiers{
															"aux",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1580),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1580),
																Column: int(14),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "a",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16983,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1580),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1580),
																			Column: int(16),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16983,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1580),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1580),
																			Column: int(19),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p16983,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1580),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1580),
																			Column: int(22),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p16857,
													FreeVars: ast.Identifiers{
														"a",
														"aux",
														"b",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1580),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1580),
															Column: int(23),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
													},
												},
												Ctx: p16857,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"la",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1573),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1580),
														Column: int(23),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"la",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1570),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1580),
													Column: int(23),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
											},
										},
										Ctx: p16857,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1569),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1580),
												Column: int(23),
											},
										},
									},
								},
								BranchFalse: &ast.Conditional{
									Cond: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1581),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1581),
															Column: int(18),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "primitiveEquals",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16857,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1581),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1581),
														Column: int(34),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "ta",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17005,
															FreeVars: ast.Identifiers{
																"ta",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1581),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(1581),
																	Column: int(37),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														Value: "object",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17005,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1581),
																	Column: int(39),
																},
																End: ast.Location{
																	Line: int(1581),
																	Column: int(47),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"std",
												"ta",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1581),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1581),
													Column: int(48),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									BranchTrue: &ast.Local{
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Body: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1582),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1582),
																		Column: int(27),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "objectFields",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17016,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1582),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1582),
																	Column: int(40),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "a",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17020,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1582),
																				Column: int(41),
																			},
																			End: ast.Location{
																				Line: int(1582),
																				Column: int(42),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17016,
														FreeVars: ast.Identifiers{
															"a",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1582),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(1582),
																Column: int(43),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												EqFodder: ast.Fodder{},
												Variable: "fields",
												CloseFodder: ast.Fodder{},
												Fun: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1582),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1582),
														Column: int(43),
													},
												},
											},
										},
										Body: &ast.Local{
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Body: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1583),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1583),
																			Column: int(28),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "length",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17030,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1583),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1583),
																		Column: int(35),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "fields",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17034,
																			FreeVars: ast.Identifiers{
																				"fields",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1583),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(1583),
																					Column: int(42),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17030,
															FreeVars: ast.Identifiers{
																"fields",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1583),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(1583),
																	Column: int(43),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													EqFodder: ast.Fodder{},
													Variable: "lfields",
													CloseFodder: ast.Fodder{},
													Fun: nil,
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1583),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1583),
															Column: int(43),
														},
													},
												},
											},
											Body: &ast.Conditional{
												Cond: &ast.Binary{
													Right: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1584),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1584),
																			Column: int(25),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "objectFields",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p16857,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1584),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(1584),
																		Column: int(38),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "b",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17047,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1584),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(1584),
																					Column: int(40),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16857,
															FreeVars: ast.Identifiers{
																"b",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1584),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(1584),
																	Column: int(41),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													Left: &ast.Var{
														Id: "fields",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16857,
															FreeVars: ast.Identifiers{
																"fields",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1584),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1584),
																	Column: int(18),
																},
															},
														},
													},
													OpFodder: ast.Fodder{},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p16857,
														FreeVars: ast.Identifiers{
															"b",
															"fields",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1584),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1584),
																Column: int(41),
															},
														},
													},
													Op: ast.BinaryOp(13),
												},
												BranchTrue: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p16857,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1585),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1585),
																Column: int(16),
															},
														},
													},
													Value: false,
												},
												BranchFalse: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: nil,
															Body: &ast.Function{
																ParenLeftFodder: ast.Fodder{},
																ParenRightFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	Cond: &ast.Binary{
																		Right: &ast.Var{
																			Id: "lfields",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17061,
																				FreeVars: ast.Identifiers{
																					"lfields",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1588),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1588),
																						Column: int(28),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17061,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1588),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(1588),
																						Column: int(17),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17061,
																			FreeVars: ast.Identifiers{
																				"i",
																				"lfields",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1588),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1588),
																					Column: int(28),
																				},
																			},
																		},
																		Op: ast.BinaryOp(8),
																	},
																	BranchTrue: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(14),
																				},
																			},
																			Ctx: p17061,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1589),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1589),
																					Column: int(19),
																				},
																			},
																		},
																		Value: true,
																	},
																	BranchFalse: &ast.Conditional{
																		Cond: &ast.Local{
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Body: &ast.Index{
																						Target: &ast.Var{
																							Id: "fields",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17073,
																								FreeVars: ast.Identifiers{
																									"fields",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1590),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(1590),
																										Column: int(37),
																									},
																								},
																							},
																						},
																						Index: &ast.Var{
																							Id: "i",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17073,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1590),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1590),
																										Column: int(39),
																									},
																								},
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						LeftBracketFodder: ast.Fodder{},
																						Id: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17073,
																							FreeVars: ast.Identifiers{
																								"fields",
																								"i",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1590),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(1590),
																									Column: int(40),
																								},
																							},
																						},
																					},
																					EqFodder: ast.Fodder{},
																					Variable: "f",
																					CloseFodder: ast.Fodder{},
																					Fun: nil,
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1590),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1590),
																							Column: int(40),
																						},
																					},
																				},
																			},
																			Body: &ast.Binary{
																				Right: &ast.Index{
																					Target: &ast.Var{
																						Id: "b",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17061,
																							FreeVars: ast.Identifiers{
																								"b",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1590),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(1590),
																									Column: int(51),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "f",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17061,
																							FreeVars: ast.Identifiers{
																								"f",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1590),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(1590),
																									Column: int(53),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17061,
																						FreeVars: ast.Identifiers{
																							"b",
																							"f",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1590),
																								Column: int(50),
																							},
																							End: ast.Location{
																								Line: int(1590),
																								Column: int(54),
																							},
																						},
																					},
																				},
																				Left: &ast.Index{
																					Target: &ast.Var{
																						Id: "a",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17061,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1590),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1590),
																									Column: int(43),
																								},
																							},
																						},
																					},
																					Index: &ast.Var{
																						Id: "f",
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17061,
																							FreeVars: ast.Identifiers{
																								"f",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1590),
																									Column: int(44),
																								},
																								End: ast.Location{
																									Line: int(1590),
																									Column: int(45),
																								},
																							},
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					LeftBracketFodder: ast.Fodder{},
																					Id: nil,
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17061,
																						FreeVars: ast.Identifiers{
																							"a",
																							"f",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1590),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1590),
																								Column: int(46),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17061,
																					FreeVars: ast.Identifiers{
																						"a",
																						"b",
																						"f",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1590),
																							Column: int(42),
																						},
																						End: ast.Location{
																							Line: int(1590),
																							Column: int(54),
																						},
																					},
																				},
																				Op: ast.BinaryOp(13),
																			},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17061,
																				FreeVars: ast.Identifiers{
																					"a",
																					"b",
																					"fields",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1590),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1590),
																						Column: int(54),
																					},
																				},
																			},
																		},
																		BranchTrue: &ast.LiteralBoolean{
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Comment: []string{},
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(14),
																					},
																				},
																				Ctx: p17061,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1591),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1591),
																						Column: int(20),
																					},
																				},
																			},
																			Value: false,
																		},
																		BranchFalse: &ast.Apply{
																			Target: &ast.Var{
																				Id: "aux",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(14),
																						},
																					},
																					Ctx: p17061,
																					FreeVars: ast.Identifiers{
																						"aux",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1593),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1593),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "a",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17101,
																								FreeVars: ast.Identifiers{
																									"a",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1593),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(1593),
																										Column: int(20),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "b",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17101,
																								FreeVars: ast.Identifiers{
																									"b",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1593),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(1593),
																										Column: int(23),
																									},
																								},
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							Right: &ast.LiteralNumber{
																								OriginalString: "1",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p17101,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1593),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1593),
																											Column: int(30),
																										},
																									},
																								},
																							},
																							Left: &ast.Var{
																								Id: "i",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p17101,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1593),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1593),
																											Column: int(26),
																										},
																									},
																								},
																							},
																							OpFodder: ast.Fodder{},
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17101,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1593),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1593),
																										Column: int(30),
																									},
																								},
																							},
																							Op: ast.BinaryOp(3),
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17061,
																				FreeVars: ast.Identifiers{
																					"a",
																					"aux",
																					"b",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1593),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1593),
																						Column: int(31),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: true,
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17061,
																			FreeVars: ast.Identifiers{
																				"a",
																				"aux",
																				"b",
																				"fields",
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1590),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1593),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(12),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																			},
																		},
																		Ctx: p17061,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"fields",
																			"i",
																			"lfields",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1588),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1593),
																				Column: int(31),
																			},
																		},
																	},
																},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "a",
																		CommaFodder: ast.Fodder{},
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1587),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1587),
																				Column: int(22),
																			},
																		},
																	},
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "b",
																		CommaFodder: ast.Fodder{},
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1587),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1587),
																				Column: int(25),
																			},
																		},
																	},
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "i",
																		CommaFodder: nil,
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1587),
																				Column: int(27),
																			},
																			End: ast.Location{
																				Line: int(1587),
																				Column: int(28),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: p17117,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"fields",
																		"lfields",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1587),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1593),
																			Column: int(31),
																		},
																	},
																},
																TrailingComma: false,
															},
															EqFodder: nil,
															Variable: "aux",
															CloseFodder: nil,
															Fun: nil,
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Body: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p16857,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1594),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1594),
																		Column: int(14),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "a",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17125,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1594),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1594),
																					Column: int(16),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "b",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17125,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1594),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1594),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17125,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1594),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1594),
																					Column: int(22),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p16857,
															FreeVars: ast.Identifiers{
																"a",
																"aux",
																"b",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1594),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1594),
																	Column: int(23),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
															},
														},
														Ctx: p16857,
														FreeVars: ast.Identifiers{
															"a",
															"b",
															"fields",
															"lfields",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1587),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1594),
																Column: int(23),
															},
														},
													},
												},
												ThenFodder: ast.Fodder{},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p16857,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"fields",
														"lfields",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1584),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1594),
															Column: int(23),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
													},
												},
												Ctx: p16857,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"fields",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1583),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1594),
														Column: int(23),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1582),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1594),
													Column: int(23),
												},
											},
										},
									},
									BranchFalse: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1596),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1596),
															Column: int(12),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "primitiveEquals",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p16857,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1596),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1596),
														Column: int(28),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "a",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17149,
															FreeVars: ast.Identifiers{
																"a",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1596),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1596),
																	Column: int(30),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "b",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17149,
															FreeVars: ast.Identifiers{
																"b",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1596),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(1596),
																	Column: int(33),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p16857,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1596),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1596),
													Column: int(34),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p16857,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
											"ta",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1581),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(1596),
												Column: int(34),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p16857,
									FreeVars: ast.Identifiers{
										"a",
										"b",
										"std",
										"ta",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1568),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1596),
											Column: int(34),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p16857,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
									"ta",
									"tb",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1565),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1596),
										Column: int(34),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p16857,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
								"ta",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1564),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1596),
									Column: int(34),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p16857,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1563),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1596),
								Column: int(34),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1562),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1562),
								Column: int(11),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1562),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1562),
								Column: int(14),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1562),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1596),
					Column: int(34),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "resolvePath",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1600),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1600),
													Column: int(20),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "split",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17177,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1600),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1600),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "f",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17181,
													FreeVars: ast.Identifiers{
														"f",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1600),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1600),
															Column: int(28),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												Value: "/",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17181,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1600),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1600),
															Column: int(33),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17177,
									FreeVars: ast.Identifiers{
										"f",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1600),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1600),
											Column: int(34),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "arr",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1600),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1600),
									Column: int(34),
								},
							},
						},
					},
					Body: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1601),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1601),
											Column: int(8),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "join",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17191,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1601),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1601),
										Column: int(13),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										Value: "/",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17195,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1601),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1601),
													Column: int(17),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										Right: &ast.Array{
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "r",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17200,
															FreeVars: ast.Identifiers{
																"r",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1601),
																	Column: int(77),
																},
																End: ast.Location{
																	Line: int(1601),
																	Column: int(78),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											CloseFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17195,
												FreeVars: ast.Identifiers{
													"r",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1601),
														Column: int(76),
													},
													End: ast.Location{
														Line: int(1601),
														Column: int(79),
													},
												},
											},
											TrailingComma: false,
										},
										Left: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1601),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(1601),
																Column: int(22),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "makeArray",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17195,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1601),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(1601),
															Column: int(32),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															Right: &ast.LiteralNumber{
																OriginalString: "1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17212,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1601),
																			Column: int(51),
																		},
																		End: ast.Location{
																			Line: int(1601),
																			Column: int(52),
																		},
																	},
																},
															},
															Left: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1601),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1601),
																					Column: int(36),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "length",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17212,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1601),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1601),
																				Column: int(43),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "arr",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17221,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1601),
																							Column: int(44),
																						},
																						End: ast.Location{
																							Line: int(1601),
																							Column: int(47),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17212,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1601),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(1601),
																			Column: int(48),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17212,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1601),
																		Column: int(33),
																	},
																	End: ast.Location{
																		Line: int(1601),
																		Column: int(52),
																	},
																},
															},
															Op: ast.BinaryOp(4),
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															ParenLeftFodder: ast.Fodder{},
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Index{
																Target: &ast.Var{
																	Id: "arr",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17228,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1601),
																				Column: int(66),
																			},
																			End: ast.Location{
																				Line: int(1601),
																				Column: int(69),
																			},
																		},
																	},
																},
																Index: &ast.Var{
																	Id: "i",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17228,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1601),
																				Column: int(70),
																			},
																			End: ast.Location{
																				Line: int(1601),
																				Column: int(71),
																			},
																		},
																	},
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17228,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1601),
																			Column: int(66),
																		},
																		End: ast.Location{
																			Line: int(1601),
																			Column: int(72),
																		},
																	},
																},
															},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	CommaFodder: nil,
																	EqFodder: nil,
																	DefaultArg: nil,
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1601),
																			Column: int(63),
																		},
																		End: ast.Location{
																			Line: int(1601),
																			Column: int(64),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17212,
																FreeVars: ast.Identifiers{
																	"arr",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1601),
																		Column: int(54),
																	},
																	End: ast.Location{
																		Line: int(1601),
																		Column: int(72),
																	},
																},
															},
															TrailingComma: false,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17195,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1601),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(1601),
														Column: int(73),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17195,
											FreeVars: ast.Identifiers{
												"arr",
												"r",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1601),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1601),
													Column: int(79),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17191,
							FreeVars: ast.Identifiers{
								"arr",
								"r",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1601),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1601),
									Column: int(80),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p17191,
						FreeVars: ast.Identifiers{
							"f",
							"r",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1600),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1601),
								Column: int(80),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1599),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1599),
								Column: int(16),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "r",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1599),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1599),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1599),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1601),
					Column: int(80),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "prune",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Body: &ast.Function{
								ParenLeftFodder: ast.Fodder{},
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.LiteralNull{
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17250,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1605),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1605),
														Column: int(19),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17250,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1605),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1605),
														Column: int(11),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17250,
											FreeVars: ast.Identifiers{
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1605),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1605),
													Column: int(19),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									BranchTrue: &ast.LiteralBoolean{
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p17250,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1606),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1606),
													Column: int(14),
												},
											},
										},
										Value: false,
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1607),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1607),
																Column: int(18),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "isArray",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17250,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1607),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1607),
															Column: int(26),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "b",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17265,
																FreeVars: ast.Identifiers{
																	"b",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1607),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(1607),
																		Column: int(28),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17250,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1607),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1607),
														Column: int(29),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										BranchTrue: &ast.Binary{
											Right: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17250,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1608),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(1608),
															Column: int(26),
														},
													},
												},
											},
											Left: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1608),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1608),
																	Column: int(12),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "length",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17250,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1608),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1608),
																Column: int(19),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17279,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1608),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1608),
																			Column: int(21),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17250,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1608),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1608),
															Column: int(22),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17250,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1608),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1608),
														Column: int(26),
													},
												},
											},
											Op: ast.BinaryOp(7),
										},
										BranchFalse: &ast.Conditional{
											Cond: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1609),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1609),
																	Column: int(18),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "isObject",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17250,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1609),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1609),
																Column: int(27),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "b",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17292,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1609),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1609),
																			Column: int(29),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17250,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1609),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1609),
															Column: int(30),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											BranchTrue: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "0",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17250,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1610),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(1610),
																Column: int(26),
															},
														},
													},
												},
												Left: &ast.Apply{
													Target: &ast.Index{
														Target: &ast.Var{
															Id: "std",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																	},
																},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1610),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1610),
																		Column: int(12),
																	},
																},
															},
														},
														Index: &ast.LiteralString{
															Value: "length",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														RightBracketFodder: ast.Fodder{},
														LeftBracketFodder: ast.Fodder{},
														Id: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17250,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1610),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1610),
																	Column: int(19),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	Id: "b",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17306,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1610),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1610),
																				Column: int(21),
																			},
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17250,
														FreeVars: ast.Identifiers{
															"b",
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1610),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1610),
																Column: int(22),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17250,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1610),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1610),
															Column: int(26),
														},
													},
												},
												Op: ast.BinaryOp(7),
											},
											BranchFalse: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p17250,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1612),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1612),
															Column: int(13),
														},
													},
												},
												Value: true,
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17250,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1609),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1612),
														Column: int(13),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17250,
											FreeVars: ast.Identifiers{
												"b",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1607),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1612),
													Column: int(13),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p17250,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1605),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1612),
												Column: int(13),
											},
										},
									},
								},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										CommaFodder: nil,
										EqFodder: nil,
										DefaultArg: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1604),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1604),
												Column: int(22),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: p17320,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1604),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1612),
											Column: int(13),
										},
									},
								},
								TrailingComma: false,
							},
							EqFodder: nil,
							Variable: "isContent",
							CloseFodder: nil,
							Fun: nil,
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1613),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1613),
												Column: int(11),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17328,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1613),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1613),
											Column: int(19),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17332,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1613),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1613),
														Column: int(21),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17328,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1613),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1613),
										Column: int(22),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchTrue: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "flatMap",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											ParenLeftFodder: nil,
											ParenRightFodder: nil,
											Body: &ast.Conditional{
												Cond: &ast.Apply{
													Target: &ast.Var{
														Id: "isContent",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17328,
															FreeVars: ast.Identifiers{
																"isContent",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1614),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(1614),
																	Column: int(44),
																},
															},
														},
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "$",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1614),
																						Column: int(45),
																					},
																					End: ast.Location{
																						Line: int(1614),
																						Column: int(46),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "prune",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: ast.Fodder{},
																		LeftBracketFodder: ast.Fodder{},
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17353,
																			FreeVars: ast.Identifiers{
																				"$",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1614),
																					Column: int(45),
																				},
																				End: ast.Location{
																					Line: int(1614),
																					Column: int(52),
																				},
																			},
																		},
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					Id: "x",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17357,
																						FreeVars: ast.Identifiers{
																							"x",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1614),
																								Column: int(53),
																							},
																							End: ast.Location{
																								Line: int(1614),
																								Column: int(54),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17353,
																		FreeVars: ast.Identifiers{
																			"$",
																			"x",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1614),
																				Column: int(45),
																			},
																			End: ast.Location{
																				Line: int(1614),
																				Column: int(55),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17328,
														FreeVars: ast.Identifiers{
															"$",
															"isContent",
															"x",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1614),
																Column: int(35),
															},
															End: ast.Location{
																Line: int(1614),
																Column: int(56),
															},
														},
													},
													TrailingComma: false,
													TailStrict: false,
												},
												BranchTrue: &ast.Array{
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1614),
																					Column: int(8),
																				},
																				End: ast.Location{
																					Line: int(1614),
																					Column: int(11),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "prune",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17368,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1614),
																				Column: int(8),
																			},
																			End: ast.Location{
																				Line: int(1614),
																				Column: int(17),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "x",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17372,
																					FreeVars: ast.Identifiers{
																						"x",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1614),
																							Column: int(18),
																						},
																						End: ast.Location{
																							Line: int(1614),
																							Column: int(19),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17368,
																	FreeVars: ast.Identifiers{
																		"std",
																		"x",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1614),
																			Column: int(8),
																		},
																		End: ast.Location{
																			Line: int(1614),
																			Column: int(20),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													CloseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
															"x",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												BranchFalse: &ast.Array{
													Elements: nil,
													CloseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												ThenFodder: nil,
												ElseFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$",
														"isContent",
														"std",
														"x",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: nil,
													Name: "x",
													CommaFodder: nil,
													EqFodder: nil,
													DefaultArg: nil,
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$",
													"isContent",
													"std",
												},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											TrailingComma: false,
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17328,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1614),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(1614),
														Column: int(31),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$",
									"$std",
									"a",
									"isContent",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1614),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1614),
										Column: int(57),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1615),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1615),
													Column: int(16),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isObject",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17328,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1615),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1615),
												Column: int(25),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "a",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17392,
													FreeVars: ast.Identifiers{
														"a",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1615),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1615),
															Column: int(27),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17328,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1615),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1615),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "$objectFlatMerge",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "$std",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"$std",
															},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "flatMap",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: nil,
													LeftBracketFodder: nil,
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"$std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																ParenLeftFodder: nil,
																ParenRightFodder: nil,
																Body: &ast.Conditional{
																	Cond: &ast.Apply{
																		Target: &ast.Var{
																			Id: "isContent",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17328,
																				FreeVars: ast.Identifiers{
																					"isContent",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1618),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(1618),
																						Column: int(19),
																					},
																				},
																			},
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Apply{
																						Target: &ast.Index{
																							Target: &ast.Var{
																								Id: "std",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1618),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1618),
																											Column: int(23),
																										},
																									},
																								},
																							},
																							Index: &ast.LiteralString{
																								Value: "prune",
																								BlockIndent: "",
																								BlockTermIndent: "",
																								NodeBase: ast.NodeBase{
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{},
																									LocRange: ast.LocationRange{
																										File: nil,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																									},
																								},
																								Kind: ast.LiteralStringKind(1),
																							},
																							RightBracketFodder: ast.Fodder{},
																							LeftBracketFodder: ast.Fodder{},
																							Id: nil,
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17420,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1618),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1618),
																										Column: int(29),
																									},
																								},
																							},
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										Target: &ast.Var{
																											Id: "a",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p17425,
																												FreeVars: ast.Identifiers{
																													"a",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1618),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(1618),
																														Column: int(31),
																													},
																												},
																											},
																										},
																										Index: &ast.Var{
																											Id: "x",
																											NodeBase: ast.NodeBase{
																												Fodder: ast.Fodder{},
																												Ctx: p17425,
																												FreeVars: ast.Identifiers{
																													"x",
																												},
																												LocRange: ast.LocationRange{
																													File: p8,
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(1618),
																														Column: int(32),
																													},
																													End: ast.Location{
																														Line: int(1618),
																														Column: int(33),
																													},
																												},
																											},
																										},
																										RightBracketFodder: ast.Fodder{},
																										LeftBracketFodder: ast.Fodder{},
																										Id: nil,
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: p17425,
																											FreeVars: ast.Identifiers{
																												"a",
																												"x",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1618),
																													Column: int(30),
																												},
																												End: ast.Location{
																													Line: int(1618),
																													Column: int(34),
																												},
																											},
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																						NodeBase: ast.NodeBase{
																							Fodder: ast.Fodder{},
																							Ctx: p17420,
																							FreeVars: ast.Identifiers{
																								"a",
																								"std",
																								"x",
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1618),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1618),
																									Column: int(35),
																								},
																							},
																						},
																						TrailingComma: false,
																						TailStrict: false,
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17328,
																			FreeVars: ast.Identifiers{
																				"a",
																				"isContent",
																				"std",
																				"x",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1618),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(1618),
																					Column: int(36),
																				},
																			},
																		},
																		TrailingComma: false,
																		TailStrict: false,
																	},
																	BranchTrue: &ast.Array{
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.DesugaredObject{
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Name: &ast.Var{
																								Id: "x",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p17328,
																									FreeVars: ast.Identifiers{
																										"x",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1616),
																											Column: int(8),
																										},
																										End: ast.Location{
																											Line: int(1616),
																											Column: int(9),
																										},
																									},
																								},
																							},
																							Body: &ast.Apply{
																								Target: &ast.Index{
																									Target: &ast.Var{
																										Id: "$",
																										NodeBase: ast.NodeBase{
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$",
																											},
																											LocRange: ast.LocationRange{
																												File: p8,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(1616),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(1616),
																													Column: int(13),
																												},
																											},
																										},
																									},
																									Index: &ast.LiteralString{
																										Value: "prune",
																										BlockIndent: "",
																										BlockTermIndent: "",
																										NodeBase: ast.NodeBase{
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{},
																											LocRange: ast.LocationRange{
																												File: nil,
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																											},
																										},
																										Kind: ast.LiteralStringKind(1),
																									},
																									RightBracketFodder: ast.Fodder{},
																									LeftBracketFodder: ast.Fodder{},
																									Id: nil,
																									NodeBase: ast.NodeBase{
																										Fodder: ast.Fodder{},
																										Ctx: p17443,
																										FreeVars: ast.Identifiers{
																											"$",
																										},
																										LocRange: ast.LocationRange{
																											File: p8,
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(1616),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(1616),
																												Column: int(19),
																											},
																										},
																									},
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												Target: &ast.Var{
																													Id: "a",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p17448,
																														FreeVars: ast.Identifiers{
																															"a",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1616),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1616),
																																Column: int(21),
																															},
																														},
																													},
																												},
																												Index: &ast.Var{
																													Id: "x",
																													NodeBase: ast.NodeBase{
																														Fodder: ast.Fodder{},
																														Ctx: p17448,
																														FreeVars: ast.Identifiers{
																															"x",
																														},
																														LocRange: ast.LocationRange{
																															File: p8,
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(1616),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(1616),
																																Column: int(23),
																															},
																														},
																													},
																												},
																												RightBracketFodder: ast.Fodder{},
																												LeftBracketFodder: ast.Fodder{},
																												Id: nil,
																												NodeBase: ast.NodeBase{
																													Fodder: ast.Fodder{},
																													Ctx: p17448,
																													FreeVars: ast.Identifiers{
																														"a",
																														"x",
																													},
																													LocRange: ast.LocationRange{
																														File: p8,
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(1616),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1616),
																															Column: int(24),
																														},
																													},
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p17443,
																									FreeVars: ast.Identifiers{
																										"$",
																										"a",
																										"x",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(1616),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(1616),
																											Column: int(25),
																										},
																									},
																								},
																								TrailingComma: false,
																								TailStrict: false,
																							},
																							LocRange: ast.LocationRange{
																								File: p8,
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(1616),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(1616),
																									Column: int(25),
																								},
																							},
																							Hide: ast.ObjectFieldHide(1),
																							PlusSuper: false,
																						},
																					},
																					Locals: ast.LocalBinds{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17328,
																						FreeVars: ast.Identifiers{
																							"$",
																							"a",
																							"x",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1615),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1619),
																								Column: int(6),
																							},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		CloseFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$",
																				"a",
																				"x",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	BranchFalse: &ast.Array{
																		Elements: nil,
																		CloseFodder: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		TrailingComma: false,
																	},
																	ThenFodder: nil,
																	ElseFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$",
																			"a",
																			"isContent",
																			"std",
																			"x",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: nil,
																		Name: "x",
																		CommaFodder: nil,
																		EqFodder: nil,
																		DefaultArg: nil,
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$",
																		"a",
																		"isContent",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																TrailingComma: false,
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1617),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1617),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "objectFields",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p17328,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1617),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1617),
																				Column: int(32),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "a",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17468,
																					FreeVars: ast.Identifiers{
																						"a",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1617),
																							Column: int(33),
																						},
																						End: ast.Location{
																							Line: int(1617),
																							Column: int(34),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17328,
																	FreeVars: ast.Identifiers{
																		"a",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1617),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1617),
																			Column: int(35),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: nil,
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$",
														"$std",
														"a",
														"isContent",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1615),
															Column: int(34),
														},
														End: ast.Location{
															Line: int(1619),
															Column: int(6),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$",
										"$std",
										"a",
										"isContent",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1615),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(1619),
											Column: int(6),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Var{
								Id: "a",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p17328,
									FreeVars: ast.Identifiers{
										"a",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1620),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1620),
											Column: int(8),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17328,
								FreeVars: ast.Identifiers{
									"$",
									"$std",
									"a",
									"isContent",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1615),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1620),
										Column: int(8),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p17328,
							FreeVars: ast.Identifiers{
								"$",
								"$std",
								"a",
								"isContent",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1613),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1620),
									Column: int(8),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p17328,
						FreeVars: ast.Identifiers{
							"$",
							"$std",
							"a",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1604),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1620),
								Column: int(8),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1603),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(1603),
								Column: int(10),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$",
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1603),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1620),
					Column: int(8),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "findSubstr",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1623),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1623),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isString",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17493,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1623),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1623),
											Column: int(21),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "pat",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17497,
												FreeVars: ast.Identifiers{
													"pat",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1623),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(1623),
														Column: int(25),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17493,
								FreeVars: ast.Identifiers{
									"pat",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1623),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1623),
										Column: int(26),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17493,
							FreeVars: ast.Identifiers{
								"pat",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1623),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1623),
									Column: int(26),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1624),
													Column: int(69),
												},
												End: ast.Location{
													Line: int(1624),
													Column: int(72),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17493,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1624),
												Column: int(69),
											},
											End: ast.Location{
												Line: int(1624),
												Column: int(77),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "pat",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17511,
													FreeVars: ast.Identifiers{
														"pat",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1624),
															Column: int(78),
														},
														End: ast.Location{
															Line: int(1624),
															Column: int(81),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17493,
									FreeVars: ast.Identifiers{
										"pat",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1624),
											Column: int(69),
										},
										End: ast.Location{
											Line: int(1624),
											Column: int(82),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "findSubstr first parameter should be a string, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17493,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1624),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1624),
											Column: int(66),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17493,
								FreeVars: ast.Identifiers{
									"pat",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1624),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1624),
										Column: int(82),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p17493,
							FreeVars: ast.Identifiers{
								"pat",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1624),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1624),
									Column: int(82),
								},
							},
						},
					},
					BranchFalse: &ast.Conditional{
						Cond: &ast.Unary{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1625),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1625),
													Column: int(17),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "isString",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17493,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1625),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(1625),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "str",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17528,
													FreeVars: ast.Identifiers{
														"str",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1625),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1625),
															Column: int(30),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17493,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1625),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(1625),
											Column: int(31),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17493,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1625),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1625),
										Column: int(31),
									},
								},
							},
							Op: ast.UnaryOp(0),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1626),
														Column: int(70),
													},
													End: ast.Location{
														Line: int(1626),
														Column: int(73),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "type",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17493,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1626),
													Column: int(70),
												},
												End: ast.Location{
													Line: int(1626),
													Column: int(78),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "str",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17542,
														FreeVars: ast.Identifiers{
															"str",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1626),
																Column: int(79),
															},
															End: ast.Location{
																Line: int(1626),
																Column: int(82),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17493,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1626),
												Column: int(70),
											},
											End: ast.Location{
												Line: int(1626),
												Column: int(83),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								Left: &ast.LiteralString{
									Value: "findSubstr second parameter should be a string, got ",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17493,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1626),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1626),
												Column: int(67),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17493,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1626),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1626),
											Column: int(83),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p17493,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1626),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1626),
										Column: int(83),
									},
								},
							},
						},
						BranchFalse: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1628),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(1628),
															Column: int(26),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "length",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17556,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1628),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(1628),
														Column: int(33),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "pat",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17560,
															FreeVars: ast.Identifiers{
																"pat",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1628),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(1628),
																	Column: int(37),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17556,
											FreeVars: ast.Identifiers{
												"pat",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1628),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1628),
													Column: int(38),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									EqFodder: ast.Fodder{},
									Variable: "pat_len",
									CloseFodder: ast.Fodder{},
									Fun: nil,
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1628),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1628),
											Column: int(38),
										},
									},
								},
							},
							Body: &ast.Local{
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1629),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1629),
																Column: int(26),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "length",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17570,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1629),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(1629),
															Column: int(33),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															Id: "str",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17574,
																FreeVars: ast.Identifiers{
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1629),
																		Column: int(34),
																	},
																	End: ast.Location{
																		Line: int(1629),
																		Column: int(37),
																	},
																},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17570,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1629),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(1629),
														Column: int(38),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										EqFodder: ast.Fodder{},
										Variable: "str_len",
										CloseFodder: ast.Fodder{},
										Fun: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1629),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1629),
												Column: int(38),
											},
										},
									},
								},
								Body: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Binary{
											Right: &ast.Var{
												Id: "str_len",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17493,
													FreeVars: ast.Identifiers{
														"str_len",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1630),
															Column: int(52),
														},
														End: ast.Location{
															Line: int(1630),
															Column: int(59),
														},
													},
												},
											},
											Left: &ast.Var{
												Id: "pat_len",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17493,
													FreeVars: ast.Identifiers{
														"pat_len",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1630),
															Column: int(42),
														},
														End: ast.Location{
															Line: int(1630),
															Column: int(49),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17493,
												FreeVars: ast.Identifiers{
													"pat_len",
													"str_len",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1630),
														Column: int(42),
													},
													End: ast.Location{
														Line: int(1630),
														Column: int(59),
													},
												},
											},
											Op: ast.BinaryOp(7),
										},
										Left: &ast.Binary{
											Right: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "0",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17493,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1630),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(1630),
																Column: int(38),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "str_len",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17493,
														FreeVars: ast.Identifiers{
															"str_len",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1630),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1630),
																Column: int(33),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17493,
													FreeVars: ast.Identifiers{
														"str_len",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1630),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1630),
															Column: int(38),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											Left: &ast.Binary{
												Right: &ast.LiteralNumber{
													OriginalString: "0",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17493,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1630),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(1630),
																Column: int(22),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "pat_len",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17493,
														FreeVars: ast.Identifiers{
															"pat_len",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1630),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1630),
																Column: int(17),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17493,
													FreeVars: ast.Identifiers{
														"pat_len",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1630),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1630),
															Column: int(22),
														},
													},
												},
												Op: ast.BinaryOp(12),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17493,
												FreeVars: ast.Identifiers{
													"pat_len",
													"str_len",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1630),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1630),
														Column: int(38),
													},
												},
											},
											Op: ast.BinaryOp(18),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17493,
											FreeVars: ast.Identifiers{
												"pat_len",
												"str_len",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1630),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1630),
													Column: int(59),
												},
											},
										},
										Op: ast.BinaryOp(18),
									},
									BranchTrue: &ast.Array{
										Elements: nil,
										CloseFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
												},
											},
											Ctx: p17493,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1631),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1631),
													Column: int(11),
												},
											},
										},
										TrailingComma: false,
									},
									BranchFalse: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1633),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1633),
															Column: int(12),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "filter",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17493,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1633),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1633),
														Column: int(19),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														ParenLeftFodder: ast.Fodder{},
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Binary{
															Right: &ast.Var{
																Id: "pat",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17611,
																	FreeVars: ast.Identifiers{
																		"pat",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1633),
																			Column: int(54),
																		},
																		End: ast.Location{
																			Line: int(1633),
																			Column: int(57),
																		},
																	},
																},
															},
															Left: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "$std",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "slice",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: nil,
																	LeftBracketFodder: nil,
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																		},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "str",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17611,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1633),
																							Column: int(32),
																						},
																						End: ast.Location{
																							Line: int(1633),
																							Column: int(35),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17611,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1633),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1633),
																							Column: int(37),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				Right: &ast.Var{
																					Id: "pat_len",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17611,
																						FreeVars: ast.Identifiers{
																							"pat_len",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1633),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1633),
																								Column: int(49),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "i",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p17611,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1633),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1633),
																								Column: int(39),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17611,
																					FreeVars: ast.Identifiers{
																						"i",
																						"pat_len",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1633),
																							Column: int(38),
																						},
																						End: ast.Location{
																							Line: int(1633),
																							Column: int(49),
																						},
																					},
																				},
																				Op: ast.BinaryOp(3),
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNull{
																				NodeBase: ast.NodeBase{
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: nil,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																FodderRight: nil,
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"i",
																		"pat_len",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1633),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(1633),
																			Column: int(50),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															OpFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17611,
																FreeVars: ast.Identifiers{
																	"$std",
																	"i",
																	"pat",
																	"pat_len",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1633),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(1633),
																		Column: int(57),
																	},
																},
															},
															Op: ast.BinaryOp(12),
														},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																CommaFodder: nil,
																EqFodder: nil,
																DefaultArg: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1633),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(1633),
																		Column: int(30),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17634,
															FreeVars: ast.Identifiers{
																"$std",
																"pat",
																"pat_len",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1633),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1633),
																	Column: int(57),
																},
															},
														},
														TrailingComma: false,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1633),
																			Column: int(59),
																		},
																		End: ast.Location{
																			Line: int(1633),
																			Column: int(62),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "range",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17634,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1633),
																		Column: int(59),
																	},
																	End: ast.Location{
																		Line: int(1633),
																		Column: int(68),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17644,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1633),
																					Column: int(69),
																				},
																				End: ast.Location{
																					Line: int(1633),
																					Column: int(70),
																				},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.Var{
																			Id: "pat_len",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17644,
																				FreeVars: ast.Identifiers{
																					"pat_len",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1633),
																						Column: int(82),
																					},
																					End: ast.Location{
																						Line: int(1633),
																						Column: int(89),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "str_len",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17644,
																				FreeVars: ast.Identifiers{
																					"str_len",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1633),
																						Column: int(72),
																					},
																					End: ast.Location{
																						Line: int(1633),
																						Column: int(79),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17644,
																			FreeVars: ast.Identifiers{
																				"pat_len",
																				"str_len",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1633),
																					Column: int(72),
																				},
																				End: ast.Location{
																					Line: int(1633),
																					Column: int(89),
																				},
																			},
																		},
																		Op: ast.BinaryOp(4),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17634,
															FreeVars: ast.Identifiers{
																"pat_len",
																"std",
																"str_len",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1633),
																	Column: int(59),
																},
																End: ast.Location{
																	Line: int(1633),
																	Column: int(90),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17493,
											FreeVars: ast.Identifiers{
												"$std",
												"pat",
												"pat_len",
												"std",
												"str",
												"str_len",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1633),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1633),
													Column: int(91),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p17493,
										FreeVars: ast.Identifiers{
											"$std",
											"pat",
											"pat_len",
											"std",
											"str",
											"str_len",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1630),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1633),
												Column: int(91),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: p17493,
									FreeVars: ast.Identifiers{
										"$std",
										"pat",
										"pat_len",
										"std",
										"str",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1629),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1633),
											Column: int(91),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p17493,
								FreeVars: ast.Identifiers{
									"$std",
									"pat",
									"std",
									"str",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1628),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1633),
										Column: int(91),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17493,
							FreeVars: ast.Identifiers{
								"$std",
								"pat",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1625),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(1633),
									Column: int(91),
								},
							},
						},
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p17493,
						FreeVars: ast.Identifiers{
							"$std",
							"pat",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1623),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1633),
								Column: int(91),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "pat",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1622),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1622),
								Column: int(17),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1622),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1622),
								Column: int(22),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1622),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1633),
					Column: int(91),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "find",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Unary{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "std",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1636),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1636),
												Column: int(12),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "isArray",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: ast.Fodder{},
								LeftBracketFodder: ast.Fodder{},
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17676,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1636),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1636),
											Column: int(20),
										},
									},
								},
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "arr",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17680,
												FreeVars: ast.Identifiers{
													"arr",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1636),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1636),
														Column: int(24),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17676,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1636),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1636),
										Column: int(25),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17676,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1636),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1636),
									Column: int(25),
								},
							},
						},
						Op: ast.UnaryOp(0),
					},
					BranchTrue: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1637),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(1637),
													Column: int(67),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17676,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1637),
												Column: int(64),
											},
											End: ast.Location{
												Line: int(1637),
												Column: int(72),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17694,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1637),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(1637),
															Column: int(76),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17676,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1637),
											Column: int(64),
										},
										End: ast.Location{
											Line: int(1637),
											Column: int(77),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "find second parameter should be an array, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17676,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1637),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1637),
											Column: int(61),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17676,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1637),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1637),
										Column: int(77),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
								},
							},
							Ctx: p17676,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1637),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1637),
									Column: int(77),
								},
							},
						},
					},
					BranchFalse: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1639),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1639),
											Column: int(10),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "filter",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17676,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1639),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1639),
										Column: int(17),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Binary{
											Right: &ast.Var{
												Id: "value",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17712,
													FreeVars: ast.Identifiers{
														"value",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1639),
															Column: int(40),
														},
														End: ast.Location{
															Line: int(1639),
															Column: int(45),
														},
													},
												},
											},
											Left: &ast.Index{
												Target: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17712,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1639),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1639),
																Column: int(33),
															},
														},
													},
												},
												Index: &ast.Var{
													Id: "i",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17712,
														FreeVars: ast.Identifiers{
															"i",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1639),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1639),
																Column: int(35),
															},
														},
													},
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17712,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1639),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1639),
															Column: int(36),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17712,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"value",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1639),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(1639),
														Column: int(45),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1639),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1639),
														Column: int(28),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17722,
											FreeVars: ast.Identifiers{
												"arr",
												"value",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1639),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1639),
													Column: int(45),
												},
											},
										},
										TrailingComma: false,
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1639),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(1639),
															Column: int(50),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "range",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p17722,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1639),
														Column: int(47),
													},
													End: ast.Location{
														Line: int(1639),
														Column: int(56),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														OriginalString: "0",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17732,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1639),
																	Column: int(57),
																},
																End: ast.Location{
																	Line: int(1639),
																	Column: int(58),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "1",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17732,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1639),
																		Column: int(78),
																	},
																	End: ast.Location{
																		Line: int(1639),
																		Column: int(79),
																	},
																},
															},
														},
														Left: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1639),
																				Column: int(60),
																			},
																			End: ast.Location{
																				Line: int(1639),
																				Column: int(63),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "length",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17732,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1639),
																			Column: int(60),
																		},
																		End: ast.Location{
																			Line: int(1639),
																			Column: int(70),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "arr",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17743,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1639),
																						Column: int(71),
																					},
																					End: ast.Location{
																						Line: int(1639),
																						Column: int(74),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17732,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1639),
																		Column: int(60),
																	},
																	End: ast.Location{
																		Line: int(1639),
																		Column: int(75),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17732,
															FreeVars: ast.Identifiers{
																"arr",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1639),
																	Column: int(60),
																},
																End: ast.Location{
																	Line: int(1639),
																	Column: int(79),
																},
															},
														},
														Op: ast.BinaryOp(4),
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17722,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1639),
													Column: int(47),
												},
												End: ast.Location{
													Line: int(1639),
													Column: int(80),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17676,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"value",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1639),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1639),
									Column: int(81),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					ThenFodder: ast.Fodder{},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Comment: []string{},
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p17676,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"value",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1636),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1639),
								Column: int(81),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1635),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1635),
								Column: int(13),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1635),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1635),
								Column: int(18),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1635),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1639),
					Column: int(81),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "all",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1642),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(1642),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isArray",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17762,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1642),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(1642),
										Column: int(23),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17766,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1642),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(1642),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17762,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1642),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(1642),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1643),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1643),
														Column: int(23),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17776,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1643),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1643),
													Column: int(30),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17780,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1643),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(1643),
																Column: int(34),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17776,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1643),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(1643),
												Column: int(35),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "arrLen",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1643),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1643),
										Column: int(35),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Body: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Var{
													Id: "arrLen",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17789,
														FreeVars: ast.Identifiers{
															"arrLen",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1645),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1645),
																Column: int(23),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "idx",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17789,
														FreeVars: ast.Identifiers{
															"idx",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1645),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1645),
																Column: int(13),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17789,
													FreeVars: ast.Identifiers{
														"arrLen",
														"idx",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1645),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1645),
															Column: int(23),
														},
													},
												},
												Op: ast.BinaryOp(8),
											},
											BranchTrue: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p17789,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1646),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1646),
															Column: int(13),
														},
													},
												},
												Value: true,
											},
											BranchFalse: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Index{
															Target: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17800,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1648),
																			Column: int(19),
																		},
																		End: ast.Location{
																			Line: int(1648),
																			Column: int(22),
																		},
																	},
																},
															},
															Index: &ast.Var{
																Id: "idx",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17800,
																	FreeVars: ast.Identifiers{
																		"idx",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1648),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1648),
																			Column: int(26),
																		},
																	},
																},
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17800,
																FreeVars: ast.Identifiers{
																	"arr",
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1648),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1648),
																		Column: int(27),
																	},
																},
															},
														},
														EqFodder: ast.Fodder{},
														Variable: "e",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1648),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1648),
																Column: int(27),
															},
														},
													},
												},
												Body: &ast.Conditional{
													Cond: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1649),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1649),
																			Column: int(19),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "isBoolean",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17789,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1649),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(1649),
																		Column: int(29),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "e",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17814,
																			FreeVars: ast.Identifiers{
																				"e",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1649),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(1649),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17789,
															FreeVars: ast.Identifiers{
																"e",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1649),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(1649),
																	Column: int(32),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchTrue: &ast.Conditional{
														Cond: &ast.Unary{
															Expr: &ast.Var{
																Id: "e",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17789,
																	FreeVars: ast.Identifiers{
																		"e",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1650),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1650),
																			Column: int(14),
																		},
																	},
																},
															},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17789,
																FreeVars: ast.Identifiers{
																	"e",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1650),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1650),
																		Column: int(14),
																	},
																},
															},
															Op: ast.UnaryOp(0),
														},
														BranchTrue: &ast.LiteralBoolean{
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p17789,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1651),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1651),
																		Column: int(16),
																	},
																},
															},
															Value: false,
														},
														BranchFalse: &ast.Apply{
															Target: &ast.Var{
																Id: "aux",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p17789,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1653),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1653),
																			Column: int(14),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17831,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1653),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1653),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "idx",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17831,
																					FreeVars: ast.Identifiers{
																						"idx",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1653),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1653),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17831,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1653),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1653),
																						Column: int(22),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17789,
																FreeVars: ast.Identifiers{
																	"aux",
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1653),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1653),
																		Column: int(23),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: true,
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p17789,
															FreeVars: ast.Identifiers{
																"aux",
																"e",
																"idx",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1650),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1653),
																	Column: int(23),
																},
															},
														},
													},
													BranchFalse: &ast.Error{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1649),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1649),
																				Column: int(38),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "format",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17789,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1649),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1649),
																			Column: int(45),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.LiteralString{
																			Value: "element \"%s\" of type %s is not a boolean",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17848,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1649),
																						Column: int(46),
																					},
																					End: ast.Location{
																						Line: int(1649),
																						Column: int(88),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "e",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17848,
																				FreeVars: ast.Identifiers{
																					"e",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1649),
																						Column: int(90),
																					},
																					End: ast.Location{
																						Line: int(1649),
																						Column: int(91),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1649),
																								Column: int(93),
																							},
																							End: ast.Location{
																								Line: int(1649),
																								Column: int(96),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "type",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17848,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1649),
																							Column: int(93),
																						},
																						End: ast.Location{
																							Line: int(1649),
																							Column: int(101),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "e",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p17859,
																								FreeVars: ast.Identifiers{
																									"e",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1649),
																										Column: int(102),
																									},
																									End: ast.Location{
																										Line: int(1649),
																										Column: int(103),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17848,
																				FreeVars: ast.Identifiers{
																					"e",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1649),
																						Column: int(93),
																					},
																					End: ast.Location{
																						Line: int(1649),
																						Column: int(104),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17789,
																FreeVars: ast.Identifiers{
																	"e",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1649),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(1649),
																		Column: int(105),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"e",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1649),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1653),
																	Column: int(23),
																},
															},
														},
													},
													ThenFodder: nil,
													ElseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"aux",
															"e",
															"idx",
															"std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p17789,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"idx",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1648),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1653),
															Column: int(23),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p17789,
												FreeVars: ast.Identifiers{
													"arr",
													"arrLen",
													"aux",
													"idx",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1645),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1653),
														Column: int(23),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "idx",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1644),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1644),
														Column: int(18),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: p17871,
											FreeVars: ast.Identifiers{
												"arr",
												"arrLen",
												"aux",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1644),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(1653),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
									},
									EqFodder: nil,
									Variable: "aux",
									CloseFodder: nil,
									Fun: nil,
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							Body: &ast.Apply{
								Target: &ast.Var{
									Id: "aux",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p17762,
										FreeVars: ast.Identifiers{
											"aux",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1654),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(1654),
												Column: int(8),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17879,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1654),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1654),
															Column: int(10),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17762,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1654),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1654),
											Column: int(11),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p17762,
								FreeVars: ast.Identifiers{
									"arr",
									"arrLen",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1644),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1654),
										Column: int(11),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p17762,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1643),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1654),
									Column: int(11),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1642),
													Column: int(76),
												},
												End: ast.Location{
													Line: int(1642),
													Column: int(79),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17762,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1642),
												Column: int(76),
											},
											End: ast.Location{
												Line: int(1642),
												Column: int(84),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17895,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1642),
															Column: int(85),
														},
														End: ast.Location{
															Line: int(1642),
															Column: int(88),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17762,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1642),
											Column: int(76),
										},
										End: ast.Location{
											Line: int(1642),
											Column: int(89),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "all() parameter should be an array, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17762,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1642),
											Column: int(31),
										},
										End: ast.Location{
											Line: int(1642),
											Column: int(73),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17762,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1642),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(1642),
										Column: int(89),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1642),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1654),
									Column: int(11),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1641),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1641),
								Column: int(10),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1641),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1654),
					Column: int(11),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "any",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1657),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(1657),
											Column: int(15),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "isArray",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17912,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1657),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(1657),
										Column: int(23),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17916,
											FreeVars: ast.Identifiers{
												"arr",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1657),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(1657),
													Column: int(27),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p17912,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1657),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(1657),
									Column: int(28),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					BranchTrue: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1658),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1658),
														Column: int(23),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "length",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p17926,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1658),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1658),
													Column: int(30),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "arr",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17930,
														FreeVars: ast.Identifiers{
															"arr",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1658),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(1658),
																Column: int(34),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17926,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1658),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(1658),
												Column: int(35),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "arrLen",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1658),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1658),
										Column: int(35),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Body: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Var{
													Id: "arrLen",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17939,
														FreeVars: ast.Identifiers{
															"arrLen",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1660),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1660),
																Column: int(23),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "idx",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p17939,
														FreeVars: ast.Identifiers{
															"idx",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1660),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1660),
																Column: int(13),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p17939,
													FreeVars: ast.Identifiers{
														"arrLen",
														"idx",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1660),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1660),
															Column: int(23),
														},
													},
												},
												Op: ast.BinaryOp(8),
											},
											BranchTrue: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p17939,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1661),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1661),
															Column: int(14),
														},
													},
												},
												Value: false,
											},
											BranchFalse: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Index{
															Target: &ast.Var{
																Id: "arr",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17950,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1663),
																			Column: int(19),
																		},
																		End: ast.Location{
																			Line: int(1663),
																			Column: int(22),
																		},
																	},
																},
															},
															Index: &ast.Var{
																Id: "idx",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17950,
																	FreeVars: ast.Identifiers{
																		"idx",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1663),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1663),
																			Column: int(26),
																		},
																	},
																},
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17950,
																FreeVars: ast.Identifiers{
																	"arr",
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1663),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1663),
																		Column: int(27),
																	},
																},
															},
														},
														EqFodder: ast.Fodder{},
														Variable: "e",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1663),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1663),
																Column: int(27),
															},
														},
													},
												},
												Body: &ast.Conditional{
													Cond: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1664),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1664),
																			Column: int(19),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "isBoolean",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17939,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1664),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(1664),
																		Column: int(29),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "e",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p17964,
																			FreeVars: ast.Identifiers{
																				"e",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1664),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(1664),
																					Column: int(31),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p17939,
															FreeVars: ast.Identifiers{
																"e",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1664),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(1664),
																	Column: int(32),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													BranchTrue: &ast.Conditional{
														Cond: &ast.Var{
															Id: "e",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17939,
																FreeVars: ast.Identifiers{
																	"e",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1665),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1665),
																		Column: int(13),
																	},
																},
															},
														},
														BranchTrue: &ast.LiteralBoolean{
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p17939,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1666),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1666),
																		Column: int(15),
																	},
																},
															},
															Value: true,
														},
														BranchFalse: &ast.Apply{
															Target: &ast.Var{
																Id: "aux",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																		},
																	},
																	Ctx: p17939,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1668),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1668),
																			Column: int(14),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17979,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1668),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1668),
																							Column: int(22),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "idx",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17979,
																					FreeVars: ast.Identifiers{
																						"idx",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1668),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1668),
																							Column: int(18),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17979,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1668),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1668),
																						Column: int(22),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17939,
																FreeVars: ast.Identifiers{
																	"aux",
																	"idx",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1668),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1668),
																		Column: int(23),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: true,
														},
														ThenFodder: ast.Fodder{},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: p17939,
															FreeVars: ast.Identifiers{
																"aux",
																"e",
																"idx",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1665),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1668),
																	Column: int(23),
																},
															},
														},
													},
													BranchFalse: &ast.Error{
														Expr: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1664),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1664),
																				Column: int(38),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "format",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p17939,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1664),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1664),
																			Column: int(45),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.LiteralString{
																			Value: "element \"%s\" of type %s is not a boolean",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17996,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1664),
																						Column: int(46),
																					},
																					End: ast.Location{
																						Line: int(1664),
																						Column: int(88),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			Id: "e",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17996,
																				FreeVars: ast.Identifiers{
																					"e",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1664),
																						Column: int(90),
																					},
																					End: ast.Location{
																						Line: int(1664),
																						Column: int(91),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			Target: &ast.Index{
																				Target: &ast.Var{
																					Id: "std",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(1664),
																								Column: int(93),
																							},
																							End: ast.Location{
																								Line: int(1664),
																								Column: int(96),
																							},
																						},
																					},
																				},
																				Index: &ast.LiteralString{
																					Value: "type",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: nil,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				RightBracketFodder: ast.Fodder{},
																				LeftBracketFodder: ast.Fodder{},
																				Id: nil,
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p17996,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1664),
																							Column: int(93),
																						},
																						End: ast.Location{
																							Line: int(1664),
																							Column: int(101),
																						},
																					},
																				},
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							Id: "e",
																							NodeBase: ast.NodeBase{
																								Fodder: ast.Fodder{},
																								Ctx: p18007,
																								FreeVars: ast.Identifiers{
																									"e",
																								},
																								LocRange: ast.LocationRange{
																									File: p8,
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(1664),
																										Column: int(102),
																									},
																									End: ast.Location{
																										Line: int(1664),
																										Column: int(103),
																									},
																								},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p17996,
																				FreeVars: ast.Identifiers{
																					"e",
																					"std",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1664),
																						Column: int(93),
																					},
																					End: ast.Location{
																						Line: int(1664),
																						Column: int(104),
																					},
																				},
																			},
																			TrailingComma: false,
																			TailStrict: false,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p17939,
																FreeVars: ast.Identifiers{
																	"e",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1664),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(1664),
																		Column: int(105),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"e",
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1664),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1668),
																	Column: int(23),
																},
															},
														},
													},
													ThenFodder: nil,
													ElseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"aux",
															"e",
															"idx",
															"std",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p17939,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"idx",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1663),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1668),
															Column: int(23),
														},
													},
												},
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p17939,
												FreeVars: ast.Identifiers{
													"arr",
													"arrLen",
													"aux",
													"idx",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1660),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1668),
														Column: int(23),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "idx",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1659),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1659),
														Column: int(18),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: p18019,
											FreeVars: ast.Identifiers{
												"arr",
												"arrLen",
												"aux",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1659),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(1668),
													Column: int(23),
												},
											},
										},
										TrailingComma: false,
									},
									EqFodder: nil,
									Variable: "aux",
									CloseFodder: nil,
									Fun: nil,
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							Body: &ast.Apply{
								Target: &ast.Var{
									Id: "aux",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p17912,
										FreeVars: ast.Identifiers{
											"aux",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1669),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(1669),
												Column: int(8),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18027,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1669),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1669),
															Column: int(10),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17912,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1669),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1669),
											Column: int(11),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p17912,
								FreeVars: ast.Identifiers{
									"arr",
									"arrLen",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1659),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1669),
										Column: int(11),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p17912,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1658),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1669),
									Column: int(11),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Binary{
							Right: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1657),
													Column: int(76),
												},
												End: ast.Location{
													Line: int(1657),
													Column: int(79),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p17912,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1657),
												Column: int(76),
											},
											End: ast.Location{
												Line: int(1657),
												Column: int(84),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18043,
													FreeVars: ast.Identifiers{
														"arr",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1657),
															Column: int(85),
														},
														End: ast.Location{
															Line: int(1657),
															Column: int(88),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17912,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1657),
											Column: int(76),
										},
										End: ast.Location{
											Line: int(1657),
											Column: int(89),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							Left: &ast.LiteralString{
								Value: "any() parameter should be an array, got ",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p17912,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1657),
											Column: int(31),
										},
										End: ast.Location{
											Line: int(1657),
											Column: int(73),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p17912,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1657),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(1657),
										Column: int(89),
									},
								},
							},
							Op: ast.BinaryOp(3),
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1657),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1669),
									Column: int(11),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1656),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1656),
								Column: int(10),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1656),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1669),
					Column: int(11),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__compare",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1674),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1674),
													Column: int(19),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18061,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1674),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1674),
												Column: int(24),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "v1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18065,
													FreeVars: ast.Identifiers{
														"v1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1674),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(1674),
															Column: int(27),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18061,
									FreeVars: ast.Identifiers{
										"std",
										"v1",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1674),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(1674),
											Column: int(28),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "t1",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1674),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1674),
									Column: int(28),
								},
							},
						},
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1674),
													Column: int(35),
												},
												End: ast.Location{
													Line: int(1674),
													Column: int(38),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "type",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18073,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1674),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(1674),
												Column: int(43),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "v2",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18077,
													FreeVars: ast.Identifiers{
														"v2",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1674),
															Column: int(44),
														},
														End: ast.Location{
															Line: int(1674),
															Column: int(46),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18073,
									FreeVars: ast.Identifiers{
										"std",
										"v2",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1674),
											Column: int(35),
										},
										End: ast.Location{
											Line: int(1674),
											Column: int(47),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "t2",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1674),
									Column: int(30),
								},
								End: ast.Location{
									Line: int(1674),
									Column: int(47),
								},
							},
						},
					},
					Body: &ast.Conditional{
						Cond: &ast.Binary{
							Right: &ast.Var{
								Id: "t2",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"t2",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1675),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(1675),
											Column: int(16),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "t1",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"t1",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1675),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1675),
											Column: int(10),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18083,
								FreeVars: ast.Identifiers{
									"t1",
									"t2",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1675),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1675),
										Column: int(16),
									},
								},
							},
							Op: ast.BinaryOp(13),
						},
						BranchTrue: &ast.Error{
							Expr: &ast.Binary{
								Right: &ast.Var{
									Id: "t2",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"t2",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1676),
												Column: int(73),
											},
											End: ast.Location{
												Line: int(1676),
												Column: int(75),
											},
										},
									},
								},
								Left: &ast.Binary{
									Right: &ast.LiteralString{
										Value: " and ",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1676),
													Column: int(63),
												},
												End: ast.Location{
													Line: int(1676),
													Column: int(70),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									Left: &ast.Binary{
										Right: &ast.Var{
											Id: "t1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"t1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1676),
														Column: int(58),
													},
													End: ast.Location{
														Line: int(1676),
														Column: int(60),
													},
												},
											},
										},
										Left: &ast.LiteralString{
											Value: "Comparison requires matching types. Got ",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1676),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1676),
														Column: int(55),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"t1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1676),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1676),
													Column: int(60),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"t1",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1676),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1676),
												Column: int(70),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"t1",
										"t2",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1676),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1676),
											Column: int(75),
										},
									},
								},
								Op: ast.BinaryOp(3),
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
									},
								},
								Ctx: p18083,
								FreeVars: ast.Identifiers{
									"t1",
									"t2",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1676),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1676),
										Column: int(75),
									},
								},
							},
						},
						BranchFalse: &ast.Conditional{
							Cond: &ast.Binary{
								Right: &ast.LiteralString{
									Value: "array",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1677),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(1677),
												Column: int(26),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								Left: &ast.Var{
									Id: "t1",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"t1",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1677),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1677),
												Column: int(15),
											},
										},
									},
								},
								OpFodder: ast.Fodder{},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"t1",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1677),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1677),
											Column: int(26),
										},
									},
								},
								Op: ast.BinaryOp(12),
							},
							BranchTrue: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1678),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1678),
													Column: int(10),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "__compare_array",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1678),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1678),
												Column: int(26),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "v1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18118,
													FreeVars: ast.Identifiers{
														"v1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1678),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1678),
															Column: int(29),
														},
													},
												},
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "v2",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18118,
													FreeVars: ast.Identifiers{
														"v2",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1678),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(1678),
															Column: int(33),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"std",
										"v1",
										"v2",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1678),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1678),
											Column: int(34),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							BranchFalse: &ast.Conditional{
								Cond: &ast.Binary{
									Right: &ast.Binary{
										Right: &ast.LiteralString{
											Value: "boolean",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1679),
														Column: int(57),
													},
													End: ast.Location{
														Line: int(1679),
														Column: int(66),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Left: &ast.Var{
											Id: "t1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"t1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1679),
														Column: int(51),
													},
													End: ast.Location{
														Line: int(1679),
														Column: int(53),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"t1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1679),
													Column: int(51),
												},
												End: ast.Location{
													Line: int(1679),
													Column: int(66),
												},
											},
										},
										Op: ast.BinaryOp(12),
									},
									Left: &ast.Binary{
										Right: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "object",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1679),
															Column: int(39),
														},
														End: ast.Location{
															Line: int(1679),
															Column: int(47),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Var{
												Id: "t1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{
														"t1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1679),
															Column: int(33),
														},
														End: ast.Location{
															Line: int(1679),
															Column: int(35),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"t1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1679),
														Column: int(33),
													},
													End: ast.Location{
														Line: int(1679),
														Column: int(47),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										Left: &ast.Binary{
											Right: &ast.LiteralString{
												Value: "function",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1679),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(1679),
															Column: int(29),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											Left: &ast.Var{
												Id: "t1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{
														"t1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1679),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1679),
															Column: int(15),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"t1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1679),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1679),
														Column: int(29),
													},
												},
											},
											Op: ast.BinaryOp(12),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"t1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1679),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1679),
													Column: int(47),
												},
											},
										},
										Op: ast.BinaryOp(18),
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"t1",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1679),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1679),
												Column: int(66),
											},
										},
									},
									Op: ast.BinaryOp(18),
								},
								BranchTrue: &ast.Error{
									Expr: &ast.Binary{
										Right: &ast.LiteralString{
											Value: " are not comparable.",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1680),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(1680),
														Column: int(60),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										Left: &ast.Binary{
											Right: &ast.Var{
												Id: "t1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{
														"t1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1680),
															Column: int(33),
														},
														End: ast.Location{
															Line: int(1680),
															Column: int(35),
														},
													},
												},
											},
											Left: &ast.LiteralString{
												Value: "Values of type ",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1680),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1680),
															Column: int(30),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"t1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1680),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1680),
														Column: int(35),
													},
												},
											},
											Op: ast.BinaryOp(3),
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"t1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1680),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1680),
													Column: int(60),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
											},
										},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"t1",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1680),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1680),
												Column: int(60),
											},
										},
									},
								},
								BranchFalse: &ast.Conditional{
									Cond: &ast.Binary{
										Right: &ast.Var{
											Id: "v2",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"v2",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1681),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(1681),
														Column: int(20),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "v1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"v1",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1681),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1681),
														Column: int(15),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"v1",
												"v2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1681),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1681),
													Column: int(20),
												},
											},
										},
										Op: ast.BinaryOp(9),
									},
									BranchTrue: &ast.Unary{
										Expr: &ast.LiteralNumber{
											OriginalString: "1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1681),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1681),
														Column: int(28),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1681),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(1681),
													Column: int(28),
												},
											},
										},
										Op: ast.UnaryOp(3),
									},
									BranchFalse: &ast.Conditional{
										Cond: &ast.Binary{
											Right: &ast.Var{
												Id: "v2",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{
														"v2",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1682),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(1682),
															Column: int(20),
														},
													},
												},
											},
											Left: &ast.Var{
												Id: "v1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18083,
													FreeVars: ast.Identifiers{
														"v1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1682),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1682),
															Column: int(15),
														},
													},
												},
											},
											OpFodder: ast.Fodder{},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{
													"v1",
													"v2",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1682),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1682),
														Column: int(20),
													},
												},
											},
											Op: ast.BinaryOp(7),
										},
										BranchTrue: &ast.LiteralNumber{
											OriginalString: "1",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1682),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(1682),
														Column: int(27),
													},
												},
											},
										},
										BranchFalse: &ast.LiteralNumber{
											OriginalString: "0",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18083,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1683),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1683),
														Column: int(11),
													},
												},
											},
										},
										ThenFodder: ast.Fodder{},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18083,
											FreeVars: ast.Identifiers{
												"v1",
												"v2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1682),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1683),
													Column: int(11),
												},
											},
										},
									},
									ThenFodder: ast.Fodder{},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18083,
										FreeVars: ast.Identifiers{
											"v1",
											"v2",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1681),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1683),
												Column: int(11),
											},
										},
									},
								},
								ThenFodder: ast.Fodder{},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18083,
									FreeVars: ast.Identifiers{
										"t1",
										"v1",
										"v2",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1679),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(1683),
											Column: int(11),
										},
									},
								},
							},
							ThenFodder: ast.Fodder{},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18083,
								FreeVars: ast.Identifiers{
									"std",
									"t1",
									"v1",
									"v2",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1677),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1683),
										Column: int(11),
									},
								},
							},
						},
						ThenFodder: ast.Fodder{},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p18083,
							FreeVars: ast.Identifiers{
								"std",
								"t1",
								"t2",
								"v1",
								"v2",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1675),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1683),
									Column: int(11),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p18083,
						FreeVars: ast.Identifiers{
							"std",
							"v1",
							"v2",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1674),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1683),
								Column: int(11),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1673),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1673),
								Column: int(15),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1673),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(1673),
								Column: int(19),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1673),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1683),
					Column: int(11),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__compare_array",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1686),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1686),
													Column: int(21),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18196,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1686),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1686),
												Column: int(28),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr1",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18200,
													FreeVars: ast.Identifiers{
														"arr1",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1686),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(1686),
															Column: int(33),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18196,
									FreeVars: ast.Identifiers{
										"arr1",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1686),
											Column: int(18),
										},
										End: ast.Location{
											Line: int(1686),
											Column: int(34),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "len1",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1686),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1686),
									Column: int(34),
								},
							},
						},
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Body: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "std",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1686),
													Column: int(43),
												},
												End: ast.Location{
													Line: int(1686),
													Column: int(46),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "length",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: ast.Fodder{},
									LeftBracketFodder: ast.Fodder{},
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18208,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1686),
												Column: int(43),
											},
											End: ast.Location{
												Line: int(1686),
												Column: int(53),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												Id: "arr2",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18212,
													FreeVars: ast.Identifiers{
														"arr2",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1686),
															Column: int(54),
														},
														End: ast.Location{
															Line: int(1686),
															Column: int(58),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18208,
									FreeVars: ast.Identifiers{
										"arr2",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1686),
											Column: int(43),
										},
										End: ast.Location{
											Line: int(1686),
											Column: int(59),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							EqFodder: ast.Fodder{},
							Variable: "len2",
							CloseFodder: ast.Fodder{},
							Fun: nil,
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1686),
									Column: int(36),
								},
								End: ast.Location{
									Line: int(1686),
									Column: int(59),
								},
							},
						},
					},
					Body: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1687),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1687),
														Column: int(23),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "min",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18222,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1687),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1687),
													Column: int(27),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "len1",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18226,
														FreeVars: ast.Identifiers{
															"len1",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1687),
																Column: int(28),
															},
															End: ast.Location{
																Line: int(1687),
																Column: int(32),
															},
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													Id: "len2",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18226,
														FreeVars: ast.Identifiers{
															"len2",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1687),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1687),
																Column: int(38),
															},
														},
													},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18222,
										FreeVars: ast.Identifiers{
											"len1",
											"len2",
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1687),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(1687),
												Column: int(39),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "minLen",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1687),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1687),
										Column: int(39),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Body: &ast.Function{
										ParenLeftFodder: ast.Fodder{},
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											Cond: &ast.Binary{
												Right: &ast.Var{
													Id: "minLen",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18237,
														FreeVars: ast.Identifiers{
															"minLen",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1689),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(1689),
																Column: int(20),
															},
														},
													},
												},
												Left: &ast.Var{
													Id: "i",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18237,
														FreeVars: ast.Identifiers{
															"i",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1689),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1689),
																Column: int(11),
															},
														},
													},
												},
												OpFodder: ast.Fodder{},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18237,
													FreeVars: ast.Identifiers{
														"i",
														"minLen",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1689),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1689),
															Column: int(20),
														},
													},
												},
												Op: ast.BinaryOp(9),
											},
											BranchTrue: &ast.Local{
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Body: &ast.Apply{
															Target: &ast.Index{
																Target: &ast.Var{
																	Id: "std",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1690),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1690),
																				Column: int(27),
																			},
																		},
																	},
																},
																Index: &ast.LiteralString{
																	Value: "__compare",
																	BlockIndent: "",
																	BlockTermIndent: "",
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{},
																		LocRange: ast.LocationRange{
																			File: nil,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																		},
																	},
																	Kind: ast.LiteralStringKind(1),
																},
																RightBracketFodder: ast.Fodder{},
																LeftBracketFodder: ast.Fodder{},
																Id: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18249,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1690),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(1690),
																			Column: int(37),
																		},
																	},
																},
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			Target: &ast.Var{
																				Id: "arr1",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18254,
																					FreeVars: ast.Identifiers{
																						"arr1",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1690),
																							Column: int(38),
																						},
																						End: ast.Location{
																							Line: int(1690),
																							Column: int(42),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18254,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1690),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(1690),
																							Column: int(44),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18254,
																				FreeVars: ast.Identifiers{
																					"arr1",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1690),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(1690),
																						Column: int(45),
																					},
																				},
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			Target: &ast.Var{
																				Id: "arr2",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18254,
																					FreeVars: ast.Identifiers{
																						"arr2",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1690),
																							Column: int(47),
																						},
																						End: ast.Location{
																							Line: int(1690),
																							Column: int(51),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18254,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1690),
																							Column: int(52),
																						},
																						End: ast.Location{
																							Line: int(1690),
																							Column: int(53),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18254,
																				FreeVars: ast.Identifiers{
																					"arr2",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1690),
																						Column: int(47),
																					},
																					End: ast.Location{
																						Line: int(1690),
																						Column: int(54),
																					},
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p18249,
																FreeVars: ast.Identifiers{
																	"arr1",
																	"arr2",
																	"i",
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1690),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1690),
																		Column: int(55),
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
														},
														EqFodder: ast.Fodder{},
														Variable: "cmpRes",
														CloseFodder: ast.Fodder{},
														Fun: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1690),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1690),
																Column: int(55),
															},
														},
													},
												},
												Body: &ast.Conditional{
													Cond: &ast.Binary{
														Right: &ast.LiteralNumber{
															OriginalString: "0",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p18237,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1691),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(1691),
																		Column: int(23),
																	},
																},
															},
														},
														Left: &ast.Var{
															Id: "cmpRes",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p18237,
																FreeVars: ast.Identifiers{
																	"cmpRes",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1691),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1691),
																		Column: int(18),
																	},
																},
															},
														},
														OpFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18237,
															FreeVars: ast.Identifiers{
																"cmpRes",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1691),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1691),
																	Column: int(23),
																},
															},
														},
														Op: ast.BinaryOp(13),
													},
													BranchTrue: &ast.Var{
														Id: "cmpRes",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																},
															},
															Ctx: p18237,
															FreeVars: ast.Identifiers{
																"cmpRes",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1692),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1692),
																	Column: int(17),
																},
															},
														},
													},
													BranchFalse: &ast.Apply{
														Target: &ast.Var{
															Id: "aux",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Comment: []string{},
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																	},
																},
																Ctx: p18237,
																FreeVars: ast.Identifiers{
																	"aux",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(1694),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1694),
																		Column: int(14),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "1",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18282,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1694),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(1694),
																						Column: int(20),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18282,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1694),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1694),
																						Column: int(16),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18282,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(1694),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1694),
																					Column: int(20),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18237,
															FreeVars: ast.Identifiers{
																"aux",
																"i",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1694),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1694),
																	Column: int(21),
																},
															},
														},
														TrailingComma: false,
														TailStrict: true,
													},
													ThenFodder: ast.Fodder{},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
															},
														},
														Ctx: p18237,
														FreeVars: ast.Identifiers{
															"aux",
															"cmpRes",
															"i",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1691),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1694),
																Column: int(21),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
														},
													},
													Ctx: p18237,
													FreeVars: ast.Identifiers{
														"arr1",
														"arr2",
														"aux",
														"i",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1690),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1694),
															Column: int(21),
														},
													},
												},
											},
											BranchFalse: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																},
															},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1696),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1696),
																	Column: int(12),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "__compare",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18237,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1696),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1696),
																Column: int(22),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "len1",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18301,
																	FreeVars: ast.Identifiers{
																		"len1",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1696),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1696),
																			Column: int(27),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																Id: "len2",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18301,
																	FreeVars: ast.Identifiers{
																		"len2",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1696),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1696),
																			Column: int(33),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18237,
													FreeVars: ast.Identifiers{
														"len1",
														"len2",
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1696),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1696),
															Column: int(34),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											ThenFodder: ast.Fodder{},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Comment: []string{},
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
												},
											},
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{
													ast.FodderElement{
														Comment: []string{},
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
													},
												},
												Ctx: p18237,
												FreeVars: ast.Identifiers{
													"arr1",
													"arr2",
													"aux",
													"i",
													"len1",
													"len2",
													"minLen",
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1689),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1696),
														Column: int(34),
													},
												},
											},
										},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												CommaFodder: nil,
												EqFodder: nil,
												DefaultArg: nil,
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1688),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1688),
														Column: int(16),
													},
												},
											},
										},
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: p18310,
											FreeVars: ast.Identifiers{
												"arr1",
												"arr2",
												"aux",
												"len1",
												"len2",
												"minLen",
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1688),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(1696),
													Column: int(34),
												},
											},
										},
										TrailingComma: false,
									},
									EqFodder: nil,
									Variable: "aux",
									CloseFodder: nil,
									Fun: nil,
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							Body: &ast.Apply{
								Target: &ast.Var{
									Id: "aux",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p18315,
										FreeVars: ast.Identifiers{
											"aux",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1697),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(1697),
												Column: int(8),
											},
										},
									},
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												OriginalString: "0",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18319,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1697),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1697),
															Column: int(10),
														},
													},
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18315,
									FreeVars: ast.Identifiers{
										"aux",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1697),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1697),
											Column: int(11),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18315,
								FreeVars: ast.Identifiers{
									"arr1",
									"arr2",
									"len1",
									"len2",
									"minLen",
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1688),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1697),
										Column: int(11),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p18315,
							FreeVars: ast.Identifiers{
								"arr1",
								"arr2",
								"len1",
								"len2",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1687),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1697),
									Column: int(11),
								},
							},
						},
					},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{
							ast.FodderElement{
								Comment: []string{},
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
							},
						},
						Ctx: p18315,
						FreeVars: ast.Identifiers{
							"arr1",
							"arr2",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1686),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1697),
								Column: int(11),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1685),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1685),
								Column: int(23),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1685),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(1685),
								Column: int(29),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1685),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1697),
					Column: int(11),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__array_less",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.Unary{
						Expr: &ast.LiteralNumber{
							OriginalString: "1",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18334,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1699),
										Column: int(66),
									},
									End: ast.Location{
										Line: int(1699),
										Column: int(67),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18334,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1699),
									Column: int(65),
								},
								End: ast.Location{
									Line: int(1699),
									Column: int(67),
								},
							},
						},
						Op: ast.UnaryOp(3),
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1699),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(1699),
											Column: int(33),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "__compare_array",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18334,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1699),
										Column: int(30),
									},
									End: ast.Location{
										Line: int(1699),
										Column: int(49),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr1",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18343,
											FreeVars: ast.Identifiers{
												"arr1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1699),
													Column: int(50),
												},
												End: ast.Location{
													Line: int(1699),
													Column: int(54),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr2",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18343,
											FreeVars: ast.Identifiers{
												"arr2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1699),
													Column: int(56),
												},
												End: ast.Location{
													Line: int(1699),
													Column: int(60),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18334,
							FreeVars: ast.Identifiers{
								"arr1",
								"arr2",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1699),
									Column: int(30),
								},
								End: ast.Location{
									Line: int(1699),
									Column: int(61),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18334,
						FreeVars: ast.Identifiers{
							"arr1",
							"arr2",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1699),
								Column: int(30),
							},
							End: ast.Location{
								Line: int(1699),
								Column: int(67),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1699),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1699),
								Column: int(20),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1699),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1699),
								Column: int(26),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1699),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1699),
					Column: int(67),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__array_greater",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralNumber{
						OriginalString: "1",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18355,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1700),
									Column: int(68),
								},
								End: ast.Location{
									Line: int(1700),
									Column: int(69),
								},
							},
						},
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1700),
											Column: int(33),
										},
										End: ast.Location{
											Line: int(1700),
											Column: int(36),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "__compare_array",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18355,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1700),
										Column: int(33),
									},
									End: ast.Location{
										Line: int(1700),
										Column: int(52),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr1",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18364,
											FreeVars: ast.Identifiers{
												"arr1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1700),
													Column: int(53),
												},
												End: ast.Location{
													Line: int(1700),
													Column: int(57),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr2",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18364,
											FreeVars: ast.Identifiers{
												"arr2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1700),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(1700),
													Column: int(63),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18355,
							FreeVars: ast.Identifiers{
								"arr1",
								"arr2",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1700),
									Column: int(33),
								},
								End: ast.Location{
									Line: int(1700),
									Column: int(64),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18355,
						FreeVars: ast.Identifiers{
							"arr1",
							"arr2",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1700),
								Column: int(33),
							},
							End: ast.Location{
								Line: int(1700),
								Column: int(69),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1700),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1700),
								Column: int(23),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1700),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(1700),
								Column: int(29),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1700),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1700),
					Column: int(69),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__array_less_or_equal",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralNumber{
						OriginalString: "0",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18376,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1701),
									Column: int(74),
								},
								End: ast.Location{
									Line: int(1701),
									Column: int(75),
								},
							},
						},
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1701),
											Column: int(39),
										},
										End: ast.Location{
											Line: int(1701),
											Column: int(42),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "__compare_array",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18376,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1701),
										Column: int(39),
									},
									End: ast.Location{
										Line: int(1701),
										Column: int(58),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr1",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18385,
											FreeVars: ast.Identifiers{
												"arr1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1701),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(1701),
													Column: int(63),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr2",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18385,
											FreeVars: ast.Identifiers{
												"arr2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1701),
													Column: int(65),
												},
												End: ast.Location{
													Line: int(1701),
													Column: int(69),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18376,
							FreeVars: ast.Identifiers{
								"arr1",
								"arr2",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1701),
									Column: int(39),
								},
								End: ast.Location{
									Line: int(1701),
									Column: int(70),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18376,
						FreeVars: ast.Identifiers{
							"arr1",
							"arr2",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1701),
								Column: int(39),
							},
							End: ast.Location{
								Line: int(1701),
								Column: int(75),
							},
						},
					},
					Op: ast.BinaryOp(10),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1701),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(1701),
								Column: int(29),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1701),
								Column: int(31),
							},
							End: ast.Location{
								Line: int(1701),
								Column: int(35),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1701),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1701),
					Column: int(75),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "__array_greater_or_equal",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralNumber{
						OriginalString: "0",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18397,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1702),
									Column: int(77),
								},
								End: ast.Location{
									Line: int(1702),
									Column: int(78),
								},
							},
						},
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1702),
											Column: int(42),
										},
										End: ast.Location{
											Line: int(1702),
											Column: int(45),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "__compare_array",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18397,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1702),
										Column: int(42),
									},
									End: ast.Location{
										Line: int(1702),
										Column: int(61),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr1",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18406,
											FreeVars: ast.Identifiers{
												"arr1",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1702),
													Column: int(62),
												},
												End: ast.Location{
													Line: int(1702),
													Column: int(66),
												},
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "arr2",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18406,
											FreeVars: ast.Identifiers{
												"arr2",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1702),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(1702),
													Column: int(72),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18397,
							FreeVars: ast.Identifiers{
								"arr1",
								"arr2",
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1702),
									Column: int(42),
								},
								End: ast.Location{
									Line: int(1702),
									Column: int(73),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18397,
						FreeVars: ast.Identifiers{
							"arr1",
							"arr2",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1702),
								Column: int(42),
							},
							End: ast.Location{
								Line: int(1702),
								Column: int(78),
							},
						},
					},
					Op: ast.BinaryOp(8),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr1",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1702),
								Column: int(28),
							},
							End: ast.Location{
								Line: int(1702),
								Column: int(32),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr2",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1702),
								Column: int(34),
							},
							End: ast.Location{
								Line: int(1702),
								Column: int(38),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1702),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1702),
					Column: int(78),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "sum",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1704),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(1704),
										Column: int(17),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "foldl",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18421,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1704),
									Column: int(14),
								},
								End: ast.Location{
									Line: int(1704),
									Column: int(23),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									ParenLeftFodder: ast.Fodder{},
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										Right: &ast.Var{
											Id: "b",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18427,
												FreeVars: ast.Identifiers{
													"b",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1704),
														Column: int(43),
													},
													End: ast.Location{
														Line: int(1704),
														Column: int(44),
													},
												},
											},
										},
										Left: &ast.Var{
											Id: "a",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18427,
												FreeVars: ast.Identifiers{
													"a",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(1704),
														Column: int(39),
													},
													End: ast.Location{
														Line: int(1704),
														Column: int(40),
													},
												},
											},
										},
										OpFodder: ast.Fodder{},
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18427,
											FreeVars: ast.Identifiers{
												"a",
												"b",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1704),
													Column: int(39),
												},
												End: ast.Location{
													Line: int(1704),
													Column: int(44),
												},
											},
										},
										Op: ast.BinaryOp(3),
									},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "a",
											CommaFodder: ast.Fodder{},
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1704),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(1704),
													Column: int(34),
												},
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "b",
											CommaFodder: nil,
											EqFodder: nil,
											DefaultArg: nil,
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1704),
													Column: int(36),
												},
												End: ast.Location{
													Line: int(1704),
													Column: int(37),
												},
											},
										},
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18433,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1704),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1704),
												Column: int(44),
											},
										},
									},
									TrailingComma: false,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									Id: "arr",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18433,
										FreeVars: ast.Identifiers{
											"arr",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1704),
												Column: int(46),
											},
											End: ast.Location{
												Line: int(1704),
												Column: int(49),
											},
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralNumber{
									OriginalString: "0",
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18433,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1704),
												Column: int(51),
											},
											End: ast.Location{
												Line: int(1704),
												Column: int(52),
											},
										},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18421,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1704),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1704),
								Column: int(53),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1704),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1704),
								Column: int(10),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1704),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1704),
					Column: int(53),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "xor",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.Var{
						Id: "y",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18444,
							FreeVars: ast.Identifiers{
								"y",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1706),
									Column: int(20),
								},
								End: ast.Location{
									Line: int(1706),
									Column: int(21),
								},
							},
						},
					},
					Left: &ast.Var{
						Id: "x",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18444,
							FreeVars: ast.Identifiers{
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1706),
									Column: int(15),
								},
								End: ast.Location{
									Line: int(1706),
									Column: int(16),
								},
							},
						},
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18444,
						FreeVars: ast.Identifiers{
							"x",
							"y",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1706),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1706),
								Column: int(21),
							},
						},
					},
					Op: ast.BinaryOp(13),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1706),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1706),
								Column: int(8),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "y",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1706),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1706),
								Column: int(11),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1706),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1706),
					Column: int(21),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "xnor",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.Var{
						Id: "y",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18454,
							FreeVars: ast.Identifiers{
								"y",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1708),
									Column: int(21),
								},
								End: ast.Location{
									Line: int(1708),
									Column: int(22),
								},
							},
						},
					},
					Left: &ast.Var{
						Id: "x",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18454,
							FreeVars: ast.Identifiers{
								"x",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1708),
									Column: int(16),
								},
								End: ast.Location{
									Line: int(1708),
									Column: int(17),
								},
							},
						},
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18454,
						FreeVars: ast.Identifiers{
							"x",
							"y",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1708),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1708),
								Column: int(22),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1708),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1708),
								Column: int(9),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "y",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1708),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1708),
								Column: int(12),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1708),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1708),
					Column: int(22),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "round",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					Target: &ast.Index{
						Target: &ast.Var{
							Id: "std",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1710),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(1710),
										Column: int(17),
									},
								},
							},
						},
						Index: &ast.LiteralString{
							Value: "floor",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: nil,
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						RightBracketFodder: ast.Fodder{},
						LeftBracketFodder: ast.Fodder{},
						Id: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18467,
							FreeVars: ast.Identifiers{
								"std",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1710),
									Column: int(14),
								},
								End: ast.Location{
									Line: int(1710),
									Column: int(23),
								},
							},
						},
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Binary{
									Right: &ast.LiteralNumber{
										OriginalString: "0.5",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18472,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1710),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(1710),
													Column: int(31),
												},
											},
										},
									},
									Left: &ast.Var{
										Id: "x",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18472,
											FreeVars: ast.Identifiers{
												"x",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1710),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(1710),
													Column: int(25),
												},
											},
										},
									},
									OpFodder: ast.Fodder{},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18472,
										FreeVars: ast.Identifiers{
											"x",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(1710),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1710),
												Column: int(31),
											},
										},
									},
									Op: ast.BinaryOp(3),
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18467,
						FreeVars: ast.Identifiers{
							"std",
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1710),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1710),
								Column: int(32),
							},
						},
					},
					TrailingComma: false,
					TailStrict: false,
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1710),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(1710),
								Column: int(10),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1710),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1710),
					Column: int(32),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
		ast.DesugaredObjectField{
			Name: &ast.LiteralString{
				Value: "isEmpty",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					Right: &ast.LiteralNumber{
						OriginalString: "0",
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18483,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1712),
									Column: int(37),
								},
								End: ast.Location{
									Line: int(1712),
									Column: int(38),
								},
							},
						},
					},
					Left: &ast.Apply{
						Target: &ast.Index{
							Target: &ast.Var{
								Id: "std",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1712),
											Column: int(18),
										},
										End: ast.Location{
											Line: int(1712),
											Column: int(21),
										},
									},
								},
							},
							Index: &ast.LiteralString{
								Value: "length",
								BlockIndent: "",
								BlockTermIndent: "",
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
								Kind: ast.LiteralStringKind(1),
							},
							RightBracketFodder: ast.Fodder{},
							LeftBracketFodder: ast.Fodder{},
							Id: nil,
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18483,
								FreeVars: ast.Identifiers{
									"std",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1712),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(1712),
										Column: int(28),
									},
								},
							},
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										Id: "str",
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18492,
											FreeVars: ast.Identifiers{
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(1712),
													Column: int(29),
												},
												End: ast.Location{
													Line: int(1712),
													Column: int(32),
												},
											},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18483,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(1712),
									Column: int(18),
								},
								End: ast.Location{
									Line: int(1712),
									Column: int(33),
								},
							},
						},
						TrailingComma: false,
						TailStrict: false,
					},
					OpFodder: ast.Fodder{},
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18483,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1712),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1712),
								Column: int(38),
							},
						},
					},
					Op: ast.BinaryOp(12),
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1712),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1712),
								Column: int(14),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p23,
					FreeVars: ast.Identifiers{
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1712),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1712),
					Column: int(38),
				},
			},
			Hide: ast.ObjectFieldHide(0),
			PlusSuper: false,
		},
	},
	Locals: ast.LocalBinds{
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.Self{
				NodeBase: ast.NodeBase{
					Fodder: ast.Fodder{},
					Ctx: p18500,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: p8,
						FileName: "",
						Begin: ast.Location{
							Line: int(25),
							Column: int(15),
						},
						End: ast.Location{
							Line: int(25),
							Column: int(19),
						},
					},
				},
			},
			EqFodder: nil,
			Variable: "std",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(25),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(25),
					Column: int(19),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Var{
					Id: "x",
					NodeBase: ast.NodeBase{
						Fodder: ast.Fodder{},
						Ctx: p18503,
						FreeVars: ast.Identifiers{
							"x",
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(26),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(26),
								Column: int(27),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(26),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(26),
								Column: int(24),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: ast.Fodder{},
					Ctx: p18506,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: p8,
						FileName: "",
						Begin: ast.Location{
							Line: int(26),
							Column: int(14),
						},
						End: ast.Location{
							Line: int(26),
							Column: int(27),
						},
					},
				},
				TrailingComma: false,
			},
			EqFodder: nil,
			Variable: "id",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(26),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(26),
					Column: int(27),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.Function{
				ParenLeftFodder: ast.Fodder{},
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					Cond: &ast.Binary{
						Right: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "16",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18512,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(77),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(34),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "base",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18514,
									FreeVars: ast.Identifiers{
										"base",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(77),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(28),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18516,
								FreeVars: ast.Identifiers{
									"base",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(77),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(77),
										Column: int(34),
									},
								},
							},
							Op: ast.BinaryOp(10),
						},
						Left: &ast.Binary{
							Right: &ast.LiteralNumber{
								OriginalString: "0",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18520,
									FreeVars: ast.Identifiers{},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(77),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(20),
										},
									},
								},
							},
							Left: &ast.Var{
								Id: "base",
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{},
									Ctx: p18522,
									FreeVars: ast.Identifiers{
										"base",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(77),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(16),
										},
									},
								},
							},
							OpFodder: ast.Fodder{},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18524,
								FreeVars: ast.Identifiers{
									"base",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(77),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(77),
										Column: int(20),
									},
								},
							},
							Op: ast.BinaryOp(7),
						},
						OpFodder: ast.Fodder{},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{},
							Ctx: p18526,
							FreeVars: ast.Identifiers{
								"base",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(77),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(77),
									Column: int(34),
								},
							},
						},
						Op: ast.BinaryOp(17),
					},
					BranchTrue: &ast.Local{
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Body: &ast.Apply{
									Target: &ast.Index{
										Target: &ast.Var{
											Id: "std",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(79),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(79),
														Column: int(26),
													},
												},
											},
										},
										Index: &ast.LiteralString{
											Value: "codepoint",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: nil,
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										RightBracketFodder: ast.Fodder{},
										LeftBracketFodder: ast.Fodder{},
										Id: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18535,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(79),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(79),
													Column: int(36),
												},
											},
										},
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													Value: "0",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18539,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(79),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(79),
																Column: int(40),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												CommaFodder: nil,
											},
										},
										Named: []ast.NamedArgument{},
									},
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{},
										Ctx: p18540,
										FreeVars: ast.Identifiers{
											"std",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(79),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(79),
												Column: int(41),
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
								},
								EqFodder: ast.Fodder{},
								Variable: "zero_code",
								CloseFodder: ast.Fodder{},
								Fun: nil,
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(79),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(79),
										Column: int(41),
									},
								},
							},
						},
						Body: &ast.Local{
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(80),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(80),
															Column: int(29),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "codepoint",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18549,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(80),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(80),
														Column: int(39),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														Value: "A",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18553,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(80),
																	Column: int(40),
																},
																End: ast.Location{
																	Line: int(80),
																	Column: int(43),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													CommaFodder: nil,
												},
											},
											Named: []ast.NamedArgument{},
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18554,
											FreeVars: ast.Identifiers{
												"std",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(80),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(80),
													Column: int(44),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									EqFodder: ast.Fodder{},
									Variable: "upper_a_code",
									CloseFodder: ast.Fodder{},
									Fun: nil,
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(80),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(80),
											Column: int(44),
										},
									},
								},
							},
							Body: &ast.Local{
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Body: &ast.Apply{
											Target: &ast.Index{
												Target: &ast.Var{
													Id: "std",
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(81),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(81),
																Column: int(29),
															},
														},
													},
												},
												Index: &ast.LiteralString{
													Value: "codepoint",
													BlockIndent: "",
													BlockTermIndent: "",
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													Kind: ast.LiteralStringKind(1),
												},
												RightBracketFodder: ast.Fodder{},
												LeftBracketFodder: ast.Fodder{},
												Id: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18563,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(81),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(81),
															Column: int(39),
														},
													},
												},
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															Value: "a",
															BlockIndent: "",
															BlockTermIndent: "",
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p18567,
																FreeVars: ast.Identifiers{},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(81),
																		Column: int(40),
																	},
																	End: ast.Location{
																		Line: int(81),
																		Column: int(43),
																	},
																},
															},
															Kind: ast.LiteralStringKind(1),
														},
														CommaFodder: nil,
													},
												},
												Named: []ast.NamedArgument{},
											},
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18568,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(81),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(81),
														Column: int(44),
													},
												},
											},
											TrailingComma: false,
											TailStrict: false,
										},
										EqFodder: ast.Fodder{},
										Variable: "lower_a_code",
										CloseFodder: ast.Fodder{},
										Fun: nil,
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(81),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(81),
												Column: int(44),
											},
										},
									},
								},
								Body: &ast.Local{
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Body: &ast.Function{
												ParenLeftFodder: ast.Fodder{},
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Local{
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Body: &ast.Apply{
																Target: &ast.Index{
																	Target: &ast.Var{
																		Id: "std",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(83),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(83),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Index: &ast.LiteralString{
																		Value: "codepoint",
																		BlockIndent: "",
																		BlockTermIndent: "",
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																		Kind: ast.LiteralStringKind(1),
																	},
																	RightBracketFodder: ast.Fodder{},
																	LeftBracketFodder: ast.Fodder{},
																	Id: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18580,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(83),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(83),
																				Column: int(33),
																			},
																		},
																	},
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				Id: "char",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18584,
																					FreeVars: ast.Identifiers{
																						"char",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(83),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(83),
																							Column: int(38),
																						},
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: []ast.NamedArgument{},
																},
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18586,
																	FreeVars: ast.Identifiers{
																		"char",
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(83),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(83),
																			Column: int(39),
																		},
																	},
																},
																TrailingComma: false,
																TailStrict: false,
															},
															EqFodder: ast.Fodder{},
															Variable: "code",
															CloseFodder: ast.Fodder{},
															Fun: nil,
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(83),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(83),
																	Column: int(39),
																},
															},
														},
													},
													Body: &ast.Local{
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	Cond: &ast.Binary{
																		Right: &ast.Var{
																			Id: "lower_a_code",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18593,
																				FreeVars: ast.Identifiers{
																					"lower_a_code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(84),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(84),
																						Column: int(44),
																					},
																				},
																			},
																		},
																		Left: &ast.Var{
																			Id: "code",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18596,
																				FreeVars: ast.Identifiers{
																					"code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(84),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(84),
																						Column: int(28),
																					},
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18598,
																			FreeVars: ast.Identifiers{
																				"code",
																				"lower_a_code",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(84),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(84),
																					Column: int(44),
																				},
																			},
																		},
																		Op: ast.BinaryOp(8),
																	},
																	BranchTrue: &ast.Binary{
																		Right: &ast.LiteralNumber{
																			OriginalString: "10",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18602,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(85),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(85),
																						Column: int(33),
																					},
																				},
																			},
																		},
																		Left: &ast.Binary{
																			Right: &ast.Var{
																				Id: "lower_a_code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18605,
																					FreeVars: ast.Identifiers{
																						"lower_a_code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(85),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(85),
																							Column: int(28),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p18609,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(85),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(85),
																							Column: int(13),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18611,
																				FreeVars: ast.Identifiers{
																					"code",
																					"lower_a_code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(85),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(85),
																						Column: int(28),
																					},
																				},
																			},
																			Op: ast.BinaryOp(4),
																		},
																		OpFodder: ast.Fodder{},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18613,
																			FreeVars: ast.Identifiers{
																				"code",
																				"lower_a_code",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(85),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(85),
																					Column: int(33),
																				},
																			},
																		},
																		Op: ast.BinaryOp(3),
																	},
																	BranchFalse: &ast.Conditional{
																		Cond: &ast.Binary{
																			Right: &ast.Var{
																				Id: "upper_a_code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18618,
																					FreeVars: ast.Identifiers{
																						"upper_a_code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(86),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(86),
																							Column: int(35),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18621,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(86),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(86),
																							Column: int(19),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18623,
																				FreeVars: ast.Identifiers{
																					"code",
																					"upper_a_code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(86),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(86),
																						Column: int(35),
																					},
																				},
																			},
																			Op: ast.BinaryOp(8),
																		},
																		BranchTrue: &ast.Binary{
																			Right: &ast.LiteralNumber{
																				OriginalString: "10",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18627,
																					FreeVars: ast.Identifiers{},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(87),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(87),
																							Column: int(33),
																						},
																					},
																				},
																			},
																			Left: &ast.Binary{
																				Right: &ast.Var{
																					Id: "upper_a_code",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p18630,
																						FreeVars: ast.Identifiers{
																							"upper_a_code",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(87),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(87),
																								Column: int(28),
																							},
																						},
																					},
																				},
																				Left: &ast.Var{
																					Id: "code",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Comment: []string{},
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																							},
																						},
																						Ctx: p18634,
																						FreeVars: ast.Identifiers{
																							"code",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(87),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(87),
																								Column: int(13),
																							},
																						},
																					},
																				},
																				OpFodder: ast.Fodder{},
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18636,
																					FreeVars: ast.Identifiers{
																						"code",
																						"upper_a_code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(87),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(87),
																							Column: int(28),
																						},
																					},
																				},
																				Op: ast.BinaryOp(4),
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18638,
																				FreeVars: ast.Identifiers{
																					"code",
																					"upper_a_code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(87),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(87),
																						Column: int(33),
																					},
																				},
																			},
																			Op: ast.BinaryOp(3),
																		},
																		BranchFalse: &ast.Binary{
																			Right: &ast.Var{
																				Id: "zero_code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18642,
																					FreeVars: ast.Identifiers{
																						"zero_code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(89),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(89),
																							Column: int(25),
																						},
																					},
																				},
																			},
																			Left: &ast.Var{
																				Id: "code",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Comment: []string{},
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																						},
																					},
																					Ctx: p18646,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(89),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(89),
																							Column: int(13),
																						},
																					},
																				},
																			},
																			OpFodder: ast.Fodder{},
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18648,
																				FreeVars: ast.Identifiers{
																					"code",
																					"zero_code",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(89),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(89),
																						Column: int(25),
																					},
																				},
																			},
																			Op: ast.BinaryOp(4),
																		},
																		ThenFodder: ast.Fodder{},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Comment: []string{},
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																			},
																		},
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18651,
																			FreeVars: ast.Identifiers{
																				"code",
																				"upper_a_code",
																				"zero_code",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(86),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(89),
																					Column: int(25),
																				},
																			},
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Comment: []string{},
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																		},
																	},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18654,
																		FreeVars: ast.Identifiers{
																			"code",
																			"lower_a_code",
																			"upper_a_code",
																			"zero_code",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(84),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(89),
																				Column: int(25),
																			},
																		},
																	},
																},
																EqFodder: ast.Fodder{},
																Variable: "digit",
																CloseFodder: ast.Fodder{},
																Fun: nil,
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(84),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(89),
																		Column: int(25),
																	},
																},
															},
														},
														Body: &ast.Conditional{
															Cond: &ast.Binary{
																Right: &ast.Binary{
																	Right: &ast.Var{
																		Id: "base",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18660,
																			FreeVars: ast.Identifiers{
																				"base",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(40),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "digit",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18663,
																			FreeVars: ast.Identifiers{
																				"digit",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(33),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18665,
																		FreeVars: ast.Identifiers{
																			"base",
																			"digit",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(90),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(90),
																				Column: int(40),
																			},
																		},
																	},
																	Op: ast.BinaryOp(9),
																},
																Left: &ast.Binary{
																	Right: &ast.LiteralNumber{
																		OriginalString: "0",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18669,
																			FreeVars: ast.Identifiers{},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(24),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "digit",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18671,
																			FreeVars: ast.Identifiers{
																				"digit",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(19),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18673,
																		FreeVars: ast.Identifiers{
																			"digit",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(90),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(90),
																				Column: int(24),
																			},
																		},
																	},
																	Op: ast.BinaryOp(8),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18675,
																	FreeVars: ast.Identifiers{
																		"base",
																		"digit",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(90),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(90),
																			Column: int(40),
																		},
																	},
																},
																Op: ast.BinaryOp(17),
															},
															BranchTrue: &ast.Binary{
																Right: &ast.Var{
																	Id: "digit",
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18679,
																		FreeVars: ast.Identifiers{
																			"digit",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(91),
																				Column: int(26),
																			},
																			End: ast.Location{
																				Line: int(91),
																				Column: int(31),
																			},
																		},
																	},
																},
																Left: &ast.Binary{
																	Right: &ast.Var{
																		Id: "aggregate",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18683,
																			FreeVars: ast.Identifiers{
																				"aggregate",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(91),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(91),
																					Column: int(23),
																				},
																			},
																		},
																	},
																	Left: &ast.Var{
																		Id: "base",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Comment: []string{},
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																				},
																			},
																			Ctx: p18687,
																			FreeVars: ast.Identifiers{
																				"base",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(91),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(91),
																					Column: int(11),
																				},
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	NodeBase: ast.NodeBase{
																		Fodder: ast.Fodder{},
																		Ctx: p18689,
																		FreeVars: ast.Identifiers{
																			"aggregate",
																			"base",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(91),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(91),
																				Column: int(23),
																			},
																		},
																	},
																	Op: ast.BinaryOp(0),
																},
																OpFodder: ast.Fodder{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18691,
																	FreeVars: ast.Identifiers{
																		"aggregate",
																		"base",
																		"digit",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(91),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(91),
																			Column: int(31),
																		},
																	},
																},
																Op: ast.BinaryOp(3),
															},
															BranchFalse: &ast.Error{
																Expr: &ast.Apply{
																	Target: &ast.Index{
																		Target: &ast.Var{
																			Id: "$std",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$std",
																				},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																		},
																		Index: &ast.LiteralString{
																			Value: "mod",
																			BlockIndent: "",
																			BlockTermIndent: "",
																			NodeBase: ast.NodeBase{
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{},
																				LocRange: ast.LocationRange{
																					File: nil,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																				},
																			},
																			Kind: ast.LiteralStringKind(1),
																		},
																		RightBracketFodder: nil,
																		LeftBracketFodder: nil,
																		Id: nil,
																		NodeBase: ast.NodeBase{
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$std",
																			},
																			LocRange: ast.LocationRange{
																				File: nil,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																			},
																		},
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					Value: "%s is not a base %d integer",
																					BlockIndent: "",
																					BlockTermIndent: "",
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p18702,
																						FreeVars: ast.Identifiers{},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(90),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(90),
																								Column: int(72),
																							},
																						},
																					},
																					Kind: ast.LiteralStringKind(1),
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Array{
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "str",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p18706,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(90),
																											Column: int(76),
																										},
																										End: ast.Location{
																											Line: int(90),
																											Column: int(79),
																										},
																									},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								Id: "base",
																								NodeBase: ast.NodeBase{
																									Fodder: ast.Fodder{},
																									Ctx: p18709,
																									FreeVars: ast.Identifiers{
																										"base",
																									},
																									LocRange: ast.LocationRange{
																										File: p8,
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(90),
																											Column: int(81),
																										},
																										End: ast.Location{
																											Line: int(90),
																											Column: int(85),
																										},
																									},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					CloseFodder: ast.Fodder{},
																					NodeBase: ast.NodeBase{
																						Fodder: ast.Fodder{},
																						Ctx: p18711,
																						FreeVars: ast.Identifiers{
																							"base",
																							"str",
																						},
																						LocRange: ast.LocationRange{
																							File: p8,
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(90),
																								Column: int(75),
																							},
																							End: ast.Location{
																								Line: int(90),
																								Column: int(86),
																							},
																						},
																					},
																					TrailingComma: false,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	FodderRight: nil,
																	TailStrictFodder: nil,
																	NodeBase: ast.NodeBase{
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$std",
																			"base",
																			"str",
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(90),
																				Column: int(43),
																			},
																			End: ast.Location{
																				Line: int(90),
																				Column: int(86),
																			},
																		},
																	},
																	TrailingComma: false,
																	TailStrict: false,
																},
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$std",
																		"base",
																		"str",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(90),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(91),
																			Column: int(31),
																		},
																	},
																},
															},
															ThenFodder: nil,
															ElseFodder: nil,
															NodeBase: ast.NodeBase{
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"$std",
																	"aggregate",
																	"base",
																	"digit",
																	"str",
																},
																LocRange: ast.LocationRange{
																	File: nil,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																},
															},
														},
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Comment: []string{},
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																},
															},
															Ctx: p18717,
															FreeVars: ast.Identifiers{
																"$std",
																"aggregate",
																"base",
																"code",
																"lower_a_code",
																"str",
																"upper_a_code",
																"zero_code",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(84),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(91),
																	Column: int(31),
																},
															},
														},
													},
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{
															ast.FodderElement{
																Comment: []string{},
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
															},
														},
														Ctx: p18720,
														FreeVars: ast.Identifiers{
															"$std",
															"aggregate",
															"base",
															"char",
															"lower_a_code",
															"std",
															"str",
															"upper_a_code",
															"zero_code",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(83),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(91),
																Column: int(31),
															},
														},
													},
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "aggregate",
														CommaFodder: ast.Fodder{},
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(82),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(82),
																Column: int(29),
															},
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "char",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(82),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(82),
																Column: int(35),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: p18723,
													FreeVars: ast.Identifiers{
														"$std",
														"base",
														"lower_a_code",
														"std",
														"str",
														"upper_a_code",
														"zero_code",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(82),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(91),
															Column: int(31),
														},
													},
												},
												TrailingComma: false,
											},
											EqFodder: nil,
											Variable: "addDigit",
											CloseFodder: nil,
											Fun: nil,
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Body: &ast.Apply{
										Target: &ast.Index{
											Target: &ast.Var{
												Id: "std",
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{
														ast.FodderElement{
															Comment: []string{},
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(4),
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(92),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(92),
															Column: int(8),
														},
													},
												},
											},
											Index: &ast.LiteralString{
												Value: "foldl",
												BlockIndent: "",
												BlockTermIndent: "",
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												Kind: ast.LiteralStringKind(1),
											},
											RightBracketFodder: ast.Fodder{},
											LeftBracketFodder: ast.Fodder{},
											Id: nil,
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18731,
												FreeVars: ast.Identifiers{
													"std",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(92),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(92),
														Column: int(14),
													},
												},
											},
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														Id: "addDigit",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18735,
															FreeVars: ast.Identifiers{
																"addDigit",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(23),
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														Target: &ast.Index{
															Target: &ast.Var{
																Id: "std",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(92),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(92),
																			Column: int(28),
																		},
																	},
																},
															},
															Index: &ast.LiteralString{
																Value: "stringChars",
																BlockIndent: "",
																BlockTermIndent: "",
																NodeBase: ast.NodeBase{
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: nil,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																	},
																},
																Kind: ast.LiteralStringKind(1),
															},
															RightBracketFodder: ast.Fodder{},
															LeftBracketFodder: ast.Fodder{},
															Id: nil,
															NodeBase: ast.NodeBase{
																Fodder: ast.Fodder{},
																Ctx: p18742,
																FreeVars: ast.Identifiers{
																	"std",
																},
																LocRange: ast.LocationRange{
																	File: p8,
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(92),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(92),
																		Column: int(40),
																	},
																},
															},
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		Id: "str",
																		NodeBase: ast.NodeBase{
																			Fodder: ast.Fodder{},
																			Ctx: p18746,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																			LocRange: ast.LocationRange{
																				File: p8,
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(92),
																					Column: int(41),
																				},
																				End: ast.Location{
																					Line: int(92),
																					Column: int(44),
																				},
																			},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: []ast.NamedArgument{},
														},
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18748,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(45),
																},
															},
														},
														TrailingComma: false,
														TailStrict: false,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														OriginalString: "0",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: p18751,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(47),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(48),
																},
															},
														},
													},
													CommaFodder: nil,
												},
											},
											Named: []ast.NamedArgument{},
										},
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
										NodeBase: ast.NodeBase{
											Fodder: ast.Fodder{},
											Ctx: p18752,
											FreeVars: ast.Identifiers{
												"addDigit",
												"std",
												"str",
											},
											LocRange: ast.LocationRange{
												File: p8,
												FileName: "",
												Begin: ast.Location{
													Line: int(92),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(92),
													Column: int(49),
												},
											},
										},
										TrailingComma: false,
										TailStrict: false,
									},
									NodeBase: ast.NodeBase{
										Fodder: ast.Fodder{
											ast.FodderElement{
												Comment: []string{},
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
											},
										},
										Ctx: p18755,
										FreeVars: ast.Identifiers{
											"$std",
											"base",
											"lower_a_code",
											"std",
											"str",
											"upper_a_code",
											"zero_code",
										},
										LocRange: ast.LocationRange{
											File: p8,
											FileName: "",
											Begin: ast.Location{
												Line: int(82),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(92),
												Column: int(49),
											},
										},
									},
								},
								NodeBase: ast.NodeBase{
									Fodder: ast.Fodder{
										ast.FodderElement{
											Comment: []string{},
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
										},
									},
									Ctx: p18758,
									FreeVars: ast.Identifiers{
										"$std",
										"base",
										"std",
										"str",
										"upper_a_code",
										"zero_code",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(81),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(92),
											Column: int(49),
										},
									},
								},
							},
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18761,
								FreeVars: ast.Identifiers{
									"$std",
									"base",
									"std",
									"str",
									"zero_code",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(80),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(92),
										Column: int(49),
									},
								},
							},
						},
						NodeBase: ast.NodeBase{
							Fodder: ast.Fodder{
								ast.FodderElement{
									Comment: []string{},
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
								},
								ast.FodderElement{
									Comment: []string{
										"// These codepoints are in ascending order:",
									},
									Kind: ast.FodderKind(2),
									Blanks: int(0),
									Indent: int(4),
								},
							},
							Ctx: p18765,
							FreeVars: ast.Identifiers{
								"$std",
								"base",
								"std",
								"str",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(79),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(92),
									Column: int(49),
								},
							},
						},
					},
					BranchFalse: &ast.Error{
						Expr: &ast.Apply{
							Target: &ast.Index{
								Target: &ast.Var{
									Id: "$std",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								Index: &ast.LiteralString{
									Value: "mod",
									BlockIndent: "",
									BlockTermIndent: "",
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
									Kind: ast.LiteralStringKind(1),
								},
								RightBracketFodder: nil,
								LeftBracketFodder: nil,
								Id: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
									},
									LocRange: ast.LocationRange{
										File: nil,
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
									},
								},
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											Value: "integer base %d invalid",
											BlockIndent: "",
											BlockTermIndent: "",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18776,
												FreeVars: ast.Identifiers{},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(77),
														Column: int(37),
													},
													End: ast.Location{
														Line: int(77),
														Column: int(62),
													},
												},
											},
											Kind: ast.LiteralStringKind(1),
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											Id: "base",
											NodeBase: ast.NodeBase{
												Fodder: ast.Fodder{},
												Ctx: p18778,
												FreeVars: ast.Identifiers{
													"base",
												},
												LocRange: ast.LocationRange{
													File: p8,
													FileName: "",
													Begin: ast.Location{
														Line: int(77),
														Column: int(65),
													},
													End: ast.Location{
														Line: int(77),
														Column: int(69),
													},
												},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							FodderRight: nil,
							TailStrictFodder: nil,
							NodeBase: ast.NodeBase{
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$std",
									"base",
								},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(77),
										Column: int(37),
									},
									End: ast.Location{
										Line: int(77),
										Column: int(69),
									},
								},
							},
							TrailingComma: false,
							TailStrict: false,
						},
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
								"base",
							},
							LocRange: ast.LocationRange{
								File: p8,
								FileName: "",
								Begin: ast.Location{
									Line: int(77),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(92),
									Column: int(49),
								},
							},
						},
					},
					ThenFodder: nil,
					ElseFodder: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
							"base",
							"std",
							"str",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						CommaFodder: ast.Fodder{},
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(76),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(76),
								Column: int(22),
							},
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "base",
						CommaFodder: nil,
						EqFodder: nil,
						DefaultArg: nil,
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(76),
								Column: int(24),
							},
							End: ast.Location{
								Line: int(76),
								Column: int(28),
							},
						},
					},
				},
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: p18784,
					FreeVars: ast.Identifiers{
						"$std",
						"std",
					},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
				TrailingComma: false,
			},
			EqFodder: nil,
			Variable: "parse_nat",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(76),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(92),
					Column: int(49),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.DesugaredObject{
				Asserts: ast.Nodes{},
				Fields: ast.DesugaredObjectFields{
					ast.DesugaredObjectField{
						Name: &ast.LiteralString{
							Value: "<",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18790,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1017),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1017),
										Column: int(8),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Body: &ast.LiteralString{
							Value: "&lt;",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18792,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1017),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1017),
										Column: int(16),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1017),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1017),
								Column: int(16),
							},
						},
						Hide: ast.ObjectFieldHide(1),
						PlusSuper: false,
					},
					ast.DesugaredObjectField{
						Name: &ast.LiteralString{
							Value: ">",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18795,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1018),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1018),
										Column: int(8),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Body: &ast.LiteralString{
							Value: "&gt;",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18797,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1018),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1018),
										Column: int(16),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1018),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1018),
								Column: int(16),
							},
						},
						Hide: ast.ObjectFieldHide(1),
						PlusSuper: false,
					},
					ast.DesugaredObjectField{
						Name: &ast.LiteralString{
							Value: "&",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18800,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1019),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1019),
										Column: int(8),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Body: &ast.LiteralString{
							Value: "&amp;",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18802,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1019),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1019),
										Column: int(17),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1019),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1019),
								Column: int(17),
							},
						},
						Hide: ast.ObjectFieldHide(1),
						PlusSuper: false,
					},
					ast.DesugaredObjectField{
						Name: &ast.LiteralString{
							Value: "\"",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18805,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1020),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1020),
										Column: int(8),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Body: &ast.LiteralString{
							Value: "&quot;",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18807,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1020),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1020),
										Column: int(18),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1020),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1020),
								Column: int(18),
							},
						},
						Hide: ast.ObjectFieldHide(1),
						PlusSuper: false,
					},
					ast.DesugaredObjectField{
						Name: &ast.LiteralString{
							Value: "'",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{
									ast.FodderElement{
										Comment: []string{},
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
									},
								},
								Ctx: p18810,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1021),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1021),
										Column: int(8),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						Body: &ast.LiteralString{
							Value: "&apos;",
							BlockIndent: "",
							BlockTermIndent: "",
							NodeBase: ast.NodeBase{
								Fodder: ast.Fodder{},
								Ctx: p18812,
								FreeVars: ast.Identifiers{},
								LocRange: ast.LocationRange{
									File: p8,
									FileName: "",
									Begin: ast.Location{
										Line: int(1021),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1021),
										Column: int(18),
									},
								},
							},
							Kind: ast.LiteralStringKind(1),
						},
						LocRange: ast.LocationRange{
							File: p8,
							FileName: "",
							Begin: ast.Location{
								Line: int(1021),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1021),
								Column: int(18),
							},
						},
						Hide: ast.ObjectFieldHide(1),
						PlusSuper: false,
					},
				},
				Locals: ast.LocalBinds{},
				NodeBase: ast.NodeBase{
					Fodder: ast.Fodder{},
					Ctx: p18813,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: p8,
						FileName: "",
						Begin: ast.Location{
							Line: int(1016),
							Column: int(23),
						},
						End: ast.Location{
							Line: int(1022),
							Column: int(4),
						},
					},
				},
			},
			EqFodder: nil,
			Variable: "xml_escapes",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1016),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(1022),
					Column: int(4),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.LiteralString{
				Value: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",
				BlockIndent: "",
				BlockTermIndent: "",
				NodeBase: ast.NodeBase{
					Fodder: ast.Fodder{},
					Ctx: p18815,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: p8,
						FileName: "",
						Begin: ast.Location{
							Line: int(1333),
							Column: int(24),
						},
						End: ast.Location{
							Line: int(1333),
							Column: int(90),
						},
					},
				},
				Kind: ast.LiteralStringKind(1),
			},
			EqFodder: nil,
			Variable: "base64_table",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1333),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(1333),
					Column: int(90),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.Apply{
				Target: &ast.Index{
					Target: &ast.Var{
						Id: "$std",
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"$std",
							},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
					},
					Index: &ast.LiteralString{
						Value: "$objectFlatMerge",
						BlockIndent: "",
						BlockTermIndent: "",
						NodeBase: ast.NodeBase{
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{},
							LocRange: ast.LocationRange{
								File: nil,
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
							},
						},
						Kind: ast.LiteralStringKind(1),
					},
					RightBracketFodder: nil,
					LeftBracketFodder: nil,
					Id: nil,
					NodeBase: ast.NodeBase{
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"$std",
						},
						LocRange: ast.LocationRange{
							File: nil,
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
						},
					},
				},
				FodderLeft: nil,
				Arguments: ast.Arguments{
					Positional: []ast.CommaSeparatedExpr{
						ast.CommaSeparatedExpr{
							Expr: &ast.Apply{
								Target: &ast.Index{
									Target: &ast.Var{
										Id: "$std",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"$std",
											},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
									},
									Index: &ast.LiteralString{
										Value: "flatMap",
										BlockIndent: "",
										BlockTermIndent: "",
										NodeBase: ast.NodeBase{
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{},
											LocRange: ast.LocationRange{
												File: nil,
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
											},
										},
										Kind: ast.LiteralStringKind(1),
									},
									RightBracketFodder: nil,
									LeftBracketFodder: nil,
									Id: nil,
									NodeBase: ast.NodeBase{
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"$std",
										},
										LocRange: ast.LocationRange{
											File: nil,
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
										},
									},
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Function{
												ParenLeftFodder: nil,
												ParenRightFodder: nil,
												Body: &ast.Array{
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.DesugaredObject{
																Asserts: ast.Nodes{},
																Fields: ast.DesugaredObjectFields{
																	ast.DesugaredObjectField{
																		Name: &ast.Index{
																			Target: &ast.Var{
																				Id: "base64_table",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18837,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1334),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1334),
																							Column: int(37),
																						},
																					},
																				},
																			},
																			Index: &ast.Var{
																				Id: "i",
																				NodeBase: ast.NodeBase{
																					Fodder: ast.Fodder{},
																					Ctx: p18840,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																					LocRange: ast.LocationRange{
																						File: p8,
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(1334),
																							Column: int(38),
																						},
																						End: ast.Location{
																							Line: int(1334),
																							Column: int(39),
																						},
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			LeftBracketFodder: ast.Fodder{},
																			Id: nil,
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18842,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1334),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1334),
																						Column: int(40),
																					},
																				},
																			},
																		},
																		Body: &ast.Var{
																			Id: "i",
																			NodeBase: ast.NodeBase{
																				Fodder: ast.Fodder{},
																				Ctx: p18845,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																				LocRange: ast.LocationRange{
																					File: p8,
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(1334),
																						Column: int(43),
																					},
																					End: ast.Location{
																						Line: int(1334),
																						Column: int(44),
																					},
																				},
																			},
																		},
																		LocRange: ast.LocationRange{
																			File: p8,
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(1334),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1334),
																				Column: int(44),
																			},
																		},
																		Hide: ast.ObjectFieldHide(1),
																		PlusSuper: false,
																	},
																},
																Locals: ast.LocalBinds{},
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18847,
																	FreeVars: ast.Identifiers{
																		"base64_table",
																		"i",
																	},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1334),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1334),
																			Column: int(72),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													CloseFodder: nil,
													NodeBase: ast.NodeBase{
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"base64_table",
															"i",
														},
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
													TrailingComma: false,
												},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: nil,
														Name: "i",
														CommaFodder: nil,
														EqFodder: nil,
														DefaultArg: nil,
														LocRange: ast.LocationRange{
															File: nil,
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
														},
													},
												},
												NodeBase: ast.NodeBase{
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"base64_table",
													},
													LocRange: ast.LocationRange{
														File: nil,
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
													},
												},
												TrailingComma: false,
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												Target: &ast.Index{
													Target: &ast.Var{
														Id: "std",
														NodeBase: ast.NodeBase{
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
															LocRange: ast.LocationRange{
																File: p8,
																FileName: "",
																Begin: ast.Location{
																	Line: int(1334),
																	Column: int(54),
																},
																End: ast.Location{
																	Line: int(1334),
																	Column: int(57),
																},
															},
														},
													},
													Index: &ast.LiteralString{
														Value: "range",
														BlockIndent: "",
														BlockTermIndent: "",
														NodeBase: ast.NodeBase{
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{},
															LocRange: ast.LocationRange{
																File: nil,
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
															},
														},
														Kind: ast.LiteralStringKind(1),
													},
													RightBracketFodder: ast.Fodder{},
													LeftBracketFodder: ast.Fodder{},
													Id: nil,
													NodeBase: ast.NodeBase{
														Fodder: ast.Fodder{},
														Ctx: p18857,
														FreeVars: ast.Identifiers{
															"std",
														},
														LocRange: ast.LocationRange{
															File: p8,
															FileName: "",
															Begin: ast.Location{
																Line: int(1334),
																Column: int(54),
															},
															End: ast.Location{
																Line: int(1334),
																Column: int(63),
															},
														},
													},
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "0",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18861,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1334),
																			Column: int(64),
																		},
																		End: ast.Location{
																			Line: int(1334),
																			Column: int(65),
																		},
																	},
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																OriginalString: "63",
																NodeBase: ast.NodeBase{
																	Fodder: ast.Fodder{},
																	Ctx: p18863,
																	FreeVars: ast.Identifiers{},
																	LocRange: ast.LocationRange{
																		File: p8,
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(1334),
																			Column: int(67),
																		},
																		End: ast.Location{
																			Line: int(1334),
																			Column: int(69),
																		},
																	},
																},
															},
															CommaFodder: nil,
														},
													},
													Named: []ast.NamedArgument{},
												},
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
												NodeBase: ast.NodeBase{
													Fodder: ast.Fodder{},
													Ctx: p18864,
													FreeVars: ast.Identifiers{
														"std",
													},
													LocRange: ast.LocationRange{
														File: p8,
														FileName: "",
														Begin: ast.Location{
															Line: int(1334),
															Column: int(54),
														},
														End: ast.Location{
															Line: int(1334),
															Column: int(70),
														},
													},
												},
												TrailingComma: false,
												TailStrict: false,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								FodderRight: nil,
								TailStrictFodder: nil,
								NodeBase: ast.NodeBase{
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$std",
										"base64_table",
										"std",
									},
									LocRange: ast.LocationRange{
										File: p8,
										FileName: "",
										Begin: ast.Location{
											Line: int(1334),
											Column: int(22),
										},
										End: ast.Location{
											Line: int(1334),
											Column: int(72),
										},
									},
								},
								TrailingComma: false,
								TailStrict: false,
							},
							CommaFodder: nil,
						},
					},
					Named: nil,
				},
				FodderRight: nil,
				TailStrictFodder: nil,
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{
						"$std",
						"base64_table",
						"std",
					},
					LocRange: ast.LocationRange{
						File: p8,
						FileName: "",
						Begin: ast.Location{
							Line: int(1334),
							Column: int(22),
						},
						End: ast.Location{
							Line: int(1334),
							Column: int(72),
						},
					},
				},
				TrailingComma: false,
				TailStrict: false,
			},
			EqFodder: nil,
			Variable: "base64_inv",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: p8,
				FileName: "",
				Begin: ast.Location{
					Line: int(1334),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(1334),
					Column: int(72),
				},
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Body: &ast.Self{
				NodeBase: ast.NodeBase{
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{},
					LocRange: ast.LocationRange{
						File: nil,
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
					},
				},
			},
			EqFodder: nil,
			Variable: "$",
			CloseFodder: nil,
			Fun: nil,
			LocRange: ast.LocationRange{
				File: nil,
				FileName: "",
				Begin: ast.Location{
					Line: int(0),
					Column: int(0),
				},
				End: ast.Location{
					Line: int(0),
					Column: int(0),
				},
			},
		},
	},
	NodeBase: ast.NodeBase{
		Fodder: ast.Fodder{
			ast.FodderElement{
				Comment: []string{},
				Kind: ast.FodderKind(0),
				Blanks: int(0),
				Indent: int(0),
			},
			ast.FodderElement{
				Comment: []string{
					"/*",
					"Copyright 2015 Google Inc. All rights reserved.",
					"",
					"Licensed under the Apache License, Version 2.0 (the \"License\");",
					"you may not use this file except in compliance with the License.",
					"You may obtain a copy of the License at",
					"",
					"    http://www.apache.org/licenses/LICENSE-2.0",
					"",
					"Unless required by applicable law or agreed to in writing, software",
					"distributed under the License is distributed on an \"AS IS\" BASIS,",
					"WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.",
					"See the License for the specific language governing permissions and",
					"limitations under the License.",
					"*/",
				},
				Kind: ast.FodderKind(2),
				Blanks: int(1),
				Indent: int(0),
			},
			ast.FodderElement{
				Comment: []string{
					"/* This is the Jsonnet standard library, at least the parts of it that are written in Jsonnet.",
					" *",
					" * There are some native methods as well, which are defined in the interpreter and added to this",
					" * file.  It is never necessary to import std.jsonnet, it is embedded into the interpreter at",
					" * compile-time and automatically imported into all other Jsonnet programs.",
					" */",
				},
				Kind: ast.FodderKind(2),
				Blanks: int(0),
				Indent: int(0),
			},
		},
		Ctx: p18872,
		FreeVars: ast.Identifiers{
			"$std",
		},
		LocRange: ast.LocationRange{
			File: p8,
			FileName: "",
			Begin: ast.Location{
				Line: int(23),
				Column: int(1),
			},
			End: ast.Location{
				Line: int(1713),
				Column: int(2),
			},
		},
	},
}
